/*	space.c	6.1	83/07/29	*/

#include "../sys/sema.h"
#include "../sys/cpu.h"
#include "../sys/param.h"
#include "../sys/immu.h"
#include "../sys/user.h"
#include "../sys/region.h"
#include "../sys/vnode.h"
#include "../sys/pfdat.h"
#include "../sys/conf.h"
#include "../sys/systm.h"
#include "../sys/proc.h"
#include "../sys/file.h"
#include "../sys/vfs.h"
#include "../sys/callout.h"
#include "../sys/mbuf.h"
#include "../sys/kernel.h"
#include "../sys/bsem.h"
#include "../sys/msg.h"
#include "../sys/sysinfo.h"
#include "../sys/buf.h"
#include "../sys/elog.h"
#include "../sys/iobuf.h"
#include "../sys/tty.h"
#include "../sys/map.h"
#include "../sys/x25conf.h"
#include "../sys/var.h"
#include "../sys/swap.h"
#include "../sys/getpages.h"
#include "../sys/tuneable.h"
#include "../sys/fbs.h"
#include "../sys/pm.h"
#include "../sys/ta.h"

#include "config.h"
#include "../cf/ipcconfig.h"

long	user_psl_template = 0;

int	rt_quantum = 6;		/* clock ticks in quantum - RT processes */
int	non_rt_quantum = 6;	/* clock ticks in quantum - non RT processes */

int	rt_disk_scheduling = 0;	/* controls disk scheduling priorities */

int	ignore_strays	= 1 ;	/* if true ignore stray interrupts. */

/*
 * System parameter formulae.
 * Change history and documentation :
 */

int     nvba = NVBA ;

int	hz = HZ;
int	nbpp = NBPP;
int	tick = 1000000 / HZ;
int	tickadj = 240000 / (60 * HZ);		/* can adjust 240ms in 60s */
char	*ptrtozero = 0;
struct	timezone tz = { TIMEZONE, DST };

#ifdef INET
#	define	NETSLOP	20			/* for all the lousy servers*/
#	ifdef GATEWAY
		int Gatewayopt = 1;
#	else
		int Gatewayopt = 0;
#	endif GATEWAY
#else
#	define	NETSLOP	0
#endif

#ifndef MBUFS
#define	MBUFS	512
#endif
int	nmbclusters = (((MBUFS * 1024)+(NBPP - 1))/NBPP);
#define	NMBCLUSTERS   (((MBUFS * 1024)+(NBPP - 1))/NBPP)
char	mclrefcnt[NMBCLUSTERS * (NBPP/MCLBYTES) + 1];
struct map	mbmap[NMBCLUSTERS/4];

/*
 * Boost values for calls to 'slumber'.  All values must be >= 0.  Note
 * that resshrboost >= resboost >= resexcboost.  To configure a system
 * with no priority adjustments, set all of the following boost values
 * to 0.
 */

int	noboost       =  0;	/* no boost */
int	resboost      =  2;	/* resource wait boost */
int	selboost      =  6;	/* select wait boost */
int	resexcboost   =  2;	/* resource wait for exclusive access boost */
int	proccomboost  =  2;	/* internal process communication boost */
int	delayboost    =  2;	/* process dealy boost */
int	iocomboost    =  2;	/* I/O completion boost */
int	devcntrlboost =  2;	/* device control function completion boost */
int	rpcboost      =  2;	/* RPC communications boost */
int	bscboost      =  2;	/* BSC device operations boost */
int	sxtboost      =  2;	/* SXT device operations boost */
int	resshrboost   =  3;	/* resource wait for shared access boost */
int	ttoboost      =  4;	/* TTY device output boost */
int	ttiboost      =  6;	/* TTY device input boost */
int	exitboost     =  6;	/* process exit boost */

/*
 *  These values define the range of process priorities on the system.
 *
 *  By default the top 5 priorities are real-time processes, whose
 *  priorities are never touched by the scheduler.
 *
 *  System processes are started at the bottom of the real-time processes.
 * 
 *  Processes sleeping in the kernel processes occupy the next 7 priorities.
 *  The top of the kernel priorites is defined by the largest boost value.
 *
 *  All priorities below the sleeping kernel processes may have their
 *  priorities temporarily adjusted by the boost values.  They will also
 *  have their priorities adjusted downward for using a lot of cpu.
 *
 *  By default all processes will start out at 20 (NZERO).  Priorities
 *  range from 0 to 31, with low numbers being scheduled first.
 */
int	real_time_pri		=  4;	/* real time quantums started here */
int	sys_proc_pri		=  4;	/* system processes started here */
int	kernel_base_pri		= 11;	/* base for kernel sleeps */
int	min_user_boost_pri 	= 12;	/* don't boost p_usrpri if < this */
int	min_whomp_pri		= 12;	/* don't float if p_prior < this */

/*
 * These are initialized at bootstrap time to values dependent on
 * memory size.  Bufpages may be set by the administrator by defining
 * BUFFERS in the configuration file.  (BUFFERS defines the size of the
 * system buffer cache in K-bytes.)
 */

#if BUFFERS > 0
int	bufpages = BUFFERS * 1024 / NBPP;
#else
#define BUFFERS 0
int	bufpages = BUFFERS;
#endif

#if PROCS > 0
int	nproc = PROCS;
#else
#define PROCS (20 + 8 * MAXUSERS)
int	nproc = PROCS;
#endif

#if INODES > 0
int	ninode = INODES;
#else
#define INODES (PROCS + 16 + MAXUSERS) + 32
int	ninode = INODES;
#endif
int	ncsize = INODES;

#if FILES > 0
int	nfile = FILES ;
#else
#define FILES 16 * (PROCS + 16 + MAXUSERS) / 10 + 32 + 2 * NETSLOP
int	nfile = FILES;
#endif

#if CALLS > 0
int	ncallout = CALLS;
#else
#define CALLS 16 + PROCS
int	ncallout = CALLS;
#endif

#if CLISTS == 0
#define CLISTS 100 + 16 * MAXUSERS
#endif
/* Allocate clist here to ensure a physical address which is low */
/* enough that the VIOC and HVIOC can DMA into them. That is, */
/* less than 15Mb. */
struct cblock Cfree[CLISTS];
struct cblock *cfree = Cfree;

/* ??? Are these reasonable defaults LEN??? */
#if MAXPROC > 0
int maxproc = MAXPROC;
#else
#define MAXPROC 25
int maxproc = MAXPROC;
#endif

#ifndef NPBUF
#define NPBUF 20
#endif

#if ! ( NREGION > 0 )
#define NREGION (PROCS * 2)
#endif

#if ! ( VHNDFRAC > 0 )
#define VHNDFRAC 16
#endif

#if ! ( IPGSLO > 0 )
#define IPGSLO 12
#endif

#if ! ( IPGSHI > 0 )
#define IPGSHI 16
#endif

#if ! ( GPGSLO > 0 )
#define GPGSLO 0
#endif

#if ! ( GPGSHI > 0 )
#define GPGSHI 0
#endif

#if ! (GPGSMAX > 0 )
#define GPGSMAX 0
#endif

#ifndef GPGSMSK
#define GPGSMSK GETPGSMSK
#endif

#if ! ( VHANDR > 0 )
#define VHANDR 5
#endif

#if ! ( VHANDL > 0 )
#define VHANDL 0
#endif

#if ! ( MAXPMEM > 0 )
#define MAXPMEM 0
#endif

#if ! ( MAXUMEM > 0 )
#define MAXUMEM (3*32*1024*1024)/NBPP
#endif

#if ! ( MINARMEM > 0 )
#define MINARMEM 25
#endif

#if ! ( MINASMEM > 0 )
#define MINASMEM 25
#endif

/*
** page locking limits
*/
int	genulockp = 10;
int	sulockp = 20;
int	u_limit, s_limit;	/* limits for general user + suser resp. */
int	plocked;		/* # pages locked */

/* Maximum physical memory to use */
int     maxpmem = MAXPMEM;

#ifdef QUOTA
#include "../ufs/quota.h"
int	ndquot = (MAXUSERS*NMOUNT)/4 + PROCS;
#else
int	ndquot = 0;
qtinit(){}
chkdq(){ return(0);}
chkiq(){ return(0);}
dqrele(){}
getinoquota(){ return(0);}
closedq(){}
#endif

#ifdef	hcx
#ifndef	FPE
#include "../fpe/fp_in_krnl.h"
fpemulate(hfsreg,acc_most,acc_least,dbl,op_most,op_least,opcode,pc,psl)
{	opcode = FPII_EXC;	}
#endif
#endif	hcx

/* initialize SystemV named-pipe (and pipe()) information structure */

#include "../specfs/fifo.h"
struct fifoinfo fifoinfo = {
      FIFOBUF,
      FIFOMAX,
      FIFOBSZ,
      FIFOMNB
};

/*
 * VFS -- virtual file systems
 */
/*	@(#)vfs_conf.c 1.1 86/02/03 SMI	*/
/*	@(#)vfs_conf.c	2.1 86/04/15 NFSSRC */

extern	struct vfsops ufs_vfsops;	/* XXX Should be ifdefed */

#ifdef NFS
extern	struct vfsops nfs_vfsops;
#endif

#ifdef PCFS
extern	struct vfsops pcfs_vfsops;
#endif

struct vfsops *vfssw[] = {
	&ufs_vfsops,		/* 0 = MOUNT_UFS */
#ifdef NFS
	&nfs_vfsops,		/* 1 = MOUNT_NFS */
#else
	(struct vfsops *)0,
#endif
#ifdef PCFS
	&pcfs_vfsops,		/* 2 = MOUNT_PC */
#else
	(struct vfsops *)0,
#endif
};

/*
* Asynchronous I/O
*/

int	max_areq = 100;

/*
**	AT&T V.3 VM
*/

reg_t	region[NREGION];	/* region entry table			*/
int	rckey[NREGION*MAX_CPU];	/* region code key array		*/

/*
 * Number of entries in sptmaps
 */
#ifdef SPTMAP
int sptmap_size = SPTMAP;
#else
int sptmap_size = 0;	/* will be calculated during startup */
#endif SPTMAP

/*	Each process has 3 pregions (text, data, and stack) plus
 *	enough for the maximum number of shared memory segments.
 *	We also need one extra null pregion to indicate the end
 *	of the list.  The maximum number of shared memory segments
 *	will be added in reginit.
 */

int	pregpp = 3 + 1;

struct pfdat	phead;		/* Head of free page list.	*/
lock_t		phead_lock;	/* spin lock for free list	*/
struct pfdat	*pfdat;		/* Page frame database.		*/
struct pfdat	*xpfdat;	/* Extend pfdat (by kpbase)	*/
int		npfdat;		/* Number of pfdat entries	*/
sema_t		*pfdat_sema;	/* Page frame semaphores	*/
struct phash	*phash;		/* Page hash access to pfdat.	*/
sema_t		*phash_sema;	/* Page hash semaphores		*/
int		phashmask;	/* Page hash mask.		*/

struct pfdat	ptfree;		/* Head of page table free list */
sema_t		ptfree_sema;	/* Page table semaphore		*/
int		nptalloced;	/* Number of alloc page tables.	*/
int		nptfree;	/* Number of free page tables.	*/


swpt_t	swaptab[MSFILES];
int	nextswap;

/*
** Disable Swap Reservation
** Set the following variable to 'TRUE' to disable swap area reservation.
** This allows programs up to 1 gigabyte in size to begin execution without the
** program loader checking memory and swap area availability.  The program will
** execute until memory and swap area are exhausted. The process will then enter
** SXBRK state waiting on memory resources.  This can result in all processes
** in the system entering this state once all memory and swap area are used 
** and the system appearing hung.  With swap area reservation enabled, programs
** are not started unless enough physical memory and swap area are available
** to meet the initial memory requirements of the process.
**
** This variable should not be changed on a running system.  The modification
** should be made to this source file and the kernel relinked.
*/
int	dis_swap_res = FALSE;


#define NOTUSED -1
struct var v = {
	BUFFERS,
	CALLS,
	INODES,
	(char *)NOTUSED,
	FILES,
	(char *)NOTUSED,
	NOTUSED,
	(char *)NOTUSED,
	PROCS,
	(char *)NOTUSED,
	NREGION,
	CLISTS,
	MAXPROC,
	NOTUSED,
	NOTUSED,
	NPBUF,
	MSFILES,
	VHNDFRAC,
	MAXPMEM,
	NOTUSED,
	NOFILE,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
	NOTUSED,
};

struct tune tune = {
	GPGSLO,
	GPGSHI,
	GPGSMAX,
	GPGSMSK,
	VHANDR,
	VHANDL,
	MAXUMEM,
	-1,
	MINARMEM,
	MINASMEM,
	IPGSLO,
	IPGSHI,
} ;

/*
 * Maps from board CPU IDs to physical CPU IDs.  These maps
 * assume that board CPU IDs are contiguous and start at 0.
 * The physical CPU ID for GCX refers to the slot number the
 * board is plugged into.  Although on both architectures the board id 
 * and physical id are the same, this structure allows for future
 * changes and architectures.
 */

#ifdef gcx
int	btop_cpuid [MAX_BOARD_CPUID + 1] = { 0, 1, 2, 3, 4, 5, 6, 7 };
#endif gcx
#ifdef hcx
#if	defined(TITAN3) || defined(ME7611)
int	btop_cpuid [MAX_BOARD_CPUID + 1] = { 0, 1 };
#else
int	btop_cpuid [MAX_BOARD_CPUID + 1] = { 0 };
#endif TITAN3
#endif hcx


/*
* Shared memory
*/

struct shmid_ds shmem[SHMMNI];
sema_t shmsem[SHMMNI];
struct	shminfo shminfo = {
	SHMMAX,
	SHMMIN,
	SHMMNI,
	SHMSEG,
	SHMALL
};


/*
 * Semaphores
 */

struct semid_ds	sema[SEMMNI];
sema_t		semsem[SEMMNI];
struct sem	sem[SEMMNS];
struct map	semmap[SEMMAP];
struct sem_undo	* sem_undo [PROCS] ;
#define	SEMUSZ	(sizeof(struct sem_undo)+sizeof(struct undo)*SEMUME)
int	semu[((SEMUSZ*SEMMNU)+NBPW-1)/NBPW];
union {
	short		semvals[SEMMSL];
	struct semid_ds	ds;
	struct sembuf	semops[SEMOPM];
}	semtmp;

struct	seminfo seminfo = {
	SEMMAP,
	SEMMNI,
	SEMMNS,
	SEMMNU,
	SEMMSL,
	SEMOPM,
	SEMUME,
	SEMUSZ,
	SEMVMX,
	SEMAEM
};


/*
 * Binary Semaphores
 */
struct bsemid_ds bsemlocal_tbl[BSEMMAP];
struct bsemsem bsemsem[BSEMMAP];
struct bseminfo	bseminfo = {
	BSEMMAP,
	BSEMGBL,
	BSEMMAP + BSEMGBL
};


/*
 * Messages
 */

int msgmappte = MSGMAPPTE;
struct map	msgmap[MSGMAP];
struct msqid_ds	msgque[MSGMNI];
struct msgsem	msgsem[MSGMNI];
struct msg	msgh[MSGTQL];
struct msginfo	msginfo = {
	MSGMAP,
	MSGMAX,
	MSGMNB,
	MSGMNI,
	MSGSSZ,
	MSGTQL,
	MSGSEG
};



/*
 *	Frequency Based Scheduler
 */

#ifdef FBS

#if FBSMNI > 12
#	define FBSMNI 12	/* system imposed limit of 12 FBSs */
#endif


struct	fbs_ds     fbs [FBSMNI];
sema_t		   fbssem [FBSMNI];
struct	fbssch_ds  fbssch [FBSMNI * FBSCYCFBS];
cyctbl_t           fbscyc [FBSCYCSYS * (FBSPROCCYC+1)];

struct	fbsproc_ds fbspid [FBSPROCSYS];

struct	fbsfl_ds	freefbsproc;	/* FBS process block free list hdr */
struct	fbsfl_ds	freecyctbl;	/* minor cycle table free list hdr */

struct	fbsinfo    fbsinfo = {
			FBSMNI,
			FBSCYCSYS,
			FBSCYCFBS,
			FBSPROCSYS,
			FBSPROCCYC
		  };

/* Performance Monitor */


struct 	pm_block  	pmblks[FBSPROCSYS];
struct 	proc		*fbsunsched[FBSMNI * (FBSPROCCYC+1)];
int			idlespare_sched = 0; /* Idle or spare scheduled */ 
int			pminfo_cpumask = 0; /* cpu mask */
int			perfmon_on = 0;	 /* PM on for any process except idle */
int			idlespare_on = 0; /* PM on for idle and/or spare */

#else   

struct	fbsproc_ds fbspid[1];
struct	fbsinfo    fbsinfo = {
			0, 0, 0, 0, 0 };

fbsinit() {}
fbsexit() {}
#ifdef	gcx
fbswait()	
{
	return(EINVAL);
}
#endif	gcx


#ifdef RT
int			perfmon_on = 0;	 /* PM on for any process except idle */
int			idlespare_on = 0; /* PM on for idle and/or spare */
#endif

#endif

/*
 * Dummy spare processes.      
 */
struct	proc		spare[MAX_CPU];

/*
 *
 * There are two options:
 * 	NFS: locking is done by kern_locktl (../os/sys_flock.c)
 *              Can't use NFS in this mode
 *
 *      !NFS: locking is done by ufs_lockctl
 *              All locking is passed up to and done by user level process.
 *		In this mode, sys_flock.o isn't linked into the kernel.
 *
 */
#include "../sys/flock.h"
#define	FLCKREC		100	/* # lock structures */

struct flckinfo flckinfo = {
	FLCKREC,
	0,
	0,
	0
};

struct filock flox[FLCKREC];		/* lock structures */

/*
**	Performance monitoring structures
*/
struct sysinfo sysinfo;
struct	syswait	syswait;
struct	syserr	syserr;
struct  minfo minfo;


/*	The following describe the physical memory configuration.
**
**		maxclick - The largest physical click number.
**			   ctob(maxclick) is the largest physical
**			   address configured plus 1.
**
**		physmem	 - The amount of physical memory configured
**			   in clicks.  ctob(maxclick) is the amount
**			   of physical memory in bytes.
**		kpbase	 - The physical address of the start of
**			   the allocatable memory area.  That is,
**			   after all kernel tables are allocated.
**			   Pfdat table entries are allocated for
**			   physical memory starting at this address.
**			   It is always on a page boundary.
*/
int	maxclick;
int	physmem;
int	kpbase;





/* ulimit value - maximum file size (in 512 byte blocks)
 * that can be written by non-super-users.
 *
 * Do not set this to a value larger than 0x01fffff
 */
#ifndef CDLIMIT
#define CDLIMIT	(0x1fffff)
#endif

daddr_t	 cdlimit = CDLIMIT;

/*
 * These have to be allocated somewhere; allocating
 * them here forces loader errors if this file is omitted.
 */
struct	proc *proc, *procNPROC ;
struct	inode *inode, *inodeNINODE;
struct	file *file, *fileNFILE;
struct 	callout *callout;
struct	buf *buf, *swbuf;
char	*buffers;
struct	dquot *dquot, *dquotNDQUOT;

/*
 * disks - maximum 32 per system
 *        vddc.c = EGL, FSD and SMD
 *        hdc.c  = HDC
 */
#include "../sys/dsk.h"
#include "../sys/vddc.h"
#include "../sys/hdc.h"
#include "../sys/vbavar.h"

#if NDSK > 0
#if NDSK > 32
	int ndsk = "error: too many disk drives configured.";
#endif NDSK > 32
	int			ndsk = NDSK;
	struct  vba_device	*vddinfo[NDSK];
	struct	iotime		vdstat[NDSK];
	struct	iobuf		vdtab[NDSK];
#endif NDSK > 0

#if NHD > 0
	hdc_ctlr_type		hdc_ctlrs[NHD];
	struct	vba_ctlr	*hdminfo[NHD];
	hdc_unit_type		hdc_units[NDSK];
#endif NHD > 0


#if NVD > 0
	struct	vba_ctlr	*vdminfo[NVD];
	vdunit_tab		vdunit_info[NDSK];

	vdctlr_tab	vdctlr_info[] = {
		 {UNKNOWN, (char *)UNKNOWN, (char *)UNKNOWN, 
			0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 1
		,{UNKNOWN, (char *)UNKNOWN, (char *)UNKNOWN, 
			0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 2
		,{UNKNOWN, (char *)UNKNOWN, (char *)UNKNOWN, 
			0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 3
		,{UNKNOWN, (char *)UNKNOWN, (char *)UNKNOWN, 
			0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#endif NVD > 3
#endif NVD > 2
#endif NVD > 1
	};
#endif NVD > 0

/*
 * cy.c -- magnetic tape driver.
 */
#include "../ml/vmparam.h"
#include "../sys/cipher.h"

#if NCY > 0
#if NCY > 8
	int ncy = "error: too many cipher tape drives";
#endif NCY > 8
	int			ncy = NCY;
	struct	vba_device	*cydinfo[NCY];
	struct	iotime		cystat[NCY];
	struct 	iobuf		cytab[NCY];
	unit_tab		unit_info[NCY];
#endif NCY > 0

#if NTM > 0
#if NTM > MAXCONTROLLERS
	int ntm = "error: too many cipher tape controllers";
#endif NTM > MAXCONTROLLERS
	int			ntm = NTM;
	struct vba_ctlr		*cyminfo[NTM];
	extern 	int		cy_normal_path();

	ctlr_tab	ctlr_info[] = {
		{NULL, NULL, cy_normal_path},
#if NTM > 1
		{NULL, NULL, cy_normal_path}
#endif NTM > 1
	};
#endif NTM > 0


/*
 * xy.c -- XYLOGICS magnetic tape driver.
 */

#if (NXY > 0) & (NXYD > 0)

#include "../sys/mtio.h"
#include "../sys/xymacro.h"

#define XYNUMIOPB 8*NXYD
	int				xynumiopb = XYNUMIOPB ;
	struct	vba_device		*xy_vba_device[NXYD] ;
	struct	iotime			xystat[NXYD];
	struct 	iobuf			xytab[NXYD];
	struct	unit_tab		xy_unit_info[NXYD];
	struct	iopb			xy_free_iopb_list_arr[XYNUMIOPB] ;
/* 
**	#defineing NOSG makes driver use physically continuous buffer, xybuffer,
**	and avoid the use of scatter/gather.
*/
#ifndef NOSG
	struct	sg_block		xy_free_sg_list_arr[XYNUMIOPB] ;
#endif
	struct	buf			xy_free_buf_list_arr[XYNUMIOPB] ;
	struct  ctlr_parms_iopb		xy_ctlr_setup_struct[NXY] ;
	struct	drive_parms_iopb	xy_drive_setup_struct[NXYD] ;
#if NXY > XYMAXCONTROLLERS
	int nxyctlr = "error: too many xylogics tape controllers";
#endif NXY > XYMAXCONTROLLERS
#if NXYD > XYMAXDRIVES
	int nxyd = "error: too many xylogics tape drives";
#endif NXYD > XYMAXDRIVES
	int			nxyctlr = NXY ;
	int			nxyd = NXYD ;
	struct vba_ctlr		xy_vba_ctlr[NXY] ;
	struct ctlr_tab		xy_ctlr_info[NXY] ;
#endif (NXY > 0) & (NXYD > 0)






/*
 * tty_pty.c -- pseudo terminal driver.
 */
#include "../sys/tty_pty.h"
#if NPTY > 0
	int			npty = NPTY;
	struct	tty		pt_tty[NPTY];
 	struct	semdrivs	ptsema[NPTY];
	struct	pt_ioctl	pt_ioctl[NPTY];
#endif


/*
 * vcc.c, vcx.c -- terminal driver.
 */
#include "../sys/vioc.h"
#if NVX > 0
#if NVX > 16
	int nvx = "error: too many viocs";
#endif NVX > 16
#	define	NVXPORTS	16*NVX
	caddr_t			VIOCBAS[NVX];
	struct	vcmds		v_cmds[NVX] ;
	struct	vcx		vcx[NVX] ;
	struct	vba_device 	*vxinfo[NVX];
	struct	vba_ctlr   	*vxminfo[NVX];
	struct	tty		vx_tty[NVXPORTS];
    struct  semdrivs vx_tty_sema[NVXPORTS];
    struct  vioc_intq vx_tty_intq[NVXPORTS];
	int			nvxports = NVXPORTS;
	int			nvx = NVX;
	char vxtype[NVX];	/* 0: viox-x/vioc-b; 1: vioc-bop */
	char vxbopno[NVX];	/* BOP board no. if indicated by vxtype[] */

	/*
	 * Used to try to handle break interrupts from the vioc sanely.
	 */
	lock_t		vx_break_lock;	/* protects updates of vx_break[] */
	int			vx_break[NVX];
#else
	init_vx_tty_semas(){}
#endif


/*
 * sxt.c --- shell layers driver.
 */
#include "../sys/sxt.h"
#if NSXT > 0
#if NSXT > MAXLINK
	int sxt_cnt = "error: too many sxts";
#endif NSXT > MAXLINK
	int	sxt_cnt = NSXT;
	char 	sxtbuf[NSXT*LINKSIZE];
#endif


/*
 * rtc.c -- real-time clock driver.
 */
#if NRTC > 0
#include "../sys/rtc.h"
	
	int			nrtc = NRTC;
	int			nclock = NRTC * CLKS_PER_CTLR;
	
	struct vba_ctlr		*rtc_vba_ctlr[NRTC];
	struct vba_device	*rtc_vba_device[NRTC];
	struct rtclock		rtc_tab_info[NRTC * CLKS_PER_CTLR];
	struct semdrivs 	rtc_sema[NRTC * CLKS_PER_CTLR];
#else
	init_rtc_semas(){}
#endif	
	
/*
 * hsd.c -- HSD driver.
 */
#if NHSD > 0
#include "../sys/hsd.h"

	int                     nhsd = NHSD;

	struct vba_ctlr         *hsd_vba_ctlr[NHSD];
	struct vba_device       *hsd_vba_device[NHSD];
	struct hsd_ctlr         hsd_ctlr[NHSD];

#endif


/*----------------------------------------------------
  Configure Network Families and Protocols
  --------------------------------------------------*/
/*
 * ../net/af.c
 *
 * Address family support routines
 */

#include "../sys/socket.h"
#include "../net/af.h"
#include "../sys/domain.h"
#include "../net/if.h"

extern	int	null_hash(),
		null_netmatch();
#define	AFNULL	{ null_hash,	null_netmatch }

char	hostname[MAXHOSTNAMELEN] = NODE ;
int	hostnamelen = sizeof(NODE) - 1 ;

/*--------------------- Internet Family and Protocol --------------*/
#ifdef INET

	/*
	 * (was from 4.2BSD ../net/if.h)
	 */
	struct	ifqueue	ipintrq;		/* ip packet input queue */
#else
	inet_hash(sin, hp)
		register struct sockaddr_in *sin;
		struct afhash *hp;
	{
		null_hash( sin, hp ) ;
	}

	inet_netmatch(sin1, sin2)
		struct sockaddr_in *sin1, *sin2;
	{
		null_netmatch( sin1, sin2 ) ;
	}

	arpioctl(cmd, data)
		int cmd;
		caddr_t data;
	{
		return (EPFNOSUPPORT);
	}

	inetADDDOMAIN() {}
	ipintr(){}
#endif

/*---------------- PUP Network Family and Protocols -------------------*/
#	define	AFPUP	AFNULL
	pupADDDOMAIN() {}

/*------------------------ IMP Network Family and Protocols --------------*/
#if NIMP > 0

	/*
	 * (was from 4.2BSD ../sys/uipc_domain.c)
	 *
	 * Define IMP Network Family.
	 */

	nimpADDDOMAIN()
	{
		extern struct domain nimpdomain;

		nimpdomain.dom_next = domains;
		domains = &nimpdomain;
	}

	/*
	 * (was from ../netinet/in_proto.c).
	 *
	 * Define IMP protocol.
	 *
	 */

	int	rip_input(),rip_output();
	extern	int raw_usrreq();
	/*
	 * IMP protocol family: raw interface.
	 * Using the raw interface entry to get the timer routine
	 * in is a kludge.
	 */
	int	rimp_output(), hostslowtimo();

	struct protosw impsw[] = {
	{ SOCK_RAW,	PF_IMPLINK,	0,		PR_ATOMIC|PR_ADDR,
	  0,		rimp_output,	0,		0,
	  raw_usrreq,
	  0,		0,		hostslowtimo,	0,
	},
	};

	struct domain impdomain =
	    { AF_IMPLINK, "imp", impsw, 
		    &impsw[sizeof (impsw)/sizeof(impsw[0])] };

#else
	nimpADDDOMAIN() {}
#endif



/*   ===== Versatec Controller ===========
*/
#if (NVP > 0) 

#include "../sys/vp.h"

	int   	nvp = NVP;

	struct	vba_device		*vp_vba_device[NVP] ;
	struct	vba_ctlr			*vp_vba_ctlr[NVP] ;
	vp_aux_type					vp_aux[NVP] ;

	extern unsigned long Xvpintr0, Xvpintr1;
 	
   unsigned long vp_intvect_proc[] = {
	            	(unsigned long)& Xvpintr0,  (unsigned long)& Xvpintr1 };

#endif (NVP > 0)
/** ======SCSI bus controller and devices======
*/

#if NSC > 0

#include "../sys/scsi.h"


int		nsc = NSC;
sc_tab	sc_info[NSC*CTLR_PER_SCSI];
struct	vba_ctlr	*scminfo[NSC];
struct	vba_device	*scdinfo[NSC]; 

struct iotime		scstat[NSC*CTLR_PER_SCSI];
struct iobuf		sctab[NSC*CTLR_PER_SCSI];
	extern unsigned long Xscintr0;
   unsigned long sc_intvect_proc[] = {
	            	(unsigned long)& Xscintr0 };

/*
**	Quarter inch cartrige Tape controller.
*/


#if NSCD > 0

#include "../sys/qt.h"


int			nqt = NSCD;

qt_tab			qt_info[NSCD] = { {NULL, NULL} };
struct iotime		qtstat[NSCD];
struct iobuf		qttab[NSCD];

#endif NSCD > 0

#endif NSC > 0
/*	======End of SCSI devices====== */
/*
 * asynchronous I/O
 */
#if NAREQ > 0
int nareq = NAREQ;	/* user defined number of areq blocks */
#else
int nareq = 128;	/* default number of areq blocks */
#endif

#ifdef TA
int num_ta_devs = TA;		/* number of TA devices */
struct ta_dev ta_dev [TA];	/* async test device structure */
#else
int num_ta_devs = 0;		/* number of TA devices */
struct ta_dev ta_dev [1];	/* dummy async test device structure */
#endif

/*
 * HCCD -- Harris Controls and Composition Division
 *
 * Hooks for procedures used by HCCD.
 */

#ifndef HCCD
hccd_exit(){}
hccd_panic(){}
#endif
/*
 * ags.c - Scientific Calculations AGS driver.
 */
#if NAGS > 0
	extern int agsprobe(), agsslave(), agsattach();
	int 		nags = NAGS ;
	struct vba_device *	agsinfo[ NAGS ] ;
	struct vba_ctlr *	agsminfo[ NAGS ] ;
	struct vba_driver	agsdriver = {
		agsprobe, agsslave, agsattach, 0, 0, "AGS ",
		agsinfo, "AGS ",agsminfo, 0
		} ;
#endif  NAGS

#if NBIU > 0
#include "../sys/biu.h"
	int nbiu=NBIU;
	biu_tab	biu_info[NBIU];
	struct	vba_device	*biu_vba_device[NBIU] ;
	struct vba_ctlr		biu_vba_ctlr[NBIU] ;
	extern unsigned long Xbiuintr0;
        unsigned long biu_intvect_proc[] = {
	            	(unsigned long)& Xbiuintr0 };
	struct semdrivs 	biu_sema[NBIU];
#else
	init_biu_semas(){}
#endif NBIU > 0
