# @(#) cf.mk

SYS = 	HCX/UX
REL = 	2.3

include mkmacros

# The following 4 "macros" are passed here via the command line in :mkuts
# and, therefore, have the values below only at installation sites.
#
ROOT = /
ROOTUSR = /usr
INCLDIR = /usr/include
C2 = /lib/c2

INCRT = ..
INS = 	/etc/install
INSDIR = $(ROOT)
CFL= -O -X -I${INCLDIR} -Wl,-L${ROOT}/lib,-L${ROOT}/usr/lib
CFLAGS = -DKERNEL ${CFL}
NAME = 	unix$(VER)
TYPE = 	hcx-9
SRC = machine

FILES = \
	name.c \
	space.c \
	conf.c \
	linesw.c \
	ioconf.c \
	swapvmunix.c

.c.o:
	@echo "          "$*.c
	@$(CC) $(CFLAGS) -c $*.c

all: 	init  ../$(NAME)

link:   ../$(NAME)

init:
	cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" "C2=$(C2)" \
		"CFL=$(CFL)" ;
	@echo ================================= cf =============================

../$(NAME): \
 	scb.o ../locore.o vbavec.o sysmap.o name.o space.o \
	conf.o linesw.o ioconf.o swapvmunix.o ../lib[0-9] \
	../lib_*
	@echo "        "loading $(NAME)
	@rm -f $(NAME)
	@LIBS='../lib0 ../lib_os1 ../lib_os2 ../lib_os3 ../lib_io ../lib9'; \
	LIBS=$$LIBS' ../lib_net ../lib_ufs'; \
	if test "$(NFS)" -gt 0; then \
		LIBS=$$LIBS' ../lib_nfs ../lib_rpc'; \
	fi; \
	if test "$(INET)" -gt 0; then \
		LIBS=$$LIBS' ../lib_hcxif ../lib_netinet'; \
	fi; \
	if test "$(NVBSC)" -gt 0; then \
		LIBS=$$LIBS' ../lib7'; \
	fi; \
	${LD} -o ../$(NAME) ../cf/ld.file -e start -x \
		scb.o ../locore.o vbavec.o sysmap.o name.o \
		space.o conf.o linesw.o ioconf.o swapvmunix.o \
		$$LIBS
	@size ../$(NAME)

name.o: cf.mk name.c mkmacros
	@echo "          "name.c
	@${CC} -c ${CFLAGS} \
		-DSYS=\"`expr $(SYS) : '\(.\{1,8\}\)'`\" \
		-DNODE=\"`expr $(NODE) : '\(.\{1,8\}\)'`\" \
		-DREL=\"`expr $(REL) : '\(.\{1,8\}\)'`\" \
		-DVER=\"`expr $(VER) : '\(.\{1,8\}\)'`\" \
		-DMACH=\"`expr $(MACH) : '\(.\{1,8\}\)'`\" \
		../cf/name.c

scb.o: ../ml/scb.s
	@echo "          "../ml/scb.s
	@cat ../ml/scb.s > scb.c
	@${CC} -E -DLOCORE ${CFLAGS} scb.c > scb.i
	@${AS} -o scb.o scb.i
	@rm scb.c scb.i

vbavec.o: ../ml/vbavec.s \
	$(INCRT)/ml/psl.h \
	config.h
	@echo "          "../ml/vbavec.s
	@cat ../ml/assym.s ../ml/vbavec.s > vbavec.c
	@${CC} -E -DLOCORE ${CFLAGS} vbavec.c > vbavec.i
	@${AS} -o vbavec.o vbavec.i
	@rm vbavec.c vbavec.i

sysmap.o: ../ml/sysmap.s ../ml/assym.s\
	config.h \
	$(INCRT)/cf/ipcconfig.h \
    	$(INCRT)/ml/param.h \
    	$(INCRT)/ml/mtpr.h \
	$(INCRT)/ml/trap.h \
	$(INCRT)/ml/psl.h \
    	$(INCRT)/ml/pte.h \
	$(INCRT)/ml/cp.h \
	$(INCRT)/ml/mem.h
	@echo "          "../ml/sysmap.s
	@cat ../ml/assym.s ../ml/sysmap.s > sysmap.c
	@${CC} -E ${CFLAGS} -DLOCORE sysmap.c > sysmap.i
	@${AS} -o sysmap.o sysmap.i
	@rm sysmap.i sysmap.c


clean:
	cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" clean ;
	-rm -f *.o lintfiles
	-rm -f ../lib*

clobber:
	cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" clobber ;
	-rm -f ../$(NAME) ioconf.c swapvmunix.c config.h
	-rm -f *.o lintfiles
	-rm -f ../lib*

lint:
	@echo ' $(FILES)' | \
	sed -e 's/ [^(]*(\([^)]*\)\.o)/ ..\/cf\/\1.c/g' > lintfiles
	@cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" lint ;
	@lint -hbxn -I. -DKERNEL -DINET -DSYS=\"\" -DNODE=\"\" -DREL=\"\" \
	-DVER=\"\" -DMACH=\"\" ${COPTS} `cat lintfiles`

install:	all
	@echo "        "install
	@cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" "C2=$(C2)" install ;
	@chmod 755 ../$(NAME)
	@chgrp bin ../$(NAME)
	@chown bin ../$(NAME)
	@$(INS) -r -f $(INSDIR) "../$(NAME)"
	@mv $(INSDIR)/$(NAME) $(INSDIR)/unix

depend:
	@cd ../..; \
	make -f uts.mk "INS=$(INS)" "INSDIR=$(INSDDIR)" "INCRT=$(INCRT)" \
		"SYS=$(SYS)" "NODE=$(NODE)" "REL=$(REL)" "VER=$(VER)" \
		"MACH=$(MACH)" "TYPE=$(TYPE)" "SRC=$(SRC)" depend ;
	@grep '^#include' `echo ' $(FILES)' | sed -e 's/ [^(]*(\([^)]*\)\.o)/ \1.c/g'`|\
	grep -v '<' | \
	sed -e 's/:[^"]*"\([^"]*\)".*/ \1/' \
	    -e 's/\.c/.o/' \
	    -e 's,../,,' | sort | \
	awk ' { \
		if ($$1 != prev) { \
			print rec2; \
			prev = $$1; \
			rec2 =  "	$$(INCRT)/"$$2" "; \
			rec1 = "\n"$$1": \\"; \
			print rec1; \
		} else { \
			rec2 = rec2 "\\"; \
			print rec2; \
			rec2 =  "	$$(INCRT)/"$$2" "; \
		} \
	} END { print rec2 } ' > makedep
	@echo '/^# DO NOT DELETE THIS LINE/+1,$$d' >eddep
	@echo '$$r makedep' >>eddep
	@echo 'w' >>eddep
	@cp cf.mk cf.mk.bak
	@ed - cf.mk < eddep
	@rm eddep makedep

# DO NOT DELETE THIS LINE - make depend uses it


conf.o: \
	$(INCRT)/cf/config.h \
	$(INCRT)/sys/buf.h \
	$(INCRT)/sys/conf.h \
	$(INCRT)/sys/param.h \
	$(INCRT)/sys/systm.h \
	$(INCRT)/sys/tty.h 

linesw.o: \
	$(INCRT)/cf/config.h \
	$(INCRT)/sys/buf.h \
	$(INCRT)/sys/conf.h \
	$(INCRT)/sys/param.h \
	$(INCRT)/sys/systm.h \
	$(INCRT)/sys/tty.h 

name.o: \
	$(INCRT)/sys/utsname.h 

space.o: \
	$(INCRT)/cf/config.h \
	$(INCRT)/cf/ipcconfig.h \
	$(INCRT)/hcxif/if_en.h \
	$(INCRT)/hcxif/if_ex.h \
	$(INCRT)/ml/pte.h \
	$(INCRT)/ml/vmparam.h \
	$(INCRT)/net/af.h \
	$(INCRT)/net/if.h \
	$(INCRT)/netinet/if_ether.h \
	$(INCRT)/sys/bsc/bsc.h \
	$(INCRT)/sys/bsc/bscio.h \
	$(INCRT)/sys/buf.h \
	$(INCRT)/sys/callout.h \
	$(INCRT)/sys/cipher.h \
	$(INCRT)/sys/cmap.h \
	$(INCRT)/sys/conf.h \
	$(INCRT)/sys/domain.h \
	$(INCRT)/sys/dsk.h \
	$(INCRT)/sys/elog.h \
	$(INCRT)/sys/file.h \
	$(INCRT)/sys/flock.h \
	$(INCRT)/sys/iobuf.h \
	$(INCRT)/sys/kernel.h \
	$(INCRT)/sys/map.h \
	$(INCRT)/sys/mbuf.h \
	$(INCRT)/sys/msg.h \
	$(INCRT)/sys/param.h \
	$(INCRT)/sys/proc.h \
	$(INCRT)/sys/quota.h \
	$(INCRT)/sys/socket.h \
	$(INCRT)/sys/stdabv.h \
	$(INCRT)/sys/sxt.h \
	$(INCRT)/sys/sysinfo.h \
	$(INCRT)/sys/systm.h \
	$(INCRT)/sys/text.h \
	$(INCRT)/sys/tty.h \
	$(INCRT)/sys/tty_pty.h \
	$(INCRT)/sys/user.h \
	$(INCRT)/sys/vbavar.h \
	$(INCRT)/sys/vddc.h \
	$(INCRT)/sys/vioc.h \
	$(INCRT)/sys/xymacro.h \
	$(INCRT)/sys/mtio.h
