/*	space.c	6.1	83/07/29	*/

#include "../sys/param.h"
#include "../sys/conf.h"
#include "../ml/vmparam.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/text.h"
#include "../sys/file.h"
#include "../sys/callout.h"
#include "../sys/cmap.h"
#include "../sys/mbuf.h"
#include "../sys/quota.h"
#include "../sys/kernel.h"
#include "../sys/msg.h"
#include "../sys/sysinfo.h"
#include "../ml/pte.h"
#include "../sys/buf.h"
#include "../sys/elog.h"
#include "../sys/iobuf.h"
#include "../sys/tty.h"
#include "../sys/map.h"
#include "../sys/flock.h"

#include "../cf/config.h"
#include "../cf/ipcconfig.h"

/*
 * System parameter formulae.
 * Change history and documentation :
 */

int     nvba = NVBA ;

int	hz = HZ;
int	tick = 1000000 / HZ;
char	*ptrtozero = 0;
struct	timezone tz = { TIMEZONE, DST };
#ifdef DMMAX
int dmmax=DMMAX ;
#endif
#ifdef DMMIN
int dmmin=DMMIN ;
#endif
#ifdef INET
#	define	NETSLOP	20			/* for all the lousy servers*/
#else
#	define	NETSLOP	0
#endif
/*
 * These are initialized at bootstrap time
 * to values dependent on memory size
 * nbuf may be set by the administrator by defining BUFFERS
 * in the configuration file.
 */
int	nswbuf;
#if BUFFERS > 0
int	nbuf = BUFFERS;
#else
int	nbuf = 0;
#endif

#if PROCS == 0
#	define	PROCS (20 + 8 * MAXUSERS)
#endif
int	nproc = PROCS;

#if INODES > 0
int	ninode = INODES;
#else
int	ninode = (PROCS + 16 + MAXUSERS) + 32;
#endif

#if FILES > 0
int	nfile = FILES;
#else
int	nfile = 16 * (PROCS + 16 + MAXUSERS) / 10 + 32 + 2 * NETSLOP;
#endif

#if CALLS > 0
int	ncallout = CALLS;
#else
int	ncallout = 16 + PROCS;
#endif

#if TEXTS > 0
int	ntext = TEXTS;
#else
int	ntext = 24 + MAXUSERS + NETSLOP;
#endif

#if CLISTS > 0
int	nclist = CLISTS;
#else
int	nclist = 100 + 16 * MAXUSERS;
#endif

#if MAXPROC > 0
int maxproc = MAXPROC;
#else
int maxproc = 25; 
#endif

/*
/* If you change this formula make sure you also update 
/* the definition of SYSPTSIZE in ../ml/sysmap.s
*/
int	sysptsize = ((128*NPTEPG/2)+(MAXUSERS*NPTEPG/16)+100);

int	nport = PROCS / 2;
#ifdef QUOTA
int	nquota = (MAXUSERS * 9)/7 + 3;
int	ndquot = (MAXUSERS*NMOUNT)/4 + PROCS;
#endif

/*
* Shared memory
*/

int	dmshm = DMSHM;
int	nshmem = SHMMNI;
struct shmid_ds *shmem, *shmemNSHMEM;
struct	shminfo shminfo = {
	SHMMAX,
	SHMMIN,
	SHMMNI,
	SHMSEG,
	SHMBRK,
	SHMALL
};

/*
 * Semaphores
 */

struct semid_ds	sema[SEMMNI];
struct sem	sem[SEMMNS];
struct map	semmap[SEMMAP];
#define	SEMUSZ	(sizeof(struct sem_undo)+sizeof(struct undo)*SEMUME)
int	semu[((SEMUSZ*SEMMNU)+NBPW-1)/NBPW];
union {
	short		semvals[SEMMSL];
	struct semid_ds	ds;
	struct sembuf	semops[SEMOPM];
}	semtmp;

struct	seminfo seminfo = {
	SEMMAP,
	SEMMNI,
	SEMMNS,
	SEMMNU,
	SEMMSL,
	SEMOPM,
	SEMUME,
	SEMUSZ,
	SEMVMX,
	SEMAEM
};

/*
 * Messages
 */

int msgmappte = MSGMAPPTE;
struct msqid_ds	msgque[MSGMNI];
struct msg	msgh[MSGTQL];
struct msginfo	msginfo = {
	MSGMAP,
	MSGMAX,
	MSGMNB,
	MSGMNI,
	MSGSSZ,
	MSGTQL,
	MSGSEG
};


/* file and record locking */
struct flckinfo flckinfo = {
	FLCKREC,
	FLCKFIL,
	0,
	0,
	0,
	0
};

struct filock flox[FLCKREC];		/* lock structures */
struct flino flinotab[FLCKFIL];		/* inode to lock assoc. structures */

struct sysinfo sysinfo;
struct syswait syswait;
struct syserr syserr;


/* ulimit value - maximum file size (in 512 byte blocks)
 * that can be written by non-super-users.
 *
 * Do not set this to a value larger than 0x01fffff
 */
#ifndef CDLIMIT
#define CDLIMIT	(0x1fffff)
#endif

daddr_t	 cdlimit = CDLIMIT;

/*
 * These have to be allocated somewhere; allocating
 * them here forces loader errors if this file is omitted.
 */
struct	proc *proc, *procNPROC ;
struct	text *text, *textNTEXT;
struct	inode *inode, *inodeNINODE;
struct	file *file, *fileNFILE;
struct 	callout *callout;
struct	cblock *cfree;
struct	buf *buf, *swbuf;
short	*swsize;
int	*swpf;
char	*buffers;
struct	cmap *cmap, *ecmap;
#ifdef QUOTA
struct	quota *quota, *quotaNQUOTA;
struct	dquot *dquot, *dquotNDQUOT;
#endif

/*
 * disks - maximum 32 per system
 *        vddc.c = EGL, FSD and SMD
 *        hdc.c  = HDC
 */
#include "../sys/dsk.h"
#include "../sys/vddc.h"
#include "../sys/hdc.h"
#include "../sys/vbavar.h"

#if NDSK > 0
	int			ndsk = NDSK;
	struct  vba_device	*vddinfo[NDSK];
	struct	iotime		vdstat[NDSK];
	struct	iobuf		vdtab[NDSK];
#endif NDSK > 0

#if NHD > 0
	hdc_ctlr_type		hdc_ctlrs[NHD];
	struct	vba_ctlr	*hdminfo[NHD];
	hdc_unit_type		hdc_units[NDSK];
#endif NHD > 0

#if NVD > 0
	extern char		vd0utl[];
	extern struct 	pte	VD0map[];
	struct	vba_ctlr	*vdminfo[NVD];
	vdunit_tab		vdunit_info[NDSK];
#if NVD > 1
	extern char		vd1utl[];
	extern struct 	pte	VD1map[];
#endif NVD > 1
#if NVD > 2
	extern char		vd2utl[];
	extern struct 	pte	VD2map[];
#endif NVD > 2
#if NVD > 3
	extern char		vd3utl[];
	extern struct 	pte	VD3map[];
#endif NVD > 3

	vdctlr_tab	vdctlr_info[] = {
		 {UNKNOWN, (char *)VD0map, vd0utl, 0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 1
		,{UNKNOWN, (char *)VD1map, vd1utl, 0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 2
		,{UNKNOWN, (char *)VD2map, vd2utl, 0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#if NVD > 3
		,{UNKNOWN, (char *)VD3map, vd3utl, 0, FALSE, FALSE, TRUE, 0,
			{UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN,
		 	UNKNOWN,UNKNOWN,UNKNOWN,UNKNOWN}}
#endif NVD > 3
#endif NVD > 2
#endif NVD > 1
	};
#endif NVD > 0

/*
**      The following is a kludge inorder to support vmstat and iostat on
**      a 32 disks system.  This will be removed in the next full release.
*/
int	dk_busy;
long	dk_time[16];
long            spare0[16];
long	dk_seek[16];
long            spare1[16];
long	dk_xfer[16];
long            spare2[16];
long	dk_wds[16];
long            spare3[16];
long	dk_bps[16];
long            spare4[16];

long	tk_nin;
long	tk_ncan;
long	tk_nout;
long	tk_rcvint;
long	tk_xmtint;
long	tk_mdmint;

/*
 * cy.c -- magnetic tape driver.
 */
#include "../sys/cipher.h"

#if NCY > 0
	int			ncy = NCY;
	struct	vba_device	*cydinfo[NCY];
	struct	iotime		cystat[NCY];
	struct 	iobuf		cytab[NCY];
	unit_tab		unit_info[NCY];
#endif NCY > 0

#if NTM > 0
	int			ntm = NTM;
	struct vba_ctlr		*cyminfo[NTM];
	extern	char		cy0utl[];
	extern	struct 	pte	CY0map[];
	extern 	int		cy_normal_path();
#if NTM > 1
	extern	char		cy1utl[];
	extern	struct 	pte	CY1map[];
#endif NTM > 1

	ctlr_tab	ctlr_info[] = {
		{CY0map, cy0utl, cy_normal_path},
#if NTM > 1
		{CY1map, cy1utl, cy_normal_path}
#endif NTM > 1
	};
#endif NTM > 0



/*
 * xy.c -- XYLOGICS magnetic tape driver.
 */

#if (NXY > 0) & (NXYD > 0)

#include "../sys/mtio.h"
#include "../sys/xymacro.h"

#define XYNUMIOPB 8*NXYD
	int				xynumiopb = XYNUMIOPB ;
	struct	vba_device		*xy_vba_device[NXYD] ;
	struct	iotime			xystat[NXYD];
	struct 	iobuf			xytab[NXYD];
	struct	unit_tab		xy_unit_info[NXYD];
	struct	iopb			xy_free_iopb_list_arr[XYNUMIOPB] ;
	struct	sg_block		xy_free_sg_list_arr[XYNUMIOPB] ;
	struct	buf			xy_free_buf_list_arr[XYNUMIOPB] ;
	struct  ctlr_parms_iopb		xy_ctlr_setup_struct[NXY] ;
	struct	drive_parms_iopb	xy_drive_setup_struct[NXYD] ;
/* 
**	#defineing NOSG makes driver use physically continuous buffer, xybuffer,
**	and avoid the use of scatter/gather (we've had some hardware problems 
**	with scatter/gather).
*/
#ifdef NOSG
	char				xybuffer[NXYD][XYASYNCBUFSZ*NBPG] ;
#endif
	struct	iopb			xybsr[NXYD] ;

	int			nxyctlr = NXY ;
	struct vba_ctlr		xy_vba_ctlr[NXY] ;
	struct ctlr_tab		xy_ctlr_info[NXY] ;
#endif (NXY > 0) & (NXYD > 0)








/*
 * tty_pty.c -- pseudo terminal driver.
 */
#include "../sys/tty_pty.h"
#if NPTY > 0
	int			npty = NPTY;
	struct	tty		pt_tty[NPTY];
	struct	pt_ioctl	pt_ioctl[NPTY];
#endif


/*
 * vcc.c, vcx.c -- terminal driver.
 */
#include "../sys/vioc.h"
#if NVX > 0
#	define	NVXPORTS	16*NVX
	caddr_t			VIOCBAS[NVX];
	struct	vcmds		v_cmds[NVX] ;
	struct	vcx		vcx[NVX] ;
	struct	vba_device 	*vxinfo[NVX];
	struct	vba_ctlr   	*vxminfo[NVX];
	struct	tty		vx_tty[NVXPORTS];
	int			nvxports = NVXPORTS;
	int			nvx = NVX;
	/*
	 * Used to try to handle break interrupts from the vioc sanely.
	 */
	int			vx_break[NVX];
#endif


/*
 * sxt.c --- shell layers driver.
 */
#include "../sys/sxt.h"
#if NSXT > 0
	int	sxt_cnt = NSXT;
	char 	sxtbuf[NSXT*LINKSIZE];
#endif


/*----------------------------------------------------
  Configure Network Families and Protocols
  --------------------------------------------------*/
/*
 * ../net/af.c
 *
 * Address family support routines
 */

#include "../sys/socket.h"
#include "../net/af.h"
#include "../sys/domain.h"
#include "../net/if.h"

extern	int	null_hash(),
		null_netmatch();
#define	AFNULL	{ null_hash,	null_netmatch }

char	hostname[MAXHOSTNAMELEN] = NODE ;
int	hostnamelen = sizeof(NODE) - 1 ;

/*--------------------- Internet Family and Protocol --------------*/
#ifdef INET

#	define	NLOOP	1		/* always included with TCP/IP */

	/*
	 * (was from 4.2BSD ../net/if.h)
	 */
	struct	ifqueue	ipintrq;		/* ip packet input queue */
#else
	inet_hash(sin, hp)
		register struct sockaddr_in *sin;
		struct afhash *hp;
	{
		null_hash( sin, hp ) ;
	}

	inet_netmatch(sin1, sin2)
		struct sockaddr_in *sin1, *sin2;
	{
		null_netmatch( sin1, sin2 ) ;
	}
	inetADDDOMAIN() {}
#endif

/*
 * if_en.c - ACC ethernet driver.
 */
#if NEN > 0 || NEX > 0
#include "../netinet/if_ether.h"
#include "../sys/stdabv.h"
#endif

#if NEN > 0
	int 		nen = NEN ;
	int 		endpm[] = { 0xF80000, 0xF90000, 0xFA0000, 0xFB0000 } ;
#include "../hcxif/if_en.h"
	struct vba_device *	eninfo[ NEN ] ;
	struct vba_ctlr *	enminfo[ NEN ] ;
	struct vba_driver	endriver = {
		enprobe, enslave, enattach, 0, 0, "en ",
		eninfo, "en ",enminfo, 0
	} ;
	en_softc_type		en_softc[ NEN ] ;
#endif  NEN

/*
 * if_ex.c - Exelan ethernet driver.
 */
#if NEX > 0
#include "../hcxif/if_ex.h"
	int 		nex = NEX ;
	struct vba_device *	exinfo[ NEX ] ;
	struct vba_ctlr *  	exminfo[ NEX ] ;
	struct vba_driver	exdriver = {
		exprobe, exslave, exattach, 0, 0, "ex ",
		exinfo, "ex ", exminfo, 0
	} ;
	ex_softc_type		ex_softc[ NEX ] ;
#endif  NEX

/*---------------- PUP Network Family and Protocols -------------------*/
#	define	AFPUP	AFNULL
	pupADDDOMAIN() {}

/*------------------------ IMP Network Family and Protocols --------------*/
#if NIMP > 0

	/*
	 * (was from 4.2BSD ../sys/uipc_domain.c)
	 *
	 * Define IMP Network Family.
	 */

	nimpADDDOMAIN()
	{
		extern struct domain nimpdomain;

		nimpdomain.dom_next = domains;
		domains = &nimpdomain;
	}

	/*
	 * (was from ../netinet/in_proto.c).
	 *
	 * Define IMP protocol.
	 *
	 */

	int	rip_input(),rip_output();
	extern	int raw_usrreq();
	/*
	 * IMP protocol family: raw interface.
	 * Using the raw interface entry to get the timer routine
	 * in is a kludge.
	 */
	int	rimp_output(), hostslowtimo();

	struct protosw impsw[] = {
	{ SOCK_RAW,	PF_IMPLINK,	0,		PR_ATOMIC|PR_ADDR,
	  0,		rimp_output,	0,		0,
	  raw_usrreq,
	  0,		0,		hostslowtimo,	0,
	},
	};

	struct domain impdomain =
	    { AF_IMPLINK, "imp", impsw, 
		    &impsw[sizeof (impsw)/sizeof(impsw[0])] };

#else
	nimpADDDOMAIN() {}
#endif

/*------------------------ Software Loopback -----------------------*/
#ifndef NLOOP
	loattach() {};
#endif NLOOP

/*	=======BYSYNC =========
*/
#include "../sys/bsc/bscio.h"
#include "../sys/bsc/bsc.h"

	char 	bscport[NVXPORTS];
	char	bscno = 0;
#if NVBSC > 0
	int 	nvbsc = NVBSC;
	struct 	bsc bsc[NVBSC*4];
	struct	bscsub bscsubs[NVBSC*4][NBSSUB];
#else
	int	nvbsc = 0;
	struct 	bsc bsc[1];
	struct	bscsub bscsubs[1][1];
	bsctxd() {};
#endif
/*	======End of BYSYNC======
*/

/*	======NFS======
*/
#ifndef NFS
nfs_mount() { nosys(); } ;
nfs_svc(){ nosys(); } ;
async_daemon(){ nosys(); } ;
nfs_getfh(){ nosys(); } ;
nfs_init(){};
#endif NFS
/*	======End of NFS======
*/
