/*
 * This following structure describes an ethernet frame.
 * This frame is defined by the the ACC V/EIU baord
 * A frame is used for either transmitting or receiving.
 * The first word of the frame is used for the command to the
 * V/EIU board and for the status from the V/EIU board.
 */

#define ENUNIT(x)   minor(x)
 
#define ENDEBUG(x)
 
#define SZ_DEBUG_DATA    30
 
struct  en_debug_struct is
   uint2 counter ;
   uint2 direction ;
#define EN_XMIT 1
#define EN_RECV 2
   uint2 cmnd_status ;
   uint2 hdr_dhost[ 3 ] ;
   uint2 hdr_shost[ 3 ] ;
   uint2 hdr_type ;
   uint2 data[ SZ_DEBUG_DATA ] ;
end_struct ;
typedef struct en_debug_struct en_debug_type ;
 
#ifdef EN_MONITOR
en_debug_type en_debug_buffer[ SZ_DEBUG_BUFFER ] ;
#endif EN_MONITOR
 
#define EN_IVCT		(uint2)0x00A0
 
#define TSEG                 8

#define SZ_RANDOM           16
#define SZ_MIN_FRAME        46
#define SZ_FRAME_DATA     2000
 
struct frame_struct is
   int2                cmnd_status ;
   int2                hdr_dhost[ 3 ] ;
   int2                hdr_shost[ 3 ] ;
   u_short             hdr_type ;
   int1                frame_data[ SZ_FRAME_DATA ] ;
   int2                random[ SZ_RANDOM ] ;
end_struct ;
 
typedef struct frame_struct frame_type ;
 
/*
 * The following are defines for bits and fields in the cmnd_status word.
 */
#define CS_BYTE_CNT     0x07FF
#define CS_RODD         0x1000
#define CS_RCRC         0x2000
#define CS_ROVRN        0x4000
#define CS_FULL         0x8000
#define CS_EMPTY        0x8000
#define CS_TRTC         0x000F
#define CS_TRTFAIL      0x4000

#define NUMBER_FRAMES 16

/*
 * There are IO port registers on the V/EIU board. The following
 * structure describes them.
 */

struct en_reg_struct ;
typedef struct en_reg_struct en_reg_type ;

struct en_reg_struct is

#define SZ_STATION_ADDR  6
#define VALID_STATION_ADDR 0x00FF

   int2 station_addr[ SZ_STATION_ADDR ] ;

#define SZ_BROADCAST_ADDR 2
#define VALID_BROADCAST_ADDR 0x00FF

   int2 broadcast_addr[ SZ_BROADCAST_ADDR ] ;

#define SZ_MULTICAST_ADDR 8
#define VALID_MULTICAST_ADDR 0x00FF

   int2 multicast_addr[ SZ_MULTICAST_ADDR ] ;

#define VALID_CSR     0x03FF
#define CSR_GO        0x0001
#define CSR_ENINT     0x0002
#define CSR_LOOP2     0x0004
#define CSR_LOOP3     0x0008
#define CSR_TCRC      0x0010
#define CSR_PROM      0x0020
#define CSR_RESET     0x0040
#define CSR_ACTV      0x0080
#define CSR_RESERVED8 0x0100
#define CSR_ENXMITODD 0x0200
#define CSR_RESERVEDA 0x0400
#define CSR_RESERVEDB 0x0800
#define CSR_RESERVEDC 0x1000
#define CSR_RESERVEDD 0x2000
#define CSR_RESERVEDE 0x4000
#define CSR_RESERVEDF 0x8000

   int2 csr ;

#define VALID_TSEG 0x7800

   int2 tseg ;

#define VALID_RSEG 0x7800

   int2 rseg ;
 
#define VALID_SEGB 0x7800

   int2 segb ;

#define VALID_RDRP 0x000F

   int2 rdrp ;

#define VALID_IVCT 0x00F8

   int2 ivct ;

#define VALID_RESERVED 0x0000

   int2 reserved ;
 
#define VALID_FCOL 0xFFFF

   int2 fcol ;

end_struct ;
 
/*
 * Ethernet software status per interface.
 *
 * Each interface is referenced by a network interface structure,
 * is_if, which the routing code uses to locate the interface.
 * This structure contains the output queue for the interface, its address, ...
 */
struct en_softc_struct ;
typedef struct en_softc_struct en_softc_type ;
 
struct en_softc_struct is
 
   struct arpcom   is_ac ;               /* Ethernet common part      */
 
   frame_type    * frame_ptr ;
   en_reg_type   * en_reg_ptr ;
   bool            xmit_avail ;          /* TRUE if there is an xmit    */
                                         /* buffer available.           */
   uint4           xmit_avail_head ;     /* The first available buffer. */
   uint4           xmit_avail_tail ;     /* Oldest full buffer.         */
   uint4           recv_buffer ;         /* The next receive buffer.    */
 
   uint4           error_odd ;           /* These errors are reported   */
   uint4           error_crc ;           /* by the V/IEU board.         */
   uint4           error_over_run ;
   uint4           error_too_short ;
 
   struct mbuf *   output_head_ptr ;
   struct mbuf *   output_tail_ptr ;
   uint4           queued_packets ;
   uint4	   queue_depth ;         /* Instantaneous output q depth*/
   uint4           max_queue_depth ;     /* Maximum outout q depth.     */
 
   bool            monitor ;             /* Monitoring variables.       */
   uint4           counter ;
   en_debug_type * debug_ptr ;
   en_debug_type * start_debug_ptr ;
   en_debug_type * end_debug_ptr ;
 
end_struct ;


#ifdef KERNEL

int4 enprobe( ), enslave( ),  enattach( ), enrintr( ), enxintr( ) ;
int4 eninit( ),  enoutput( ), enioctl( ),  enreset( ) ;
struct mbuf * enget( ) ;

extern int4			nen ;
extern struct vba_device *	eninfo[ ] ;
extern en_softc_type		en_softc[ ] ;

extern int4			endpm[ ] ;

extern struct vba_driver	endriver ;

#endif KERNEL
