
/*
 * Ethernet software status per interface.
 *
 * Each interface is referenced by a network interface structure, ex_ptr_if, which 	/* 
 * the routing code uses to locate the interface.  This structure contains the 
 * output queue for the interface, its address, ... NOTE: To configure multiple
 * controllers, the sizeof this structure must be a multiple of 16 (ex_ptr_h2xhdr).
 */

#define DEBUG	(boothowto & RB_DEBUG)

#define PHYS_ADDR( addr )	(uint4)((uint4)(addr) & (uint4)0x00ffffff)
#define VME_ADDRESS_MODIFIER	(uint4)0x39
#define EX_IVCT			(uint2)0x00B8
#define	NH2X 4			/* Host to eXcelan request buffers */
#define	NX2H 4			/* eXcelan to Host reply buffers */

struct ex_reg_struct ;
typedef struct ex_reg_struct ex_reg_type ;

struct	ex_reg_struct is
	uint2	ex_porta;	/* write on porta resets EXOS */
	uint2	ex_portb;	/* write on portb interrupts EXOS */
end_struct ;

#define	NFRAGMENTS 1	/* number fragments that the EXOS will scatter/gather */

#define	EXMAXRBUF 1520	/* per EXOS 101 manual 5.3.7 (maybe 1518 would do) */

struct	buf_blk {		/* packet/buffer block descriptor */
	uint2	bb_len;			/* length of block, in bytes */
	uint2	bb_addr[2];		/* address of block */
	/*
	 * Array above is really a single uint4 field.
	 * We kludge its definition to defeat word-alignment.
	 * Access would look like:
	 *     longaddr = *(uint4 *)bp->.mb_er.er_blks[0].bb_addr;
	 */
};

struct	net_mode {		/* read/write mode control objects */
/*12*/	uint1	nm_request;	/* request code */
/*13*/	uint1	nm_reply;	/* reply code */
/*14*/	uint1	nm_mask;	/* bit-wise switches for read, write */
/*15*/	uint1	nm_optn;		/* acceptable packet reception errors */
/*16*/	uint1	nm_mode;		/* EXOS filtering mode */
/*17*/
};

struct	net_addrs {		/* read/write receive address slots */
/*12*/	uint1	na_request;	/* request code */
/*13*/	uint1	na_reply;	/* reply code */
/*14*/	uint1	na_mask;	/* bit-wise switches for read, write */
/*15*/	uint1	na_slot;		/* index of address slot */
/*16*/	uint1	na_addrs[6];		/* address read and/or written */
/*22*/
};

struct	net_recv {		/* read/alter receive slot enable bit */
/*12*/	uint1	nr_request;	/* request code */
/*13*/	uint1	nr_reply;	/* reply code */
/*14*/	uint1	nr_mask;	/* bit-wise switches for read, write */
/*15*/	uint1	nr_slot;		/* index of address slot */
/*16*/
};

struct	net_ststcs {		/* read/reset network statistics objects */
/*12*/	uint1	ns_request;	/* request code */
/*13*/	uint1	ns_reply;	/* reply code */
/*14*/	uint1	ns_mask;	/* bit-wise switches for read, write */
/*15*/	uint1	ns_rsrv;		/* reserved for EXOS */
/*16*/	uint2	ns_nobj;		/* number of objects to work on */
/*18*/	uint2	ns_xobj;		/* index of first object to work on */
/*20*/	uint4	ns_bufp;		/* pointer to statistics buffer */
/*24*/
};

struct	enet_xmit {		/* send a packet on the Ethernet */
/*12*/	uint1	et_request;	/* request code */
/*13*/	uint1	et_reply;	/* reply code */
/*14*/	uint1	et_slot;
/*15*/	uint1	et_nblock;		/* number of blocks composing packet */
/*16*/	struct	buf_blk et_blks[NFRAGMENTS];	/* array of block descriptors */
/*22-64*/
};

struct	enet_recv {		/* receive a packet on the Ethernet */
/*12*/	uint1	er_request;	/* request code */
/*13*/	uint1	er_reply;	/* reply code */
/*14*/	uint1	er_slot;
/*15*/	uint1	er_nblock;		/* number of blocks composing buffer */
/*16*/	struct	buf_blk er_blks[NFRAGMENTS];	/* array of block descriptors */
/*22-64*/
};

struct	stat_array {
	uint4	sa_fsent;	/* frames sent without errors */
	uint4	sa_xsclsn;	/* frames aborted excess collisions */
	uint4	sa_nsqe;	/* frames subject to heartbeat failure */
	uint4	sa_undef;	/* undefined (TDR on EXOS 101) */
	uint4	sa_frcvd;	/* frames received no errors */
	uint4	sa_align;	/* frames received alignment error */
	uint4	sa_crc;		/* frames received crc error */
	uint4	sa_flost;	/* frames lost */
};

struct ex_config_struct ;
typedef struct ex_config_struct ex_config_type ;

struct	ex_config_struct is
/*00*/	uint2	cm_0rsrv;	/* reserved, must be 1 */
/*02*/	char	cm_vc[4];	/* returns ASCII version code */
/*06*/	uint1	cm_cc;		/* returns config completion code */
/*07*/	uint1	cm_opmode;	/* specifies operation mode */
/*08*/	uint2	cm_dfo;		/* specifies host data format option */
/*10*/	uint1	cm_1rsrv ;	/* reserved, must be 0 */
/*11*/  uint1   cm_context ;    /* must be 0 for request, xos supplies */
/*12*/	uint1	cm_2rsrv ;  	/* reserved, must be 0 */
/*13*/	uint1	cm_ham;		/* specifies host address mode */
/*14*/	uint1	cm_3rsrv;	/* reserved, must be 0 */
/*15*/	uint1	cm_mapsiz;	/* reserved, must be 0 */
/*16*/	uint1	cm_byteptrn[4];	/* host data format option test pattern */
/*20*/	uint2	cm_wordptrn[2];
/*24*/	uint4	cm_lwordptrn;
/*28*/	uint1	cm_rsrvd[20];
/*48*/	uint4	cm_mba;		/* use 0xFFFFFFFF in link level mode */
/*52*/	uint1	cm_nproc;	/* use 0xFF in link level mode */
/*53*/	uint1	cm_nmbox;	/* use 0xFF in link level mode */
/*54*/	uint1	cm_nmcast;	/* use 0xFF in link level mode */
/*55*/	uint1	cm_nhost;	/* use 1 in link level mode */

	/* the next five parameters define the request message queue */
/*56*/	uint4	cm_h2xba;	/* base address of message queue */
/*60*/	uint2	cm_h2xhdr;	/* address offset of msg Q header */
/*62*/	uint1	cm_h2xtyp;	/* interrupt type */
/*63*/	uint1	cm_h2xval;	/* interrupt value (not used) */
/*64*/	uint2	cm_h2xaddr;	/* interrupt vector */
/*66*/	uint2	cm_h2xpad;	/* pad out unused portion of vector */

	/* the next five parameters define the reply message queue */
/*68*/	uint4	cm_x2hba;	/* base address of message queue */
/*72*/	uint2	cm_x2hhdr;	/* address offset of msg Q header */
/*74*/	uint1	cm_x2htyp;	/* interrupt type */
/*75*/	uint1	cm_x2hval;	/* interrupt value (not used) */
/*76*/	uint2	cm_x2haddr;	/* interrupt vector */
/*78*/	uint2	cm_x2hpad;	/* pad out unused portion of vector */
/*80*/
end_struct ;

struct ex_msg_struct ;
typedef struct ex_msg_struct ex_msg_type ;

/* we send requests and receive replys with the EXOS using this structure */
struct	ex_msg_struct is
/*00*/	uint2	mb_link;	/* address of next message buffer */
/*02*/	uint1	mb_rsrv;	/* reserved for use by EXOS */
/*03*/	uint1	mb_status;	/* used bit-wise for message protocol */
/*04*/	uint2	mb_length;	/* length, in bytes, of the rest */
/*06*/	uint2	mb_1rsrv;	/* reserved for used by EXOS */
/*08*/	int4	mb_mid;		/* available to user */
/*12*/	union	mb_all {
		struct	net_mode	mb_net_mode;
		struct	net_addrs	mb_net_addrs;
		struct	net_recv	mb_net_recv;
		struct	net_ststcs	mb_net_ststcs;
		struct	enet_xmit	mb_enet_xmit;
		struct	enet_recv	mb_enet_recv;
	} mb_all;
/* following field is used only by host, not read by board */
	ex_msg_type *	mb_next;	/* host's pointer to next message */
end_struct ;

struct ex_queue_base_struct ;
typedef struct ex_queue_base_struct ex_queue_base_type ;

struct ex_queue_base_struct is

	uint2		h2xhdr ;	/* EXOS's request queue header */
	uint2		x2hhdr ;	/* EXOS's reply queue header */
	ex_msg_type 	h2xmsg[NH2X] ;	/* request msg buffers */
	ex_msg_type 	x2hmsg[NX2H] ;	/* reply msg buffers */
	ex_config_type	cm ;		/* configuration message */
	struct stat_array sa ;		/* EXOS writes stats here */
end_struct ;

struct  ex_debug_struct is
	uint2 counter ;
	uint2 direction ;
#define EX_XMIT 1
#define EX_RECV 2
	uint2 cmnd_status ;
	uint1 hdr_dhost[ 6 ] ;
	uint1 hdr_shost[ 6 ] ;
	uint2 hdr_type ;
	uint1 data[ 60 ] ;
end_struct ;
typedef struct ex_debug_struct ex_debug_type ;

#ifdef EX_MONITOR
ex_debug_type ex_debug_buffer[ SZ_DEBUG_BUFFER ] ;
#endif EX_MONITOR

struct ex_softc_struct ;
typedef struct ex_softc_struct ex_softc_type ;

struct	ex_softc_struct is

	struct arpcom	is_ac ;		/* Ethernet common part      */
	ex_reg_type *	reg_ptr ;
	int4		is_flags ;	/* private flags */
	uint4           unit ;          /* Unit number of this EXOS board */
#define	EX_XPENDING	1		/* xmit rqst pending on EXOS */
#define	EX_STATPENDING	2		/* stats rqst pending on EXOS */
#define	EX_RPENDING	4		/* recv rqst pending on EXOS */
	ex_msg_type *	is_h2xnext ;	/* host pointer to request queue */
	ex_msg_type *	is_x2hnext ;	/* host pointer to reply queue */
	ex_queue_base_type *	ex_qb_ptr ;	/* map info for structs below */
	ex_reg_type *   ex_reg_ptr ;    /* points to porta and portb */
	/* the following is not referenced directly. The address of the queue */
	/* base must fall on a 16 byte boundary. ex_qb_ptr will point at      */
	/* the first address after pad that meets this criterion. After       */
	/* ex_qb_ptr is initialized, the following should not be referenced.  */
	char			pad[ 16 ] ;
	ex_queue_base_type	queue_base ;
	/* end mapped area */
	char			ex_rbuf[ ETHERMTU+50 ] ;
	char			ex_tbuf[ ETHERMTU+50 ] ;

#ifdef EX_MONITOR
	uint4			counter ;
	ex_debug_type *		debug_ptr ;
	ex_debug_type *		start_debug_ptr ;
	ex_debug_type *		end_debug_ptr ;
#endif

end_struct  ;

#ifdef KERNEL

int4 exprobe( ), exslave(),   exattach( ), exintr( ) ;
int4 exinit( ),  exoutput( ), exioctl( ),  exreset( ) ;
struct mbuf * exget( ) ;

extern int4			nex ;
extern struct vba_device *	exinfo[ ] ;
extern struct vba_ctlr *	exminfo[ ] ;
extern ex_softc_type		ex_softc[ ] ;

extern struct vba_driver	exdriver ;

#endif KERNEL

#define EXWATCHINTVL	20	/* call exwatch every x secs */

/* EXOS I/O PORT A write definitions */
#define	EX_RESET     0x00	/* value doesn't really matter... */

/* EXOS I/O PORT B write definitions */
#define	EX_INTERRUPT 0x00

/* EXOS I/O PORT B read definitions */
#define	EX_TESTOK    0x01	/* set when self-diagnostics passed */
#define	EX_UNREADY   0x08     	/* set until EXOS ready to read from B */

/* message buffer status field definitions */
#define	MH_OWNER	1	/* mask for status bit for owner */
#define	MH_HOST		0	/* if 0, the host owns the buffer */
#define	MH_EXOS		1	/* if 1, the EXOS owns the buffer */

/* EXOS Link Level request codes */
#define	LLTRANSMIT	0xC	/* send a packet */
#define	LLRTRANSMIT	0xE	/* send a packet, and self-receive */
#define	LLRECEIVE	0xD	/* receive a packet */
#define	LLNET_MODE	0x8	/* read/write mode control objects */
#define	LLNET_ADDRS	0x9	/* read/write receive address slots */
#define	LLNET_RECV	0xA	/* read/alter receive slot enable bit */
#define	LLNET_STSTCS	0xB	/* read/reset network statistics objects */

/* Link Level return codes common to all requests */
#define	LL_OK		0	/* successful completion */
#define	LLX_MODE	0xA1	/* EXOS not in link level mode. */

/* LLTRANSMIT unique return codes */
#define	LLXM_1RTRY	0x1	/* successful xmission, 1 retry */
#define	LLXM_RTRYS	0x2	/* successful xmission, more than 1 retry */
#define	LLXM_NSQE	0x8	/* successful xmission, no SQE TEST signal */
#define	LLXM_CLSN	0x10	/* xmission failed, excess retries */
#define	LLXM_NCS	0x20	/* xmission failed, no carrier sense */
#define	LLXM_LNGTH	0x40	/* xmission failed, bad packet length */
#define	XMIT_BITS	"\7\7LENGTH\6CARRIER\5XCLSNS\4SQETST"
#define	LLXM_ERROR	(LLXM_NSQE|LLXM_CLSN|LLXM_NCS|LLXM_LNGTH)

/* LLRECEIVE unique return codes */
#define	LLRC_TRUNC	0x4	/* pkt received, but truncated to fit buffer */
#define	LLRC_ALIGN	0x10	/* pkt received, but with alignment error */
#define	LLRC_CRC	0x20	/* pkt received, but with CRC error */
#define	LLRC_BUFLEN	0x40	/* no pkt received, buffer less than 64 bytes */
				/* this should never happen here */
#define	RECV_BITS	"\7\7BUFLEN\6CRC\5ALIGN\3TRUNC"

/* LLNET_ADDRS unique return codes */
#define	LLNA_BADSLOT	0xD1	/* slot doesn't exist or can't be accessed */
#define	LLNA_BADADDR	0xD3	/* invalid address for designated slot */

/* LLNET_RECV unique return codes */
#define	LLNR_BADSLOT	0xD1	/* slot doesn't exist or can't be accessed */
#define	LLNR_BADADDR	0xD2	/* designated slot was empty */


/* address slot object indices */
#define	NULLSLOT	0	/* the null slot */
#define	MINMCSLOT	1	/* minimum multicast slot index */
#define	MAXMCSLOT	8	/* default maximum multicast slot index */
#define	PHYSSLOT	253	/* physical slot index */
#define	UNVRSSLOT	254	/* universal slot index */
#define	BROADSLOT	255	/* broadcast slot index */

/* request mask bit definitions */
#define	WRITE_OBJ	1	/* write the object */
#define	READ_OBJ	2	/* read the object */
#define	ENABLE_RCV	4	/* enable reception on designated slot */

/* NET_MODE options mask bit definitions */
#define	OPT_ALIGN	0x10	/* receive packets with alignment errors */
#define	OPT_CRC		0x20	/* receive packets with CRC errors */
#define	OPT_DSABLE	0x80	/* disconnect controller hardware */

/* NET_MODE mode field value definitions */
#define	MODE_OFF	0	/* stop transmission and reception */
#define	MODE_PERF	1	/* perfect multicast address filtering */
#define	MODE_HW		2	/* hardware-only multicast address filtering */
#define	MODE_PROM	3	/* promiscuous reception */

/*
 * N.B.  Structures below are carefully constructed so that
 * they correspond to the message formats that NX firmware
 * defines.  None of them should contain any compiler-instigated
 * padding.  Be especially careful about VAX C longword alignment!
 */

#define	mb_nm	mb_all.mb_net_mode
#define	mb_na	mb_all.mb_net_addrs
#define	mb_nr	mb_all.mb_net_recv
#define	mb_ns	mb_all.mb_net_ststcs
#define	mb_et	mb_all.mb_enet_xmit
#define	mb_er	mb_all.mb_enet_recv
#define	MBDATALEN (sizeof(union mb_all)+6)
