/*	locore.s	4.50	81/05/18	*/
 # -------------------

#include "../ml/mtpr.h"
#include "../ml/trap.h"
#include "../ml/psl.h"
#include "../ml/pte.h"
#include "../ml/cp.h"
#include "../ml/mem.h"

#include "../sys/errno.h"

	.set	HIGH,0x1f	# mask for total disable
	.set	BERVEC,0x80	# offset into scb of the bus error vector 
	.set	RESTVEC,0x8	# offset into scb of the restart vector 
	.set	MAXPHYSMEM,128*1024*1024-1 # max physical memory 
	.set	MEMUNIT,64*1024 # minimum memory increment

	.set	NISP,3		# number of interrupt stack pages
	.set	SYSTEM,0xC0000000 # virtual address of system start
	.set	PPAGES,0x100000  # Number of possible pages in P0,P1, etc.

/* ACBL for non-negative '_add' */
#define ACBL(_limit,_add,_index,_displ) \
	adda 	_add,_index; \
	cmpl	_index,_limit; \
	bleq	_displ

/* _ACBL for negative '_add' */
#define _ACBL(_limit,_add,_index,_displ) \
	adda	_add,_index; \
	cmpl	_index,_limit; \
	bgeq	_displ

/* Keep address of psl if coming from user mode */
#define CHECK_SFE(_delta,name) \
	bitl	$PSL_CURMOD,_delta(sp); \
	jeql	name; \
	moval	_delta(sp),_user_psl; \
name:

/*
 * User structure is UPAGES at top of user space.
 */
	.globl	_u
	.set	_u,SYSTEM - UPAGES*NBPG

/*
 * Restart stack. Used on power recovery or panic.
 * Takes a core-dump and then halts.
 */ 
	.globl	_rsstk

_rsstk: 	.space	1024-8
	.globl	pwfl_stk	
pwfl_stk:	.space  4
dumpflag:	.space	4

	.globl	_intstack
_intstack:
	.space	NISP*NBPG
eintstack:

	.data
	.globl	pwfl_r0
pwfl_r0:	.space	14*4		# Enough for r0 - r13
	.globl	pwfl_sp
pwfl_sp:	.long	0x12345678	# r14
	.globl	pwfl_SCBB
pwfl_SCBB:	.long	0x12345678
	.globl	pwfl_SBR
pwfl_SBR:	.long	0x12345678
	.globl	pwfl_SLR
pwfl_SLR:	.long	0x12345678
	.globl	pwfl_P0BR
pwfl_P0BR:	.long	0x12345678
	.globl	pwfl_P0LR
pwfl_P0LR:	.long	0x12345678
	.globl	pwfl_P1BR
pwfl_P1BR:	.long	0x12345678
	.globl	pwfl_P1LR
pwfl_P1LR:	.long	0x12345678
	.globl	pwfl_P2BR
pwfl_P2BR:	.long	0x12345678
	.globl	pwfl_P2LR
pwfl_P2LR:	.long	0x12345678
	.globl	pwfl_IPL
pwfl_IPL:	.long	0x12345678
/*
*       Since DCK and CCK are write only they cannot be read and saved, making
*       the following locations useless.
*
	.globl	pwfl_DCK
pwfl_DCK:	.long	0x12345678
	.globl	pwfl_CCK
pwfl_CCK:	.long	0x12345678
*
*/
	.globl	pwfl_PCBB
pwfl_PCBB:	.long	0x12345678
	.globl	pwfl_ISP
pwfl_ISP:	.long	0x12345678
	.globl	pwfl_KSP
pwfl_KSP:	.long	0x12345678
	.globl	pwfl_USP
pwfl_USP:	.long	0x12345678
	.globl	pwfl_MME
pwfl_MME:	.long	0x12345678
	.globl	pwfl_PSL
pwfl_PSL:	.long	0x12345678

/*
 * Do a dump.
 * Called by auto-restart.
 * May be called manually.
 */
	.align	4
	.text
	.globl	_Xdoadump
	.globl	_doadump
_Xdoadump:			# CP brings HCX here on power recovery
	loadr	$0x3fff,pwfl_r0	# Restore r0 - r13
	movl	pwfl_sp,sp	# Restore sp ( = r14 )
	mtpr	pwfl_SCBB,$SCBB
	mtpr	pwfl_SBR,$SBR	# Restore all re_loadable registers
	mtpr	pwfl_SLR,$SLR
	mtpr	pwfl_P0BR,$P0BR
	mtpr	pwfl_P0LR,$P0LR
	mtpr	pwfl_P1BR,$P1BR
	mtpr	pwfl_P1LR,$P1LR
	mtpr	pwfl_P2BR,$P2BR
	mtpr	pwfl_P2LR,$P2LR
	mtpr	pwfl_IPL,$IPL
/*	mtpr	pwfl_DCK,$DCK      DCK and CCK are write only so this won't
	mtpr	pwfl_CCK,$CCK      work.   */
	mtpr	pwfl_PCBB,$PCBB
	mtpr	pwfl_ISP,$ISP
	mtpr	pwfl_KSP,$KSP
	mtpr	pwfl_USP,$USP

	bicpsw	$0xff		# Restore PSW.
	bispsw	pwfl_PSL+2	# Set original bits back (just in case..)
/* now go to mapped mode */
/* Have to change PC to system addresses */
	mtpr	$1,$PACC	# Thoroughly clean up caches.
	mtpr	$1,$PADC
	mtpr	$1,$TBIA
	mtpr	pwfl_MME,$MME  	# Restore MME. Last thing to be done.
	jmp 	*$f0
_doadump:
	.word 0
#define	_rsstkmap	_Sysmap+12			
			# HCX storage, scb, rsstk, interrupt stack
f0:
	mtpr	$HIGH,$IPL
	andl2	$0~PG_PROT,_rsstkmap
	orl2	$PG_KW,_rsstkmap		# Make dump stack r/w
	tstl	dumpflag			# dump only once!
	bneq	f11
	movl	$1,dumpflag
	mtpr	$0,$TBIA
	movab	dumpflag,sp
	callf	$4,_dumpsys
f11:
	halt

/*
 * Interrupt vector routines
 */ 
	.globl	_waittime

	.align	4
	.globl 	_cputype
_cputype:
	.word	632

#define	SCBVEC(name) \
	.align 4; \
	.globl _X/**/name; \
_X/**/name
#define	PRINTF(n,msg)

			# these registers are not restored by the C-compiler.
#define	PUSHR		pushl r0; pushl r1;
#define	POPR		movl (sp)+, r1; movl (sp)+, r0;
			# mask for error code on stack.

#define SAVE_FPSTAT(_delta,name1,name2,name3)\
				bitl	$PSL_DBL,_delta(sp); \
				beql	name1; \
				pushl	$1; \
				pushd; \
				jmp	name2; \
			name1:	pushl	$0; \
				pushl	$0; \
				stf	-(sp); \
			name2:	pushl	$0;

#define REST_FPSTAT(name1,name2,name3)\
				tstl	(sp)+; \
				bitl	$1,8(sp); \
				beql	name2; \
				ldd	(sp); \
				jmp	name3; \
			name2:	ldf	(sp); \
			name3:	moval	12(sp),sp;

#define PUSHBPAR	pushab	6*4(sp) /* Push address of buserr paramters */
#define BPAR1		28		/* Offset to first hardware parameter */

SCBVEC(buserr):
	CHECK_SFE(12,buser)
	SAVE_FPSTAT(12,buer1,buer2,buer3);
	PUSHR
	andl3	BPAR1(sp),$ERRCD,r0
	jeql	go_on
	cmpl	r0,$APE
	jneq	.L111
	halt		# Address parity error !!!
.L111:	cmpl	r0,$VBE	
	jneq	go_on
	halt		# Versabus error !!!
go_on:
	PUSHBPAR
	callf 	$8,_buserror
	POPR
	REST_FPSTAT(_buer1,_buer2,_buer3);
	movab	8(sp),sp	# Remove hardware parameters
  	rei
 
SCBVEC(powfail):		# We should be on interrupt stack now.
	movpsl	pwfl_PSL	# Keeps all flags, etc.
	storer	$0x3fff,pwfl_r0	# Saves r0 - r13
	moval	0(sp),pwfl_sp	# Saves sp ( = r14 )
	callf   $4,_savprv      # Saves priviledged registers
	moval	_Xdoadump-SYSTEM,_scb+RESTVEC
	halt

SCBVEC(stray):
	PUSHR
	subl3	_scb,8(sp),-(sp)
	mfpr	$IPL,-(sp)
	callf	$12,_logstray
	pushab .FFF1
	callf	$8, _printf
	.data
.FFF1:	.byte	'*,'*,'*,'*,'*,'*,'*,' ,'U,'n,'d,'e,'f,'i,'n,'e,'d,' ,'i,'n,'t,'e,'r,'r,'u,'p,'t,' ,'*,'*,'*,'*,'*,'*,'*,012,0
	.text
	POPR
	rei

SCBVEC(soft15):
SCBVEC(soft14):
SCBVEC(soft13):
/*SCBVEC(soft12):*/		/* network software interrupt. see vbavec.s */
SCBVEC(soft11):
SCBVEC(soft10):
#ifndef SIMIO
SCBVEC(soft9):
#endif
SCBVEC(soft7):
SCBVEC(soft6):
SCBVEC(soft5):
SCBVEC(soft4):
#ifndef SIMIO
SCBVEC(soft3):
SCBVEC(soft2):
SCBVEC(soft1):
#endif
	PUSHR
	pushab .FFFF
	callf	$8, _printf
	.data
.FFFF:	.byte	'*,'*,'*,'*,'*,'*,'*,' ,'U,'n,'d,'e,'f,'i,'n,'e,'d,' ,'s,'o,'f,'t,'w,'a,'r,'e,' ,'i,'n,'t,'e,'r,'r,'u,'p,'t,' ,'*,'*,'*,'*,'*,'*,'*,012,0
	.text
	POPR
	rei

#ifdef SIMIO
SCBVEC(soft2):
#endif
SCBVEC(cnrint):
	CHECK_SFE(4,cnint)
	SAVE_FPSTAT(4,cnint1,cnint2,cnint3);PUSHR; 
	pushl $CPCONS; callf $8,_cnrint; POPR; incl _cnt+V_INTR;
	REST_FPSTAT(_cnint1,_cnint2,_cnint3); rei
#ifdef SIMIO
SCBVEC(soft3):
#endif
SCBVEC(cnxint):
	CHECK_SFE(4,xint)
	SAVE_FPSTAT(4,xint1,xint2,xint3);PUSHR; 
	pushl $CPCONS; callf $8,_cnxint; POPR; REST_FPSTAT(_xint1,_xint2,_xint3);
	incl _cnt+V_INTR; rei
SCBVEC(rmtrint):
	CHECK_SFE(4,rmint)
	SAVE_FPSTAT(4,rmint1,rmint2,rmint3); PUSHR; 
	pushl $CPREMOT; callf $8,_cnrint; POPR; REST_FPSTAT(_rmint1,_rmint2,_rmint3);
	incl _cnt+V_INTR; rei
SCBVEC(rmtxint):
	CHECK_SFE(4,rmxint)
	SAVE_FPSTAT(4,rmxint1,rmxint2,rmxint3); PUSHR; 
	pushl $CPREMOT; callf $8,_cnxint; POPR; REST_FPSTAT(_rmxint1,_rmxint2,_rmxint3);
	incl _cnt+V_INTR; rei
#ifdef SIMIO
SCBVEC(soft9):
#endif
#define PUSHPCPSL	pushl 5*4+2*4(sp); pushl 5*4+2*4(sp);
SCBVEC(hardclock):
	CHECK_SFE(4,hdclk)
	SAVE_FPSTAT(4,hdclk1,hdclk2,hcclk3)
	PUSHR
	PUSHPCPSL
	callf $12,_hardclock			# hardclock(pc,psl)
	POPR;
	REST_FPSTAT(_hdclk1,h_dclk2,_hdclk3)
	incl	_cnt+V_INTR		## temp so not to break vmstat -= HZ
	rei
SCBVEC(softclock):
	CHECK_SFE(4,sftclk)
	SAVE_FPSTAT(4,sftcl1,sftcl2,stfcl3)
	PUSHR
	PUSHPCPSL
	callf $12,_softclock			# softclock(pc,psl)
	POPR; 
	REST_FPSTAT(_sftcl1,_sftcl2,_sftcl3)
	rei

/*
 * Trap and fault vector routines
 */ 
#define	TRAP(a)	pushl $T_/**/a; jbr alltraps

/*
 * Ast delivery (profiling and/or reschedule)
 */
/*
 * When we want to reschedule we will force a memory fault by setting the m.s.b
 *  of P0LR. Then , on memory fault if m.s.b of P0LR is on we will clear it and
 *  TRAP(astflt).
 *
 */
 
					# if this bit is on it is an ast.
#define 	ASTBIT	0x00100000  	
#define		P0MASK	0xc0000000

SCBVEC(kspnotval):
	CHECK_SFE(4,ksp)
	pushl	$0
	SAVE_FPSTAT(8,ksp1,ksp2,ksp3)
	TRAP(KSPNOTVAL)
SCBVEC(privinflt):
	CHECK_SFE(4,priv)
	pushl	$0
	SAVE_FPSTAT(8,priv1,priv2,priv3)
	TRAP(PRIVINFLT)
SCBVEC(resopflt):
	CHECK_SFE(4,resop)
	pushl	$0
	SAVE_FPSTAT(8,resop1,resop2,resop3)
	TRAP(RESOPFLT)
SCBVEC(resadflt):
	CHECK_SFE(4,resad)
	pushl	$0
	SAVE_FPSTAT(8,resad1,resad2,resad3)
	TRAP(RESADFLT)
SCBVEC(bptflt):
	CHECK_SFE(4,bpt)
	pushl	$0
	SAVE_FPSTAT(8,bpt1,bpt2,bpt3)
	TRAP(BPTFLT)
SCBVEC(tracep):
	CHECK_SFE(4,trace)
	pushl	$0
	SAVE_FPSTAT(8,trace1,trace2,trace3)
	TRAP(TRCTRAP)
SCBVEC(alignflt):
	CHECK_SFE(4,algn)
	pushl	$0
	SAVE_FPSTAT(8,align1,align2,align3)
	TRAP(ALIGNFLT)
SCBVEC(arithtrap):
	CHECK_SFE(8,arith)
	SAVE_FPSTAT(8,arith1,arith2,arith3)
	TRAP(ARITHTRAP)

SCBVEC(protflt):
	CHECK_SFE(12,prot)
	bitl	$1,(sp)+
	jneq	segflt
	SAVE_FPSTAT(8,prot1,prot2,prot3)
	TRAP(PROTFLT)
segflt:
	SAVE_FPSTAT(8,segflt1,segflt2,segflt3)
	TRAP(SEGFLT)

SCBVEC(fpm):			# Floating Point eMulation
	.globl	_fpemulate	
	CHECK_SFE(12,fpm)
	SAVE_FPSTAT(16,fpm1,fpm2,fpm3)
	callf	$4,_fpemulate	# We want psl, pc and opcode
				# not to be removed by the ret.

				# Don't modify registers now !!
	REST_FPSTAT(_fpm1,_fpm2,_fpm3)
	moval	8(sp),sp
	tstl	(sp)		# Stack= PSL, PC, return_code
	jneq	_Xarithtrap	# If not OK, emulate F.P. exception
	movab	4(sp),sp	# Else remove return_code and
	rei

SCBVEC(sfexcep):
	CHECK_SFE(4,sfe)
 	pushl $0
	SAVE_FPSTAT(8,sfe1,sfe2,sfe3)
	TRAP(ASTFLT)

SCBVEC(transflt):
	CHECK_SFE(12,trans)
	bitl	$1,(sp)+
	bneq	tableflt
pageflt: 
	SAVE_FPSTAT(8,trans1,trans2,trans3)
	TRAP(PAGEFLT)
tableflt: 
	SAVE_FPSTAT(8,_trans1,_trans2,_trans3)
	TRAP(TABLEFLT)

alltraps:
	mfpr	$USP,-(sp)
	callf	$4,_trap
	mtpr	(sp)+,$USP
	incl	_cnt+V_TRAP
	movab	4(sp),sp
	REST_FPSTAT(trap1,trap2,trap3)
	movab	4(sp),sp
	mtpr	$HIGH,$IPL		## dont go to a higher IPL (GROT)
	rei

SCBVEC(syscall):
	moval	8(sp),_user_psl
	SAVE_FPSTAT(8,sys1,sys2,sys3)
	pushl	$T_SYSCALL
	mfpr	$USP,-(sp)
	callf	$4,_syscall
	mtpr	(sp)+,$USP
	incl	_cnt+V_SYSCALL
	movab	4(sp),sp
	REST_FPSTAT(_sys1,_sys2,_sys3)
	movab	4(sp),sp
	rei
	.text
/*
 * Initialization
 *
 * IPL 0x1f; MME 0; scbb, pcbb, sbr, slr, isp, ksp not set
 *
 */

	.align	4
	.globl	start
start:
	.word	0

/* set system control block base and system page table params */

	mtpr	$_scb-SYSTEM,$SCBB
	mtpr	$_Sysmap-SYSTEM,$SBR
	mtpr	$_Syssize,$SLR

/* double map the kernel into the virtual user addresses of phys mem */
/* (to be on the safe side.This is supposed to run in system sace. )  */
	mtpr	$_Sysmap,$P0BR
	mtpr	$_Syssize,$P0LR

	mtpr	$_Sysmap,$P1BR		# Against Murphy
	mtpr	$_Syssize,$P1LR

/* set ISP */
	movl	$_intstack-SYSTEM+NISP*NBPG,sp	# Still physical !
	mtpr	$_intstack+NISP*NBPG,$ISP

/* count up memory */

	clrl	r7
.lab1:	pushl	$0 
	pushl	$1 
	pushl	r7 
	callf	$16,_badaddr 
	tstl	r0 
	bneq	.lab9
	adda	$MEMUNIT,r7
	cmpl	r7,$MAXPHYSMEM
	bleq	.lab1
.lab9:
/* clear memory from kernel bss and pages for proc 0 u. and page table */
	movab	_edata,r6
	movab	_end,r5
	andl2	$0~SYSTEM,r6
	andl2	$0~SYSTEM,r5
	adda	$(UPAGES*NBPG)+NBPG+NBPG,r5
.label1:	clrl	(r6); ACBL( r5,$4,r6,.label1)

/* trap(), syscall() save r0-r13 in entry mask (per ../sys/reg.h) */
/* For floating point emulation, we do same for 'fpemulate' */
	orw2	$0x01fff,_trap
	orw2	$0x01fff,_syscall
	orw2	$0x01fff,_fpemulate

/* initialize system page table: scb and int stack writeable */
	clrl	r2
	movab	eintstack,r1 
	andl2	$0~SYSTEM,r1
	shrl 	$PGSHIFT,r1,r1		# r1-page number of eintstack.

/* Start by making the processor storage read/only */

	orl3	$PG_V|PG_KR,r2,_Sysmap[r2]; incl r2;
	orl3	$PG_V|PG_KR,r2,_Sysmap[r2]; incl r2;

/* Other parts of the system are read/write for kernel */

.LAB1:	orl3	$PG_V|PG_KW,r2,_Sysmap[r2]; # data:kernel write + phys=virtual
	aoblss r1,r2,.LAB1

/* make rsstk read-only as red zone for interrupt stack */
	andl2	$0~PG_PROT,_rsstkmap
	orl2	$PG_V|PG_KR,_rsstkmap		# Make dump stack r/w

/* make kernel text space read-only */
/*
 * HAVE TO CHECK ALL THE MAGIC CONSTANTS USED HERE : $xxxxxx */

	movab	_etext+NBPG-1,r1
	andl2	$0~SYSTEM,r1
	shrl 	$PGSHIFT,r1,r1
.LLL:	orl3	$PG_V|PG_KR,r2,_Sysmap[r2]
	aoblss r1,r2,.LLL

/* make kernel data, bss, read-write */
	movab	_end+NBPG-1,r1
	andl2	$0~SYSTEM,r1
	shrl 	$PGSHIFT,r1,r1
.lll:	orl3	$PG_V|PG_KW,r2,_Sysmap[r2]
	aoblss r1,r2,.lll

/* now go to mapped mode */
/* Have to change both PC and SP to system addresses */
	mtpr	$1,$TBIA
	mtpr	$1,$PADC	/* needed by HW parity & ECC logic */
	mtpr	$1,$PACC	/* just in case */
	mtpr 	$1,$MME
	movab	SYSTEM(sp),sp
	.globl	go_virt
go_virt:	
	jmp 	*$.LLL1
.LLL1:

/* disable any interrupts */
	movl	$0,_intenable
/* init mem sizes */
	shrl	$PGSHIFT,r7,_maxmem
	movl	_maxmem,_physmem
	movl	_maxmem,_freemem

/* setup context for proc[0] == Scheduler */
	movab	_end-SYSTEM+NBPG-1,r6
	andl2	$0~(NBPG-1),r6		# make page boundary

/* setup page table for proc[0] */
	shrl	$PGSHIFT,r6,r3			# r3 = btoc(r6)
	orl3	$PG_V|PG_KW,r3,_Usrptmap	# init first upt entry
	incl	r3				# r3 - next page
	movab	_usrpt,r0			# r0 - first user page
	mtpr	r0,$TBIS

/* init p0br, p0lr */
	mtpr	r0,$P0BR	# No P0 for proc[0]
	mtpr	$0,$P0LR

	mtpr	r0,$P1BR	# No P1 either
	mtpr	$0,$P1LR


/* init p2br, p2lr */
	movab	NBPG(r0),r0
	movl	$PPAGES-UPAGES,r1
	mtpr	r1,$P2LR
	moval	-4*PPAGES(r0),r2
	mtpr	r2,$P2BR

/* setup mapping for UPAGES of _u */
	clrl	r2; 
	movl 	$SYSTEM,r1
	adda 	$UPAGES,r3
	jbr .ff2
.ff11:	decl	r3
	moval	-NBPG(r1),r1;	# r1 = virtual add of next (downward) _u page
	suba	$4,r0		# r0 = pte address
	orl3	$PG_V|PG_URKW,r3,(r0)
	mtpr	r1,$TBIS
.ff2:	aobleq	$UPAGES,r2,.ff11

/* initialize (slightly) the pcb */
	movab	UPAGES*NBPG(r1),PCB_KSP(r1)	# KSP starts at end of _u
	movl	r1,PCB_USP(r1)			# USP starts just below _u
	mfpr	$P0BR,PCB_P0BR(r1)
	mfpr	$P0LR,PCB_P0LR(r1)
	mfpr	$P1BR,PCB_P1BR(r1)
	mfpr	$P1LR,PCB_P1LR(r1)
	mfpr	$P2BR,PCB_P2BR(r1)
	mfpr	$P2LR,PCB_P2LR(r1)
	movl	$CLSIZE,PCB_SZPT(r1)		# init u.u_pcb.pcb_szpt
	movl	r11,PCB_R11(r1)			# r11 obtained from CP on boot
  	movl	r10,PCB_R10(r1)			# r10 obtained from CP on boot
	movab	.for1,PCB_PC(r1)			# initial pc
	clrl	PCB_PSL(r1)			# kernel mode, ipl=0
	shll	$PGSHIFT,r3,r3
	mtpr	r3,$PCBB			# first pcbb (physical ! )

/* Go to a 'normal' process mode (kernel) */

	ldpctx
	rei		# Actually 'returns' to the next instruction:

/* put signal trampoline code in u. area */
.for1:	movab	_u,r0
	movl	sigcode+0,PCB_SIGC+0(r0)
	movl	sigcode+4,PCB_SIGC+4(r0)
	movl	sigcode+8,PCB_SIGC+8(r0)
	movl	sigcode+12,PCB_SIGC+12(r0)

/* save reboot flags in global _boothowto */
	movl	r11,_boothowto
  	movl	r10,_boot_device

/* calculate firstaddr, and call main() */
	movab	_end-SYSTEM+NBPG-1,r0
	shrl	$PGSHIFT,r0,-(sp)
	adda	$UPAGES+1,(sp)		# First physical unused page number
	callf 	$8,_main

/* proc[1] == /etc/init now running here in kernel mode; run icode */
	pushl	$PSL_CURMOD		# User mode PSL
	pushl $0			# PC = 0 (virtual now)
	rei

/* signal trampoline code: it is known that this code takes up to 16    */
/* bytes in pcb.h and in the code above 				*/
/*  The following user stack layout was set up by machdep.c/sendsig 	*/
/*	routine:							*/
/*									*/
/*	+---------------+						*/
/*	|  Last PSL	|\						*/
/*	+---------------+ > for rei					*/
/*  :-->|  Last PC	|/						*/
/*  |	+---------------+						*/
/*  :___|__*   SP	|\						*/
/*	+---------------+ |						*/
/*	|  sigmask	| |						*/
/*	+---------------+  > cleaned by kcall $139 (sigcleanup)		*/
/*  :-->|  u.u_onstack	| |						*/
/*  |	+---------------+ |						*/
/*  :___|_* copy of SCP	|/						*/
/*  |	+---------------+						*/
/*  |	|  Process' r0	| 						*/
/*  |	+---------------+						*/
/*  |	|  Process' r1	|						*/
/*  |	+---------------+						*/
/*  |	| Handler addr. |\						*/
/*  |	+---------------+ |						*/
/*  :___|_*   SCP	| |						*/
/*	+---------------+  > cleaned by ret from calls			*/
/*	|  u.u_code	| |						*/
/*	+---------------+ |						*/
/*	|  signal number|/						*/
/*	+---------------+						*/
/*									*/
/*   * Stack when entering sigcode; setup by sendsig();			*/
/*									*/
	.align	4
sigcode:			# When the process executes this code
				#  (located in its u. structure), it
				#  is in user mode !
	calls	$4*4+4,*12(sp)	# 4 words popped from stack when this call returns
	movl	(sp)+,r1
	movl	(sp)+,r0
	kcall	$139		# Signal cleanup
	rei			# From user mode to user mode !

/*
 * Primitives
 */ 

/*
 * badaddr(addr, len, value)
 *	see if access addr with a len type instruction causes a machine check
 *	len is length of access (1=byte, 2=short, 4=long)
 *	value determines whether to read or write:
 *		if value=0, read.
 *		if value!=0, write (appropriate bytes of) value to addr.
 *	r0 = 0 means good(exists); r0 =1 means does not exist.
 */
	.globl	_badaddr
_badaddr:
	.word	0x7c	# Keep r6,r5,r4,r3,r2
	mfpr	$IPL,r1
	mtpr	$HIGH,$IPL
	movl	_scb+BERVEC,r2
	movl	4(fp),r3
	movl	8(fp),r4
	movab	.FOR9,_scb+BERVEC
	movl	0xc(fp),r5
	jneq	.wrtit
/* 
 * Test address by reading.
*/
.rdit:	bbc	$0,r4,.RD1; tstb	(r3)
.RD1:	bbc	$1,r4,.RD2; tstw	(r3)
.RD2:	bbc	$2,r4,.OK; tstl	(r3)
	jmp	.OK			# made it w/o machine checks
/*
 * Test address by writing.
*/
.wrtit:	bbc	$0,r4,.WR1; 
	movob	r5,(r3)
	clrl	r6
.loop1:	aobleq	$1000,r6,.loop1
.WR1:	bbc	$1,r4,.WR2; 
	movow	r5,(r3)
	clrl	r6
.loop2:	aobleq	$1000,r6,.loop2
.WR2:	bbc	$2,r4,.OK; 
	movl	r5,(r3)
	clrl	r6
.loop3:	aobleq	$1000,r6,.loop3
.OK:	clrl	r0			# made it w/o machine checks
.FOR6:	movl	r2,_scb+BERVEC
	mtpr	r1,$IPL
	ret

	.align	4
.FOR9:			# Here we catch bus error (if it comes)
	andl3	4(sp),$ERRCD,r0
	cmpl	r0,$HES
	jneq	.VIB
.HVIB:	andl2	$FATALBUSERR,4(sp) # This destroys item on stack!
	beql	.FOR5
	halt			# Fatal Bus Error (not timeout)
.VIB:	cmpl	r0,$APE
	jneq	.FOR4
	halt			# Address parity error !!!
.FOR4:	cmpl	r0,$VBE
	jneq	.FOR5
	halt			# Versabus error
.FOR5:
	movl	$1,r0		# Anything else = bad address
	movab	8(sp),sp	# discard buss error trash
	movab	.FOR6,(sp)		# new program counter on stack.
	rei


/*
 * big_bcopy: same interface as bcopy but faster for large/aligned blocks.
 */
_big_bcopy:	.globl	_big_bcopy
	.word	0x4		# save r2
	movl	4(fp),r0
	movl	8(fp),r1
	movl	12(fp),r2
	/* Drop through into movblk: */
/*
 * The code below implements an optimized movblk instruction. It is faster
 * than a movblk instruction by a factor of three for long blocks.
 *
 * Enter with:
 *      r0: source address
 *      r1: destination address
 *      r2: # of bytes to move
 *
 * Return with:
 *      r0: 0
 */
movblk:
	bitl	$3,r0		# Is the source address longword aligned?
	bneq	5f		# .. no, go do movblk instruction
	bitl	$3,r1		# Is the destination address longword aligned?
	bneq	5f		# .. no, go do movblk instruction
	bitl	$3,r2		# Is the byte count a multiple of 4?
	beql	1f		# .. yes, go try optimized loop
5:
	movblk
   	clrl	r0
	ret

2:
	cmpl    $32,r2		# Are there atleast 32 more bytes to copy
	bgtr    5b              # .. no, do rest with movblk instruction
	movl    (r0),(r1)       # Copy 32 bytes from source to destination
	movl    4(r0),4(r1)
	movl    8(r0),8(r1)
	movl    12(r0),12(r1)
	movl    16(r0),16(r1)
	movl    20(r0),20(r1)
	movl    24(r0),24(r1)
	movl    28(r0),28(r1)
	adda	$32,r0
	adda	$32,r1
	suba	$32,r2		# decrement byte count to transfer
	jmp     2b              # .. loop

1:
	cmpl    $128,r2		# Are there atleast 128 more bytes to copy
	bgtr    2b              # .. no, do rest with smaller loop
	movl    (r0),(r1)       # Copy 128 bytes from source to destination
	movl    4(r0),4(r1)
	movl    8(r0),8(r1)
	movl    12(r0),12(r1)
	movl    16(r0),16(r1)
	movl    20(r0),20(r1)
	movl    24(r0),24(r1)
	movl    28(r0),28(r1)
	movl    32(r0),32(r1)
	movl    36(r0),36(r1)
	movl    40(r0),40(r1)
	movl    44(r0),44(r1)
	movl    48(r0),48(r1)
	movl    52(r0),52(r1)
	movl    56(r0),56(r1)
	movl    60(r0),60(r1)
	movl    64(r0),64(r1)
	movl    68(r0),68(r1)
	movl    72(r0),72(r1)
	movl    76(r0),76(r1)
	movl    80(r0),80(r1)
	movl    84(r0),84(r1)
	movl    88(r0),88(r1)
	movl    92(r0),92(r1)
	movl    96(r0),96(r1)
	movl    100(r0),100(r1)
	movl    104(r0),104(r1)
	movl    108(r0),108(r1)
	movl    112(r0),112(r1)
	movl    116(r0),116(r1)
	movl    120(r0),120(r1)
	movl    124(r0),124(r1)
	adda	$128,r0
	adda	$128,r1
	suba	$128,r2		# decrement byte count to transfer
	jmp     1b              # .. loop

/*
 * ovbcopy(source,dest,len): overlapping bcopy
 *
 * This implementation of ovbcopy assumes that the 'movblk' instruction
 * is implemented as copying a single byte at a time in the forward
 * direction. As such, if the strings are overlapping and if the
 * source string follows the destination string, a simple 'movblk'
 * instruction is used regardless of the warnings in the architecture
 * manual.
 */

_ovbcopy:	.globl	_ovbcopy
	.word	0x001c
	movl	4(fp),r0        # r0: source ptr.
	movl	8(fp),r1        # r1: destination ptr.
	movl	12(fp),r2       # r2: len
	cmpl	r0,r1           # Compare source and destination
	bgtru	.F_1		# .. normal forward case
	beql	.F_2		# .. equal, nothing to do
	adda	r2,r0		# r0: ptr to first byte after source string
	cmpl	r0,r1           # Is source string entirely before destination?
	bgtru	.F_3            # .. no, do backwards
	suba	r2,r0		# normal forward case. r0: source ptr.
.F_1:
	movblk
.F_2:
	ret
.F_3:                           # Overlapping and must do backwards
	adda	r2,r1		# r1: ptr to first byte following dest string
	subl3	r0,r1,r3        # r3: Distance between strings
				# (i.e. amount to mount on each iteration)
	movl	r2,r4           # r4: total length
	jbr	.F_5
.F_4:
	suba	r3,r0           # r0: ptr to source to copy this time
	suba	r3,r1           # r1: ptr to dest. to copy this time
	movl	r3,r2           # r2: amount to copy this time
	movblk                  # Copy the chunk
	suba	r3,r0           # r0: ptr to source just copied
	suba	r3,r1           # r1: ptr to dest just copied
	suba	r3,r4           # r4: total length left to copy
.F_5:
	cmpl	r4,r3           # Would another iteration be too much?
	jgtr	.F_4            # .. no, copy another r3 bytes
	movl	r4,r2           # r2: number of bytes left to copy
	suba	r2,r0           # r0: ptr to source left to copy
	suba	r2,r1           # r1: ptr to dest left to copy
	movblk                  # copy the last little bit
	ret

/*
 * big_bzero (base, count): same interface as bzero but faster (10 times)
 *  for large/aligned blocks.
 */
_big_bzero: .globl	_big_bzero
	.word	0x4		# save r2
	movl	4(fp),r1	# r1 = destination address
	movl	8(fp),r2	# r2 = count

	bitl	$3,r1		# Is the destination address longword aligned?
	bneq	5f		# .. no, go do movs3 instruction
	bitl	$3,r2		# Is the byte count a multiple of 4?
	bneq	5f		# .. no, go do movs3 instruction

1:
	cmpl    $64,r2          # Are there atleast 64 more bytes to clear
	bgtr    5f              # .. no, do rest with movs3 instruction
	clrl    (r1)		# Clear 64 bytes in destination
	clrl    4(r1)
	clrl    8(r1)
	clrl    12(r1)
	clrl    16(r1)
	clrl    20(r1)
	clrl    24(r1)
	clrl    28(r1)
	clrl    32(r1)
	clrl    36(r1)
	clrl    40(r1)
	clrl    44(r1)
	clrl    48(r1)
	clrl    52(r1)
	clrl    56(r1)
	clrl    60(r1)
	adda	$64,r1
	suba	$64,r2          # decrement byte count to transfer
	brb     1b              # .. loop

5:
	movl	$.FF,r0		# r0 = null source string
	movs3
	ret
.FF:
	.byte 0

/*
 * Copy a null terminated string from the user address space into
 * the kernel address space.
 *
 * copyinstr(fromaddr, toaddr, maxlength, &lencopied)
 */
.globl	_copyinstr
_copyinstr:
	.word	0xfc
	movl	12(fp),r4		# r4 = maximum length
	movl	4(fp),r6		# r6 = user address
	movl	8(fp),r3		# r3 = kernel address
	andl3	$(NBPG*CLSIZE-1),r6,r7	# r7 = bytes to process this page
	subl3	r7,$NBPG*CLSIZE,r7
1:
	cmpl	r4,r7			# r7 = min( bytes on page, length left)
	bgeq	2f
	movl	r4,r7
2:
	prober	$1,(r6),$1		# page accessible ?
	beql	7f
	movl	r6,r0			# prepare to find the string's length
	movl	r6,r1
	movl	r7,r2
	cmps3				# find the string's length
	tstl	r2			# is the string NULL terminated ?
	beql	5f			# if so, continue here
	movl	r6,r0			# prepare to move the string
	movl	r3,r1
	subl3	r2,r7,r5		# r5 = string's length, including NULL
	incl	r5
	movl	r5,r2
	movblk				# move the block, without the NULL
	suba	r5,r4			# update maximum length left
	clrl	r0
3:
	tstl	16(fp)			# return length ?
	beql	4f
	subl3	r4,12(fp),*16(fp)	# actual len = maxlen - unused length
4:
	ret
5:			# the set of bytes does not contain a NULL
	movl	r6,r0
	movl	r3,r1
	movl	r7,r2
	movblk
	adda	r7,r6			# update user address
	adda	r7,r3			# update kernel address
	subl2	r7,r4			# update maximum length left
	beql	6f			# maximum count reached ?
	movl	$NBPG*CLSIZE,r7		# prepare to process next page
	brb	1b
6:
	movl	$ENOENT,r0		# reached maximum count
	brb	3b
7:
	movl	$EFAULT,r0
	brb	3b

/*
 * Copy a null terminated string from the kernel
 * address space to the user address space.
 *
 * copyoutstr(fromaddr, toaddr, maxlength, &lencopied)
 */
.globl	_copyoutstr
_copyoutstr:
	.word	0xfc
	movl	12(fp),r4		# r4 = maximum length
	movl	4(fp),r6		# r6 = kernel address
	movl	8(fp),r3		# r3 = user address
	andl3	$(NBPG*CLSIZE-1),r6,r7	# r7 = bytes to process this page
	subl3	r7,$NBPG*CLSIZE,r7
1:
	cmpl	r4,r7			# r7 = min( bytes on page, length left)
	bgeq	2f
	movl	r4,r7
2:
	probew	$1,(r3),$1		# page accessible ?
	beql	7f
	movl	r6,r0			# prepare to find the string's length
	movl	r6,r1
	movl	r7,r2
	cmps3				# find the string's length
	tstl	r2			# is the string NULL terminated ?
	beql	5f			# if so, continue here
	movl	r6,r0			# prepare to move the string
	movl	r3,r1
	subl3	r2,r7,r5		# r5 = string's length, including NULL
	incl	r5
	movl	r5,r2
	movblk				# move the block, without the NULL
	suba	r5,r4			# update maximum length left
	clrl	r0
3:
	tstl	16(fp)			# return length ?
	beql	4f
	subl3	r4,12(fp),*16(fp)	# actual len = maxlen - unused length
4:
	ret
5:			# the set of bytes does not contain a NULL
	movl	r6,r0
	movl	r3,r1
	movl	r7,r2
	movblk
	adda	r7,r6			# update user address
	adda	r7,r3			# update kernel address
	subl2	r7,r4			# update maximum length left
	beql	6f			# maximum count reached ?
	movl	$NBPG*CLSIZE,r7		# prepare to process next page
	brb	1b
6:
	movl	$ENOENT,r0		# reached maximum count
	brb	3b
7:
	movl	$EFAULT,r0
	brb	3b

/*
 * Copy a null terminated string from one point to another in
 * the kernel address space.
 *
 * copystr(fromaddr, toaddr, maxlength, &lencopied)
 */
.globl	_copystr
_copystr:
	.word	0xfc
	movl	12(fp),r4		# r4 = maximum length
	movl	4(fp),r6		# r6 = source address
	movl	8(fp),r3		# r3 = destination address
	andl3	$(NBPG*CLSIZE-1),r6,r7	# r7 = bytes to process this page
	subl3	r7,$NBPG*CLSIZE,r7
1:
	cmpl	r4,r7			# r7 = min( bytes on page, length left)
	bgeq	2f
	movl	r4,r7
2:
	movl	r6,r0			# prepare to find the string's length
	movl	r6,r1
	movl	r7,r2
	cmps3				# find the string's length
	tstl	r2			# is the string NULL terminated ?
	beql	5f			# if so, continue here
	movl	r6,r0			# prepare to move the string
	movl	r3,r1
	subl3	r2,r7,r5		# r5 = string's length, including NULL
	incl	r5
	movl	r5,r2
	movblk				# move the block, without the NULL
	suba	r5,r4			# update maximum length left
	clrl	r0
3:
	tstl	16(fp)			# return length ?
	beql	4f
	subl3	r4,12(fp),*16(fp)	# actual len = maxlen - unused length
4:
	ret
5:			# the set of bytes does not contain a NULL
	movl	r6,r0
	movl	r3,r1
	movl	r7,r2
	movblk
	adda	r7,r6			# update user address
	adda	r7,r3			# update kernel address
	subl2	r7,r4			# update maximum length left
	beql	6f			# maximum count reached ?
	movl	$NBPG*CLSIZE,r7		# prepare to process next page
	brb	1b
6:
	movl	$ENOENT,r0		# reached maximum count
	brb	3b

/* 
 * Copy specified amount of data from user space into the kernel.
 *
 * copyin(from, to, len)
 */
_copyin:	.globl	_copyin
	.word	0x0004
#ifdef GPROF
	.data
	.align  4
.gpr2:	.long   0
	.text
	pushl   $.gpr2
	callf   $8,mcount
#endif
	movl	12(fp),r2		# copy length
	blss	ersb
	movl	4(fp),r0		# copy user address
	cmpl	$NBPG,r2		# probing one page or less ?
	bgeq	cishort			# yes
ciloop:
	prober	$1,(r0),$1		# first byte accessible ?
	beql	ersb			# no
	adda	$NBPG,r0		# incr user address ptr
	_ACBL	($NBPG+1,$-NBPG,r2,ciloop)	# reduce count and loop
cilast:
	prober	$1,(r0),$1		# first byte accessible ?
	beql	ersb			# no
	prober	$1,-1(r0)[r2],$1	# last byte accessible ?
	beql	ersb			# no
	movl	4(fp),r0   		# r0= source address
	movl	8(fp),r1  		# r1 = destination address
	movl	12(fp),r2               # r2 = byte count
	jmp     movblk                  # use the optimized movblk routine
cishort:
	prober	$1,(r0),$1		# first byte accessible ?
	beql	ersb			# no
	prober	$1,-1(r0)[r2],$1	# last byte accessible ?
	beql	ersb			# no
	movl	8(fp),r1		# prepare for copy: r0, r2 already set
	jmp     movblk                  # Use the optimized movblk routine

ersb:
	movl	$EFAULT,r0
	ret

/* 
 * Copy specified amount of data from kernel to the user space.
 *
 * copyout(from, to, len)
 */
_copyout:	.globl	_copyout
	.word	0x04
#ifdef GPROF
	.data
	.align  4
.gpr1:	.long   0
	.text
	pushl   $.gpr1
	callf   $8,mcount
#endif
	movl	12(fp),r2		# get count
	blss	ersb
	movl	8(fp),r1		# get user address
	cmpl	$NBPG,r2		# probing one page or less ?
	bgeq	coshort			# yes
coloop:
	probew	$1,(r1),$1		# first byte accessible ?
	beql	ersb			# no
	adda	$NBPG,r1		# incr user address ptr
	_ACBL	($NBPG+1,$-NBPG,r2,coloop)	# reduce count and loop
colast:
	probew	$1,(r1),$1		# first byte accessible ?
	beql	ersb			# no
	probew  $1,-1(r1)[r2],$1        # last byte accessible ?
	beql	ersb			# no
	movl	4(fp),r0   		# r0= source address
	movl	8(fp),r1  		# r1 = destination address
	movl	12(fp),r2               # r2 = byte count
	jmp     movblk                  # use the optimized movblk routine
coshort:
	probew	$1,(r1),$1		# first byte accessible ?
	beql	ersb			# no
	probew  $1, -1(r1)[r2],$1       # last byte accessible ?
	beql	ersb			# no
	movl	4(fp),r0		# prepare for copy: r1, r2 already set
	jmp     movblk                  # use the optimzied movblk routine

/*
 * non-local goto's
 */
	.globl	_setjmp			# the _Setjmp subroutine in VAX
					# became _setjmp procedure for HCX
_setjmp:
	.word	0x0
	movl	4(fp),r2
	storer	$0x1ff8,(r2)
	adda	$40,r2
	movl	(fp),(r2)
	adda	$4,r2
	movab	8(fp),(r2)
	adda	$4,r2
	movl	-8(fp),(r2)
	clrl	r0
	ret

	.globl	_longjmp		# the _Longjmp subroutine in VAX
					# became _longjmp procedure for HCX
_longjmp:
	.word	0x0000
	movl	4(fp),r2
_Longjmp:			# called from swtch
	loadr	$0x3ff8,(r2)
	adda	$44,r2
	movl	(r2),r1
	adda	$4,r2
	movab	(sp),r0
	cmpl	r1,r0				# must be a pop
	bgequ	lj2
	pushab	lj1
	callf	$8,_panic
lj2:
	movl	r1,sp
	jmp	*(r2)

lj1:	.byte	'l,'o,'n,'g,'j,'m,'p,0


	.globl	_whichqs
	.globl	_qs
	.globl	_cnt

	.globl	_noproc
	.comm	_noproc,4
	.globl	_runrun
	.comm	_runrun,4

/*
 * The following primitives use the fancy HCX instructions.
 * _whichqs tells which of the 32 queues _qs
 * have processes in them.  setrq puts processes into queues, remrq
 * removes them from queues.  The running process is on no queue,
 * other processes are on a queue related to p->p_pri, divided by 4
 * actually to shrink the 0-127 range of priorities into the 32 available
 * queues.
 */

/*
 * setrq(p), using fancy HCX instructions.
 *
 * Call should be made at spl8(), and p->p_stat should be SRUN
 */
	.globl	_setrq
_setrq:
	.word	0x4
	movl	4(fp),r0
	movzbl	P_PRI(r0),r1		# put on queue which is p->p_pri / 4
	shar	$2,r1,r1
	shal	$1,r1,r2
	moval	_qs[r2],r2
	insque	(r0),*4(r2)		# at end of queue
	shal	r1,$1,r1
	orl2	r1,_whichqs		# mark queue non-empty
	ret

/*
 * remrq(p), using fancy HCX instructions
 *
 * Call should be made at spl8().
 */
	.globl	_remrq
_remrq:
	.word	0x0
	movl	4(fp),r0
	movzbl	P_PRI(r0),r1
	shar	$2,r1,r1
	bbs	r1,_whichqs,rem1
	pushab	rem3			# it wasn't recorded to be on its q
	callf	$8,_panic
rem1:
	remque	(r0)
	bneq	rem2			# q not empty yet
	shal	r1,$1,r1
	mcoml	r1,r1
	andl2	r1,_whichqs		# mark queue empty
rem2:
	ret

rem3:	.byte	'r,'e,'m,'r,'q,0

#ifdef notdef
	.globl __insque
__insque:
	.word 0
	insque	*4(fp), *8(fp)
	ret

	.globl __remque
__remque:
	.word 0
	remque	*4(fp)
	ret
#endif

/*
 * Masterpaddr is the p->p_addr of the running process on the master
 * processor.  When a multiprocessor system, the slave processors will have
 * an array of slavepaddr's.
 */
	.data
	.align	4
	.globl	_masterpaddr
_masterpaddr:
	.long	0

	.text
/*
 * swtch(), using fancy HCX instructions
 */
	.globl	_swtch
_swtch:
	.word	0x0
	movl	(fp),fp			# prepare for rei
	movl	(sp),4(sp)		# saved pc
	tstl	(sp)+
	movpsl	4(sp)
	mtpr	$0x18,$IPL		# no interrupts, please
	ffs	_whichqs,r0		# look for non-empty queue
	bgeq	sw1c
	movl	$1,_noproc
mtpr0:	mtpr	$0,$IPL			# must allow interrupts here
sw1a:	ffs	_whichqs,r0		# look for non-empty queue
	bgeq	sw1b
waitl:	brb	sw1a			# still idle
sw1b:	mtpr	$0x18,$IPL
	bbc	r0,_whichqs,mtpr0	# proc moved via lbolt interrupt
sw1c:	clrl	_noproc
	clrl	_runrun
	shal	$1,r0,r1
	moval	_qs[r1],r1
	movl	(r1),r2			# r2 = p = highest pri process
	remque	*(r1)
	bneq	sw2
	shal	r0,$1,r1
	mcoml	r1,r1
	andl2	r1,_whichqs		# no more procs in this queue
sw2:	movl	*P_ADDR(r2),r0
#ifdef notdef
	movl	r0,_masterpaddr
#endif notdef
	shal	$PGSHIFT,r0,r0		# r0 = pcbb(p)
/*	mfpr	$PCBB,r1		# resume of current proc is easy
 *	cmpl	r0,r1
 */	beql	res0
	incl	_cnt+V_SWTCH
	incl	_sysinfo+PSWTCH
swresm:
	movl	_CMAP2,_u+PCB_CMAP2	# yech
	svpctx
	mtpr	r0,$PCBB
	ldpctx
	movl	_u+PCB_CMAP2,_CMAP2	# yech
	mtpr	$_CADDR2,$TBIS
res0:
	tstl	_u+PCB_SSWAP
	beql	res1
	movl	_u+PCB_SSWAP,r2
	clrl	_u+PCB_SSWAP
	movab	_Longjmp,(sp)
	clrl	4(sp)			# PSL = kernel mode, IPL=0
res1:
/*
 * Now we have to care about code and data keys.
 * 2 functions may be called, so r0 and r1 must be
 * saved across the 'callf'. Also, r2 is used as index
 * since r1 may be changed by the functions (just a
 * precaution).
 */
	PUSHR				# ro and r1
	pushl	r2
	movl	_u+U_PROCP,r2		# r2 = & proc[?]
	tstl	P_CKEY(r2)		# does he have a code key ?
	bneq	ckey_ok
	callf	$4,_getcodekey		# no, we have to give him one.
	movl	r0,P_CKEY(r2)
ckey_ok:
	tstl	P_DKEY(r2)		# same stuff for data key
	bneq	dkey_ok
	callf	$4,_getdatakey
	movl	r0,P_DKEY(r2)
dkey_ok:
	mtpr	P_CKEY(r2),$CCK
	mtpr	P_DKEY(r2),$DCK

	movl	(sp)+,r2		# restore his registers
	POPR
	rei

/*
 * idle location
 */
	.globl	_waitloc1
	.data
_waitloc1:
	.long	mtpr0

	.globl	_waitloc2
	.data
_waitloc2:
	.long	waitl
	.text

/*
 * resume(pf)
 */
	.globl	_resume
_resume:
	.word	0x0
	shal	$PGSHIFT,4(fp),r0	# r0 = pcbb(p)
	movl	(fp),fp			# prepare for rei
	movl	(sp)+,4(sp)		# saved pc
	tstl	(sp)+
	movpsl	4(sp)
	mtpr	$0x18,$IPL		# no interrupts, please
	brw	swresm

/*
 * {fu,su},{byte,word}
 */
	.globl	_fuiword
	.globl	_fuword
_fuiword:
_fuword:
	.word	0x4			# save r2
	movl	4(fp), r1
	prober	$1,(r1),$1		# check access on first byte
	beql	fserr			# page unreadable
	bitl	$1,r1			# check byte alignment
	bneq	fuw2			# odd, do byte-word-byte
	bitl	$2,r1			# check word alignment
	bneq	fuw1			# odd, do in 2 words
	movl	(r1),r0			# move longword
	ret
fuw1:
	prober	$1,3(r1),$1		# check access on last byte
	beql	fserr			# page unreadable
	movw	(r1),r0			# move two words
	shal	$16,r0,r0
	movzwl	2(r1),r1		# orw2 sign extends
	orl2	r1,r0
	ret
fuw2:
	prober	$1,3(r1),$1		# check access on last byte
	beql	fserr			# page unreadable
	movb	(r1),r0			# move byte-word-byte
	shal	$24,r0,r0
	movzwl	1(r1),r2		# orw2 sign extends
	shal	$8,r2,r2
	movzbl	3(r1),r1		# orb2 sign extends
	orl2	r2,r1
	orl2	r1,r0
	ret
fserr:
	mnegl	$1,r0
	ret

	.globl	_fuibyte
	.globl	_fubyte
_fuibyte:
_fubyte:
	.word	0x0
	prober	$1,*4(fp),$1
	beql	fserr
	movzbl	*4(fp),r0
	ret

	.globl	_suiword
	.globl	_suword
_suiword:
_suword:
	.word	0x0
	movl	4(fp), r0
	probew	$1,(r0),$1		# check access on first byte
	beql	fserr			# page unwritable
	bitl	$1,r0			# check byte alignment
	bneq	suw1			# odd byte boundary
	bitl	$2,r0			# check word alignment
	beql	suw2			# longword aligned
	movw	8(fp),(r0)		# move two words
	movw	10(fp),2(r0)
	jbr	suw3
suw1:
	probew	$1,3(r0),$1		# check access on last byte
	beql	fserr			# page unreadable
	movb	8(fp),(r0)
	movb	9(fp),1(r0)
	movb	10(fp),2(r0)
	movb	11(fp),3(r0)
	jbr	suw3
suw2:
	probew	$1,3(r0),$1		# check access on last byte
	beql	fserr			# page unreadable
	movl	8(fp),(r0)
suw3:
	clrl	r0
	ret

	.globl	_suibyte
	.globl	_subyte
_suibyte:
_subyte:
	.word	0x0
	probew	$1,*4(fp),$1
	beql	fserr
	movb	11(fp),*4(fp)
	clrl	r0
	ret

/*
 * Copy 1 relocation unit (NBPG bytes)
 * from user virtual address to physical address
 */
_copyseg: 	.globl	_copyseg
	.word	0x4
	orl3	$PG_V|PG_KW,8(fp),_CMAP2
	mtpr	$_CADDR2,$TBIS		# invalidate entry for copy
	movl	4(fp),r0   		# r0= source address
	movl	$_CADDR2,r1  		# r1 = destination address
	movl	$NBPG,r2		# r2 = byte count
	jmp     movblk                  # use the optimized movblk routine

/*
 * clearseg(physical_page_number);
 *
 * zero out physical memory
 * specified in relocation units (NBPG bytes)
 * This routine was optimized for speed on HCX.
 */
_clearseg: 	.globl	_clearseg
	.word	0
	orl3	$PG_V|PG_KW,4(fp),_CMAP1	# Maps to virtual addr CADDR1
	mtpr	$_CADDR1,$TBIS
	clrl	r1			# r1 = index of cleared long
.BACK1:
	clrl	_CADDR1(r1)             # Clear 64 bytes pointed to by r1
	clrl    _CADDR1+4(r1)
	clrl    _CADDR1+8(r1)
	clrl    _CADDR1+12(r1)
	clrl    _CADDR1+16(r1)
	clrl    _CADDR1+20(r1)
	clrl    _CADDR1+24(r1)
	clrl    _CADDR1+28(r1)
	clrl    _CADDR1+32(r1)
	clrl    _CADDR1+36(r1)
	clrl    _CADDR1+40(r1)
	clrl    _CADDR1+44(r1)
	clrl    _CADDR1+48(r1)
	clrl    _CADDR1+52(r1)
	clrl    _CADDR1+56(r1)
	clrl    _CADDR1+60(r1)
	adda	$64,r1                  # Update r1 to point to next 64 bytes
	cmpl    r1,$1024                # Have we cleared all 1024 bytes yet?
	blss    .BACK1                  # .. no, loop
	ret

/*
 * if ( useracc(address, count, mode) ) ....
 * Check address.
 * Given virtual address, byte count, and rw flag
 * returns 0 on no access.
 * Note : it is assumed that on all calls to this routine,
 *  mode=0 means write access, mode=1 means read access.
 */
_useracc:	.globl	_useracc
	.word	0x4
	movl	$1,r2			# r2 = 'user mode' for probew/probew
probes:
	movl	4(fp),r0		# get va
	movl	8(fp),r1		# count
	tstl	12(fp)			# test for read access ?
	bneq	userar			# yes
	cmpl	$NBPG,r1		# can we do it in one probe ?
	bgeq	uaw2			# yes
uaw1:
	probew	r2,(r0),$1
	beql	uaerr			# no access
	adda	$NBPG,r0
	_ACBL($NBPG+1,$-NBPG,r1,uaw1)
uaw2:
	probew	r2,(r0),$1		# check first byte
	beql	uaerr
	probew	r2,-1(r0)[r1],$1	# check last byte
	beql	uaerr
	movl	$1,r0
	ret

userar:
	cmpl	$NBPG,r1
	bgeq	uar2
uar1:
	prober	r2,(r0),$1
	beql	uaerr
	adda	$NBPG,r0
	_ACBL($NBPG+1,$-NBPG,r1,uar1)
uar2:
	prober	r2,(r0),$1		# check first byte
	beql	uaerr
	prober	r2,-1(r0)[r1],$1	# check last byte
	beql	uaerr
	movl	$1,r0
	ret
uaerr:
	clrl	r0
	ret

/*
 * if ( kernacc(address, count, mode) ) ....
 * Check address.
 * Given virtual address, byte count, and rw flag
 * returns 0 on no access.
 * Same as useracc routine but checks for kernel access rights.
 */

_kernacc:	.globl	_kernacc
	.word	0x4
	clrl	r2		# r2 = 0 means kernel mode probe.
	jbr	probes		# Dijkstra would get gastric distress here.

/*
 * addupc - increment some histogram counter
 *	in the profiling buffer
 *
 * addupc(pc, prof, counts)
 * long	pc , counts;	Only least significant word of 'counts' is added.
 * struct	uprof *prof;
 * 
 * struct uprof {		# profile arguments 
 * 	short	*r_base;	# buffer base 
 * 	unsigned pr_size;	# buffer size 
 * 	unsigned pr_off;	# pc offset 
 * 	unsigned pr_scale;	# pc scaling 
 * }
 */
	.globl	_addupc
_addupc:
	.word	4
	movl	8(fp),r2	# r2 points to structure
	subl3	8(r2),4(fp),r0	# r0 = PC - lowpc
	jlss	.LAB9		# PC < lowpc , out of range !
	shrl	$1,r0,r0	# the unit is words
	shrl	$1,12(r2),r1	# ditto for scale
	emul	r1,r0,$0,r0
	shrq	$14,r0,r0
	tstl	r0		# too big
	jneq	.LAB9
	cmpl	r1,4(r2)	# Check buffer overflow
	jgequ	.LAB9
	probew	$1,*0(r2)[r1],$1	# first byte of counter accessible?
	jeql	.LAB9
	# We can get away with one probe here since 'profil()' ensured
	# that the address was on an even boundary.
	shrl	$1,r1,r1	# make r1 word index
	addw2	14(fp),*0(r2)[r1]
.LAB9:	ret
/*
*       Save all registers
*/
        .globl  _savreg
_savreg:
        .word   0
	movpsl	pwfl_PSL	# Keeps all flags, etc.
	storer	$0x1fff,pwfl_r0	# Saves r0 - r12
	movl    (fp),pwfl_r0+4*13 # Save fp of caller of savreg (e.g. panic)
	moval	0(sp),pwfl_sp	# Saves sp ( = r14 )
	callf   $4,_savprv      # Saves priviledged registers
	ret
/*
*       Save priviledged registers
*/
        .globl  _savprv
_savprv:
        .word   0
	mfpr	$SBR,pwfl_SBR	# Save all re_loadable registers
	mfpr	$SLR,pwfl_SLR
	mfpr	$P0BR,pwfl_P0BR
	mfpr	$P0LR,pwfl_P0LR
	mfpr	$P1BR,pwfl_P1BR
	mfpr	$P1LR,pwfl_P1LR
	mfpr	$P2BR,pwfl_P2BR
	mfpr	$P2LR,pwfl_P2LR
	mfpr	$IPL,pwfl_IPL
	mfpr	$MME,pwfl_MME  
/*	mfpr	$DCK,pwfl_DCK   DCK and CCK are write only so this won't
	mfpr	$CCK,pwfl_CCK   work. */
	mfpr	$PCBB,pwfl_PCBB
	mfpr	$ISP,pwfl_ISP
	mfpr	$SCBB,pwfl_SCBB
	mfpr	$KSP,pwfl_KSP
	mfpr	$USP,pwfl_USP
        ret
