/*
**	 Common Disk Drive Structure Definitions
*/
#define BPS		512			/* bytes per sector	      */

#define GB_ID		"geometry"
#define GB_ID_LEN 	sizeof(GB_ID)-1
#define GB_MAXPART	0x08
#define GB_VERSION	0x01

/*
**  Partition Definition structure
*/
typedef struct {
	long	start ;      	/* starting 1k block number for partition */
	long	length ; 	/* partition size in 1k blocks */
} par_tab ;

/*
** Geometry Block:
**
** The geometry block defines partition offsets and information
** about the flaw maps on the flaw map track.
** It resides on the first sector of the flaw map track.
** This structure is also used by vddc disk controllers.  In this case,
** the block resides at sector 0 of the disk.
**
** The geometry_sector structure defines the sector containing the geometry
** block.  This sector is checksumed independent of the geometry information.
** The fields in these structured which should never be moved are the id and
** version fields in the geometry_block structure and the checksum field in
** the geometry_sector structure.  This will provide for easy extensions in
** the future.
*/

#define DRIVE_TYPE      flaw_offset     /* For VDDC Geometry Blocks Only      */

typedef struct {
	char id [GB_ID_LEN];		/* identifies the geometry block      */
	long version;			/* geometry block version number      */
	long flaw_offset;		/* flaw map byte offset in partition7 */
	long flaw_size;			/* harris flaw map size in bytes      */
	long flaw_checksum;		/* sum of bytes in harris flaw map    */
	long unused[3];			/* --- available for use              */
	par_tab partition [GB_MAXPART]; /* partition definitions              */
} geometry_block ;

typedef struct {
	geometry_block	geometry_block;	/* disk geometry */
	char		filler [BPS - sizeof(geometry_block) - sizeof(long)] ;
	long           	checksum ;	/* sector checksum	*/
} geometry_sector ;

/*
** GB_CHECKSUM:
**
** This macro computes the checksum for the geometry sector and returns the
** value.  Input to this macro is a pointer to the geometry_sector.
*/
#define GB_CHECKSUM( _gs_ptr, _checksum )  				       \
{									       \
	register unsigned char *	_ptr ;				       \
	register unsigned long		_i, _xsum ;                            \
									       \
	_xsum = 0 ;							       \
	_ptr = (unsigned char *)(_gs_ptr) ;				       \
									       \
	for (_i = 0 ; _i < (sizeof(geometry_sector) - sizeof(long)) ; _i ++) { \
		_xsum += * _ptr ++ ;					       \
	} ;								       \
									       \
	_checksum = _xsum ;						       \
}

#ifdef DSKGENDATA
/*
 * The following data structures are defined for
 * sadp(1). If the maximum number of cylinders (842)
 * changes, CHPERCYL should also change.
 *
 * Also, the assumption is that we have an 8 bit minor
 * number (3 bits used for partition and 5 bit used for
 * unit number). Hence a maximum of 32 drives per system.
 */
#define	CHPERCYL	((842 >> 3) + 1)
#define	MAXDRV		32

int	lastcyl[MAXDRV];
long	dkprf[MAXDRV][CHPERCYL],
	skprf[MAXDRV][CHPERCYL];

/*
 * Hard error is the preface to plaintive error messages
 * about failing disk transfers.
 */
#define harderr(_bp, _cp)                                       \
{                                                               \
	printf("dsk/%ds%d: error bn%d ",                        \
	    dkunit(_bp), (minor(_bp->b_dev)&07), _bp->b_blkno); \
}
#endif DSKGENDATA
