/*	dskio.h	2.2	86/03/06	*/

/*
** Structures and definitions for disk io control commands
*/

/*
** DSK_STATUS Structure:
**
** The dskstatus structure is information about a drive, returned
** by the DSKIOCSTATUS ioctl procedure. Information is obtained from
** the geometry block and the profile PROM.
*/

typedef struct {
	int  id;               	/* identifies the disk model        */
	int  drive_status;	/*                                  */
	int  rpm;		/* disk rpm, normally 3600          */
	int  cylinders;		/* number of logical cylinders      */
	int  heads;		/* number of logical heads          */
	int  sectors;		/* number of logical sectors/track  */
	int  phys_cylinders;	/* number of physical cylinders     */
	int  phys_heads;	/* number of physical heads         */
	int  phys_sectors;	/* number of physical sectors/trk   */
	int  def_cyl;		/* logical cylinder of def area     */
	int  def_cylinders;	/* number of definition cylinders   */
	int  diag_cyl;		/* logical cylinder of diag area    */
	int  diag_cylinders;	/* number of diagnostic cylinders   */
	int  bytes_per_sec;	/* bytes/sector vendor flaw conversn*/
} dsk_status ;


/*
** GEOMETRY_STATUS Structure:
**
** The geometry_status structure is information about partition definitions
** returned by the DSKIOCSTATUS ioctl procedure. Information is obtained from
** the geometry block and the profile PROM.
*/

typedef struct {
	par_tab partition [GB_MAXPART];	/* partition definitions         */
} geometry_status ;

/*
 * These structures define vendor flaw (vflaw) information returned
 * by the DSKIOCVENDOR ioctl, and the harris format flaw (hflaw)
 * information as it exists in the flaw map files (editable).
 */

#define MAXVFLAW	8		/* maximum# vendor flaws reported   */

struct flaw {				/* location of flaws within a track */
	unsigned short	offset;		/* start byte offset of the flaw    */
	unsigned short	length;		/* number of bits in the flaw       */
} flaw;

typedef struct {			/* defines which tracks have flaws  */
	int		cylinder;	/* cylinder number of the track     */
	int		head;		/* head number of the track         */
	struct flaw	flaw[MAXVFLAW];	/* a list of flaws on the track     */
} vflaw_type;

typedef struct {			/* a harris format flaw map entry   */
	char		text[132];	/* each hflaw entry is max 132 chars*/
} hflaw_type;

/*
 * The formatop structure is passed for the DSKIOCFORMAT ioctl;
 * it defines a track to be formatted.
 * The verifyop structure is passed for the DSKIOCVERIFY ioctl;
 * it defines a track to be verified.
 */

struct formatop {
	int		cylinder;	/* physical cyl number of the track */
	int		head;		/* physical head number of the track*/
	int		flaw_count;	/* number of flaws on the track     */
	int		flaw[MAXVFLAW];	/* relative sector of the flaws  */
};

struct verifyop {
	int		cylinder;	/* physical cyl number of the track */
	int		head;		/* physical head number of the track*/
	int		err_sector;	/* returned bad sector# if error    */
};

/*
** Disk io control commands
*/

#ifndef _IOCTL_
#ifdef  KERNEL
#include "../sys/ioctlp.h"
#else
#include <sys/ioctlp.h>
#endif
#endif

#define	DSKIOCFORMAT	  _IOW(d, 1, struct formatop)	/* Format disk track  */
#define	DSKIOCFORMATCTL	  _IO(d, 2)            		/* Format control     */
#define	DSKIOCSTATUS	  _IOR(d, 3, dsk_status)	/* Disk status        */
#define	DSKIOCVENDORFLAW  _IOR(d, 4, vflaw_type)	/* Vendor Flaw info   */
#define DSKIOCGEOMETRY    _IOR(d, 5, geometry_status) 	/* get geometries     */
#define DSKIOCSETGEOMETRY _IOW(d, 6, geometry_status) 	/* set geometries     */
#define	DSKIOCVERIFY	  _IOW(d, 7, struct verifyop)	/* Verify disk track  */
