/*	ioctl.h	6.1	83/07/29	*/
/*
 * Ioctl definitions
 */

#ifndef	_IOCTL_

#ifndef _IOCTLP_
#ifdef  KERNEL
#include "../sys/ioctlp.h"
#else
#include <sys/ioctlp.h>         /* ioctl parameter definitions */
#endif  KERNEL
#endif _IOCTLP_

/*
 *      ioctl.h should be included only for programs compiled in the "ucb
 *      universe".  This file is equivalent to an empty file for programs
 *      compiled in the "att universe".  The kernel needs the file included
 *      independent of the universe switch.
 */
#if defined(KERNEL) || defined(ucbhcx)
#define	_IOCTL_

#ifdef  KERNEL
#include "../sys/ttychars.h"
#include "../sys/ttydev.h"
#else
#include <sys/ttychars.h>
#include <sys/ttydev.h>
#endif

/*
 * Structure for TIOCGETP and TIOCSETP ioctls.
 */

struct sgttyb {
	char	sg_ispeed;		/* input speed */
	char	sg_ospeed;		/* output speed */
	char	sg_erase;		/* erase character */
	char	sg_kill;		/* kill character */
	int  	sg_flags;		/* mode flags */
};

struct tchars {
	char	t_intrc;	/* interrupt */
	char	t_quitc;	/* quit */
	char	t_startc;	/* start output */
	char	t_stopc;	/* stop output */
	char	t_eofc;		/* end-of-file */
	char	t_brkc;		/* input delimiter (like nl) */
};
struct ltchars {
	char	t_suspc;	/* stop process signal */
	char	t_dsuspc;	/* delayed stop process signal */
	char	t_rprntc;	/* reprint line */
	char	t_flushc;	/* flush output (toggles) */
	char	t_werasc;	/* word erase */
	char	t_lnextc;	/* literal next character */
};
/*
 * tty ioctl commands
 */
#define	TIOCGETD	_IOR(t, 0, int)		/* get line discipline */
#define	TIOCSETD	_IOW(t, 1, int)		/* set line discipline */
#define	TIOCHPCL	_IO(t, 2)		/* hang up on last close */

/*
 * TIOCMODG and TIOCMODS are not supported in HCX/UX 2.2
 */

#if     0
#define	TIOCMODG	_IOR(t, 3, int)		/* get modem control state */
#define	TIOCMODS	_IOW(t, 4, int)		/* set modem control state */
#define		TIOCM_LE	0001		/* line enable */
#define		TIOCM_DTR	0002		/* data terminal ready */
#define		TIOCM_RTS	0004		/* request to send */
#define		TIOCM_ST	0010		/* secondary transmit */
#define		TIOCM_SR	0020		/* secondary receive */
#define		TIOCM_CTS	0040		/* clear to send */
#define		TIOCM_CAR	0100		/* carrier detect */
#define		TIOCM_CD	TIOCM_CAR
#define		TIOCM_RNG	0200		/* ring */
#define		TIOCM_RI	TIOCM_RNG
#define		TIOCM_DSR	0400		/* data set ready */
#endif

#define	TIOCGETP	_IOR(t, 8,struct sgttyb)/* get parameters -- gtty */
#define	TIOCSETP	_IOW(t, 9,struct sgttyb)/* set parameters -- stty */
#define	TIOCSETN	_IOW(t,10,struct sgttyb)/* as above, but no flushtty */
#define	TIOCEXCL	_IO(t, 13)		/* set exclusive use of tty */
#define	TIOCNXCL	_IO(t, 14)		/* reset exclusive use of tty */
#define	TIOCFLUSH	_IOW(t, 16, int)	/* flush buffers */
#define	TIOCSETC	_IOW(t,17,struct tchars)/* set special characters */
#define	TIOCGETC	_IOR(t,18,struct tchars)/* get special characters */
#define		TANDEM		0x00000001	/* send stopc on out q full */
#define		CBREAK		0x00000002	/* half-cooked mode */
#define		LCASE		0x00000004	/* simulate lower case */
#ifndef _TERMIO_                                /* avoid redefining ECHO */
#define		ECHO		0x00000008	/* echo input */
#endif  _TERMIO_
#define		CRMOD		0x00000010	/* map \r to \r\n on output */
#define		RAW		0x00000020	/* no i/o processing */
#define		ODDP		0x00000040	/* get/send odd parity */
#define		EVENP		0x00000080	/* get/send even parity */
#define		ANYP		0x000000c0	/* get any parity/send none */
#define		NLDELAY		0x00000300	/* \n delay */
#define			O_NL0	0x00000000
#define			O_NL1	0x00000100	/* tty 37 */
#define			O_NL2	0x00000200	/* vt05 */
#define			O_NL3	0x00000300
#define		TBDELAY		0x00000c00	/* horizontal tab delay */
#define			O_TAB0	0x00000000
#define			O_TAB1	0x00000400	/* tty 37 */
#define			O_TAB2	0x00000800
#define		XTABS		0x00000c00	/* expand tabs on output */
#define		CRDELAY		0x00003000	/* \r delay */
#define			O_CR0	0x00000000
#define			O_CR1	0x00001000	/* tn 300 */
#define			O_CR2	0x00002000	/* tty 37 */
#define			O_CR3	0x00003000	/* concept 100 */
#define		VTDELAY		0x00004000	/* vertical tab delay */
#define			O_FF0	0x00000000
#define			O_FF1	0x00004000	/* tty 37 */
#define		BSDELAY		0x00008000	/* \b delay */
#define			O_BS0	0x00000000
#define			O_BS1	0x00008000
#define 	ALLDELAY	(NLDELAY|TBDELAY|CRDELAY|VTDELAY|BSDELAY)
/*
 *      The following are kept for compatibility with old stty and gtty calls.
 */
#define         O_HUPCL         0x00010000
#define         O_NOAL          0x00020000
/*
 *      The following definitions should only be defined for user programs.
 *      The kernel cannot include these definitions because of conflicts
 *      with definitions in <sys/termio.h>.
 */
#if !defined(KERNEL) && !defined(_TERMIO_)
#define			NL0	O_NL0
#define			NL1	O_NL1
#define			NL2	O_NL2
#define			NL3	O_NL3
#define			TAB0	O_TAB0
#define			TAB1	O_TAB1
#define			TAB2	O_TAB2
#define			CR0	O_CR0
#define			CR1	O_CR1
#define			CR2	O_CR2
#define			CR3	O_CR3
#define			FF0	O_FF0
#define			FF1	O_FF1
#define			BS0	O_BS0
#define			BS1	O_BS1
#define         NOFLSH          0x00008000      /* no output flush on signal */
#endif
/*
 *      The local modes are stored in a separate short variable in HCX/UX
 *      instead in the upper 2 bytes of an integer.  Thus, the following
 *      definitions are shifted 16 bits to the right of the corresponding
 *      BSD 4.2 definitions.
 */
#define		CRTBS		0x00000001	/* do backspacing for crt */
#define		PRTERA		0x00000002	/* \ ... / erase */
#define		CRTERA		0x00000004	/* " \b " to wipe out char */
#define		TILDE		0x00000008	/* hazeltine tilde kludge */
#define		MDMBUF		0x00000010	/* start/stop output on carrier intr */
#define		LITOUT		0x00000020	/* literal output */
#define		TOSTOP		0x00000040	/* SIGSTOP on background output */
#define		FLUSHO		0x00000080	/* flush output to terminal */
#define		NOHANG		0x00000100	/* no SIGHUP on carrier drop */
#define         L001000         0x00000200
#define		CRTKIL		0x00000400	/* kill line with " \b " */
#define         PASS8           0x00000800
#define		CTLECH		0x00001000	/* echo control chars as ^X */
#define		PENDIN		0x00002000	/* tp->t_rawq needs reread */
#define		DECCTQ		0x00004000	/* only ^Q starts after ^S */
/* locals, from 127 down */
#define	TIOCLBIS	_IOW(t, 127, int)	/* bis local mode bits */
#define	TIOCLBIC	_IOW(t, 126, int)	/* bic local mode bits */
#define	TIOCLSET	_IOW(t, 125, int)	/* set entire local mode word */
#define	TIOCLGET	_IOR(t, 124, int)	/* get local modes */
/*
 *      Since the definitions are for the local modes above are already
 *      shifted over 16 bits, the following definitions are defined to
 *      be equivalent and not shifted 16 bits to the right as they are
 *      in BSD 4.2.  All applications should use the following definitions
 *      when modifying the local modes.
 */
#define		LCRTBS		CRTBS
#define		LPRTERA		PRTERA
#define		LCRTERA		CRTERA
#define		LTILDE		TILDE
#define		LMDMBUF		MDMBUF
#define		LLITOUT		LITOUT
#define		LTOSTOP		TOSTOP
#define		LFLUSHO		FLUSHO
#define		LNOHANG		NOHANG
#define		LCRTKIL		CRTKIL
#define         LPASS8          PASS8
#define		LCTLECH		CTLECH
#define		LPENDIN		PENDIN
#define		LDECCTQ		DECCTQ
#define		LNOFLSH		0x00008000

/*
 *  TIOCSBRK, TIOCCBRK, TIOCSDTR, and TIOCCDTR are not supported
 *  in HCX/UX 2.2
 */

#if     0
#define	TIOCSBRK	_IO(t, 123)		/* set break bit */
#define	TIOCCBRK	_IO(t, 122)		/* clear break bit */
#define	TIOCSDTR	_IO(t, 121)		/* set data terminal ready */
#define	TIOCCDTR	_IO(t, 120)		/* clear data terminal ready */
#endif

#define	TIOCGPGRP	_IOR(t, 119, int)	/* get pgrp of tty */
#define	TIOCSPGRP	_IOW(t, 118, int)	/* set pgrp of tty */
#define	TIOCSLTC	_IOW(t,117,struct ltchars)/* set local special chars */
#define	TIOCGLTC	_IOR(t,116,struct ltchars)/* get local special chars */
#define	TIOCOUTQ	_IOR(t, 115, int)	/* output queue size */
#define	TIOCSTI		_IOW(t, 114, char)	/* simulate terminal input */
#define	TIOCNOTTY	_IO(t, 113)		/* void tty association */
#define	TIOCPKT		_IOW(t, 112, int)	/* pty: set/clear packet mode */
#define		TIOCPKT_DATA		0x00	/* data packet */
#define		TIOCPKT_FLUSHREAD	0x01	/* flush packet */
#define		TIOCPKT_FLUSHWRITE	0x02	/* flush packet */
#define		TIOCPKT_STOP		0x04	/* stop output */
#define		TIOCPKT_START		0x08	/* start output */
#define		TIOCPKT_NOSTOP		0x10	/* no more ^S, ^Q */
#define		TIOCPKT_DOSTOP		0x20	/* now do ^S ^Q */
#define	TIOCSTOP	_IO(t, 111)		/* stop output, like ^S */
#define	TIOCSTART	_IO(t, 110)		/* start output, like ^Q */
#define	TIOCMSET	_IOW(t, 109, int)	/* set all modem bits */
#define	TIOCMBIS	_IOW(t, 108, int)	/* bis modem bits */
#define	TIOCMBIC	_IOW(t, 107, int)	/* bic modem bits */
#define	TIOCMGET	_IOR(t, 106, int)	/* get all modem bits */
#define	TIOCREMOTE	_IO(t, 105)		/* remote input editing */

#define	OTTYDISC	0		/* old, v7 std tty driver */
#define	NETLDISC	1		/* line discip for berk net */
#define	NTTYDISC	2		/* new tty discipline */
#define NEWTTY          0x00008000

#define	FIOCLEX		_IO(f, 1)		/* set exclusive use on fd */
#define	FIONCLEX	_IO(f, 2)		/* remove exclusive use */
/* another local */
#define	FIONREAD	_IOR(f, 127, int)	/* get # bytes to read */
#define	FIONBIO		_IOW(f, 126, int)	/* set/clear non-blocking i/o */
#define	FIOASYNC	_IOW(f, 125, int)	/* set/clear async i/o */
#define	FIOSETOWN	_IOW(f, 124, int)	/* set owner */
#define	FIOGETOWN	_IOR(f, 123, int)	/* get owner */

#endif  defined(KERNEL) || defined(ucbhcx)

/*
 * Allow programs compiled in the 'att' universe to still have access
 * to the socket i/o controls.
 */

#define	SIOCSHIWAT	_IOW(s,  0, int)		/* set high watermark */
#define	SIOCGHIWAT	_IOR(s,  1, int)		/* get high watermark */
#define	SIOCSLOWAT	_IOW(s,  2, int)		/* set low watermark */
#define	SIOCGLOWAT	_IOR(s,  3, int)		/* get low watermark */
#define	SIOCATMARK	_IOR(s,  7, int)		/* at oob mark? */
#define	SIOCSPGRP	_IOW(s,  8, int)		/* set process group */
#define	SIOCGPGRP	_IOR(s,  9, int)		/* get process group */

#define	SIOCADDRT	_IOW(r, 10, struct rtentry)	/* add route */
#define	SIOCDELRT	_IOW(r, 11, struct rtentry)	/* delete route */

#define	SIOCSIFADDR	_IOW(i, 12, struct ifreq)	/* set ifnet address */
#define	SIOCGIFADDR	_IOWR(i,13, struct ifreq)	/* get ifnet address */
#define	SIOCSIFDSTADDR	_IOW(i, 14, struct ifreq)	/* set p-p address */
#define	SIOCGIFDSTADDR	_IOWR(i,15, struct ifreq)	/* get p-p address */
#define	SIOCSIFFLAGS	_IOW(i, 16, struct ifreq)	/* set ifnet flags */
#define	SIOCGIFFLAGS	_IOWR(i,17, struct ifreq)	/* get ifnet flags */
#define	SIOCGIFCONF	_IOWR(i,20, struct ifconf)	/* get ifnet list */
#define	SIOCIFDETACH	_IOWR(i,21, struct ifconf)	/* detach interface */
#define	SIOCSETETADDR	_IOWR(i,30, struct ifreq)  /* Set E.T. station addr.*/

/*---- serial line commands ----*/
#define SLIOC		('S'<<8)
#define SLIOCUNIT	_IOR(S, 1, int)		/* get serial line unit # */

#endif  _IOCTL_
