/* @(#)termio.h	6.3 */
#ifndef _TERMIO_
#define _TERMIO_

#define	NCC	8

#ifndef _IOCTLP_
#ifdef  KERNEL
#include "../sys/ioctlp.h"
#else
#include <sys/ioctlp.h>         /* ioctl parameter definitions */
#endif  KERNEL
#endif _IOCTLP_

/* control characters */
#define	VINTR	0
#define	VQUIT	1
#define	VERASE	2
#define	VKILL	3
#define	VEOF	4
#define	VEOL	5
#define	VEOL2	6
#define	VBRK	VEOL2
#define	VMIN	VEOF
#define	VTIME	VEOL
#define	VSWTCH	7
#define	VSUSP	8
#define	VDSUSP	9
#define	VRPRNT	10
#define	VFLUSH	11
#define	VWERASE	12
#define	VLNEXT	13
#define VSTART  14
#define VSTOP   15

/* default control chars */
#define	CESC	'\\'
#define	CNUL	0
#define	CDEL	0377

/*   If _TTYCHARS_ is defined, let its default values take precedence */

#ifndef _TTYCHARS_
#define	CINTR	0177	/* DEL */
#define	CERASE	'#'
#define	CKILL	'@'
#define	CQUIT	034	/* FS, cntl | */
#define	CEOF	04	/* cntl d */
#define	CEOT	CEOF
#define	CBRK	0377
#define	CSTART	021	/* cntl q */
#define	CSTOP	023	/* cntl s */
#endif

#define CEOL	CNUL
#define	CSWTCH	032	/* cntl z */
#define	CNSWTCH	CNUL

/* input modes */
#define	IGNBRK	0000001
#define	BRKINT	0000002
#define	IGNPAR	0000004
#define	PARMRK	0000010
#define	INPCK	0000020
#define	ISTRIP	0000040
#define	INLCR	0000100
#define	IGNCR	0000200
#define	ICRNL	0000400
#define	IUCLC	0001000
#define	IXON	0002000
#define	IXANY	0004000
#define	IXOFF	0010000

/* output modes */
#define	OPOST	0000001
#define	OLCUC	0000002
#define	ONLCR	0000004
#define	OCRNL	0000010
#define	ONOCR	0000020
#define	ONLRET	0000040
#define	OFILL	0000100
#define	OFDEL	0000200
/*
 *      The following definitions conflict with some ucb definitinos.
 *      if sgtty is already defined, let it take precedecnt over these
 *      definitions.  Note:  if this is the case, these delay settings
 *      need to be accessed using the "ucb" ioctl function codes.
 */
#ifndef _IOCTL_
#define	NLDLY	0000400
#define	NL0	0
#define	NL1	0000400
#define	CRDLY	0003000
#define	CR0	0
#define	CR1	0001000
#define	CR2	0002000
#define	CR3	0003000
#define	TABDLY	0014000
#define	TAB0	0
#define	TAB1	0004000
#define	TAB2	0010000
#define	TAB3	0014000
#define	BSDLY	0020000
#define	BS0	0
#define	BS1	0020000
#define	VTDLY	0040000
#define	VT0	0
#define	VT1	0040000
#define	FFDLY	0100000
#define	FF0	0
#define	FF1	0100000
#endif  _IOCTL_

/* control modes */
#define	CBAUD	0000017

#ifndef	_TTYDEV_
#ifdef  KERNEL
#include "../sys/ttydev.h"
#else
#include <sys/ttydev.h>         /* Baud rate definitions */
#endif  KERNEL
#endif  _TTYDEV_

#define	CSIZE	0000060
#define	CS5	0
#define	CS6	0000020
#define	CS7	0000040
#define	CS8	0000060
#define	CSTOPB	0000100
#define	CREAD	0000200
#define	PARENB	0000400
#define	PARODD	0001000
#define	HUPCL	0002000
#define	CLOCAL	0004000
#define	LOBLK	0010000

/* line discipline 0 modes */
#define	ISIG	0000001
#define	ICANON	0000002
#define	XCASE	0000004
#define	ECHO	0000010
#define	ECHOE	0000020
#define	ECHOK	0000040
#define	ECHONL	0000100
#define	NOFLSH	0000200
#define	IBRK	0000400

#define	SSPEED	015	/* default speed: 9600 baud */

/*
 * Ioctl control packet
 */
struct termio {
	unsigned short	c_iflag;	/* input modes */
	unsigned short	c_oflag;	/* output modes */
	unsigned short	c_cflag;	/* control modes */
	unsigned short	c_lflag;	/* line discipline modes */
	char	c_line;			/* line discipline */
	unsigned char	c_cc[NCC];	/* control chars */
};

#define	TIOC	('T'<<8)
#define	TCGETA	_IOR(T, 1, struct termio)
#define	TCSETA	_IOW(T, 2, struct termio)
#define	TCSETAW	_IOW(T, 3, struct termio)
#define	TCSETAF	_IOW(T, 4, struct termio)
#define	TCSBRK	_IO(T, 5)
#define	TCXONC	_IO(T, 6)
#define	TCFLSH	_IO(T, 7)
#define	TCDSET	_IO(T, 8)

/* used internally */
#define	LDIOC	('D'<<8)
#define	LDOPEN	_IO(D,0)
#define	LDCLOSE	_IO(D,1)
#define	LDCHG	_IO(D,2)
#define	LDGETT	_IO(D,8)
#define	LDSETT	_IO(D,9)

/*
 * Terminal types
 */
#define	TERM_NONE	0	/* tty */
#define	TERM_TEC	1	/* TEC Scope */
#define	TERM_V61	2	/* DEC VT61 */
#define	TERM_V10	3	/* DEC VT100 */
#define	TERM_TEX	4	/* Tektronix 4023 */
#define	TERM_D40	5	/* TTY Mod 40/1 */
#define	TERM_H45	6	/* Hewlitt-Packard 45 */
#define	TERM_D42	7	/* TTY Mod 40/2B */

/*
 * Terminal flags
 */
#define TM_NONE		0000	/* use default flags */
#define TM_SNL		0001	/* special newline flag */
#define TM_ANL		0002	/* auto newline on column 80 */
#define TM_LCF		0004	/* last col of last row special */
#define TM_CECHO	0010	/* echo terminal cursor control */
#define TM_CINVIS	0020	/* do not send esc seq to user */
#define TM_SET		0200	/* must be on to set/res flags */

/*
 * structure of ioctl arg for LDGETT and LDSETT
 */
struct	termcb	{
	char	st_flgs;	/* term flags */
	char	st_termt;	/* term type */
	char	st_crow;	/* gtty only - current row */
	char	st_ccol;	/* gtty only - current col */
	char	st_vrow;	/* variable row */
	char	st_lrow;	/* last row */
};

#endif  _TERMIO_
