/* @(#)tty.h	6.2 */

/*
 * A clist structure is the head of a linked list queue of characters.
 * The routines getc* and putc* manipulate these structures.
 */
struct clist {
	int	c_cc;		/* character count */
	struct cblock *c_cf;	/* pointer to first */
	struct cblock *c_cl;	/* pointer to last */
};

#ifdef KERNEL
int	nclist;			/* number of clist blocks allocated */
#endif

/* Macro to find clist structure given pointer into it	*/
#define CMATCH(pointer)		(cfree + (pointer - cfree))

/*
 * This structure is used to walk through a clist without taking
 * any characters off it.
 */
struct cur_clist {
	struct cblock *cr_cb;	/* points to the current cblock in chain */
	int	cr_ndx;		/* index of the current character */
};

/* Character control block for interrupt level control	*/

struct ccblock {
	caddr_t	c_ptr;		/* buffer address	*/
	ushort	c_count;	/* character count	*/
	ushort	c_size;		/* buffer size		*/
};

/*
 * A tty structure is needed for each UNIX character device that
 * is used for normal terminal IO.
 */
#define	NCC	8
#define	NXCC	16      /* number of control chars including 4.2 extensions */
struct tty {
	struct	clist t_rawq;	/* raw input queue */
	struct	clist t_canq;	/* canonical queue */
	struct	clist t_outq;	/* output queue */
	struct	ccblock	t_tbuf;	/* tx control block */
	struct	ccblock t_rbuf;	/* rx control block */
	int	(* t_proc)();	/* routine for device functions */
	struct	proc *t_rsel;	/* wake up when have received data */
	struct	proc *t_wsel;	/* wake up when can accept data */
	char	*t_slsoftc;	/* pointer to IP serial line tables */
	ushort	t_iflag;	/* input modes */
	ushort	t_oflag;	/* output modes */
	ushort	t_cflag;	/* control modes */
	ushort	t_lflag;	/* line discipline modes */
	ushort	t_xflag;	/* extra modes for new tty driver */
	long	t_state;	/* internal state */
	short	t_pgrp;		/* process group name */
	dev_t	t_dev;		/* major/minor for this tty */
	char	t_line;		/* line discipline */
	char	t_delct;	/* delimiter count */
	char	t_term;		/* terminal type */
	char	t_tmflag;	/* terminal flags */
	char	t_col;		/* current column */
	char	t_row;		/* current row */
	char	t_vrow;		/* variable row */
	char	t_lrow;		/* last physical row */
	struct cur_clist t_beglin;	/* points into beginning of current line in t_rawq */
	char	t_lncount;	/* number of characters in current input line */
	char	t_rocount;	/* number of characters in current input line since last write */
	char	t_rocol;	/* column that current input line began at */
	char	t_dscnt;	/* number of delayed suspend characters in input queue */
	char	t_hqcnt;	/* no. high queue packets on t_outq */	
	char	t_dstat;	/* used by terminal handlers
						and line disciplines */
	unsigned char	t_cc[NXCC];	/* settable control chars */
};

/*
 * The structure of a clist block
 */
#define	CBSIZE	128		/* number of chars in a clist block */
struct cblock {
	struct cblock *c_next;
	short	c_first;
	short	c_last;
	char	c_data[CBSIZE];
};

#ifdef KERNEL
extern struct cblock *cfree;
extern struct cblock * getcb();
extern struct cblock * getcf();
extern struct clist ttnulq;
#endif

struct chead {
	struct cblock *c_next;
	int	c_size;
	int	c_flag;
};

#ifdef KERNEL
extern struct chead cfreelist;
#endif

struct inter {
	int	cnt;
};

#define	QESC	0200	/* queue escape */
#define	HQEND	01	/* high queue end */

#define	TTIPRI	28
#define	TTOPRI	29

/* limits */
#ifdef KERNEL
extern int ttlowat[], tthiwat[];
#endif
#define	TTYHOG	256
#define	TTXOLO	132
#define	TTXOHI	180
#define	TTECHI	80

/*
 * Hardware bits.
 */
#define	DONE	0200
#define	IENABLE	0100
#define	OVERRUN	040000
#define	FRERROR	020000
#define	PERROR	010000

/*
 * Modem control commands.
 */
#define	DMSET		0
#define	DMBIS		1
#define	DMBIC		2
#define	DMGET		3

/* Internal state */
#define	TIMEOUT	01		/* Delay timeout in progress */
#define	WOPEN	02		/* Waiting for open to complete */
#define	ISOPEN	04		/* Device is open */
#define	TBLOCK	010		/* Input stopped by ctl-s */
#define	CARR_ON	020		/* Software copy of carrier-present */
#define	BUSY	040		/* Output in progress */
#define	OASLP	0100		/* Wakeup when output done */
#define	IASLP	0200		/* Wakeup when input done */
#define	TTSTOP	0400		/* Output stopped by ctl-s */
#define	EXTPROC	01000		/* External processing */
#define	TACT	02000		/* t_cc[VTIME] timeout active */
#define	CLESC	04000		/* Last char was escape (\) */
#define	RTO	010000		/* Raw Timeout */
#define	TTIOW	020000		/* Waiting for output to drain */
#define	TTXON	040000		/* Should send ctl-q ASAP */
#define	TTXOFF	0100000		/* Should send ctl-s ASAP */
#define	RCOLL	0200000		/* multiple processes waiting on data */
#define	WCOLL	0400000		/* multiple processes wanting to send data */
#define	NTBUF	01000000	/* transmitting something not from t_tbuf */
#define	ERASE	02000000	/* Within a \.../ for PRTRUB */
#define	LNCH	04000000	/* Next character is literal */
#define	NBIO	010000000	/* Tty in non-blocking mode (tty, not file descriptor!) */
#define	ASYNC	020000000	/* Tty in async i/o mode */
#define	XCLUDE  040000000       /* Exclusive use flag against open */
#define	OPROC   0100000000      /* Set when pty master device is open */
#define DTRON	0200000000      /* Set when DTR has been raised */
#define DTRTOUT 0400000000      /* Timeout flag for DTR raise or drop */

#define	LOCAL	(CLESC|ERASE|LNCH)

/* l_output status */
#define	CPRES	0100000

/* device commands */
#define	T_OUTPUT	0
#define	T_TIME		1
#define	T_SUSPEND	2
#define	T_RESUME	3
#define	T_BLOCK		4
#define	T_UNBLOCK	5
#define	T_RFLUSH	6
#define	T_WFLUSH	7
#define	T_BREAK		8
#define	T_INPUT		9
#define	T_PARM		11
#define	T_SWTCH		12

/*
 * Terminal flags (set in t_tmflgs).
 */

#define SNL	1		/* non-standard new-line needed */
#define ANL	2		/* automatic new-line */
#define LCF	4		/* Special treatment of last col, row */
#define TERM_CTLECHO	010	/* Echo terminal control characters */
#define TERM_INVIS	020	/* do not send escape sequences to user */
#define QLOCKB		040	/* high queue locked for base level */
#define QLOCKI		0100	/* high queue locked for interrupts */
#define	TERM_BIT 0200		/* Bit reserved for terminal drivers. */
				/* Usually used to indicate that an esc*/
				/* character has arrived and that the  */
				/* next character is special.          */
				/* This bit is the same as the TM_SET  */
				/* bit which may never be set by a user*/
/*
 *	device report
 */
#define	L_BUF	0
#define	L_BREAK	3

/*
 * User visible structures and constants
 * related to terminal line disciplines.
 */

/*-------------- don't update these values without changing linesw.c----*/
#define	TTYDISC		0		/* Terminal line discipline */
/* REMEMBER: line discipline 2 is reserved for Berkeley line discipline */
#define	SLIPDISC	3		/* Serial Line IP discipline */
