/*	user.h	6.1	83/07/29	*/

#ifdef KERNEL
#include "../ml/pcb.h"
#include "../sys/dmap.h"
#include "../sys/time.h"
#include "../sys/resource.h"
#include "../sys/errno.h"
#else
#include <errno.h>
#include <ml/pcb.h>
#include <sys/dmap.h>
#include <sys/time.h>
#include <sys/resource.h>
#endif

/*
 * Per process structure containing data that
 * isn't needed in core when the process is swapped out.
 * Documentation Block and Change history :
 *     bernie 4/28/85  Add audittrail structures
 */
 
#define	SHSIZE		32
#define	MAXCOMLEN	16		/* <= MAXNAMLEN, >= sizeof(a_comm) */

struct auddir {
	char * aud_buf ;                /* Pointer to buffer */
	int aud_len ;                   /* Length of data including 0 byte */
	/* allocated buffer is 'auddirlen(aud_len)' bytes long */
} ;

/* Audit directory buffer is alway a multiple of 32 bytes long */
#define auddirlen(x)     (((x)+63) & ~63)
 
struct	user {
	struct	pcb u_pcb;
	struct	proc *u_procp;		/* pointer to proc structure */
	int	*u_ar0;			/* address of users saved R0 */
	char	u_comm[MAXCOMLEN + 1];

	char	u_univ[MAXUNIVLEN];	/* current universe */
	short	u_univlen;		/* length of universe w/null byte */

/* syscall parameters, results and catches */
	int	u_arg[8];		/* arguments to current system call */
	int	*u_ap;			/* pointer to arglist */
	label_t	u_qsave;		/* for non-local gotos on interrupts */
	char	u_error;		/* return error code */
	union {				/* syscall return values */
		int	r_val1;
		off_t	r_off;
		time_t	r_time;
	} u_r;
	char	u_eosys;		/* special action on end of syscall */
	char	u_madesyscalls;		/* 0 iff no system calls by process */

/* 1.1 - processes and protection */
	struct ucred *u_cred;		/* user credentials (uid, gid, etc) */
#define	u_uid	u_cred->cr_uid		/* effective user id */
#define	u_gid	u_cred->cr_gid		/* effective group id */
#define	u_groups u_cred->cr_groups	/* groups, 0 terminated */
#define	u_ruid	u_cred->cr_ruid		/* real user id */
#define	u_rgid	u_cred->cr_rgid		/* real group id */

/* 1.2 - memory management */
	size_t	u_tsize;		/* text size (clicks) */
	size_t	u_dsize;		/* data size (clicks) */
	size_t	u_ssize;		/* stack size (clicks) */
	struct	dmap u_dmap;		/* disk map for data segment */
	struct	dmap u_smap;		/* disk map for stack segment */
	struct	dmap u_cdmap, u_csmap;	/* shadows of u_dmap, u_smap, for
					   use of parent during fork */
	label_t u_ssave;		/* label variable for swapping */
	size_t	u_odsize, u_ossize;	/* for (clumsy) expansion swaps */
	time_t	u_outime;		/* user time at last sample */

/* 1.3 - signal management */
	int	(*u_signal[NSIG])();	/* disposition of signals */
	int	u_sigmask[NSIG];	/* signals to be blocked */
	int	u_sigonstack;		/* signals to take on sigstack */
	int	u_oldmask;		/* saved mask from before sigpause */
	int	u_code;			/* ``code'' to trap */
	struct	sigstack u_sigstack;	/* sp & on stack state variable */
#define	u_onstack	u_sigstack.ss_onstack
#define	u_sigsp		u_sigstack.ss_sp
	int	u_saveacc[3];		/* saves accumulator on signals */

/* 1.4 - descriptor management */
	struct	file *u_ofile[NOFILE];	/* file structures for open files */
	char	u_pofile[NOFILE];	/* per-process flags of open files */
	int	u_lastfile;		/* high-water mark of u_ofile */
#define	UF_EXCLOSE 	0x1		/* auto-close on exec */
#define	UF_MAPPED 	0x2		/* mapped from device */
	struct	vnode *u_cdir;		/* current directory */
	struct	vnode *u_rdir;		/* root directory of current process */
	struct	tty *u_ttyp;		/* controlling tty pointer */
	dev_t	u_ttyd;			/* controlling tty dev */
	short	u_cmask;		/* mask for file creation */

/* 1.5 - timing and statistics */
	struct	rusage u_ru;		/* stats for this proc */
	struct	rusage u_cru;		/* sum of stats for reaped children */
	struct	itimerval u_timer[3];
	int	u_XXX[3];
	struct	timeval u_start;
	short	u_acflag;

/* 1.6 - resource controls */
	struct	rlimit u_rlimit[RLIM_NLIMITS];
	struct	quota *u_quota;		/* user's quota structure */
	int	u_qflags;		/* per process quota flags */

/* BEGIN TRASH */
	char	u_segflg;		/* 0:user D; 1:system; 2:user I */
	caddr_t	u_base;			/* base address for IO */
	unsigned int u_count;		/* bytes remaining for IO */
	off_t	u_offset;		/* offset in file for IO */
	unsigned int u_fmode;		/* f_flag of current file */
	union {
	   struct {			/* header of executable file */
		short	Ux_mag;		/* magic number */
		short	Ux_nscns;	/* number of sections */
		unsigned Ux_timdat;	/* creation time */
		unsigned Ux_symptr;	/* pointer to symbol table */
		unsigned Ux_nsyms;	/* number of symbols */
		short	Ux_opthdr;	/* size of optional header */
		short	Ux_flags;	/* flags */
		short	Aux_mag;	/* magic number */
		short	Ux_stamp;	/* version stamp */
		unsigned Ux_tsize;	/* text size */
		unsigned Ux_dsize;	/* data size */
		unsigned Ux_bsize;	/* bss size */
		unsigned Ux_entloc;	/* entry location */
		unsigned Ux_textstart;	/* text start location */
		unsigned Ux_datastart;	/* data start location */
	   } Ux_A;
	   char ux_shell[SHSIZE];	/* #! and name of interpreter */
	} u_exdata;
#define	aux_mag		Ux_A.Aux_mag
#define	ux_mag		Ux_A.Ux_mag
#define	ux_nscns	Ux_A.Ux_nscns
#define	ux_stamp	Ux_A.Ux_stamp
#define	ux_tsize	Ux_A.Ux_tsize
#define	ux_dsize	Ux_A.Ux_dsize
#define	ux_bsize	Ux_A.Ux_bsize
#define	ux_ssize	Ux_A.Ux_ssize
#define	ux_entloc	Ux_A.Ux_entloc
#define	ux_tstart	Ux_A.Ux_textstart
#define	ux_dstart	Ux_A.Ux_datastart
/* END TRASH */
	struct uprof {			/* profile arguments */
		short	*pr_base;	/* buffer base */
		unsigned pr_size;	/* buffer size */
		unsigned pr_off;	/* pc offset */
		unsigned pr_scale;	/* pc scaling */
	} u_prof;
/* begin Sys 5 stuff */
	daddr_t	u_limit;		/* max. write address for NBPG = 512 */
	short	u_lock;			/* process/text/data locking flags */
/* end Sys 5 stuff */
  	struct auddir	u_audrootdir;	/* pathname of current rootdir */
  	struct auddir	u_audcwdir;	/* pathname of current directory */
	int	u_stack[1];
	struct crcb *u_crcbp;
};

struct ucred {
	u_short	cr_ref;			/* reference count */
	short   cr_uid;			/* effective user id */
	short   cr_gid;			/* effective group id */
	int     cr_groups[NGROUPS];	/* groups, 0 terminated */
	short   cr_ruid;		/* real user id */
	short   cr_rgid;		/* real group id */
};
#ifdef KERNEL
struct ucred *crget();
struct ucred *crcopy();
struct ucred *crdup();
#endif

/* u_eosys values */
#define	JUSTRETURN	0
#define	RESTARTSYS	1

#ifdef KERNEL
extern	struct user u;
extern	struct user swaputl;
extern	struct user forkutl;
extern	struct user xswaputl;
extern	struct user xswap2utl;
extern	struct user pushutl;
extern	struct user vfutl;
char	ckey_cache[NOFCKEY]; /* = 1 if code key i may be in code cache */
char	dkey_cache[NOFDKEY]; /* = 1 if data key i may be in data cache */
short	ckey_cnt[NOFCKEY];   /* # of processes using code key i */
short	dkey_cnt[NOFDKEY];   /* processes using data key i (should be 1) */
#endif
