/*	vmmac.h	6.1	83/07/29	*/

/*
 * Virtual memory related conversion macros
 */

/* Core clicks to number of pages of page tables needed to map that much */
#define	ctopt(x)	(((x)+NPTEPG-1)/NPTEPG)

/* Virtual page numbers to text|data|stack segment page numbers and back */
#define	vtotp(p, v)	((int)(v)-LOWPAGES)
#define	vtodp(p, v)	((int)((v) - stoc(ctos((p)->p_tsize)) - LOWPAGES))
#define	vtosp(p, v)	((int)(btop(USRSTACK) - 1 - (v)))
#define	tptov(p, i)	((unsigned)(i) + LOWPAGES)
#define	dptov(p, i)	((unsigned)(stoc(ctos((p)->p_tsize)) + (i) + LOWPAGES))
#define	sptov(p, i)	((unsigned)(btop(USRSTACK) - 1 - (i)))

/* Tell whether virtual page numbers are in text|data|stack segment */
#define	isassv(p, v)	((v) & 0x200000)
#define	isatsv(p, v)	(((v) - LOWPAGES) < (p)->p_tsize)
#define	isadsv(p, v)	(((v) - LOWPAGES) >= stoc(ctos((p)->p_tsize)) && \
				!isassv(p, v))

/* Tell whether pte's are text|data|stack */
#define	isaspte(p, pte)		((pte) > sptopte(p, (p)->p_ssize))
#define	isatpte(p, pte)		((pte) < dptopte(p, 0))
#define	isadpte(p, pte)		(!isaspte(p, pte) && !isatpte(p, pte))

/* Text|data|stack pte's to segment page numbers and back */
#define	ptetotp(p, pte)		((pte) - (p)->p_p0br)
#define	ptetodp(p, pte)		((pte) - ((p)->p_p0br + (p)->p_tsize))
#define	ptetosp(p, pte)	\
	(((p)->p_p0br - (pte)) + ((p)->p_szpt * NPTEPG - UPAGES - 1))

#define	tptopte(p, i)		((p)->p_p0br + (i))
#define	dptopte(p, i)		((p)->p_p0br + (p)->p_tsize + (i))
#define	sptopte(p, i) \
	((p)->p_p0br + ((p)->p_szpt * NPTEPG - UPAGES - 1 - (i)))

/* Virtual page numbers to shared memory segment numbers and back */
#define vtoshmp(p, v)		vtodp(p, v)
#define shmptov(p, i)		dptov(p, i)

/* Tell if a virtual data page is in a shared-memory memory segment or not */
#define	isashmsv(p, v)		((isadsv(p, v)) && \
	(((v) - LOWPAGES) >= stoc(ctos((p)->p_tsize + (p)->p_dsize))))
#define	isansdsv(p, v)		((isadsv(p, v)) && \
	(((v) - LOWPAGES) < stoc(ctos((p)->p_tsize + (p)->p_dsize))))

/* Tell if a data pte is a shared or non-shared data pte */
#define isashmpte(p, pte)	((isadpte(p, pte)) && \
					((pte) >= dptopte(p, (p)->p_dsize)))
#define isansdpte(p, pte)	((isadpte(p, pte)) && \
					((pte) < dptopte(p, (p)->p_dsize)))

/* Shared memory ptes to segment page numbers and back */
#define	ptetoshmp(p, pte)	ptetodp(p, pte)
#define	shmptopte(p, i)		dptopte(p, i)

/* Invalidate the hardware translation buffer for a single pte */
#if CLSIZE==1
#define newpte(v) \
	mtpr(ptob((unsigned int)(v)), TBIS)
#endif

/* Invalidate the hardware translation buffer for a set of pte's */
#if CLSIZE==1
#define newptes(v, sz) \
{ \
	if ((sz) >= 8) \
		mtpr(0, TBIA); \
	else { \
		register caddr_t addrss = ptob((unsigned int)(v)); \
		while ((sz) > 0) { \
			mtpr(addrss, TBIS); \
			addrss += NBPG; \
			(sz)--; \
		} \
	} \
}
#endif

/* Convert a pte pointer to a virtual page number. */
#define ptetov(p, pte) \
	(   (isatpte((p), (pte))) ? \
		(tptov((p), ptetotp((p), (pte)))) : \
	    (isaspte((p), (pte))) ? \
		(sptov((p), ptetosp((p), (pte)))) : \
		(dptov((p), ptetodp((p), (pte))))   )

/* Convert a virtual page number to a pte address. */
/* Modified for HCX, for HCX drivers */
#define vtopte(p, v) \
	(   (((v) & 0x300000) == 0x300000) ? \
		(struct pte *)(mfpr (SBR) + 0xc0000000 + (((v)&0xfffff)<<2)) : \
	    (isassv((p), (v))) ? \
		(sptopte((p), vtosp((p), (v)))) : \
	    (isatsv((p), (v))) ? \
		(tptopte((p), vtotp((p), (v)))) : \
		(dptopte((p), vtodp((p), (v))))   )

/* Bytes to pages without rounding, and back */
#define	btop(x)		(((unsigned)(x)) >> PGSHIFT)
#define	ptob(x)		((caddr_t)((x) << PGSHIFT))

/* Turn virtual addresses into kernel map indices */
#define	kmxtob(a)	(usrpt + (a) * NPTEPG)
#define	btokmx(b)	(((b) - usrpt) / NPTEPG)

/* User area address and pcb bases */
#define	uaddr(p)	(&((p)->p_p0br[(p)->p_szpt * NPTEPG - UPAGES]))
#define	pcbb(p)		((p)->p_addr[0].pg_pfnum)

/* Average new into old with aging factor time */
#define	ave(smooth, cnt, time) \
	smooth = ((time - 1) * (smooth) + (cnt)) / (time)

/* Abstract machine dependent operations */

#define	setp0br(x)	(u.u_pcb.pcb_p0br = (x), mtpr(x, P0BR))
#define	setp0lr(x)	(u.u_pcb.pcb_p0lr = (x), mtpr(x, P0LR))
#define	setp1br(x)	(u.u_pcb.pcb_p1br = (x), mtpr(x, P1BR))
#define	setp1lr(x)	(u.u_pcb.pcb_p1lr = (x), mtpr(x, P1LR))
#define	setp2br(x)	(u.u_pcb.pcb_p2br = (x), mtpr(x, P2BR))
#define	setp2lr(x)	(u.u_pcb.pcb_p2lr = (x), mtpr(x, P2LR))
#define	initp2br(x)	((x) - P2PAGES)


#define	outofmem()	wakeup((caddr_t)&proc[2]);

/*
 * Page clustering macros.
 * 
 * dirtycl(pte)			is the page cluster dirty?
 * anycl(pte,fld)		does any pte in the cluster has fld set?
 * zapcl(pte,fld) = val		set all fields fld in the cluster to val
 * distcl(pte)			distribute high bits to cluster; note that
 *				distcl copies everything but pg_pfnum,
 *				INCLUDING pg_m!!!
 *
 * In all cases, pte must be the low pte in the cluster, even if
 * the segment grows backwards (e.g. the stack).
 */
#define	H(pte)	((struct hpte *)(pte))

#if CLSIZE==1
#define	dirtycl(pte)	dirty(pte)
#define	anycl(pte,fld)	((pte)->fld)
#define	zapcl(pte,fld)	(pte)->fld
#define	distcl(pte)
#endif

#if CLSIZE==2
#define	dirtycl(pte)	(dirty(pte) || dirty((pte)+1))
#define	anycl(pte,fld)	((pte)->fld || (((pte)+1)->fld))
#define	zapcl(pte,fld)	(pte)[1].fld = (pte)[0].fld
#endif

#if CLSIZE==4
#define	dirtycl(pte) \
    (dirty(pte) || dirty((pte)+1) || dirty((pte)+2) || dirty((pte)+3))
#define	anycl(pte,fld) \
    ((pte)->fld || (((pte)+1)->fld) || (((pte)+2)->fld) || (((pte)+3)->fld))
#define	zapcl(pte,fld) \
    (pte)[3].fld = (pte)[2].fld = (pte)[1].fld = (pte)[0].fld
#endif

#ifndef distcl
#define	distcl(pte)	zapcl(H(pte),pg_high)
#endif

/*
 * Lock a page frame.
 */
#define MLOCK(c) { \
	while ((c)->c_lock) { \
		(c)->c_want = 1; \
		sleep((caddr_t)(c), PSWP+1); \
	} \
	(c)->c_lock = 1; \
}
/*
 * Unlock a page frame.
 */
#define MUNLOCK(c) { \
	if (c->c_want) { \
		wakeup((caddr_t)c); \
		c->c_want = 0; \
	} \
	c->c_lock = 0; \
}
