/*	@(#)inode.h 1.1 85/05/30 SMI; from UCB 4.24 83/07/01	*/

/*
 * The I node is the focus of all file activity in UNIX.
 * There is a unique inode allocated for each active file,
 * each current directory, each mounted-on file, text file, and the root.
 * An inode is 'named' by its dev/inumber pair. (iget/iget.c)
 * Data in icommon is read in from permanent inode on volume.
 */

#define	NDADDR	12		/* direct addresses in inode */
#define	NIADDR	3		/* indirect addresses in inode */

struct inode {
	struct	inode *i_chain[2];	/* must be first */
	struct	vnode i_vnode;	/* vnode associated with this inode */
	struct	vnode *i_devvp;	/* vnode for block I/O */
	u_short	i_flag;
	dev_t	i_dev;		/* device where inode resides */
	ino_t	i_number;	/* i number, 1-to-1 with device address */
	long	i_id;		/* unique identifier */
	int	i_diroff;	/* offset in dir, where we found last entry */
	struct	fs *i_fs;	/* file sys associated with this inode */
	struct	dquot *i_dquot;	/* quota structure controlling this file */
	union {
		struct	{
			daddr_t	ir_lastr;		/* last read (read-ahead) */
			struct inode  *ir_shadow;	/* inode that this shadows */
		} i_regular;			/* IFREG, IFDIR */
		struct	socket *is_socket;	/* IFSOCK */
		struct	{
			struct socket *if_rsocket;	/* read socket */
			struct socket *if_wsocket;	/* write socket */
			short	if_frcnt;	/* number of active readers */
			short	if_fwcnt;	/* number of active writers */
		} i_fifo;			/* IFIFO */
	} i_un;
	struct	{
		struct inode  *if_freef;	/* free list forward */
		struct inode **if_freeb;	/* free list back */
	} i_fr;
	struct 	icommon
	{
		u_short	ic_mode;	/*  0: mode and type of file */
		short	ic_nlink;	/*  2: number of links to file */
		short	ic_uid;		/*  4: owner's user id */
		short	ic_gid;		/*  6: owner's group id */
		quad	ic_size;	/*  8: number of bytes in file */
#ifdef KERNEL
		struct timeval ic_atime;/* 16: time last accessed */
		struct timeval ic_mtime;/* 24: time last modified */
		struct timeval ic_ctime;/* 32: last time inode changed */
#else
		time_t	ic_atime;	/* 16: time last accessed */
		long	ic_atspare;
		time_t	ic_mtime;	/* 24: time last modified */
		long	ic_mtspare;
		time_t	ic_ctime;	/* 32: last time inode changed */
		long	ic_ctspare;
#endif
		daddr_t	ic_db[NDADDR];	/* 40: disk block addresses */
		daddr_t	ic_ib[NIADDR];	/* 88: indirect blocks */
		long	ic_flags;	/* 100: status, currently unused */
		long	ic_blocks;	/* 104: blocks actually held */
		long	ic_gen;		/* 108: generation number */
		long	ic_spare[4];	/* 112: reserved, currently unused */
	} i_ic;
};

struct dinode {
	union {
		struct	icommon di_icom;
		char	di_size[128];
	} di_un;
};

#define	i_mode		i_ic.ic_mode
#define	i_nlink		i_ic.ic_nlink
#define	i_uid		i_ic.ic_uid
#define	i_gid		i_ic.ic_gid
/* ugh! -- must be fixed */
#define	i_size		i_ic.ic_size.val[0]
#define	i_db		i_ic.ic_db
#define	i_ib		i_ic.ic_ib
#define	i_atime		i_ic.ic_atime
#define	i_mtime		i_ic.ic_mtime
#define	i_ctime		i_ic.ic_ctime
#define i_blocks	i_ic.ic_blocks
#define	i_rdev		i_ic.ic_db[0]
#define	i_lastr		i_un.i_regular.ir_lastr
#define	i_shadow	i_un.i_regular.ir_shadow
#define	i_gen		i_ic.ic_gen
#define	i_socket	i_un.is_socket
#define	i_rsocket	i_un.i_fifo.if_rsocket
#define	i_wsocket	i_un.i_fifo.if_wsocket
#define	i_frcnt		i_un.i_fifo.if_frcnt
#define	i_fwcnt		i_un.i_fifo.if_fwcnt
#define	i_forw		i_chain[0]
#define	i_back		i_chain[1]
#define	i_freef		i_fr.if_freef
#define	i_freeb		i_fr.if_freeb

#define di_ic		di_un.di_icom
#define	di_mode		di_ic.ic_mode
#define	di_nlink	di_ic.ic_nlink
#define	di_uid		di_ic.ic_uid
#define	di_gid		di_ic.ic_gid
#define	di_size		di_ic.ic_size.val[0]
#define	di_db		di_ic.ic_db
#define	di_ib		di_ic.ic_ib
#define	di_atime	di_ic.ic_atime
#define	di_mtime	di_ic.ic_mtime
#define	di_ctime	di_ic.ic_ctime
#define	di_rdev		di_ic.ic_db[0]
#define	di_blocks	di_ic.ic_blocks
#define	di_gen		di_ic.ic_gen

struct inostat {
	long	is_iget;	/* number of calls to "iget" */
};

#ifdef KERNEL

long	nextinodeid;		/* unique id generator */
struct inode *inode;			/* the inode table itself */
struct inode *inodeNINODE;		/* the end of the inode table */
int	ninode;				/* number of slots in the table */

struct inostat inostat;
extern struct vnodeops ufs_vnodeops;	/* vnode operations for ufs */

struct	inode *imake();
extern struct inode	*ialloc();
extern struct inode	*iget();
extern ino_t		dirpref();
#endif

/* flags */
#define	ILOCKED		0x1		/* inode is locked */
#define	IUPD		0x2		/* file has been modified */
#define	IACC		0x4		/* inode access time to be updated */
#define	IWANT		0x10		/* some process waiting on lock */
#define	ITEXT		0x20		/* inode is pure text prototype */
#define	ICHG		0x40		/* inode has been changed */
#define	ILWAIT		0x200		/* someone waiting on file lock */
#define	IINCORE		0x400		/* never flush this to disk */
#define IREF		0x800		/* inode is being referenced */
#define	IRENAME		0x1000		/* inode is being renamed */

/* modes */
#define IFMTSHFT        12              /* Number of bits to shift type by */
#define	IFMT		0170000		/* type of file */
#define	IFIFO		0010000		/* fifo special */
#define	IFCHR		0020000		/* character special */
#define	IFDIR		0040000		/* directory */
#define	IFBLK		0060000		/* block special */
#define	IFREG		0100000		/* regular */
#define	IFLNK		0120000		/* symbolic link */
#define	IFSOCK		0140000		/* socket */

#define	ISUID		04000		/* set user id on execution */
#define	ISGID		02000		/* set group id on execution */
#define	ISVTX		01000		/* save swapped text even after use */
#define	IREAD		0400		/* read, write, execute permissions */
#define	IWRITE		0200
#define	IEXEC		0100

/* ic_flags */
#define AUDLOG		0x01		/* audit trail logging bit */
#ifdef KERNEL

/* Return true if this file is for a local inode, sockets don't count and
   neither do NFS files. */
#define INODEFP( fp_ ) \
   ((fp_)->f_type==DTYPE_VNODE && VNISIN((struct vnode *)((fp_)->f_data)))

/* Return true if this vnode is for a local inode, NFS files don't count */
#define VNISIN( vp_ ) \
	((vp_)->v_op == &ufs_vnodeops )
/*
 * Convert between inode pointers and vnode pointers
 */
#define VTOI(VP)	((struct inode *)(VP)->v_data)
#define ITOV(IP)	((struct vnode *)&(IP)->i_vnode)

/*
 * Convert between vnode types and inode formats
 */
extern enum vtype	iftovt_tab[];
extern int		vttoif_tab[];
#define IFTOVT(M)	(iftovt_tab[((M) & IFMT) >> IFMTSHFT])
#define VTTOIF(T)	(vttoif_tab[(int)(T)])

#define MAKEIMODE(T, M)	(VTTOIF(T) | (M))

/*
 * Lock and unlock inodes.
 */
#ifdef ITRACE

extern int ilpanic();

#define	ILOCK(ip) { \
	while ((ip)->i_flag & ILOCKED) { \
		(ip)->i_flag |= IWANT; \
		sleep((caddr_t)(ip), PINOD); \
	} \
	itrace((ip), caller(), 1); \
	timeout(ilpanic, (ip), 30*hz); \
	(ip)->i_flag |= ILOCKED; \
}

#define	IUNLOCK(ip) { \
	(ip)->i_flag &= ~ILOCKED; \
	untimeout(ilpanic, (ip)); \
	itrace((ip), caller(), 0); \
	if ((ip)->i_flag&IWANT) { \
		(ip)->i_flag &= ~IWANT; \
		wakeup((caddr_t)(ip)); \
	} \
}
#else
#define	ILOCK(ip) { \
	while ((ip)->i_flag & ILOCKED) { \
		(ip)->i_flag |= IWANT; \
		sleep((caddr_t)(ip), PINOD); \
	} \
	(ip)->i_flag |= ILOCKED; \
}

#define	IUNLOCK(ip) { \
	(ip)->i_flag &= ~ILOCKED; \
	if ((ip)->i_flag&IWANT) { \
		(ip)->i_flag &= ~IWANT; \
		wakeup((caddr_t)(ip)); \
	} \
}
#endif

/*
 * Mark the accessed, updated, or changed times in an inode
 * with the current (unique) time
 */
#define IMARK(ip, flag) { \
	struct timeval ut; \
	UNIQTIME(&ut); \
	ip->i_flag |= flag; \
	if (flag & IACC) \
		ip->i_atime = ut; \
	if (flag & IUPD) \
		ip->i_mtime = ut; \
	if (flag & ICHG) { \
		ip->i_diroff = 0; \
		ip->i_ctime = ut; \
	} \
} \

/*
 * Check that file is owned by current user or user is su.
 */
#define OWNER(CR, IP) \
	(((CR)->cr_uid == (IP)->i_uid)? 0: (suser()? 0: u.u_error))

/*
 * enums
 */
enum de_op { DE_CREATE, DE_CREATE_DIR, DE_LINK, DE_RENAME }; /* direnter ops */

#endif
