/*
 * This following structure describes an ethernet frame.
 * This frame is defined by the the ACC V/EIU baord
 * A frame is used for either transmitting or receiving.
 * The first word of the frame is used for the command to the
 * V/EIU board and for the status from the V/EIU board.
 */

#define ENUNIT(x)   minor(x)
 
#define ENDEBUG(x)
 
#define SZ_DEBUG_DATA    30
 
struct  en_debug {
   unsigned short counter ;
   unsigned short direction ;
#define EN_XMIT 1
#define EN_RECV 2
   unsigned short cmnd_status ;
   unsigned short hdr_dhost[ 3 ] ;
   unsigned short hdr_shost[ 3 ] ;
   unsigned short hdr_type ;
   unsigned short data[ SZ_DEBUG_DATA ] ;
} ;
 
#ifdef EN_MONITOR
struct en_debug en_debug_buffer[ SZ_DEBUG_BUFFER ] ;
#endif EN_MONITOR
 
#define TSEG                 8

#define SZ_RANDOM           16
#define SZ_MIN_FRAME        46
#define SZ_FRAME_DATA     2000
 
struct frame {
   short                cmnd_status ;
   short                hdr_dhost[ 3 ] ;
   short                hdr_shost[ 3 ] ;
   unsigned short       hdr_type ;
   char                 frame_data[ SZ_FRAME_DATA ] ;
   short                random[ SZ_RANDOM ] ;
} ;
 
 
/*
 * The following are defines for bits and fields in the cmnd_status word.
 */
#define CS_BYTE_CNT     0x07FF
#define CS_RODD         0x1000
#define CS_RCRC         0x2000
#define CS_ROVRN        0x4000
#define CS_FULL         0x8000
#define CS_EMPTY        0x8000
#define CS_TRTC         0x000F
#define CS_TRTFAIL      0x4000

#define NUMBER_FRAMES 16

/*
 * There are IO port registers on the V/EIU board. The following
 * structure describes them.
 */

struct en_reg {

#define SZ_STATION_ADDR  6
#define VALID_STATION_ADDR 0x00FF

   short station_addr[ SZ_STATION_ADDR ] ;

#define SZ_BROADCAST_ADDR 2
#define VALID_BROADCAST_ADDR 0x00FF

   short broadcast_addr[ SZ_BROADCAST_ADDR ] ;

#define SZ_MULTICAST_ADDR 8
#define VALID_MULTICAST_ADDR 0x00FF

   short multicast_addr[ SZ_MULTICAST_ADDR ] ;

#define VALID_CSR     0x03FF
#define CSR_GO        0x0001
#define CSR_ENINT     0x0002
#define CSR_LOOP2     0x0004
#define CSR_LOOP3     0x0008
#define CSR_TCRC      0x0010
#define CSR_PROM      0x0020
#define CSR_RESET     0x0040
#define CSR_ACTV      0x0080
#define CSR_RESERVED8 0x0100
#define CSR_ENXMITODD 0x0200
#define CSR_RESERVEDA 0x0400
#define CSR_RESERVEDB 0x0800
#define CSR_RESERVEDC 0x1000
#define CSR_RESERVEDD 0x2000
#define CSR_RESERVEDE 0x4000
#define CSR_RESERVEDF 0x8000

   short csr ;

#define VALID_TSEG 0x7800

   short tseg ;

#define VALID_RSEG 0x7800

   short rseg ;
 
#define VALID_SEGB 0x7800

   short segb ;

#define VALID_RDRP 0x000F

   short rdrp ;

#define VALID_IVCT 0x00F8

   short ivct ;

#define VALID_RESERVED 0x0000

   short reserved ;
 
#define VALID_FCOL 0xFFFF

   short fcol ;

} ;
 
/*
 * Ethernet software status per interface.
 *
 * Each interface is referenced by a network interface structure,
 * is_if, which the routing code uses to locate the interface.
 * This structure contains the output queue for the interface, its address, ...
 */
 
struct en_softc {
 
   struct arpcom	is_ac ;               /* Ethernet common part      */
 
   struct frame *	frame_ptr ;
   struct en_reg *	en_reg_ptr ;
   char			xmit_avail ;          /* TRUE if there is an xmit    */
                                         /* buffer available.           */
   unsigned int		xmit_avail_head ;     /* The first available buffer. */
   unsigned int		xmit_avail_tail ;     /* Oldest full buffer.         */
   unsigned int		recv_buffer ;         /* The next receive buffer.    */
 
   unsigned int		error_odd ;           /* These errors are reported   */
   unsigned int		error_crc ;           /* by the V/IEU board.         */
   unsigned int		error_over_run ;
   unsigned int		error_too_short ;
 
   struct mbuf *	output_head_ptr ;
   struct mbuf *	output_tail_ptr ;
   unsigned int		queued_packets ;
   unsigned int		queue_depth ;         /* Instantaneous output q depth*/
   unsigned int		max_queue_depth ;     /* Maximum outout q depth.     */
 
   char			monitor ;             /* Monitoring variables.       */
   unsigned int		counter ;
   struct en_debug *	debug_ptr ;
   struct en_debug *	start_debug_ptr ;
   struct en_debug *	end_debug_ptr ;
 
} ;


#ifdef KERNEL

int enprobe( ), enslave( ),  enattach( ), enrintr( ), enxintr( ) ;
int eninit( ),  enoutput( ), enioctl( ),  enreset( ) ;
struct mbuf * enget( ) ;

extern int			nen ;
extern struct vba_device *	eninfo[ ] ;
extern struct en_softc		en_softc[ ] ;

extern int			endpm[ ] ;

extern struct vba_driver	endriver ;

#endif KERNEL
