
/*
 * Ethernet software status per interface.
 */

#define DEBUG	(boothowto & RB_DEBUG)

#define PHYS_ADDR( addr )	(unsigned int)((unsigned int)(addr) & (unsigned int)0x00ffffff)
#define VME_ADDRESS_MODIFIER	(unsigned int)0x39
#define	NH2X 4			/* Host to eXcelan request buffers */
#define	NX2H 4			/* eXcelan to Host reply buffers */

struct	ex_reg {
	unsigned short	ex_porta;	/* write on porta resets EXOS */
	unsigned short	ex_portb;	/* write on portb interrupts EXOS */
} ;

#define	NFRAGMENTS 1	/* number fragments that the EXOS will scatter/gather */

#define	EXMAXRBUF 1520	/* per EXOS 101 manual 5.3.7 (maybe 1518 would do) */

struct	buf_blk {		/* packet/buffer block descriptor */
	unsigned short	bb_len;			/* length of block, in bytes */
	unsigned short	bb_addr[2];		/* address of block */
	/*
	 * Array above is really a single unsigned int field.
	 * We kludge its definition to defeat word-alignment.
	 * Access would look like:
	 *     longaddr = *(unsigned int *)bp->.mb_er.er_blks[0].bb_addr;
	 */
} ;

struct	net_mode {		/* read/write mode control objects */
/*12*/	char	nm_request;	/* request code */
/*13*/	char	nm_reply;	/* reply code */
/*14*/	char	nm_mask;	/* bit-wise switches for read, write */
/*15*/	char	nm_optn;		/* acceptable packet reception errors */
/*16*/	char	nm_mode;		/* EXOS filtering mode */
/*17*/
} ;

struct	net_addrs {		/* read/write receive address slots */
/*12*/	char	na_request;	/* request code */
/*13*/	char	na_reply;	/* reply code */
/*14*/	char	na_mask;	/* bit-wise switches for read, write */
/*15*/	char	na_slot;		/* index of address slot */
/*16*/	char	na_addrs[6];		/* address read and/or written */
/*22*/
} ;

struct	net_recv {		/* read/alter receive slot enable bit */
/*12*/	char	nr_request;	/* request code */
/*13*/	char	nr_reply;	/* reply code */
/*14*/	char	nr_mask;	/* bit-wise switches for read, write */
/*15*/	char	nr_slot;		/* index of address slot */
/*16*/
} ;

struct	net_ststcs {		/* read/reset network statistics objects */
/*12*/	char	ns_request;	/* request code */
/*13*/	char	ns_reply;	/* reply code */
/*14*/	char	ns_mask;	/* bit-wise switches for read, write */
/*15*/	char	ns_rsrv;		/* reserved for EXOS */
/*16*/	unsigned short	ns_nobj;		/* number of objects to work on */
/*18*/	unsigned short	ns_xobj;		/* index of first object to work on */
/*20*/	unsigned int	ns_bufp;		/* pointer to statistics buffer */
/*24*/
} ;

struct	enet_xmit {		/* send a packet on the Ethernet */
/*12*/	char	et_request;	/* request code */
/*13*/	char	et_reply;	/* reply code */
/*14*/	char	et_slot;
/*15*/	char	et_nblock;		/* number of blocks composing packet */
/*16*/	struct	buf_blk et_blks[NFRAGMENTS];	/* array of block descriptors */
/*22-64*/
} ;

struct	enet_recv {		/* receive a packet on the Ethernet */
/*12*/	char	er_request;	/* request code */
/*13*/	char	er_reply;	/* reply code */
/*14*/	char	er_slot;
/*15*/	char	er_nblock;		/* number of blocks composing buffer */
/*16*/	struct	buf_blk er_blks[NFRAGMENTS];	/* array of block descriptors */
/*22-64*/
} ;

struct	stat {
	unsigned int	sa_fsent;	/* frames sent without errors */
	unsigned int	sa_xsclsn;	/* frames aborted excess collisions */
	unsigned int	sa_nsqe;	/* frames subject to heartbeat failure */
	unsigned int	sa_undef;	/* undefined (TDR on EXOS 101) */
	unsigned int	sa_frcvd;	/* frames received no errors */
	unsigned int	sa_align;	/* frames received alignment error */
	unsigned int	sa_crc;		/* frames received crc error */
	unsigned int	sa_flost;	/* frames lost */
} ;

struct	ex_config {
/*00*/	unsigned short	cm_0rsrv;	/* reserved, must be 1 */
/*02*/	char	cm_vc[4];	/* returns ASCII version code */
/*06*/	char	cm_cc;		/* returns config completion code */
/*07*/	char	cm_opmode;	/* specifies operation mode */
/*08*/	unsigned short	cm_dfo;		/* specifies host data format option */
/*10*/	char	cm_1rsrv ;	/* reserved, must be 0 */
/*11*/  char   cm_context ;    /* must be 0 for request, xos supplies */
/*12*/	char	cm_2rsrv ;  	/* reserved, must be 0 */
/*13*/	char	cm_ham;		/* specifies host address mode */
/*14*/	char	cm_3rsrv;	/* reserved, must be 0 */
/*15*/	char	cm_mapsiz;	/* reserved, must be 0 */
/*16*/	char	cm_byteptrn[4];	/* host data format option test pattern */
/*20*/	unsigned short	cm_wordptrn[2];
/*24*/	unsigned int	cm_lwordptrn;
/*28*/	char	cm_rsrvd[20];
/*48*/	unsigned int	cm_mba;		/* use 0xFFFFFFFF in link level mode */
/*52*/	char	cm_nproc;	/* use 0xFF in link level mode */
/*53*/	char	cm_nmbox;	/* use 0xFF in link level mode */
/*54*/	char	cm_nmcast;	/* use 0xFF in link level mode */
/*55*/	char	cm_nhost;	/* use 1 in link level mode */

	/* the next five parameters define the request message queue */
/*56*/	unsigned int	cm_h2xba;	/* base address of message queue */
/*60*/	unsigned short	cm_h2xhdr;	/* address offset of msg Q header */
/*62*/	char	cm_h2xtyp;	/* interrupt type */
/*63*/	char	cm_h2xval;	/* interrupt value (not used) */
/*64*/	unsigned short	cm_h2xaddr;	/* interrupt vector */
/*66*/	unsigned short	cm_h2xpad;	/* pad out unused portion of vector */

	/* the next five parameters define the reply message queue */
/*68*/	unsigned int	cm_x2hba;	/* base address of message queue */
/*72*/	unsigned short	cm_x2hhdr;	/* address offset of msg Q header */
/*74*/	char	cm_x2htyp;	/* interrupt type */
/*75*/	char	cm_x2hval;	/* interrupt value (not used) */
/*76*/	unsigned short	cm_x2haddr;	/* interrupt vector */
/*78*/	unsigned short	cm_x2hpad;	/* pad out unused portion of vector */
/*80*/
} ;

/* we send requests and receive replys with the EXOS using this structure */
struct	ex_msg {
/*00*/	unsigned short	mb_link;	/* address of next message buffer */
/*02*/	char	mb_rsrv;	/* reserved for use by EXOS */
/*03*/	char	mb_status;	/* used bit-wise for message protocol */
/*04*/	unsigned short	mb_length;	/* length, in bytes, of the rest */
/*06*/	unsigned short	mb_1rsrv;	/* reserved for used by EXOS */
/*08*/	int	mb_mid;		/* available to user */
/*12*/	union	mb_all {
		struct	net_mode	mb_net_mode;
		struct	net_addrs	mb_net_addrs;
		struct	net_recv	mb_net_recv;
		struct	net_ststcs	mb_net_ststcs;
		struct	enet_xmit	mb_enet_xmit;
		struct	enet_recv	mb_enet_recv;
	} mb_all;
/* following field is used only by host, not read by board */
	struct ex_msg *	mb_next;	/* host's pointer to next message */
} ;

struct ex_queue_base {

	unsigned short		h2xhdr ;	/* EXOS's request queue header */
	unsigned short		x2hhdr ;	/* EXOS's reply queue header */
	struct ex_msg		h2xmsg[NH2X] ;	/* request msg buffers */
	struct ex_msg		x2hmsg[NX2H] ;	/* reply msg buffers */
	struct ex_config	cm ;		/* configuration message */
	struct stat		sa ;		/* EXOS writes stats here */
} ;

struct  ex_debug {
	unsigned short counter ;
	unsigned short direction ;
#define EX_XMIT 1
#define EX_RECV 2
	unsigned short cmnd_status ;
	char hdr_dhost[ 6 ] ;
	char hdr_shost[ 6 ] ;
	unsigned short hdr_type ;
	char data[ 60 ] ;
} ;

#ifdef EX_MONITOR
struct ex_debug ex_debug_buffer[ SZ_DEBUG_BUFFER ] ;
#endif EX_MONITOR

struct	ex_softc {
	struct arpcom	is_ac ;		/* Ethernet common part      */
	int		is_flags ;	/* private flags */
	unsigned int    unit ;          /* Unit number of this EXOS board */
#define	EX_XPENDING	1		/* xmit rqst pending on EXOS */
#define	EX_STATPENDING	2		/* stats rqst pending on EXOS */
#define	EX_RPENDING	4		/* recv rqst pending on EXOS */
	struct ex_msg *	is_h2xnext ;	/* host pointer to request queue */
	struct ex_msg *	is_x2hnext ;	/* host pointer to reply queue */
	struct ex_queue_base *	ex_qb_ptr ;	/* map info for structs below */
	struct ex_reg *   ex_reg_ptr ;    /* points to porta and portb */
	/* the following is not referenced directly. The address of the queue */
	/* base must fall on a 16 byte boundary. ex_qb_ptr will point at      */
	/* the first address after pad that meets this criterion. After       */
	/* ex_qb_ptr is initialized, the following should not be referenced.  */
	char			pad[ 16 ] ;
	struct ex_queue_base	queue_base ;
	/* end mapped area */
	char			ex_rbuf[ ETHERMTU+50 ] ;
	char			ex_tbuf[ ETHERMTU+50 ] ;
/* EX_MONITOR variables */
	unsigned int		counter ;
	struct ex_debug *	debug_ptr ;
	struct ex_debug *	start_debug_ptr ;
	struct ex_debug *	end_debug_ptr ;
} ;

#ifdef KERNEL

int exprobe( ), exslave(),   exattach( ), exintr( ) ;
int exinit( ),  exoutput( ), exioctl( ),  exreset( ) ;
struct mbuf * exget( ) ;

extern int			nex ;
extern struct vba_device *	exinfo[ ] ;
extern struct vba_ctlr *	exminfo[ ] ;
extern struct ex_softc		ex_softc[ ] ;
extern struct vba_driver	exdriver ;

#endif KERNEL

#define EXWATCHINTVL	20	/* call exwatch every x secs */

/* EXOS I/O PORT A write definitions */
#define	EX_RESET     0x00	/* value doesn't really matter... */

/* EXOS I/O PORT B write definitions */
#define	EX_INTERRUPT 0x00

/* EXOS I/O PORT B read definitions */
#define	EX_TESTOK    0x01	/* set when self-diagnostics passed */
#define	EX_UNREADY   0x08     	/* set until EXOS ready to read from B */

/* message buffer status field definitions */
#define	MH_OWNER	1	/* mask for status bit for owner */
#define	MH_HOST		0	/* if 0, the host owns the buffer */
#define	MH_EXOS		1	/* if 1, the EXOS owns the buffer */

/* EXOS Link Level request codes */
#define	LLTRANSMIT	0xC	/* send a packet */
#define	LLRTRANSMIT	0xE	/* send a packet, and self-receive */
#define	LLRECEIVE	0xD	/* receive a packet */
#define	LLNET_MODE	0x8	/* read/write mode control objects */
#define	LLNET_ADDRS	0x9	/* read/write receive address slots */
#define	LLNET_RECV	0xA	/* read/alter receive slot enable bit */
#define	LLNET_STSTCS	0xB	/* read/reset network statistics objects */

/* Link Level return codes common to all requests */
#define	LL_OK		0	/* successful completion */
#define	LLX_MODE	0xA1	/* EXOS not in link level mode. */

/* LLTRANSMIT unique return codes */
#define	LLXM_1RTRY	0x1	/* successful xmission, 1 retry */
#define	LLXM_RTRYS	0x2	/* successful xmission, more than 1 retry */
#define	LLXM_NSQE	0x8	/* successful xmission, no SQE TEST signal */
#define	LLXM_CLSN	0x10	/* xmission failed, excess retries */
#define	LLXM_NCS	0x20	/* xmission failed, no carrier sense */
#define	LLXM_LNGTH	0x40	/* xmission failed, bad packet length */
#define	XMIT_BITS	"\7\7LENGTH\6CARRIER\5XCLSNS\4SQETST"
#define	LLXM_ERROR	(LLXM_NSQE|LLXM_CLSN|LLXM_NCS|LLXM_LNGTH)

/* LLRECEIVE unique return codes */
#define	LLRC_TRUNC	0x4	/* pkt received, but truncated to fit buffer */
#define	LLRC_ALIGN	0x10	/* pkt received, but with alignment error */
#define	LLRC_CRC	0x20	/* pkt received, but with CRC error */
#define	LLRC_BUFLEN	0x40	/* no pkt received, buffer less than 64 bytes */
				/* this should never happen here */
#define	RECV_BITS	"\7\7BUFLEN\6CRC\5ALIGN\3TRUNC"

/* LLNET_ADDRS unique return codes */
#define	LLNA_BADSLOT	0xD1	/* slot doesn't exist or can't be accessed */
#define	LLNA_BADADDR	0xD3	/* invalid address for designated slot */

/* LLNET_RECV unique return codes */
#define	LLNR_BADSLOT	0xD1	/* slot doesn't exist or can't be accessed */
#define	LLNR_BADADDR	0xD2	/* designated slot was empty */


/* address slot object indices */
#define	NULLSLOT	0	/* the null slot */
#define	MINMCSLOT	1	/* minimum multicast slot index */
#define	MAXMCSLOT	8	/* default maximum multicast slot index */
#define	PHYSSLOT	253	/* physical slot index */
#define	UNVRSSLOT	254	/* universal slot index */
#define	BROADSLOT	255	/* broadcast slot index */

/* request mask bit definitions */
#define	WRITE_OBJ	1	/* write the object */
#define	READ_OBJ	2	/* read the object */
#define	ENABLE_RCV	4	/* enable reception on designated slot */

/* NET_MODE options mask bit definitions */
#define	OPT_ALIGN	0x10	/* receive packets with alignment errors */
#define	OPT_CRC		0x20	/* receive packets with CRC errors */
#define	OPT_DSABLE	0x80	/* disconnect controller hardware */

/* NET_MODE mode field value definitions */
#define	MODE_OFF	0	/* stop transmission and reception */
#define	MODE_PERF	1	/* perfect multicast address filtering */
#define	MODE_HW		2	/* hardware-only multicast address filtering */
#define	MODE_PROM	3	/* promiscuous reception */

/*
 * N.B.  Structures below are carefully constructed so that
 * they correspond to the message formats that NX firmware
 * defines.  None of them should contain any compiler-instigated
 * padding.  Be especially careful about VAX C longword alignment!
 */

#define	mb_nm	mb_all.mb_net_mode
#define	mb_na	mb_all.mb_net_addrs
#define	mb_nr	mb_all.mb_net_recv
#define	mb_ns	mb_all.mb_net_ststcs
#define	mb_et	mb_all.mb_enet_xmit
#define	mb_er	mb_all.mb_enet_recv
#define	MBDATALEN (sizeof(union mb_all)+6)
