/*
 * Header file for VME BUS proNET-80 driver.
 */

/*
 *      host-board communication modes
 *        XMT_DATA_BY_DMA       enable dma of data on output
 *        XMT_CMD_BY_DMA        enable programming of board registers by dma
 *                              . on output ( use in conjunction with data
 *                              . dma only -- ie. XMT_DATA_BY_DMA defined )
 *        RCV_DATA_BY_DMA       enable dma of data on input
 */
#define         XMT_DATA_BY_DMA
#define         XMT_CMD_BY_DMA
#define         RCV_DATA_BY_DMA
#define         INTERRUPT

#define PNIVCT          (0xe8)

#define		is_if		is_ac.ac_if	/* network-visible interface */
#define		is_addr		is_ac.ac_enaddr	/* hardware Ethernet address */

#ifndef FALSE
#define FALSE		0
#endif

#ifndef TRUE
#define TRUE		1
#endif

#define PNMTU		(1536)
#define PNMRU		(PNMTU + 16)
#define PNTBUF		(PNMRU + sizeof( struct pn_header ))
#define PNBUFSIZE	(PNMRU + sizeof( struct pn_header ))

/*
 * Number of intermediate receive buffers. These are buffer area for
 * "funny" mbufs managed by the driver.
 */
#define PNBUF		20

#define PNUNIT(x)	minor(x)

/*
 *   debugging and tracing stuff
 */
extern	int	pn_tracehdr ;	/* 1 => trace headers (slowly!!) */
extern	int	pn_logerrors ;	/* 1 => log all i/o errors */

#define pntracehdr  if (pn_tracehdr) pnprt_hdr
#define pnprintf    if (pn_logerrors) printf


#define PN_HEADER_SIZE  6
struct pn_header {
	/* the first two fields are required by the hardware */
	u_char		pn_dhost;	/* destination address */
	u_char		pn_shost;	/* source address */
	/* the next three fields are the local network header */
  	u_char		pn_version;	/* header version */
	u_char		pn_type;	/* packer type => protocol number */
  	u_char		pn_info;	/* protocol-specific information */
  	u_char		pn_tr_sz ;	/* if ( pn_type == RING_TRAIL )  */
  					/*	off = pn_tr_sz * 512 ;   */
};

/*
 * Debugging aid
 */
#define TCSR    	0x01
#define TBUFAD  	0x02
#define TBUFDA  	0x03    /* WO */
#define RFSREG          0x04    /* RO */
#define DMAINT          0x05
#define RCSR    	0x06
#define RCNTR           0x07    /* RO */
#define RCOUNTER        0x08    /* WO */
#define RSTATR  	0x09    /* RO */
#define RSTCNTR 	0x0a    /* WO */
#define RBUFDMAPTR      0x0b    /* WO */
#define NODEADDR        0x0c    /* WO */
#define RBUFFER         0x0f    /* RO */

#define PN_READ 	0x80
#define PN_WRITE        0x40
#define PN_INVALID      0x20

struct pn_reg_struct {
	int     reg ;
	int     addr ;
	int     value ;
} pn_reg[ 100 ] ;

int	pn_index ;
int	pn_rindex ;

struct	ccb_t1	{
	 short	command;
	 short	dum1;
	 short	srcpntrl;
	 short	dum2;
	 short	srcpntrh;
	 short	dum3;
	 short	dstpntrl;
	 short	dum4;
	 short	dstpntrh;
	 short	dum5;
	 short	bytcntl;
	 short	dum6;
	 short	bytcnth;
	 short	dum7;
	 short	chanstatus;
	 short	dum8;
	};

struct	ccb_t1e	{
	 short	command;
	 short	dum1;
	 short	srcpntrl;
	 short	dum2;
	 short	srcpntrh;
	 short	dum3;
	 short	dstpntrl;
	 short	dum4;
	 short	dstpntrh;
	 short	dum5;
	 short	bytcntl;
	 short	dum6;
	 short	bytcnth;
	 short	dum7;
	 short	chanstatus;
	 short	dum8;
	 short	commext;
	 short	dum9;
	 short	mask;
	 short	dum10;
	 short	comp;
	 short	dum11;
	 short	trnpntrl;
	 short	dum12;
	 short	trnpntrh;
	 short	dum13;
	};

struct	ccb_t2	{
	 short	chaincom;
	 short	dum9;
	 short	jumpl;
	 short	dum10;
	 short	jumph;
	 short	dum11;
	};


struct	dc0 {			/* Channel 0 cmd blk */
	long	align;
	struct	ccb_t2	ccb1 ;
	struct	ccb_t1	ccb2 ;
	struct	ccb_t2	ccb3 ;
	} ;

struct	dc1 {			/* Channel 1 cmd blk */
	long	align;
	struct	ccb_t1	ccb1 ;
	struct	ccb_t2	ccb2 ;
	} ;

struct  dc2 {			/* Channel 2 cmd blk */
	long	align;
	struct	ccb_t1	ccb1 ;
	struct	ccb_t2	ccb2 ;
	struct	ccb_t1	ccb3 ;
	struct	ccb_t1	ccb4 ;
	struct	ccb_t2	ccb5 ;
	} ;

struct  dc3 {			/* Channel 3 cmd blk */
	long	align;
	struct	ccb_t1	ccb1 ;
	struct	ccb_t2	ccb2 ;
	} ;


/*
 * Software status of each interface.
 *
 * Each interface is referenced by a network interface structure,
 * pn_if, which the routing code uses to locate the interface.
 * Structure contains the output queue for the interface, its address, ...
 */
struct pn_softc {

	struct arpcom	is_ac ;		/* Ethernet common part      */

	struct pn_reg *	pn_reg_ptr ;
	short		pn_oactive;	/* is output active */
	u_short		pn_ostart;	/* output start, for retransmission */
	short		pn_tries;	/* transmit current retry count */
	short		pn_init;	/* number of ring inits */
	short		pn_refused;	/* number of packets refused */
	short		pn_timeouts;	/* number of transmit timeouts */
	short		pn_otimeout;	/* number of output timeouts */
	short		pn_ibadf;	/* number of input bad formats */
	short		pn_parity;	/* number of parity errors */
	u_short		rcvhd;		/* beginning of the rcv blk, for RBUFDMAPTR */
	u_long *	pn_rcv_ptr;	/* rcv buffer pointer */
	u_long          pn_rcv[ (PNBUFSIZE >> 2) + 1 ] ;
	u_long *	pn_tbuf_ptr;	/* xmit buffer pointer */
	u_long		pn_tbuf[ (PNTBUF >> 2) + 1 ] ;   /* xmit buffer */
	int             pn_rcv_len ;
	struct dc0      pn_dmac_ch0 ;   /* dmac channel 0 command buffer */
	struct dc1      pn_dmac_ch1 ;   /* dmac channel 1 command buffer */
	struct dc2      pn_dmac_ch2 ;   /* dmac channel 2 command buffer */
	struct dc3      pn_dmac_ch3 ;   /* dmac channel 3 command buffer */

} ;

#define RING_VERSION	2	/* current version of v2lni header */

/*
 * Packet types (protocol numbers) in proNET protocol header
 * Other types are defined, but are proprietary.
 */
#define RING_IP		1
#define RING_TRAIL	2
#define RING_ARP	3
#define RING_HDLC	4
#define RING_VAXDB	5
#define RING_RINGWAY	6
#define RING_RINGWAYM	8
#define RING_NOVELL	10
#define RING_PUP	12
#define RING_XNS	14
#define RING_DIAGNOSTICS 15		/* protocol type for testing */
#define RING_ECHO	16

#define PN_BROADCAST	0xFF		/* hardware-defined broadcast address */
#define PN_GRPOF(c)	((c >> 4) & 0xF)/* extract group number */
#define PN_MBROF(c)	(c & 0xF)	/* extract member number */
#define PN_GRPCAST(c)	(!PN_MBROF(c))	/* is it a group broadcast */
#define RCMD_CHAN	0
#define RCV_CHAN	1
#define XCMD_CHAN	2
#define XMT_CHAN	3
#define TBUFSIZE	0x0400		/* in lwords */
#define RBUFSIZE	0x4000		/* in bytes */

#define RDMA		0x8000		/* DMA_INT bit 15 */
#define TDMA		0x0080		/* DMA_INT bit 7  */
#define NXM		0x0010		/* DMA_INT bit 4  */
#define RSTRDMA		0x4000		/* DMA_INT bit 14 resets RDMA */
#define RSTTDMA 	0x0040		/* DMA_INT bit 6 resets TDMA  */
#define RSTNXM  	0x0010		/* DMA_INT bit 4 resets NXM   */

/* the following structure is the 82258 register map
 * address offset for these registers is 0x4000
 * sizeof(dmac_regs) should be 0x100
 * therefore sizeof(struct dmac_regs) should be 0x40
 */
struct dmac_regs {			/* all dummy fields are place holders */
/*4000*/u_short		gcr;		/* general command register */
/*4002*/u_short		scr;		/* subchannel register */
/*4004*/u_short		gsr;		/* general status register */
	u_short		dummy2;		/* missing in 82258 register map */
/*4008*/u_short		gmr;		/* general mode register */
/*400a*/u_short		gbr;		/* general burst register */
/*400c*/u_short		gdr;		/* general delay register */
	u_short		dummy5;		/* missing in 82258 register map */
/*4010*/u_short		csr;		/* channel[0-3] status registers */
/*4012*/u_short		dar;		/* channel[0-3] data assembly registers */
/*4014*/u_short		maskr;		/* channel[0-3] mask registers */
/*4016*/u_short		cmpr;		/* channel[0-3] compare registers */
/*4018*/u_short		mivr;		/* multiplexer interrupt vector register */
/*401a*/u_short		lvr;		/* last vector register */
	u_short		dummy9;		/* missing in 82258 register map */
	u_short		dummy10;	/* missing in 82258 register map */
/*4020*/u_short		cprL;		/* channel[0-3] command pointer registers */
/*4022*/u_short		cprH;
/*4024*/u_short		sprL;		/* channel[0-3] source pointer registers */
/*4026*/u_short		sprH;
/*4028*/u_short		dprL;		/* channel[0-3] dest pointer registers */
/*402a*/u_short		dprH;
/*402c*/u_short		ttprL;		/* channel[0-3] translate table pointer regs */
/*402e*/u_short		ttprH;
/*4030*/u_short		lprL;		/* channel[0-3] list pointer registers */
/*4032*/u_short		lprH;
	u_short		dummy16;	/* missing in 82258 register map */
	u_short		dummy17;	/* missing in 82258 register map */
/*4038*/u_short		bcrL;		/* channel[0-3] byte count registers */
/*403a*/u_short		bcrH;
/*403c*/u_short		ccrL;		/* channel[0-3] command registers */
/*403e*/u_short		ccrH;
} ;
/* these registers only appear once in the dmac_regs array */
#define GCR_OF(c)	((c)->pn_dmac[0].gcr)
#define SCR_OF(c)	((c)->pn_dmac[0].scr)
#define GSR_OF(c)	((c)->pn_dmac[0].gsr)
#define GMR_OF(c)	((c)->pn_dmac[0].gmr)
#define GBR_OF(c)	((c)->pn_dmac[0].gbr)
#define GDR_OF(c)	((c)->pn_dmac[0].gdr)
#define MIVR_OF(c)	((c)->pn_dmac[3].mivr)
#define LVR_OF(c)	((c)->pn_dmac[3].lvr)


#define CHAN0   0
#define CHAN1   1
#define CHAN2   2
#define CHAN3   3


#define CPRL_OF(i)	(pn_reg_ptr->pn_dmac[i].cprL)
#define CPRH_OF(i)	(pn_reg_ptr->pn_dmac[i].cprH)
#define SPRL_OF(i)	(pn_reg_ptr->pn_dmac[i].sprL)
#define SPRH_OF(i)	(pn_reg_ptr->pn_dmac[i].sprH)
#define DPRL_OF(i)	(pn_reg_ptr->pn_dmac[i].dprL)
#define DPRH_OF(i)	(pn_reg_ptr->pn_dmac[i].dprH)
#define TTPRL_OF(i)	(pn_reg_ptr->pn_dmac[i].ttprL)
#define TTPRH_OF(i)	(pn_reg_ptr->pn_dmac[i].ttprH)
#define LPRL_OF(i)	(pn_reg_ptr->pn_dmac[i].lprL)
#define LPRH_OF(i)	(pn_reg_ptr->pn_dmac[i].lprH)
#define BCRL_OF(i)	(pn_reg_ptr->pn_dmac[i].bcrL)
#define BCRH_OF(i)	(pn_reg_ptr->pn_dmac[i].bcrH)
#define CCRL_OF(i)	(pn_reg_ptr->pn_dmac[i].ccrL)
#define CCRH_OF(i)	(pn_reg_ptr->pn_dmac[i].ccrH)
#define MASKR_OF(i)	(pn_reg_ptr->pn_dmac[i].maskr)
#define CMPR_OF(i)	(pn_reg_ptr->pn_dmac[i].cmpr)
#define DAR_OF(i)	(pn_reg_ptr->pn_dmac[i].dar)
#define CSR_OF(i)	(pn_reg_ptr->pn_dmac[i].csr)

/* the following are interrupt vector, dma page, and csr registers
 * address offset for this structure is 0x4000 + sizeof(dmac_regs)
 */
struct ctl_regs {		/* all dummy fields are place holders */
/*4100*/u_short		rdmaintr;	/* rcv dma interrupt vector */
        u_short		dummy0;
/*4104*/u_short		xdmaintr;	/* xmt dma and nxm interrupt vector */
        u_short		dummy1;
/*4108*/u_short		rintr;		/* rcv done interrupt vector */
        u_short		dummy2;
/*410c*/u_short		xintr;		/* xmt done interrupt vector */
	u_short		dummy3;
/*4110*/u_short		pronet_pg;	/* high 8 bits for pronet page addr in DMA */
	u_short		dummy4;
/*4114*/u_short		cmd_pg;		/* high 8 bits for cmd page addr in DMA */
	u_short		dummy5;
/*4118*/u_short		rcvdata_pg;	/* high 8 bits for rcv data page addr in DMA */
	u_short		dummy6;
/*411c*/u_short		xmtdata_pg;	/* high 8 bits for xmt data page addr in DMA */
	u_short		dummy7;
/*4120*/u_short		tcsr;		/* xmt csr */
#define PN_ORIGR	0x0001	/* originate/xmt/send pkt */
#define PN_INITR	0x0010
#define PN_TRESET  	0x0020	/* reset xmt/rcv to known state */
#define PN_TINTEN	0x0040	/* IEN: ORIGC(CTL), OTIM(TCSR) or CNTR(RCSR) */
#define PN_TREADY	0x0080	/* xmt - TCSR */
#define PN_TREFUSE	0x0100	/* last pkt sent wasn't accepted */
#define PN_TBDFORM	0x0400
#define PN_OTIM		0x0800	/* timed out waiting for ORIGC(CTL) */
#define PN_JOIND	0x1000	/* CTL is acknowledging a JOINR(RCSR) cmd */
#define PN_RNOK		0x2000	/* NOTOKEN or NOFLAG timer has expired */
#define PN_RTINT	0x4000	/* reset xmt/rcv int */
#define PN_TINT		0x8000	/* xmt/rcv csr int request */
	u_short		dummy8;
/*4124*/u_short		tbufad;		/* xmt buffer addr reg */
	u_short         dummy9;
/*4128*/u_long 		tbufda;		/* xmt buffer data reg (write-only) */
#define rfsreg  tbufda                  /* refuse register (read-only) */
/*412c*/u_short		dmaint;	/* dma and interrupt control reg */
	u_short         dummy11;
/*4130*/u_short		rcsr;		/* rcv csr */
#define PN_COPYEN	0x0001	/* enable copy of rcv pkt into RBUFFER */
#define PN_JOINR	0x0002	/* enable relay in wire center for join ring */
#define PN_ENMOD	0x0004	/* enable CTL modem */
#define PN_RRESET	0x0020
#define PN_RINTEN	0x0040
#define PN_CNTR		0x0080
#define PN_CNTW		0x0100
#define PN_PERR		0x0800	/* parity error detected */
/*      PN_JOIND	0x1000	*/
/*      PN_RNOK		0x2000	*/
#define PN_RRINT	0x4000
#define PN_RINT		0x8000
	u_short         dummy12;
/*4134*/u_short		rcntr;	/* rcv count reg (read-only) */
#define	rcounter rcntr		/* rcv buffer input counter (write-only) */
	u_short         dummy13;
/*4138*/u_short		rstatr;	/* rcv status reg (reaqd-only) */
	u_short         dummy14;
#define rstcntr	rstatr		/* reset rcntr reg (write-only) */
/*413c*/u_short		rbufdmaptr;	/* rcv buffer DMA pointer (write_only)*/
#define nodeaddr	rbufdmaptr	/* (write-only) */
	u_short         dummy15;
};

/*********************************/
/*   The three operating modes   */
/*********************************/
static u_short pn_modes[] = {
	PN_JOINR,			/* digital loopback */
	PN_ENMOD,			/* analog loopback */
	PN_JOINR | PN_ENMOD		/* network mode */
};

#define PN_TBITS \
	"\10\20IRQ\17RTI\16NOK\15JND\14TMO\13BDF\11REF\10RDY\7IEN\6RST\5INI\1ORI"
#define PN_RBITS \
	"\10\20IRQ\17RTI\16NOK\15JND\14PER\11CSW\10RDY\7IEN\6RST\3MEN\2JNR\1CEN"
#define PN_SBITS 	"\10\3PERR\2RBDF\1ROVR"

#define PERR	04		/* parity error bit in RSTATR */
#define RBDFM	02		/* bad format error bit in RSTATR */
#define ROVR	01		/* pkt length overflow error bit in RSTATR */

struct pn_reg {
/*0000*/u_long			pn_rbuffer[ 0x1000 ] ;	/* rcv buffer */
/*4000*/struct dmac_regs	pn_dmac[4];	/* 82258 reg map */
/*4100*/struct ctl_regs		pn_ctl;		/* int vec, dma pg, and csrs reg map */
};

#define PNTERR	(PN_OTIM | PN_TBDFORM | PN_TREFUSE)	/* any xmt error */
#define PNRERR	(PERR | RBDFM | ROVR)		/* any rcv error */

#define PNRETRY		10	/* xmt retry limit */
#define PNIDENTRETRY	10	/* identify retry limit */
#define PNIDENTSUCC	 5	/* number of successful sends required */
#define PNTIMEOUT	60	/* seconds before a xmt timeout */

/* 82258 DMAC specific stuff */
struct short_cmd {
	u_short		type1_cmd;
	u_short		sprL;
	u_char		sprH;
	u_char		dummy0;
	u_short		dprL;
	u_char		dprH;
	u_char		dummy1;
	u_short		bcrL;
	u_char		bcrH;
	u_char		dummy2;
	u_short		csr;
};

struct long_cmd {
	struct short_cmd	shcmd;
	u_short		ext_cmd;
	u_short		mask;
	u_short		cmpr;
	u_short		ttprL;
	u_char		ttprH;
	u_char		dummy0;
};

struct debug_cmd {
	struct long_cmd		cmdbl;
	u_short		updated_sprL;
	u_char		updated_sprH;
	u_char		dummy0;
	u_short		updated_dprL;
	u_char		updated_dprH;
	u_char		dummy1;
	u_short		updated_bcrL;
	u_char		updated_bcrH;
	u_char		dummy2;
};

#define gsr_of(chan,addr)	((addr)[0].gsr >> (chan))

#ifdef KERNEL

int pnprobe( ), pnslave(),   pnattach( ), pnintr( ) ;
int pninit( ),  pnoutput( ), pnioctl( ),  pnreset( ) ;

extern int			npn ;
extern struct vba_device *	pninfo[ ] ;
extern struct vba_ctlr *	pnminfo[ ] ;
extern struct pn_softc		pn_softc[ ] ;
extern struct vba_driver	pndriver ;

#endif KERNEL




#define	DI_DEN (short) 0x01
#define	DI_TRN (short) 0x02
#define	DI_DCM (short) 0x08
#define	DI_NXM (short) 0x10
#define	DI_RNX (short) 0x10
#define	DI_RTI (short) 0x40
#define	DI_TIT (short) 0x80
#define	DI_RRI (short) 0x4000
#define	DI_RIT (short) 0x8000



/* ADMAC General Mode Register Bits */

#define	DGM_SCW (short) 0x01		/* system word bus */
#define	DGM_RCW (short) 0x02		/* resident word bus */
#define	DGM_REM (short) 0x04		/* remote mode */
#define	DGM_MUL (short) 0x08		/* multiplexor channel */
#define	DGM_CY0 (short) 0x10		/* single cycle channel 0 */
#define	DGM_CY1 (short) 0x20		/* single cycle channel 1 */
#define	DGM_CY2 (short) 0x40		/* single cycle channel 2 */
#define	DGM_CY3 (short) 0x80		/* single cycle channel 3 */
#define	DGM_PR0 (short) 0x0100		/* priority scheme 0 = 0,1,2,3 */
#define	DGM_PR1 (short) 0x0200		/* 1 = 012,3 2 = 01,23  3 = 0123 */
#define	DGM_IN0 (short) 0x0400		/* disable channel 0 interrupt */
#define	DGM_IN1 (short) 0x0800		/* disable channel 1 interrupt */
#define	DGM_IN2 (short) 0x1000		/* disable channel 2 interrupt */
#define	DGM_IN3 (short) 0x2000		/* disable channel 3 interrupt */
#define	DGM_CMI (short) 0x4000		/* common interrupt on EOD2 */

/* ADMAC General Status Register Bits */

#define	DS_REQ0 (short) 0x01		/* DREQ pending if inactive
					 * transfer in progress if active */
#define	DS_ACT0 (short) 0x02		/* channel active */
#define	DS_INT0 (short) 0x04		/* interrupt pending */
#define	DS_MEM0 (short) 0x08		/* control in memory/io- space */
#define	DS_REQ1 (short) 0x10		/* DREQ pending if inactive
					 * transfer in progress if active */
#define	DS_ACT1 (short) 0x20		/* channel active */
#define	DS_INT1 (short) 0x40		/* interrupt pending */
#define	DS_MEM1 (short) 0x80		/* control in memory/io- space */
#define	DS_REQ2 (short) 0x0100		/* DREQ pending if inactive
					 * transfer in progress if active */
#define	DS_ACT2 (short) 0x0200		/* channel active */
#define	DS_INT2 (short) 0x0400		/* interrupt pending */
#define	DS_MEM2 (short) 0x0800		/* control in memory/io- space */
#define	DS_REQ3 (short) 0x1000		/* DREQ pending if inactive
					 * transfer in progress if active */
#define	DS_ACT3 (short) 0x2000		/* channel active */
#define	DS_INT3 (short) 0x4000		/* interrupt pending */
#define	DS_MEM3 (short) 0x8000		/* control in memory/io- space */

/* ADMAC Channel Command Type I Register Bits */

#define	DCM_SME (short) 0x01		/* source in memory/io-  */
#define	DCM_SDE (short) 0x02		/* decrement source address */
#define	DCM_SIN (short) 0x04		/* increment source address */
#define	DCM_SWD (short) 0x08		/* source data in words/bytes- 
					 * for data channels this translates
					 * to longs/words- and must match
					 * jumper */
#define	DCM_DME (short) 0x10		/* destination in memory/io-  */
#define	DCM_DDE (short) 0x20		/* decrement destination address */
#define	DCM_DIN (short) 0x40		/* increment destination address */
#define	DCM_DWD (short) 0x80		/* destination data in words/bytes- 
					 * for data channels this translates
					 * to longs/words- and must match
					 * jumper */
#define	DCM_DLC (short) 0x0100		/* data list chaining */
#define	DCM_LLC (short) 0x0200		/* data list linked */
#define	DCM_SC (short) 0x0400		/* data list source/dest- */
#define	DCM_EOD (short) 0x0800		/* enable EOD pin output on last xfer*/
#define	DCM_EXT (short) 0x1000		/* enable termination on external EOD*/
#define	DCM_ECX (short) 0x2000		/* extended type I command block */
#define	DCM_SYS (short) 0x4000		/* source synchronization */
#define	DCM_SYD (short) 0x8000		/* destination synchronization */
					/* if DCM_SYS and DCM_SYD are both set
					 * then no synchronization
					 * if neither then type II command */

/* ADMAC Channel Command Type I Extended Register Bits */

#define	DCE_MA1	(short) 0x01
#define	DCE_MA2	(short)	0x02
#define	DCE_VE1	(short)	0x04
#define	DCE_VE2	(short)	0x08
#define	DCE_TRA	(short)	0x10
#define	DCE_STA	(short)	0x20

#define	DCE_WMA	(DCE_MA1 | DCE_MA2)
#define	DCE_BMA	DCE_MA2


/* ADMAC Channel Command Type II Register Bits */

#define	DC2_BC	(short) 0x01		/* if byte count exceeded */
#define DC2_ET	(short) 0x02		/* if external termination */
#define	DC2_MA	(short)	0x04		/* if match termination */
#define	DC2_VE	(short)	0x08		/* if verify mismatch */
#define	DC2_INV	(short)	0x10		/* invert status register bits */
#define	DC2_INT	(short)	0x0400		/* send interrupt same as EOD
					 * except bit is set in gsr */
#define	DC2_EOD	(short)	0x0800		/* send EOD */
#define	DC2_OP0	(short)	0x1000		
#define	DC2_OP1	(short)	0x2000		/* type II commands
					 * 00 = stop
					 * 01 = conditional stop
					 * 10 = conditional jump relative
					 * 11 = conditional jump direct */
#define	DC2_MAJ	(DC2_MA|DC2_OP0|DC2_OP1)/* jump on match condition */
#define	DC2_MAHLT	(DC2_MA|DC2_OP0)/* halt on match condition */
#define	DC2_UNMAJ	(DC2_INV|DC2_MA|DC2_OP0|DC2_OP1)
		/* jump on not match condition */
#define	DC2_UMHLT	(DC2_INV|DC2_MA|DC2_OP0)
		/* halt on unmatch condition */
#define	DC2_JMP	(DC2_BC|DC2_ET|DC2_MA|DC2_VE|DC2_OP0|DC2_OP1)
					/* jump always */
#define	DC2_NOP	(DC2_OP1)		/* no operation
					 * (continue to next command */
#define	DC2_HLT	(0)			/* halt channel program */

#define	DMA_BURST	11
#define	DMA_DELAY	0

/* ADMAC General Command Register Bits */

#define	DGC_NOP		0		/* nop */
#define	DGC_CON		1		/* continue channel */
#define	DGC_STM		2		/* start channel from memory space */
#define	DGC_STI		3		/* start channel from i/o space */
#define	DGC_STP		4		/* stop channel */
#define	DGC_HLT		7		/* halt/single step channel */
#define	DGC_RIT		0x08		/* clear channel interrupt */
#define	DGC_CH0		0x10
#define	DGC_CH1		0x20
#define	DGC_CH2		0x40
#define	DGC_CH3		0x80		/* affected channels */

#define	DAD_B23		0x80		/* address bit 23 does data xfers */

#define	DCB_TBUFF	0
#define	DCB_RBUFF	1
#define	DCB_MEMWD	2
#define	DCB_DIRWD	3
#define	DCB_MATCH	4
#define	DCB_TEST	5
#define	DCB_TRAN	6
#define	DCB_TRAMAT	7
#define	DCB_TRATST	8
#define	DCB_BKDUM	9
#define	DCB_MEMBT	10
#define	DCB_BTMAT	11
/*
#define	DCO_MEMWD	22
#define	DCO_DIRWD	23
 */
#define	DCO_MATCH	24
#define	DCO_TEST	25
#define	DCO_TRAN	26
#define	DCO_TRAMAT	27
#define	DCO_TRATST	28
/*
#define	DCO_BKDUM	29
#define	DCO_MEMBT	30
 */
#define	DCO_BTMAT	31




/*
 *      DMA controller command addressing translation
 */
#define REG_MODE_L(addr) 						\
        ( (short)  (((vtoph(0,addr)>>1) & 0x00fffe) | ((addr^1) & 1)) )
#define REG_MODE_H(addr)                                                \
        ( (short)  (((vtoph(0,addr)>>1) & 0x7f0000) >> 16) )

#define CMD_MODE_L(addr)  ( (short)   ((vtoph(0,addr)>>1) & 0x00fffe) )
#define CMD_MODE_H(addr)  ( (short)  (((vtoph(0,addr)>>1) & 0x7f0000) >> 16) )

#define DATA_MODE_L(addr)      (CMD_MODE_L(addr))
#define DATA_MODE_H(addr)      (CMD_MODE_H(addr) | DAD_B23)



#define XMIT_TIMEOUT         1000000L


