#include <sys/param.h>
#include <stdio.h>
#include <nlist.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <hcxif/if_com.h>
#include <hcxif/if_en.h>
#include <hcxif/if_ex.h>
#include <pronet/if_pn.h>
#include <errno.h>
extern int errno;
 
#define TCP_DATA_OFFSET 10
#define IP_OFFSET_MASK ~(IP_DF & IP_MF)

extern int 		opterr ;
extern int 		optind ;
extern char *		optarg ;
 
struct en_softc		en_softc ;
struct ex_softc		ex_softc ;
struct pn_softc		pn_softc ;
struct arpcom *		softc_ptr ;
struct en_debug		debug_buffer[ SZ_DEBUG_BUFFER ] ;
struct en_debug *	en_debug_ptr ;
struct ex_queue_base	ex_qb ;
 
char			up ;
char			broadcast ;
char			debug ;
char			loopback ;
char			point ;
char			trailers ;
char			running ;
char			arp ;

struct ether_arp *	arp_ptr ;
struct ip *		ip_ptr ;
struct tcphdr *         tcp_ptr;
int 			type ;
int 			length ;
unsigned short *	short_ptr ;
short			word ;
int 			words ;
int                     size ;

struct nlist		nl[] = {
#define	X_EN_SOFTC	0
	{ "_en_softc" },
#define	X_EX_SOFTC	1
	{ "_ex_softc" },
#define	X_PN_SOFTC	2
	{ "_pn_softc" },
	{ "" },
} ;

unsigned int 		i ;
unsigned int 		j ;
int	 		mf ;
struct	ifreq		ifr;
char			name[ 30 ] ;
int 			s ;
int                     error ;
 

void main( argc, argv )
	int 	argc;
	char **	argv;
{
	char	address ;
	char	en_debug = 0 ;
	char	addr_buf[ 20 ] ;
	int 	unit ;
	int 	c ;
	char *	ptr ;
	int 	counter ;
	int 	i ;
	char *  device_ptr ;
	int     debug_no ;

 
	if ( argc == 1 ) {
		printf("usage: %s controller [ -a xx.xx.xx.xx.xx.xx ] [ -d ] [ system ] [ core ]\n", argv[ 0 ] ) ;
		exit( 1 ) ;
	}

	device_ptr = argv[ 1 ] ;
	switch ( device_ptr[ 0 ] ) {
	case 'e' :
	case 'E' :
		switch ( device_ptr[ 1 ] ) {
		case 'n' :
		case 'N' :
			type = X_EN_SOFTC ;
			break ;
		case 'x' :
		case 'X' :
			type = X_EX_SOFTC ;
			break ;
		default :
			printf( "invalid type of controller\n" ) ;
			exit( 1 ) ;
		}
		break ;
	case 'p' :
	case 'P' :
		switch ( device_ptr[ 1 ] ) {
		case 'n' :
		case 'N' :
			type = X_PN_SOFTC ;
			break ;
		default :
			printf( "invalid type of controller\n" ) ;
			exit( 1 ) ;
		}
		break ;
	default :
		printf( "invalid type of controller\n" ) ;
		exit( 1 ) ;
	}

	unit = atoi( & device_ptr[ 2 ] ) ;
	argc -- ;
	argv ++ ;

	for ( ;; ) {

		c = getopt( argc, argv, "a:d" ) ;
		if ( c == EOF )
			break ;
		switch ( c ) {

		case 'a' :
			address = TRUE ;
			for ( i = 0 ;
			      (addr_buf[ i ] = optarg[ i ]) != 0 ;
			      ++ i ) ;
			break ;

		case 'd' :
		        switch ( type ) {
		        case X_EN_SOFTC:
		        case X_EX_SOFTC:
				en_debug = TRUE ;
				break ;
			case X_PN_SOFTC:
	                 	printf( "there is no debug for pn\n" ) ;
	                 	break ;
	                }
			break ;

		}
 
	}
	s = socket( AF_INET, SOCK_DGRAM, 0 ) ;
	if ( s < 0 ) {
		printf( "enconfig: socket error: " ) ;
		switch ( errno ) {
		case	ENOTSOCK :
			printf( "Not sock\n" ) ;
			break ;
		case	EDESTADDRREQ :
			printf( "Dest addr req\n" ) ;
			break ;
		case	EMSGSIZE :
			printf( "msg size\n" ) ;
			break ;
		case	EPROTOTYPE :
			printf( "Proto type\n" ) ;
			break ;
		case	ENOPROTOOPT :
			printf( "Proto opt\n" ) ;
			break ;
		case	EPROTONOSUPPORT :
			printf( "Proto no support\n" ) ;
			break ;
		case	ESOCKTNOSUPPORT :
			printf( "socket no support\n" ) ;
			break ;
		case	EOPNOTSUPP :
			printf( "Op not support\n" ) ;
			break ;
		case	EPFNOSUPPORT :
			printf( "PF no support\n" ) ;
			break ;
		case	EAFNOSUPPORT :
			printf( "AF no support\n" ) ;
			break ;
		case	EADDRINUSE :
			printf( "Addr in use\n" ) ;
			break ;
		case	EADDRNOTAVAIL :
			printf( "Addr not avail\n" ) ;
			break ;
		default:
			printf( "Socket error=%x\n", errno ) ;
			break ;
		}
		exit( 1 ) ;
	}
	if ( address ) {
		strncpy( ifr.ifr_name, device_ptr, sizeof( device_ptr ) ) ;
		if ( ioctl( s, SIOCGIFFLAGS, (caddr_t) & ifr ) < 0 ) {
			printf( "enconfig: ioctl( SIOCGIFFLAGS ) failed\n" ) ;
			exit( 1 ) ;
		}

		counter = 0 ;
		ptr = addr_buf ;
		while ( c = *ptr ) {
			if ( c != '.' ) {
				if ( (c >= '0') && (c <= '9') )
					c -= '0' ;
				else if ( (c >= 'A') && (c <= 'F') )
					c -= 0x37 ;
				else if ( (c >= 'a') && (c <= 'f') )
					c -= 0x57 ;
				else {
					ptr[ 1 ] = 0 ;
					printf(
				"enconfig: invalid hex number %s\n", ptr ) ;
					exit( 1 ) ;
				}
			}
			*(ptr ++) = c ;
			counter ++ ;
		}
		if ( counter > 17 ) {
			printf( "enconfig: address to large\n" ) ;
			exit( 1 ) ;
		}

		ptr = ifr.ifr_addr.sa_data ;
		addr_buf[ counter ] = '.' ;
		for ( i = 0 ; i < counter ; ++ i ) {
			c = 0 ;
			while ( addr_buf[ i ] != '.' ) {
				c <<= 4 ;
				c |= addr_buf[ i ] ;
				++ i ;
			}
			*(ptr ++) = c ;
		}

		strncpy( ifr.ifr_name, device_ptr, sizeof( device_ptr ) ) ;
		if ( ioctl( s, SIOCSETETADDR, (caddr_t)& ifr ) < 0 ) {
			printf( "enconfig: ioctl( SIOCSETETADDR ) failed\n");
			exit( 1 ) ;
		}
	} else {
		ptr = "/unix" ;
		if ( argc - optind > 0 ) {
			ptr = argv[ optind ] ;
			optind++;
		}
		nlist( ptr, nl ) ;
		if ( nl[ type ].n_value == 0 && nl[type].n_type == 0) {
			printf( "%s: no namelist.\n",ptr ) ;
			exit(1);
		}
		ptr = "/dev/kmem";
		if ( argc - optind > 0 ) {
			ptr = argv[ optind ] ;
		}
		mf = open( ptr, 0 ) ;
		if ( mf < 0 ) {
			printf( "cannot open %s\n",ptr ) ;
			exit( 1 ) ;
		}

		switch ( type ) {
		case X_EN_SOFTC :
			size = sizeof en_softc ;
			softc_ptr = (struct arpcom *) & en_softc ;
			break ;
		case X_EX_SOFTC :
			size = sizeof ex_softc ;
			softc_ptr = (struct arpcom *) & ex_softc ;
			break ;
		case X_PN_SOFTC :
			size = sizeof pn_softc ;
			softc_ptr = (struct arpcom *) & pn_softc ;
			break ;
		default:
			printf( "internal logic error\n" ) ;
			exit( 1 ) ;
			break ;
		}
		lseek( mf, (long)nl[ type ].n_value, 0 ) ;
		for ( i = unit + 1 ; i ; -- i ) {
			counter = read( mf, softc_ptr, size ) ;
			if ( counter < 0 ) {
				perror( "trouble reading core image" ) ;
				exit( 1 ) ;
			}
		}

		/*
		 * Print out the common portion of the softc struct.
		 */
		switch ( type ) {
		case X_EN_SOFTC :
		case X_EX_SOFTC :
   		printf( "ethernet address = %02x.%02x.%02x.%02x.%02x.%02x\n",
			softc_ptr ->ac_enaddr[ 0 ], softc_ptr ->ac_enaddr[ 1 ],
			softc_ptr ->ac_enaddr[ 2 ], softc_ptr ->ac_enaddr[ 3 ],
			softc_ptr ->ac_enaddr[ 4 ], softc_ptr ->ac_enaddr[ 5 ] ) ;
			break ;
		case X_PN_SOFTC :
   		printf( "pronet address = %02x\n",
			softc_ptr ->ac_enaddr[ 5 ] ) ;
			break ;
		}

		up        =  ((softc_ptr->ac_if.if_flags & IFF_UP) != 0);
		broadcast =  ((softc_ptr->ac_if.if_flags & IFF_BROADCAST) != 0);
		debug     =  ((softc_ptr->ac_if.if_flags & IFF_DEBUG) != 0);
		loopback  =  ((softc_ptr->ac_if.if_flags & IFF_LOOPBACK) != 0);
		printf(
	"FLAGS: up    = %x, broadcast = %x, debug   = %x, loopback = %x\n",
			up, broadcast, debug, loopback ) ;
		point     =  ((softc_ptr->ac_if.if_flags & IFF_POINTOPOINT) != 0);
		trailers  = !((softc_ptr->ac_if.if_flags & IFF_NOTRAILERS) != 0);
		running   =  ((softc_ptr->ac_if.if_flags & IFF_RUNNING) != 0);
		arp       = !((softc_ptr->ac_if.if_flags & IFF_NOARP) != 0);
		printf(
	"       point = %x, trailers  = %x, running = %x, arp   = %x\n\n",
			point, trailers, running, arp ) ;

		printf( "input packets     = %7d,  ",
			softc_ptr->ac_if.if_ipackets ) ;
		printf( "output packets   = %7d\n",
			softc_ptr->ac_if.if_opackets);
		printf( "      errors      = %7d, ",
			softc_ptr->ac_if.if_ierrors);
		printf( "      errors      = %7d\n",
			softc_ptr->ac_if.if_oerrors);
		printf( "                           " ) ;

		printf( "        collisions  = %7d\n",
			softc_ptr->ac_if.if_collisions ) ;
		/*
		* End of the common portion of the soft structures.
		*/

		switch ( type ) {
		case X_EN_SOFTC :

		printf( "      odd         = %7d, ", en_softc.error_odd ) ;

		printf( "      queued      = %7d\n", en_softc.queued_packets ) ;

		printf( "      crc         = %7d, ", en_softc.error_crc ) ;

		printf( "      queue depth = %7d\n", en_softc.max_queue_depth ) ;

		printf( "      over run    = %7d\n", en_softc.error_over_run ) ;

		printf( "      too short   = %7d\n\n",
			en_softc.error_too_short ) ;

		printf( "xmit_avail_head = %2d, ", en_softc.xmit_avail_head ) ;

		printf( "xmit_avail_tail = %2d, ", en_softc.xmit_avail_tail ) ;

		printf( "recv_buffer = %2d\n", en_softc.recv_buffer ) ;

			if ( en_softc.monitor )
				printf( "monitor is set\n" ) ;
			break ;
		case X_EX_SOFTC:
			strncpy( ifr.ifr_name, device_ptr, sizeof(device_ptr) );
			error = ioctl( s, SIOCGSTAT, (caddr_t)& ifr ) ;
			if ( error < 0 ) {
				printf("enconfig: ioctl( SIOCGSTAT ) failed\n");
				printf("error = 0x%x, errno = 0x%x\n",
					error, errno ) ;
			} else {
			        /* print the excelan statistics */
				lseek(mf, (long)ex_softc.ex_qb_ptr, 0);
				for (i = unit + 1; i; -- i) {
					counter = read(mf, &ex_qb, sizeof ex_qb);
					if (counter < 0) {
						perror("trouble reading /dev/kmem");
					}
				}
			        printf( "\n\t%7d\tframes sent without errors\n",
                     			 ex_qb.sa.sa_fsent ) ;
		        	printf( "\t%7d\tframes aborted excess collisions\n",
                     			 ex_qb.sa.sa_xsclsn ) ;
		        	printf( "\t%7d\tframes subject to heartbeat failure\n",
                    			ex_qb.sa.sa_nsqe ) ;
		        	printf( "\t%7d\tframes received no errors\n",
                     			 ex_qb.sa.sa_frcvd ) ;
		        	printf( "\t%7d\tframes received alignment error\n",
                    			 ex_qb.sa.sa_align ) ;
		        	printf( "\t%7d\tframes received crc error\n",
                     			 ex_qb.sa.sa_crc ) ;
		        	printf( "\t%7d\tframes lost\n",
                    			 ex_qb.sa.sa_flost ) ;
                    	}
                    	break ;
                case X_PN_SOFTC:
			printf( "\t%7d\tnumber of ring inits\n",
				pn_softc.pn_init ) ;
			printf( "\t%7d\tnumber of packets refused \n",
				pn_softc.pn_refused ) ;
			printf( "\t%7d\tnumber of transmit timeouts \n",
				pn_softc.pn_timeouts ) ;
			printf( "\t%7d\tnumber of output timeouts \n",
			 	pn_softc.pn_otimeout ) ;
			printf( "\t%7d\tnumber of bad input formats \n",
				pn_softc.pn_ibadf ) ;
			printf( "\t%7d\tnumber of parity errors \n",
				pn_softc.pn_parity ) ;
                        break ;
                }

		if ( ! en_debug )
			exit( 1 ) ;

		switch ( type ) {
		case X_EN_SOFTC :
			if ( en_softc.start_debug_ptr ) {
			        debug_no = en_softc.debug_ptr -
			        	   en_softc.start_debug_ptr ;
				lseek( mf, (long)en_softc.start_debug_ptr, 0 ) ;
			} else {
			        printf( "there is no en debug buffer\n" ) ;
				exit( 1 ) ;
			}
			break ;
		case X_EX_SOFTC :
			if ( ex_softc.start_debug_ptr ) {
			        debug_no = ex_softc.debug_ptr -
			        	   ex_softc.start_debug_ptr ;
				lseek( mf,
				       (long)ex_softc.start_debug_ptr,
				       0 ) ;
			} else {
			        printf( "there is no ex debug buffer\n" ) ;
				exit( 1 ) ;
			}
			break ;
		}
		printf( "Current buffer# %d\n", debug_no ) ;
		read( mf, debug_buffer, sizeof debug_buffer ) ;
		en_debug_ptr = debug_buffer ;

		for ( i = 0 ;
		      i < SZ_DEBUG_BUFFER ;
		      ++ i, ++ en_debug_ptr ) {
			if ( en_debug_ptr -> cmnd_status == 0 )
				break ;
			printf( "buffer #%d, frame #%d ", i,
				en_debug_ptr -> counter ) ;
			if ( en_debug_ptr -> direction == EN_XMIT )
				printf( "XMIT " ) ;
			else if ( en_debug_ptr -> direction == EN_RECV )
				printf( "RECV " ) ;
			else
				printf( "UNKN " ) ;
			length = en_debug_ptr->cmnd_status&CS_BYTE_CNT ;
			printf( "csr = 0x%04x, length = %x\n",
				en_debug_ptr -> cmnd_status, length ) ;

			printf("   dest = %02x.%02x.%02x.%02x.%02x.%02x, ",
			en_debug_ptr -> hdr_dhost[ 0 ] >> 8 & 0xff,
			en_debug_ptr -> hdr_dhost[ 0 ] & 0xff,
			en_debug_ptr -> hdr_dhost[ 1 ] >> 8 & 0xff,
			en_debug_ptr -> hdr_dhost[ 1 ] & 0xff,
			en_debug_ptr -> hdr_dhost[ 2 ] >> 8 & 0xff,
			en_debug_ptr -> hdr_dhost[ 2 ]  & 0xff ) ;

			printf("source = %02x.%02x.%02x.%02x.%02x.%02x, ",
				en_debug_ptr -> hdr_shost[ 0 ] >> 8 & 0xff,
				en_debug_ptr -> hdr_shost[ 0 ] & 0xff,
				en_debug_ptr -> hdr_shost[ 1 ] >> 8 & 0xff,
				en_debug_ptr -> hdr_shost[ 1 ] & 0xff,
				en_debug_ptr -> hdr_shost[ 2 ] >> 8 & 0xff,
				en_debug_ptr -> hdr_shost[ 2 ] & 0xff ) ;

			type = en_debug_ptr -> hdr_type ;
			printf("type = 0x%04x ", type ) ;

			switch ( type ) {

			case ETHERTYPE_PUP:
				printf( "(pup)\n" ) ;
				short_ptr = en_debug_ptr -> data ;
				length = SZ_DEBUG_DATA ;
				break ;

			case ETHERTYPE_IP:
				printf( "(ip)\n" ) ;
				ip_ptr=(struct ip *)en_debug_ptr->data ;
				printf(
			"   header length = %01x, header version =%01x\n",
					ip_ptr -> ip_hl, ip_ptr -> ip_v ) ;
				printf( "   type { service = %02x, length = %04x\n",
					ip_ptr -> ip_tos, ip_ptr -> ip_len ) ;
				printf( "   id = %04x, {fset = %04x\n",
					ip_ptr -> ip_id, ip_ptr -> ip_off ) ;
				printf( "   time to live = %02x, protocol = %02x\n",
					ip_ptr -> ip_ttl, ip_ptr -> ip_p ) ;
				printf( "   checksum = %04x\n",
						ip_ptr -> ip_sum ) ;
				printf("   source address      = 0x%08x\n",
						ip_ptr -> ip_src.s_addr ) ;
				printf("   destination address = 0x%08x\n   ",
						ip_ptr -> ip_dst.s_addr ) ;

				/* Output TCP header only if ip "data" {fset is
				 * zero.  When ip data {fset is not zero, {
				 * this is an ip datagram fragment and the TCP
				 * header is not contained in this datagram.
				 */
				if ( ( (ip_ptr->ip_off & IP_OFFSET_MASK) == 0) &&
						(ip_ptr->ip_p == IPPROTO_TCP) ) {
					tcp_ptr = (struct tcphdr *)
						& en_debug_ptr -> data[ TCP_DATA_OFFSET ] ;
					printf(
				"\n   tcp source port = %x, tcp dest port = %x\n",
						tcp_ptr->th_sport,
						tcp_ptr->th_dport);
					printf(
				"   tcp seq number = %x, tcp ack number = %x\n",
						tcp_ptr->th_seq, tcp_ptr->th_ack);
					printf(
				"   tcp data {fset = %x, tcp flags = %x ",
						tcp_ptr->th_off, tcp_ptr->th_flags);
					if (tcp_ptr->th_flags & TH_FIN)
						printf("FIN ");
					if (tcp_ptr->th_flags & TH_SYN)
						printf("SYN ");
					if (tcp_ptr->th_flags & TH_RST)
						printf("RST ");
					if (tcp_ptr->th_flags & TH_PUSH)
						printf("PUSH ");
					if (tcp_ptr->th_flags & TH_ACK)
						printf("ACK ");
					if (tcp_ptr->th_flags & TH_URG)
						printf("URG ");
					printf(
		"\n   tcp window = %x, tcp checksum = %x, tcp urgent ptr = %x\n",
						tcp_ptr->th_win,
						tcp_ptr->th_sum,
						tcp_ptr->th_urp);

					length = MIN(length, 74);
					length = (length - 14 - 40 +1) >> 1;
					short_ptr = &en_debug_ptr->data[TCP_DATA_OFFSET + 10];
				/* Here for protocols other than TCP */
				} else {
					length = MIN( length, 74 ) ;
					length = (length-14-20+1) >> 1 ;
					short_ptr = & en_debug_ptr->data[ 10 ] ;
				}
				break;
			case ETHERTYPE_ARP:
				printf( "(arp)\n" ) ;
				arp_ptr = (struct ether_arp *)
					en_debug_ptr -> data ;
				printf(
		"   address format  = %04x, proto format      = %04x\n",
					arp_ptr -> arp_hrd,
					arp_ptr -> arp_pro ) ;
				printf(
		"   hdr addr length = %02x, proto addr length = %02x\n",
					arp_ptr -> arp_hln,
					arp_ptr -> arp_pln ) ;
				if ( arp_ptr -> arp_op == ARPOP_REQUEST )
					printf( "   operation = request\n" ) ;
				else
					printf( "   operation = reply\n" ) ;
				printf(
		"   sender hardware address = %02x.%02x.%02x.%02x.%02x.%02x\n",
					arp_ptr -> arp_sha[ 0 ],
					arp_ptr -> arp_sha[ 1 ],
					arp_ptr -> arp_sha[ 2 ],
					arp_ptr -> arp_sha[ 3 ],
					arp_ptr -> arp_sha[ 4 ],
					arp_ptr -> arp_sha[ 5 ] ) ;
				printf(
		"   sender protocol address = %02x.%02x.%02x.%02x\n",
					arp_ptr -> arp_spa[ 0 ],
					arp_ptr -> arp_spa[ 1 ],
					arp_ptr -> arp_spa[ 2 ],
					arp_ptr -> arp_spa[ 3 ] ) ;
				printf(
		"   target hardware address = %02x.%02x.%02x.%02x.%02x.%02x\n",
					arp_ptr -> arp_tha[ 0 ],
					arp_ptr -> arp_tha[ 1 ],
					arp_ptr -> arp_tha[ 2 ],
					arp_ptr -> arp_tha[ 3 ],
					arp_ptr -> arp_tha[ 4 ],
					arp_ptr -> arp_tha[ 5 ] ) ;
				printf(
		"   target hardware address = %02x.%02x.%02x.%02x\n",
					arp_ptr -> arp_tpa[ 0 ],
					arp_ptr -> arp_tpa[ 1 ],
					arp_ptr -> arp_tpa[ 2 ],
					arp_ptr -> arp_tpa[ 3 ] ) ;
				length = MIN( length, 74 ) ;
				length = (length - 14 - 28 + 1) >> 1 ;
				short_ptr = & en_debug_ptr -> data[ 14 ];
				break ;
			default :
				if (type & 0xfff0 == ETHERTYPE_TRAIL){
					printf( "(trailer)" ) ;
				}
				printf( "\n" ) ;
				short_ptr = en_debug_ptr -> data ;
				length = SZ_DEBUG_DATA ;
				break ;
			}
			while ( length > 0 ) {
				printf( "\n   " ) ;
				words = 0 ;
				for ( ;; ) {
					word = *short_ptr ;
					printf( "%02.2x %02.2x ",
				(word >> 8) & 0xff, word & 0xff);
					++ short_ptr ;
					-- length ;
					++ words ;
					if ( (length < 1) || (words > 7) )
						break ;
				}
			}
			printf( "\n\n" ) ;
		}
	}
}
