#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

/*
 * Strings which identify various microcode files.
 */
char cpuwcs[]  = { 0x00, 0x40, 0x01, 0x0E, 0x05, 0x40, 0x00, 0x00,
		   0x00, 0x43 } ;
char cpupoc[]  = { 0x00, 0x40, 0x01, 0x0E, 0x05, 0x40, 0x00, 0x00,
		   0x30, 0x40 } ;
char ohdcwcs[] = { 0x00, 0x00, 0x0E, 0x28, 0x00, 0x00, 0x00, 0x18,
		   0x00, 0x00, 0x61, 0xFE } ;
char hdcwcs[]  = { 0x00, 0x00, 0x1F, 0x1C, 0x00, 0x00, 0x20, 0x18,
		   0x00, 0x00, 0x61, 0xFE } ;


int fd ;

main( argc, argv)
char ** argv;
{
	register int i ;
	register int ret ;
	struct stat statbuf;
	char buf[100];


	for ( i= 1 ; i < argc ; i ++ ) {

	 	printf( "%12s: ", argv[i]);

	        ret = lstat( argv[i], &statbuf );
	        if ( ret < 0 ) {
	                perror("");
	                continue;
	        }

	        if ( (statbuf.st_mode & S_IFMT) != S_IFREG ) {
	        	printf( "not a regular file\n" );
	        	continue;
	        }

		fd = open( argv[i], O_RDONLY );
		if ( fd < 0 ) {
			perror("");
			continue;
		}

		ret = read( fd, buf, 100);
		if ( ret < 0 ) {
		        perror("read");
		        continue ;
		}

		/*
		 * HCX5/7/9 cpu microcode
		 */
	        if ( bcmp( buf, cpuwcs, sizeof( cpuwcs ) ) == 0 ) {
	        	docpu( "wcs" );

		/*
		 * HCX5/7/9 cpu power-on-confidence
		 */
	        } else if ( bcmp( buf, cpupoc, sizeof( cpupoc ) ) == 0 ) {
	                docpu( "poc" );

		/*
		 * Old HDC microcode
		 */
	        } else if ( bcmp( buf, ohdcwcs, sizeof( ohdcwcs ) ) == 0 ) {
	        	dohdc( 0x4B );

		/*
		 * New HDC microcode
		 */
	        } else if ( bcmp( buf, hdcwcs, sizeof( hdcwcs ) ) == 0 ) {
	        	dohdc( 1 );

		} else {
		        printf ( "not a microcode file\n" ) ;
		}

	}
}

docpu( wcstype )
char * wcstype ;
{
	int ret ;
	int year ;
	int day ;
	char buf[100];

	printf( "cpu %s version: ", wcstype );

	ret = lseek( fd, 0264, 0);
	if ( ret < 0 ) {
	        perror("lseek");
	        return ;
	}

	ret = read( fd, buf, 4);
	if ( ret < 0 ) {
	        perror("read");
	        return ;
	}


	year =  ((buf[1]&0xF)*10) + ((buf[2]>>4)&0xf);
	day =   ((buf[0]&0xF)*10) + ((buf[1]>>4)&0xf);
	if ( year < 80 || year > 99 || day < 0 || day > 31 ) {
		printf( "unknown\n" );
		return ;
	}

	printf( "%x\.%2.2x (%x/%d/%d)\n", buf[2]&0xf, buf[3],
	        (buf[0]>>4)&0xF, day, year );

}

dohdc( word )
int word ;
{
/*
 * A word of hdc wcs is 48 bits long with 16 bits packed
 * into each longword of the file. In other words, 12 bytes per wcs location.
 */
#define HDCWORD 12

	int ret ;
	int ver ;
	char buf[HDCWORD];

	printf( "hdc wcs version: " );

	ret = lseek( fd, word*HDCWORD, 0);
	if ( ret < 0 ) {
	        perror("lseek");
	        return;
	}

	ret = read( fd, buf, HDCWORD);
	if ( ret < 0 ) {
	        perror("read");
	        return;
	}

	ver=((buf[2]&0x1F)<<3)|((buf[3]>>5)&0x7);
	printf( "%2.2x\n", ver );

}
extern int errno, sys_nerr;
extern char *sys_errlist[];


perror(s)
char	*s;
{
	register char *c;
	register int n;

	c = "Unknown error";
	if(errno < sys_nerr)
		c = sys_errlist[errno];

	n = s ? strlen(s) : 0;
	if(n) {
	        printf( "%s: ", s );
	}

	printf( "%s\n", c );
}
