static	char *sccsid = "@(#)mt.c	4.8 (Berkeley) 83/05/08";

/*
 * mt -- magnetic tape manipulation program
 */
#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/mtio.h>
/*
#include <sys/xymacro.h>
#include <sys/xydata.h>
*/
#include	<sys/param.h>
#include	<sys/systm.h>
#include	<ml/mtpr.h>
#include	<sys/vm.h>
#include	<sys/buf.h>
#include 	<sys/scsi.h>
#include 	<sys/qt.h>

#include <sys/ioctl.h>
#include <sys/file.h>

#define	equal(s1,s2)	(strcmp(s1, s2) == 0)

struct commands {
	char *c_name;
	int c_code;
	int c_ronly;
} com[] = {
	{ "weof",	MTWEOF,	0 },
	{ "eof",	MTWEOF,	0 },
	{ "fsf",	MTFSF,	1 },
	{ "bsf",	MTBSF,	1 },
	{ "fsr",	MTFSR,	1 },
	{ "bsr",	MTBSR,	1 },
	{ "rewind",	MTREW,	1 },
	{ "offline",	MTOFFL,	1 },
	{ "rewoffl",	MTOFFL,	1 },
	{ "status",	MTNOP,	1 },
	{ "settype",	MTSETTYPE, 1 },
	{ "retension",	MTRETEN, 1 },
	{ 0 }
};

/* These names correspond to #defines in mtio.h.*/
char *drivename[NUMDRIVETYPE] =
{
/*CIPHERF880*/ 		"cipher880",
/*CIPHERF890*/ 		"cipher890",
/*CIPHERM990*/ 		"cipher990",
/*CDC92181*/   		"cdc92181",
/*CDC92185*/   		"cdc92185",
/*FUJI2242*/   		"fuji2242",
/*KENNEDY9400*/		"kennedy9400",
/*KENNEDY9600*/		"kennedy9600",
/*PERTECFS1000*/	"pertecfs1000",
/*PERTECFS2000*/	"pertecfs2000",
/*STC2920*/     	"stc2920",
/*TELEXSHAMROCK	*/	"telexshamrock",
/*THORN9800*/     	"thorn9800",
/*THORN9900*/     	"thorn9900",
/*HARRIS6330*/		"6330",
/*HARRIS6331*/		"6331",
/*HARRIS6350*/		"6350",
/*HARRIS6351*/		"6351",
/*HARRIS6740*/		"6740",
/*CIPHERF891*/ 		"cipher891",
} ;

int mtfd;
struct mtop mt_com;
struct mtget mt_status;
char *tape;

void main(argc, argv)
	char **argv;
{
	char *getenv();
	register char *cp;
	register struct commands *comp;

	if (argc > 2 && (equal(argv[1], "-t") || equal(argv[1], "-f"))) {
		argc -= 2;
		tape = argv[2];
		argv += 2;
	}
	else {
		if ((tape = getenv("TAPE")) == NULL)
			tape = DEFTAPE;
	}

	if (argc < 2) {
		fprintf(stderr, "usage: mt [ -f device ] command [ count ]\n");
		exit(1);
	}

	cp = argv[1];
	for (comp = com; comp->c_name != NULL; comp++)
		if (strncmp(cp, comp->c_name, strlen(cp)) == 0)
			break;
	if (comp->c_name == NULL) {
		fprintf(stderr, "mt: don't grok \"%s\"\n", cp);
		exit(1);
	}

	if ( comp->c_code == MTSETTYPE )
	{
		int i ;

		for (i=0; i<NUMDRIVETYPE; i++ )
			if ( strcmp(drivename[i], argv[2]) == 0 )
				break ;
		if ( i==NUMDRIVETYPE )
		{
			fprintf(stderr, "mt: unknown drive type : '%s'\n",
				argv[2] ) ;
			exit(1);
		}
		mt_com.mt_count = i ;
		if ((mtfd = open(tape, O_SETTYPE | 
			(comp->c_ronly ? O_RDONLY : O_RDWR) ) )<0)
		{
			perror(tape);
			exit(1);
		}
	}
	else
	{
		if ((mtfd = open(tape, comp->c_ronly ? 0 : 2)) < 0)
		{
			perror(tape);
			exit(1);
		}

		if (comp->c_code != MTNOP)
		{
			mt_com.mt_count = (argc > 2 ? atoi(argv[2]) : 1);
			if (mt_com.mt_count < 0)
			{
				fprintf(stderr, "mt: negative repeat count\n");
				exit(1);
			}
		}
		else
		{
			if (ioctl(mtfd, MTIOCGET, (char *)&mt_status) < 0)
			{
				perror("mt");
				exit(2);
			}
			status(&mt_status);
			exit(0);
		}
	}

	mt_com.mt_op = comp->c_code;
	if (ioctl(mtfd, MTIOCTOP, &mt_com) < 0)
	{
		fprintf(stderr, "%s %s %d ", tape, comp->c_name, mt_com.mt_count);
		perror("failed");
		exit(2);
	}

}

#ifdef vax
#include <vaxmba/mtreg.h>
#include <vaxmba/htreg.h>
#include <vaxuba/utreg.h>
#include <vaxuba/tmreg.h>
#undef b_repcnt	/* argh */
#include <vaxuba/tsreg.h>
#endif

#ifdef sun
#include <sundev/tmreg.h>
#include <sundev/arreg.h>
#endif

#ifdef hcx
/* disgusting */
#define CYER_BITS "\006CR\007CC\010CE\012P\013FB\014RDY\015EOT\016LP\017OL\020FM"
#define CYDS_BITS "\001BANK\003REV\004100ips\005STREAM\01016bits\015MAIL\016INTR\017LINK\020LOCK"
#define XYER_BITS ""
#define XYDS_BITS "\001ONLIN\002DRRDY\003FBSY\004DBSY\005REW\006BOT\007HISD\010GC\011WPT\012EOT\013PEID\014FMK\015RLL\016RLS\017CER\020HER"
#endif

struct tape_desc {
	short	t_type;		/* type of magtape device */
	char	*t_name;	/* printing name */
	char	*t_dsbits;	/* "drive status" register */
	char	*t_erbits;	/* "error" register */
} tapes[] = {

#ifdef vax
	{ MT_ISTS,	"ts11",		0,		TSXS0_BITS },
	{ MT_ISHT,	"tm03",		HTDS_BITS,	HTER_BITS },
	{ MT_ISTM,	"tm11",		0,		TMER_BITS },
	{ MT_ISMT,	"tu78",		MTDS_BITS,	0 },
	{ MT_ISUT,	"tu45",		UTDS_BITS,	UTER_BITS },
#endif

#ifdef sun
	{ MT_ISCPC,	"TapeMaster",	TMS_BITS,	0 },
	{ MT_ISAR,	"Archive",	ARCH_CTRL_BITS,	ARCH_BITS },
#endif

#ifdef hcx
	{ MT_ISCY,	"Cipher",	CYDS_BITS,	CYER_BITS },
	{ MT_ISXYLOGICS,"Xylogics",	XYDS_BITS,	XYER_BITS },
	{ MT_ISQT,"Quarter Inch Cartridge", 0, 0	 },
#endif

	{ 0 }
};

/*
 * Interpret the status buffer returned.
 */
status(bp)
	register struct mtget *bp;
{
	register struct tape_desc *mt;

	for (mt = tapes; mt->t_type; mt++)
		if (mt->t_type == bp->mt_type)
			break;
	if (mt->t_type == 0) {
		printf("unknown tape drive type (%d)\n", bp->mt_type);
		return;
	}

	printf("%s tape drive, residual=%d\n", mt->t_name, bp->mt_resid);
	if (bp->mt_type == MT_ISQT ) {
		printf("ds=%x",bp->mt_dsreg);
		switch (bp->mt_dsreg) {
			case S_DRIVE :
				printf(">DRIVE\n");
				printf("er=%x ",bp->mt_erreg);
				switch(bp->mt_erreg) {
					case SC_NS :
						printf("No error");
						break;
					case SC_HE :
						printf("Non-recoverable hardware error");
						break;
					case SC_DNR :
						printf("Drive not ready");
						break;
					case SC_MNL :
						printf("Media not loaded");
						break;
					case SC_EOT :
						printf("End of tape");
						break;
					case SC_DT :
						printf("Drive timeout");
						break;
					default:
						printf("Unknown drive code");
						break;
				}
				break;
			case S_TARGET :
				printf(">TARGET\n");
				printf("er=%x",bp->mt_erreg);
				switch(bp->mt_erreg) {
					case SC_UDE :
					case SC_BBF :
						printf("Uncorrectable data error");
						break;
					case SC_BNF :
						printf("Block not found");
						break;
					case SC_DTE :
						printf("DMA timeout");
						break;
					case SC_WP :
						printf("Drive write protected");
						break;
					case SC_CDC :
						printf("Corrected data error");
						break;
					case SC_FMD :
						printf("EOF");
						break;
					case SC_CE :
						printf("Data compare error");
						break;
					default:
						printf("Unknown target code");
						break;
				}
				break;
			case S_SYS :
				printf(">SYS\n");
				printf("er=%x",bp->mt_erreg);
				switch(bp->mt_erreg) {
					case SC_IC :
						printf("Unauthorized media or invalid command");
						break;
					default:
						printf("Unknown system code");
						break;
				}
				break;
			case S_VENDOR :
				printf(">VENDOR\n");
				printf("er=%x",bp->mt_erreg);
				switch(bp->mt_erreg) {
					case SC_UA :
						printf("UA");
						break;
					case SC_CT :
						printf("Command timeout");
						break;
					case SC_AE :
						printf("Can only write at end of media");
						break;
					case SC_REOM :
						printf("REOM");
						break;
					default:
						printf("Unknown vendor code");
						break;
				}
				break;
			default :
				printf(">UNKNOWN\n");
				printf("er=%x",bp->mt_erreg);
				printf("Unknown error class\n");
				break;
		}
		printf("\n");
		return;
	}
	printreg("ds", bp->mt_dsreg, mt->t_dsbits);
	if ( bp->mt_type == MT_ISXYLOGICS )
		printf("\ner %x", bp->mt_erreg ) ;
	else
		printreg("\ner", bp->mt_erreg, mt->t_erbits);
	putchar('\n');
}

/*
 * Print a register a la the %b format of the kernel's printf.
 */
printreg(s, v, bits)
	char *s;
	register unsigned short v;
	register char *bits;
{
	register int i, any = 0;
	register char c;

	if (bits && *bits == 8)
		printf("%s=%o", s, v);
	else
		printf("%s=%x", s, v);
	bits++;

	if (v && bits) {
		putchar('<');
		while (i = *bits++) {
			if (v & (1 << (i-1))) {
				if (any)
					putchar(',');
				any = 1;
				for (; (c = *bits) > 32; bits++)
					putchar(c);
			} else
				for (; *bits > 32; bits++)
					continue;
		}
		putchar('>');
	}
}
