/*	sadc.c 1.10 of 9/23/83	*/
/*
	sadc.c - writes system activity binary data from /dev/kmem to a
		file or stdout.
	Usage: sadc [t n] [file]
		if t and n are not specified, it writes
		a dummy record to data file. This usage is
		particularly used at system booting.
		If t and n are specified, it writes system data n times to
		file every t seconds.
		In both cases, if file is not specified, it writes
		data to stdout.
*/

#include <nlist.h>
#include <sys/time.h>
#include <sys/param.h>
#define KERNEL
#include <sys/file.h>
#undef KERNEL
#include <sys/vnode.h>
#include <ufs/inode.h>
#include <sys/text.h>
#include <sys/proc.h>
#include <sys/iobuf.h>
#include <sys/stat.h>
#include <sys/elog.h>
#include <sys/sysinfo.h>
#include "sa.h"

struct stat ubuf,syb;
struct syserr err;

struct iotime idisk[NDCTR * NDDRV];
struct iotime icytape[NCYTCTR * NCYTDRV];
struct iotime ixytape[NXYTCTR * NXYTDRV];
int 	ndisk, ntape, ncytape, nxytape, inodes, files, procs, texts, atbl;

struct sa d;
char *flnm = "/tmp/sa.adrfl";
char *loc, *origloc;
static int fa = 0;
unsigned tblloc;
char *malloc();
static int tblmap[X_SYSINFO];

static int recsz,tmpsz;
long time();
int i,j,k;
long lseek();
int f;

static char Sccsid[]="@(#)sadc.c	1.10";

main(argc, argv)
char **argv;
int argc;
{

	int ct, n;
	unsigned ti;
	int fd;
	long min;
	long rem;
	struct stat buf;
	char *fname;
	struct inode 	*ip;
	struct file	*fp;
	struct text	*tp;
	struct proc 	*pp;

	ct = argc >= 3? atoi(argv[2]): 0;
	min = time((long *) 0);
	ti = argc >= 3? atoi(argv[1]): 0;

/*	check if the address file is there and not older
	than /unix, if so, read the offsets in.
	Otherwise, search /unix name list to get the offsets
	and create an address file for later usage.
	*/
	stat ("/unix",&syb);
	if ((stat(flnm,&ubuf) == -1) ||
		 (ubuf.st_mtime <= syb.st_ctime))
			goto creafl;
	fa = open(flnm,2);
	if (read(fa,setup,sizeof setup) != sizeof setup){
		close(fa);
		unlink(flnm);
creafl:
		i = umask(0);
		fa = open(flnm, O_RDWR|O_CREAT|O_TRUNC, 0664);
		umask(i);
		chown(flnm, 0, getegid());

/*		search name list to get offsets	*/
		nlist("/unix",setup);

/*		write offsets to address file	*/
		write(fa,setup,sizeof setup);
		close (fa);
	}

/*	open /dev/kmem	*/
	if((f = open("/dev/kmem", 0)) == -1)
		perrexit();

/*	get number of disk drives and tape drives */
	lseek(f, (long)setup[X_NDSK].n_value, 0);
	if (read(f, &ndisk, sizeof(ndisk)) == -1)
		perrexit();
	if ( (long)setup[X_NCY].n_value ) {
		lseek(f, (long)setup[X_NCY].n_value, 0);
		if (read(f, &ncytape, sizeof(ncytape)) == -1)
			perrexit();
	}
	else 
		ncytape = 0 ;
	if ( (long)setup[X_NXYD].n_value ) {
		lseek(f, (long)setup[X_NXYD].n_value, 0);
		if (read(f, &nxytape, sizeof(nxytape)) == -1)
			perrexit();
	}
	else
		nxytape = 0 ;
	ntape = ncytape + nxytape ;


/*	construct tblmap and compute record size	*/
	tblmap[X_VDSTAT] = ndisk;
	tblmap[X_CYSTAT] = ncytape;
	tblmap[X_XYSTAT] = nxytape;
	recsz = sizeof(struct sa) - sizeof d.devio + (ndisk + ntape) *
		sizeof(d.devio[0]);

	if (argc == 3 || argc == 1){

/*	no data file is specified, direct data to stdout	*/
		fd = 1;
		/*	write header record	*/
		write(fd,tblmap,sizeof tblmap);
	}
	else {
		fname = argc==2? argv[1]: argv[3];
		/*	check if the data file is there	*/
		/*	check if data file is too old	*/
		if ((stat(fname,&buf) == -1) ||
		 ((min - buf.st_mtime) > 86400))
			goto credfl;
		if ((fd = open(fname,2)) == -1){
credfl:

/*			data file does not exist:
			create one and write the header record.	*/
			if ((fd = creat(fname,00644)) == -1)
				perrexit();
			close(fd);
			fd = open (fname,2);
			lseek(fd,0L,0);
			write (fd,tblmap,sizeof tblmap);
		}
		else{

/*			data file exist:
			position the write pointer to the last good record.  */
			lseek(fd,-(long)((buf.st_size - sizeof tblmap) % recsz),2);
		}
	}

/*	if n =0 , write the additional dummy record	*/
	if (ct == 0){
		for (i=0;i<4;i++)
			d.si.cpu[i] = -300;
		d.ts = min;
		write(fd,&d,recsz);
	}

/*	get memory for tables	*/

	lseek(f, (long)setup[X_NINODE].n_value, 0);
	if (read(f, &inodes, sizeof(int)) == -1)
		perrexit();
	lseek(f, (long)setup[X_NFILE].n_value, 0);
	if (read(f, &files, sizeof(int)) == -1)
		perrexit();
	lseek(f, (long)setup[X_NTEXT].n_value, 0);
	if (read(f, &texts, sizeof(int)) == -1)
		perrexit();
	lseek(f, (long)setup[X_NPROC].n_value, 0);
	if (read(f, &procs, sizeof(int)) == -1)
		perrexit();

	if (tblloc < sizeof(struct inode) * inodes);
		tblloc += sizeof(struct inode) * inodes;
	if (tblloc < sizeof(struct file) * files);
		tblloc += sizeof (struct file) * files;
	if (tblloc < sizeof (struct text) * texts);
		tblloc += sizeof (struct text) * texts;
	if (tblloc < sizeof (struct proc) * procs);
		tblloc += sizeof (struct proc) * procs;
	origloc = loc = malloc(tblloc);
	if (loc == NULL)
		perrexit();

	for(;;) {

/*		read data from /dev/kmem to structure d	*/

		loc = origloc;
		lseek(f,(long)setup[X_SYSINFO].n_value,0);
		if (read(f, &d.si, sizeof d.si) == -1)
			perrexit();
		d.si.bswapin = ctod(d.si.bswapin);
		d.si.bswapout = ctod(d.si.bswapout);

		lseek(f, (long)setup[X_VDSTAT].n_value, 0);
		if (read(f, idisk, sizeof(struct iotime) * ndisk) == -1)
			perrexit();
		lseek(f, (long)setup[X_CYSTAT].n_value, 0);
		if (read(f, icytape, sizeof(struct iotime) * ncytape) == -1)
			perrexit();
		lseek(f, (long)setup[X_XYSTAT].n_value, 0);
		if (read(f, ixytape, sizeof(struct iotime) * nxytape) == -1)
			perrexit();

/*		compute size of inode table 	*/
		if (gettbl(setup[X_INODE].n_value,
		    inodes * sizeof(struct inode)) == 0)
			perrexit();
		for (i = n = 0, ip = (struct inode *)loc; i < inodes; i++, ip++)
			if (ip->i_vnode.v_count)
				n++;
		loc = (char *)ip;
		d.szinode = n;

/*		compute size of file table 	*/
		if (gettbl(setup[X_FILE].n_value,
		    files * sizeof(struct file)) == 0)
			perrexit();
		for (i = n = 0, fp = (struct file *)loc; i < files; i++, fp++)
			if (fp->f_count)
				n++;
		loc = (char *)fp;
		d.szfile = n;

/*		compute size of text table 	*/
		if (gettbl(setup[X_TEXT].n_value,
		    texts * sizeof(struct text)) == 0)
			perrexit();
		for (i = n = 0, tp = (struct text *)loc; i < texts; i++, tp++)
			if (tp->x_vptr)
				n++;
		loc = (char *)tp;
		d.sztext = n;

/*		compute size of proc table 	*/
		if (gettbl(setup[X_PROC].n_value,
		    procs * sizeof(struct proc)) == 0)
			perrexit();
		for (i = n = 0, pp = (struct proc *)loc; i < procs; i++, pp++)
			if (pp->p_stat)
				n++;
		loc = (char *)pp;
		d.szproc = n;

/*		record maximum sizes of system tables	*/
		d.mszinode = inodes;
		d.mszfile = files;
		d.msztext = texts;
		d.mszproc = procs;

/*		record system tables overflows	*/
		lseek(f,(long)setup[X_ERR].n_value,0);
		if (read(f,&err,sizeof (struct syserr)) == -1)
			perrexit();
		d.inodeovf = err.inodeovf;
		d.fileovf = err.fileovf;
		d.textovf = err.textovf;
		d.procovf = err.procovf;
		/*	get time stamp	*/
		d.ts = time ((long *) 0);

		for (i = 0; i < ndisk; i++) {
			d.devio[i][IO_OPS] = idisk[i].io_cnt;
			d.devio[i][IO_BCNT] = idisk[i].io_bcnt;
			d.devio[i][IO_ACT] = idisk[i].io_act;
			d.devio[i][IO_RESP] = idisk[i].io_resp;
		}

		for (i = ndisk; i < ndisk + ntape; i++) {
			d.devio[i][IO_OPS] = icytape[i - ndisk].io_cnt +
					ixytape[i-ndisk].io_cnt;
			d.devio[i][IO_BCNT] = icytape[i - ndisk].io_bcnt +
					ixytape[i-ndisk].io_bcnt;
			d.devio[i][IO_ACT] = icytape[i - ndisk].io_act +
					ixytape[i-ndisk].io_act;
			d.devio[i][IO_RESP] = icytape[i - ndisk].io_resp +
					ixytape[i-ndisk].io_resp;
		}

/*	write data to data file from structure d	*/
		write(fd,&d,recsz);
		if(--ct > 0)
			sleep(ti);
		else {
			close(fd);
			exit(0);
		}
	}
}

gettbl(addr, size)
unsigned long addr, size;
{

	lseek(f, (long)addr, 0);
	if (read(f, (char *)&atbl, sizeof(int)) == -1)
		return(0);
	lseek(f, (long)atbl, 0);
	if (read(f, loc, size) == -1)
		return(0);
	return(1);
}


perrexit()
{
	perror("sadc");
	exit(2);
}
