#include <stdio.h>
#include <nlist.h>
#include <sys/utsname.h>
#include <sys/param.h>
#include <sys/buf.h>
#include <sys/vbavar.h>

#define DSKGENDATA
#include <sys/dsk.h>
#undef DSKGENDATA

#define VDGENDATA
#include <sys/vddc.h>
#undef VDGENDATA


#define	UNIX		"/unix"
#define	DISKNM(i)	(vdst[i].type_name)
#define	NDSK		0
#define	VDDINFO		1
#define	DKPRF		2
#define	SKPRF		3


/*
 * These constants are taken from 'sadp.c'
 * on System V.
 */
#define BLANK		' '
#define BLOB		'*'
#define TRACE		'.'
#define BRK		'='
#define FOOT		'-'
#define CYLNO		1
#define SEEKD		2


int	fd,			/* /dev/kmem */
	dev,			/* 0, 1 or 2 for fsd, smd or xfd */
	ndsk,			/* kernel variable: number of disks */
	nonum,			/* set to 0 if no number for drives is given */
	tflag,			/* -t flag */
	hflag,			/* -h flag */
	dflag,			/* -d flag */
	drvlist[MAXDRV],	/* drvlist[i] = 1 iff i is profiled */
	alvlist[MAXDRV],	/* list of 'alive' disks */
	dkcyl[MAXDRV][CHPERCYL],/* picture 1 of dkprf */
	skcyl[MAXDRV][CHPERCYL];/* picture 1 of skprf */


struct utsname		name;
struct vba_device	**vddinfo,
			vba_device;
struct nlist		nl[] = {
				{"_ndsk"},
				{"_vddinfo"},
				{"_dkprf"},
				{"_skprf"},
				{0}
			};

char	*ctime();


/*
 * main -- sadp.c.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	int	s,
		n;

	nonum = 1;		/* assume no numbers for drives is given */
	dev = -1;		/* non-existent device type */
	s = n = 0;		/* invalid values */
	while (--argc) {
		if (**++argv == '-') {
			switch ((*argv)[1]) {
			case 't':
				tflag++;
				break;
			case 'h':
				hflag++;
				break;
			case 'd':
				dflag++;
				if (--argc == 0) {
					usage();
					exit (1);
				}
				if (setdrvlist (*++argv)) {
					printf ("%s: bad device\n", *argv);
					exit (1);
				}
				break;
			default:
				usage();
				exit (1);
			}
		} else if (s == 0) {
			if ((s = atoi (*argv)) <= 0) {
				printf ("%s: bad s value\n", s);
				exit (1);
			}
		} else if (n == 0) {
			if ((n = atoi (*argv)) <= 0) {
				printf ("%n: bad n value\n", n);
				exit (1);
			}
		} else {
			usage();
			exit (1);
		}
	}

	if (s == 0) {
		usage();
		exit (1);
	}

	if (n == 0)
		n++;

	if (hflag == 0)
		tflag++;


	init();
	sadp (s, n);
	exit (0);
}


/*
 * init -- make sure that 'drvlist' is ok and initialize
 *	   'dev' and 'drvlist' if they are not already
 *	   inittialized (i.e., no -d flag).
 */
init()
{
	int	i,
		j,
		size;
	struct vba_device	*p;

	if (nlist (UNIX, nl) < 0) {
		perror ("nlist");
		exit (1);
	}

	j = 0;
	for (i = 0; nl[i].n_name; i++)
		if (nl[i].n_value == NULL) {
			printf ("%s: not defined in %s\n", nl[i].n_name, UNIX);
			j++;
		}
	if (j)
		exit (1);

	if ((fd = open ("/dev/kmem", 0)) < 0) {
		perror ("open");
		exit (1);
	}

	/*
	 * 'ndsk' is the number of disks this system
	 * is configured to support.
	 */
	Lseek (fd, (long) nl[NDSK].n_value, 0);
	Read (fd, (char *) &ndsk, sizeof ndsk);

	/*
	 * Allocate space to read in kernel's 'vddinfo'
	 * data structure.
	 */
	size = sizeof (struct vba_device **) * ndsk;
	if ((vddinfo = (struct vba_device **) malloc (size)) == NULL) {
		perror ("malloc");
		exit (1);
	}
	Lseek (fd, (long) nl[VDDINFO].n_value, 0);
	Read (fd, (char *) vddinfo, size);

	/*
	 * Now, check the validity of 'dev' and 'drvlist'.
	 * This is done by initializing 'alvlist[i]' to 1
	 * for each alive disk and then comparing it against
	 * 'drvlist'.
	 */
	p = &vba_device;
	for (i = 0; i < ndsk; i++) {
		if (vddinfo[i]) {
			/*
			 * Read in the 'vba_device' data structure.
			 */
			Lseek (fd, (long) vddinfo[i], 0);
			Read (fd, (char *) p, sizeof (*p));

			/*
			 * If there is only one device type (i.e., only
			 * fsd), the -d flag may be omitted on the command
			 * line. In this case, neither 'dev' nor 'drvlist'
			 * is initialized. If this is the case, go ahead
			 * and initialize them according to the first
			 * drive type.
			 */
			if (dflag == 0) {
				if (dev == -1)
					dev = p->ui_type;

				if (dev != p->ui_type) {
					printf ("Device type not given\n");
					exit (1);
				}
			}

			if (p->ui_alive && p->ui_type == dev)
				alvlist[p->ui_unit] = 1;
		}
	}

	/*
	 * 'alvlist' now has the list of all alive
	 * disks for the device type 'dev'.
	 */
	j = 0;
	for (i = 0; i < MAXDRV; i++) {
		if (nonum)
			drvlist[i] = alvlist[i];
		else if (drvlist[i] && alvlist[i] == 0) {
			printf ("%s%d: not alive\n", DISKNM(dev), i);
			j++;
		}
	}
	if (j)
		exit (1);
}


/*
 * sadp --
 */
sadp (s, n)
int	s,
	n;
{
	int	i,
		j;
	char	*stime;
	long	curt,
		dkio[MAXDRV],
		skio[MAXDRV];

	while (n-- > 0) {
	    /*
	     * Take a picture. After 's' seconds,
	     * take another picture and compute the 
	     * elevant statistics.
	     */
	    Lseek (fd, (long) nl[DKPRF].n_value, 0);
            Read (fd, (char *) dkcyl, sizeof dkcyl);
	    Lseek (fd, (long) nl[SKPRF].n_value, 0);
            Read (fd, (char *) skcyl, sizeof skcyl);

	    sleep (s);

	    Lseek (fd, (long) nl[DKPRF].n_value, 0);
            Read (fd, (char *) dkprf, sizeof dkprf);
	    Lseek (fd, (long) nl[SKPRF].n_value, 0);
            Read (fd, (char *) skprf, sizeof skprf);

            for (i = 0; i < MAXDRV; i++) {
                if (drvlist[i] == 0)
                    continue;

		dkio[i] = skio[i] = 0;
                for (j = 0; j < CHPERCYL; j++) {
	            dkcyl[i][j] = dkprf[i][j] - dkcyl[i][j];
		    dkio[i] += dkcyl[i][j];
	            skcyl[i][j] = skprf[i][j] - skcyl[i][j];
		    skio[i] += skcyl[i][j];
	        }
	    }

	    /*
	     * Current time and system name.
	     */
	    uname (&name);
	    curt = time ((long *) 0);
	    stime = ctime (&curt);
	    printf ("\n\n%s\n", stime);
	    printf ("%s %s %s %s %s\n", name.sysname, name.nodename,
				name.release, name.version, name.machine);

	    /*
	     * Print out the computed statistics.
	     */
	    if (tflag) {
		printf ("\nCYLINDER ACCESS PROFILE\n");
		for (i = 0; i < MAXDRV; i++) {
		    if (drvlist[i] == 0)
			continue;

		    printf ("\n%s-%d:\n", DISKNM(dev), i);
		    printf ("Cylinders\tTransfers\n");
		    for (j = 0; j < CHPERCYL; j++) {
			if (dkcyl[i][j])
			    printf ("%3d - %3d\t%ld\n", j*8, j*8+7,dkcyl[i][j]);
		    }

		    printf ("\nSampled I/O = %ld", dkio[i]);
		    printf (", Actual I/O = %ld\n", dkio[i]);
		    printf ("Percentage of I/O sampled = %2.2f\n", 100.0);
		}

		printf ("\n\n\nSEEK DISTANCE PROFILE\n");
		for (i = 0; i < MAXDRV; i++) {
            	    if (drvlist[i] == 0)
                	continue;

                    printf ("\n%s-%d:\n", DISKNM(dev), i);
                    printf ("Seek Distance\tSeeks\n");
                    for (j = 0; j < CHPERCYL; j++)
                        if (skcyl[i][j]){
                            if (j == 0)
                                printf ("        0\t%ld\n", skcyl[i][j]);
                            else
                               printf("%3d - %3d\t%ld\n",j*8-7,j*8,skcyl[i][j]);
                	}
                    printf ("Total Seeks = %ld\n", skio[i]);
	        }
	    } else if (hflag) {
		for (i = 0; i < MAXDRV; i++) {
		    if (drvlist[i] == 0)
			continue;

	 	    if (dkio[i]) {
			cylhdr (CYLNO, dkio[i], i);
			cylhist (dkio[i], dkcyl[i]);
			cylftr (CYLNO);
		    }
		}

		for (i = 0; i < MAXDRV; i++) {
		    if (drvlist[i] == 0)
			continue;

		    if (skio[i]){
			cylhdr (SEEKD, skio[i], i);
			cylhist (skio[i], skcyl[i]);
			cylftr (SEEKD);
		    }
		}
	    }
	}
}


/*
 * setdrvlist -- set drvlist[i] to 1 if it is to be profiled.
 *		 If the user has not specified a drive range,
 *		 we will find out what are the valid drive
 *		 numbers for him. In this case, we set the
 *		 variable 'nonum' to 1. Otherwise, if the user
 *		 has provided drive numbers (e.g., fsd-0-3 or
 *		 fsd-1,4, ...), we reset 'nonum'.
 */
setdrvlist (s)
char	*s;
{
	int	i;

	for (i = 0; i < nvddrv; i++) {
		if (equal (DISKNM(i), s)) {
			dev = i;
			break;
		}
	}

	if (i == nvddrv)
		return (-1);

	s += 3;
	if (*s == '\0')
		return (0);

	nonum = 0;
	if (*s == '-') {
		int	n1, n2;

		if (*++s < '0' || *s > '9' || (n1 = *s - '0') >= MAXDRV)
			return (-1);
		drvlist[n1] = 1;
		if (*++s == '\0')
			return (0);
		if (*s == '-') {
			if (*++s<'0' || *s>'9' || (n2= *s-'0')>=MAXDRV || n1>=n2)
				return (-1);
			for (i = n1; i <= n2; i++)
				drvlist[i] = 1;
			return (0);
		}
		while (1) {
			if (*s != ',')
				return (-1);
			if (*++s < '0' || *s > '9' || (n1 = *s - '0') >= MAXDRV)
				return (-1);
			drvlist[n1] = 1;
			if (*++s == '\0')
				return (0);
		}
	}

	return (-1);
}


/*
 * equal -- return 1 if string 't' is equal or longer than 's'.
 */
equal (s, t)
char	*s,
	*t;
{
	while (*s && *s == *t) {
		s++;
		t++;
	}

	return (*s == '\0');
}


/*
 * Lseek -- call lseek and exit upon error.
 */
Lseek (fildes, offset, whence)
int	fildes;
long	offset;
int	whence;
{
	if (lseek (fildes, offset, whence) == -1) {
		perror ("lseek");
		exit (1);
	}
}


/*
 * Read -- call read and exit upon error.
 */
Read (fildes, buf, n)
int	fildes;
char	*buf;
int	n;
{
	if (read (fildes, buf, n) != n) {
		perror ("read");
		exit (1);
	}
}


/*
 * usage -- print the usage message.
 */
usage()
{
	printf ("Usage: sadp [-th] [-d device[-drive]] s [n]\n");
}


/*
 * The code from here to the end of the file is
 * taken directly from 'sadp.c' on System V.
 * The only modification is making it readable!
 */

struct HISTDATA {
	long	hdata[CHPERCYL];
};


/*
 * cylhist -- Cylinder profiling histogram
 */
cylhist (at, dp)
long	at;
register struct HISTDATA *dp;
{
	register	ii;
	int	maxrow;
	long	graph[CHPERCYL];
	long    max, max2;
	long    data;
	long    scale;

	max = 0;
	for (ii = 0; ii < CHPERCYL; ii++) {
		if (data = dp->hdata[ii]) {
			maxrow = ii;
			if (data > max) {
				max2 = max;
				max = data;
			} else if (data > max2 && data != max)
				max2 = data;
		}
	}
	maxrow++;

	/* determine scaling */
	scale = 1;
	if (max2) {
		scale = at / (max2 * 2);
		if (scale > 48)
			scale = 48;
	}

	for (ii = 0; ii < maxrow; ii++) {
		if (dp->hdata[ii])
			graph[ii] = (scale * 100 * dp->hdata[ii]) / at;
		else
			graph[ii] = -1;
	}

	prthist(graph, maxrow, scale, (long) (max*100*scale/at));
}


/*
 * prthist -- print histogram.
 */
prthist(array, mrow, scale, gmax)
long array[], scale, gmax;
register mrow;
{
	long    line;

	line = 50;
	/* handle overflow in scaling */
	if (gmax > 51) {
		line = 52;
		printf ("\n%2ld%% -|", gmax/scale);
		pline(line--, array, mrow, BLOB);
		printf ("\n     %c", BRK);
		pline(line--, array, mrow, BRK);
	} else if ( gmax = 51 )
		line = 51;
	while (line > 0) {
		if ((line & 07) == 0)
			printf ("\n%2ld%% -|", line/scale);
		else
			printf ("\n     |");
		pline(line--, array, mrow, BLOB);
	}
	printf ("\n 0%% -+");
	line = -1;
	pline( line, array, mrow, FOOT);
}


/*
 * pline -- print histogram line.
 */
pline(line, array, mrow, dot)
long line, array[];
int mrow;
char dot;
{
	register ii;
	register char *lp;
	char lbuff[132];

	lp = lbuff;
	for (ii = 0; ii < mrow; ii++)
		if (array[ii] < line)
			if (line == 1 && array[ii] == 0)
				*lp++ = TRACE;
			else
				*lp++ = BLANK;
		else
			*lp++ = dot;
	*lp++ = 0;
	printf ("%s", lbuff);
}


/*
 * cylhdr -- print histogram header.
 */
cylhdr (flag, total, unit)
int	flag;
long	total;
int	unit;
{
	if (flag == CYLNO)
		printf ("\nCYLINDER ACCESS HISTOGRAM\n");
	if (flag == SEEKD)
		printf ("\nSEEK DISTANCE HISTOGRAM\n");
	printf ("\n%s-%d:\n", DISKNM(dev), unit);
	printf ("Total %s = %ld\n", flag==CYLNO ? "transfers" : "seeks", total);
}


/*
 * cylftr -- print histogram footer.
 */
cylftr (flag)
{
	if (flag == CYLNO)
		printf ("\n      \t\t\tCylinder number, granularity=8");
	else
		printf ("\n      =<<\t<\t<\t<\t<\t<\t<\t<\t<");
	printf ("\n      081\t8\t1\t2\t2\t3\t4\t4\t5");
	printf ("\n        6\t0\t4\t0\t7\t3\t0\t6\t2");
	printf ("\n         \t \t4\t8\t2\t6\t0\t4\t8");
	printf ("\n");
}
