#ifndef lint
static	char *sccsid = "@(#)savecore.c 1.1 85/05/31 SMI"; /* from UCB 4.13 */
#endif

/*
 * savecore
 */
#include <stdio.h>
#include <nlist.h>
#include <sys/param.h>
#include <sys/dir.h>
#include <ml/pte.h>
#include <sys/stat.h>
#include <sys/vfs.h>
#include <sys/time.h>
#include <sys/utsname.h>
#include <sys/cmap.h>
#include <sys/vm.h>

#define	DAY	(60L*60L*24L)
#define	LEEWAY	(3*DAY)
#define V_OFFSET (uts.version - (char *)&uts)

#define eq(a,b) (!strcmp(a,b))
#define ok(number) ((number)&0x3fffffff)

#define SHUTDOWNLOG "/usr/adm/shutdownlog"

/* NCMAPBF must be a multiple of BLKSIZ */
#define NBUFPG 32       		/* # buffer pages */
#define NCMAPBF 1024			/* number of cmap entries in the buffer */
#define BLKSIZ  8       		/* filesystem block size */
#define BLKMSK  0xfffffff8              /* mask to 1st page of block */

struct nlist nl[] = {
#define X_DUMPDEV	0
	{ "_dumpdev" },
#define X_DUMPLO	1
	{ "_dumplo" },
#define X_TIME		2
	{ "_time" },
#define	X_DUMPSIZE	3
	{ "_dumpsize" },
#define X_UTSNAME	4
	{ "_utsname" },
#define X_PANICSTR	5
	{ "_panicstr" },
#define	X_DUMPMAG	6
	{ "_dumpmag" },
#define	X_SYSMAP        7
	{ "_Sysmap" },
#define	X_SYSSIZE       8
	{ "_Syssize" },
#define	X_FIRSTFREE	9
	{ "_firstfree" },
#define	X_CMAP    	10
	{ "_cmap" },
	{ "" },
};

char	*system;
char	*dirname;			/* directory to save dumps in */
char	*ddname;			/* name of dump device */
char	*find_dev();
dev_t	dumpdev;			/* dump device */
time_t	dumptime;			/* time the dump was taken */
int	dumplo;				/* where dump starts on dumpdev */
int	dumpsize;			/* amount of memory dumped */
int	dumpmag;			/* magic number in dump */
time_t	now;				/* current date */
char	*path();
unsigned malloc();
char	*ctime();
char	vers[9];
char	core_vers[9];
char	panic_mesg[80];
int	panicstr;
int	aflg;
off_t	lseek();
off_t	Lseek();
struct utsname uts;	/* only for calculating the offset! */

main(argc, argv)
 	char **argv;
	int argc;
{

	argc--, argv++;
	if (argc > 0 && argv[0][0] == '-') {
		register char *cp = *argv++;
		argc--;
		switch (*++cp) {
			case 'a':
				aflg++;
				break;
			default:
				fprintf(stderr,
				"usage: savecore [-a] dirname [ system ]\n");
				exit(1);
		}
	} /* end of if options */

	if (argc != 1 && argc != 2) {
		fprintf(stderr,
			"usage: savecore [-a] dirname [ system ]\n");
		exit(1);
	}
	dirname = argv[0];
	if (argc == 2)
		system = argv[1];
	if (access(dirname, 2) < 0) {
		perror(dirname);
		exit(1);
	}
	read_kmem();
	if (dump_exists()) {
		(void) time(&now);
		check_kmem();
		log_entry();
		if (get_crashtime() && check_space()) {
			save_core();
			clear_dump();
		} else
			exit(1);
	}
	return 0;
}

int
dump_exists()
{
	register int dumpfd;
	int word;

	dumpfd = Open(ddname, 0);
	Lseek(dumpfd, (off_t)(dumplo + ok(nl[X_DUMPMAG].n_value)), 0);
	Read(dumpfd, (char *)&word, sizeof word);
	close(dumpfd);
	return (word == dumpmag);
}

clear_dump()
{
	register int dumpfd;
	int zero = 0;

	dumpfd = Open(ddname, 1);
	Lseek(dumpfd, (off_t)(dumplo + ok(nl[X_DUMPMAG].n_value)), 0);
	Write(dumpfd, (char *)&zero, sizeof zero);
	close(dumpfd);
}

char *
find_dev(dev, type)
	register dev_t dev;
	register int type;
{

	char *dp;
	extern char * find_dev2() ;

	if ( (dp = find_dev2( "/dev/", dev, type ) ) != 0 )
	        return ( dp ) ;

	if ( (dp = find_dev2( "/dev/dsk/", dev, type ) ) != 0 )
	        return ( dp ) ;

	if ( (dp = find_dev2( "/dev/mt/", dev, type ) ) != 0 )
	        return ( dp ) ;

	fprintf(stderr, "savecore: Can't find device %d,%d\n",
		major(dev), minor(dev));
	exit(1);
	/*NOTREACHED*/
}

char *
find_dev2(dirname, dev, type)
	register char *dirname ;
	register dev_t dev;
	register int type;
{
	register DIR *dfd;
	struct direct *dir;
	struct stat statb;
	static char devname[MAXPATHLEN + 1];
	char *dp;

	if ( (dfd = opendir( dirname )) == 0 ) {
		return 0 ;
	}

	while ((dir = readdir(dfd))) {
		strcpy(devname, dirname);
		strcat(devname, dir->d_name);
		if (stat(devname, &statb)) {
			perror(devname);
			continue;
		}
		if ((statb.st_mode&S_IFMT) != type)
			continue;
		if (dev == statb.st_rdev) {
			closedir(dfd);
			dp = (char *)malloc(strlen(devname)+1);
			strcpy(dp, devname);
			return dp;
		}
	}
	closedir(dfd);
	return 0 ;
}

read_kmem()
{
	int kmem;
	FILE *fp;
	register char *cp;

	nlist("/unix", nl);
	if (nl[X_DUMPDEV].n_value == 0) {
		fprintf(stderr, "savecore: /unix: dumpdev not in namelist\n");
		exit(1);
	}
	if (nl[X_DUMPLO].n_value == 0) {
		fprintf(stderr, "savecore: /unix: dumplo not in namelist\n");
		exit(1);
	}
	if (nl[X_TIME].n_value == 0) {
		fprintf(stderr, "savecore: /unix: time not in namelist\n");
		exit(1);
	}
	if (nl[X_DUMPSIZE].n_value == 0) {
		fprintf(stderr, "savecore: /unix: dumpsize not in namelist\n");
		exit(1);
	}
	if (nl[X_UTSNAME].n_value == 0) {
		fprintf(stderr, "savecore: /unix: version not in namelist\n");
		exit(1);
	}
	if (nl[X_PANICSTR].n_value == 0) {
		fprintf(stderr, "savecore: /unix: panicstr not in namelist\n");
		exit(1);
	}
	if (nl[X_DUMPMAG].n_value == 0) {
		fprintf(stderr, "savecore: /unix: dumpmag not in namelist\n");
		exit(1);
	}
	if (nl[X_SYSMAP].n_value == 0) {
		fprintf(stderr, "savecore: /unix: sysmap not in namelist\n");
		exit(1);
	}
	if (nl[X_SYSSIZE].n_value == 0) {
		fprintf(stderr, "savecore: /unix: syssize not in namelist\n");
		exit(1);
	}
	if (nl[X_FIRSTFREE].n_value == 0) {
		fprintf(stderr, "savecore: /unix: firstfree not in namelist\n");
		exit(1);
	}
	if (nl[X_CMAP].n_value == 0) {
		fprintf(stderr, "savecore: /unix: cmap not in namelist\n");
		exit(1);
	}
	kmem = Open("/dev/kmem", 0);
	Lseek(kmem, (long)nl[X_DUMPDEV].n_value, 0);
	Read(kmem, (char *)&dumpdev, sizeof (dumpdev));
	Lseek(kmem, (long)nl[X_DUMPLO].n_value, 0);
	Read(kmem, (char *)&dumplo, sizeof (dumplo));
	Lseek(kmem, (long)nl[X_DUMPMAG].n_value, 0);
	Read(kmem, (char *)&dumpmag, sizeof (dumpmag));
	dumplo *= 1024L;
	ddname = find_dev(dumpdev, S_IFBLK);
	if ((fp = fdopen(kmem, "r")) == NULL) {
		fprintf(stderr, "savecore: Couldn't fdopen kmem\n");
		exit(1);
	}
	if (system) {
		nlist(system, nl);
		return;
	}
 	fseek(fp, (long)(nl[X_UTSNAME].n_value + V_OFFSET), 0);
	fgets(vers, sizeof vers, fp);
	fclose(fp);
}

check_kmem()
{
	FILE *fp;
	register char *cp;

	if ((fp = fopen(ddname, "r")) == NULL) {
		perror(ddname);
		exit(1);
	}
	if (!system) {
		fseek(fp, (off_t)(dumplo+ok(nl[X_UTSNAME].n_value + V_OFFSET)), 0);
		fgets(core_vers, sizeof core_vers, fp);
		fclose(fp);
		if (!eq(vers, core_vers))
			fprintf(stderr,
			   "savecore: Warning: unix version mismatch:\n\t%sand\n\t%s",
			   vers, core_vers);
	}
	fp = fopen(ddname, "r");
	fseek(fp, (off_t)(dumplo + ok(nl[X_PANICSTR].n_value)), 0);
	fread((char *)&panicstr, sizeof panicstr, 1, fp);
	if (panicstr) {
		fseek(fp, dumplo + ok(panicstr), 0);
		cp = panic_mesg;
		do
			*cp = getc(fp);
		while (*cp++);
	}
	fclose(fp);
}

get_crashtime()
{
	int dumpfd;
	time_t clobber = (time_t)0;

	dumpfd = Open(ddname, 0);
	Lseek(dumpfd, (off_t)(dumplo + ok(nl[X_TIME].n_value)), 0);
	Read(dumpfd, (char *)&dumptime, sizeof dumptime);
	close(dumpfd);
	if (dumptime == 0)
		return (0);
	printf("System went down at %s", ctime(&dumptime));
	if (dumptime < now - LEEWAY || dumptime > now + LEEWAY) {
		printf("Dump time is unreasonable\n");
		return (0);
	}
	return (1);
}

char *
path(file)
	char *file;
{
	register char *cp = (char *)malloc(strlen(file) + strlen(dirname) + 2);

	(void) strcpy(cp, dirname);
	(void) strcat(cp, "/");
	(void) strcat(cp, file);
	return (cp);
}

check_space()
{
	struct statfs fsb;
	int dfd, spacefree;

	if (statfs(dirname, &fsb) < 0) {
		perror(dirname);
		exit(1);
	}
	spacefree = fsb.f_bfree * fsb.f_bsize / 1024;
	if (read_number("minfree") > spacefree) {
		fprintf(stderr,
		   "savecore: Dump omitted, not enough space on device\n");
		return (0);
	}
	if (fsb.f_bavail < 0)
		fprintf(stderr,
			"Dump performed, but free space threshold crossed\n");
	return (1);
}

read_number(fn)
	char *fn;
{
	char lin[80];
	register FILE *fp;

	if ((fp = fopen(path(fn), "r")) == NULL)
		return (0);
	if (fgets(lin, 80, fp) == NULL) {
		fclose(fp);
		return (0);
	}
	fclose(fp);
	return (atoi(lin));
}


save_core()
{
	char buffer[NBUFPG*NBPG];
	char *cp = buffer;

	char cmapbuf[NCMAPBF*sizeof(struct cmap)]; /* cmap entries buffer */
	char *cmp = cmapbuf; 	/* for reading entries in */
	struct cmap *cmap_p; 	/* points to entries in cmapbuf */

	int 	ifd, ofd, bounds, n, dmpprev = 0,
        	firstfree, kernelsize, numpgs,
		curpg, cmapinb, blk_strt, ncmap;
	unsigned cmaploc;
	FILE *fp;


	/* Process the unix.x file */

	ifd = Open(system?system:"/unix", 0);

	bounds = read_number("bounds");
	sprintf(cp, "unix.%d", bounds);
	ofd = Create(path(cp), 0644);

	while((n = Read(ifd, cp, BUFSIZ)) > 0)
		Write(ofd, cp, n);

	close(ifd);
	close(ofd);


	/* Process the vmcore.x file */

	ifd = Open(ddname, 0);
	sprintf(cp, "vmcore.%d", bounds);
	ofd = Create(path(cp), 0644);

	/* Get the number of pages in the dump core image */
	Lseek(ifd, (off_t)(dumplo + ok(nl[X_DUMPSIZE].n_value)), 0);
	Read(ifd, (char *)&dumpsize, sizeof (dumpsize));

	/* if -a option, set up to copy all of memory
	 * image to vmcore.x
	 */
	if (aflg) {
		kernelsize = dumpsize;
		printf("Saving %d bytes of image in vmcore.%d\n",
			NBPG*dumpsize, bounds);
	}

	/* else, set up to copy just kernel space */
	else {
		/* Get 1st page # past the kernel space. */
		Lseek(ifd, (off_t)(dumplo + ok(nl[X_FIRSTFREE].n_value)), 0);
		Read(ifd, (char *)&firstfree, sizeof (firstfree));

		/* Round up kernel space to a disk block boundary */
		kernelsize = curpg =
		        ( (firstfree + BLKSIZ - 1) / BLKSIZ ) * BLKSIZ;

		printf("Saving %d bytes of kernel space image in vmcore.%d\n",
			NBPG*firstfree, bounds);
	}

	/* Point to front of dump image */
	Lseek(ifd, (off_t)dumplo, 0);

	/* Always dump kernel space out to vmcore.x file.
	 * If -a option, then dump all of memory image to vmcore.x.
	 */
	while (kernelsize > 0) {
		n = Read(ifd, cp,
				(kernelsize > NBUFPG ? NBUFPG : kernelsize) * NBPG);
		if (n <= 0) {
			printf("Dump area was too small ");
			if (aflg)
				printf("- %d pages not saved\n", kernelsize);
			else
				printf("- %d kernel space pages not saved\n",
					kernelsize);
			goto dumpall;
		}
		/* output to vmcore.x */
 		Write(ofd, cp, n);

		/* pages left to copy */
		kernelsize -= n/NBPG;
	}

	/* if -a option, then all done dumping to vmcore.x */
	if (aflg)
	        goto dumpall;


	/* Dump the rest of the core image -- only dump those pages
	 * that are CSYS or CSTACK and not c_gone or c_free
	 * in the CMAP table.
	 * Put holes in the file for those pages not copied into
	 * vmcore.x.
	 */
	printf("Saving additional system and stack pages to vmcore.%d\n",
	          bounds);

	/* start page # of a block of pages to dump */
	blk_strt = 0;

	/* Get the location of the cmap table. */
	Lseek(ifd, (off_t)(dumplo + ok(nl[X_CMAP].n_value)), 0);
	Read(ifd, (char *)&cmaploc, sizeof(cmaploc));

	/* Skip first cmap entry -- dummy entry used to point
	 * to cmhash table.  Also skip pages already dumped
	 * along with the kernel space when the kernel space
	 * was rounded up to a BLKSIZ boundary.
	 */
	cmaploc += sizeof(struct cmap) * (1 + curpg - firstfree);

	/* total number of cmap entries to scan */
	ncmap = dumpsize - curpg;

	/* scan entire cmap table for CSYS pages */
	while (ncmap > 0) {

		/* Read in cmap entries into the cmap buffer */
		klseek(ifd, cmaploc, 0);
		cmaploc += NCMAPBF * sizeof(struct cmap);

		/* set up remaining cmap entries to read and
		 * number of cmap entries in cmap buffer.
		 */
		if (ncmap < NCMAPBF) {
		     cmapinb = ncmap;
		     ncmap = 0;
		}
		else {
		     ncmap -= NCMAPBF;
		     cmapinb = NCMAPBF;
		}

		/* Read in cmap entries into buffer */
		Read(ifd, cmp, (cmapinb*sizeof(struct cmap)) );
		cmap_p = (struct cmap *)cmapbuf;

		/* Scan all cmap entries in cmap buffer */
		while (cmapinb > 0) {

			/* Check for proper page to dump */
	        	if ( ((cmap_p->c_type == CSYS) || (cmap_p->c_type == CSTACK))
				&& (cmap_p->c_free == 0) && (cmap_p->c_gone == 0) ) {

				/* if previous block was not dumped */
	        		if (blk_strt == 0) {	
					/* set to start of current block */
					blk_strt = curpg & BLKMSK;
	        		    	numpgs = BLKSIZ;
				}

				/* Previous block(s) are also to be dumped, so
				 * increment # blocks to dump
				 */
				else
					numpgs += BLKSIZ;

				/* Skip to next block of cmap entries. */
				cmap_p += BLKSIZ - (curpg - (curpg & BLKMSK));
				cmapinb -= BLKSIZ - (curpg - (curpg & BLKMSK));
				curpg = (curpg & BLKMSK) + BLKSIZ;

				/* previous block should be dumped */
				dmpprev = 1;

			} /* end if block of pages to dump */

			/* Move to next page to see if it should be dumped. */
	        	else {
				cmap_p++; 	/* point to next cmap entry */
				curpg++;  	/* increment page # */
				cmapinb--; 	/* # cmap entries left in buffer */

				/* If a start of next block, then we got through
    				 * an entire block without a page to dump and we mark
    				 * that previous block is not to be dumped.
				 */
				if (curpg % BLKSIZ == 0)
					dmpprev = 0;
			}

			/* If:
			 *  1) we are at the end of a block of pages
			 *  2) and there are previous block(s) to dump,
			 *  3) and the block of pages we just finished
			 *     is not to be dumped,
			 * then:
			 * dump the page block(s) to the dump file.
			 */
			if ( ((curpg % BLKSIZ) == 0) && (blk_strt != 0)
				&& (dmpprev == 0) ) {

				/* Set dump device file and dump output
				 * file to proper positions.
				 */
				Lseek(ifd, (off_t)(dumplo + (blk_strt*NBPG)), 0);	
				Lseek(ofd, (off_t)(blk_strt*NBPG), 0);

				/* dump page(s) */
				while (numpgs > 0) {

					n = Read(ifd, cp,
						(numpgs > NBUFPG ? NBUFPG : numpgs) * NBPG);

					if (n <= 0) {
						printf("Dump area was too small");
						printf(" - at least %d pages not saved\n",
						     numpgs );
						goto dumpall;
					}

					/* output to vmcore.x */
			 		Write(ofd, cp, n);

					/* decrement pages left to copy */
					numpgs -= n/NBPG;

				} /* end while */

				/* indicate no pages yet found to dump */
				blk_strt = 0;

	        	} /* end if pages to dump */

		} /* end while cmapinb */

	} /* end while ncmap */


	/* Write out the last page block(s).  Always write the last
	 * page block out to the dump file so that the size of the file
	 * is the same as dumpsize.
	 * If blk_strt != 0 , then last page block(s) include last page
	 * block of the dump file.
	 */
	if (blk_strt == 0) {
		blk_strt = dumpsize - BLKSIZ;
		numpgs = BLKSIZ;
	}

	/* Seek to proper location in both files */
 	Lseek(ifd, (off_t)(dumplo + (blk_strt*NBPG)), 0);
	Lseek(ofd, (off_t)(blk_strt*NBPG), 0);

	while (numpgs > 0) {
		n = Read(ifd, cp,
			(numpgs > NBUFPG ? NBUFPG : numpgs) * NBPG);
		if (n <= 0) {
			printf("Dump area too small - %d system pages not saved\n",
				numpgs);
			goto dumpall;
		}

		/* Output to vmcore.x */
		Write(ofd, cp, n);

		/* decrement pages left to copy */
		numpgs -= n/NBPG;

	} /* end while numpgs */


dumpall:
	close(ifd);
	close(ofd);

	/* update the bounds file */
	fp = fopen(path("bounds"), "w");
	fprintf(fp, "%d\n", bounds+1);
	fclose(fp);

}


char *days[] = {
	"Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat"
};

char *months[] = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep",
	"Oct", "Nov", "Dec"
};

log_entry()
{
	FILE *fp;
	struct tm *tm, *localtime();

	tm = localtime(&now);
	fp = fopen("/usr/adm/shutdownlog", "a");
	if (fp == 0)
		return;
	fseek(fp, 0L, 2);
	fprintf(fp, "%02d:%02d  %s %s %2d, %4d.  Reboot", tm->tm_hour,
		tm->tm_min, days[tm->tm_wday], months[tm->tm_mon],
		tm->tm_mday, tm->tm_year + 1900);
	if (panicstr)
		fprintf(fp, " after panic: %s\n", panic_mesg);
	else
		putc('\n', fp);
	fclose(fp);
}

/*
 * Versions of std routines that exit on error.
 */

Open(name, rw)
	char *name;
	int rw;
{
	int fd;

	if ((fd = open(name, rw)) < 0) {
		perror(name);
		exit(1);
	}
	return fd;
}

Read(fd, buff, size)
	int fd, size;
	char *buff;
{
	int ret;

	if ((ret = read(fd, buff, size)) < 0) {
		perror("read");
		exit(1);
	}
	return ret;
}

off_t
Lseek(fd, off, flag)
	int fd, flag;
	long off;
{
	long ret;

	if ((ret = lseek(fd, off, flag)) == -1L) {
		perror("lseek");
		exit(1);
	}
	return ret;
}

Create(file, mode)
	char *file;
	int mode;
{
	register int fd;

	if ((fd = creat(file, mode)) < 0) {
		perror(file);
		exit(1);
	}
	return fd;
}

Write(fd, buf, size)
	int fd, size;
	char *buf;
{

	if (write(fd, buf, size) < size) {
		perror("write");
		exit(1);
	}
}

/* seek to a physical address location in the 'fd'
 * file, given a system logical address, 'loc'.
 */
klseek(fd, loc, off)
	int fd;
	long loc;       /* logical physical address */
	int off;        /* from current position, or absolute */
{

	long v;
	long addr;
	struct pte pte;

	loc &= ~0xc0000000;
	v = btop(loc);
	if(v >= nl[X_SYSSIZE].n_value) {
		printf("address botch %x\n", loc);
		exit(1);
	}
	addr = (long)((struct pte *)nl[X_SYSMAP].n_value + v);
	lseek(fd, (addr & ~0xc0000000) + dumplo, 0);
	if(read(fd, (char *)&pte, sizeof(pte)) != sizeof(pte)) {
		printf("Error reading kmem for pte at %x\n", addr);
		exit(1);
	}
	if (pte.pg_v == 0 && (pte.pg_fod || pte.pg_pfnum == 0)) {
		printf("pte bad for %x\n", addr);
		exit(1);
	}
	loc = (long)ptob(pte.pg_pfnum) + (loc & PGOFSET);
	(void) lseek(fd, (long)(loc + dumplo), off);
}
