/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

/*    	"@(#)sdb:com/machdep.c	1.18"   */

/*
 ****		MACHINE and OPERATING SYSTEM DEPENDENT
 ****		Routines which deal with the run-time stack
 */

#include "head.h"
#include "coff.h"

extern BKPTR	bkpthead;
extern MSG		NOPCS;

long rdsym();			/* in symt.c */
extern SYMENT syment;		/* rdsym() stores symbol table entry */
extern AUXENT auxent[];		/* rdsym() stores auxiliary entry */
extern int gflag;		/* initfp() in symt.c sets */
extern FILHDR filhdr;
static int frameflg;		/* set if now past last frame */

#define EVEN	(-2)

/* initialize frame pointers to top of call stack */
/*  MACHINE DEPENDENT */
struct proct *
initframe() {
	register struct proct *procp;

#if DEBUG
	if (debugflag == 1)
		enter1("initframe");
	else if (debugflag == 2)
	{
		enter2("initframe");
		closeparen();
	}
#endif
	if ((pid == 0) && (fcor < 0 || fakecor)) {
		frameflg = 1;
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("initframe");
			printf("0x%x",badproc);
			endofline();
		}
#endif
		return(badproc);
	}

	frame = SDBREG(FP);
	callpc = SDBREG(PC);


	/* FP in Power 6/32 starts at maxstor = 1L << 31 and decreases */
	if ((frame & 0xf0000000) != 0xB0000000) {
		frameflg = 1;
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("initframe");
			printf("0x%x",badproc);
			endofline();
		}
#endif
		return(badproc);
	}

	procp = adrtoprocp(callpc);
	frameflg = 0;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("initframe");
		printf("0x%x",procp);
		endofline();
	}
#endif
	return(procp);
}

struct proct *
nextframe() {
	register struct proct *procp;
	union word word;

#if DEBUG
	if (debugflag == 1)
		enter1("nextframe");
	else if (debugflag == 2)
	{
		enter2("nextframe");
		closeparen();
	}
#endif
	callpc = get(frame+8, DSP);
	frame = get(frame, DSP) & EVEN;
	if (callpc > 0xB0000000)	/* error handler kludge */
	{
		fprintf( FPRT1, "nextframe(): callpc = %#x; \n", callpc );
		callpc = get(frame+24, DSP);
	}

	if ((frame & 0xf0000000) != 0xB0000000) {
		frameflg = 1;
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("nextframe");
			printf("0x%x",badproc);
			endofline();
		}
#endif
		return(badproc);
	}
	procp = adrtoprocp(callpc);

#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("nextframe");
		printf("0x%x",procp);
		endofline();
	}
#endif
	return(procp);
}

/* returns core image address for variable */
/* formaddr() should be in symt.c ?? */
ADDR
formaddr(class, addr)
register char class;
ADDR addr; 
{
	ADDR retaddr;

#if DEBUG
	if (debugflag == 1)
		enter1("formaddr");
	else if (debugflag == 2)
	{
		enter2("formaddr");
		arg("class");
		printf("0x%x",class);
		comma();
		arg("addr");
		printf("0x%x",addr);
		closeparen();
	}
#endif
	switch(class) {
	case C_REG:
	case C_REGPARM:
		retaddr = stackreg(addr);
		break;

	case C_EXT:
	case C_STAT:
	case C_MOS:
	case C_MOU:
	case C_MOE:
	case C_FIELD:			/* ?? */
		retaddr = addr;
		break;
		
	case C_ARG:
	case C_AUTO:
		retaddr = frame+addr;	/* addr was negated in outvar() */
		break;

	default:
		fprintf(FPRT1,"Bad class in formaddr: %d(%#x)\n", class, class);
		retaddr = 0;
		break;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("formaddr");
		printf("0x%x",retaddr);
		endofline();
	}
#endif
	return (retaddr);
}

/*
 *  stackreg(reg):
 * This routine is intended to be used to find the address of a register
 * variable.  The stack is searched backwards from the current frame
 * until the procedure with the register variable is found.  If the
 * register assigned to the variable appears on the stack, then its
 * address is returned.  Otherwise, the register variable is still in
 * the named register, so the register number is returned.  We distinguish
 * addresses from register numbers by noting that register numbers are less
 * than NUMREGS (16) and that stack addresses are greater.
 *
 *  MACHINE DEPENDENT
 */
/* new stackreg() for 3B Unix
 *	C stores 'n' registers starting with %R8 and going backwards
 *	can only get 'n' by looking at text -- bits 8->11 from beginning of proc
 */

/*  Power 632 version - C stores 'n' registers arbitrarily.
 *	A mask (16 bits) is stored as the high order short in the word which is
 *	offset one word from the beginning of the frame (frame + WORDSIZE).
 *	The 0'th bit of this mask tells whether register 0 has been saved,
 *	the 1'st bit tells wheter register 1 has been saved, etc.
 *	Each saved register in the frame occupies another word of space,
 *	and they are stored in order, from lowest to highest numbered.
 *	The first saved register is at frame + 5 * WORDSIZE.
 */

#define REGOFF 12
#define MASKOFF 4
ADDR
stackreg(reg) {
	register int regfl, mask, i;
	register struct proct *procp;
	ADDR regaddr;

#if DEBUG
	if (debugflag == 1)
		enter1("stackreg");
	else if (debugflag == 2)
	{
		enter2("stackreg");
		arg("reg");
		printf("0x%x",reg);
		closeparen();
	}
#endif
	if (proc[0] == '\0')
	{
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("stackreg");
			printf("0x%x",reg);
			endofline();
		}
#endif
		return(reg);
	}
	regaddr = reg;
	regfl = 0x10000 << reg;
	for (procp=initframe(); procp!=badproc; procp=nextframe()) {
		if (sl_procp == procp)
			break;
		if ((mask = get(frame + MASKOFF, DSP)) & regfl) {
			regaddr = frame + REGOFF;
			for (i=0; i < reg; i++) {
				if (mask & 0x10000) regaddr += WORDSIZE;
				mask >>= 1;
			}
		}
	}
	if (procp == badproc) {
		fprintf(FPRT1, "Stackreg() error: frame=%#x\n", frame);
		regaddr = ERROR;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("stackreg");
		printf("0x%x",regaddr);
		endofline();
	}
#endif
	return(regaddr);
}

/* Print the stack trace: 't' and 'T' commands
 * set top (i.e. top != 0) to print just the top procedure
 * modified to print something even if a.out stripped
 */

prfrx(top) {
	int narg, argp;		/* number of words that arguments occupy */
	long offs;		/* offset into symbol table */
	register char class;	/*storage class of symbol */
	register int endflg;	/* 1 iff cannot find symbolic names of
					more arguments */
	int subsproc = 0;	/* 1 iff not top function on stack */
	register char *p;	/* pointer to procedure (function) name */
	int line_num;		/* line number in calling function */
	register struct proct *procp;
	SYMENT *syp = &syment;
	AUXENT *axp = auxent;
	
#if DEBUG
	if (debugflag == 1)
		enter1("prfrx");
	else if (debugflag == 2)
	{
		enter2("prfrx");
		arg("top");
		printf("0x%x",top);
		closeparen();
	}
#endif
	procp = initframe();
	if (frameflg) {		/*  no initial frame */
		if (pid == 0 && (fcor < 0 || fakecor))	{  /* usual error */
			errflg = "No process and no core file.";
			chkerr();
		}
		else {				/* unusual initial frame */
#if DEBUG
			if (debugflag == 1)
				exit1();
			else if (debugflag == 2)
			{
				exit2("prfrx");
				endofline();
			}
#endif
			return;
		}
	}
	do {
		/*  Power 6/32 crt0 (start) gets a current fp of zero */
		if (get (frame + 12, DSP) == 0) 
		{
#if DEBUG
			if (debugflag == 1)
				exit1();
			else if (debugflag == 2)
			{
				exit2("prfrx");
				endofline();
			}
#endif
			return;
		}
		p = procp->pname;
		if (eqstr("__dbsubc", p))	/*  3B ?? */
		{
#if DEBUG
			if (debugflag == 1)
				exit1();
			else if (debugflag == 2)
			{
				exit2("prfrx");
				endofline();
			}
#endif
			return;
		}
		if (procp == badproc) {		/*  if stripped a.out */
			printf("pc: 0x%lx;	args: (", callpc);
			endflg = 1;
		}
		else if (p[0] == '_') {
			printf("%s(", p+1);
			endflg = 0;
		}
		else {
			printf("%s(", p);
			endflg = 0;
		}
		if (endflg == 0) {
			offs = procp->st_offset;
			do {		/*  in COFF, always just 1 ? */
				if( (offs = rdsym(offs)) == ERROR) {
					endflg++;
					break;
				}
			} while (ISFCN(syp->n_type));
			class = syp->n_sclass;
			while (! ISARGV(class)) {
				if ((offs = rdsym(offs)) == ERROR) {
					endflg++;
					break;
				}
				class = syp->n_sclass;
				if (ISFCN(syp->n_type)) {
					endflg++;
					break;
				}
			}
		}

		narg = ((get (frame - 4, DSP) & 0xffff) - 4)/4;
		if (narg & ~0xff)	narg = 0;
		argp = frame + WORDSIZE;
		while (narg) {
			if (endflg) {
				printf("%d", get(argp, DSP));
				argp += 4;
			} else {
				int length;
				if ((syp->n_type == T_STRUCT) ||
				    (syp->n_type == T_UNION))   {
				    /* print address of structure
				     * (so that structures of, e.g.
				     * 100 element arrays, are not dumped)
				     */
				    /* Power 6/32: skip leading underscore */
				    if (syp->n_nptr[0] == '_')
					printf("&%s=", syp->n_nptr+1);
				    else
					printf("&%s=", syp->n_nptr);
				    dispx(argp, "x", C_EXT, 
						(short) (-1), 0, DSP);
				    length = axp->x_sym.x_misc.x_lnsz.x_size;
				}
				else {
				    if (syp->n_type == T_ENUM)
					length =
					    axp->x_sym.x_misc.x_lnsz.x_size;
				    else
					length = typetosize(syp->n_type, 0);
				    /* Power 6/32: skip leading underscore */
				    if (syp->n_nptr[0] == '_')
				    {
					printf( "%s=", syp->n_nptr + 1 );
				    }
				    else
				    {
					printf( "%s=", syp->n_nptr );
				    }
				    dispx(argp, "", C_EXT, 
					    (short) syp->n_type, 0, DSP);
				}
				if (length > WORDSIZE) {
					argp += length;
					/*  adjust for more than 1 word */
					/* narg -= length/WORDSIZE -1; */
				}
				/* bytes, shorts, longs pushed as ints */
				else {
					argp += WORDSIZE;
				}
			}
			do {
				if (endflg) break;
				if ((offs = rdsym(offs)) == ERROR) {
					endflg++;
					break;
				}
				class = syp->n_sclass;
				if (ISFCN(syp->n_type)) {
					endflg++;
					break;
				}
			} while (! ISARGV(class));
			if (--narg != 0) printf(",");
		}
		printf(")");
		if (procp->sfptr != badfile)
			printf("   [%s",
				(procp->sfptr)->sfilename);
		if(gflag) {
			if ((line_num = adrtolineno(callpc-subsproc,procp)) > 0)
				printf(":%d", line_num);
		}
		if(procp->sfptr != badfile)
			printf("]");
		printf("\n");
		subsproc = 1;
		procp = nextframe();
	} while (!top && !frameflg);	/*  only one frame desired, or
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("prfrx");
		endofline();
	}
#endif
						no frames left in backtrace */
}


/* machine dependent initialization */
sdbenter(xflag) {
	mkioptab();
}

/* machine dependent exit code */
sdbexit() {
	exit(0);
}

