	/* @(#) opset.c: 1.12         */
/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

/*
 *
 *	UNIX debugger
 *
 *		Instruction printing routines.
 *		MACHINE DEPENDENT.
 */

#include "head.h"

#define		CASEL	0xFC
#define		MAXOFF	0x400

STRING		errflg;
L_INT		vvar[36];


/* instruction printing */

/*
 * Argument access types
 */
#define ACCA	(8<<3)	/* address only */
#define ACCR	(1<<3)	/* read */
#define ACCW	(2<<3)	/* write */
#define ACCM	(3<<3)	/* modify */
#define ACCB	(4<<3)	/* branch displacement */
#define ACCI	(5<<3)	/* XFC code */

/*
 * Argument data types
 */
#define TYPB	0	/* byte */
#define TYPW	1	/* word */
#define TYPL	2	/* long */
#define TYPQ	3	/* quad */
#define TYPF	4	/* floating */
#define TYPD	5	/* double floating */

#define	DSYM	4
#define	ISYM	4
#define	shorten(s)	((s >> 16) & 0xFFFF)

TYPE	struct optab	*OPTAB;
struct optab {
	char *iname;
	char val;
	char nargs;
	char argtype[6];
} optab[];
STRING	regname[];
POS	type, space, incp;

int ioptab[256]; /* index by opcode to optab */

mkioptab() {/* set up ioptab */
REG OPTAB p=optab; while (p->iname) {ioptab[p->val&LOBYTE]=p-optab; p++;}
}

extern char *fmtr;
extern char *fmtR;

printins(fmt,idsp,ins)
char fmt;
REG L_INT	ins;
{
	short i,b,mode; char **r; char *fmat;
	struct proct *procp;
	REG char *	ap;
	REG OPTAB	ip;
	REG	regno;
	REG	d;	/* assembled byte, word, long, or float */

#if DEBUG
	if (debugflag == 1)
		enter1("printins");
	else if (debugflag == 2)
	{
		enter2("printins");
		arg("fmt");
		printf("0x%x",fmt);
		comma();
		arg("idsp");
		printf("0x%x",idsp);
		comma();
		arg("ins");
		printf("0x%x",ins);
		closeparen();
	}
#endif
	procp = adrtoprocp(dot);
	if (procp->paddress == dot) {
		printf("0x%04.4x", ins & 0xffff);
		oincr = 2;
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("printins");
			endofline();
		}
#endif
		return;
	}

	type=DSYM; space=idsp;
	ins = byte (ins);
	ip=optab+ioptab[ins]; printf("%s\t",ip->iname); incp=1;
	ap=ip->argtype;
	for (i=0; i<ip->nargs; i++,ap++) {
		vvar[i]=0x80000000;
		if (i!=0) printc(',');
	  top:
		if (*ap&ACCB) 
			b= 0xAF + ((*ap&7)<<5);  /* branch displacement */
		else {
			b=bchkget(inkdot(incp),idsp);
			++incp;
			}
		regno = b & 0xF;
		mode= b >>= 4;
	  mid:
			switch ((int)mode) {
				case 0:
				case 1:
				case 2:
				case 3: /* short literal */
					printc ('$');
					d = mode << 4 | regno;
					goto disp;
				case 4: /* [r] */ 
					printf("[%s]",regname[regno]); goto top;
				case 5: /* r */ 
					printf("%s",regname[regno]); break;
				case 7: /* -(r) */ 
					printc('-');
			  base:
				case 6: /* (r) */ 
					printf("(%s)",regname[regno]); break;
				case 9: /* *(r)+ */ 
					printc('*');
				case 8: /* (r)+ */
				  	if (regno==0xF || mode ==8 && 
						(regno==8 || regno==9)) 
				  	{   /* PC: immediate or absolute */
						printc('$'); 
						d = snarf ((regno&03)+1, idsp);
						goto disp;
				  	}
				  	else
				  	{
				  		printf("(%s)+",regname[regno]); 
						break;
				  	}
				  
				case 0xB: 		/* byte disp deferred */
				case 0xD:		/* word disp deferred */
				case 0xF:		/* long disp deferred */
					printc('*');
				case 0xA: 		/* byte disp */
				case 0xC:		/* word disp */
				case 0xE:		/* long disp */
					d = snarf (1<<((mode>>1&03)-1), idsp);
					if (regno == 0xF) 
					{ /* PC offset addressing */
						d += dot + incp;
						psymoff (d, ISYM, fmt);
						vvar[i] = d;
						break;
					}
			  disp:
					if (d >= 0 && d < MAXOFF)
						printf ("%d", d);
					else
						psymoff (d, type, fmt);
					if (mode >= 0xA)
						printf ("(%s)", regname[regno]);
					vvar[i] = d;
				  	break;
			}
	}
	if (ins == CASEL) {	/* CASEL instruction */
		if (inkdot(incp) & 01)		/* align */
			incp++;
		for (i=0; i<=vvar[2]; ++i) {
			printc(EOR); printf("    %d:  ",i+vvar[1]);
			d = shorten (get (inkdot (incp+i+i),idsp));
			if (d&0x8000) d -= 0x10000;
			psymoff(inkdot(incp)+d,type,fmt);
		}
		incp += vvar[2]+vvar[2]+2;
	}
	oincr=incp;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("printins");
		endofline();
	}
#endif
}

L_INT	inkdot(incr)
{
	L_INT		newdot;

#if DEBUG
	if (debugflag == 1)
		enter1("inkdot");
	else if (debugflag == 2)
	{
		enter2("inkdot");
		arg("incr");
		printf("0x%x",incr);
		closeparen();
	}
#endif
	newdot=dot+incr;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("inkdot");
		printf("0x%x",newdot);
		endofline();
	}
#endif
	return(newdot);
}

printc(c)
char c; {
	printf("%c", c);
}

snarf (nbytes, idsp)
{
	register int value;

#if DEBUG
	if (debugflag == 1)
		enter1("snarf");
	else if (debugflag == 2)
	{
		enter2("snarf");
		arg("nbytes");
		printf("0x%x",nbytes);
		comma();
		arg("idsp");
		printf("0x%x",idsp);
		closeparen();
	}
#endif
	value = chkget (inkdot (incp), idsp);
	incp += nbytes;
	value = value >> (4-nbytes)*8;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("snarf");
		printf("0x%x",value);
		endofline();
	}
#endif
	return (value);
}


psymoff(val, regno, fmt)
L_INT val; char fmt, **regno; {
	struct proct *procp;
	register int diff;

#if DEBUG
	if (debugflag == 1)
		enter1("psymoff");
	else if (debugflag == 2)
	{
		enter2("psymoff");
		arg("val");
		printf("0x%x",val);
		comma();
		arg("regno");
		printf("0x%x",regno);
		comma();
		arg("fmt");
		printf("0x%x",fmt);
		closeparen();
	}
#endif
	if (fmt == 'i') {
		if (regno == regname + 13) {   /* local */
			if ((diff = adrtolocal((ADDR) -val, adrtoprocp(dot))
					) != -1) {
				printf("%s", sl_name);
				prdiff(diff);
#if DEBUG
				if (debugflag == 1)
					exit1();
				else if (debugflag == 2)
				{
					exit2("psymoff");
					printf("0x%x",0);
					endofline();
				}
#endif
				return(0);
			}
		}
		if (val < firstdata) {
			if ((procp = adrtoprocp((ADDR) val)) != badproc) {
				prlnoff(procp, val);
#if DEBUG
				if (debugflag == 1)
					exit1();
				else if (debugflag == 2)
				{
					exit2("psymoff");
					printf("0x%x",0);
					endofline();
				}
#endif
				return(0);
			}
		} else {
			if ((diff = adrtoext((ADDR) val)) != -1) {
				printf("%s", sl_name);
				prdiff(diff);
#if DEBUG
				if (debugflag == 1)
					exit1();
				else if (debugflag == 2)
				{
					exit2("psymoff");
					printf("0x%x",0);
					endofline();
				}
#endif
				return(0);
			}
		}
	}
	prhex(val);
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("psymoff");
		printf("0x%x",1);
		endofline();
	}
#endif
	return(1);
}


prdiff(diff) {
#if DEBUG
	if (debugflag == 1)
		enter1("prdiff");
	else if (debugflag == 2)
	{
		enter2("prdiff");
		arg("diff");
		printf("0x%x",diff);
		closeparen();
	}
#endif
	if (diff) {
		printf("+");
		prhex(diff);
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("prdiff");
		endofline();
	}
#endif
}
