/*	Copyright (c) 1984 AT&T	*/
/*	  All Rights Reserved  	*/

/*	THIS IS UNPUBLISHED PROPRIETARY SOURCE CODE OF AT&T	*/
/*	The copyright notice above does not evidence any   	*/
/*	actual or intended publication of such source code.	*/

/*	"@(#)sdb:com/runpcs.c	1.23"   */

/*
 *	UNIX debugger
 */

#include "head.h"


/* breakpoint instruction */
#define BPT	0x30

extern MSG		NOFORK;
extern MSG		ENDPCS;
extern MSG		BADWAIT;
extern MSG		NOPCS;

extern ADDR		sigint;
extern ADDR		sigqit;

/* breakpoints */
extern BKPTR		bkpthead;

extern CHAR		lastc;
extern int		errno;
extern INT		pid;
extern L_INT		loopcnt;

extern ADDR	userpc;



getsig(sig)
{	return(sig);
}

/* runpcs - control the user process, running it in single step or
 *	continuous mode, depending on runmode, and with signal execsig.
 *
 *	It has the side effect of deleting all breakpoints (calls delbp()).
 */
runpcs(runmode,execsig)
{
	REG BKPTR	bkpt;
	int cmdval;

#if DEBUG
	if (debugflag == 1)
		enter1("runpcs");
	else if (debugflag == 2)
	{
		enter2("runpcs");
		arg("runmode");
		printf("0x%x",runmode);
		comma();
		arg("execsig");
		printf("0x%x",execsig);
		closeparen();
	}
#endif
	/* for 'g' (goto) command, set userpc to address to which to jump */
	if (adrflg) userpc = dot;

	/* execute breakpoints (or single step) loopcnt number of times */
	while (loopcnt-- > 0) {
		errno = 0;
		if (runmode == SINGLE) {
			/* hardware handles single-stepping */
			delbp();
			ptrace(SINGLE, pid, userpc, execsig);
			bpwait(); chkerr(); execsig=0; readregs();

		}

		else {		/* continuing from a breakpoint is hard */

			/* if truly continuing from bkpt, execute instruction
			 * that should be where breakpoint was set	    */
			if   ((bkpt=scanbkpt(userpc))
			  && !(adrflg && args[0]=='\0') /*  'g' cmd */
/* ?? 0 bkpt */		  && !(userpc==0 && (cmd=='r'||cmd=='R')))  
			{
				execbkpt(bkpt, execsig);
				execsig = 0;
			}
			setbp();
			ptrace(runmode, pid, userpc, execsig);
			bpwait(); chkerr(); execsig=0;
			delbp();
			readregs();
		}

		/* stopped by BPT instruction */

		if ((signo==0) && (bkpt=scanbkpt(userpc))) {
	loop1:

			dot=bkpt->loc;
			if (bkpt->comm[0] != '\n') {
				cmdval = acommand(bkpt->comm);
				/* if not "really" breakpoint, i.e.
				 * user did not give k command, then
				 * cmdval == 1, and should execute
				 * breakpoint instruction, not count
				 * as a breakpoint in the loop, and
				 * continue to the next breakpoint
				 */
				if (cmdval) {
					execbkpt(bkpt,execsig);
					execsig=0;
					/* if at another breakpoint,
					 * handle as before (check for
					 * user command, by going to
					 * loop1).
					 */
					if ((signo == 0) &&
					    (bkpt=scanbkpt(userpc)))
						goto loop1;
					loopcnt++;  /* Don't count iteration */
					continue;
				}
			}
		}

		/* else not stopped by a BPT instruction */
		else {
			execsig=signo;
			if (execsig) break;
		}
	}	/* end of while loop */
#if DEBUG
	if (debugflag == 1)
		exit1();
		else if (debugflag == 2)
	{
		exit2("runpcs");
		endofline();
	}
#endif
}


#define BPOUT 0
#define BPIN 1
INT bpstate = BPOUT;

/* endpcs() - kill child (user) process */
endpcs()
{
	REG BKPTR	bkptr;

#if DEBUG
	if (debugflag == 1)
		enter1("endpcs");
	else if (debugflag == 2)
	{
		enter2("endpcs");
		closeparen();
	}
#endif
	if (pid) {
		errno = 0;
   		ptrace(EXIT,pid,0,0); pid=0;
		chkerr();

		/*  cleanup of breakpoint states not necessary.
		*	Only two modes being used are BKPTSET and 0.
		*/
	}
	bpstate=BPOUT;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("endpcs");
		endofline();
	}
#endif
}

/* setup() -initializes ptrace, forks the child (user) process, and
 *		waits for child.
 */
setup()
{
#if DEBUG
	if (debugflag == 1)
		enter1("setup");
	else if (debugflag == 2)
	{
		enter2("setup");
		closeparen();
	}
#endif
	close(fsym); fsym = -1;
	if ((pid = fork()) == 0)
	{
		ptrace(SETTRC,0,0,0);

		/* restore user interrupt handling mode */
		signal(SIGINT,sigint); signal(SIGQUIT,sigqit);
		doexec(); exit(0);
	}
	else if (pid == -1) {
		error(NOFORK);
		longjmp(env);
	}
	else {
		bpwait();

		readregs();
		fsym=open(symfil,0);
		if (errflg) {
			fprintf(FPRT1, "%s: cannot execute\n",symfil);
			endpcs();
		}
	}
	bpstate=BPOUT;
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("setup");
		endofline();
	}
#endif
}

/* execbkpt(bkptr, execsig) - execute the one instruction that was replaced
 *		by a preakpoint instruction.  It has the side effect of
 *		removing all breakpoints, by calling delbp().
 *
 *		In the case of the 3B, a NULL bkptr means that single
 *		stepping is requested, and that a breakpoint instruction
 *		has not just been encountered.
 */

static
execbkpt(bkptr, execsig)
BKPTR bkptr;
{
#if DEBUG
	if (debugflag == 1)
		enter1("execbkpt");
	else if (debugflag == 2)
	{
		enter2("execbkpt");
		arg("bkptr");
		printf("0x%x",bkptr);
		comma();
		arg("execsig");
		printf("0x%x",execsig);
		closeparen();
	}
#endif
	delbp();
	ptrace(SINGLE, pid, bkptr->loc, execsig);
	bkptr->flag=BKPTSET;
	bpwait(); chkerr(); readregs();
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("execbkpt");
		endofline();
	}
#endif
}

/* doexec() - performs exec call, after parsing arguments on sdb "run" command
 *		line.
 */
extern STRING environ;

doexec()
{
	char *argl[MAXARG], args[LINSIZ];
	register char c, redchar, *argsp, **arglp, *filnam;

#if DEBUG
	if (debugflag == 1)
		enter1("doexec");
	else if (debugflag == 2)
	{
		enter2("doexec");
		closeparen();
	}
#endif
	arglp = argl;
	argsp = args;
	*arglp++ = symfil;
	c = ' ';

	do {
		while (eqany(c, " \t")) {
			c = rdc();	/* get char from extern args[] */
		} 
		if (eqany(c, "<>")) {	/* redirecting I/O argument */
			redchar = c;
			do {
				c = rdc();
			} while (eqany(c, " \t"));
			filnam = argsp;
			do {
				*argsp++ = c;
				c = rdc();
			} while (!eqany(c, " <>\t\n"));
			*argsp++ = '\0';
			if (redchar == '<') {
				close(0);
				if (open(filnam,0) < 0) {
					fprintf(FPRT1,
						"%s: cannot open\n",filnam);
					 exit(0);
				}
			} else {
				close(1);
				if (creat(filnam,0666) < 0) {
					fprintf(FPRT1,
						"%s: cannot create\n",filnam);
					 exit(0);
				}
			}
		} else if (c == '"' || c == '\'') {  /*  quoted argument */
			char sc = c;
			*arglp++ = argsp;
			for (c = rdc(); c != sc; c = rdc()) {
				if (c == '\\')
					c = bsconv(rdc());
				*argsp++ = c;
			}
			c = rdc();
			*argsp++ = '\0';
		} else if (c != '\n') {		/* a regular argument */
			*arglp++ = argsp;
			do {
				*argsp++ = c;
				c = rdc();
			} while(!eqany(c, " <>\t\n"));
			*argsp++ = '\0';
		}
	} while (c != '\n');
	*arglp = (char *) 0;
	/* exect(symfil, argl, environ); */
	/* perror("Returned from exect"); */
	execve(symfil, argl, environ);	/* exect -> execve (gila) */
	perror("Returned from exec");
}

/* scanpkpt() - returns a pointer to the entry in the breakpoint list
 *		corresponding to the address given, if that is a
 *		breakpoint address.  Otherwise returns NULL pointer.
 */
BKPTR
scanbkpt(adr)
ADDR adr;
{
	REG BKPTR	bkptr;

#if DEBUG
	if (debugflag == 1)
		enter1("scanbkpt");
	else if (debugflag == 2)
	{
		enter2("scanbkpt");
		arg("adr");
		printf("0x%x",adr);
		closeparen();
	}
#endif
	for(bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt) {
		if (bkptr->flag && bkptr->loc==adr) break;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("setbp1");
		printf("0x%x",bkptr);
		endofline();
	}
#endif
	return(bkptr);
}

/* delbp(). Remove breakpoints [delbp1() puts original instruction back] */

delbp()
{
	REG BKPTR	bkptr;

#if DEBUG
	if (debugflag == 1)
		enter1("delbp");
	else if (debugflag == 2)
	{
		enter2("delbp");
		closeparen();
	}
#endif
	if (bpstate != BPOUT) {
		for (bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt) {
			if (bkptr->flag) {

				delbp1(bkptr);	/*  separate subr */
			}
		}
		bpstate=BPOUT;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("delbp");
		endofline();
	}
#endif
}

/* delbp1(bkptr) - deletes breakpoint given by bkptr */

static
delbp1(bkptr)
BKPTR	bkptr;
{
	register long adr;
	register int ptv;
	struct proct *procp;

#if DEBUG
	if (debugflag == 1)
		enter1("delbp1");
	else if (debugflag == 2)
	{
		enter2("delbp1");
		arg("bkptr");
		printf("0x%x",bkptr);
		closeparen();
	}
#endif
	adr = bkptr->loc;
	ptv = ptrace(RIUSER, pid, adr, 0);
	if (!(errno && ptv == -1)) {
		ptv = ptrace(WIUSER, pid, adr,
			(bkptr->ins & 0xff000000) | (ptv & 0xffffff));
		if (!(errno && ptv == -1)) return;
	}
	procp = adrtoprocp(adr);
	if (procp->pname[0] == '_')
		fprintf(FPRT1, "Cannot delete breakpoint:  %s:%d @ %d\n",
			(procp->pname)+1, adrtolineno(adr, procp), dot);
	else
		fprintf(FPRT1, "Cannot delete breakpoint:  %s:%d @ %d\n",
			procp->pname, adrtolineno(adr, procp), dot);
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("delbp1");
		endofline();
	}
#endif
}



/* setbp() -
 * Insert breakpoints [setbp1() overwrites instruction space with bpt instr] *
 */
setbp()
{
	REG BKPTR	bkptr;

#if DEBUG
	if (debugflag == 1)
		enter1("setbp");
	else if (debugflag == 2)
	{
		enter2("setbp");
		closeparen();
	}
#endif
	if (bpstate != BPIN) {
		for (bkptr=bkpthead; bkptr; bkptr=bkptr->nxtbkpt) {
			if (bkptr->flag) {
				setbp1(bkptr);	/*  separate subr */
			}
		}
		bpstate=BPIN;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("setbp");
		endofline();
	}
#endif
}

/* setpb1(bkptr) - sets breakpoint given by bkptr */

static
setbp1(bkptr)
BKPTR	bkptr;
{
	register long adr;
	register int ptv;
	struct proct *procp;

#if DEBUG
	if (debugflag == 1)
		enter1("setbp1");
	else if (debugflag == 2)
	{
		enter2("setbp1");
		arg("bkptr");
		printf("0x%x",bkptr);
		closeparen();
	}
#endif
	adr = bkptr->loc;
	errno = 0;
	ptv = ptrace(RIUSER, pid, adr, 0);

	if (!(errno && ptv == -1)) {
		bkptr->ins = ptv;
		ptv = ptrace(WIUSER, pid, adr, 
			BPT << 24 | (ptv & 0xffffff));
		if (!(errno && ptv == -1)) return;
	}

	procp = adrtoprocp(adr);
	if (procp->pname[0] == '_')
		fprintf(FPRT1, "Cannot set breakpoint:  %s:%d @ %d\n",
			(procp->pname)+1, adrtolineno(adr, procp), dot);
	else
		fprintf(FPRT1, "Cannot set breakpoint:  %s:%d @ %d\n",
			procp->pname, adrtolineno(adr, procp), dot);
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("setbp1");
		endofline();
	}
#endif
}

bpwait()
{
	REG ADDR w;
	INT stat;

#if DEBUG
	if (debugflag == 1)
		enter1("bpwait");
	else if (debugflag == 2)
	{
		enter2("bpwait");
		closeparen();
	}
#endif
	signal(SIGINT, 1);
	while ((w = wait(&stat)) != pid  &&  w!=-1) ;
	signal(SIGINT,sigint);
	if (w == -1) {
		pid = 0;
		errflg=BADWAIT;
	}
	else if ((stat & 0177) != 0177) {
		if (signo = stat&0177) sigprint();
		if (stat&0200) {
			error(" - core dumped");
			close(fcor);
			corfil = "core";
			setcor();	/* get most recent core dump */
		}
		pid=0;
		errflg=ENDPCS;
	}
	else {
		signo = stat>>8;
		if (signo != SIGTRAP &&
		    ptrace(RUREGS, pid, SYSREG(PC), 0) != adsubn)
		{
			sigprint();
		}
		else	signo = 0;
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("bpwait");
		endofline();
	}
#endif
}

readregs()
{
	/*get REG values from pcs*/
	REG i;
	REG REGPTR p;

#if DEBUG
	if (debugflag == 1)
		enter1("readregs");
	else if (debugflag == 2)
	{
		enter2("readregs");
		closeparen();
	}
#endif
	for (p=reglist; p->rname; p++)
	{
        	SDBREG(p->roffs) =
                    ptrace(RUREGS, pid, SYSREG(p->roffs)
			*NBPW
		    , 0);
	}
	if( pid == 0 && PC < 0 )
	{
		error( NOPCS );
#if DEBUG
		if (debugflag == 1)
			exit1();
		else if (debugflag == 2)
		{
			exit2("readregs");
			printf("0x%x",ERROR);
			endofline();
		}
#endif
		return(ERROR);
	}
	else
	{
		dot = userpc = SDBREG(PC);
	}
#if DEBUG
	if (debugflag == 1)
		exit1();
	else if (debugflag == 2)
	{
		exit2("readregs");
		endofline();
	}
#endif
}

char 
readchar() {
	lastc = *argsp++;
	if (lastc == '\0') lastc = '\n';
	return(lastc);
}

char
rdc()
{
	register char c;

	c = *argsp++;
	return(c == '\0' ? '\n' : c);
}

