%{	/*	HCX assembler parse logic */

#include "stdio.h"

#include "symbols.h"
#include "instab.h"
#include "gendefs.h"
#include "ctype.h"
#include <syms.h>

#define	TXTCHARS	1024       /* size of lexical buffer */

#define defsym(x)	((x.exptype==TSYM)&&((x.symptr->styp&TYPE)!=UNDEF))

#define	isoctdigit(c)	(isdigit(c) && (c<='7'))  /* check for octal digit */

extern unsigned short
		cline, /* line number from the .ln pseudo op */
		line; /* current input file line number */

extern short
		readonly, /* put .data stuff into .text segment */
		opt; /* sdi optimization flag */

extern FILE
		*fdin; /* input source */

extern char
		file[], /* assembly input file name */
		cfile[]; /* original C file name */

extern int local_label_number[10];        /* local label sequence numbers   */
extern short local_label_fref_flag[10];   /* local label forward references */
short local_label_def_flag[10];           /* local label defined indicators */

char asciifile[TXTCHARS];         /* buffer to hold .ascii, .asciz strings  */

int quadflag = 0;          /* flag to indicate parsing of quad operand */

extern symbol	*dot;	   /* pc pointer symbol */


extern char *infile;       /* name of source input file */

extern long	newdot,    /* up to date value of pc pointer symbol */
		dotbss;    /* bss pointer symbol */

BYTE	bytesout;	   /* output displacement */
short	bitsout = 0;	   /* nbr of bits accumulated with bit fields */
long	outword = 0;	   /* bit field accumulator */
long	is_tdiff = 0;      /* was there a case of sym - sym ? */
long	is_opcode = 0;     /* is it an operand expression ? */
int     asciflag;          /* 1 => .asciz; 0 => .ascii      */

rexpr	operate(); /* procedure to perform expression operations */

extern upsymins	*lookup(); /* symbol table access routine */

static BYTE defseen = 0; /* flag set if .def expression found on this line */

%}

%union	{
		long	unlong;
		struct { long    unquad[2]; } qs;
		double	unfloat;
		rexpr	unrexpr;
		symbol	*unsymptr;
		instr	*uninsptr;
		addrmode unaddr;
		char	*unstrptr;
	}


%token	<uninsptr>	ZEROP
%token	<uninsptr>	ONEOP
%token	<uninsptr>	TWOOP
%token	<uninsptr>	THROP
%token	<uninsptr>	FOUOP
%token	<uninsptr>	FIVOP
%token	<uninsptr>	SIXOP

%token	<unlong>	ISPACE
%token	<unlong>	IBYTE
%token	<unlong>	IWORD
%token	<unlong>	IQUAD
%token	<unlong>	ILONG
%token	<unlong>	IINT
%token	<unlong>	IDATA
%token	<unlong>	IBSS
%token	<unlong>	IGLOBL
%token	<unlong>	ISET
%token	<unlong>	ITEXT
%token	<unlong>	ICOMM
%token	<unlong>	IALIGN
%token	<unlong>	IFLOAT
%token	<unlong>	IDOUBLE
%token	<unlong>	IORG
%token	<unlong>	IASCII
%token	<unlong>	IASCIZ

/* Symbolic Debugger Support Pseudo Operations */

%token	<unlong>	IFILE ILN IDEF IENDEF
%token	<unlong>	IVAL ISCL ITYPE ILINE ITAG ISIZE IDIM

/* Lexical Tokens recognized by yylex */

%token	<unlong>	SP NL SH LITOP REG REGOP AND SQ LP RP MUL PLUS COMMA 
%token	<unlong>	MINUS ALPH DIV DIG COLON SEMI LSH RSH LB RB XOR
%token	<unlong>	IOR TILDE DQ

/* define operator precedences so that yacc can make a clean parser */

%nonassoc	COLON
%left		PLUS MINUS
%left		IOR XOR AND
%left		MUL DIV LSH RSH TILDE

%token	<unlong>	INTNUM
%token	<qs>    	QUADNUM
%token	<unfloat>	FLTNUM
%token	<unsymptr>	IDENT
%token	<unsymptr>	LOCLABDEF
%token	<unsymptr>	LOCLABBREF
%token	<unsymptr>	LOCLABFREF
%token	<unlong>	ERROR
%token	<unstrptr>	STRING

%type 	<unstrptr>      istring

%type	<unrexpr>	iexpr fexpr expr qexpr

%type	<unaddr>	operand subop1 subop2

%type	<unlong>	reg

%%

file:		/* empty */ 
					
	| 	file labinst NL =	{ line++;
					  dot->value = newdot;
            			          generate(0,NEWSTMT,(long)line,NULLSYM);
					}
	| 	file labinst SEMI =	{ dot->value = newdot;
					  generate(0,NEWSTMT,(long)line,NULLSYM);
					}
	|	file error NL =		{ line++;
					  yyerrok;
					  dot->value = newdot;
					}
		;

labinst:	labels instruction
		;

labels:		/* no label is okay */
          /* allow either local labels or regular labels */

	|	labels LOCLABDEF COLON =	{ if (($2->styp&TYPE) != UNDEF)
						yyerror("multiply defined label");
					  $2->value = newdot;
					  $2->styp |= dot->styp;
					  if (opt & ((dot->styp&TYPE)==TXT)) /* sdi */

						deflab($2);
					}

	|	labels IDENT COLON =	{ if (($2->styp&TYPE) != UNDEF)
						yyerror("multiply defined label");
					  $2->value = newdot;
					  $2->styp |= dot->styp;
					  if (opt & ((dot->styp&TYPE)==TXT)) /* sdi */

						deflab($2);
					}
		;

instruction:	/* no instruction is okay */

/* pseudo operations - assembler directives */

/* .space */	|
		ISPACE expr =	{ if (($2.exptype!=TYPL) ||
				      ($2.expval.lngval < 0))
					yyerror("illegal .space size expression");
 				  spfill($2.expval.lngval);
				}

/* .byte */	|
		IBYTE setbyte iexprlist = { if (bitsout) {
					       generate((BITSPBY*bytesout),
						GENVALUE,outword,NULLSYM);
					       outword = 0;
					       bitsout = 0;
					       };
					  }

/* .word */	|
		IWORD setword iexprlist = { if (bitsout) {
					       generate((BITSPBY*bytesout),
						GENVALUE,outword,NULLSYM);
					       outword = 0;
					       bitsout = 0;
					       };
					  }

/* .ascii */	|
		IASCII setascii istringlist =  { }

/* .asciz */	|
		IASCIZ setasciz istringlist =  { }

/* .float */	|
		IFLOAT setfloat fexprlist

/* .double */	|
		IDOUBLE setdouble fexprlist

/* .quad */	|
		IQUAD setquad qexprlist = { quadflag = 0; }

/* .long */	|
		ILONG setlong iexprlist = { if (bitsout) {
					       generate((BITSPBY*bytesout),
						GENVALUE,outword,NULLSYM);
					       outword = 0;
					       bitsout = 0;
					       };
					  }

/* .int */	|
		IINT setint iexprlist = { if (bitsout) {
					       generate((BITSPBY*bytesout),
						GENVALUE,outword,NULLSYM);
					       outword = 0;
					       bitsout = 0;
					       };
					  }

/* .text */	|
		ITEXT = { cgsect(TXT);
			}		 

/* .data */	|
		IDATA = { if (!readonly)
				cgsect(DAT);
			  else
				cgsect(TXT);
			}

/* .bss */	|
		IBSS IDENT COMMA expr COMMA expr = 
			{ long mod;
			  if ($4.exptype != TYPL)
				yyerror(".bss size not absolute");
			  if ($6.exptype != TYPL)
				yyerror(".bss alignment not absolute");
			  if ($4.expval.lngval <= 0)
				yyerror("illegal .bss size");
			  if ($6.expval.lngval <= 0)
				yyerror("illegal .bss alignment");
			  if (($2->styp&TYPE) != UNDEF)
				yyerror("multiply defined .bss label");
			  if (mod = dotbss % $6.expval.lngval)
				dotbss += $6.expval.lngval - mod;
			  $2->value = dotbss;
			  dotbss += $4.expval.lngval;
			  $2->styp |= BSS;
			}
		

/* .globl */	|
		IGLOBL IDENT =	{ $2->styp = $2->styp | EXTERN;
				}

/* .set */	|
		ISET IDENT COMMA expr = {
         				  if ($4.exptype < 0 || $4.exptype > TMAX)
						yyerror("illegal expression in .set");
		 			  else if ($4.exptype&TDIFF)
					  {
						$2->value = (long)( ((symbol *)($4.expval.lngval)) ->value) - (long)($4.symptr->value);
						$2->styp |= ABS;
					  }
   					  else if ($4.exptype == TSYM && !($4.symptr->styp == EXTERN) &&
 						   !($4.symptr->styp == UNDEF))
					  {
						$2->value = (long)($4.symptr->value) +
							$4.expval.lngval;
						$2->styp = $4.symptr->styp;
					  }
					  else if ($4.exptype != TYPL)
						yyerror("illegal expression in .set");
					  else if ($2==NULL)
						yyerror("unable to define symbol");
					  else {
						$2->value = $4.expval.lngval;
						$2->styp |= ABS;
						};
					 }

/* .comm */	|
		ICOMM IDENT COMMA expr = { if ($4.exptype!=TYPL)
						yyerror("illegal .comm size expression");
					    else if (($2->styp&TYPE)!=UNDEF)
						yyerror("illegal attempt to redefine symbol");
					    else {
						$2->styp = (EXTERN | UNDEF);
						$2->value = $4.expval.lngval;
						};
					  }

/* .align */	|
		IALIGN expr = { if (($2.exptype!=TYPL) ||
				     ($2.expval.lngval < 1) ||
				     ($2.expval.lngval > 1024))
					yyerror("illegal .align expression");
				  else
				     if ($2.expval.lngval != 1)
				     {
					if (opt)
					{
					  if (dot->styp == TXT)
				     	      	jalign($2.expval.lngval);
					  else 
					      	ckalign($2.expval.lngval);
					}
					else 
					    ckalign($2.expval.lngval);
				     }
			       }

/* .org */	|
		IORG expr = { int incr;
			       if (($2.exptype!=TYPL)&&(!defsym($2)))
					yyerror("illegal .org expression");
			       else if (defsym($2)) {	
					if ($2.symptr->styp!=dot->styp)
						yyerror("incompatible .org expression type");
				        incr = $2.expval.lngval + $2.symptr->value -
						dot->value;
						}
			       else { /* absolute origins */
				        werror("usage of absolute origins");						incr = $2.expval.lngval - dot->value;
				  	};
			       if (incr < 0)
					yyerror("cannot decrement value of '.'");
			       else
					fill(incr);

			     }

/* Symbolic Debugger (SDB) support pseudo operations */

/* .file */	|
		IFILE STRING = { int cfile_len;
				 char buf[20];
				 if (cfile[0] != '\0')
					yyerror("only 1 .file allowed");
				 if (strlen($2) > FILNMLEN)
				 {
				 /*
					sprintf (buf, ".file name exceeds %d characters", FILNMLEN);
					werror(buf);
				 */
					cfile_len = FILNMLEN;
				 }
				 else
				        cfile_len = strlen($2);
				 strncpy(cfile, $2, cfile_len);
				 if (cfile_len == FILNMLEN)
				        strcat (cfile, "\0");
				 generate(0,SETFILE,NULLVAL,NULLSYM);
			       }

/* .ln */	|
		ILN expr = { cline = (short)$2.expval.lngval;
			     generate(0,LINENBR,$2.expval.lngval,$2.symptr);
			   }

		|
		ILN expr COMMA expr = { cline = (short)$2.expval.lngval;
					generate(0,LINENUM,$2.expval.lngval,
								$2.symptr);
					generate(0,LINEVAL,$4.expval.lngval,
								$4.symptr);
				      }

/* .def */	|
		IDEF {defseen = 1;} IDENT = { generate(0,DEFINE,NULLVAL,$3);
					      defseen = 0;
			     }

/* .endef */	|
		IENDEF = { generate(0,ENDEF,NULLVAL,NULLSYM);
		}

/* .val */	|
		IVAL expr = { generate(0,SETVAL,$2.expval.lngval,$2.symptr);
			    }

/* .type */	|
		ITYPE expr = { generate(0,SETTYP,$2.expval.lngval,$2.symptr);
			     }

/* .line */	|
		ILINE expr = { generate(0,SETLNO,$2.expval.lngval,$2.symptr);
			     }

/* .size */	|
		ISIZE expr = { generate(0,SETSIZ,$2.expval.lngval,$2.symptr);
			     }

/* .scl */	|
		ISCL expr = { generate(0,SETSCL,$2.expval.lngval,$2.symptr);
			    }

/* .tag */	|
		ITAG IDENT = { generate(0,SETTAG,NULLVAL,$2);
			     }

/* .dim */	| 
		dotdim /* can be .dim expr [, expr ] */



/* Instructions */

/* INST */	|

		ZEROP 
			= {
if(is_tdiff)
	is_tdiff = 0;
			   generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
			  }

/* INST op1 */	|

		ONEOP operand 
			= {
			    if ($1->tag==EXTIN1 || $1->tag==EXTIN2)
{
if(is_tdiff)
	is_tdiff = 0;
				gensdi($1,$2,NULL,NULL,$1->opermodes);
}
			    else {
is_opcode = 1;
			    	generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
	genoprdff($2,1,$1->tag,$1->opermodes);
}
else
			    	genopr($2,1,$1->tag,$1->opermodes);
is_opcode = 0;
				};
			  }

/* INST op1, op2 */ |

		TWOOP operand COMMA operand
			= {
is_opcode = 1;
			   
			    	generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
	genoprdff($2,1,$1->tag,$1->opermodes);
	genoprdff($4,2,$1->tag,$1->opermodes);
}
else
{
			    	genopr($2,1,$1->tag,$1->opermodes);
			    	genopr($4,2,$1->tag,$1->opermodes);
}
is_opcode = 0;
			  }

/* INST op1, op2, op3 */ |

		THROP operand COMMA operand COMMA operand
 			= {
is_opcode = 1;
			    	generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
	genoprdff($2,1,$1->tag,$1->opermodes);
	genoprdff($4,2,$1->tag,$1->opermodes);
	genoprdff($6,3,$1->tag,$1->opermodes);
}
else
{
	is_tdiff = 0;
	genopr($2,1,$1->tag,$1->opermodes);
	genopr($4,2,$1->tag,$1->opermodes);
	genopr($6,3,$1->tag,$1->opermodes);
}
is_opcode = 0;
			  }

/* INST op1, op2, op3, op4 */ |

		FOUOP operand COMMA operand COMMA operand COMMA operand
			= {
is_opcode = 1;
			    generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
        genoprdff($2,1,$1->tag,$1->opermodes);
        genoprdff($4,2,$1->tag,$1->opermodes);
        genoprdff($6,3,$1->tag,$1->opermodes);
        genoprdff($8,4,$1->tag,$1->opermodes);
}
else
{
			    genopr($2,1,$1->tag,$1->opermodes);
			    genopr($4,2,$1->tag,$1->opermodes);
			    genopr($6,3,$1->tag,$1->opermodes);
			    genopr($8,4,$1->tag,$1->opermodes);
}
is_opcode = 0;
			  }

/* INST op1, op2 op3, op4, op5 */ |

		FIVOP operand COMMA operand COMMA operand COMMA operand COMMA
		      operand
			= {
is_opcode = 1;
			    generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
        genoprdff($2,1,$1->tag,$1->opermodes);
        genoprdff($4,2,$1->tag,$1->opermodes);
        genoprdff($6,3,$1->tag,$1->opermodes);
        genoprdff($8,4,$1->tag,$1->opermodes);
        genoprdff($10,5,$1->tag,$1->opermodes);
}
else
{
			    genopr($2,1,$1->tag,$1->opermodes);
			    genopr($4,2,$1->tag,$1->opermodes);
			    genopr($6,3,$1->tag,$1->opermodes);
			    genopr($8,4,$1->tag,$1->opermodes);
			    genopr($10,5,$1->tag,$1->opermodes);
}
is_opcode = 0;
			  }

/* INST op1, op2, op3, op4, op5, op6 */ |

		SIXOP operand COMMA operand COMMA operand COMMA operand COMMA
		      operand COMMA operand
			= {
is_opcode = 1;
			    generate($1->nbits,GENVALUE,$1->opcode,NULLSYM);
if(is_tdiff)
{
	is_tdiff = 0;
        genopr($2,1,$1->tag,$1->opermodes);
        genopr($4,2,$1->tag,$1->opermodes);
        genopr($6,3,$1->tag,$1->opermodes);
        genopr($8,4,$1->tag,$1->opermodes);
        genopr($10,5,$1->tag,$1->opermodes);
        genopr($12,6,$1->tag,$1->opermodes);
}
else
{
                            genopr($2,1,$1->tag,$1->opermodes);
                            genopr($4,2,$1->tag,$1->opermodes);
                            genopr($6,3,$1->tag,$1->opermodes);
                            genopr($8,4,$1->tag,$1->opermodes);
                            genopr($10,5,$1->tag,$1->opermodes);
                            genopr($12,6,$1->tag,$1->opermodes);
}
is_opcode = 0;
			  }
			;


setbyte:	= { bytesout = 1;
		  };

setquad:	= { bytesout = 8;
		    quadflag = 1;
		  };

setlong:	= { bytesout = 4;
		  };

setint:		= { bytesout = 4;
		  };

setword:	= { bytesout = 2;
		  };

setfloat:	= { bytesout = 4;
		  };

setdouble:	= { bytesout = 8;
		  };

setascii: 	= { asciflag = 0;        /* indicate .ascii */
		  };

setasciz: 	= { asciflag = 1;        /* indicate .asciz */
		  };

/* continuation of .dim pseudo operation definitions */

dotdim:		IDIM expr = { generate(0,SETDIM1,$2.expval.lngval,$2.symptr);
			    }
		|
		dotdim COMMA expr = { generate(0,SETDIM2,$3.expval.lngval,
							$3.symptr);
				    }
		;


	
operand:	/* %r */
		reg = { $$.admode = AREG;
			$$.adreg1 = $1;
		      }

	|	/* ... [%r] */
		subop1 LB reg RB = { $$.admode = $1.admode | AINDX;
		 	if ($3 == 0xE)
				yyerror("SP used as index register");
		 	else
		 	if ($3 == 0xF)
				yyerror("PC used as index register");
		 	else
				     $$.adreg2 = $3;
				   }

	|	subop1
	;


subop1:		/* -(%r) */
		MINUS LP reg RP = { $$.admode = ADECR;
		 	if ($3 != 0xE)
				yyerror("Autodecrement register must be SP");
		 	else
				    $$.adreg1 = $3;
				  }

	|	/* * ... */
		MUL subop2 = { $$ = $2;
		               if ($2.admode==ABASE) {
				  /* *(%r) is really *0(%r) */
				  $$.admode = (ADISP | ASTAR);
				  $$.adexpr.exptype = TYPL;
				  $$.adexpr.expval.lngval = 0L;
			     	  }
			       else
			       	$$.admode = $2.admode | ASTAR;
		             }

	| subop2
	;

subop2:		/* expression */
		expr = { $$.admode = AEXP;
			 $$.adexpr = $1;
			 $$.adreg1 = $$.adreg2 = 0;
		       }

	|	/* (%r) */
		LP reg RP = { $$.admode = ABASE;
			      $$.adreg1 = $2;
			    }

	|	/* expression (%r)  */
		expr LP reg RP = { $$.admode = ADISP;
				   $$.adreg1 = $3;
				   $$.adexpr = $1;
				 }

	|	/* (%r)+ */
		LP reg RP PLUS = { $$.admode = AINCR;
		 	if ($2 != 0xE)
				yyerror("Autoincrement register must be SP");
		 	else
				   $$.adreg1 = $2;
				 }

	|	/* $ expression */
		LITOP expr = { $$.admode = AIMM;
			       $$.adreg1 = $$.adreg2 = 0;
			       $$.adexpr = $2;
			     }
	;



reg:		REG = { $$ = $1;
		      }
	
	|
		REGOP expr = { if ((($2.exptype&TINT)==0) ||
				   ($2.expval.lngval < 0) ||
				   ($2.expval.lngval > 15))
					yyerror("illegal register specification");
			       else
				 $$ = $2.expval.lngval;
			     }
	;


			    
iexprlist:	iexpr
		|	iexprlist COMMA iexpr
		;
			    
istringlist:	istring
		|	istringlist COMMA istring
		;
			    

fexprlist:	fexpr
		|	fexprlist COMMA fexpr
		;



fexpr:		expr = { if (($1.exptype&TFLT)==0)
				yyerror("illegal floating point expression");
			  $$=$1;
			  if (bytesout == 4)
			  generate((BITSPBY*bytesout),GENVALUE,$1.expval.dblval,
					NULLSYM);
			  else if (bytesout == 8) { int *ptr;
			  ptr=(int *)&$1.expval.dblval;
			  generate((BITSPBY*4),GENVALUE,*ptr,NULLSYM);
			  ptr++;
			  generate((BITSPBY*4),GENVALUE,*ptr,NULLSYM); }
			  else aerror("illegal floating point size");
		       }
		;

qexprlist:	qexpr
		|	qexprlist COMMA qexpr
		;



qexpr:		expr = {  int *ptr;
                          if (($1.exptype&TFLT)!=0)
			        yyerror("illegal quad expression");
			  ptr=(int *)$1.expval.qs.quadval;
			  generate((BITSPBY*4),GENVALUE,*ptr,NULLSYM);
			  ptr++;
			  generate((BITSPBY*4),GENVALUE,*ptr,NULLSYM);
		       }
		;

istring:        STRING = { strcpy (asciifile, $1);
        { int i = 0; char c;        /* buffer pointer and temporary character */
           while (i < strlen(asciifile))   /* step through each character in operand */
	      switch (asciifile[i]) /* check the next character */
	      {
	         case '\\':         /* backslash -- special character */
	         {
	            if (++i<strlen(asciifile))  /* if backslash not last character */
	            {
	               switch (asciifile[i])  /* check char after backslash */
	               {
	                  case 'b': c = '\b';  /* backspace */
	                     break;
	                  case 't': c = '\t';  /* horizontal tab */
	                     break;
	                  case 'n': c = '\n';  /* linefeed */
	                     break;
	                  case 'f': c = '\f';  /* form feed */
	                     break;
	                  case 'r': c = '\r';  /* carriage return */
	                     break;
   /*	not obtainable    case '"': c = '\"';     double quote           */
   /*  	because of parser    break;                                      */
	                  case '\'': c = '\''; /* single quote */
	                     break;
	                  case '\\': c = '\\'; /* backslash */
	                     break;
	                  default:    /* a slash with no special characters   */
	                     if (isoctdigit(asciifile[i]))
	                     {        /* if octal number follows */
	                        unsigned char value = 0; int j = 1;
	                        while ((j++ <= 3) && (i<strlen(asciifile))
	                              && isoctdigit(asciifile[i]))
	                        {
	                           value *= 8;    /* accumulate the number    */
	                           value += (asciifile[i++] - '0');
	                        }
	                        if (value == 0)   /*  \000 ~ '\0'  */
	                           c = '\0';
	                        else              /* return the octal number  */
	                           c = value ;
	                        --i;              /* point back one character */
	                     }
	                     else   /* No conversion needed on the slash     */
	                     {      /* Output the slash and proceed normally */
		                generate(BITSPBY,GENVALUE,'\\',NULLSYM);
		                c = asciifile[i];  /* back up one character */
		             }
		             break;
                    }       /*  output the escape sequence */
		    generate(BITSPBY,GENVALUE,c,NULLSYM);
		    i++;    /* proceed to the next character */
		    break;
	         default:   /* normal character -- output it */
		    generate(BITSPBY,GENVALUE,asciifile[i++],NULLSYM);
		 }
		 else    /* output terminating slash */
		    generate(BITSPBY,GENVALUE,'\\',NULLSYM);
	      }
	   }
        }
	if (asciflag) generate(BITSPBY,GENVALUE,'\0',NULLSYM);
     }
		;

iexpr:		expr = { if (($1.exptype&TFLT)!=0)
				yyerror("illegal fixed point expression");
			 if (bitsout) {
				generate((BITSPBY*bytesout),
				 	 GENVALUE,outword,NULLSYM);
				outword = 0;
				bitsout = 0;
				};
			  $$=$1;
			  if ($1.exptype&TDIFF) /* process the special case of
					    (sym - sym) type expression */
			  	generate((BITSPBY*bytesout),GENABSSYM,
					$1.expval.lngval,$1.symptr);
			  else /* generate a relocation entry if you use a
				  symbol in the expression */
			  	generate((BITSPBY*bytesout),GENRELOC,
					$1.expval.lngval,$1.symptr);
		       }

	| /* bit field expression nn:expr */
		expr COLON expr = { if ($1.exptype != TYPL)
					yyerror("illegal bit field expression");
			 	    if (($3.exptype&TINT)==0)
					yyerror("cannot relocate bit field expresion");	
				    if ($1.expval.lngval > (BITSPBY*bytesout))
					yyerror("bit field crosses data boundary");
				    if (($1.expval.lngval+bitsout) > (BITSPBY*bytesout)) {
					  generate(BITSPBY*bytesout,GENVALUE,
						   outword,NULLSYM);
					  bitsout = 0;
					  outword = 0;
					};
				    $3.expval.lngval <<= bitsout;
				    outword |= $3.expval.lngval;
				    bitsout += $1.expval.lngval;
				  }

		;


expr:
		IDENT = { if (($1->styp&TYPE)!=ABS) {
				$$.exptype = TSYM;
			  	$$.symptr = $1;
			  	$$.expval.lngval = 0L;
			  } else { /* an abs symbol is just like an intnum */
				$$.exptype = TYPL;
				$$.symptr = NULL;
				$$.expval.lngval = $1->value;
			  };
			}

	|
		LOCLABBREF = {
				$$.exptype = TSYM;
			  	$$.symptr = $1;
			  	$$.expval.lngval = 0L;
		             }

	|
		LOCLABFREF = {
				$$.exptype = TSYM;
			  	$$.symptr = $1;
			  	$$.expval.lngval = 0L;
			     }

	|
		INTNUM = { $$.exptype = TYPL;
			   $$.expval.lngval = $1;
			 }
	|
		QUADNUM = { $$.exptype = TYPQ;
			   $$.expval.qs.quadval[0] = $1.unquad[0];
			   $$.expval.qs.quadval[1] = $1.unquad[1];
			 }

	|	FLTNUM = { $$.exptype = TYPD;
			   $$.expval.dblval = $1;
			 }

	|	LP expr RP = { $$=$2;
			     }

	|	expr PLUS expr = { $$ = operate(PLUS,$1,$3);
				 }

	|	expr MINUS expr = { $$ = operate(MINUS,$1,$3);
				    if ($$.exptype & TDIFF)
					if (is_opcode)
					{
					   is_tdiff = 1;
					   is_opcode = 0;
					}
				  }

	|
		expr TILDE expr = { $$ = operate(TILDE,$1,$3);
				  }

	|
		expr IOR expr = { $$ = operate(IOR,$1,$3);
				}

	|
		expr XOR expr = { $$ = operate(XOR,$1,$3);
				}

	| 
		expr AND expr = { $$ = operate(AND,$1,$3);
				}

	|
		expr LSH expr = { $$ = operate(LSH,$1,$3);
				}

	|
		expr RSH expr = { $$ = operate(RSH,$1,$3);
				}

	|
		expr MUL expr = { $$ = operate(MUL,$1,$3);
				}

	|
		expr DIV expr = { $$ = operate(DIV,$1,$3);
				}

	|
		MINUS expr = { $$ = operate(-MINUS,$2,NULL);
			     }

	|
		TILDE expr = { $$ = operate(-TILDE,$2,NULL);
			     }

		;



%%

int curr_buff_off = -2048;     /* source file offset of current input buffer */
int new_line_off = 0;          /* source file offset of next line            */

rexpr operate(op,left,right)

int op;
rexpr left, right;

	{ rexpr result;
	  
	  if (op < 0) { /* unary operator  - or ~ */
		  op = - op;
		  switch (op) {

			case MINUS: switch (left.exptype) {

					case TSYM: result.exptype = TSYM;
						   /* a negative symbol ptr
						      indicates symbol subtract
						      to code gen routines */
						   result.symptr =
						   (symbol*)(-(int)left.symptr);
						   result.expval.lngval =
							left.expval.lngval;
						   return(result);
						   
					case TYPL: result.exptype = TYPL;
						   result.symptr = NULL;
						   result.expval.lngval =
						   - left.expval.lngval;
						   return(result);

					case TYPD: result.exptype = TYPD;
						   result.symptr = NULL;
						   result.expval.dblval =
						   - left.expval.dblval;
						   return(result);

			    		case TYPQ: {
        					   union {unsigned int i[2]; unsigned short w[4];} t;
   						   unsigned int sum, carry, value;
   						   int j;

	             		 	   	   result.exptype = TYPQ;
		   			   	   result.symptr = NULL;
   						   t.i[0] = left.expval.qs.quadval[0];
   						   t.i[1] = left.expval.qs.quadval[1];

				   		   /* negate the operand */
   						   carry = 0;
   						   value = 1;
      						   for (j = 3; j >= 0; --j)
      						   {
    						      sum = (~t.w[j] & 0xffff) + carry + value;
                                                      value = 0;
                                                      carry = sum >> 16;
 					      	      t.w[j] = sum - (carry << 16);
      						   }

            					   result.expval.qs.quadval[0] = t.i[0];
            					   result.expval.qs.quadval[1] = t.i[1];
		   				   return(result);
				       }

					default: yyerror("illegal expression operand");
				    };

			case TILDE: switch (left.exptype) {

					case TYPL: result.exptype = TYPL;
						   result.symptr = NULL;
						   result.expval.lngval =
						   ~ left.expval.lngval;
						   return(result);

					case TYPQ: result.exptype = TYPQ;
						   result.symptr = NULL;

						   result.expval.qs.quadval[0] = ~ left.expval.qs.quadval[0];
						   result.expval.qs.quadval[1] = ~ left.expval.qs.quadval[1];
						   return(result);

					default: yyerror("illegal expression operand");
				    };

			default: yyerror("illegal expression");
			
			};
	   	}
	  else  /* binary operator */
		{ switch (op)  {

			case DIV: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval/right.expval.lngval;
					return(result);

				} else

				if (left.exptype==TYPQ&&right.exptype==TYPQ) {
					union {unsigned int i[2]; unsigned short w[4];} t;
          				union {unsigned int i[2]; unsigned short w[4];} u;
          				union {unsigned int i[2]; unsigned short w[4];} divr;
          				union {unsigned int i[2]; unsigned short w[4];} quot;
          				union {unsigned int i[2]; unsigned short w[4];} work;
   					unsigned int sum, carry, value, temp;
   					int i, j, sign;
	   				int num_divisor_digits;
	  				int num_iterations;
	   				int quotient_bit;

	             			result.exptype = TYPQ;
		   			result.symptr = NULL;
   					t.i[0] = left.expval.qs.quadval[0];
   					t.i[1] = left.expval.qs.quadval[1];
   	   				u.i[0] = right.expval.qs.quadval[0];
   	   				u.i[1] = right.expval.qs.quadval[1];

	     				/* make dividend positive */
					sign = 0;
	     				if ((t.i[0] & 0x80000000) != 0)
	     				{
	     				        sign++;
	     					carry = 0;
	     					value = 1;
      					   	for (j = 3; j >= 0; --j)
      						{
    						      sum = (~t.w[j] & 0xffff) + carry + value;
                                                      value = 0;
                                                      carry = sum >> 16;
 						      t.w[j] = sum - (carry << 16);
      						}
	     				}

	     				/* make divisor positive */
	     				if ((u.i[0] & 0x80000000) != 0)
	     				{
	     				        sign++;
	     					carry = 0;
	     					value = 1;
      						for (j = 3; j >= 0; --j)
      						{
    				      		      sum = (~u.w[j] & 0xffff) + carry + value;
                                                      value = 0;
                                                      carry = sum >> 16;
 						      u.w[j] = sum - (carry << 16);
      						}
	     				}

	   				/* find first non-zero bit in divisor */
           				if (u.i[0] != 0)
           				{
              					j = 0;
              					temp = u.i[0];
             			 		while ((temp & 0x80000000) == 0)
              					{
              						temp <<= 1;
	      						++j;
	      					}
	   				}
	   				else if (u.i[1] != 0)
	   				{
              					j = 32;
              					temp = u.i[1];
              					while ((temp & 0x80000000) == 0)
              					{
              						temp <<= 1;
	      						++j;
	      					}
	   				}
	   				else
	        				yyerror("division by zero");

	   				num_divisor_digits = 64 - j;
	   				num_iterations = 65 - num_divisor_digits;
	   				quotient_bit = num_divisor_digits;

	   				/* negate the divisor -- for subtracting from dividend */
   					carry = 0;
   					value = 1;
      					for (j = 3; j >= 0; --j)
      					{
    					      	sum = (~u.w[j] & 0xffff) + carry + value;
                                              	value = 0;
                                              	carry = sum >> 16;
 					      	divr.w[j] = sum - (carry << 16);
      					}

					/* left-justify the negated divisor */
      			       		for (j=num_divisor_digits; j<=62; ++j)
      			       		{
	   			     		divr.i[0] <<= 1;
	   		  	     		if ((divr.i[1] & 0x80000000) == 0x80000000)
	      						divr.i[0] |= 1;
	   			     		divr.i[1] <<= 1;
	   		       		}

	                                /* initialize the quotient */
      			       		quot.i[0] = 0;
      			       		quot.i[1] = 0;

	       /* Iterate, subtracting the negating divisor from the
	          shifting dividend, thus setting bits of the quotient. */
           				for (i = 1; i <= num_iterations; ++i)
           				{
           					/* get working dividend */
	      					work.i[0] = t.i[0];
	      					work.i[1] = t.i[1];

                  				/* add negated, left-justified divisor to it */
						carry = 0;
						for (j = 3; j >= 0; --j)
      	   					{
    	      						sum = work.w[j] + divr.w[j] + carry;
              						carry = sum >> 16;
 	      						work.w[j] = sum - (carry << 16);
      	   					}

      	   					/* carry out => quotient bit should be set */
	     					if (carry > 0)
	     					{
	        					t.i[0] = work.i[0];
	        					t.i[1] = work.i[1];

	        					temp = 0x80000000;
	        					if (quotient_bit < 32)
	        					{
	           						if (quotient_bit > 0)
	           						{
	           							for (j = 1; j<=quotient_bit; ++j)
	              								temp >>= 1;
	           						}
		   						quot.i[0] |= temp;
	        					}
	        					else
	        					{
	           						if (quotient_bit > 32)
	           						{
	           							for (j=33; j<=quotient_bit; ++j)
	              								temp >>=1;
	           						}
		   						quot.i[1] |= temp;
	        					}
	     					}
	     					quotient_bit++;

          					/* shift the dividend left one bit */
	   			     		t.i[0] <<= 1;
	   		  	     		if ((t.i[1] & 0x80000000) == 0x80000000)
	      						t.i[0] |= 1;
	   			     		t.i[1] <<= 1;
           				}

					/* make quotient negative, if necessary */
	     				if (sign == 1)
	     				{
	     					carry = 0;
	     					value = 1;
      						for (j = 3; j >= 0; --j)
      						{
    				      		      sum = (~quot.w[j] & 0xffff) + carry + value;
                                                      value = 0;
                                                      carry = sum >> 16;
 						      quot.w[j] = sum - (carry << 16);
      						}
	   				}

           				result.expval.qs.quadval[0] = quot.i[0];
           				result.expval.qs.quadval[1] = quot.i[1];
	  				return(result);

	   			} else


				if (left.exptype==TYPD&&right.exptype==TYPD) {
					result.exptype = TYPD;
					result.symptr = NULL;
					result.expval.dblval =
					left.expval.dblval/right.expval.dblval;
					return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case MUL: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval*right.expval.lngval;
					return(result);

				} else

				if (left.exptype==TYPQ&&right.exptype==TYPQ) {
        		   	   union {unsigned int i[2]; unsigned short w[4];} t;
          			   union {unsigned int i[2]; unsigned short w[4];} u;
          			   union {unsigned int i; unsigned short w[2];} p1;
          			   union {unsigned int i; unsigned short w[2];} p2;
          			   union {unsigned int i; unsigned short w[2];} p3;
          			   union {unsigned int i; unsigned short w[2];} p4;
          			   union {unsigned int i; unsigned short w[2];} p5;
          			   union {unsigned int i[2]; unsigned short w[4];} s;
   				   unsigned int sum, carry, value;
   				   int j;

	             		   result.exptype = TYPQ;
		   		   result.symptr = NULL;
				   /* initialize multiplier and multiplicand */
   				   t.i[0] = left.expval.qs.quadval[0];
   				   t.i[1] = left.expval.qs.quadval[1];
   	   			   u.i[0] = right.expval.qs.quadval[0];
   	   			   u.i[1] = right.expval.qs.quadval[1];

	     /* compute partial products and keep the last four words */
	   			   p1.i = t.w[3] * u.w[3];
	   			   s.w[3] = p1.w[1];

	   			   p2.i = t.w[2] * u.w[3];
	   			   p3.i = t.w[3] * u.w[2];
    	   			   sum = p1.w[0] + p2.w[1] + p3.w[1];
	   			   s.w[2] = sum & 0xffff;
           			   carry = sum >> 16;
    	   			   p1.i = p2.w[0] + p3.w[0] + carry;

	   			   p2.i = t.w[1] * u.w[3];
	   			   p3.i = t.w[2] * u.w[2];
	   			   p4.i = t.w[3] * u.w[1];
    	   			   sum = p1.w[1] + p2.w[1] + p3.w[1] + p4.w[1];
	   			   s.w[1] = sum & 0xffff;
           			   carry = sum >> 16;
    	   			   p1.i = p2.w[0] + p3.w[0] + p4.w[0] + carry;

	   			   p2.i = t.w[0] * u.w[3];
	   			   p3.i = t.w[1] * u.w[2];
	   			   p4.i = t.w[2] * u.w[1];
	   			   p5.i = t.w[3] * u.w[0];
    	   			   sum = p1.w[1] + p2.w[1] + p3.w[1] + p4.w[1] + p5.w[1];
    	   			   s.w[0] = sum & 0xffff;

           			   result.expval.qs.quadval[0] = s.i[0];
           			   result.expval.qs.quadval[1] = s.i[1];
	   			   return(result);

	   			} else

				if (left.exptype==TYPD&&right.exptype==TYPD) {
					result.exptype = TYPD;
					result.symptr = NULL;
					result.expval.dblval =
					left.expval.dblval*right.expval.dblval;
					return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case RSH: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval>>right.expval.lngval;
					return(result);

				} else

	        		   if (left.exptype==TYPQ&&right.exptype==TYPQ) {

        		   	   	union {unsigned int i[2];} t;
          			   	union {int i[2];} u;
   				   	t.i[0] = left.expval.qs.quadval[0];
   				   	t.i[1] = left.expval.qs.quadval[1];
   	   			   	u.i[0] = right.expval.qs.quadval[0];
   	   			   	u.i[1] = right.expval.qs.quadval[1];
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;

	       				if (u.i[1] > 64 || u.i[0] != 0)
	       				{
          					result.expval.qs.quadval[0] = 0;
          					result.expval.qs.quadval[1] = 0;
	   					return(result);
	       				}
					while (u.i[1]-- > 0)
					{
	   					t.i[1] >>= 1;
	   					if ((t.i[0] & 1) == 1)
	      						t.i[1] |= 0x80000000;
	   					t.i[0] >>= 1;
					}

          				result.expval.qs.quadval[0] = t.i[0];
          				result.expval.qs.quadval[1] = t.i[1];
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case LSH: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval<<right.expval.lngval;
					return(result);

				} else

	        		   if (left.exptype==TYPQ&&right.exptype==TYPQ) {

        		   	   	union {unsigned int i[2];} t;
          			   	union {int i[2];} u;
   				   	t.i[0] = left.expval.qs.quadval[0];
   				   	t.i[1] = left.expval.qs.quadval[1];
   	   			   	u.i[0] = right.expval.qs.quadval[0];
   	   			   	u.i[1] = right.expval.qs.quadval[1];
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;

	       				if (u.i[1] > 64 || u.i[0] != 0)
	       				{
          					result.expval.qs.quadval[0] = 0;
          					result.expval.qs.quadval[1] = 0;
	   					return(result);
	       				}
					while (u.i[1]-- > 0)
					{
	   					t.i[0] <<= 1;
	   					if ((t.i[1] & 0x80000000) == 0x80000000)
	      						t.i[0] |= 1;
	   					t.i[1] <<= 1;
					}

          				result.expval.qs.quadval[0] = t.i[0];
          				result.expval.qs.quadval[1] = t.i[1];
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case AND: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval&right.expval.lngval;
					return(result);

				} else

	        		   if (left.exptype==TYPQ&&right.exptype==TYPQ) {
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;
          				result.expval.qs.quadval[0] = left.expval.qs.quadval[0]&right.expval.qs.quadval[0];
          				result.expval.qs.quadval[1] = left.expval.qs.quadval[1]&right.expval.qs.quadval[1];
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case XOR: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval^right.expval.lngval;
					return(result);

				} else

	        		   if (left.exptype==TYPQ&&right.exptype==TYPQ) {
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;
          				result.expval.qs.quadval[0] = left.expval.qs.quadval[0]^right.expval.qs.quadval[0];
          				result.expval.qs.quadval[1] = left.expval.qs.quadval[1]^right.expval.qs.quadval[1];
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case IOR: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval|right.expval.lngval;
					return(result);

				} else

	        		   if (left.exptype==TYPQ&&right.exptype==TYPQ) {
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;
          				result.expval.qs.quadval[0] = left.expval.qs.quadval[0]|right.expval.qs.quadval[0];
          				result.expval.qs.quadval[1] = left.expval.qs.quadval[1]|right.expval.qs.quadval[1];
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case TILDE: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval|(~right.expval.lngval);
					return(result);

				} else

	         	           if (left.exptype==TYPQ&&right.exptype==TYPQ) {
	             		   	result.exptype = TYPQ;
		   		   	result.symptr = NULL;
					result.expval.qs.quadval[0] = left.expval.qs.quadval[0]|(~right.expval.qs.quadval[0]);
					result.expval.qs.quadval[1] = left.expval.qs.quadval[1]|(~right.expval.qs.quadval[1]);
	   			   	return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case MINUS: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval-right.expval.lngval;
					return(result);

				} else

					if (left.exptype==TYPQ&&right.exptype==TYPQ) {
        					   union {unsigned int i[2]; unsigned short w[4];} t;
          					   union {unsigned int i[2]; unsigned short w[4];} u;
   						   unsigned int sum, carry, value;
   						   int j;

	             			 	   result.exptype = TYPQ;
		   				   result.symptr = NULL;
   						   t.i[0] = left.expval.qs.quadval[0];
   						   t.i[1] = left.expval.qs.quadval[1];
   	   					   u.i[0] = right.expval.qs.quadval[0];
   	   					   u.i[1] = right.expval.qs.quadval[1];

						   /* negate the subtrahend */
   						   carry = 0;
   						   value = 1;
      						   for (j = 3; j >= 0; --j)
      						   {
    						      sum = (~u.w[j] & 0xffff) + carry + value;
                                                      value = 0;
                                                      carry = sum >> 16;
 						      u.w[j] = sum - (carry << 16);
      						   }

						   /* add the minuend and the negated subtrahend */
   	   					   carry = 0;
      	   					   for (j = 3; j >= 0; --j)
      	   					   {
    	      					      sum = t.w[j] + u.w[j] + carry;
              				    	      carry = sum >> 16;
 	      				  	      t.w[j] = sum - (carry << 16);
      	   					   }

           					   result.expval.qs.quadval[0] = t.i[0];
           					   result.expval.qs.quadval[1] = t.i[1];
	   					   return(result);

	   			} else

				if (left.exptype==TYPD&&right.exptype==TYPD) {
					result.exptype = TYPD;
					result.symptr = NULL;
					result.expval.dblval =
					left.expval.dblval-right.expval.dblval;
					return(result);

				} else
				
				if ((left.exptype == TSYM) &&
				    (right.exptype == TSYM)) {
					result.exptype = (TSYM | TDIFF);
					result.symptr = right.symptr;
					result.expval.lngval=(long)left.symptr;
					return(result);

				} else
				
				if ((left.exptype == TYPL) &&
				    (right.exptype == TSYM)) {
					result.exptype = TSYM;
					/* negative symbol ptr indicates symbol
					   subtract to code gen routines */
					result.symptr = 
						(symbol*)(-(int)right.symptr);
					result.expval.lngval = 
					left.expval.lngval-right.expval.lngval;
					return(result);

				} else

				if ((left.exptype == TSYM) &&
				    (right.exptype == TYPL)) {
					result.exptype = TSYM;
					result.symptr = left.symptr;
					result.expval.lngval = 
					left.expval.lngval-right.expval.lngval;
					return(result);

				} else

					yyerror("illegal expression operands");
				break;

			case PLUS: if (left.exptype==TYPL&&right.exptype==TYPL) {
					result.exptype = TYPL;
					result.symptr = NULL;
					result.expval.lngval =
					left.expval.lngval+right.expval.lngval;
					return(result);

				} else

				if (left.exptype==TYPD&&right.exptype==TYPD) {
					result.exptype = TYPD;
					result.symptr = NULL;
					result.expval.dblval =
					left.expval.dblval+right.expval.dblval;
					return(result);

				} else

           				if (left.exptype==TYPQ&&right.exptype==TYPQ) {
          					union {unsigned int i[2]; unsigned short w[4];} t;
          					union {unsigned int i[2]; unsigned short w[4];} u;
   	  					unsigned int sum, carry;
   	  					int j;

	             			 	result.exptype = TYPQ;
		   				result.symptr = NULL;
   	   					t.i[0] = left.expval.qs.quadval[0];
   	   					t.i[1] = left.expval.qs.quadval[1];
   	   					u.i[0] = right.expval.qs.quadval[0];
   	   					u.i[1] = right.expval.qs.quadval[1];

   	   					/* add the addends */
   	   					carry = 0;
      	   					for (j = 3; j >= 0; --j)
      	   					{
    	      						sum = t.w[j] + u.w[j] + carry;
              						carry = sum >> 16;
 	      						t.w[j] = sum - (carry << 16);
      	   					}

           					result.expval.qs.quadval[0] = t.i[0];
           					result.expval.qs.quadval[1] = t.i[1];
	   					return(result);

	   			} else

				if (((left.exptype==TYPL)&&(right.exptype==TSYM)) ||
				    ((left.exptype==TSYM)&&(right.exptype==TYPL))) {
					result.exptype = TSYM;
					result.expval.lngval = 
				        left.expval.lngval+right.expval.lngval;
					if (left.exptype==TSYM) 
						result.symptr = left.symptr;
					else 
						result.symptr = right.symptr;
					return(result);

				} else

					yyerror("illegal expression operands");
				break;

			      default: yyerror("illegal expression");

			      };
		  
	  	};

	}




#if FLEXNAMES
char	yytext[BUFSIZ];
#else
char	yytext[TXTCHARS+2];
#endif

short type[] = {
	EOF,
	SP,	0,	0,	0,	0,	0,	0,	0,
	0,	SP,	NL,	0,	0,	SP,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	0,	0,	0,	0,	0,	0,	0,	0,
	SP,	0,	DQ,	SH,	LITOP,	REG,	AND,	SQ,
	LP,	RP,	MUL,	PLUS,	COMMA,	MINUS,	ALPH,	DIV,
	DIG,	DIG,	DIG,	DIG,	DIG,	DIG,	DIG,	DIG,
	DIG,	DIG,	COLON,	SEMI,	LSH,	0,	RSH,	0,
	0,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	LB,	0,	RB,	XOR,	ALPH,
	0,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,	ALPH,
	ALPH,	ALPH,	ALPH,	0,	IOR,	0,	TILDE,	0,
};

#define INBUFSIZ	2048

static char	inbuf[INBUFSIZ + 2];
static char	*inptr = &inbuf[INBUFSIZ];

#define mygetc(x)	{ if (++inptr > &inbuf[INBUFSIZ]) { \
			  inbuf[fread(&inbuf[1],1,INBUFSIZ,fdin)+1]=EOF; \
			  curr_buff_off += 2048; \
			  new_line_off = 0; \
			  inptr = &inbuf[1]; \
			  }; \
			  x = *inptr; }

#define unmygetc(x)	*inptr-- = x


yylex()
{
        extern int curr_buff_off; /* source file offset of current buffer     */
        extern int new_line_off;  /* source file offset of current (new) line */
	long value, oldvalue;
	short base;
	register short c;
	register short ctype;
	char fltchr[64];
	char *cptr;
	upsymins	id;
	static BYTE instseen = 0; /* instruction seen on line */
        static short ccomment = 0;   /* indicate no c-style comment seen */

	double atof();

	mygetc(c);          /* Get the next lexical character.               */
	if (ccomment)       /* If we are in the middle of a c-style comment, */
  	   for (;;)         /* keep searching for some terminator.           */
  	   {
	      if (c == '\n')   /* If end of line  */
	      {
	         instseen = 0;         /* flag an "instruction" seen */
       	         new_line_off = inptr - &inbuf[1] + curr_buff_off;
	         return(NL);           /* determine offset of next line */
              }
              else if (c == '*')  /* If asterisk  */
              {
                 mygetc(c);       /* look at next character. */
                 if (c == '/')    /* If slash, comment is terminated. */
                 {
                    ccomment = 0;  /* Turn off the c-style comment flag */
                    mygetc(c);     /* and get the next character.       */
                    break;         /* Join normal lexical analysis.     */
                 }
              }
              else if (type[c+1] == EOF)  /* If end of source file */
              {
                 yyerror ("final comment not terminated");
                 return(0);    /* warn of unterminated comment  */
              }
	      mygetc(c);    /* Still in comment -- get next character */
           }

RELEX:	while (type[c+1] == SP) mygetc(c);  /* Skip over blanks */

	ctype = type[c+1];
	value = c;

	switch (ctype) {

			case REG: /* %n indicating register n */
				mygetc(value);
				if (value >= '0' && value <= '9') {
					yylval.unlong = value - '0';
					if (value=='1') {
						mygetc(value);
						if (value >= '0' &&
						    value <= '5') 
							yylval.unlong =
								10+value-'0';
						else
							unmygetc(value);
						};
					return(REG);
					}
				else	{
					unmygetc(value);
					return(REGOP);
					};
				
			case ALPH: /* some kind of alpha-numeric string */
				cptr = yytext;
				do {
#if FLEXNAMES
					if (cptr < &yytext[BUFSIZ-1])
#else
					if (cptr <= &yytext[TXTCHARS])
#endif
						*cptr++ = value;
					mygetc(value);
					ctype = type[value+1];
	    			} while (ctype == DIG || ctype == ALPH || ctype == LITOP);
#if !FLEXNAMES
				yytext[8] =
#endif
				*cptr = '\0';
				/* look to see if it is a label */
				while (value == ' ' || value == '\t')
					mygetc(value);
				unmygetc(value); /* throw it back */
				if (value == ':') { /* it's a label */
					id = *lookup(yytext,INSTALL,USRNAME);
					yylval.unsymptr = id.stp;
					return(IDENT);
					};
				/* look to see if it is a special symbol
				   or an instruction mnemonic */
				if (!instseen) { /* look for an instruction */
					id = *lookup(yytext,N_INSTALL,MNEMON);
					if (id.stp==NULL || id.stp->tag==0) {
						yyerror("invalid instruction mnemonic");
						return(ERROR);
						};
					/* it is some kind of instruction */
					yylval.uninsptr = id.itp;
					instseen = 1;
					return(id.itp->val+256);
					}
				else { /* some kind of operand ?? */
				       /* is it a reserved symbol? */
					if (!defseen) {
					  id = *lookup(yytext,N_INSTALL,MNEMON);
					  if (id.stp!=NULL&&id.stp->tag==SPSYM){
						yylval.unlong = id.itp->opcode;
						return(REG);
						};
					};
					/* it is just a user's symbol */
					id = *lookup(yytext,INSTALL,USRNAME);
					yylval.unsymptr = id.stp;
					return(IDENT);
					};


			case DIG: /* some kind of number */
				yylval.unfloat = 0.0;
				yylval.unlong = value-'0';
				if (value=='0') {
					mygetc(value);
					if (value=='x' || value=='X') {
						base = 16;
					} else if (value=='d' || value=='D' ||
						    value=='f' || value=='F') {
						cptr = &fltchr[0];
						mygetc(value);
						while (cptr<&fltchr[63]&&
						    value =='.'
						 || value=='e' || value=='d' ||
						    value=='E' || value=='D'
						 || value=='-' || value=='+' ||
							 (type+1)[value]==DIG) {
							*cptr++ = value;
							mygetc(value);
						};
						unmygetc(value);
						*cptr++ = '\0';
						yylval.unfloat = atof(fltchr);
						return(FLTNUM);
					} else if (value == ':') {  /* local */
					      oldvalue = '0';       /* label */
			 	              goto loclabdef;       /* def.  */
		                        } else if (value == 'b') {  /* local */
			 	              oldvalue = '0';       /* label */
				              goto loclabbref;      /* ref.  */
/*    Unfortunately, 0f		                        } else if (value == 'f') {
      is a valid floating        		              oldvalue = '0';
      point constant			                      goto loclabfref;*/
					} else {
						unmygetc(value);
						base = 8;
					}
				} else
					base = 10;
			 	oldvalue = value;
				mygetc(value);
				if (value == ':') /* if local label definiton */
				{
                                   short regnum;
loclabdef:
                                   regnum = oldvalue - '0';  /* determine label */
                                   local_label_def_flag[regnum] = 1;
                                   if (local_label_fref_flag[regnum])  /* if already forward referenced */
                                   {
                                      local_label_fref_flag[regnum] = 0;   /* use the number determined at forward reference */
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, local_label_number[regnum]);
                                      id = *lookup(yytext,N_INSTALL,USRNAME);
                                   }
                                   else  /* if not forward referenced since last occurance */
                                   {
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, ++local_label_number[regnum]);
                                      id = *lookup(yytext,INSTALL,USRNAME);
                                   }
				   yylval.unsymptr = id.stp;
				   unmygetc(value);
				   return(LOCLABDEF);
			        }
		                else if ((value == 'b') && (base != 16))
		                {
                                   short regnum;
loclabbref:
                                   regnum = oldvalue - '0';  /* determine label */
                                   if (!local_label_def_flag[regnum]) /* if backward ref. to undefined label */
                                   {
                                      yyerror ("backward reference to undefined local label");
                                      return(ERROR);
                                   }
                                   if (local_label_fref_flag[regnum]) /* if backward ref following a forward ref */
                                   {
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, local_label_number[regnum]-1);
                                      id = *lookup(yytext,N_INSTALL,USRNAME);
                                   }
                                   else /* backward ref before any forward ref */
                                   {
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, local_label_number[regnum]);
                                      id = *lookup(yytext,N_INSTALL,USRNAME);
                                   }
				   yylval.unsymptr = id.stp;
				   return(LOCLABBREF);
				}
		                else if ((value == 'f') && (base != 16))
		                {
                                   short regnum;
loclabfref:
                                   regnum = oldvalue - '0';   /* determine label */
                                   if (local_label_fref_flag[regnum])      /* if forward reference already seen */
                                   {
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, local_label_number[regnum]);
                                      id = *lookup(yytext,N_INSTALL,USRNAME);
                                   }
                                   else  /* if first forward reference */
                                   {
                                      local_label_fref_flag[regnum] = 1;   /* indicate a forward reference */
                                      (void)sprintf(yytext, "L%1d\001%5d\0", regnum, ++local_label_number[regnum]);
                                      id = *lookup(yytext,INSTALL,USRNAME);
                                   }
				   yylval.unsymptr = id.stp;
				   return(LOCLABFREF);
                                }
                                else
                                {
                                   if (quadflag)
                                   {
                                      union {unsigned int i[2]; unsigned short w[4];} t;
                                      int sum, save_sum, carry, i;
                                      t.i[0] = 0;
                                      t.i[1] = yylval.unlong;
                                      while ((type+1)[value]==DIG
                                              || (base==16 &&
                                              (value>='a' && value<='f'||value>='A' && value<='F'))) {
                                          carry = 0;
                                          if (base==8)
                                          {
                                             value -= '0';
                                             for (i = 3; i >= 0; --i)
                                             {
	 					sum = (((unsigned int) t.w[i] << 3) + carry ) + value;
	 					value = 0;
	 					carry = sum >> 16;
	 					t.w[i] = sum - (carry << 16);
      					     }
   					  }
   				          else if (base==10)
   					  {
      					     value -= '0';
      					     for (i = 3; i >= 0; --i)
      					     {
	 					sum = (((unsigned int) t.w[i] * 10) + carry ) + value;
	 					value = 0;
	 					carry = sum >> 16;
	 					t.w[i] = sum - (carry << 16);
      					     }
   					  }
 					  else
   					  {
      				 	     if (value>='a' && value<='f')
					        value -= 'a' - 10 - '0';
      					     else if (value>='A' && value<='F')
						value -= 'A' - 10 - '0';
      					     value -= '0';
      					     for (i = 3; i >= 0; --i)
      					     {
	 					sum = (((unsigned int) t.w[i] << 4) + carry ) + value;
	 					value = 0;
	 					carry = sum >> 16;
	 					t.w[i] = sum - (carry << 16);
      					     }
   					  }



					  mygetc(value);
   				      }

   				      yylval.qs.unquad[0] = t.i[0];
   				      yylval.qs.unquad[1] = t.i[1];
				      unmygetc (value);
		                      return(QUADNUM);

				}
				else
				{

				while ((type+1)[value]==DIG
				    || (base==16 &&
				    (value>='a' && value<='f'||value>='A' && value<='F'))) {
					if (base==8)
						yylval.unlong *= 8;
					else if (base==10)
						yylval.unlong *= 10;
					else
					{
						yylval.unlong <<= 4;
						if (value>='a' && value<='f')
							value -= 'a' - 10 - '0';
						else if (value>='A' && value<='F')
							value -= 'A' - 10 - '0';
					}
					yylval.unlong += value-'0';
					mygetc(value);
				}
				}
				unmygetc (value);
		                return(INTNUM);
                                }
			case SQ: /* '? --> return the integer value of ? */
				mygetc(yylval.unlong);
				if (yylval.unlong == '\n')
					unmygetc('\n'); /* get the NL next time */
				return(INTNUM);

			case DQ: /* this is the beginning of a string "..." */
				yylval.unstrptr = cptr = yytext;
				mygetc(c);
				while ((c != '"') && (c != '\n')) {
					if (cptr <= &yytext[TXTCHARS])
						*cptr++ = c;
					mygetc(c);
					};
				*cptr = '\0';
				if (c == '\n')
					yyerror("string not terminated");
				return(STRING);

			case EOF: /* no more input, that's all folks */
			        return(0);

			case SH: /* # comment to end of line */
				mygetc(c);
				while (c != '\n') 
					mygetc(c);
				instseen = 0;
         	                new_line_off = inptr - &inbuf[1] + curr_buff_off;
				return(NL);

			case NL: /* new line */
				instseen = 0;
       	                        new_line_off = inptr - &inbuf[1] + curr_buff_off;
				return(NL);

			case SEMI: /* logical end of statement */
				instseen = 0;
				return(SEMI);

                        case DIV:
                                mygetc(c);      /* Get the next character     */
                                if (c == '*')   /* slash asterisk => comment  */
                                {
                                   ccomment = 1;  /* Flag c-style comment     */
    	                           for (;;)       /* Look for some terminator */
  	                           {
	                              mygetc(c);  /* Get the next character   */
	                              if (c == '\n')   /* If end of line      */
	                              {
	                                 instseen = 0;
                                         new_line_off = inptr - &inbuf[1] + curr_buff_off;
	                                 return(NL);  /* Point to next line   */
                                      }
                                      else if (c == '*')  /* If asterisk      */
                                      {
                                         mygetc(c);    /* get the next char.  */
                                         if (c == '/')  /* If end of comment  */
                                         {
                                            ccomment = 0;  /* Turn off flag   */
                                            mygetc(c);  /* Get next character */
                                            goto RELEX;  /* Continue normal   */
                                         }               /* lexical analysis  */
                                      }
                                   }
                                }
                                else if (type[c+1] == 0)
                                {
				   yyerror("illegal input character");
				   return(ERROR);
                                }
                                else
                                {
                                   unmygetc (c);
                                   return(ctype);
                                }

			case 0: /* ????? */
				yyerror("illegal input character");
				return(ERROR);

			default: /* if all else fails, return token */
				return(ctype);
			};
}

fill(nbytes)
register long nbytes;
{
	register long fillval;

	if (nbytes == 0)
		return;
	fillval = (dot->styp == TXT)	? ((TXTFILL<<8)|TXTFILL)
					: ((FILL<<8)|FILL);
	while (nbytes >= 2) {
		generate(2*BITSPBY,0,fillval,NULLSYM);
		nbytes -= 2;
	}
	if (nbytes)
		generate(BITSPBY,0,fillval,NULLSYM);
} /* fill */

spfill(nbytes)                  /* for repetition of "space" operand */
register long nbytes;
{
	register long fillval;     /* value to be output */

	if (nbytes == 0)           /* if no space to fill */
		return;
	fillval = (dot->styp == TXT)	? ((TXTFILL<<8)|TXTFILL)
					: ((FILL<<8)|FILL);
	if (nbytes > 4)            /* if more than 4 bytes, use repetition */
	   generate(nbytes*BITSPBY,0,fillval,NULLSYM);
        else                       /* follow normal fill procedures */
        {
	   while (nbytes >= 2) {
		generate(2*BITSPBY,0,fillval,NULLSYM);
		nbytes -= 2;
	   }
	   if (nbytes)
		generate(BITSPBY,0,fillval,NULLSYM);
	}

} /* spfill */


ckalign(size)
long size;
{
	long mod;
	if ((mod = newdot % size) != 0) {
		fill(size - mod);
	}
} /* ckalign */
