/* UNIX HEADER */
#include	<stdio.h>

/* COMMON SGS HEADERS */
#include	"filehdr.h"
#include	"ldfcn.h"


/*  checkflags(filename, &filehead)
 *
 *  Checks and sets object file header flags appropriately.
 *  Checkflags returns SUCCESS or FAILURE.
 *
 *  Checkflags will FAIL (and strip won't strip) in the following cases:
 *  - if there is no symbol table (strip announces the fact if flags
 *    imply that symbols were expected)
 *  - if lflag or xflag is ON and there are no local symbols 
 *  - if a non-executable contains relocation entries and the r, f, s,
 *    and l flags are OFF
 */

int
checkflags(filename, filehead)

char	*filename;
FILHDR	*filehead;

{
    /* STRIP FUNCTION CALLED */
    extern		error( );

    /* EXTERNAL VARIABLES USED */
    extern LDFILE	*ldptr;
    extern int		lflag,
			xflag,
			sflag,
			rflag,
    			fflag,
			norelflag;

    /* LOCAL VARIABLE */
    unsigned short	filflags;

    if (filehead->f_nsyms == 0L) {
	if (lflag || xflag || rflag || sflag || fflag) {
	    error(filename, "no symbol table", 1);
	    onintr();
	}
	return(FAILURE);
    }
    else if (!(lflag || xflag || rflag || sflag || fflag)) {
	filehead->f_nsyms = 0L;
    }

    filflags = filehead->f_flags;

    if (sflag && (filflags & F_LSYMS) != 0) {
	error(filename, "no local symbols", 1);
	return(FAILURE);
    }

    if (norelflag && !(filflags & F_EXEC) && !(filflags & F_RELFLG)) {
	error(filename, "relocation entries; cannot strip", 1);
	return(FAILURE);
    }

#if !UNIX
    if (fflag && !(filflags & F_EXEC)) {
	error( filename, "not executable, cannot strip", 1);
	return( FAILURE );
    }
#endif

    return(SUCCESS);
}
/*
 *	static char ID[] = "@(#) checkflags.c: 1.5 8/18/83";
 */
