/*
 * Copyright (c) 1983 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char sccsid[] = "@(#)look_up.c	5.1 (Berkeley) 6/6/85";
#endif not lint

#include "talk_ctl.h"


/*
 * See if the local daemon has a invitation for us
 */
check_local()
{
	CTL_RESPONSE response;

	/* the rest of msg was set up in get_names */
	msg.ctl_addr = ctl_addr;
	/* must be initiating a talk */
	if (!look_for_invite(&response))
		return (0);
	/*
	 * There was an invitation waiting for us, 
	 * so connect with the other (hopefully waiting) party 
	 */
	current_state = "Waiting to connect with caller";
again:
	if (connect(sockt, &response.addr, sizeof(response.addr)) != -1)
		return (1);
	if (errno == EINTR)
		goto again;
	if (errno == ECONNREFUSED) {
		/*
		 * The caller gave up, but his invitation somehow
		 * was not cleared. Clear it and initiate an 
		 * invitation. (We know there are no newer invitations,
		 * the talkd works LIFO.)
		 */
		ctl_transact(his_machine_addr, msg, DELETE, &response);
		close(sockt);
		open_sockt();
		return (0);
	}
	p_error("Unable to connect with initiator");
	/*NOTREACHED*/
}

/*
 * Look for an invitation on 'machine'
 */
look_for_invite(response)
	CTL_RESPONSE *response;
{
	struct in_addr machine_addr;

	current_state = "Checking for invitation on caller's machine";
	ctl_transact(his_machine_addr, msg, LOOK_UP, response);
	/* the switch is for later options, such as multiple invitations */
	switch (response->answer) {

	case SUCCESS:
		msg.id_num = response->id_num;
		return (1);

	default :
		/* there wasn't an invitation waiting for us */
		return (0);
	}
}

/*  
 * heuristic to detect if need to reshuffle CTL_RESPONSE structure
 *
 * What a royal pain. It seems that vax responses and sun responses
 * are different. Not only that but neither of those responses match
 * what hcx structures look like. In response, the HCX will accept
 * either 'sun' or 'vax' responses and will generate responses that
 * look like 'sun' responses. (Atleast the sun has it's bytes in the
 * right order, even if they're not spaced right.)
 */

#define swapshort(a) (((a << 8) | ((unsigned short) a >> 8)) & 0xffff)
#define swaplong(a) ((swapshort(a) << 16) | (swapshort(((unsigned)a >> 16))))

#ifdef hcx

swapresponse(rsp)
	CTL_RESPONSE *rsp;
{

	/*
	 * If this is a vax response structure,
	 *      Just swap the byte ordering in the sin_family and id fields.
	 *	Luckily the vax has the fields of the structure at the same
	 *              offsets as the hcx.
	 */

	if ( rsp->addr.sin_family == swapshort(AF_INET)) {
		rsp->addr.sin_family = swapshort( AF_INET ) ;
		rsp->id_num = swaplong(rsp->id_num);

	/*
	 * Otherwise, just assume this is a sun response structure.
	 *
	 * On the sun, there is no hole between the 'answer' and 'idnum'
	 * fields. However, the byte ordering in a single field are
	 * already correct.
	 */

	} else {
		struct ctl_response_sun {
			char type;
			char answer;
			char id_num1;
			char id_num2;
			unsigned short id_num3;
			short sin_family;
			short sin_port;
			short sin_addr2;
			short sin_addr1;
		} swaprsp ;
		bcopy(rsp, &swaprsp, sizeof(struct ctl_response_sun));

        	/* move the id_num field, skipping the two byte hole
        	 * in the ctl_response struct.
        	 */
        	rsp->id_num = (swaprsp.id_num1<<24) | (swaprsp.id_num2<<16) |
        	              swaprsp.id_num3 ;

		/*
		 * Now for the socket address.
		 */
		rsp->addr.sin_family = swaprsp.sin_family;
 		rsp->addr.sin_port = swaprsp.sin_port;
		rsp->addr.sin_addr.s_addr =
			    swaprsp.sin_addr1 | (swaprsp.sin_addr2 << 16);

	}

}
#endif

#ifdef sun
struct ctl_response_vax {
	char type;
	char answer;
	short junk;
	int id_num;
	struct sockaddr_in addr;
};

swapresponse(rsp)
	CTL_RESPONSE *rsp;
{
	struct ctl_response_vax swaprsp;
	
	if (rsp->addr.sin_family != AF_INET) {
		bcopy(rsp, &swaprsp, sizeof(CTL_RESPONSE));
		swaprsp.addr.sin_family = swapshort(swaprsp.addr.sin_family);
		if (swaprsp.addr.sin_family == AF_INET) {
			rsp->addr = swaprsp.addr;
			rsp->type = swaprsp.type;
			rsp->answer = swaprsp.answer;
			rsp->id_num = swaplong(swaprsp.id_num);
		}
	}
}
#endif

#ifdef vax
struct ctl_response_sun {
	char type;
	char answer;
	unsigned short id_num2;
	unsigned short id_num1;
	short sin_family;
	short sin_port;
	short sin_addr2;
	short sin_addr1;
};

swapresponse(rsp)
	CTL_RESPONSE *rsp;
{
	struct ctl_response_sun swaprsp;
	
	if (rsp->addr.sin_family != AF_INET) {
		bcopy(rsp, &swaprsp, sizeof(struct ctl_response_sun));
		if (swaprsp.sin_family == swapshort(AF_INET)) {
			rsp->type = swaprsp.type;
			rsp->answer = swaprsp.answer;
			rsp->id_num = swapshort(swaprsp.id_num1)
			    | (swapshort(swaprsp.id_num2) << 16);
			rsp->addr.sin_family = swapshort(swaprsp.sin_family);
 			rsp->addr.sin_port = swaprsp.sin_port;
			rsp->addr.sin_addr.s_addr =
			    swaprsp.sin_addr2 | (swaprsp.sin_addr1 << 16);
		}
	}
}
#endif
