static	char *sccsid = "@(#)mon.c	2.0 (HCX) 5/7/84";

/*
 * mon.c -- Combined UCB and ATT monitoring, courtesy of the
 * universe switch project.
 */

#include <mon.h>

#define PROFDIR		"PROFDIR"
#define	NOSPACE		"No space for monitor buffer(s)\n"
#define	CNTROVF		"mcount: counter overflow\n"
#define TOOSMALL	"Profiling buffer too small\n"

#define ARCDENSITY	5	/* an estimate of the number of entry
				points in 100 bytes of text */
#define MINARCS		50	/* the minimum number of entry point
				counters to allocate */
#define XWRDS_PER_CNTR	2	/* the number of words of text to associate
				with a single pc counter */

char **___Argv;			/* Set to argv array by mcrt0 */
static int Profiling = 3;	/* On/off switch used by mcount */
static char Mon_out[128];	/* Name of the "mon.out" file */
static struct hdr *Monbase;	/* Start of monitor buffer */
static int Monsize;		/* Size of entire monitor buffer */
static char *Profbase;		/* Start of profiling section */
static int Profsize;		/* Size of profiling section */
static int Scale;		/* Profiling scale */
static struct cnt *Countbase;	/* Next free counter */

extern char *strrchr (), *getenv (), *sbrk (), *minbrk;
extern void monstartup (), monitor (), moncontrol (), mcount ();

/*
 * monstartup -- High level interface to monitor.
 *
 * Given low and high pc addresses, monstartup computes the
 * number of entry point and pc counters to use, allocates
 * the buffer, and calls monitor. Text locations lpc to hpc-1
 * inclusive are profiled.
 */
void monstartup (lpc, hpc)
	char *lpc;
	char *hpc;
{
	int nfns;		/* Number of entry point counters */
	int npc;		/* Number of pc counters */
	char *buffer;		/* Ptr to monitor buffer */
	int bufsize;		/* Size of monitor buffer */

	/* Compute the number of entry point counters to
	allocate. */

	nfns = (hpc - lpc) * ARCDENSITY / 100;
	if (nfns < MINARCS)
		nfns = MINARCS;

	/* Compute the number of pc counters to allocate. */

	npc = ((hpc - lpc) + sizeof (WORD) - 1) / sizeof (WORD);
	npc /= XWRDS_PER_CNTR;

	/* Allocate the buffer. */

	bufsize = sizeof (struct hdr) + (nfns * sizeof (struct cnt)) +
		(npc * sizeof (WORD));
	bufsize = ((bufsize + sizeof (int) - 1) / sizeof (int)) * sizeof (int);

	if ((buffer = sbrk (bufsize)) == (char *) -1) {
		write (2, NOSPACE, sizeof (NOSPACE));
		return;
	}
	minbrk = sbrk (0);

	monitor (lpc, hpc, (WORD *) buffer, bufsize / sizeof (WORD), nfns);
}

/*
 * monitor -- Arrange for the monitoring of a program.
 */
void monitor (lpc, hpc, wordbuf, nwords, nfns)
	char	*lpc, *hpc;	/* boundaries of text to be profiled */
	WORD	*wordbuf;	/* ptr to space for monitor data (WORDs) */
	int	nwords;		/* size of above space (in WORDs) */
	int	nfns;		/* max number of functions to counted */
{
	int fd, tmp, nbytes;
	char *s, *bytebuf;
	
	/* Convert the description of the word buffer into a
	description of a byte buffer. */

	bytebuf = (char *) wordbuf;
	nbytes = nwords * sizeof (WORD);

	/* If lpc is null, write the monitor buffer out to the
	"mon.out" file  and return. */

	if (lpc == NULL) {
		if (Monbase != NULL) {
			moncontrol(0);
			if ((fd = creat (Mon_out, 0666)) < 0 ||
				write (fd, (char *) Monbase, Monsize) !=
				Monsize)
				(void) perror (Mon_out);
			if (fd >= 0)
				(void) close (fd);
		}
		return;
	}

	/* Generate the name of the "mon.out" file from the
	PROFDIR environment variable.

	If PROFDIR doesn't exist, "mon.out" is produced as before.
 	If PROFDIR = NULL, no profiling output is produced.
	If PROFDIR = string, "string/pid.progname" is produced,
	where name consists of argv[0] suitably massaged. */

	if ((s = getenv (PROFDIR)) == NULL)
		strcpy (Mon_out, MON_OUT);

	else if (*s == '\0')
		return;

	else {
		register int pid, n;
		register char *t = Mon_out;

		/* The code below essentially does an sprintf (Mon_out,
		"%s/%d.%s", getenv (PROFDIR), getpid (), ___Argv[0]). */

		while (*s != '\0')
			*t++ = *s++;
		*t++ = '/';

		if ((pid = getpid ()) <= 0)
			pid = 1;

		for (n = 10000; n > pid; n /= 10);

		while (pid) {
			*t++ = (pid / n) + '0';
			pid %= n;
			n /= 10;
		}

		*t++ = '.';

		if (___Argv != NULL)
			strcpy (t, (s = strrchr (___Argv[0], '/')) != NULL ?
			s+1 : ___Argv[0]);
	}

	/* Initialize globals. */

	Monbase = (struct hdr *) bytebuf;
	Monsize = nbytes;
	Monbase -> lpc = lpc;
	Monbase -> hpc = hpc;
	Monbase -> nfns = nfns;

	Countbase = (struct cnt *) (bytebuf + sizeof (struct hdr));

	tmp = sizeof (struct hdr) + (nfns * sizeof (struct cnt));
	Profbase = bytebuf + tmp;
	Profsize = nbytes - tmp;
	if (Profsize <= 0) {
		write (2, TOOSMALL, sizeof (TOOSMALL));
		return;
	}

	/* Compute the infamous scale. If called from monstartup
	with XWRDS_PER_CNTR=2, scale should come out to 0x8000. */

	tmp = hpc - lpc;
	if (Profsize < tmp) {
		Scale = ((float) Profsize / tmp) * 65536;
	}
	else { /* Profiling buffer is larger than we need. */
		Scale = 65536;
		tmp = ((Profsize - tmp) / sizeof (WORD)) * sizeof (WORD);
		Profsize -= tmp;
		Monsize -= tmp;
	}

	/* Start the monitoring. */
	moncontrol(1);
}

/*
 * moncontrol -- Control Profiling
 */
void moncontrol (on)
	int on;
{
	if (on) { /* start profiling */
		profil (Profbase, Profsize, Monbase -> lpc, Scale);
		Profiling = 0;
	}
	else { /* stop profiling */
		Profiling = 3;
		profil ((char *)0, 0, 0, 0);
	}
}

/*
 * mcount -- Entry point counter.
 *
 * The -p option on the compiler generates, in the prologue
 * of every routine, a call to mcount. Mcount is passed the
 * address of a word also in the prologue of the routine.
 * Said word points to the counter to increment. The first
 * time through mcount the pointer will be zero and a counter
 * will be allocated from the monitor buffer.
 */
void mcount (mcntpp)
	long **mcntpp;
{
	char *selfpc = *(char **)((&mcntpp)-3);	/* -8(fp) */

	/* Check that we aren't recursively invoked. */

	if (Profiling)
		return;
	Profiling++;

	/* If the counter pointer is null, we need to
	allocate a counter from the monitor buffer. */

	if (*mcntpp == 0) {

		if (((char *) Countbase) >= Profbase) {
			write (2, CNTROVF, sizeof (CNTROVF));
			Profiling--;
			return;
		}

		Countbase -> fnpc = selfpc;
		*mcntpp = & (Countbase -> mcnt);
		Countbase++;

	}

	/* Increment the counter. */
	(**mcntpp)++;
	Profiling--;

}
