#	@(#)libc.mk	1.9
#
# Makefile for libc
#
# The variable PRF is null by default, causing both the standard C library
# and a profiled library to be maintained.  If profiled object is not
# desired, the reassignment PRF=@# should appear in the make command line.
#
# This subtree contains the source for the two (AT&T and UCB) C libraries.
# Some of the source applies only to the AT&T library, some applies only to
# the UCB library, and some is common. All the source in a single directory
# is the same type (AT&T, UCB, or both). MAKE variables defined below detail
# which directories comprise each library.
#
# The directory structure is that of AT&T's original structure with UCB
# directories added where appropriate:
#
#    hcx                       port
#       crt                       sys
#       csu                       gen
#       gen                          att
#          att                       ucb
#          ucb                    print
#       sys                       stdio
#          att                    ucbstdio
#          ucb                    compat-4.1
#       compat-4.1                compat-sys5
#	compat-sys		  rpc
#				  yp
#				  net

.UNIVERSE=att
MACHINE=hcx
DONE=
INS=/etc/install
CFLAGS=$(CFL)
LDFLAGS=$(LFL)

AR=ar
STRIP=strip
LORDER=lorder
LIB=$(ROOT)/lib
LIBP=$(ROOT)/lib/libp

# AT&T directories.
ATTPORT=stdio print sys gen gen/att rpc yp net
ATTMACH=crt csu gen gen/att sys sys/att

# UCB directories.
UCBPORT=compat-4.1 compat-sys5 ucbstdio sys gen gen/ucb rpc yp net
UCBMACH=compat-4.1 compat-sys5 crt csu gen gen/ucb sys sys/ucb

all: compile attlibc ucblibc
att: compile attlibc
ucb: compile ucblibc

compile:
#
# Compile machine-dependent library modules.
	cd $(MACHINE); $(MAKE) all
#
# Compile portable library modules.
	cd port; $(MAKE) all

attlibc:
#
# Place portable modules in "object" directory, then overlay
# the machine-dependent modules.
	-rm -rf object
	mkdir object
	for i in $(ATTPORT); do cp port/$$i/*.o object; done
	$(PRF)for i in $(ATTPORT); do cp port/$$i/*.p object; done
	for i in $(ATTMACH); do cp $(MACHINE)/$$i/*.o object; done
	$(PRF)for i in $(ATTMACH); do cp $(MACHINE)/$$i/*.p object; done
#
# Delete temporary libraries.
	-rm -f att.libc
	$(PRF)-rm -f attp.libc
#
# Set aside run-time modules, which don't go in library archive!
	cd object; mv *crt0.o ..
#
# Build the archive out of the remaining modules.
#
# Figure out the correct ordering for all the archive modules except
# cuexit and fakcu, which must go at end of archive.
	(cd object; \
	mv cuexit.o cuexit.x; \
	mv fakcu.o fakcu.x; \
	$(LORDER) *.o | tsort >objlist; \
	mv cuexit.x cuexit.o; \
	mv fakcu.x fakcu.o; \
	$(AR) q ../att.libc `cat objlist` cuexit.o fakcu.o; \
	$(AR) t ../att.libc >../att.contents; \
	$(PRF)for i in *.p ; do mv $$i `basename $$i .p`.o ; done; \
	$(PRF)$(AR) q ../attp.libc `cat objlist` cuexit.o fakcu.o; \
	$(PRF)$(AR) t ../attp.libc >../attp.contents)
#
# all done
	-rm -rf object
	$(DONE)

ucblibc:
#
# Place portable modules in "object" directory, then overlay
# the machine-dependent modules.
	-rm -rf object
	mkdir object
	for i in $(UCBPORT); do cp port/$$i/*.o object; done
	$(PRF)for i in $(UCBPORT); do cp port/$$i/*.p object; done
	for i in $(UCBMACH); do cp $(MACHINE)/$$i/*.o object; done
	$(PRF)for i in $(UCBMACH); do cp $(MACHINE)/$$i/*.p object; done
#
# Delete temporary libraries.
	-rm -f ucb.libc
	$(PRF)-rm -f ucbp.libc
#
# Set aside run-time modules, which don't go in library archive!
	cd object; mv *crt0.o ..
#
# Build the archive out of the remaining modules.
	(cd object; \
	$(LORDER) *.o | tsort >objlist; \
	$(AR) q ../ucb.libc `cat objlist`; \
	$(AR) ma flsbuf.o ../ucb.libc exit.o; \
	$(AR) t ../ucb.libc >../ucb.contents; \
	$(PRF)for i in *.p ; do mv $$i `basename $$i .p`.o ; done; \
	$(PRF)$(AR) q ../ucbp.libc `cat objlist`; \
	$(PRF)$(AR) ma flsbuf.o ../ucbp.libc exit.o; \
	$(PRF)$(AR) t ../ucbp.libc >../ucbp.contents)
#
# all done
	-rm -rf object
	$(DONE)

move:
#
# move the library or libraries into the correct directory
	for i in *crt0.o; do $(INS) -r -f $(LIB) -m 0644 -u bin -g bin $$i; done
	cp att.libc $(LIB)/libattc.a
	chmod 644 $(LIB)/libattc.a
	chown bin $(LIB)/libattc.a
	chgrp bin $(LIB)/libattc.a
	cp ucb.libc $(LIB)/libucbc.a
	chmod 644 $(LIB)/libucbc.a
	chown bin $(LIB)/libucbc.a
	chgrp bin $(LIB)/libucbc.a
	ln -sf 'lib{}c.a' $(LIB)/libc.a
	$(PRF)if test ! -d $(LIBP); then mkdir $(LIBP); fi
	$(PRF)cp attp.libc $(LIBP)/libattc.a
	$(PRF)chmod 644 $(LIBP)/libattc.a
	$(PRF)chown bin $(LIBP)/libattc.a
	$(PRF)chgrp bin $(LIBP)/libattc.a
	$(PRF)cp ucbp.libc $(LIBP)/libucbc.a
	$(PRF)chmod 644 $(LIBP)/libucbc.a
	$(PRF)chown bin $(LIBP)/libucbc.a
	$(PRF)chgrp bin $(LIBP)/libucbc.a
	$(PRF)ln -sf 'lib{}c.a' $(LIBP)/libc.a

install: all move

clean:
	-rm -rf object
	-rm -f *.contents
	cd port; $(MAKE) clean
	cd $(MACHINE); $(MAKE) clean

clobber:
	-rm -rf object
	-rm -f *.o *.libc *.contents
	cd port; $(MAKE) clobber
	cd $(MACHINE); $(MAKE) clobber
