#ifndef lint
static	char sccsid[] = "@(#)mntent.c 1.1 86/02/03 SMI";
/* @(#)mntent.c	2.1 86/04/11 NFSSRC */
#endif

#include <stdio.h>
#include <ctype.h>
#include <mntent.h>
#include <sys/file.h>

static	struct mntent mnt;
static	char line[BUFSIZ+1];

static char *
mntstr(p)
	register char **p;
{
	char *cp = *p;
	char *retstr;

	while (*cp && isspace(*cp))
		cp++;
	retstr = cp;
	while (*cp && !isspace(*cp))
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (retstr);
}

static int
mntdigit(p)
	register char **p;
{
	register int value = 0;
	char *cp = *p;
	char *retstr;

	while (*cp && isspace(*cp))
		cp++;
	for (; *cp && isdigit(*cp); cp++) {
		value *= 10;
		value += *cp - '0';
	}
	while (*cp && !isspace(*cp))
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (value);
}

static
mnttabscan(mnttabp, mnt)
	FILE *mnttabp;
	struct mntent *mnt;
{
	char *cp;

	do {
		cp = fgets(line, 256, mnttabp);
		if (cp == NULL) {
			return (EOF);
		}
	} while (*cp == '#');
	mnt->mnt_fsname = mntstr(&cp);
	if (*cp == '\0')
		return (1);
	mnt->mnt_dir = mntstr(&cp);
	if (*cp == '\0')
		return (2);
	mnt->mnt_type = mntstr(&cp);
	if (*cp == '\0')
		return (3);
	mnt->mnt_opts = mntstr(&cp);
	if (*cp == '\0')
		return (4);
	mnt->mnt_freq = mntdigit(&cp);
	if (*cp == '\0')
		return (5);
	mnt->mnt_passno = mntdigit(&cp);
	return (6);
}
	
FILE *
setmntent(fname, flag)
	char *fname;
	char *flag;
{
	FILE *mnttabp;
	int lock;

	if ((mnttabp = fopen(fname, flag)) == NULL) {
		return (NULL);
	}
	lock = LOCK_SH;
	while (*flag) {
		if (*flag == 'w' || *flag == 'a' || *flag == '+') {
			lock = LOCK_EX;
		}
		flag++;
	}
	if (flock(fileno(mnttabp), lock) < 0) {
		fclose(mnttabp);
		return (NULL);
	}
	return (mnttabp);
}

int
endmntent(mnttabp)
	FILE *mnttabp;
{

	if (mnttabp) {
		fclose(mnttabp);
	}
	return (1);
}

struct mntent *
getmntent(mnttabp)
	FILE *mnttabp;
{
	int nfields;

	if (mnttabp == 0)
		return ((struct mntent *)0);
	nfields = mnttabscan(mnttabp, &mnt);
	if (nfields == EOF || nfields != 6)
		return ((struct mntent *)0);
	return (&mnt);
}

addmntent(mnttabp, mnt)
	FILE *mnttabp;
	register struct mntent *mnt;
{
	if (fseek(mnttabp, 0, 2) < 0) {
		return (1);
	}
	mntprtent(mnttabp, mnt);
	return (0);
}

static char tmpopts[256];

static char *
mntopt(p)
	char **p;
{
	char *cp = *p;
	char *retstr;

	while (*cp && isspace(*cp))
		cp++;
	retstr = cp;
	while (*cp && *cp != ',')
		cp++;
	if (*cp) {
		*cp = '\0';
		cp++;
	}
	*p = cp;
	return (retstr);
}

static int
mntprtent(mnttabp, mnt)
	FILE *mnttabp;
	register struct mntent *mnt;
{
	fprintf(mnttabp, "%s %s %s %s %d %d\n",
	    mnt->mnt_fsname,
	    mnt->mnt_dir,
	    mnt->mnt_type,
	    mnt->mnt_opts,
	    mnt->mnt_freq,
	    mnt->mnt_passno);
	return(0);
}

char *
hasmntopt(mnt, opt)
	register struct mntent *mnt;
	register char *opt;
{
	char *f, *opts;

	strcpy(tmpopts, mnt->mnt_opts);
	opts = tmpopts;
	f = mntopt(&opts);
	for (; *f; f = mntopt(&opts)) {
		if (strncmp(opt, f, strlen(opt)) == 0)
			return (f - tmpopts + mnt->mnt_opts);
	} 
	return (NULL);
}

int
vfymntopt (mnt, bad_opt)

        register struct mntent *mnt;
        register char *bad_opt;

{
	/* allow each function to verify it's own options.
	   If it is not a valid filesystem type then return
	   that as an error code.
	*/

           if (strcmp (mnt->mnt_type, MNTTYPE_42) == 0)
                return (vfy_42(mnt, bad_opt));

	   if (strcmp (mnt->mnt_type,  MNTTYPE_NFS)== 0)
	   	return (vfy_nfs (mnt, bad_opt));

	   if (strcmp (mnt->mnt_type, MNTTYPE_PC) == 0)
	   	return (vfy_pc (mnt, bad_opt));

	   if (strcmp (mnt->mnt_type, MNTTYPE_SWAP) == 0)
	   	return (vfy_swap (mnt, bad_opt));

	   if (strcmp (mnt->mnt_type, MNTTYPE_IGNORE) == 0)
		return (vfy_ignore (mnt, bad_opt));

	   else /* return error condition - no such filesystem */
	   	return (MNTERR_BADFS);
}

static int
vfy_42 (mnt, bad_opt)

	struct mntent *mnt;
	char *bad_opt;

{

	if (!valid_options (mnt->mnt_opts, bad_opt))
	   return (MNTERR_BADOPT); /* there was an error */
	else
	{
	   /* verify that the mnt entry contains only 4.2 options */
	   char *opts, *f;

	   strcpy (tmpopts, mnt->mnt_opts);
	   opts = tmpopts;
	   f = mntopt(&opts);
	   for (; *f; f = mntopt (&opts))
	   {
	      if (strcmp (f,MNTOPT_RW) == 0) ;
	      else
	      if (strcmp (f, MNTOPT_RO) == 0) ;
	      else
	      if (strcmp (f, MNTOPT_QUOTA) == 0) ;
	      else
	      if (strcmp (f, MNTOPT_NOQUOTA) == 0) ;
	      else
	      {
	        strcpy (bad_opt, f);
	        return (MNTERR_NOPT);
	      }
	   }
	}
	return (0);  /* indicate that all is ok */
}

static int
vfy_nfs (mnt, bad_opt)

	struct mntent *mnt;
	char *bad_opt;

{
	if (!valid_options (mnt->mnt_opts, bad_opt))
	   return (MNTERR_BADOPT); /* there was an error */
	else
	{
	   /* all options are valid under NFS */
	   return (0);
	}
}

static int
vfy_pc (mnt, bad_opt)

	struct mntent *mnt;
	char *bad_opt;

{
	if (valid_options (mnt->mnt_opts, bad_opt))
	   return (0);  /* indicate that all options are valid */
	else
	   return (MNTERR_BADOPT); /* there was an error */
}

static int
vfy_swap (mnt, bad_opt)

	struct mntent *mnt;
	char *bad_opt;

{
	return (0);  /* indicate that all options are valid */
}

static int
vfy_ignore (mnt, bad_opt)

	struct mntent *mnt;
	char *bad_opt;

{
	return (0);  /* indicate that all options are valid */
}

static int
valid_options (options, buff)

	register char *options;
	register char *buff;

{
	char *opts, *f;

  	strcpy (tmpopts, options);
  	opts = tmpopts;
  	f = mntopt (&opts);
  	for (; *f; f=mntopt (&opts))
	{
	/* is it a valid option on any file system ??? */

	   if (strcmp (f,MNTOPT_RO) == 0) ;	/* read only */
	   else
	   if (strcmp (f, MNTOPT_RW) == 0) ;	/* read/write */
	   else
	   if (strcmp (f, MNTOPT_QUOTA) == 0) ;/* quotas */
	   else
	   if (strcmp (f, MNTOPT_NOQUOTA) == 0) ;/* no quotas */
	   else
	   if (strcmp (f, MNTOPT_SOFT) == 0) ;	/* soft mount */
	   else
	   if (strcmp (f, MNTOPT_HARD) == 0) ;	/* hard mount */
	   else
	   if (strncmp (f, "retry",5) == 0) ;/* retry mount count */
	   else
	   if (strncmp (f, "suid",4) == 0) ; /* set uid allowed */
	   else
	   if (strncmp (f, MNTOPT_NOSUID,6) == 0) ;/* set-uid disallowed */
	   else
	   if (strncmp (f, MNTOPT_NOAUTO,6) == 0) ;/* don"t auto mount fs */
	   else
	   if (strncmp (f, "fg",2) == 0) ;	/* retry in foreground */
	   else
	   if (strncmp (f, "bg",2) == 0) ;	/* retry in background */
	   else
	   if (strncmp (f, "rsize",4) == 0) ;/* set read f size */
	   else
	   if (strncmp (f, "wsize",4) == 0) ;/* set write f size */
	   else
	   if (strncmp (f, "timeo",5) == 0) ;/* NFS timeout count */
	   else
	   if (strncmp (f, "retrans",7) == 0) ;/* NFS retrans. count */
	   else
	   if (strncmp (f, "port",4) == 0)  ;/* IP port number */
	   else
	   if (strncmp (f, MNTOPT_INTR,4) == 0) ;	/* kb interrupts on mount */
	   else
	   {
	      strcpy (buff, f);
	      return (0);
	   }
	}
	return (1);  /* tell caller that all parameters check out */
}
