/*
 * Sun RPC is a product of Sun Microsystems, Inc. and is provided for
 * unrestricted use provided that this legend is included on all tape
 * media and as a part of the software program in whole or part.  Users
 * may copy or modify Sun RPC without charge, but are not authorized
 * to license or distribute it to anyone else except as part of a product or
 * program developed by the user.
 * 
 * SUN RPC IS PROVIDED AS IS WITH NO WARRANTIES OF ANY KIND INCLUDING THE
 * WARRANTIES OF DESIGN, MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE, OR ARISING FROM A COURSE OF DEALING, USAGE OR TRADE PRACTICE.
 * 
 * Sun RPC is provided with no support and without any obligation on the
 * part of Sun Microsystems, Inc. to assist in its use, correction,
 * modification or enhancement.
 * 
 * SUN MICROSYSTEMS, INC. SHALL HAVE NO LIABILITY WITH RESPECT TO THE
 * INFRINGEMENT OF COPYRIGHTS, TRADE SECRETS OR ANY PATENTS BY SUN RPC
 * OR ANY PART THEREOF.
 * 
 * In no event will Sun Microsystems, Inc. be liable for any lost revenue
 * or profits or other special, indirect and consequential damages, even if
 * Sun has been advised of the possibility of such damages.
 * 
 * Sun Microsystems, Inc.
 * 2550 Garcia Avenue
 * Mountain View, California  94043
 */
#ifndef lint
static char sccsid[] = "@(#)xdr_float.c 1.1 86/09/24 Copyr 1984 Sun Micro";
#endif

/*
 * xdr_float.c, Generic XDR routines impelmentation.
 *
 * Copyright (C) 1984, Sun Microsystems, Inc.
 *
 * These are the "floating point" xdr routines used to (de)serialize
 * most common data items.  See xdr.h for more info on the interface to
 * xdr.
 */

#include <rpc/types.h>
#include <rpc/xdr.h>
#include <stdio.h>


/*
 * NB: Not portable.
 * This routine works on Suns (Sky / 68000's) and Vaxen.
 * Now works for HCX.
 */
   
#ifdef vax

/* What IEEE single precision floating point looks like on a Vax */
struct  ieee_single {
	unsigned int	mantissa: 23;
	unsigned int	exp     : 8;
	unsigned int	sign    : 1;
};

/* Vax single precision floating point */
struct  vax_single {
	unsigned int	mantissa1 : 7;
	unsigned int	exp       : 8;
	unsigned int	sign      : 1;
	unsigned int	mantissa2 : 16;
		
};

#define VAX_SNG_BIAS    0x81
#define IEEE_SNG_BIAS   0x7f

static struct sgl_limits {
	struct vax_single s;
	struct ieee_single ieee;
} sgl_limits[2] = {
	{{ 0x3f, 0xff, 0x0, 0xffff },	/* Max Vax */
	{ 0x0, 0xff, 0x0 }},		/* Max IEEE */
	{{ 0x0, 0x0, 0x0, 0x0 },	/* Min Vax */
	{ 0x0, 0x0, 0x0 }}		/* Min IEEE */
};
#endif /* vax */

#ifdef hcx

/* What IEEE single precision floating point looks like on an HCX */
/* HCX single precision floating point */
struct  hcx_single {
	unsigned int	sign    : 1;
	unsigned int	exp     : 8;
	unsigned int	mantissa: 23;
};

#define HCX_SNG_BIAS    0x81
#define IEEE_SNG_BIAS   0x7f

#endif hcx

bool_t 
xdr_float(xdrs, fp)
	register XDR *xdrs;
	register float *fp;
{
#ifdef vax
	struct ieee_single is;
	struct vax_single vs, *vsp;
	struct sgl_limits *lim;
	int i;
#endif vax
#ifdef hcx
	struct hcx_single *hsp;
	struct hcx_single hs;
#endif hcx
	switch (xdrs->x_op) {

	case XDR_ENCODE:
#ifdef hcx
		/*
		 * HCX single precision format and IEEE single precision
		 * format have the same field definitions but not the same
		 * values.
		 */
		hs = *((struct hcx_single *)fp);
		/*
		 * Since the exponent ranges don't match,
		 *      Translation HCX largest no. to IEEE infinity.
		 */
		if (hs.exp == 0xff && hs.mantissa >= 0x7ffff0) {
			hs.mantissa = 0 ;
		/*
		 * Translate HCX reserved to IEEE not-a-number.
		 */
		} else if ( hs.exp == 0 && hs.sign == 1 ) {
		        hs.exp = 0xff ;
		        hs.mantissa = 1 ;
		/*
		 * Translate HCX zero to IEEE zero.
		 */
		} else if ( hs.exp == 0 && hs.sign == 0 ) {
		        hs.mantissa = 0 ;
		/*
		 * Numbers which are too small to translate get
		 * converted to zero.
		 */
		} else if ( hs.exp < (HCX_SNG_BIAS - IEEE_SNG_BIAS)) {
		        hs.exp = 0 ;
		        hs.mantissa = 0 ;
		/*
		 * All other numbers are different merely in value of the
		 * exponent.
		 */
	        } else {
			hs.exp -= HCX_SNG_BIAS - IEEE_SNG_BIAS;
		}
		return (XDR_PUTLONG(xdrs, (long *)&hs));
#endif hcx
#ifdef mc68000
		return (XDR_PUTLONG(xdrs, (long *)fp));
#endif mc68000
#ifdef vax
		vs = *((struct vax_single *)fp);
		for (i = 0, lim = sgl_limits;
			i < sizeof(sgl_limits)/sizeof(struct sgl_limits);
			i++, lim++) {
			if ((vs.mantissa2 == lim->s.mantissa2) &&
				(vs.exp == lim->s.exp) &&
				(vs.mantissa1 == lim->s.mantissa1)) {
				is = lim->ieee;
				goto shipit;
			}
		}
		is.exp = vs.exp - VAX_SNG_BIAS + IEEE_SNG_BIAS;
		is.mantissa = (vs.mantissa1 << 16) | vs.mantissa2;
	shipit:
		is.sign = vs.sign;
		return (XDR_PUTLONG(xdrs, (long *)&is));
#endif vax

	case XDR_DECODE:
#if defined(hcx)
		/*
		 * HCX single precision format and IEEE single precision
		 * format have the same field definitions but not the same
		 * values.
		 */
		hsp = (struct hcx_single *)fp;
		if (!XDR_GETLONG(xdrs, (long *)hsp))
			return (FALSE);
		/*
		 * Since the exponent ranges don't match,
		 *      Translation IEEE infinity and other large IEEE numbers
		 *      to largest HCX number.
		 */
		if ( hsp->exp > 0xff - (HCX_SNG_BIAS - IEEE_SNG_BIAS) ) {
			hsp->mantissa = 0x7fffff ;
			hsp->exp = 0xff ;
		/*
		 * Translate IEEE not-a-number to HCX reserved.
		 */
		} else if ( hsp->exp == 0xff && hsp->mantissa != 0 ) {
			hsp->exp = 0 ;
			hsp->sign = 1 ;
		/*
		 * Translate IEEE zero to HCX zero.
		 */
		} else if ( hsp->exp == 0 && hsp->mantissa == 0 ) {
		        hsp->sign = 0 ;
		/*
		 * All other numbers are different merely in value of the
		 * exponent.
		 */
	        } else {
			hsp->exp += HCX_SNG_BIAS - IEEE_SNG_BIAS;
		}
		return (TRUE);
#endif hcx
#ifdef mc68000
		return (XDR_GETLONG(xdrs, (long *)fp));
#endif mc68000
#ifdef vax
		vsp = (struct vax_single *)fp;
		if (!XDR_GETLONG(xdrs, (long *)&is))
			return (FALSE);
		for (i = 0, lim = sgl_limits;
			i < sizeof(sgl_limits)/sizeof(struct sgl_limits);
			i++, lim++) {
			if ((is.exp == lim->ieee.exp) &&
				(is.mantissa == lim->ieee.mantissa)) {
				*vsp = lim->s;
				goto doneit;
			}
		}
		vsp->exp = is.exp - IEEE_SNG_BIAS + VAX_SNG_BIAS;
		vsp->mantissa2 = is.mantissa;
		vsp->mantissa1 = (is.mantissa >> 16);
	doneit:
		vsp->sign = is.sign;
		return (TRUE);
#endif vax

	case XDR_FREE:
		return (TRUE);
	}
	return (FALSE);
}

/*
 * This routine works on Suns (Sky / 68000's) and Vaxen.
 */

#ifdef vax
/* What IEEE double precision floating point looks like on a Vax */
struct  ieee_double {
	unsigned int	mantissa1 : 20;
	unsigned int	exp	  : 11;
	unsigned int	sign	  : 1;
	unsigned int	mantissa2 : 32;
};
 
/* Vax double precision floating point */
struct  vax_double {
	unsigned int	mantissa1 : 7;
	unsigned int	exp       : 8;
	unsigned int	sign      : 1;
	unsigned int	mantissa2 : 16;
	unsigned int	mantissa3 : 16;
	unsigned int	mantissa4 : 16;
};

#define VAX_DBL_BIAS    0x81
#define IEEE_DBL_BIAS   0x3ff
#define MASK(nbits)     ((1 << nbits) - 1)

static struct dbl_limits {
	struct  vax_double d;
	struct  ieee_double ieee;
} dbl_limits[2] = {
	{{ 0x7f, 0xff, 0x0, 0xffff, 0xffff, 0xffff },	/* Max Vax */
	{ 0x0, 0x7ff, 0x0, 0x0 }},			/* Max IEEE */
	{{ 0x0, 0x0, 0x0, 0x0, 0x0, 0x0},		/* Min Vax */
	{ 0x0, 0x0, 0x0, 0x0 }}				/* Min IEEE */
};

#endif /* vax */


bool_t           
xdr_double(xdrs, dp)
	register XDR *xdrs;
	double *dp;
{
	register long *lp;
#ifdef vax
	struct  ieee_double id;
	struct  vax_double vd;
	register struct dbl_limits *lim;
	int i;
#endif vax

	switch (xdrs->x_op) {

	case XDR_ENCODE:
#ifdef mc68000
		lp = (long *)dp;
#endif mc68000
#ifdef vax
		vd = *((struct  vax_double *)dp);
		for (i = 0, lim = dbl_limits;
			i < sizeof(dbl_limits)/sizeof(struct dbl_limits);
			i++, lim++) {
			if ((vd.mantissa4 == lim->d.mantissa4) &&
				(vd.mantissa3 == lim->d.mantissa3) &&
				(vd.mantissa2 == lim->d.mantissa2) &&
				(vd.mantissa1 == lim->d.mantissa1) &&
				(vd.exp == lim->d.exp)) {
				id = lim->ieee;
				goto shipit;
			}
		}
		id.exp = vd.exp - VAX_DBL_BIAS + IEEE_DBL_BIAS;
		id.mantissa1 = (vd.mantissa1 << 13) | (vd.mantissa2 >> 3);
		id.mantissa2 = ((vd.mantissa2 & MASK(3)) << 29) |
				(vd.mantissa3 << 13) |
				((vd.mantissa4 >> 3) & MASK(13));
	shipit:
		id.sign = vd.sign;
		lp = (long *)&id;
#endif vax
		return (XDR_PUTLONG(xdrs, lp++) && XDR_PUTLONG(xdrs, lp));

	case XDR_DECODE:
#ifdef mc68000
		lp = (long *)dp;
		return (XDR_GETLONG(xdrs, lp++) && XDR_GETLONG(xdrs, lp));
#endif mc68000
#ifdef vax
		lp = (long *)&id;
		if (!XDR_GETLONG(xdrs, lp++) || !XDR_GETLONG(xdrs, lp))
			return (FALSE);
		for (i = 0, lim = dbl_limits;
			i < sizeof(dbl_limits)/sizeof(struct dbl_limits);
			i++, lim++) {
			if ((id.mantissa2 == lim->ieee.mantissa2) &&
				(id.mantissa1 == lim->ieee.mantissa1) &&
				(id.exp == lim->ieee.exp)) {
				vd = lim->d;
				goto doneit;
			}
		}
		vd.exp = id.exp - IEEE_DBL_BIAS + VAX_DBL_BIAS;
		vd.mantissa1 = (id.mantissa1 >> 13);
		vd.mantissa2 = ((id.mantissa1 & MASK(13)) << 3) |
				(id.mantissa2 >> 29);
		vd.mantissa3 = (id.mantissa2 >> 13);
		vd.mantissa4 = (id.mantissa2 << 3);
	doneit:
		vd.sign = id.sign;
		*dp = *((double *)&vd);
		return (TRUE);
#endif vax

	case XDR_FREE:
		return (TRUE);
	}
	return (FALSE);
}
