#include <stdio.h>
#include "define.h"
#include "defs.h"
#include "extern.h"
/******************************************************************************
 *                                                                            *
 *  hash - hash a string by adding up it's characters, dividing by            *
 *         TABLE_SIZE, and returning the remainder                            *
 *                                                                            *
 *  calling syntax: integer = hash (string pointer)                           *
 *                                                                            *
 *  returns: integer                                                          *
 *                                                                            *
 *  globals used: TABLE_SIZE                                                  *
 *                                                                            *
 ******************************************************************************/

int hash (string)
char  *string;
{
   int i;                               /* for-loop counter                   */
   int len;                             /* length of string                   */
   int sum = 0;                         /* partial sum                        */

   len = strlen (string);
   for (i=1; i <= len; i++)
      sum += *(string++);               /* add in one char at a time          */

   return (sum % TABLE_SIZE);           /* return remainder                   */
}



/******************************************************************************
 *                                                                            *
 *  add - add a hash table element into tableh                                *
 *                                                                            *
 *  calling syntax: add (int, pointer to struct def_el)                       *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: tableh                                                      *
 *                                                                            *
 ******************************************************************************/

add (index, element)
int index;
struct def_el *element;

{
   if (tableh [index] == 0) {           /* if bucket empty put the first      */
      tableh [index] = element;         /* element in                         */
      element->next = NULL;             /* terminate next pointer             */
   }
   else {
      element->next = tableh [index];   /* else add to head of list           */
      tableh [index] = element;
   }
}



/******************************************************************************
 *                                                                            *
 *  lookup - given a possible cap name, find it's element in tableh, return   *
 *           null pointer if not found.                                       *
 *                                                                            *
 *  calling syntax: pointer to def_el = lookup (string pointer)               *
 *                                                                            *
 *  returns: struct def_el *                                                  *
 *                                                                            *
 *  globals used: tableh                                                      *
 *                                                                            *
 ******************************************************************************/

struct def_el *lookup (name)
char *name;
{
   int h;                               /* hash bucket number */
   struct def_el *hook;                 /* pointer to hook out element */

   h = hash (name);                     /* find bucket */
   hook = tableh [h];

   while (hook != NULL) {               /* run down list */
      if (strcmp (hook->name, name) == 0)   /* when names match */
         return (hook);                     /* return pointer */
      else
         hook = hook->next;             /* else keep going */
   }

   return (NULL);                       /* return NULL when element not found */
}



/******************************************************************************
 *                                                                            *
 *  build - build a hash table of def_el elements defined by the file         *
 *          "TINF*DEF_LIST"; every element in this list should have a         *
 *          name, a type, and the integer offset of that capability from      *
 *          from the start of it's particular section.                        *
 *                                                                            *
 *  calling syntax: build ()                                                  *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

build ()

{
   FILE *def_list_fp;                   /* list of recognized capabilities    */
   char name [CAP_LENGTH];              /* capability name                    */
   char type [CAP_LENGTH];              /* capability type                    */
   int h, offset;                       /* hash value, section offset         */
   struct def_el  *new_el;              /* definition element pointer         */
   char filename [FILE_LENGTH];     	/* name of object file                */

   if (def_list_qual)                                        		   	/*BJP*/
		sprintf (filename, "%s/%s", def_list_qual, DEF_LIST_FILE); 			/*BJP*/
   else                                                         			/*BJP*/
		sprintf (filename, "%s/%s", OBJECT_QUAL, DEF_LIST_FILE); 			/*BJP*/

   /* open the definition list file, if not possible forget the whole deal
    */
   if ((def_list_fp = fopen (filename, "r")) == NULL) {
      error (1, filename);
      exit ();
   }

   /* scan through defintion list file picking up name, type and the offset of
      this capability into its particular section.
    */
   while (fscanf (def_list_fp, "%s %s %d\n", name, type, &offset) != EOF) {

         /* allocate new hash table element, if no space exit porgram
         */
         if ((new_el = (struct def_el *)malloc (sizeof (struct def_el))) == 0)
            error (2);                         /* print error message, exit  */

         strcpy (new_el->name, name);          /* copy name into element */
         strcpy (new_el->type, type);             /* copy type into element */
         new_el->offset = offset;                 /* copy offset            */
         h = hash (name);                         /* hash element           */
         add (h, new_el);                         /* and link into table    */

   }
}
