#include <stdio.h>
#include "define.h"
#include "defs.h"
#include "extern.h"
/******************************************************************************
 *                                                                            *
 *  error - print an error message to the users terminal.                     *
 *                                                                            *
 *  calling syntax: error (int type, [char *], [char *]                       *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

error (type, s1, s2)
int type;
char *s1;
char *s2;

{
   fprintf (stderr, "tdc: ");                           /* common start       */
   switch (type) {                                      /* switch on type     */
      case 0:
         fprintf (stderr,
                  "Usage: tdc [-v] file [file ...]\n");
         exit(-1);                                  /* exit on incorrect use  */
      case 1:
         fprintf (stderr,
                  "error: open of %s failed.\n", s1);
         break;
      case 2:
         fprintf (stderr,
                  "error: unable to allocate space for hash table entry.\n");
         exit (-1);                                 /* exit on these errors   */
      case 3:
         fprintf (stderr,
                  "error: %s is not found in %s.\n", s1, s2);
         break;
      case 4:
         fprintf (stderr,
                  "error: illegal state parsing %s.\n", s1);
         break;
      case 5:
         fprintf (stderr,
                  "error: cap name %s not defined.\n", s1);
         break;
      case 6:
         fprintf (stderr,
                  "error: capability %s of undefined type.\n", s1);
         break;
      case 7:
         fprintf (stderr,
                  "error: buffer overun reading value for %s.\n", s1);
         break;
      case 8:
         fprintf (stderr,
                  "error: invalid numeric string as value for %s.\n", s1);
         break;
      case 9:
         fprintf (stderr,
                  "warning: EOF detected while reading terminal definition.\n");
         break;
      case 10:
         fprintf (stderr,
                 "error: unable to find \"use=\" terminal %s.\n", s1);
         break;
      case 11:
         fprintf (stderr,
                  "error: reading existing tdc object file %s.\n", s1);
         break;
      case 12:
         fprintf (stderr,
                  "error: link of %s failed.\n", s1);
         break;
      case 13:
         fprintf (stderr,
                  "error: attempt to link %s to itself.\n", s1);
         break;
      case 14:
         fprintf (stderr,
                  "error: no non-blank-containing name for terminal.\n");
         exit(-1);                                /* nothing to call terminal */
      default:
         fprintf (stderr,
                  "unrecognized error code.\n");
         break;

   } /* end switch */

} /* end error */



/******************************************************************************
 *                                                                            *
 *  clear -  clear any or all of the global data structures;                  *
 *              boolean_section,                                              *
 *              numeric_section,                                              *
 *              string_section.                                               *
 *                                                                            *
 *  calling syntax: clear (int level of clear desired)                        *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: boolean_section, numeric_section, string_section            *
 *                                                                            *
 ******************************************************************************/

clear (level)
int level;

{
   int i;                               /* index                              */

   switch (level) {                     /* switch on level                    */
      case 3:
         string_list_offset = 0;              /* init offset into string list */
         for (i = 0; i < NUM_OF_STR; i++)     /* init strings  "not defined"  */
            string_section [i] = UNDEFINED;
      case 2:
         for (i = 0; i < NUM_OF_NUM; i++)     /* init numerics "not defined"  */
            numeric_section [i] = UNDEFINED;
      case 1:
         for (i = 0; i < NUM_OF_BOOL; i++)    /* init boolean section FALSE   */
            boolean_section [i] = (char)FALSE;
   }
}
