#include <stdio.h>
#include "define.h"
#include "defs.h"

extern char *getenv();
extern int getopt();
extern int optind;
 
/******************************************************************************
 *                                                                            *
 *  Global Definitions                                                        *
 *                                                                            *
 ******************************************************************************/

struct def_el *tableh[TABLE_SIZE];      /* hash table                         */

short header_section [HEADER_SIZE];     /* ouput file header section          */
char term_name [NAME_LENGTH];           /* terminal name section              */
char boolean_section [NUM_OF_BOOL+2];   /* boolean defs section               */
short numeric_section [NUM_OF_NUM];     /* numeric defs section               */
short string_section [NUM_OF_STR];      /* offsets into string_list           */
char string_list [STRING_LIST_LENGTH];  /* null termn., binary strings        */
int string_list_offset;                 /* running offset into string list    */
char *object_qual;                      /* object qualifier from environ.     */
char *def_list_qual;                    /* def_list qualifier from environ.   */

char *CURRENT_BUF;                      /* pointer to current parse buffer    */
int BUF_OFFSET;                         /* current offset into buffer         */
int BUF_SIZE;                           /* size of this buffer                */
char EOB_FLAG = (char)-1;               /* end of buffer flag                 */

 
/******************************************************************************
 *                                                                            *
 *  main - main routine for tdc. Checks number of arguments, opens input      *
 *         file, builds hash tables for cap_name recognition, initializes     *
 *         numeric_section and string_section as undefined, runs through      *
 *         arguments either compiling every terminal entry in the input file, *
 *         or only compiling selected terminal entries, and writing these out.*
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

main (argc, argv)
int argc;
char *argv[];

{
   int i, j;                        /* integer indeces                        */
   char c;                          /* option character                       */
   char main_buf [PARSE_BUF_SIZE];  /* buffer for the primary terminal entry  */
   int main_buf_size = 0;           /* size of this primary entry             */
   FILE *termdef_fp;                /* termdef input file                     */
   char *use_vec[VEC_SIZE];         /* vector of "use=" specification values  */
   int num_use;                     /* number of "use=" specified             */
   int verbose = FALSE;             /* verbose flag                           */

   if ((c = getopt(argc, argv, "v")) != EOF) {   /* check for verbose option  */
      if ((argc < 3) || (c == '?'))              /* check option, argc        */
         error(0);                               /* error and exit            */
      verbose = TRUE;                            /* set verbose flag          */
   }
   else if (argc < 2)                            /* no option, check argc     */
      error(0);                                  /* error and exit            */
   else
      optind = 1;                                /* no option set index       */

   object_qual = getenv("TERMINFO");    /* pick up object qualifier from env  */
   def_list_qual = getenv("DEFLIST");   /* pick up def_list qualifier from env*/

   build ();                            /* build tables                       */

   clear (3);                           /* clear global data structs          */

   /* go through each of the input file arguments; pick out the terminal
      entries and put them in the parse buffer; parse them building the
      various output sections while so doing; write these sections to an
      output file under the correct qualifier and form the appropriate
      alias links.
    */
   for (i = optind; i < argc; i++) {

      /* open termdef input file                                              */
      if ((termdef_fp = fopen (argv [i], "r")) == NULL) {
         error (1, argv[i]);
         exit (-1);
      }

      while (termseek(0, termdef_fp) == TRUE) { /* position at next entry     */

         /* read entry into main_buf checking for "use=", placing the name of
            the use terminal in use_name and returning > 0 if this occurs.
          */
         if (num_use =
             fill_parse_buf (main_buf, &main_buf_size, termdef_fp, use_vec)) {
            /* process the use names in reverse order to ensure correct
               overlay of the attributes.
             */
            if (num_use > 1) {
               for (j = num_use-1; j > -1; j-- )
                  use(*(use_vec+j), argv [i], 0);     /* non-quick version    */
            }
            else
               use(*use_vec, argv[i], 1);             /* quick version        */
         }

         parse (main_buf, main_buf_size, verbose);    /* process this entry   */
         write_object ();                       /* output object file         */
         main_buf_size = 0;                     /* rest main buffer index     */
         clear (3);                             /* clear out main structs     */
      } /* end while */

   } /* end for */

} /* end main */

