#include <stdio.h>
#include <ctype.h>
#include "define.h"
#include "defs.h"
#include "extern.h"
#include "state.h"

extern struct def_el *lookup();

/******************************************************************************
 *                                                                            *
 *  set_buffer - set the global variables, CURRENT_BUF, BUF_SIZE, and reset   *
 *               the index BUF_OFFSET to zero.                                *
 *                                                                            *
 *  calling syntax: set_buffer (char * buffer to be parsed, int size)         *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: CURRENT_BUF, BUF_SIZE, BUF_OFFSET                           *
 *                                                                            *
 ******************************************************************************/

set_buffer (loc, size)
char *loc;
int size;

{
   CURRENT_BUF = loc;                   /* set current parse buffer           */
   BUF_SIZE = size;                     /* set current parse buffer size      */
   BUF_OFFSET = 0;                      /* reset current parse buffer offset  */
}



/******************************************************************************
 *                                                                            *
 *  buf_char - get the next character out of the buffer defined by the        *
 *             "set_buffer" routine and return the "END OF BUFFER" flag when  *
 *             ther end of the buffe rhas been reached                        *
 *                                                                            *
 *  calling syntax: char = buf_char ()                                        *
 *                                                                            *
 *  returns: char                                                             *
 *                                                                            *
 *  globals used: CURRENT_BUFFER, BUF_SIZE, BUF_OFFSET                        *
 *                                                                            *
 ******************************************************************************/

char
buf_char ()

{
#if DEBUG
printf("%%");
fflush(stdout);
#endif
   if (BUF_OFFSET == BUF_SIZE)                  /* we have reached end of buf */
      return (EOB_FLAG);                        /* return flag value          */
   else
      return (CURRENT_BUF [BUF_OFFSET++]);      /* return next character      */
}



/******************************************************************************
 *                                                                            *
 *  skip_cap - skip past a paticular capability in the current parse buffer   *
 *                                                                            *
 *  calling syntax: skip_cap ()                                               *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: CURRENT_BUF, BUF_OFFSET                                     *
 *                                                                            *
 ******************************************************************************/

skip_cap ()

{
   char in_char;    /* input char */
   int es_flag = FALSE; /* escape flag */

   if (CURRENT_BUF [BUF_OFFSET-1] == ',')       /* may be at the end of a     */
      return;                                   /* capability def already     */
   else {
      while (((in_char = buf_char ()) != ',') ||    /* otherwise skip to next */
             (es_flag)) {                           /* comma                  */
         if (in_char == '\\') {                     /* check for backslash    */
            if (es_flag)
               es_flag = FALSE;                     /* already saw backslash  */
            else
               es_flag = TRUE;                      /* set escape flag        */
         }
         else
            es_flag = FALSE;                        /* reset escape flag      */
      } /* end while*/
   } /* end if */
} /* end skip_cap */



/******************************************************************************
 *                                                                            *
 *  read_def_val - read characters from the current parse buffer up to the    *
 *                 next comma and place them in the buffer pointed to by the  *
 *                 argument pointer. The "\" character hides commas. This     *
 *                 routine is called by both num and string so a type         *
 *                 parameter is passed and checking done for non_digits in    *
 *                 case of a numeric cap value being read                     *
 *                                                                            *
 *  calling syntax: read_def_val (pointer to buffer,                          *
 *                                int type of cap,                            *
 *                                char * name of cap being eval'ed)           *
 *                                                                            *
 *  returns: int (FALSE = error reading numeric, TRUE = okay)                 *
 *                                                                            *
 *  globals used:                                                             *
 *                                                                            *
 ******************************************************************************/

int
read_def_val (place, type, name)
char *place;
int type;
char *name;

{
   int index = 0;                   /* index into buffer                      */
   int es_flag = 0;                 /* escape sequence flag                   */
   int result = TRUE;               /* holder for returned result             */
   char in_char;                    /* input chracter                         */

   while ((((in_char = buf_char ()) != ',') || (es_flag)) &&    /* until comma*/
          (in_char != EOB_FLAG)) {                      /* or buffer overun   */

      if (type == NUM) {                                /* numeric definition */
         if ( ! isdigit (in_char))
            result = FALSE;                             /* invalid numeric    */
      }
      else if ((in_char == BACKSLASH) && ( ! es_flag))  /* check for escape   */
         es_flag = 1;                                   /* set escape flag    */
      else
         es_flag = 0;                                   /* reset escape flag  */

      *(place+index) = in_char;                         /* put in buffer      */
      index++;
   }

   if (in_char == EOB_FLAG)                             /* if end of buffer   */
      error (7, name);                                  /* signal error       */

   *(place+index) = '\0';                           /* null terminate string  */
   return (result);
}



/******************************************************************************
 *                                                                            *
 *  bool - procedure called when boolean cap has been parsed in buffer: flag  *
 *         appropriate filed in boolean_section.                              *
 *                                                                            *
 *  calling syntax: bool (pointer to def_el, int on/off flag)                 *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: boolean_section                                             *
 *                                                                            *
 ******************************************************************************/


bool (elm, on_flag)
struct def_el *elm;
int on_flag;

{
   if ( ! on_flag)
      boolean_section [elm->offset] = FALSE;         /* flag false            */
   else
      boolean_section [elm->offset] = TRUE;          /* flag true             */
}



/******************************************************************************
 *                                                                            *
 *  num - procedure called when numeric capability has been parsed in the     *
 *        buffer: read value from parse buffer, convert to integer and put in *
 *        appropriate filed in numeric_section.                               *
 *                                                                            *
 *  calling syntax: num (pointer to def_el,                                   *
 *                       int on/off flag,                                     *
 *                       char * name of capability)                           *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: numeric_section                                             *
 *                                                                            *
 ******************************************************************************/

num (elm, on_flag, name)
struct def_el *elm;
int on_flag;
char *name;

{
   char num_str [NUM_LENGTH];       /* string value of this capability        */

   if (! on_flag) {                                     /* disable this cap   */
      numeric_section [elm->offset] = UNDEFINED;
      skip_cap ();                                      /* skip its value     */
   }
   else if ( ! read_def_val (num_str, NUM, name)) {
      error (8, name);                                  /* invalid numeric    */
      return;                                           /* def already read   */
   }
   else {                           /* put integer value into numeric_section */
#if DEBUG
printf("num() => num_str = %s\n", num_str);
fflush(stdout);
#endif
      numeric_section [elm->offset] = (short)atoi (num_str);
   }
#if DEBUG
printf("num() <=\n");
fflush(stdout);
#endif
}



/******************************************************************************
 *                                                                            *
 *  string - procedure called when string capability has been recognized in   *
 *           the parse buffer: read value from buffer, translate value into   *
 *           its binary equivalent, add this binary string to the end of the  *
 *           string list, and enter offset of this string in string list into *
 *           string_section.                                                  *
 *                                                                            *
 *  calling syntax: string (pointer to def_el,                                *
 *                          int on/off flag,                                  *
 *                          char * name of capability)                        *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: string_section, string_list, string_list_offset             *
 *                                                                            *
 ******************************************************************************/

string (elm, on_flag, name)
struct def_el *elm;
int on_flag;
char *name;

{
   char string_value [STRING_LENGTH];   /* source string                      */
   char trans_string [STRING_LENGTH];   /* object string                      */

   if (! on_flag) {                                     /* negate cap         */
      string_section [elm->offset] = UNDEFINED;
      skip_cap ();                                      /* skip its value     */
   }
   else {                                                   /* string ok      */
      read_def_val (string_value, STR, name);               /* read value     */
      trans (string_value, trans_string);                   /* translate      */
      string_section [elm->offset] = string_list_offset;    /* enter offset   */
      strcpy (string_list+string_list_offset, trans_string); /* copy          */
      string_list_offset += strlen (trans_string) +1;       /* update offset  */
   }

}



/******************************************************************************
 *                                                                            *
 *  name - routine called when a terminal name string has been parsed from    *
 *         the parse buffer: fills in term_name string from the string        *
 *         parameter.                                                         *
 *                                                                            *
 *  calling syntax: name (pointer to string, character count)                 *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: term_name                                                   *
 *                                                                            *
 ******************************************************************************/

name (strang, chr_cnt, flag)
char *strang;                       /* pointer to input buffer                */
int chr_cnt;                        /* character count of name string         */
int flag;                           /* print-name flag                        */

{
  strncpy (term_name, strang, chr_cnt);     /* fill in terminal name field    */
  strncpy (term_name+chr_cnt, "\0", 1);     /* null terminate this string     */
  if (flag)
     fprintf (stdout, "\n%s\n", term_name); /* print the name of entry to     */
                                            /* stdout for tracking            */
} /* end name */



/******************************************************************************
 *                                                                            *
 *  cap_found - procedure called when a capability name has been recognized   *
 *              by the state machine in "parse": check name for the '@'       *
 *              character on the end indicating this cap is to be deleted and *
 *              null terminate the name, check for a '.' in front of the name *
 *              to see if this entry is commented out, look up name in hash   *
 *              table and determine type of capability, call "bool", "num",   *
 *              or "string" depending on this type.                           *
 *                                                                            *
 *  calling syntax: cap_found (char * potential capability name,              *
 *                             int count of chars in this name)               *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

cap_found (str, count)
char *str;
int count;

{
   struct def_el *el;       /* pointer to element in hash table               */
   int turn_on = TRUE;      /* enable/disable cap flag, TRUE = enable         */

   if (*(str+count-1) == '@') {     /* look for negate modifier               */
      *(str+count-1) = '\0';        /* null terminate over the negate modifier*/
      turn_on = FALSE;              /* FALSE = negate capability              */
   }
   else
      *(str+count) = '\0';          /* null terminate                         */

#if DEBUG
printf ("cap found %s.\n", str);
#endif

   if (*str != '.') {                   /* if not commented out               */

      if ((el = lookup (str)) == (struct def_el *)NULL) { /* look up in table */
         error (5, str);                            /* not there              */
         skip_cap ();                               /* skip this capability   */
      }
      else {
         if (strcmp (el->type, BOOL_TYPE) == 0)     /* call type dependent    */
            bool (el, turn_on);                     /* procedures             */
         else if (strcmp (el->type, NUM_TYPE) == 0)
            num (el, turn_on, str);
         else if (strcmp (el->type, STR_TYPE) == 0)
            string (el, turn_on, str);
         else {                                     /* we looked up an element*/
            error (6,str);                          /* with undefined type    */
            skip_cap ();                            /* skip this capability   */
         }
      }
   }
   else                                 /* if commented out                   */
      skip_cap ();                                  /* skip it                */
#if DEBUG
printf("cap_found: bye\n" );
#endif
}



/******************************************************************************
 *                                                                            *
 *  parse - given a file pointer to a buffer containing a terminal defintion  *
 *          and the size of this definition, process this using the state     *
 *          machine "state" calling "name" when the name has been extracted   *
 *          and "cap_found" every time a possible capability name is          *
 *          extracted. To make things easier this routine calls "set_buffer"  *
 *          to make the buffer containing the terminal definiton, the size of *
 *          of this definition, and the current offset into this definition,  *
 *          global. This to allow routines called by routines called by       *
 *          "parse" (namely "read_def_val") to extract the values for numeric *
 *          and string capabilities without having to pass this info all the  *
 *          way down. Only three routines manipulate these globals:           *
 *          "set_buffer", "buf_char", and "skip_cap".                         *
 *                                                                            *
 *  calling syntax: parse (char * to buffer containing terminal def,          *
 *                         int size of terminal def)                          *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: state                                                       *
 *                                                                            *
 ******************************************************************************/

parse (p_buf, size, flag)
char *p_buf;
int size;
int flag;

{
   int old_state = 0;                       /* old state holder               */
   int new_state;                           /* next state holder              */
   int str_len = 0;                         /* length of saved input stream   */
   char in_char;                            /* input char                     */
   char buff [MAX_LINE_LENGTH];             /* input stream buffer            */
 
   set_buffer (p_buf, size);                /* set parse buf globals          */

   while ((in_char = buf_char ()) != EOB_FLAG) {        /* until end of buf   */
#if DEBUG
printf("%c", in_char);
fflush(stdout);
#endif
      new_state = state [old_state] [(int)in_char];     /* figure new state   */
      switch (new_state)  {                             /* switch on it       */
         case -1:
            error (4, buff);
            break;
         case 0:
         case 3:
            buff [str_len++] = in_char;     /* collect input                  */
            old_state = new_state;
            break;
         case 1:                            /* skip input                     */
         case 4:
         case 5:
            old_state = new_state;
            break;
         case 2:                            /* termdef name recognized        */
            name (buff,str_len, flag);      /* process it                     */
            str_len = 0;                    /* reset string length            */
            old_state = 3;                  /* get first cap name             */
            break;
         case 6:                            /* cap name recognized            */
            cap_found (buff, str_len);      /* go process it      */
#if DEBUG
write(1,"parse: after cap_found\n", strlen( "parse: after cap_found\n" ) );
#endif
            str_len = 0;                    /* reset string length            */
            old_state = 3;                  /* get next cap name              */
            break;
      }
   }
}
