#include <stdio.h>
#include "define.h"

extern char *malloc();

/******************************************************************************
 *                                                                            *
 *  fill_parse_buf - given a file pointer pointing to the begining of a       *
 *                   terminal defintion, and a pointer to a buffer, extract   *
 *                   1 (one) terminal definition from the file and place it   *
 *                   into the buffer, returning the size (in bytes) of the    *
 *                   definition in "size". If a "use=" string was included in *
 *                   the definition return the condition TRUE and the name of *
 *                   the def to be used in "name".                            *
 *                                                                            *
 *  calling syntax: int = fill_parse_buff (char * buffer to fill,             *
 *                                        int * size of defintion,            *
 *                                        FILE * file to read from,           *
 *                                        char ** vector of use names)        *
 *                                                                            *
 *  returns: int number of use names found                                    *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

int fill_parse_buf (buffer, size, fp, names)
char *buffer;
int *size;
FILE *fp;
char **names;

{
   char in_char;                            /* input character                */
   int sstate = 0;                          /* state in lookin for "use=..."  */
   int index = 0;                           /* index for use name             */
   int result = 0;                          /* result no. of "use="'s found   */
   char *gen;                               /* pointer to "use=" value        */

   do {         /* do until end of this terminal definition                   */

    /* fill parse_buf a line at a time looking for "use=....."                */
      while (((in_char = getc (fp)) != EOF) &&
             (in_char != '\n')) {

         switch (sstate) {                 /* switch on "sstate"              */
            case 0:
               if (in_char == 'u')          /* first letter ?                 */
                  sstate = 1;               /*    go to next state            */
               break;
            case 1:
               if (in_char == 's')          /* second letter ?                */
                  sstate = 2;               /*    next state                  */
               else
                  sstate = 0;               /* begin again                    */
               break;
            case 2:
               if (in_char == 'e')          /* third letter ?                 */
                  sstate = 3;               /*    next state                  */
               else
                  sstate = 0;               /* begin again                    */
               break;
            case 3:
               if (in_char == '=') {        /* equal sign                     */
                  gen = malloc(FILE_LENGTH);
                  sstate = 4;               /*   acceptance state             */
               }
               else
                  sstate = 0;               /* begin again                    */
               break;
            case 4:
               if (in_char == ',') {        /* end of "use=...."              */
                  gen [index] = '\0';       /* null terminate use name        */
                  index = 0;                /* reset index                    */
                  *names++ = gen;           /* add to vector                  */
                  *names = NULL;            /* terminate vector (in case)     */
                  result++;                 /* increment result               */
                  sstate = 0;               /* reset state                    */
               }
               else
                  gen [index++] = in_char;  /* put away term name for use     */
               break;
         }

         buffer [(*size)++] = in_char;              /* put in buffer          */
      }

      if (in_char != EOF) {
         buffer [(*size)++] = in_char;              /* put in parse buffer    */
         in_char = getc(fp);                        /* check next character   */

         if (in_char != EOF)                        /* check for EOF          */
            ungetc(in_char, fp);                    /* put it back if not     */
      }

   } while ((in_char == ' ') ||     /* first char after new line non-blank or */
            (in_char == '\t'));     /* non-tab indicates start of next def    */
 
   return (result);
}



/******************************************************************************
 *                                                                            *
 *  termseek - look through the termdef file for either the next terminal     *
 *             definition encountered (null argument) or a terminal definit-  *
 *             ion matching with a name matching that of the argument,        *
 *             positon termdef_fp at the begining of that defintion, and      *
 *             return TRUE. If the desired definition cannot be found return  *
 *             FALSE.                                                         *
 *                                                                            *
 *  calling syntax: termseek (char * to desired term's name or null,          *
 *                            FILE * file to look in)                         *
 *                                                                            *
 *  returns: int (TRUE/FALSE)                                                 *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

int termseek (tern, seek_fp)
char *tern;
FILE *seek_fp;

{
   int nindex = 0;                  /* name index                             */
   char in_char, look_char;         /* input chars                            */
   long backcount;                  /* negative position count for fseek call */
   char maybe_name [NAME_LENGTH];   /* name matching buffer                   */
   char trash [5*MAX_LINE_LENGTH];  /* throw away line                        */
 
   /* look through termdef file until EOF                                     */
   while ((in_char = getc (seek_fp)) != EOF) {
 
      if ((in_char == '#') || (in_char == ' ') ||
          (in_char == '\n')) {                      /* comment or junk ?      */
         fgets (trash, 5*MAX_LINE_LENGTH, seek_fp); /* waste comment          */
      }

      else if (tern == 0) {                         /* no particular term     */
         ungetc (in_char, seek_fp);                 /* put char back          */
         return (TRUE);                             /* return TRUE = found    */
      }

      /* a particular terminal is sought so scan the name line for a match on */
      /* tern, keeping in mind we are at the terminal name string already     */
      else {

         maybe_name [nindex++] = in_char;           /* already have first char*/
         backcount = -1;                            /* init backwards count   */
 
         do {

            /* fill in maybe_name up to first delimeter                       */
            while (((look_char = getc (seek_fp)) != '|') &&     /* bar or     */
                   (look_char != ',')) {                        /* comma      */

               if (look_char == '\n') {             /* if newline, skip it    */
                  while ((look_char = getc(seek_fp)) == ' ')    /* and balnks */
                     backcount--;
               }

               maybe_name [nindex++] = look_char;
               backcount--;                         /* decrement back count   */
            }
            backcount--;                            /* also for delim         */
            maybe_name [nindex] = '\0';             /* null terminate         */
 
            if (strcmp (tern, maybe_name) == 0) {   /* compare                */
               fseek (seek_fp, backcount, 1);       /* if good repos fp and   */
               return (TRUE);                       /* return TRUE = found    */
            }
            else
               nindex = 0;                          /* zero counter           */

         } while (look_char != ',');                /* do to end of names sec.*/

         /* this entry is a dud, so skip it                                   */
         fgets (trash, 5*MAX_LINE_LENGTH, seek_fp);     /* goto end of line   */
         while (((in_char = getc (seek_fp)) == ' ') ||  /* first char of line */
                (in_char == '\t')) {                    /* blank or tab then  */
            fgets (trash, 5*MAX_LINE_LENGTH, seek_fp);  /* burn line          */
         }

         if (in_char != EOF)                        /* if not at end of file  */
            ungetc (in_char, seek_fp);              /* put first char back    */

      } /* end if */

   } /* end while */

   return (FALSE);      /* if we have reached EOF return FALSE = not found    */

} /* end termseek */
