#include <ctype.h>
#include "define.h"
#if DEBUG
#include <stdio.h>
#endif
/******************************************************************************
 *                                                                            *
 *  octal - convert an ASCII string of digits representing an octal number    *
 *          into the number it represnts and pass this back as a one byte     *
 *          character. When this routine is called we have allready seen one  *
 *          character and the char * pointed to by duo is pointing one past   *
 *          the first digit we want to grab.                                  *
 *                                                                            *
 *  calling syntax: octal (pointer to char *)                                 *
 *                                                                            *
 *  returns: char                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

char octal (duo)
char **duo;

{
   char val [OCT_LENGTH];       /* buffer for octal number                    */
   int octits = 1;              /* number of octal digits, initally one       */
   int back;                    /* back counter                               */
   int power = 0;               /* powers of eight counter                    */
   int result = 0;              /* end result                                 */
   static int pow [OCT_LENGTH] = {1, 8, 64};    /* powers of eight array      */

#if DEBUG
printf( "octal: enter\n" );
fflush( stdout );
#endif

   val [0] = *((*duo)-1);                           /* we know we have one    */
                                                    /* total up the rest      */
   while ((octits < OCT_LENGTH) && isdigit (**duo)) {
      val [octits] = **duo;
      (*duo)++;
      octits++;
   }

   for (back = octits-1; back >= 0; back--, power++)    /* convert to octal   */
      result += (((int)val [back])-060) * pow [power];

#if DEBUG
printf( "octal: returns %o\n", result );
fflush( stdout );
#endif

   return ((char) result);                          /* return result as char  */
} /* end octal */



/******************************************************************************
 *                                                                            *
 *  escape - translate a termdef source sequence that began with a "\" into   *
 *           the correct object sequence. Pointers to char *'s are used as    *
 *           arguments to allow escape to change the contents of the pointers *
 *           to the source and object strings so that the processing of these *
 *           strings may proceed in stream fashion after returning.           *
 *                                                                            *
 *  calling syntax: escape (pointer to char * pointing to source string,      *
 *                            "  "  "   "   "   "  "   "  object   "    )     *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

escape (espoint, eopoint)
char **espoint;                             /* pointer to pointer to source   */
char **eopoint;                             /* pointer to pointer to object   */

{
   char echar;

   echar = *(*espoint)++;                   /* grab char after "\"            */
#if DEBUG
   printf( "escape: echar is '%c'\n", echar );
   fflush( stdout );
#endif
   switch (echar) {                         /* switch on it                   */
      case 'b':                             /* backspace                      */
         *(*eopoint)++ = BACKSPACE;
         break;
      case 'e':                             /* two forms of escape            */
      case 'E':
         *(*eopoint)++ = ESCAPE;
         break;                             /* formfeed                       */
      case 'f':
         *(*eopoint)++ = FORMFEED;
         break;
      case 'l':                             /* linefeed                       */
      case 'n':                             /* newline                        */
         *(*eopoint)++ = '\n';
         break;
      case 'r':                             /* return                         */
         *(*eopoint)++ = RETURN;
         break;
      case 's':                             /* space                          */
         *(*eopoint)++ = ' ';
         break;
      case 't':                             /* tab                            */
         *(*eopoint)++ = TAB;
         break;
      case '0':
#if DEBUG
printf( "escape: case '0'\n" );
fflush( stdout );
#endif
         if (! isdigit (*(*espoint+1))) {   /* may be fake null = '200        */
            *(*eopoint)++ = FAKENULL;
            break;
         }                                  /* else is octal                  */
#if DEBUG
printf( "escape: case '0': falling through\n" );
fflush( stdout );
#endif
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':                             /* start of octal string          */
         *(*eopoint)++ = octal (espoint);   /* call octal to convert          */
#if DEBUG
printf( "escape: back from octal\n" );
fflush( stdout );
#endif
         break;
      default:                              /* plain character                */
         *(*eopoint)++ = echar;
         break;
   } /* end switch */
} /* end escape */



/******************************************************************************
 *                                                                            *
 *  trans - given an ASCII source string translate it (in a termdef specific  *
 *          manner) into a binary object string.                              *
 *                                                                            *
 *  calling syntax: trans (pointer to source string, pointer to object string)*
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

trans (source, object)
char *source;
char *object;

{
   char schar;

   do {
      schar = *source++;                    /* grab source char               */
      switch (schar) {                      /* switch on it                   */
         case '\\':                         /* escape type sequence           */
            escape (&source, &object);      /* call escape                    */
            break;
         case '^':                          /* possible control character     */
            if (*source=='?') {             /* special case DEL               */
               *object++ = '\177';
               ++source;
            }
            else {                          /* regular control character      */
               *object++ = *(source++) & LOW5; /* mask off high three bits    */
            }
            break;
         case '\n':
            while (*source == ' ')          /* skip blanks after newlines     */
               source++;
            break;                          /* and skip the newline           */
         default:                           /* otherwise just copy char       */
            *object++ = schar;
            break;
      } /* end switch */
   } while (schar != '\0');                 /* up until null terminator       */
} /* end trans */
