#include <stdio.h>
#include "define.h"
#include "defs.h"
#include "extern.h"
/******************************************************************************
 *                                                                            *
 *  read_object - given a file pointer to an existing tdc output file, read   *
 *                this file into the global data structures;                  *
 *                "boolean_section", "numeric_section", "string_section", and *
 *                "string_list".                                              *
 *                                                                            *
 *  calling syntax: read_object (FILE * object file to read,                  *
 *                               char * name of definition being read)        *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: boolean_section, numeric_section, string_section,           *
 *                string_list, string_list_offset                             *
 *                                                                            *
 ******************************************************************************/

read_object (ob_fp, exist_name)
FILE *ob_fp;
char *exist_name;

{
   int i;                               /* incremental counter                */
   short head [HEADER_SIZE];            /* use obj file header                */
   char tname [NAME_LENGTH];            /* use obj file name                  */
   char chead [CHEAD];                  /* header conversion array            */
   char cnum [CNUM];                    /* numeric section conversion array   */
   char cstring [CSTRING];              /* string section conversion array    */

   /* read in each of the sections, header, name, boolean, numeric, string,   */
   /* and string list.                                                        */

   if (CHEAD != fread (chead, sizeof (char), CHEAD, ob_fp)) {   /* header     */
      error (11, exist_name);
      return;
   }
   else 
     stoi (chead, head, HEADER_SIZE);           /* convert to internal rep*/

   if (head [1] != fread (tname, sizeof (char), head [1], ob_fp)) { /* name   */
      error (11, exist_name);
      return;
   }

   /* read in boolean section being carefull not to read too far if this files
      boolean section is shorter than ours, or read too short if this files
      boolean section is larger tahn ours
    */
   head [2] = (head [2] < NUM_OF_BOOL + 2) ? head [2] : NUM_OF_BOOL+2;
   if (head [2] != fread (boolean_section, sizeof (char), head [2], ob_fp)) {
      error (11, exist_name);
      clear (1);                                    /* clear boolean section  */
      return;                                       /* quit to save damage    */
   }

   for (i = 0; i < head [2]-(NUM_OF_BOOL+2); i++)   /* get remaining chars if */
      getc (ob_fp);                                 /* there are any          */

   /* read in numeric section being carefull not to read too far if this files
      numeric section is shorter than ours, or read short if this files numeric
      section is larger than ours
    */
   head [3] = (head [3] < NUM_OF_NUM) ? head [3] : NUM_OF_NUM;
   if (2*head [3] != fread (cnum, sizeof (char), 2*head [3], ob_fp)) {
      error (11, exist_name);
      clear (2);                                    /* clear nums and bools   */
      return;                                       /* quit to save damage    */
   }
   else                                             /* if read worked         */
      stoi (cnum, numeric_section, head [3]);       /* convert to internal rep*/

   for (i=0; i < (2*head [3] - CNUM); i++)          /* position to next       */
      getc (ob_fp);                                 /* section                */

   /* read in string section being carefull not to read too far if this files
      string section is shorter than ours, or read short if this file's string
      section is larger than ours
    */
   head [4] = (head [4] < NUM_OF_STR) ? head [4] : NUM_OF_STR;
   if (2*head [4] != fread (cstring, sizeof (char), 2*head [4], ob_fp)) {
      error (11, exist_name);
      clear (3);                                /* clear nums, bools, strings */
      return;                                   /* quit to save on damage     */
   }
   else                                                     /* if read worked */
      stoi (cstring, string_section, head [4]);             /* convert        */

   for (i=0; i < (2*head[4]-CSTRING); i++)              /* position to start  */
      getc (ob_fp);                                     /* of string list     */

   /* read in string list, don't worry about size cause if we don't use some
      strings its no hanging matter
    */
   if (head [5] != fread (string_list, sizeof (char), head [5], ob_fp)) {
      error (11, exist_name);
      clear (3);                                        /* clear everything   */
      return;                                           /* blow away          */
   }

   string_list_offset = head [5];       /* want to add to end of string_list  */
}



/******************************************************************************
 *                                                                            *
 *  use - given the name of a terminal definition be to used in building      *
 *        another terminal definition, and given the name of the file in      *
 *        which to look for it; first try finding an existing tdc object file *
 *        corresponding to the defintion to be used. If one exists use it by  *
 *        reading its contents into the global data structures to be built.   *
 *        If one doesn't exist, look for the entry to be used in the current  *
 *        file. If it is found here read it in, parse it, and return. If it   *
 *        it can't be found in the current source file look in the file named *
 *        by DEFAULT_NAME and treat the same as before.                       *
 *                                                                            *
 *  calling syntax: (char * name of def to use,                               *
 *                   char * name of file to look in)                          *
 *                                                                            *
 *  returns: void                                                             *
 *                                                                            *
 *  globals used: none                                                        *
 *                                                                            *
 ******************************************************************************/

use (use_name, termdef_name, quick_flag)
char *use_name;
char *termdef_name;
int quick_flag;

{
   int i;                       /* integer index                              */
   char filename [FILE_LENGTH]; /* filename to look for existing object under */
   FILE *therefile_fp;          /* fp to existing object file                 */
   FILE *samefile_fp;           /* alternate fp to termdef source file        */
   FILE *default_fp;            /* fp to default termdef source file          */
   char *new_vec[VEC_SIZE];         /* nested use vector area                 */
   int new_num;                     /* number of new "use=" specifications    */
   char use_buf [PARSE_BUF_SIZE];   /* buffer for use term entry              */
   int use_buf_size = 0;            /* size of use term entry                 */
   int len;                         /* length of object qualifier/directory   */

#if DEBUG
   printf ("\nuse called: use terminal %s\n", use_name);
#endif

   /* if the quick form is usable (i.e. no multiple "use="s around) do it     */
   if (quick_flag) {

      /* create file name from use_name                                       */
      if (object_qual)
         sprintf(filename, "%s/%c/%s", object_qual, *use_name, use_name);
      else
         sprintf(filename, "%s/%c/%s", OBJECT_QUAL, *use_name, use_name);

#if DEBUG
      printf ("looking for object under %s\n", filename);
#endif

      /* try opening the above created file, success means there is an existing
         object file for the use terminal
       */
      if ((therefile_fp = fopen (filename, "r")) != NULL) {
#if DEBUG
         printf ("reading existing object file.\n");
#endif
         read_object (therefile_fp, filename);              /* read this file */
         fclose (therefile_fp);                             /* close this file*/
         return;                                            /* return         */
      }
   }

   /* otherwise re-open the termdef input file and
      look for the use terminal entry
    */
   if ((samefile_fp = fopen (termdef_name, "r")) == NULL) {
      error (1, termdef_name);                              /* open error     */
      exit(-1);
   }

   else if (termseek (use_name, samefile_fp)) {             /* entry found    */
      /* read this entry into use_buf, check for nested use call, return use
         terminal name(s)in new_vec and number of names if so.
       */
#if DEBUG
      printf ("processing source definition from %s.\n", termdef_name);
#endif
      if (new_num =
          fill_parse_buf (use_buf, &use_buf_size, samefile_fp, new_vec)) {
         /* process the use names in reverse order to preserve overlay of
            attributes
          */
         if (new_num > 1) {
            for (i = new_num-1; i > -1; i--)
               use(*(new_vec+i), termdef_name, 0);      /* non-quick          */
         }
         else
            use(*new_vec, termdef_name, 1);             /* quick version      */
      }

      parse (use_buf, use_buf_size, 0);         /* parse use buffer, no names */
      fclose (samefile_fp);                             /* close termdef file */
   }
 
   /* check if the default file is the same as the one we've been working with
    */
   else if (! strcmp(termdef_name, DEFAULT_NAME)) {
      error(10, use_name);                              /* no entry found     */
      fclose(samefile_fp);                              /* close open file    */
   }
 
   /* otherwise open the default termdef input file and
      look for the use terminal entry
    */
   else if ((default_fp = fopen (DEFAULT_NAME, "r")) == NULL) {
      error (1, DEFAULT_NAME);                              /* open error     */
      fclose (samefile_fp);                              /* close source file */
   }

   else if (termseek (use_name, default_fp)) {              /* entry found    */
      /* read this entry into use_buf, check for nested use call, return use
         terminal name(s) in new_vec and number of names if so.
       */
#if DEBUG
      printf ("processing source definition from %s.\n", DEFAULT_NAME);
#endif
      if (new_num =
          fill_parse_buf (use_buf, &use_buf_size, default_fp, new_vec)) {
         /* process the use names in reverse order to preserve overlay of
            attributes
          */
         if (new_num > 1) {
            for (i = new_num-1; i > -1; i--)
               use(*(new_vec+i), termdef_name, 0);      /* non-quick          */
         }
         else
            use(*new_vec, termdef_name, 1);             /* quick version      */
      }

      parse (use_buf, use_buf_size, 0);         /* parse use buffer, no names */
      fclose (samefile_fp);                            /* close termdef files */
      fclose (default_fp);
   }

   else {                                      /* no entry nowhere            */
      error (10, use_name);                    /* write error message         */
      fclose (samefile_fp);                    /* close source files          */
      fclose (default_fp);
   }
}
