#include <stdio.h>
#include <errno.h>
#include "define.h"
#include "defs.h"
#include "extern.h"
/******************************************************************************
 *                                                                            *
 *  write_object - open/create an unblocked file under the TERMFILE account/  *
 *                 qualifier with a file name specified by the first section  *
 *                 of the term_name string; calculate and fill in the         *
 *                 fields of header_section; and ouput in turn:               *
 *                 header_section, term_name,  boolean_section (with padding),*
 *                 numeric_section, string_section, and string_list.          *
 *                                                                            *
 *  calling syntax: write_object ()                                           *
 *                                                                            *
 *  returns: null                                                             *
 *                                                                            *
 *  globals used: header_section, term_name, boolean_section, numeric_section,*
 *                string_section, string_list, string_list_offset             *
 *                                                                            *
 ******************************************************************************/

write_object ()

{
   char filename [FILE_LENGTH];     /* name of object file                    */
   char linkname [FILE_LENGTH];     /* name of synonomous link'ed files       */
   char dirname [FILE_LENGTH];	    /* directory name                         */
   char temp [FILE_LENGTH];         /* temp holder for file name creation     */
   char systring [MAX_LINE_LENGTH]; /* line for system "mkdir" call           */
   char chead [CHEAD];              /* conversion array for header section    */
   char cnum [CNUM];                /* conversion array for numeric section   */
   char cstring [CSTRING];          /* conversion array for string section    */
   FILE *output_fp;                 /* output file pointer                    */
   int blank_flag;                  /* blank encountered in file name         */
   int index = 0;                   /* index into output file name buffer     */
   int lindex = 0;                  /* link name index                        */
   int pad = 0;                     /* index, boolean pad factor              */
   extern int errno;                /* external error number                  */

   if (object_qual)                                   /* if qual in environ   */
      strcpy (dirname, object_qual);                  /* copy into dirname    */
   else
      strcpy (dirname, OBJECT_QUAL);                  /* else use default     */

   /* create file name from first section of terminal name string             */
   do {
      blank_flag = FALSE;	
      while ((term_name[index] != '|') && (term_name[index] != '\0')) {
         if (term_name[index] == ' ')
            blank_flag = TRUE;
         temp[index] = term_name[index];
         ++index;
      }
      if (term_name[index] == '\0')
         break;
      else if (blank_flag)
	++index;
   } while (blank_flag);

   if (blank_flag && (term_name[index] == '\0'))
	error(14);

   temp[index] = '\0';                                  /* null terminate     */

   sprintf (filename, "%s/%c/%s", dirname, temp[0], temp);/* create filename  */

   /* create/open a file of this name                                         */
   if ((output_fp = fopen (filename, "w")) == NULL) {

      if (errno == ENOENT) {                        /* no directory           */

#if DEBUG
fprintf(stderr, "creating directory: %s/%c\n", dirname, temp[0]);
#endif

         /* set up string for system call to mkdir                            */
         sprintf (systring, "mkdir %s/%c", dirname, temp[0]);
         system (systring);
         output_fp = fopen (filename, "w");         /* create file            */
      }
      else {
         error (1, filename);
         return;
      }
   }

#if DEBUG
fprintf(stderr, "output file: %s opened\n", filename);
#endif

   /* calculate values for header section                                     */
   header_section [0] = MAGIC;                  /* magic number               */
   header_section [1] = strlen (term_name)+1;   /* size (chars) name string   */

   /* figure padding needed for boolean_section considering it doesn't start  */
   /* on an even word boundary                                                */
   if (pad = ((NUM_OF_BOOL+header_section[1]) % WS))/* if remaider exists     */
      pad = WS-pad;                                 /* sub from bytes/word    */

   header_section [2] = NUM_OF_BOOL+pad;    /* size (chars) boolean section   */
   header_section [3] = NUM_OF_NUM;         /* size (short int) numeric sec.  */
   header_section [4] = NUM_OF_STR;         /* size (short int) string offsets*/
   header_section [5] = string_list_offset; /* size (chars) string table      */

   itos (header_section, chead, HEADER_SIZE);   /* convert to external two    */
   itos (numeric_section, cnum, NUM_OF_NUM);    /* byte-reversed rep.         */
   itos (string_section, cstring, NUM_OF_STR);

   /* write all this out                                                      */
   fwrite (chead, sizeof (char), CHEAD, output_fp);
   fwrite (term_name, sizeof (char), header_section [1], output_fp);
   fwrite (boolean_section, sizeof (char), header_section [2], output_fp);
   fwrite (cnum, sizeof (char), CNUM, output_fp);
   fwrite (cstring, sizeof (char), CSTRING, output_fp);
   fwrite (string_list, sizeof (char), string_list_offset, output_fp);

   /* close output file                                                       */
   fclose (output_fp);

   /* make hard links out of all the aliases (except the last) for this file  */
   while (term_name[index] != '\0') {
      index++;					/* increment past bar         */
      lindex = 0;                               /* rest lindex                */
      blank_flag = FALSE;                       /* rest blank flag            */

      /* grab an alias out of term_name and put in temp                       */
      while ((term_name[index] != '|') && (term_name[index] != '\0')) {
         if (term_name[index] == ' ')
            blank_flag = TRUE;
         temp[lindex++] = term_name[index++];
      }

      if (term_name[index] != '|') 		/* last alias; blow it off    */
         break;

      if ( ! blank_flag ) {
         temp[lindex] = '\0';                      /* null terminate          */
         sprintf(linkname, "%s/%c/%s", dirname, temp[0], temp); /* form name  */

#if DEBUG
fprintf(stderr, "linking %s to %s\n", filename, linkname);
#endif

         if (link(filename, linkname) < 0) {           /* link               */

            if (strcmp(filename, linkname) == 0)       /* link to itself ??? */
               error(13, filename);
            else if (errno == ENOENT) {                /* no directory       */

#if DEBUG
fprintf(stderr, "creating directory: %s/%c\n", dirname, temp[0]);
#endif

               /* set up string for system call to mkdir                     */
               sprintf (systring, "mkdir %s/%c",
                                  dirname,
                                  temp[0]);
               system (systring);
               link(filename, linkname);               /* link file          */
            }
	    else if (errno == EEXIST) {                /* link exists        */
 
#if DEBUG
fprintf(stderr, "removing old existing link: %s\n", linkname);
#endif

               unlink (linkname);                      /* remove the existing */
               link(filename, linkname);               /* link; link again    */
            }
            else {
               error (12, linkname);                   /* print error message */
               return;                                 /* and return          */
            } /* end if */

         } /* end if */

      } /* end if */

   } /* end while */

   return;

} /* end write_object */
