/*	@(#)pow.c	1.14	*/
/*LINTLIBRARY*/
/*
 *	pow(x, y) returns x ** y.
 *	Returns EDOM error and value 0 for 0 to a non-positive power
 *	or negative to a non-integral power;
 *	ERANGE error and value HUGE or -HUGE when the correct value
 *	would overflow, or 0 when the correct value would underflow.
 *	uses log and exp
 *	This version accepts negative x and integral y,
 *	apparently in violation of the ANSI FORTRAN standard for x ** y.
 *	There is a much more accurate but much more elaborate algorithm
 *	in Cody and Waite, which should be substituted for this.
 */

#include <math.h>
#include <values.h>
#include <errno.h>

double
pow(x, y)
register double x, y;
{

register int neg;
struct exception exc;

#include "pow_inc1.c"

  	if (x != 1) { /* x isn't the final result */
		/* the following code protects against multiplying x and y
		 * until there is no chance of multiplicative overflow */
		if ((x = log(x)) < 0) { /* preserve sign of product */
			x = -x;
			y = -y;
		}

#include "pow_inc2.c"

		x = exp(x * y); /* finally; no mishap can occur */
	}

#include "pow_inc3.c"

}

float xx;
double
powr(x, y)
register double x, y;
{

register int neg;
struct exception exc;

#include "pow_inc1.c"

	if (x != 1) { /* x isn't the final result */
		/* the following code protects against multiplying x and y
		 * until there is no chance of multiplicative overflow */

                xx = x;
	        asm ("  ldf  _xx");
	        asm ("  logf  ");
	        asm ("  stf  _xx");

		if (xx < 0) { /* preserve sign of product */
		        x = -xx;
			y = -y;
		}
	        else
	           x = xx;

#include "pow_inc2.c"

		xx = x * y;
		asm ("  ldf  _xx");
		asm ("  expf  ");
		asm ("  stf  _xx");
	   	x = xx;
	}

#include "pow_inc3.c"

}
