
#ifndef   VOS
#include "all.h"
#endif

/*
   #title termdef Character array allocation routines
*/
/*----------------------------------------------------------------------*
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

   /*------------------------------------------------------------*
    *          ++++++ alloc_char_array routine ++++++            *
    *  called from :                                             *
    *    get_key_sequences, make_tables                          *
    *                                                            *
    *  function :                                                *
    *    allocate a 2 dimensional character array in the heap    *
    *                                                            *
    *  arguments :                                               *
    *    int : # of rows                                         *
    *    int : # of columns                                      *
    *                                                            *
    *  returns char ** :                                         *
    *    pointer to array                                        *
    *------------------------------------------------------------*/

unsigned char **
alloc_char_array (rows, cols)
   int rows;
   int cols;
 {
   /* called routines */
   int set_row_pointers ();                                         /*<TKW-08>*/

   /* local variables  */
   unsigned char **return_value;                                    /*<TKW-08>*/
   unsigned char *data_ptr;                                         /*<TKW-08>*/

   /*----------------------------------------------*
    *  allocate the table of pointers to the rows  *
    *  and the actual char data area               *
    *----------------------------------------------*/

   return_value = (unsigned char **) calloc (rows, sizeof (int));   /*<TKW-08>*/
   data_ptr = (unsigned char *) calloc (rows * cols, sizeof (char));/*<TKW-08>*/

   /*-------------------------------------*
    *  assign values to the row pointers  *
    *-------------------------------------*/

   set_row_pointers (rows, cols, data_ptr, return_value);

   return return_value;
 }

/* #page */
   /*------------------------------------------------------------*
    *        ++++++ realloc_char_array routine ++++++            *
    *  called from :                                             *
    *    start_table_overflow, next_table_overflow               *
    *                                                            *
    *  function :                                                *
    *    reallocate a 2 dimensional character array in the heap  *
    *                                                            *
    *  arguments :                                               *
    *    int : # of rows                                         *
    *    int : # of columns                                      *
    *    char ** : current pointer to 2-d array                  *
    *                                                            *
    *  returns char ** :                                         *
    *    pointer to array                                        *
    *------------------------------------------------------------*/

unsigned char **                                                   /*<TKW-08>*/
realloc_char_array (rows, cols, current_ptr)                       /*<TKW-08>*/
   int  rows;                                                      /*<TKW-08>*/
   int  cols;                                                      /*<TKW-08>*/
   char **current_ptr;

 {
   /* called routines */
   int set_row_pointers ();

   /* local variables  */
   unsigned char **return_value;
   unsigned char *data_ptr;

   /*------------------------------------------------*
    *  reallocate the table of pointers to the rows  *
    *  and the actual char data area                 *
    *------------------------------------------------*/

   data_ptr =                                                      /*<TKW-08>*/
      (unsigned char *)newalloc (current_ptr [0], rows * cols * sizeof (char));
   return_value =                                                  /*<TKW-08>*/
      (unsigned char **) newalloc ( (char *) current_ptr, rows * sizeof (int));

   /*-------------------------------------*
    *  assign values to the row pointers  *
    *-------------------------------------*/

   set_row_pointers (rows, cols, data_ptr, return_value);

   return return_value;
 }
/*  #page */
   /*------------------------------------------------------------*
    *          ++++++ set_row_pointers routine ++++++            *
    *  called from :                                             *
    *    alloc_char_array, realloc_char_array                    *
    *                                                            *
    *  function :                                                *
    *    set the pointers to the rows of a 2d char array         *
    *                                                            *
    *  arguments :                                               *
    *    int : # of rows                                         *
    *    int : # of columns                                      *
    *    unsigned char *  : pointer to data portion of the array *
    *    unsigned char ** : pointer to pointers to array         *
    *------------------------------------------------------------*/

set_row_pointers (rows, cols, data_ptr, ptr_ptr)
   int  rows;
   int  cols;
   unsigned char *data_ptr;                                         /*<TKW-08>*/
   unsigned char **ptr_ptr;                                         /*<TKW-08>*/
 {
   int  index;

   for (index = 0;  index < rows;  ++index)
      ptr_ptr [index] = &data_ptr [cols * index];

   return;
 }
   /*------------------------------------------------------------*
    *        ++++++ dealloc_char_array routine ++++++            *
    *  called from :                                             *
    *                                                            *
    *                                                            *
    *  function :                                                *
    *    deallocate a 2 dimensional character array in the heap  *
    *                                                            *
    *  arguments :                                               *
    *    char ** : current pointer to 2-d array                  *
    *------------------------------------------------------------*/

dealloc_char_array (old_ptr)
   char **old_ptr;
 {
   /* free the data portion and the pointers to the rows */

   cfree (old_ptr [0]);       /* frees the data */
   cfree ((char *) old_ptr);  /* frees the pointers */
   return;
 }

/*
   This routine replaces the realloc routine in stdio
*/
char *
newalloc ( ptr, size)
char * ptr;
int size;

{
#ifdef USE_NOIO
   int * iptr;
   char * new_ptr;
   int old_size;
   int new_size;

   if ( (((int)ptr) & 070000000) == 0)  ++ptr;
   iptr = (int *) ptr;
   old_size = - *(--iptr);  /* number of words in original block */
   new_size = 2 + (size+2)/3;
   if ( old_size == new_size) return (ptr);
   old_size = (old_size - 2) * 3;
   new_ptr = calloc(size);
   if (old_size < size) size = old_size;
   for (old_size=0;old_size<size;old_size++)new_ptr[old_size] = ptr[old_size];
   cfree(ptr);
   return(new_ptr);
#else
   return ( (char *)realloc (ptr, size) );                         /*<TKW-08>*/
#endif
}
