
#ifndef   VOS
#include "all.h"
#endif

/*
   #title termdef getkey routine
*/
/*----------------------------------------------------------------------*
 *                                                                      *
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/
/*
*A* Mod Number: <862199> - By: Tim Wallace - On: Fri Jun 20 17:40:07 1986
*A*
*B* Modified Modules:
*B*      GETKEY.C
*B*
*C* Problem Description:
*C*      The tilde character was special cased as it was processed by
*C*      the getkey routine, meaning that an ASCII tilde was returned
*C*      even when the tilde was part of an escape sequence.  This
*C*      mostly affected VT200 terminals whose fuction keys were
*C*      defined to contain a tilde.  This bug was a carry-over from
*C*      TERMDEF.
*C*
*C* Resolution Description:
*C*      Took out tilde special casing.
*C*
*D* Comments:
*D*
*/
   /*------------------------------------------------------------*
    *         ++++++ getkey routine ++++++                       *
    *  function :                                                *
    *    identifies a key sequence or updates the                *
    *    key sequence search state based on the                  *
    *    passed input byte.                                      *
    *  arguments :                                               *
    *    unsigned char : new input byte from terminal            *
    *    unsigned char * : pointer for key info return           *
    *    struct tdef_heap *  : structure containing pointers to  *
    *                          FSA start table, FSA next table,  *
    *                          and terminal definition sections  *
    *  returns int :                                             *
    *    INVALID_SEQUENCE if an invalid sequence is found.       *
    *    UNDETERMINED if a sequence terminator is not reached.   *
    *    ASCII if an ascii character is entered in state 0.      *
    *    positive : the return value contained in the input      *
    *               FSA.                                         *
    *                                                            *
    *  *key_return (unsigned char) is set for the following      *
    *         return conditions :                                *
    *    INVALID_SEQUENCE : undefined                            *
    *    UNDETERMINED : undefined                                *
    *    ASCII : the ascii character found                       *
    *    CONTROL : undefined                                     *
    *------------------------------------------------------------*/




int getkey (byte_in, key_return, table_ptrs)
   unsigned char        byte_in;
   unsigned char        *key_return;
   struct tdef_heap     *table_ptrs;
 {
   /*---------------------*
    *  local definitions  *
    *---------------------*/

   static int     current_state = 0;   /* initialize state to 0 */
   int            current_index;       /* index into next_table */
   int            entries_left;        /* entries left in next_table */
   int            return_value;        /* value for function return */





   /*---------------------------------------------------------------*
    *  Processing begins here.  If the current state is 0, we have  *
    *  to special case ASCII input.  Otherwise, we enter the FSA    *
    *  processing.                                                  *
    *---------------------------------------------------------------*/

   if (current_state == 0  &&  byte_in >= ' '  && byte_in <= '~')   /*<862199>*/
    {
      *key_return = byte_in;
      return ASCII;
    }

   /*---------------------------------------------------------------*
    *  Get the first entry and the number of entries to examine     *
    *  next_table from start_table.                                 *
    *---------------------------------------------------------------*/

   current_index = table_ptrs->start_table [current_state].first_entry;
   entries_left  = table_ptrs->start_table [current_state].number_of_entries;

   /*------------------------------------------------------------*
    *  Loop here until we find the byte for this state or we run *
    *  out of entries to examine.                                *
    *------------------------------------------------------------*/

   while(( table_ptrs->next_table [current_index].input_byte != byte_in)
           && ( entries_left > 1 ))
      {
      --entries_left;
      ++current_index;
      }

   /*---------------------------------------------------------*
    *  current_index now points to the next_table entry for   *
    *  the current state and input byte.  Act here based on   *
    *  the next state (either terminal or non-terminal).      *
    *---------------------------------------------------------*/

   /* set return value */
   return_value = table_ptrs->next_table [current_index].return_key;

   if (table_ptrs->next_table [current_index].next_state == TERMINAL)
      /* reset current state to 0 */
      current_state = 0;
   else
      current_state = table_ptrs->next_table [current_index].next_state;

   /* now, simply return */
   return return_value;
 }
