
#ifndef   VOS
#include "all.h"
#endif

/*
   #title Read  termdef file  into heap
*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

/************************************************************************
 *                                                                      *
 *  get_term_def - reads a terminal definition file into the heap and   *
 *                 returns a structure of pointers to the various       *
 *                 sections of the definition.                          *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 *   The purpose of this routine is to make a terminal definition file  *
 *  ready for processing.  First, we check to see if a file containing  *
 *  the definition for a particular terminal exists.  If not, an error  *
 *  message is generated.  Otherwise, the header portion is processed   *
 *  so that the size of the file may be obtained and space allocated.   *
 *  Then the file is rewound and read into the allocated heap space.    *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * INPUT PARAMETER(S):                                                  *
 * -------------------                                                  *
 *         char *termdef_filename  - character pointer to name of file  *
 *                                   containing terminal definition     *
 *                                   information or a terminal type     *
 *                                   name.  If NULL, then information   *
 *                                   from the environment will be used. *
 *                                                                      *
 * VALUE RETURNED:                                                      *
 * ---------------                                                      *
 *          NULL  -  if the file corresponding to termdef_filename is   *
 *                   empty or nonexistent.                              *
 *      or                                                              *
 *          struct_tdef_heap* - pointer to structure on heap containing *
 *                              pointers to various sections of the     *
 *                              terminal definition.                    *
 *                                                                      *
 ************************************************************************/

/************************************************************************
 *                                                                      *
 * !!!SPECIAL NOTE !!! :                                                *
 * ---------------------                                                *
 *       The termdef_file contains integers in the short integer format *
 *       (two byte reversed) corresponding to the hardware of the VAX   *
 *       and PDP-11 hardware. Get_term_def reads the file into a heap   *
 *       as is, then converts the integers into host machine format     *
 *       ( 3 bytes/integer - VOS/6024  ;  (4 bytes/integer  - HCX   )   *
 *       and puts  all the information in a second heap which is kept   *
 *       for further processing.  The first heap is deallocated and the *
 *       pointer to the start of the remaining heap is returned.        *
 *                                                                      *
 ************************************************************************/

#define ALIGN 2                                                     /*<TKW-02>*/

#define MAGIC  0432     /* magic number in 1st two bytes of file */ /* 872750 */


struct tdef_heap  *get_term_def( termdef_filename )
char   *termdef_filename;

{

int            get_terminal_name ();
char           buffer[512];                                         /* 872750 */
struct tdef_heap  *terminal_def  /*  structure of pointers into heap */
        = NULL;                  /*  .. initial condition */        /* 872750 */
unsigned char  *temp_heap_start; /*  pointer to the start of heap space */
unsigned char  *save_start     ; /*  pointer to the start of heap space */
unsigned char  *perm_heap_start; /*  pointer to the start of heap space */
int            *perm_int_start;  /*  pointer to the start of heap space */
unsigned char  *temp_heap_ptr  ; /*  offset pointer into the heap       */
unsigned char  *perm_heap_ptr  ; /*  offset pointer into the heap       */
int            *perm_int_ptr   ; /*  offset pointer into the heap       */
unsigned char  *temp_num_start ; /*  pointer to the start of Numbers sect*/
int            *perm_num_start ; /*  pointer to the start of Numbers sect*/
unsigned char  *temp_str_start ; /*  pointer to the start of Strings sect*/
unsigned char  *perm_str_start ; /*  pointer to the start of Strings sect*/
unsigned char  *temp_str_end   ; /*  pointer to the end   of Strings sect*/
int    temp_heap_size ;          /*  size of the heap                   */
int    perm_heap_size ;          /*  size of the heap                   */
int    struct_size    ;          /*  size of the heap                   */
unsigned char  *head_ptr ;       /*  pointer to start of header data    */
unsigned int    magic_num ;      /*  to hold magic number from file     */
int    count    ;          /*  number of words read from file into heap */
int    i        ;
int    how_many ;          /*  number of short ints to convert */
int    null_padding;                                                /*<TKW-02>*/
int    section_size[HEADER_ENTRY_COUNT];/* converted section sizes */

#ifdef USE_NOIO
unsigned char   head_array[HEADER_ENTRY_COUNT * SHORT_INT_SIZE + SKIP_SIZE];
                                      /* buffer to hold section sizes */
                                      /*(short integer format)        */
#else
unsigned char   head_array[HEADER_ENTRY_COUNT * SHORT_INT_SIZE];
                                      /* buffer to hold section sizes */
                                      /*(short integer format)        */
#endif

#ifdef USE_NOIO
int   termdef_file;        /* LFN on VOS */
#else
FILE  *termdef_file ;      /* pointer to terminal definition file */
#endif


/******************************  CODE  ******************************/

/*
 * Get the terminfo filename from the environment, but only if it was
 * not passed in by the caller.  Stop here if we can't get a name.
 */
if (termdef_filename == NULL)
   {
   termdef_filename = buffer;
   *termdef_filename = '\0';                                        /* 872750 */
   }
i = get_terminal_name( termdef_filename );
if (i == FALSE)
   {

#ifdef USE_NOIO
   IOW(&SCREEN, &SYM_WRITE, &26, " No terminal name found. \n",&IO_A_reg);
#else
   fprintf(stderr, " No terminal name found. \n");
#endif

   return(NULL);
   }


/*
 * Attempt to access and open the constructed file name.
 */

#ifdef USE_NOIO

/*
 * Blank pad file name.  First, make sure it's on our local stack.
 * Here, the first 8 characters used to be converted from nulls to
 * blanks, and from lowercase to uppercase.  I thought that useless.   872750
 */
if ( termdef_filename != buffer )                                   /* 872750 */
   {                                                                /* 872750 */
   strcpy( buffer, termdef_filename );                              /* 872750 */
   strcat( buffer, "   " );                                         /* 872750 */
   termdef_filename = buffer;                                       /* 872750 */
   }                                                                /* 872750 */
FOPEN (&termdef_file, termdef_filename);
if (termdef_file == -1)     /* Error returns -1 */
   {
   IOW(&SCREEN, &SYM_WRITE, &36,
       " No terminal definition file found. ",&IO_A_reg);
   return( NULL );                                                  /* 872750 */
   }

#else   /* using STDIO */

#if VOS                                                             /*<TKW-04>*/
$IOMESG = 0;
#endif                                                              /*<TKW-04>*/

termdef_file = fopen(termdef_filename,"r");

#if VOS                                                             /*<TKW-04>*/
$IOMESG = 1;
#endif                                                              /*<TKW-04>*/

if (termdef_file == NULL)               /* not found, no entry */   /* 872750 */
{
   *termdef_filename = '\0';
   i = get_terminal_name( termdef_filename );
   if (i == FALSE)
   {
      fprintf(stderr, " No terminal name found. \n");
      return(NULL);
   }
   else
   {
      if ((termdef_file = fopen(termdef_filename,"r")) == NULL)
      {
         printf(" No terminal definition file found. \n");
         perm_heap_ptr = NULL;
         return( NULL );
      }
   }
}
#endif  /* using STDIO vs. NOIO */

/*
 * Start reading the file that we found.
 */

#ifdef USE_NOIO
   i = HEADER_ENTRY_COUNT * SHORT_INT_SIZE + SKIP_SIZE;
   IOW (&termdef_file, &SYM_READ, &i, (char *) head_array, &count);
   head_ptr = &(head_array[SKIP_SIZE]);                             /* 872750 */
#else
   count = fread((char *)head_array,SHORT_INT_SIZE,
                         HEADER_ENTRY_COUNT,termdef_file);
   head_ptr = head_array;                                           /* 872750 */
#endif

   if (count <= 0)                                                  /* 872750 */
      {
#ifdef USE_NOIO
      IOW (&SCREEN, &SYM_WRITE, &33,
                " Format error in HEADER section.  ", &IO_A_reg);
#else
      fprintf(stderr, " Format error in HEADER section.\n");
#endif
      return( NULL);
      }

   /*
    * Validate the magic number.                                       872750
    */
   magic_num = (head_ptr[1] << 8) + head_ptr[0];                    /* 872750 */
   if ( magic_num != MAGIC )                                        /* 872750 */
      {                                                             /* 872750 */
#ifdef USE_NOIO                                                     /* 872750 */
      IOW (&SCREEN, &SYM_WRITE, &42,                                /* 872750 */
                " Invalid magic number in HEADER section.  ", &IO_A_reg);
#else                                                               /* 872750 */
      fprintf(stderr, " Invalid magic number in HEADER section.\n");/* 872750 */
#endif                                                              /* 872750 */
      return( NULL);                                                /* 872750 */
      }                                                             /* 872750 */


      /*  convert header from short integer to usable format */

      stoi(head_ptr,section_size,HEADER_ENTRY_COUNT);               /* 872750 */

      /* calculate heap space required for original file      */
      temp_heap_size = SHORT_INT_SIZE *
                               (HEADER_ENTRY_COUNT
                              + section_size[NUMERIC_SECTION_SIZE]
                              + section_size[STRING_OFFSETS_SECTION_SIZE] )
                          +  section_size[NAMES_SECTION_SIZE]
                          +  section_size[BOOLEAN_SECTION_SIZE]
                          +  section_size[STRING_SECTION_SIZE]  ;
      /*                                                             / < TKW > /
       * The numeric section should start at an even short int       / < TKW > /
       * (2 byte) boundary.  If required, TDC or TIC should stick a  / < TKW > /
       * null byte at the end of the boolean section to accomplish   / < TKW > /
       * this.                                                       / < TKW > /
       *        The difference is that TDC counts the null byte(s) in/ < TKW > /
       * the Header section entry for the Boolean section size.  TIC / < TKW > /
       * does not. Therefore we should test and add an extra byte if / < TKW > /
       * necessary. TDC generated files should always pass the test. / < TKW > /
       * TIC files may or may not.                                   / < TKW > /
       */

       null_padding =                                               /*<TKW-02>*/
            (section_size[NAMES_SECTION_SIZE] +                     /*<TKW-02>*/
             section_size[BOOLEAN_SECTION_SIZE] ) % ALIGN;          /*<TKW-02>*/
       if (null_padding)                                            /*<TKW-02>*/
          temp_heap_size += ALIGN - null_padding;                   /*<TKW-02>*/
      temp_heap_size += 1; /* Add a cushion  */                     /*<TKW-02>*/

      /* calculate heap space required for heap in this machines format */
      perm_heap_size = HOST_INT_SIZE *
                               (HEADER_ENTRY_COUNT  +
                                section_size[NUMERIC_SECTION_SIZE] +
                                section_size[STRING_OFFSETS_SECTION_SIZE] )
                         +   section_size[NAMES_SECTION_SIZE] +
                             section_size[BOOLEAN_SECTION_SIZE] +
                             section_size[STRING_SECTION_SIZE]  ;
      perm_heap_size = perm_heap_size + HOST_INT_SIZE ;  /* to make sure we can
                                                            align on a word
                                                            boundary         */
      struct_size = sizeof(*terminal_def);
      terminal_def    = (struct tdef_heap *)malloc(struct_size);
      perm_heap_start = (unsigned char *)malloc(perm_heap_size);

#ifdef USE_NOIO
      temp_heap_start = (unsigned char *)malloc(temp_heap_size+SKIP_SIZE);
      IOS (&termdef_file, &REWIND,&IO_A_reg);
#else
      temp_heap_start = (unsigned char *)malloc( temp_heap_size );
      rewind(termdef_file);
#endif

      save_start = temp_heap_start;
      temp_heap_ptr = temp_heap_start;
      perm_heap_ptr = perm_heap_start;
      perm_int_start = (int *)perm_heap_start;

      /*  Move contents of file into temporary heap */

#ifdef USE_NOIO
      temp_heap_size += SKIP_SIZE;                                  /*<TKW-03>*/
      IOW (&termdef_file, &SYM_READ,
       &(temp_heap_size),(char *)temp_heap_ptr,&count);
      IOS (&termdef_file, &CLOSE, &IO_A_reg);
#else
      count = fread ((char *) temp_heap_ptr,
                    (temp_heap_size - 1),1,termdef_file);
      fclose(termdef_file);
#endif

      if (count == 0)
         {

#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &23, " File format in error.\n",&IO_A_reg);
#else
         printf(" File format in error.\n");
#endif
         return( NULL);
         }

      /* Save Header section in permanent heap */

#ifdef USE_NOIO
      temp_heap_start += SKIP_SIZE;
#endif

      perm_int_ptr = perm_int_start;
      for( i=0; i<HEADER_ENTRY_COUNT; i++)
         *perm_int_ptr++ = section_size[i];

      /*  Move remaining contents of temp heap into permanent heap  */
      /* ---------------------------------------------------------- */

      temp_num_start = temp_heap_start + SHORT_INT_SIZE * HEADER_ENTRY_COUNT
                       + *(perm_int_start + NAMES_SECTION_SIZE)
                       + *(perm_int_start + BOOLEAN_SECTION_SIZE);
      /*                                                             / < TKW > /
       * The Numerics section in the compiled terminal definition    / < TKW > /
       * file must begin on an even byte boundary. If necessary, a   / < TKW > /
       * null byte is appended to the end of the Booleans section to / < TKW > /
       * make sure. The TDC generated files include this null byte   / < TKW > /
       * in the Header section entry for Boolean section size. TIC   / < TKW > /
       * files do not. In order for TIC generated files to work with / < TKW > /
       * TERMLIB , we should test that the Numerics section starts   / < TKW > /
       * on an even byte and pad if necessary.                       / < TKW > /
       * <<<****    Actually, there is a bug/feature in TDC that     / < TKW > /
       * the the amount of padding is tied to the machine's short    / < TKW > /
       * integer size.  So, with 3-byte words on VOS, either one     / < TKW > /
       * or two null bytes may be required for padding. This         / < TKW > /
       * introduces an architecture dependency that is not necessary / < TKW > /
       * but since the retrofit would be painful to existing VOS     / < TKW > /
       * customers, the dependency will stand.  Therefore, do not    / < TKW > /
       * expect that TIC files will works on VOS; however, Terml(i)b / < TKW > /
       * should accept either TIC or TDC files on a Unix based OS.   / < TKW > /
       */
      if ( null_padding )                                           /*<TKW-02>*/
         temp_num_start += ALIGN - null_padding;                    /*<TKW-02>*/

      perm_num_start = (int *)(perm_heap_start
                            + (HOST_INT_SIZE * HEADER_ENTRY_COUNT)
                       + *(perm_int_start + NAMES_SECTION_SIZE)
                       + *(perm_int_start + BOOLEAN_SECTION_SIZE));
      perm_num_start = (int *)((unsigned char *)perm_num_start
                           + (HOST_INT_SIZE
                              - ((int)perm_num_start % (int)HOST_INT_SIZE)));

      temp_heap_ptr = temp_heap_start + SHORT_INT_SIZE * HEADER_ENTRY_COUNT ;
      perm_heap_ptr = perm_heap_start + HOST_INT_SIZE  * HEADER_ENTRY_COUNT ;
      while (temp_heap_ptr < temp_num_start)     /* NAMES & BOOLEAN SECTIONS */
         *perm_heap_ptr++ = *temp_heap_ptr++;

      how_many = perm_int_start [NUMERIC_SECTION_SIZE]
               + perm_int_start [STRING_OFFSETS_SECTION_SIZE];
      stoi(temp_num_start,perm_num_start,how_many);  /* NUMBERS &
                                                        STRING OFFSETS
                                                        SECTIONS      */
   /*                                                                / < TKW > /
    * temp_str_start = temp_heap_start + SHORT_INT_SIZE              / < TKW > /
    *                  * HEADER_ENTRY_COUNT                          / < TKW > /
    *                  + *(perm_int_start + NAMES_SECTION_SIZE)      / < TKW > /
    *                  + *(perm_int_start + BOOLEAN_SECTION_SIZE)    / < TKW > /
    *                  + SHORT_INT_SIZE *                            / < TKW > /
    *            (  *(perm_int_start + NUMERIC_SECTION_SIZE)         / < TKW > /
    *            + *(perm_int_start + STRING_OFFSETS_SECTION_SIZE)); / < TKW > /
    */                                                              /*<TKW-02>*/
      temp_str_start = (unsigned char *)( temp_num_start            /*<TKW-02>*/
                       + SHORT_INT_SIZE *                           /*<TKW-02>*/
                 ( *(perm_int_start + NUMERIC_SECTION_SIZE)         /*<TKW-02>*/
                 + *(perm_int_start + STRING_OFFSETS_SECTION_SIZE)));/*< TKW >*/

      temp_str_end   = temp_str_start
                       + *(perm_int_start + STRING_SECTION_SIZE);
/*    perm_str_start = perm_heap_start
==                          + (HOST_INT_SIZE * HEADER_ENTRY_COUNT)
==                     + *(perm_int_start + NAMES_SECTION_SIZE)
==                     + *(perm_int_start + BOOLEAN_SECTION_SIZE)
==                     + HOST_INT_SIZE *
==                       (  *(perm_int_start + NUMERIC_SECTION_SIZE)
==                        + *(perm_int_start + STRING_OFFSETS_SECTION_SIZE));
 */
      perm_str_start = (unsigned char *)( perm_num_start
                                        + perm_int_start [NUMERIC_SECTION_SIZE]
                          + perm_int_start [STRING_OFFSETS_SECTION_SIZE]);

      perm_heap_ptr = perm_str_start;
      temp_heap_ptr = temp_str_start;
      while (temp_heap_ptr < temp_str_end)     /* STRING SECTION */
         *perm_heap_ptr++ = *temp_heap_ptr++;

      terminal_def->header_section = (int *)perm_heap_start;
      terminal_def->names_section =
                    perm_heap_start + (HEADER_SIZE * HOST_INT_SIZE);
      terminal_def->boolean_section =
                    terminal_def->names_section +
                    *(perm_int_start + NAMES_SECTION_SIZE);
      terminal_def->numbers_section = perm_num_start;
      terminal_def->string_offsets_section = perm_num_start +
                          *(perm_int_start + NUMERIC_SECTION_SIZE);
      terminal_def->strings_section = perm_str_start;
      terminal_def->tdef_heap_end =
                   terminal_def->strings_section +
                   *(perm_int_start + STRING_SECTION_SIZE);

      cfree(save_start);

return(terminal_def);
}
