/*
 *A* Mod Number: <872750> - By: Lee Brown - On: Tue Jun 16 16:40:31 1987
 *A*
 *B* Modified Modules:
 *B*      GETTD.C  GETTER.C GETENV.C V:TLIB   V:HC
 *B*
 *C* Problem Description:
 *C*      1) TERMLB does not query the VUE TERM environment variable.
 *C*      2) TERMLB cannot be used if a file with the same name as the
 *C*         desired TERI file exists under the default qualifier.
 *C*      3) When setting the #TER register, TERMLB must use
 *C*         permanent registers.
 *C*      4) TERMDEF parameter is not queried for interactive pgms.
 *C*      5) The magic number in the TERI file ought to be validated.
 *C*
 *C* Resolution Description:
 *C*      1) Taught TERMLB how to query environment variables.
 *C*      2) We no longer look in the default qualifier.  The TERMINFO
 *C*         environment variable or #TIN register may be used to
 *C*         specify qualifiers other than 0000TERI.
 *C*      3) Taught TERMLB how to specify permanent jcl registers.
 *C*      4) Added a call to $TDEF for interactive programs.
 *C*      5) Added code to validate the TERI file magic number.
 *C*
 *D* Comments:
 *D*      This involved a not insignificant re-write for 3 reasons:
 *D*      flow of control changed due to selecting one and only one
 *D*      qualifier to search, I tried to make the VOS and CX versions
 *D*      more consistent, and I knew some ways to take advantage of
 *D*      the C library that were not being used.
 *D*
 *D*      I added 4 user-callable subroutines: Get_env_var,
 *D*      Get_termdef, Get_jcl_reg, and Set_jcl_reg.  These are for
 *D*      our own internal use and will not be documented.
 *D*
 *A* Mod Number: <872727> - By: Lee Brown - On: Wed May 27 14:54:38 1987
 *A*
 *B* Modified Modules:
 *B*      PARSTR.C
 *B*
 *C* Problem Description:
 *C*      TERMLB does not use the proper cursor positioning sequence
 *C*      for a HMW terminal for any address with column=0.  In that
 *C*      case, the column (which comes first in the sequence) is
 *C*      omitted and the row is used as the column followed by a
 *C*      zero byte used as the row.
 *C*
 *C* Resolution Description:
 *C*      A %c was "printed" as if it were a null terminated string --
 *C*      when the value is binary zero, the string pointer is not
 *C*      advanced beyond what it thinks is the string terminator.
 *C*      A much simpler method of fulfilling a %c is now used.
 *C*
 *D* Comments:
 *D*      This is likely to be a patch for 5.1.1 and 6.1, and released
 *D*      with 6.1.1 and 6.2.
 *D*
 *A* Mod Number: <TKW-0n> - By: Tim Wallace - On: Tue May 05 14:27:08 1987
 *A*
 *B* Modified Modules:
 *B*      alloc.c  bool.c   echopr.c fillst.c getkey.c gettd.c
 *B*      getter.c global.c num.c    parse.c  parstr.c prterm.c
 *B*      sort.c   stacks.c string.c swap.c   tables.c tlib.h
 *B*
 *C* Problem Description:
 *C*      1) In the process of updating the source on the HCX, it was
 *C*         necessary to remove some compiler directives in order to
 *C*         provide common source on both VOS and UN*X machines.
 *C*         Conditional compilation units take care of OS dependent
 *C*         features.
 *C*      2) In the process of creating the compiled terminal definition
 *C*         file, TIC/TDC must ensure that the NUMERICS section starts
 *C*         on an even byte boundary. So, the BOOLEAN section may have to
 *C*         have a single NULL byte appended to it.  TDC includes any NULL
 *C*         padding in the size of the BOOLEAN section that it writes to
 *C*         the HEADER.  TIC does not include padding bytes in the size
 *C*         that it writes to the header. Therefore, a test should be
 *C*         performed in gettd.c where the start of the Numeric section
 *C*         is determined and also where the space required for the heap is
 *C*         calculated.
 *C*         ****NOTE****: TDC has an error in that it pads the
 *C*         BOOLEAN section so that the NUMERIC section starts on an even
 *C*         word boundary on VOS. As a result, BOOLEAN section on a VOS
 *C*         environment  may have 0,1, or 2 bytes padding.  This will not
 *C*         be corrected at this time because it would require all customers
 *C*         to recompile their TERI files. Old files would not be consistent
 *C*         with the new TDC or TERMLB.
 *C*      3) The NOIO (no stdio) version of the library does not read in
 *C*         the entire definition file because the 6 words VOS places at
 *C*         the beginning of the file were not considered when the I/O
 *C*         call was made to read the compiled definition file into the
 *C*         heap.
 *C*      4) Also with the NOIO version,  when a terminal name is passed to
 *C*         get_term_def(), the terminal name is blank filled
 *C*         to 8 characters, but the size of the string containing the
 *C*         is not considered. Overwriting of other data/code could occur.
 *C*      5) For the UN*X version, if a terminal name was passed to
 *C*         get_term_def() it was only searched for under the local directory.
 *C*      6) Numerous changes were made to the diagnostic routines,
 *C*         echo_print() and print_fts() for more informative output.
 *C*      7) A bug was uncovered in fillst.c, and string.c in which
 *C*         an integer variable  was used to hold an offset into the
 *C*         string section, then its address was checked to see
 *C*         if it was a valid address.  An integer pointer should
 *C*         have been used.
 *C*      8) Modifications to tables.c and alloc.c to support the
 *C*         NOIO version used "vhil" type function declarations which
 *C*         didn't compile well (sic) under UN*X.
 *C*
 *C* Resolution Description:
 *C*      1) #page, #title, #expandinclude were removed and,
 *C*         #loadtable's were replaced with #include 's. ( <TKW-01> )
 *C*      2) Numeric section is checked to start on the proper byte
 *C*         an necessary padding bytes are added to size for heap. ( <TKW-02> )
 *C*      3) Add SKIP_SIZE to size passed to I/O call to read file into heap.
 *C*         ( <TKW-03> )
 *C*      4) When a non-NULL terminal name is passed to get_term_def(), the
 *C*         NOIO version will copy the name to a local buffer before blank
 *C*         padding the name to 8 characters. ( <TKW-04> )
 *C*      5) When a non-NULL terminal name is passed to get_term_def(), the
 *C*         UN*X version search under the local directory first, then
 *C*         it will search under the default path or, if the TERMINFO
 *C*         environment  variable is set, it will search under the
 *C*         specified path. ( <TKW-05> )
 *C*      6) echo_print() and print_fts() now print attribute indices as
 *C*         well as other additional info. ( <TKW-06> )
 *C*      7) Declared variable to be integer pointer. (  <TKW-07> )
 *C*      8) Changed function declarations to be "C" compatible ( <TKW-08> )
 *C*
 *D* Comments:
 *D*         None
 */
