
#ifndef   VOS
#include "all.h"
#endif

/*
   #title Apply parameters to string
*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *  apply_parameters_to_string : applies a list of parameters to a      *
 *       string from the terminal definition and creates an interpreted *
 *       string which can be sent to the terminal.                      *
 *                                                                      *
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *  INPUT :                                                             *
 *     int num_parms     - number of parameters which are to be passed  *
 *     int *parm_ptr     - pointer to parameter list                    *
 *     int max_len       - max length of (returned string +             *
 *                                        Null terminator)              *
 *     unsigned char   *p_string                                        *
 *                       - string from terminal definition stripped of  *
 *                         padding information                          *
 *                                                                      *
 *----------------------------------------------------------------------*
 *                                                                      *
 *   OUTPUT :                                                           *
 *     unsigned char   *return_string                                   *
 *                       - user string to which the interpreted string  *
 *                         is returned                                  *
 *                                                                      *
 *----------------------------------------------------------------------*
 *                                                                      *
 *   RETURN : (int)                                                     *
 *      positive - length of returned string excluding null terminator  *
 *      0        - error max length exceeded                            *
 *      negative - an error was detected in interpreting the input      *
 *                 string (see below)                                   *
 *                                                                      *
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 *                                                                      *
 * The following is a directory of error conditions (negative return) : *
 *                                                                      *
 *       LENGTH_ERROR         0    - max length exceeded or NULL string *
 *       IF_THEN_ELSE_ERROR  -1    - improper string of the form :      *
 *                              " %? expr %t thenpart %e elsepart ..%;" *
 *       PERCENT_ERR         -2    - improper string of the form :      *
 *                                   "%..."                             *
 *       PR_0ND_ERR          -3    - improper string of the form :      *
 *                                   "%0.."                             *
 *       PR_2D_ERR           -4    - improper string of the form :      *
 *                                   "%2.." or "%02.."                  *
 *       PR_3D_ERR           -5    - improper string of the form :      *
 *                                   "%3.." or "%03.."                  *
 *       PUSH_ERR            -6    - improper string of the form :      *
 *                                   "%p.."                             *
 *       POP_VAR_ERR         -7    - improper string of the form :      *
 *                                   "%P.."                             *
 *       PUSH_VAR_ERR        -8    - improper string of the form :      *
 *                                   "%g.."                             *
 *       CHAR_CONST_ERR      -9    - improper string of the form :      *
 *                                   "%'.."                             *
 *       INT_CONST_ERR       -10   - improper string of the form :      *
 *                                   "%{.."                             *
 *       VAR_NUM_ERR         -11   - invalid variable index :           *
 *                                   valid range is "a..z"              *
 *       PARM_NUM_ERR        -12   - invalid parameter index :          *
 *                                   valid range is "1..9"              *
 *       PUSH_POP_ERR        -13   - stack error - probably exceeding   *
 *                                   defined stack limits.              *
 *       IF_PUSH_POP_ERR     -14   - stack error in processing if-then- *
 *                                   else - probably exceeding defined  *
 *                                   stack limits.                      *
 *                                                                      *
 *----------------------------------------------------------------------*/

  /* global variables accessed */

/*  int      end_of_string_;                      */
/*  int      int_c_;                              */
/*  int      error_;                              */
/*  int     *mystack_top_;                        */
/*  int     *max_mystack_;                        */
/*  unsigned char    *if_stack_top_;              */
/*  unsigned char    *max_if_stack_;              */


#define  get_digits                                            \
            if (number == 0)                                   \
               digits = 1;                                     \
            else                                               \
               {                                               \
               if (number < 0)                                 \
                  {                                            \
                  negate = TRUE;                               \
                  number = - number;                           \
                  }                                            \
               base = 1;                                       \
               digits = 0;                                     \
               quotient = 1;                                   \
               while (quotient > 0)                            \
                  {                                            \
                  quotient = number / base;                    \
                  if (quotient > 0)                            \
                     {                                         \
                     digits++;                                 \
                     base = base * 10;                         \
                     }                                         \
                  }                                            \
               if (negate)                                     \
                  {                                            \
                  negate = FALSE;                              \
                  number = - number;                           \
                  }                                            \
               }                                               \





int apply_parameters_to_string
                    (num_parms, parm_ptr, max_len , return_string, p_string  )

int    num_parms;
int   *parm_ptr ;
int    max_len   ;
unsigned char  *return_string;
unsigned char  *p_string     ;

{
int      get_op_type();
int      keep;
int      len;
int      operation;
int      varray[26];
int      doneflush;
unsigned char  *popped_string     ;
unsigned char   save;
int      number,digits,quotient,negate,base;

len = 0;
negate = FALSE;
num_parms;
mystack_top_ = mystack - 1;
max_mystack_ = mystack + MYSTACK_SIZE;
if_stack_top_ = if_stack - 1;
max_if_stack_ = if_stack + IF_STACK_SIZE;
if_push(TRUE);
if (len < max_len)
   {
   end_of_string_ = FALSE;
   error_ = NO_ERROR;
   }
else
   {
   end_of_string_ = TRUE;
   error_ = LENGTH_ERR;
   operation = OP_ERROR;
   }
while( !end_of_string_)
   {
   keep = if_peek(1);
   operation = get_op_type(&return_string,&p_string,max_len,&len,keep);
   if (operation != END_STRING)
      ++p_string;
   if (if_peek(1))
      {
      switch( operation )
         {
         case PERCENT :
            ++len;
            if (len < max_len)
               *return_string++ = '%';
            break;
         case PRINT_POP_D :
            number = pop();
            get_digits;                 /* determines magnitude of number */
            len += digits;
            if (len < max_len)
               {

#ifdef USE_NOIO
               sprintd(return_string,1, number);
#else
               sprintf(return_string,"%d",number);
#endif
               while (*return_string != NULL)
                  ++return_string;
               }
            break;
         case PRINT_POP_2D :
            number = pop();
            get_digits;
            if (digits < 2)
               digits = 2;
            len += digits;
            if (len < max_len)
               {

#ifdef USE_NOIO
               sprintd(return_string,2, number);
#else
               sprintf(return_string,"%2d",number);
#endif
               while (*return_string != NULL)
                  ++return_string;
               }
            break;
         case PRINT_POP_3D :
            number = pop();
            get_digits;
            if (digits < 3)
               digits = 3;
            len += digits;
            if (len < max_len)
               {

#ifdef USE_NOIO
               sprintd(return_string,3, number);
#else
               sprintf(return_string,"%3d",number);
#endif
               while (*return_string != NULL)
                  ++return_string;
               }
            break;
         case PRINT_POP_C :
            ++len;
            if (len < max_len)
               {
               *return_string = (unsigned char) pop();              /* 872727 */
               ++return_string;                                     /* 872727 */
               *return_string = '\0';                               /* 872727 */
               }
            break;
         case PRINT_POP_S :                              /* " */
            popped_string = (unsigned char *)pop();
            len += strlen(popped_string);
            if (len < max_len)
               {
#ifdef USE_NOIO
               while (*popped_string != NULL)
                  *return_string++ = *popped_string++;
               *return_string = NULL;
#else
               sprintf(return_string,"%s",popped_string);
               while (*return_string != NULL)
                  ++return_string;
#endif
               }
            break;
         case PUSH_OP :
            if ((int_c_ >= 0) && (int_c_ <= 8))
               push(parm_ptr[int_c_]);
            else
               {
               operation = OP_ERROR;
               error_ = VAR_NUM_ERR;
               }
            break;
         case POP_VAR_OP :
            if ((int_c_ >= 0) && (int_c_ <= 25))
               varray[int_c_] = pop();
            else
               {
               operation = OP_ERROR;
               error_ = VAR_NUM_ERR;
               }
            break;
         case PUSH_VAR_OP :
            if ((int_c_ >= 0) && (int_c_ <= 25))
               push(varray[int_c_]);
            else
               {
               operation = OP_ERROR;
               error_ = VAR_NUM_ERR;
               }
            break;
         case CHAR_CONST_OP :
            push(int_c_);
            break;
         case INT_CONST_OP :
            push(int_c_);
            break;
         case PLUS :
            push (pop() + pop());
            break;
         case MINUS :
            push (pop() - pop());
            break;
         case TIMES :
            push (pop() * pop());
            break;
         case DIVIDE :
            push (pop() / pop());
            break;
         case MOD   :
            push (pop() % pop());
            break;
         case BIT_AND:
            push (pop() & pop());
            break;
         case BIT_OR :
            push (pop() | pop());
            break;
         case BIT_XOR :
            push (pop() ^ pop());
            break;
         case LOG_EQUAL :
            push (pop() == pop());
            break;
         case LESS_THAN :
            push (pop() < pop());
            break;
         case GREATER_THAN :
            push (pop() > pop());
            break;
         case LOG_NOT :
            push( ! pop());
            break;
         case BIT_NOT :
            push( ~ pop());
            break;
         case INCREM :
            parm_ptr[0]++;
            parm_ptr[1]++;
            break;
         }
      }
   switch(operation)
      {
      case IF_OP :
         save = if_peek(1);
         if_push('?');
         if_push(save);
         break;
      case THEN_OP :
         if ((if_peek(2) == 'e' ) || (if_peek(2) == '?'))
            {
            if_push('t');
            if (if_peek(2))
               if_push ((char)pop());
           /*  {                           */
           /*  if (if_pop(if_stack))       */
           /*     if_push('\1',if_stack);  */
           /*  else                        */
           /*     if_push('\0',if_stack);  */
           /*  }                           */
            else
               if_push('\0');
            }
         else
            {
            error_ = IF_THEN_ELSE_ERROR;
            operation = OP_ERROR;
            }
         break;
      case ELSE_OP :
         if (if_peek(2) == 't' )
            {
            save = if_pop();
            if_pop();               /* pops 't' */
            if_push('e');
            if (if_peek(2))
               if_push (!save);
            else
               if_push('\0');
            }
         else
            {
            error_ = IF_THEN_ELSE_ERROR;
            operation = OP_ERROR;
            }
         if (error_ == PUSH_POP_ERR)
            operation = OP_ERROR;
         break;
      case END_IF_OP :
         doneflush = if_flush();
         if (doneflush != TRUE)
            {
            error_ = IF_THEN_ELSE_ERROR;
            operation = OP_ERROR;
            }
         break;
      case OP_ERROR :
         end_of_string_ = TRUE;
         break;
      case END_STRING :
         end_of_string_ = TRUE;
         break;
      }
   if (len >= max_len)
      {
      if (error_ == NO_ERROR)
         error_ = LENGTH_ERR;
      operation = OP_ERROR;
      end_of_string_ = TRUE;
      }
   if ((error_ == PUSH_POP_ERR) || (error_ == IF_PUSH_POP_ERR))
            operation = OP_ERROR;
   }
if (error_ == NO_ERROR)
   *return_string = NULL;
if (operation == OP_ERROR)
   return(error_);
else
   return(len);
}



int get_op_type(preturn_string,pp_string,max_len,len_ptr,keep)

unsigned char    **preturn_string;
unsigned char    **pp_string;
int      max_len;
int     *len_ptr;
int      keep;
{
unsigned char     temp[132] ;
unsigned char     ccc;
int      state;
int      op_type;
int      i,j;
int      temp_len;


state = START;
temp_len = *len_ptr;
i = 0;
op_type = 0;
while(!op_type)
  {
   ccc = **pp_string;
   switch(state)
      {
      case START :
         while(( ccc != '\0') && ( ccc != '%'))
            {
            if (keep)
               {
                ++temp_len;
               if (temp_len < max_len)
                  {
                  **preturn_string = ccc;
                  ++(*preturn_string);
                  }
               else
                  {
                  op_type = OP_ERROR;
                  error_ = LENGTH_ERR;
                  }
               }
            ++(*pp_string);
            ccc = **pp_string;
            }
         if (ccc == '%')
            {
            state = PERCENT;
            ++(*pp_string);
            temp[ i++ ] = ccc;
            }
         else
            op_type = END_STRING;
         break;

      case PERCENT :
         switch(ccc)
            {
            case '%' :
               op_type = PERCENT;
               break;
            case '0' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = ITSA0;
               break;
            case '2' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = ITSA2;
               break;
            case '3' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = ITSA3;
               break;
            case 'd' :
               op_type = PRINT_POP_D;
               break;
            case 'c' :
               op_type = PRINT_POP_C;
               break;
            case 's' :
               op_type = PRINT_POP_S;
               break;
            case 'p' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = SMALLP;
               break;
            case 'P' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = BIGP;
               break;
            case 'g' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = SMALLG;
               break;
            case '\'' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = TICK1 ;
               break;
            case '{' :
               temp[ i++ ] = ccc;
               ++(*pp_string);
               state = L_BRACE;
               break;
            case '+' :
               op_type = PLUS ;
               break;
            case '-' :
               op_type = MINUS ;
               break;
            case '*' :
               op_type = TIMES ;
               break;
            case '/' :
               op_type = DIVIDE;
               break;
            case 'm' :
               op_type = MOD   ;
               break;
            case '&' :
               op_type = BIT_AND;
               break;
            case '|' :
               op_type = BIT_OR;
               break;
            case '^' :
               op_type = BIT_XOR;
               break;
            case '=' :
               op_type = LOG_EQUAL;
               break;
            case '<' :
               op_type = LESS_THAN;
               break;
            case '>' :
               op_type = GREATER_THAN;
               break;
            case '!' :
               op_type = LOG_NOT ;
               break;
            case '~' :
               op_type = BIT_NOT;
               break;
            case 'i' :
               op_type = INCREM ;
               break;
            case '?' :
               op_type = IF_OP ;
               break;
            case 't' :
               op_type = THEN_OP ;
               break;
            case 'e' :
               op_type = ELSE_OP ;
               break;
            case ';' :
               op_type = END_IF_OP;
               break;
            default :
               temp[i] = ccc ;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               error_   = PERCENT_ERR;
               op_type = OP_ERROR;
               break;
            }
         break;
      case ITSA0 :
         if ( ccc == '2')
            {
            temp[i++] = ccc;
            state = ITSA2;
            ++(*pp_string);
            }
         else if ( ccc == '3')
            {
            temp[i++] = ccc;
            state = ITSA3;
            ++(*pp_string);
            }
         else
            {
            temp[i] = ccc;
            if (keep)
               {
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = PR_0ND_ERR;
               op_type = OP_ERROR;
            }
         break;
      case ITSA2 :
         if ( ccc == 'd')
            op_type = PRINT_POP_2D;
         else
            {
            temp[i] = ccc;
            if (keep)
               {
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = PR_2D_ERR;
               op_type = OP_ERROR;
            }
         break;
      case ITSA3 :
         if ( ccc == 'd')
            op_type = PRINT_POP_3D;
         else
            {
            temp[i] = ccc;
            if (keep)
               {
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = PR_3D_ERR;
               op_type = OP_ERROR;
            }
         break;
      case SMALLP :
         if (( ccc >= '1') && ( ccc <= '9'))
            {
            int_c_ = (int)(ccc - '1');
            op_type = PUSH_OP;
            }
         else
            {
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = PUSH_ERR;
               op_type = OP_ERROR;
            }
         break;
      case BIGP :
         if (( ccc >= 'a') && ( ccc <= 'z'))
            {
            int_c_ = (int)(ccc - 'a');
            op_type = POP_VAR_OP;
            }
         else
            {
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = POP_VAR_ERR;
               op_type = OP_ERROR;
            }
         break;
      case SMALLG :
         if (( ccc >= 'a') && ( ccc <= 'z'))
            {
            int_c_ = (int)(ccc - 'a');
            op_type = PUSH_VAR_OP;
            }
         else
            {
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               }
               error_   = PUSH_VAR_ERR;
               op_type = OP_ERROR;
            }
         break;
      case TICK1 :
         if (( ccc >= ' ') && ( ccc <= '~'))
            {
            int_c_ = (int)ccc;
            state = TICK2;
            temp[ i++ ] = ccc;
            ++(*pp_string);
            }
         else
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
                  error_   = CHAR_CONST_ERR;
                  op_type = OP_ERROR;
               }
         break;
      case TICK2 :
         if ( ccc >= '\'')
            {
            op_type = CHAR_CONST_OP;
            }
         else
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               error_   = CHAR_CONST_ERR;
               op_type = OP_ERROR;
               }
         break;
      case L_BRACE :
         int_c_ = 0;
         temp[ i++ ] = ccc;
         ++(*pp_string);
         if (( ccc >= '0') && ( ccc <= '9'))
            {
            int_c_ = 10 * int_c_ + (int)(ccc - '0');
            state = GET_INT_CONST;
            }
         else
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               error_   = INT_CONST_ERR;
               op_type = OP_ERROR;
               }
         break;
      case GET_INT_CONST :
         temp[ i++ ] = ccc;
         if (( ccc >= '0') && ( ccc <= '9'))
            {
            int_c_ = 10 * int_c_ + (int)(ccc - '0');
            state = GET_INT_CONST;
            ++(*pp_string);
            }
         else if ( ccc == '}')
            op_type = INT_CONST_OP;
         else
            if (keep)
               {
               temp[i] = ccc;
               for( j = 0; j <= i; j++)
                  {
                  ++temp_len;
                  if (temp_len < max_len)
                     {
                     **preturn_string = temp[j];
                     ++(*preturn_string);
                     }
                  }
               error_   = INT_CONST_ERR;
               op_type = OP_ERROR;
               }
         break;
      }    /* end switch */
   if (temp_len > max_len)
      {
      if (error_ == NO_ERROR)
         error_ = LENGTH_ERR;
      op_type = OP_ERROR;
      end_of_string_ = TRUE;
      }
   }       /* end while */
*len_ptr = temp_len;
return(op_type);
}

#ifdef USE_NOIO

int sprintd (char * buffer, int num_digs, int value)
{
  char digits[9];
  int i;
  int returnval = 0;

  /* these are used for BLU calls */

#ifdef VOS
int regs [NUM_REGS];    /* registers I, J, K, E, A */
#endif


   regs[A] = value ;
   BLU (BLU_BINASC, regs, regs);
   digits[0] = (regs[K] & 077600000) >> 16 ;
   digits[1] = (regs[K] & 000177400) >> 8  ;
   digits[2] =  regs[K] & 000000377        ;
   digits[3] = (regs[E] & 077600000) >> 16 ;
   digits[4] = (regs[E] & 000177400) >> 8  ;
   digits[5] =  regs[E] & 000000377        ;
   digits[6] = (regs[A] & 077600000) >> 16 ;
   digits[7] = (regs[A] & 000177400) >> 8  ;
   digits[8] =  regs[A] & 000000377        ;

   if (num_digs < 1 || num_digs > 9) num_digs = 1;

   for (i=0 ; i<9 ; i++ ) {
      if (digits[i] != ' ' || 9 - num_digs <= i) {
         *buffer++ = digits[i];
         returnval++;
      }
   }
   *buffer = NULL;
   return (returnval+1);
}
#endif
