
#ifndef   VOS
#include "all.h"
#endif

/*
   #title termdef key sequence sort routine
 */

/*----------------------------------------------------------------------*
 *                                                                      *
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

   /*------------------------------------------------------------*
    *             ++++++ key_sort routine ++++++                 *
    *  called from :                                             *
    *    build_tables                                            *
    *                                                            *
    *  function :                                                *
    *    Eliminates all undefined key sequences, then sorts      *
    *    the remainder.  number_of_keys contains the new number  *
    *    of keys on return.                                      *
    *                                                            *
    *  arguments :                                               *
    *    int *      : array of key names                         *
    *    int *      : array of key return values                 *
    *    int *      : pointer to number of keys defined          *
    *    unsigned char ** : unsorted key sequences               *
    *                                                            *
    *  returns :                                                 *
    *    TRUE if sort went ok.                                   *
    *    FALSE if an error occurred that will prohibit unique    *
    *       key determination.                                   *
    *------------------------------------------------------------*/


int key_sort (key_name, key_return, number_of_keys, sequences)
   int           key_name [] ;
   int  *        key_return;
   int  *        number_of_keys;
   unsigned char **sequences;
 {

   int   index_1, index_2;             /* handy indices */
   int   compares;                     /* # of compares to do this pass */
   int   switch_made;                  /* sort switch flag */
   int   return_value;                 /* function return value */

   unsigned char temp_sequence [KEY_SEQUENCE_SIZE];
                                       /* temp key sequence holder */
   int   temp_int;                     /* temp int holder */
   int   key_count;                    /* count of actual keys */
   int   return_sort;                  /* return value from sort */

   return_sort = TRUE;
   key_count = 0;

   /* eliminate any undefined keys */
   for (index_1 = 0;  index_1 < *number_of_keys;  ++index_1)
    {
      if (sequences [index_1] [0] == '\0')
       {
         /* it's a null sequence, move the next one here
            (if we're not at the end) */
         if (index_1 != *number_of_keys)
          {
            /* find the next defined sequence */
            for (index_2 = index_1 + 1;
                 index_2 < *number_of_keys;
                 ++index_2)
             {
               if (sequences [index_2] [0] != '\0')
                  break;
             }

            /* did we get one ? */
            if (index_2 < *number_of_keys)
             {
               /* yes, copy that one here */
               key_return [index_1] = key_return [index_2];
               strcpy (sequences [index_1], sequences [index_2]);
               key_name [index_1] = key_name [index_2];

               sequences [index_2] [0] = 0; /* null the other one out */
               key_count = index_1 + 1;
                     /* adjust number of actual keys defined */
             }
          } /* end if not last one */
       } /* end if null sequence */
      else
         /* add one to the number of keys */
         ++key_count;
    } /* end for loop */

   /* sort 'em - quick 'n' dirty rock (and roll) sort */
   switch_made = TRUE;
   index_1 = 0;
   *number_of_keys = key_count;  /* update parameter key count */
   compares = key_count;

   while (index_1 < key_count - 1  && switch_made == TRUE)
    {
      --compares;
      switch_made = FALSE;
      for (index_2 = 0;  index_2 <= compares-1;  ++index_2)
       {
         /* see if the next two should be switched */
         return_value = strcmp (sequences [index_2], sequences [index_2+1]);

         if (return_value > 0)
          {
            /* "bigger" string is first, switch_made */
            switch_made = TRUE;

            temp_int = key_return [index_2];
            key_return [index_2] = key_return [index_2 + 1];
            key_return [index_2 + 1] = temp_int;

            strcpy (temp_sequence, sequences [index_2]);
            strcpy (sequences [index_2], sequences [index_2 + 1]);
            strcpy (sequences [index_2 + 1], temp_sequence);

            temp_int = key_name [index_2];
            key_name [index_2] = key_name [index_2 + 1];
            key_name [index_2 + 1] = temp_int;
          }
         else if (return_value == 0)
            /* two keys with same sequence - flag it */
               return_sort = FALSE;
       }  /* end for */
    } /* end while */

   /* if we saw some duplicate sequences, check them out now
      and write messages to the user */

   if (return_sort != TRUE)
    {
      /* Two keys can have the same sequence. */

      return_sort = TRUE;

    }  /* end if return not TRUE */
   return return_sort;
 }
