
#ifndef   VOS
#include "all.h"
#endif


/*
   #title Get String Attribute
*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *   get_string_attrib -- Gets a string stripped of padding info.       *
 *                        Returns length of string.                     *
 *                                                                      *
 *   This routine is called to get the string associated with a partic- *
 *   ular string attribute. It strips off padding information , fills   *
 *   the string into a user defined area, and returns the number of     *
 *   characters in the string.  The returned length is the total length *
 *   of the translated string , not exceeding the maximum return string *
 *   length parameter.  If the returned string does not overflow, the   *
 *   returned string is NULL terminated, but the null byte is not       *
 *   counted in the returned length.                                    *
 *                                                                      *
 *   If the returned length = 0 , then the string was NULL.             *
 *   If the returned length = -1, then the maximum length               *
 *                                was exceeded.                         *
 *                                                                      *
 *----------------------------------------------------------------------*/

/*-----------------------------------------------------------------------*
 * Parameter List:                                                       *
 * ---------------                                                       *
 *                                                                       *
 * terminal_def (input)struct tdef_heap *                                *
 *              structure of pointers into heap[ containing the terminal *
 *              definition.                                              *
 *                                                                       *
 * str_attrib   (input) int                                              *
 *              offset into string offsets section corresponding to the  *
 *              desired attribute.                                       *
 *                                                                       *
 *  out_str     (output) unsigned char *                                 *
 *              user area in which to store translated string.           *
 *  max_out_len (input) int                                              *
 *              maximum string length returned.                          *
 *  pad_ptr     (output) int *                                           *
 *              pointer to area in which to store padding factor.        *
 *                         -will be zero if no padding required.         *
 *                         -if NULL, then no padding info is given.      *
 *  line_ptr    (input/output) int *                                     *
 *              - INPUT  : if this is a NULL pointer, then the amount of *
 *                         padding cannot depend on line length, and a   *
 *                         "*" following the pad count or a "." in the   *
 *                         pad count will be ignored.                    *
 *              - OUTPUT : if a NULL is returned then no padding         *
 *                         information is given. Otherwise it points to  *
 *                         the area to store the padding - line count    *
 *                         factor.                                       *
 *   (Example:                                                           *
 *      attribute xxxxx will give pad_ptr =0,  line_ptr =0,  out =xxxxx  *
 *             $<16>xxx will give pad_ptr =16, line_ptr =0,  out =xxx    *
 *             $<16*>xx will give pad_ptr =16, line_ptr =1,  out =xx     *
 *             $<1.6*>x will give pad_ptr =16, line_ptr =10, out =x      *
 *                                                                       *
 *      if line_ptr is NULL the last two examples would give instead:    *
 *                                                                       *
 *             $<16*>xx will give pad_ptr =16, out_attrib =*xx           *
 *             $<1.6*>x will give pad_ptr =1,  out_attrib =.6*x          *
 *                                                                       *
 *-----------------------------------------------------------------------*/


int get_string_attrib( term_def,
                       string_attrib,
                       out_string,
                       max_out_length,
                       pad_ptr,
                       line_ptr)
struct tdef_heap  *term_def      ;
int                string_attrib ;
unsigned char     *out_string    ;
int                max_out_length;
int               *pad_ptr    ;
int               *line_ptr   ;

{
unsigned char  *offset    ;
int            *str_offset;                                         /*<TKW-07>*/
char   pad_holder[132]    ;
char   c                  ;
int    state              ;
int    pad                ;
int    got_padding        ;
int    pad_len            ;
int    lines              ;
int    len                ;
int    i                  ;



str_offset = term_def->string_offsets_section + string_attrib;      /*<TKW-07>*/
offset =  term_def->strings_section + *str_offset;                  /*<TKW-07>*/
len = 0;
pad = 0;
lines = 0;
if (pad_ptr != NULL)
   *pad_ptr = pad;
if (line_ptr != NULL)
   *line_ptr = lines;
got_padding = !DONE;
pad_len = 0;
if ((*str_offset == -1) ||                                          /*<TKW-07>*/
    ( str_offset >= (int *)term_def->strings_section))              /*<TKW-07>*/
   /* string undefined  or not recognized  by tic or tdc     */
   state = END_OF_STRING;
else
   state = NORMAL;
while (state != END_OF_STRING)
   {
   if (offset > term_def->tdef_heap_end)
      {
      state = END_OF_STRING;

#ifdef USE_NOIO
      IOW (&SCREEN, &SYM_WRITE, &53,
           " Tried to search in string section past end of heap.\n",&IO_A_reg);
#else
      printf(" Tried to search in string section past end of heap.\n");
#endif
      }
   if (len >= max_out_length)
      {
      state = END_OF_STRING;
      got_padding = DONE;
   /* printf(" String length exceeds maximum specified.\n"); */
      }
   c = *offset++;
   if (c == '\0')
      state = END_OF_STRING;
   switch(state)
      {
      case NORMAL :
         switch(c)
            {
            case '$' :
               if (got_padding != DONE)
                  {
                  state = DOLLAR_SIGN;
                  pad_holder[pad_len++] = c;
                  }
               break;
            default :
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               break;
            }
         break;

      case DOLLAR_SIGN :
         switch(c)
            {
            case '<' :
               state = L_BRACKET;
               pad_holder[pad_len++] = c;
               break;
            case '$' :
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               pad_len = 0;
               break;
            }
         break;

      case L_BRACKET :
         switch(c)
            {
            case '0' :
            case '1' :
            case '2' :
            case '3' :
            case '4' :
            case '5' :
            case '6' :
            case '7' :
            case '8' :
            case '9' :
               state = DIGIT;
               pad_holder[pad_len++] = c;
               pad = (pad * 10) + (int)(c - '0');
               break;
            case '$' :
               state = DOLLAR_SIGN;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               pad_len = 0;
               break;
            }
         break;
      case DIGIT :
         switch(c)
            {
            case '0' :
            case '1' :
            case '2' :
            case '3' :
            case '4' :
            case '5' :
            case '6' :
            case '7' :
            case '8' :
            case '9' :
               state = DIGIT;
               pad_holder[pad_len++] = c;
               pad = (pad * 10) + (int)(c - '0');
               break;
            case '$' :
               state = DOLLAR_SIGN;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            case '.' :
               state = DEC_POINT;
               pad_holder[pad_len++] = c;
               break;
            case '*' :
               state = STAR;
               pad_holder[pad_len++] = c;
               lines = 1 ;
               break;
            case '>' :
               state = NORMAL;
               got_padding = DONE;
               lines = 0;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               pad_len = 0;
               break;
            }
         break;
      case DEC_POINT :
         switch(c)
            {
            case '0' :
            case '1' :
            case '2' :
            case '3' :
            case '4' :
            case '5' :
            case '6' :
            case '7' :
            case '8' :
            case '9' :
               state = FRACTION;
               pad_holder[pad_len++] = c;
               pad = (pad * 10) + (int)(c - '0');
               break;
            case '$' :
               state = DOLLAR_SIGN;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               pad_len = 0;
               break;
            }
         break;
      case FRACTION :
         switch(c)
            {
            case '*' :
               state = STAR;
               pad_holder[pad_len++] = c;
               lines = 10 ;
               break;
            case '$' :
               state = DOLLAR_SIGN;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               pad_len = 0;
               break;
            }
         break;
      case STAR :
         switch(c)
            {
            case '>' :
               state = NORMAL;
               got_padding = DONE;
               break;
            case '$' :
               state = DOLLAR_SIGN;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               pad_len = 0;
               pad_holder[pad_len++] = c;
               break;
            default :
               state = NORMAL;
               for( i = 0; i < pad_len; i++)
                  {
                  len++;
                  if (len < max_out_length)
                     *out_string++ = pad_holder[i];
                  }
               len++ ;
               if (len < max_out_length)
                  *out_string++ = c;
               pad_len = 0;
               break;
            }
         break;
      case END_OF_STRING :
         if (got_padding != DONE)
            for( i = 0; i < pad_len; i++)
               {
               len++;
               if (len < max_out_length)
                  *out_string++ = pad_holder[i];
               }
          break;
      }
   }
if (line_ptr != NULL)
   {
   *line_ptr = lines;
   if ((pad_ptr != NULL)&&(got_padding == DONE))
      *pad_ptr = pad;
   }
if (len >= max_out_length)
   len = -1;
else
   *out_string = '\0';             /* null terminate  string */

return(len);
}
