
#ifndef   VOS
#include "all.h"
#endif

/*
   #title key sequence FSA table builder
*/

/*----------------------------------------------------------------------*
 *                                                                      *
 *                                                                      *
 *         P R O P R I E T A R Y  AND  C O N F I D E N T I A L          *
 *                                                                      *
 *                                                                      *
 *       HARRIS COMPUTER SYSTEMS DIVISION OF HARRIS CORPORATION         *
 *                                                                      *
 *    The software contained herein is furnished pursuant a license     *
 *    agreement for use with a single HCSD computer and cannot be       *
 *    copied except as provided in the license agreement.               *
 *                                                                      *
 *----------------------------------------------------------------------*/

   /*----------------------------------------*
    *  terminal definition reading routines  *
    *----------------------------------------*/
/*********************************************************
   extern struct tdef_heap  *get_term_def (char *);
      /* Finds a terminal definition and stores it
         on the heap.
         Returns a pointer to a stucture containing
         pointers to various sections of the definition.
   extern int     get_boolean_attrib (struct tdef_heap *, int);
      /* Gets the value of a boolean entry from the
         terminal definition entry.
         Returns TRUE or FALSE.
         Parameters are pointer to structure of pointers into terminal
         definition entry and integer defined to represent the desired
         boolean attribute.
   extern int     get_numeric_attrib (struct tdef_heap *, int);
      /* Gets the value of a numeric entry from the
         terminal definition entry.
         Returns value for attribute or negative one (if undefined).
         Parameters are pointer to structure of pointers into terminal
         definition entry and integer defined to represent the desired
         numeric attribute.
   extern int   get_string_attrib (struct tdef_heap *,int,unsigned char [],
                                   int,int *,int*);
      /* Finds and translates a string attribute.
         Returns the length of the translated string.
         Parameters are pointer to terminal entry,
         integer corresponding to attribute to search for,
         pointer to string to store the translated definition in,
         the maximum number of characters to output to the
         translated string, pointers to places to store number of
         pads needed and ratio per line.

         ***********************************************************/


   /*------------------------------------------------------------*
    *          ++++++ build_tables routine ++++++                *
    *  called from :                                             *
    *    read_configuration                                      *
    *                                                            *
    *  function :                                                *
    *    retrieves key sequences from the terminal description   *
    *    entry and creates the FSA tables from them.             *
    *                                                            *
    *  arguments :                                               *
    *    int []     : attributes for keys in terminal desc.      *
    *    int []     : value FSA should return for each key       *
    *    int []     : "required" key flag                        *
    *    int        : number of key attributes                   *
    *    struct tdef_heap *  : structure containing pointers to  *
    *                          FSA start table, FSA next table,  *
    *                          and terminal definition sections  *
    *                                                            *
    *  returns :                                                 *
    *    TRUE if tables are created ok.                          *
    *    FALSE if an error occurred that prohibits creation of   *
    *       FSA tables.                                          *
    *------------------------------------------------------------*/

int build_tables (key_attribs, key_return, key_mandatory, attrib_count,
                  terminal_def)

   int                     key_attribs [] ;
   int                     key_return [];
   int                     key_mandatory [];
   int                     attrib_count;
   struct tdef_heap       *terminal_def;
 {
   /*-------------------*
    *  called routines  *
    *-------------------*/

   int            get_key_sequences ();
                  /* retrieves key sequences from the terminal
                     definition.  Returns TRUE if all
                     went ok, FALSE if not. */

   int            key_sort ();
                  /* eliminates undefined keys and sorts the
                     sequences array.  key_attribs and key_return
                     are updated to match the new placement in
                     sequences.  attrib_count is updated.  */

   int            make_tables ();
                  /* creates the actual FSA tables */

   int            dealloc_char_array ();
                  /* deallocates a 2-d char array */


   /*-------------------*
    *  local variables  *
    *-------------------*/

   unsigned char **sequences;     /* ptr to array of ptrs to key sequences */
   int           return_value;    /* value to return to caller */

   /*------------------------------------------------------*
    *  Get the key sequences.  If all mandatory ones were  *
    *  retrieved, the FSA table creation is attempted.     *
    *------------------------------------------------------*/

   if ((get_key_sequences (key_attribs, key_mandatory,
                           attrib_count, &sequences, terminal_def))
      &&
       (key_sort (key_attribs, key_return, &attrib_count,
                    sequences)))

          return_value = make_tables (key_attribs, key_return,
                                      attrib_count, sequences,
                                      terminal_def);
   else
      return_value = FALSE;

   /*----------------------------------*
    *  deallocate the sequences array  *
    *----------------------------------*/

   dealloc_char_array ( (char **) sequences);

   return (return_value);
 }






   /*------------------------------------------------------------*
    *          ++++++ get_key_sequences routine ++++++           *
    *  called from :                                             *
    *    build_tables                                            *
    *                                                            *
    *  function :                                                *
    *    retrieve the indicated key sequences from the terminal  *
    *    definition string.                                      *
    *                                                            *
    *  arguments :                                               *
    *    int []     : attributes for keys in terminal desc.      *
    *    int []     : "required" key flag                        *
    *    int        : number of key attributes                   *
    *    unsigned char ***   : where to stick pointer to array   *
    *                          of pointers to key sequences.     *
    *    struct tdef_heap *  : structure of pointers to sections *
    *                          of terminal definition            *
    *                                                            *
    *  returns int :                                             *
    *    TRUE if all sequences were found                        *
    *    FALSE if not.                                           *
    *                                                            *
    *------------------------------------------------------------*/





int get_key_sequences (key_attrib, key_mandatory, attrib_count,
                       sequences, terminal_def)
   int                     key_attrib [];
   int                     key_mandatory [];
   int                     attrib_count;
   unsigned char * * *     sequences;
   struct tdef_heap       *terminal_def;

 {
   /*-------------------*
    *  called routines  *
    *-------------------*/

   unsigned char **alloc_char_array ();                             /*<TKW-08>*/
      /* allocates a 2 dimensional character array */

   /*-------------------*
    *  local variables  *
    *-------------------*/

   int            found_string_size;   /* size of string from term def */
   int            index_1;             /* general purpose indices */
   int            return_value;        /* value to return to caller */

   /*--------------------------------------------------*
    *  allocate the space for the key sequence table.  *
    *--------------------------------------------------*/


   *sequences = alloc_char_array (attrib_count, KEY_SEQUENCE_SIZE); /*<TKW-08>*/


   /*--------------------------------------------------------*
    *  Loop here for each key.  Assign the sequence pointer  *
    *  and put the sequence in the table.                    *
    *--------------------------------------------------------*/
   return_value = TRUE;
   for (index_1 = 0; index_1 < attrib_count; ++index_1)
    {
      found_string_size =
            get_string_attrib (terminal_def, key_attrib[index_1],
                         (unsigned char *)(*sequences)[index_1],
                          KEY_SEQUENCE_SIZE, NULL, NULL);
      if ((found_string_size == 0) && (key_mandatory [index_1]))
         {
#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &41,
                  " String attribute for key was not found.\n",&IO_A_reg);
#else
         printf (" String attribute for key %d was not found.\n",
                 key_attrib [index_1]);
#endif
         return_value = FALSE;
         }
      else if (found_string_size == -1)
         {
#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &39,
                  " String attribute for key is too long.\n",&IO_A_reg);
         IOW (&SCREEN, &SYM_WRITE, &34,
                  " Maximum allowed length exceeded.\n",&IO_A_reg);
#else
         printf (" String attribute for key %d is too long.\n",
                 key_attrib [index_1]);
         printf ("    Maximum allowed length is %d, attribute length is %d.\n",
                 KEY_SEQUENCE_SIZE, found_string_size);
#endif
         return_value = FALSE;
         }
      else if (found_string_size == KEY_SEQUENCE_SIZE)
         {
#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &39,
                  " String attribute for key is too long.\n",&IO_A_reg);
         IOW (&SCREEN, &SYM_WRITE, &34,
                  " Maximum allowed lenght exceeded.\n",&IO_A_reg);
         IOW (&SCREEN, &SYM_WRITE, &34,
                  "    -no room for null terminator.\n",&IO_A_reg);
#else
         printf (" String attribute for key %d is too long.\n",
                 key_attrib [index_1]);
         printf ("    Maximum allowed length is %d, attribute length is %d.\n",
                 KEY_SEQUENCE_SIZE, found_string_size);
         printf ("    -no room for null terminator.\n");
#endif
         return_value = FALSE;
         }

    } /* end for each key */
   return (return_value);
 }




/* #title make_tables     */

   /*------------------------------------------------------------*
    *           ++++++ make_tables routine ++++++                *
    *  function :                                                *
    *    Creates start_table and next_table from the key         *
    *    definitions the passed arbuments.                       *
    *                                                            *
    *  arguments :                                               *
    *    int []     : attributes for keys in terminal desc.      *
    *    int []     : value FSA should return for each key       *
    *    char * *   : pointer to sequence pointers array         *
    *    int        : number of key attributes                   *
    *    struct tdef_heap *  : structure of pointers to sections *
    *                          of terminal definition & FSA      *
    *                          start_ and next_ tables           *
    *                                                            *
    *  returns :                                                 *
    *    TRUE if all went ok.                                    *
    *    FALSE if an error occurred that prohibits creation of   *
    *       FSA tables.                                          *
    *------------------------------------------------------------*/

int make_tables (key_attribs, key_return, attrib_count, sequences,
                 terminal_def)

   int                     key_attribs [] ;
   int                     key_return  [] ;
   int                     attrib_count;
   unsigned char * *       sequences;
   struct tdef_heap       *terminal_def;

 {
   /*-------------------*
    *  called routines  *
    *-------------------*/

   unsigned char           **alloc_char_array ();
            /* allocates a 2-d unsigned char array in the heap */

   struct START_TABLE      *start_table_overflow ();
            /* re-allocates the start table */

   struct NEXT_TABLE       *next_table_overflow ();
            /* re-allocates the next table */

   int                     dealloc_char_array ();
            /* de-allocates a 2 dimensional char array */

#ifndef USE_NOIO
   int                     print_tables ();
            /* prints out the tables to w1 */
#endif

   /*-------------------*
    *  local variables  *
    *-------------------*/

   unsigned char  * * state_prefix;
                           /* defines the byte sequence that leads into the
                              given state */
   int            current_state;
                           /* the state whose exiting
                              transitions are being formed */
   int            next_undefined_state;
                           /* the state number to assign
                              to a newly found state */
   int            num_entries;
                           /* number of entries for the
                              current state in next_table */
   int            next_index;         /* index into next_table */
   int            sequences_index_1 ; /* first index into sequences */
   int            compare_length;     /* length of current state prefix */
   unsigned char  last_unique_byte;
                           /* last unique byte following current sequence */
   int            start_table_size;
                           /* current size of start table */
   int            next_table_size;
                           /* current size of next table */
   int            start_table_space;
                           /* current space (bytes) alloc'ed */
   int            next_table_space;
                           /* current space (bytes) alloc'ed */
   int            return_value;
                           /* value to return to caller */
   int            compare ;
                           /* results of string comparison */

#ifndef USE_NOIO
   int            debug;
#endif



   /* see if there is anything to do */
   if (attrib_count == 0)
      return FALSE;  /* no keys defined, just get out */

   /*-----------------------*
    *  allocate some space  *
    *-----------------------*/

   terminal_def->start_table =
                             (struct START_TABLE *) malloc (START_TABLE_SPACE);
   terminal_def->next_table  =
                             (struct NEXT_TABLE  *) malloc (NEXT_TABLE_SPACE);

   start_table_size = START_TABLE_SIZE;
   next_table_size = NEXT_TABLE_SIZE;
   start_table_space = START_TABLE_SPACE;
   next_table_space = NEXT_TABLE_SPACE;

   /*----------------------------------------------------------*
    *  alloc space for next table and initialize the pointers  *
    *  within.                                                 *
    *----------------------------------------------------------*/
                                                                    /*<TKW-08>*/
   state_prefix = alloc_char_array (START_TABLE_SPACE / 2, KEY_SEQUENCE_SIZE);

   /* initialize */

#ifndef USE_NOIO
   debug = FALSE;
#endif

   return_value = TRUE;
   next_index = 0;
   next_undefined_state = 1;
   state_prefix [0] [0] = NULL;  /* start state has no prefix */
   sequences_index_1 = 0;

   /*--------------------------------------------------*
    *  One pass of this loop defines all the outgoing  *
    *  transitions for the current state.              *
    *--------------------------------------------------*/

   for (current_state = 0 ;  current_state < next_undefined_state ;
        ++current_state)
    {
      /*--------------------------------------------------*
       * find each unique byte that follows the sequence  *
       * to this state. If it is terminal, insert the     *
       * return information into next_table.  Otherwise,  *
       * create a new state for the new sequence and      *
       * insert the non_terminal information into         *
       * next_table.  num_entries is always incremented.  *
       *------------------------------------------------- */

      num_entries = 0;
      terminal_def->start_table[current_state].first_entry = next_index;
      compare_length = strlen (state_prefix [current_state]);
      last_unique_byte = NULL;

      /* position ourselves in sequences so we're looking at
         the first entry with state_prefix [current_state] */

      sequences_index_1 = 0;

      compare = newstrncmp ( (char *) state_prefix [current_state] ,
              (char *) sequences [sequences_index_1], compare_length) ;

      while((sequences_index_1 < attrib_count) && (compare != 0))
         {
         ++sequences_index_1;

         compare = newstrncmp ( (char *) state_prefix [current_state] ,
                 (char *) sequences [sequences_index_1], compare_length) ;

         }

      if (sequences_index_1 >= attrib_count)
       {
#ifdef USE_NOIO
         IOW (&SCREEN, &SYM_WRITE, &54,
         " ***** INTERNAL ERROR - cannot create key FSA tables.\n",&IO_A_reg);
#else
         printf ( " ***** INTERNAL ERROR - cannot create key FSA tables.\n");
#endif
         return FALSE;
        }

      /* now, we can loop */
      while (sequences_index_1 < attrib_count  &&
             newstrncmp ( (char *) &state_prefix [current_state] [0],
             (char*) &sequences [sequences_index_1] [0], compare_length) == 0)
        {
         /* does this byte define a new transition from current state? */

          if (sequences [sequences_index_1] [compare_length] !=
              last_unique_byte)
           {
            /* yes - make an entry for it in next table */
            last_unique_byte =
                sequences [sequences_index_1] [compare_length];

            terminal_def->next_table[next_index].input_byte =
                                                last_unique_byte;
            /* does this byte lead to a terminal state ? */
            if (sequences [sequences_index_1] [compare_length + 1]
                == NULL)
             {
               /* yes, mark as terminal */
               terminal_def->next_table[next_index].next_state = -1;
               terminal_def->next_table[next_index].return_key =
                                        key_return [sequences_index_1];
             }
            else
             {
               /* no, create a new state and indicate the prefix to
                  get there */
               /* check start table overflow */
               if (next_undefined_state >= start_table_size)
                  terminal_def->start_table =
                                start_table_overflow (&start_table_size,
                                                      &start_table_space,
                                                   terminal_def->start_table,
                                                      &state_prefix);

               terminal_def->next_table[next_index].next_state =
                                                     next_undefined_state;
               terminal_def->next_table[next_index].return_key =
                                                             UNDETERMINED;
               newstrncpy ((char *) &state_prefix [next_undefined_state] [0],
                        (char *) &sequences [sequences_index_1] [0],
                        compare_length + 1);
               /* add a NULL so a compare will work */
               state_prefix [next_undefined_state] [compare_length+1]
                  = (unsigned char) NULL;
               ++next_undefined_state;
             }
            ++num_entries; /* inc num of entries for this state */
            ++next_index;  /* next entry goes in next slot */

            /* check for next_table overflow */
            if (next_index >= next_table_size)
               terminal_def->next_table =
                             next_table_overflow (&next_table_size,
                                                  &next_table_space,
                                                  terminal_def->next_table);
          }
         else
          {
            /* check for non-determinism with previous sequence */
            if (sequences [sequences_index_1 - 1] [compare_length + 1] == NULL
               && sequences [sequences_index_1] [compare_length + 1] != NULL)
             {
#ifdef USE_NOIO
               IOW (&SCREEN, &SYM_WRITE, &32,
                        " ***** ERROR - defined sequences\n",&IO_A_reg);
               IOW (&SCREEN, &SYM_WRITE, &41,
                        "      make key determination impossible.\n",&IO_A_reg);
               key_attribs[1] = key_attribs[1]; /* get rid of WaRNING */
#else
               printf (
                  " ***** ERROR - the defined sequences for %5d and %5d\n",
                  key_attribs [sequences_index_1 - 1],
                  key_attribs [sequences_index_1] );
               printf (
                  "       make key determination impossible.\n");
#endif
               return_value = FALSE;
             }
          }
         ++sequences_index_1; /* examine next sequence */
       } /* end while */

      /* make the "else" entry into the table */
      terminal_def->next_table[next_index].input_byte = (unsigned char) -1;
      terminal_def->next_table[next_index].next_state =  -1;
      terminal_def->next_table[next_index].return_key = INVALID_SEQUENCE;
      ++num_entries;
      ++next_index;

      /* check for next_table overflow */
      if (next_index >= next_table_size)
         terminal_def->next_table =
                           next_table_overflow (&next_table_size,
                                                &next_table_space,
                                                terminal_def->next_table);

      /* complete the start table entry */
      terminal_def->start_table[current_state].number_of_entries =
                                                       num_entries;
    } /* end of for loop */

   /*-------------------------------------*
    *  deallocate the state prefix table  *
    *-------------------------------------*/

   dealloc_char_array ( (char **) state_prefix);

#ifndef USE_NOIO

   if (debug)
      print_tables (terminal_def, start_table_size,next_table_size);

#endif

   return return_value;
 }






/*   #title start table overflow  */

   /*------------------------------------------------------------*
    *         ++++++ start_table_overflow routine ++++++         *
    *  called from :                                             *
    *    make_tables                                             *
    *                                                            *
    *  function :                                                *
    *    re-allocate the start table and state prefix table      *
    *                                                            *
    *  arguments :                                               *
    *    int * : pointer to current size of table                *
    *    int * : pointer to current space in bytes of table      *
    *    struct START_TABLE * : ptr to array current allocated   *
    *                           space                            *
    *    unsigned char *** : ptr to state prefix string array    *
    *                                                            *
    *  returns struct START_TABLE * :                            *
    *    pointer to new allocation                               *
    *------------------------------------------------------------*/
struct START_TABLE * start_table_overflow (current_size, current_space,
                                           current_ptr, current_state_prefix)
   int                  *current_size;
   int                  *current_space;
   struct START_TABLE   *current_ptr;
   unsigned char        ***current_state_prefix;
 {
   unsigned char        **realloc_char_array ();
            /* reallocates a 2-d unsigned char array in the heap */

   /*---------------------------*
    *  calculate the new sizes  *
    *---------------------------*/

   *current_size += START_TABLE_SIZE_INCR;
   *current_space += START_TABLE_SPACE_INCR;

   /*-------------------------------------*
    *  reallocate the state prefix table  *
    *-------------------------------------*/

   *current_state_prefix = (unsigned char **)
      realloc_char_array (*current_space / 2,
                          KEY_SEQUENCE_SIZE,
                          (char **) *current_state_prefix);

   /*--------------------------------------------------------*
    *  reallocate and return the pointer to the start table  *
    *--------------------------------------------------------*/

   return (struct START_TABLE *)
      newalloc ( (char *) current_ptr, *current_space);
 }





/* #title next table overflow   */

   /*------------------------------------------------------------*
    *          ++++++ next_table_overflow routine ++++++         *
    *  called from :                                             *
    *    make_tables                                             *
    *                                                            *
    *  function :                                                *
    *    re-allocate the next table                              *
    *                                                            *
    *  arguments :                                               *
    *    int * : pointer to current size of table                *
    *    int * : pointer to current space in bytes of table      *
    *    struct NEXT_TABLE * : ptr to array current allocated    *
    *                           space                            *
    *                                                            *
    *  returns struct NEXT_TABLE * :                             *
    *    pointer to new allocation                               *
    *------------------------------------------------------------*/
struct NEXT_TABLE * next_table_overflow (current_size, current_space,
                                         current_ptr)
   int               *current_size;
   int               *current_space;
   struct NEXT_TABLE *current_ptr;
 {
   /*---------------------------*
    *  calculate the new sizes  *
    *---------------------------*/

   *current_size += NEXT_TABLE_SIZE_INCR;
   *current_space += NEXT_TABLE_SPACE_INCR;

   /*-------------------------------------------------------*
    *  reallocate and return the pointer to the next table  *
    *-------------------------------------------------------*/

   return (struct NEXT_TABLE *)
      newalloc ( (char *) current_ptr, *current_space);
 }

int                                                                 /*<TKW-08>*/
newstrncmp ( s1, s2, n)                                             /*<TKW-08>*/
char *s1;                                                           /*<TKW-08>*/
char *s2;                                                           /*<TKW-08>*/
int n;                                                              /*<TKW-08>*/

{
#ifdef USE_NOIO
   int i;

   for (i=0 ; i<n ; i++) {
      if (s1[i] < s2[i]) return (-1);
      if (s1[i] > s2[i]) return (1);
   }
   return (0);
#else
   return ( (int)strncmp( s1, s2, n ) );                            /*<TKW-08>*/
#endif
}

char *
newstrncpy ( s1, s2, n)
char *s1;
char *s2;
int n;

{


#ifdef USE_NOIO

   int i, eos2;

   eos2 = FALSE;
   for (i=0 ; i<n ; i++) {
      if (eos2)
       s1[i] = NULL;
      else {
         s1[i] = s2[i];
         if (s2[i] == NULL) eos2 = TRUE;
      }
   }
   return (s1);

#else

extern char *strncpy();                                             /*<TKW-08>*/
   return ( strncpy ( s1, s2, n ) );

#endif
}





#ifndef USE_NOIO
/* #title print_tables */

   /*------------------------------------------------------------*
    *             ++++++ print_tables routine ++++++             *
    *  called from :                                             *
    *    make_tables                                             *
    *                                                            *
    *  function :                                                *
    *    print the tables to w1                                  *
    *                                                            *
    *  arguments :                                               *
    *    struct START_TABLE * : start table                      *
    *    int : current size of start table                       *
    *    struct NEXT_TABLE * : next table                        *
    *    int * : current size of next table                      *
    *------------------------------------------------------------*/

print_tables (table_ptrs, start_table_size, next_table_size)

   struct tdef_heap   *table_ptrs;
   int                start_table_size;
   int                next_table_size;
 {
   int      index;
   FILE *   listing;


   /* open the listing file */

   listing = fopen ("W5", 'w');
   if (listing == NULL)
      return;

   /* start table */
   fprintf (listing,
     "1  start_table \n state first entry   number of entries\n\n");
   for (index = 0;  index < start_table_size; ++index)
    {
      fprintf (listing,
       "    %-4d  %-4d           %-4d\n",
       index,
       table_ptrs->start_table [index].first_entry,
       table_ptrs->start_table [index].number_of_entries);
    }


   /* start table */
   fprintf (listing,
     "0  next_table \n entry  input char   next state   return key\n\n");
   for (index = 0;  index < next_table_size; ++index)
    {
      fprintf (listing,
         "    %-4d  '%-3o         %-4d        %-4d\n",
         index,
         table_ptrs->next_table [index].input_byte,
         table_ptrs->next_table [index].next_state,
         table_ptrs->next_table [index].return_key);
    }
#ifdef VOS
   HEAP_CHECK(listing);
#endif
   return;
 }
#endif
