/*	boot.c	6.1	83/07/29	*/

#define KERNEL
#include "../uts/machine/sys/param.h"
#undef  KERNEL
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../uts/machine/sys/vm.h"
#include "../uts/machine/ml/mtpr.h"
#include <filehdr.h>
#include <aouthdr.h>
#include <scnhdr.h>
#include "../stand/saio.h"
#include "../uts/machine/sys/reboot.h"

/*
 * Boot program... arguments passed in r10 and r11 determine
 * whether boot stops to ask for system name and which device
 * boot comes from.
 */

/*	r11 = 0 -> automatic boot, load file '/unix' */
/*	r11 = 1 -> ask user for file to load */

/* r10 specifies the boot device for auto-booting. */
/* CP version 1 has the following definition:      */
/*+--------------------------------------------------------------------------+*/
/*| reserved|version|partition|bus types| type   | bus #  | ctlr # | drive # |*/
/*|  4 bits | 4 bits|  4 bits |  4 bits | 4 bits | 4 bits | 4 bits |  4 bits |*/
/*+--------------------------------------------------------------------------+*/

/* CP version 2 has the following defintion:       */
/*+--------------------------------------------------------------------------+*/
/*| reserved|version|partition|bus types| type   | bus #  | ctlr # | drive # |*/
/*|  4 bits | 4 bits|  4 bits |  4 bits | 4 bits | 2 bits | 6 bits |  4 bits |*/
/*+--------------------------------------------------------------------------+*/

#define TEXT_BEGIN (FILHSZ + sizeof(struct aouthdr) + 3*sizeof(struct scnhdr))

int     howto ;
int     boot_device ;
int	version ;
int     buses[ 2 ] ;
char	line[ 100 ] ;
char *	line_ptr ;
int	retry = 0 ;
int	io ;

main()
{
	register int	r12 ;		/* skip r12    */
	register int	r11 ;		/* r11 from cp */
	register int	r10 ;		/* r10 from cp */

	howto = r11 ;
	boot_device = r10 ;
	printf("\nBoot\n");
	version = (boot_device & 0xf000000) >> 24 ;
	if ( version == 0 ) {
		printf( "************************************************\n");
		printf( "*                                              *\n");
		printf( "*            I M P O R T A N T !               *\n");
		printf( "*                                              *\n");
		printf( "*        CP proms are out of revision.         *\n");
		printf( "*                                              *\n");
		printf( "*    Please, upgrade to the latest revision.   *\n");
		printf( "*                                              *\n");
		printf( "************************************************\n");
		howto |= RB_ASKNAME ;
	}

	buses[ 0 ] = RB_PRIMARY_IO( boot_device ) ;
	if ( buses[ 0 ] == RB_NO_BUS )
		buses[ 0 ] = RB_VERSA_BUS ;
	buses[ 1 ] = RB_SECONDARY_IO( boot_device ) ;

	for (;;) {
		printf( ": " ) ;
		if ( howto & RB_ASKNAME ) {
			line_ptr = line ;
			gets( line_ptr ) ;
		} else {
			line_ptr = "unix" ;
			printf( "%s\n", line_ptr ) ;
		}
		while ( *line_ptr == ' ' )
			line_ptr ++ ;
		if ( *line_ptr != 0 ) {
			io = open( line_ptr, 0 ) ;
			if ( io >= 0 )
				copyunix( io, howto ) ;
		}
		if ( ++ retry > 2 )
			howto |= RB_SINGLE | RB_ASKNAME ;
	}
}

/*ARGSUSED*/
copyunix( io, howto )
	register int io ;
	register int howto;
{
	register int r10 ;	/* preserve r10 from CP */
	FILHDR filehdr;
	AOUTHDR x;
	register int i;
	char *addr;

	i = read(io, (char *)&filehdr, FILHSZ);
	if (i != FILHSZ ||
	    (filehdr.f_magic != HCXWRMAGIC && filehdr.f_magic != HCXROMAGIC))
		_stop("Bad format\n");
	i = read(io, (char *)&x, sizeof x);
	if (i != sizeof x ||
	    (x.magic != 0407 && x.magic != 0413 && x.magic != 0410))
		_stop("Bad format\n");
	printf("%d", x.tsize);
	/*
	 * A magic number of 0413 indicates that the text and data
	 * sections start on page boundaries in the disk image.
	 */
	if (x.magic == 0413 && lseek(io, 0x400, 0) == -1)
		goto shread;
	if (x.magic != 0413 && lseek(io, TEXT_BEGIN, 0) == -1)
		goto shread;
	if (x.magic == 0413 && (read(io, (char *)0x800,
		    (x.tsize+CLOFSET)&~CLOFSET) != (x.tsize+CLOFSET)&~CLOFSET))
		goto shread;
	if (x.magic != 0413 && read(io, (char *)0x800, x.tsize) != x.tsize)
		goto shread;
	addr = (char *)(x.tsize + 0x800);
	if (x.magic == 0413 || x.magic == 0410)
		while ((int)addr & CLOFSET)
			*addr++ = 0;
	printf("+%d", x.dsize);
	if (read(io, addr, x.dsize) != x.dsize)
		goto shread;
	addr += x.dsize;
	printf("+%d", x.bsize);
	x.bsize += 32*1024;	/* slop */
	for (i = 0; i < x.bsize; i++)
		*addr++ = 0;
	x.entry &= 0x1fffffff;
	printf(" start 0x%x\n", x.entry);
	mtpr (0, PADC);		/* Purge data cache */
	mtpr (0, PACC);		/* Purge code cache */

	if (howto & RB_DCOFF) {
		mtpr (2, DCR); /* Disable data cache */
	} else {
		mtpr (1, DCR);	/* Enable data cache */
	}
	dummy();

	if ( howto & RB_LOAD_ONLY ) {
		printf("Load-Only option set (VM NOT enabled).\n\n") ;
	        asm( "halt" ) ;
	}

	(*((int (*)()) x.entry))();
shread:
	_stop("Short read\n");
}

dummy() {}
