/*
 *  format.c  --  format/verify/etc program for disks on an HDC
 *
 *		  This source file is used to produce either a unix
 *		  version or a standalone version. If STANDALONE is
 *		  defined, then the standalone version is produced,
 *		  otherwise the unix version.
 */

#define KERNEL

#include "../../include/values.h"
#include "../../include/fcntl.h"
#include "../uts/machine/sys/param.h"
#include "../uts/machine/sys/buf.h"

#ifdef STANDALONE
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../stand/saio.h"
#include "../uts/machine/sys/reboot.h"

#else
#include "../uts/machine/sys/errno.h"
#endif

#include "../uts/machine/sys/types.h"
#include "../../include/stdio.h"
#include "../../include/ctype.h"
#include "../uts/machine/sys/signal.h"
#include "../uts/machine/sys/ioctl.h"
#include "../uts/machine/sys/dsk.h"
#include "../uts/machine/sys/dskio.h"
#include "../uts/machine/sys/hdc.h"


int	certify();
int	certify_option();
int	data_com();
int	data_option();
int	data_proc();
int	format();
int	format_option();
int	format_proc();
int	help();
int	intr();
int	list();
int	list_option();
int	list_proc();
int	partition();
int	pcom();
int	quit();
int	readg();
int	read_vendor_data();
int	status();
int	unit();
int	verify();
int	verify_option();
int	verify_proc();
int	writeg();
int	yes();

char    usage_err[] = "usage: format [option] diskunit\n";

/*
 * This global data is general information about the disk.
 * They are initialized when the program starts and are
 * intended to be read-only.
 */

int	boot_device ;		/* used only by stand-alone version      */
int	version ;		/* used only by stand-alone version      */
int     buses[ 2 ] ;		/* used only by stand-alone version      */
dsk_status disk_status;		/* disk status from DSKIOCSTATUS ioctl   */
int	def_sectors;		/* no. of sectors in disk def partition  */
int	spc;			/* total number of sectors per cylinder  */
int	total_sectors;		/* total number of sectors on the disk   */
int	fd;			/* file descriptor of the disk           */

char    dsk_device[132];	/* If STANDALONE then this is the dsk    */
				/* device specified by the user; e.g.    */
				/* "dsk(0,0,0)".   If this is the unix   */
				/* version of "format", this is the      */
				/* special file name of the dsk device;  */
				/* e.g. "/dev/rdsk/0s7".                 */

/*
 * Geometry block global data.
 */

geometry_sector disk_geometry_sector;	/* image of disk version of geo block */
geometry_block  *disk_geo;		/* pointer to disk version geo block  */
geometry_sector incore_geometry_sector;	/* the geo block modified by the user */
geometry_block  *geo;			/* ptr to incore version geo block    */
int		disk_geometry_block;	/* TRUE if geo block exists on disk   */
int		geometry_modified;	/* TRUE if p command was ever used    */
char		partition_unit='k';	/* c,m,k,s,p - current partition unit */

/*
 * Flaw map global data.
 *
 * Note: flaw records from the flaw backup file are always read into
 * an intermediate buffer and copied into the real buffer, because the
 * standalone tapemaster controller cannot access an address > 20 bits.
 * This "hflaw_input" buffer is initialized so that the loader will
 * assign it a low address.
 */

char	hflaw_input[4*1024]=0;	/* flaw records are input to here   */
char	hflaw[HDC_MAXFLAWS*82];	/* the harris-format ascii flaw map */
int	hflaw_bytes;		/* number of bytes in "hflaw"       */
int	hflaw_count;		/* number of flaws in "hflaw"       */
char	hflaw_backup_file[100];	/* pathname of the flaw backup file */
vflaw_type vflaw[HDC_MAXFLAWS];	/* the vendor-format flaw map       */
int	vflaw_count;		/* number of flaws in "vflaw"       */
char	flawrecord[] = "cyl %-5d head %-4d offset %-7d length %-7d      %s\n";
char    firstflawrec[] = "flaw map for %s\n";

#define FFLAWSIZE 132           /* this intermediate flawmap buffer */
typedef struct {                /*    is used only by "format_proc";*/
	char text[FFLAWSIZE];   /*    it is too large to be defined */
} fflaw_struct;                 /*    locally in the procedure      */
fflaw_struct fflaw[HDC_MAXFLAWS];

/*
 * The following data structures define the various commands.
 * The "q!" command is not in these tables - to make the "q"
 * command unambiguous. Note that all commands are in the table
 * of commands, but the ones in the option table are only those
 * which can be invoked via a format program option.
 *
 * TO ADD A NEW COMMAND, add the appropriate entries to these
 * tables, then add the procedure(s) which processes the command.
 */

char	certifyform[] =	"certify [flawmap filepath]";
char	dataform[] =	"data <flawmap filepath>";
char	formatform[] =	"format [flawmap filepath]";
char    helpform[]=	"? [commandname]";
char	listform[] =	"list [flawmap filepath]";
char	parform[]=	"partition [partition#] [size]";
char	quitform[] =	"quit";
char	readform[] =	"read [flawmap filepath]";
char    statusform[]=	"status";
char	unitform[] =	"unit {cylinders,sectors,kbytes,megabytes,percent}";
char	verifyform[] =	"verify";
char	writeform[] =	"write [flawmap filepath]";

char	certifyhelp[] =	"certify the disk to produce a flawmap backup file";
char	datahelp[] =	"read vendor data to produce a flawmap backup file";
char	formathelp[] =	"format disk then write flawmap and geometry block";
char	helphelp[] =	"print help information";
char	listhelp[] =	"list disk flaws in cyl-head-sector format";
char	parhelp[]=	"define or print partition sizes";
char	quithelp[] =	"terminate the format program";
char	readhelp[] =	"read geometry block and flawmap (optional) from disk";
char	statushelp[] =	"print detailed information about the disk";
char	unithelp[] =	"set or print the unit for printing/setting partitions";
char	verifyhelp[] =	"verify the disk";
char	writehelp[] =	"write geometry block and flawmap (optional) to disk";

char	certifywarn[] =	"overwrites the flawmap backup file and the disk";
char	formatwarn[] =	"overwrites all information on the disk";
char	readwarn[] =	"overwrites the flawmap backup file";
char	datawarn[] =	"overwrites the flawmap backup file";
char	writewarn[] =	"overwrites the geometry block and flawmap (optional)";

typedef struct {
	char	*name;		/* command name */
	char    *cmdformmsg;	/* command format message string */
	char	*helpmsg;	/* help message string */
	char	*warnmsg;	/* warning message string */
	int	(*handler)();	/* routine which executes the command */
	int	minargs;	/* minimum command argument count */
	int	maxargs;	/* maximum command argument count */
} cmd;
cmd	*getcmd();

cmd cmdtab[] = {
/* CERTIFY AND DATA ARE NOT IMPLEMENTED IN HDC MICRO-CODE YET
{ "certify  ",	certifyform, certifyhelp, certifywarn, certify,    1, 2 },
{ "data     ",	dataform,    datahelp,    datawarn,    data_com,   2, 2 },
*/
{ "format   ",	formatform,  formathelp,  formatwarn,  format,     1, 2 },
{ "?        ",	helpform,    helphelp,    "",          help,       1, 2 },
{ "list     ",	listform,    listhelp,    "",          list,       1, 2 },
{ "partition",	parform,     parhelp,     "",          partition,  1, 3 },
{ "quit     ",	quitform,    quithelp,    "",          quit,       1, 1 },
{ "read     ",	readform,    readhelp,    readwarn,    readg,      1, 2 },
{ "status   ",	statusform,  statushelp,  "",          status,     1, 1 },
{ "unit     ",	unitform,    unithelp,    "",          unit,       1, 2 },
{ "verify   ",	verifyform,  verifyhelp,  "",          verify,     1, 1 },
{ "write    ",	writeform,   writehelp,   writewarn,   writeg,     1, 2 },
0
};

#ifndef STANDALONE
typedef struct {
	char option;          /* format program option          */
	int  (*handler)();    /* routine to process the command */
	int  minargs;         /* minimum command argument count */
	int  maxargs;         /* maximum command argument count */
} opt;

opt opttab[] = {
/* CERTIFY AND DATA ARE NOT IMPLEMENTED IN HDC MICRO-CODE YET
	{ 'c', certify_option, 3, 4 },
	{ 'd', data_option,    4, 4 },
*/
	{ 'f', format_option,  3, 4 },
	{ 'l', list_option,    3, 4 },
	{ 'v', verify_option,  3, 3 },
	0
};
#endif

/*************************************************************************
*  Procedure:	main
*
*  Description: Program start.
*
*  Returns:	n/a
**************************************************************************/

#ifdef STANDALONE
main( )

{
	register int	r12 ;
	register int	r11 ;
	register int	r10 ;

#else
main( argc, argv )

 	int  		argc ; 		/* number of arguments on the
                                         * "format" command invocation line
                                         */
        char 		*argv[] ;	/* array of pointers to the argument
        				 * values on the command line
					 */

{
#endif

	register cmd *c_ptr;	/* temp command pointer            */
	int 	count;
	char 	*cp;		/* temporary character pointer     */
	int 	i;
	int	disk_unit;	/* the unit number of the disk     */
	char    line[132];	/* user commands are input to here */
	int	margc;		/* number of arguments             */
	char	*margv[60];	/* pointers to argument values     */



#ifdef STANDALONE
	/*
	 * Standalone programs don't receive parameters.
	 * Request a disk specification and open the dsk file.
	 */

	boot_device = r10 ;
	version = (boot_device & 0xf000000) >> 24 ;
	buses[ 0 ] = RB_PRIMARY_IO( boot_device ) ;
	if ( buses[ 0 ] == RB_NO_BUS )
		buses[ 0 ] = RB_VERSA_BUS ;
	buses[ 1 ] = RB_SECONDARY_IO( boot_device ) ;

	printf("Standalone hdc disk formatter\n");
	dsk_device[0] = 0;
	for(;;) {
		printf("dsk device: ");
		gets(line);
		strcpy(dsk_device,line);
		if (strlen(dsk_device) == 0) continue;
		printf("Opening '%s'\n", dsk_device);
		if((fd = open(line, 2)) > 0)
			break;
		printf("Cannot open '%s'.\n", dsk_device);
		printf("Device format is 'dsk(ctlr,drive,par,bus)'.\n");
	}


#else
	/*
	 * The unix version of "format" receives a disk specification
	 * on the command line. Validate the command line, then open
	 * the disk special file.
	 */

	if (argc < 2  ||  argc > 4) {
		printf(usage_err);
		exit(1);
	};
     	count = sscanf( argv[argc-1], "%d%*c", &disk_unit);
     	if (count!=1 || disk_unit<0 || disk_unit>99) {
     		printf("Invalid disk unit number.\n");
     		exit(1);
     	}
	sprintf(dsk_device, "/dev/rdsk/%ds7", disk_unit);
	printf("Opening '%s'.\n",dsk_device);
	fd = open( dsk_device, 2 );
 	if (fd <= 0) {
 		perror("format: cannot open the dsk unit");
 		exit(1);
 	}
 	signal(SIGINT,intr);
#endif



	/*
	 * The disk device has been opened succesfully.
	 *
	 * Initialize; i.e.
	 *      o  indicate we are the format program
	 *         (to get special privileges)
	 *      o  get disk info
	 *      o  initialize pointers and data
	 *      o  read geometry block
 	 */

 	if ( ioctl(fd,DSKIOCFORMATCTL,1) < 0 ) {
 		perror("format: cannot enable formatting");
 		exit(1);
 	}
 	if ( ioctl(fd,DSKIOCSTATUS,&disk_status) < 0 ) {
 		perror("format: cannot read disk status");
 		terminate(1);
 	}
 	disk_geometry_block = FALSE;
 	geometry_modified = FALSE;
 	geo = &incore_geometry_sector.geometry_block;
 	disk_geo = &disk_geometry_sector.geometry_block;
	spc = disk_status.heads * disk_status.sectors;
	total_sectors = spc * disk_status.cylinders;
	def_sectors = (disk_status.def_cylinders +
	         disk_status.diag_cylinders) * spc;
	read_geometry_block();



#ifndef STANDALONE
  	/*
  	 * If this is the unix version of "format" and an option
  	 * is present on the command line, then execute the command
  	 * specified by the option and terminate.
	 */

	if (argc>2) {
		cp = argv[1];
		if (*cp++ != '-') {
			printf(usage_err);
			terminate(1);
		}
		for (i=0; opttab[i].option != 0; i++)
			if (opttab[i].option == *cp) {
				if (argc < opttab[i].minargs) break;
				if (argc > opttab[i].maxargs) break;
				(opttab[i].handler)(argc,argv);
				terminate(0);
			}
		printf(usage_err);
		terminate(1);
	}
#endif



        /*
         * If there were any problems, wait for a carriage return.
         * This is so that all the error messages don't get scrolled
         * off the screen when we print the startup disk status.
         *
         * We assume there are problems if any of:
         *      o  could not read profile prom - nsect, hhead, and
         *         ncyl will probably be zero, thus a small spc;
         *      o  not a valid geo block - a default internal geo
         *         was created, thus it differs from the disk geo;
         */

	if (spc<10 || !geo_compare()) {
		printf("Type <CR> to continue  ");
		if (gets(line) == 0)
			if (feof(stdin))
			       	clearerr(stdin);
		printf("\n\n\n");
	}
	status(1,1);

        /*
         * Read and execute user commands forever.
         * Note that a specific check is made for the "q!" command so
         * that the "q" command won't be ambiguous.
         */

	while (TRUE) {
   		printf("format:");
   		if (gets(line) == 0) {
			if (feof(stdin)) {
   			       	clearerr(stdin);
   				putchar('\n');
   			}
   			continue;
   		}
   		if (line[0] == 0) continue;
   		margc = makearg(line,margv);
   		cp = margv[0];
   		if (*cp++=='q' && *cp=='!')
   			terminate(0);
   		c_ptr = getcmd(margv[0]);
   		if (c_ptr == (cmd *)-1) {
   			printf("Ambiguous command.\n");
   			continue;
   		}
   		if (c_ptr == 0) {
   			printf("Invalid command.\n");
   			continue;
   		}
   		if (margc < c_ptr->minargs  ||  margc > c_ptr->maxargs) {
   			printf("usage:  %s\n", c_ptr->cmdformmsg);
   			continue;
   		}
   		(*c_ptr->handler)(margc, margv);
        }
}

/*************************************************************************
*  Procedure:	certify
*
*  Description: This procedure processes the "certify" command.
*
*  Returns:
**************************************************************************/

certify( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "certify" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	char *filename;

	/*
	 * If the optional flawmapfile parameter was specified, certify
	 * then write the produced flawmap into the file.
	 * Since the certify command takes a long time, insure that
	 * the flaw file is accessible before certifying.
	 */

	if (argc == 2) {
		if (!command_warning(certifywarn)) return;
        	strcpy(hflaw_backup_file, argv[1]);
		strcpy(filename,hflaw_backup_file);
		fd = open( filename, O_RDWR+O_CREAT+O_TRUNC, 0444 );
        	if (fd<0) {
        		printf("Could not open the flaw map file.\n");
        		terminate(1);
        	}
        	close(fd);
        	if ( certify_proc() )
        		write_hflaw_backup();
        	return;
	};

	/*
	 * No flawmapfile parameter was specified so certify, format
	 * with the produced flawmap, then write the flawmap and geometry
	 * block to the disk description tracks.
	 */

	printf("No file was specified for the flawmap.\n");
	printf("Do you wish to both certify and format the disk?");
	if (!yes()) return;
	if (!certify_proc()) return;
	if (!format_proc(TRUE)) return;
	write_hflaw();
	write_geometry_block();
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	certify_option
*
*  Description: This procedure processes the "certify" command when
*               the command was invoked with an option on the "format"
*               command line; i.e. "format -c [filepath] <disk>"
*
*  Returns:
**************************************************************************/

certify_option( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	char *filename;

	/*
	 * If the optional flawmapfile parameter was specified, certify
	 * then write the produced flawmap into the file.
	 * Since the certify command takes a long time, insure that
	 * the flaw file is accessible before certifying.
	 */

	if (argc == 4) {
        	strcpy(hflaw_backup_file, argv[2]);
		strcpy(filename,hflaw_backup_file);
		fd = open( filename, O_RDWR+O_CREAT+O_TRUNC, 0444 );
        	if (fd<0) {
        		printf("Could not open the flaw map file.\n");
        		terminate(1);
        	}
        	close(fd);
        	if (!certify_proc()) terminate(1);
        	write_hflaw_backup();
        	return;
	};

	/*
	 * No flawmapfile parameter was specified so certify, format
	 * with the produced flawmap, then write the flawmap and geometry
	 * block to the disk description tracks.
	 */

	if (!certify_proc()) terminate(1);
	if (!format_proc(TRUE)) terminate(1);
	write_hflaw();
	write_geometry_block();
}
#endif

/*************************************************************************
*  Procedure:	certify_proc
*
*  Description: This procedure certifies the disk to produce a flawmap.
*
*  Returns:     FALSE = error
*               TRUE  = ok, and
*                       "hflaw" contains the produced flawmap
*                       "hflaw_count" is the number of flaws in "hflaw"
*                       "hflaw_bytes" is the number of bytes in "hflaw"
*
**************************************************************************/

certify_proc()
{
	int cyl,head,i,bps;
	char *cp;
	struct formatop track;

	bps = disk_status.bytes_per_sec;

	/*
	 * Add the initial comment record to the flawmap
	 */

	cp = (char *)hflaw;
	sprintf(cp, firstflawrec, dsk_device);
	while (*cp++ != '\n');
	hflaw_count = 1;

        /*
         * Print a header every 50 cylinders and a dot after each cylinder.
         */

        printf("Certifying.\n");
        for ( cyl=0; cyl<disk_status.phys_cylinders; cyl++ ) {
        	if ( !(cyl % 50) )
        		printf("\ncylinder %4d of %4d ",
        			cyl, disk_status.phys_cylinders);
        	printf(".");
        	fflush(stdout);

		/*
		 * For each track on this cylinder, issue a certify ioctl.
		 * For each returned flaw on the track, generate an ascii
		 * flaw record in "hflaw".
		 */

        	for ( head=0; head<disk_status.phys_heads; head++ ) {
        		track.cylinder  = cyl;
        		track.head      = head;
        		track.flaw_count= 0;
 			if ( ioctl(fd,DSKIOCCERTIFY,&track) ) {
 				perror("\nformat: cannot certify track");
 				printf("Certification aborted.\n");
 				return FALSE;
 			}
			for (i=0; i<track.flaw_count; i++) {
				sprintf(cp, flawrecord,
					track.cylinder,
					track.head,
					track.flaw[i]*bps+bps/2,
					1,
					"certify" );
				while (*cp++ != '\n');
				hflaw_count++;
				if (hflaw_count >= HDC_MAXFLAWS) {
					printf("\nToo many flaws.\n");
 					printf("Certification aborted.\n");
 					return FALSE;
				}
			}
        	}
        }
	hflaw_bytes = (int)cp - (int)hflaw;
        printf("\nCertification complete.\n");
        return TRUE;
}

/*************************************************************************
*  Procedure:	checksum
*
*  Description: This procedure computes a checksum. It adds
*               all the bytes in a specified buffer.
*
*  Returns:	The checksum
**************************************************************************/

unsigned int
checksum( buffer, count )

 	unsigned char	*buffer ;	/* starting address for checksum.
                                         */
        int             count ;		/* number of bytes to checksum.
					 */
{
	int i;
	unsigned sum;

	sum = 0;
	for ( i=0; i<count; i++ )
		sum += buffer[i];
	return(sum);
}

/*************************************************************************
*  Procedure:	command_warning
*
*  Description: This procedure prints a warning message and then asks
*               the user to verify that he really wants to continue.
*
*  Returns:	TRUE means the user wants to continue.
*               FALSE means the user wants to abort the command.
**************************************************************************/

command_warning( warning_message )

	char	*warning_message;	/* the warning message to print */

{
	printf("This command %s.\n", warning_message);
	printf("Do you want to continue?");
	return (yes()  ?  TRUE : FALSE);
}

/*************************************************************************
*  Procedure:	data_com
*
*  Description: This procedure processes the "data" command.
*
*  Returns:
*************************************************************************/

data_com( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "data" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	if (!command_warning(datawarn)) return;
        strcpy(hflaw_backup_file, argv[1]);
	data_proc();
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	data_option
*
*  Description: This procedure processes the "data" command when
*               the command was invoked with an option on the "format"
*               command line; i.e. "format -d <filepath> <disk>"
*
*  Returns:
**************************************************************************/

data_option( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
       	strcpy(hflaw_backup_file, argv[2]);
        if (!data_proc())
        	terminate(1);
}
#endif

/*************************************************************************
*  Procedure:	data_proc
*
*  Description: This procedure reads all vendor data from the disk
*               and produces a flawmap backup file.
*
*  Returns:
**************************************************************************/

data_proc( argc, argv )
{
	int i,cyl,head;
	char *cp;
	vflaw_type track;

	/*
	 * Add the initial comment record to the flawmap
	 */

	cp = (char *)hflaw;
	sprintf(cp, firstflawrec, dsk_device);
	while (*cp++ != '\n');
	hflaw_count = 1;

        /*
         * Print a header every 50 cylinders and a dot after each cylinder.
         */

        printf("Reading vendor flaw data.\n");
        for ( cyl=0; cyl<disk_status.phys_cylinders; cyl++ ) {
        	if ( !(cyl % 50) )
        		printf("\ncylinder %4d of %4d ",
        			cyl, disk_status.phys_cylinders);
        	printf(".");
        	fflush(stdout);

		/*
		 * For each track on this cylinder, read vendor data.
		 * For each returned flaw on the track, generate an ascii
		 * flaw record in "hflaw".
		 */

        	for ( head=0; head<disk_status.phys_heads; head++ ) {
        		track.cylinder  = cyl;
        		track.head      = head;
        		track.flaw[0].length = 0;
		 	if ( ioctl(fd,DSKIOCVENDORFLAW,&track) ) {
 				perror("\nformat: cannot read vendor data");
        			printf("cyl %s head %s\n", cyl,head);
 				printf("Vendor data read aborted.\n");
 				return FALSE;
 			}
			for (i=0; i<MAXVFLAW; i++) {
				if (track.flaw[i].length == 0) break;
				sprintf(cp, flawrecord,
					track.cylinder,
					track.head,
					track.flaw[i].offset,
					track.flaw[i].length,
					"vendor" );
				while (*cp++ != '\n');
				hflaw_count++;
				if (hflaw_count >= HDC_MAXFLAWS) {
					printf("\nToo many flaws.\n");
 					printf("Vendor data read aborted.\n");
 					return FALSE;
				}
			}
        	}
        }
	hflaw_bytes = (int)cp - (int)hflaw;
        printf("\nVendor data read complete.\n");
        write_hflaw_backup();
        return TRUE;
}

/*************************************************************************
*  Procedure:	disk_def_io
*
*  Description: This procedure reads and writes the disk
*               definition track - geometry block, flaw map.
*
*  Returns: 0 = normal
*           nz = error
**************************************************************************/

disk_def_io( read_write, buffer, byte_offset, byte_count )

int 	read_write;	/* HDC_READ or HDC_WRITE
			 */
char	*buffer;	/* buffer to read/write to/from
			 */
int	byte_offset;	/* Relative address within disk def tracks
			 */
int	byte_count;	/* Number of bytes to read/write.
			 * hdc requires a long word multiple.
			 */

{
	int err;
	byte_count = (byte_count+3) & 0xfffffffc;
	err = lseek( fd, byte_offset, 0 );
	if (err<0) {
		perror("format: cannot lseek in disk definition track");
		return(1);
	}
	switch (read_write) {
	case HDC_READ:
		err= read( fd, buffer, byte_count );
		break;
	case HDC_WRITE:
		err= write( fd, buffer, byte_count );
		break;
	}
	if (err > 0) return(0);
	if (err==0)
		printf("format: zero bytes transferred\n");
	else
		perror("format: disk definition track i/o");
	return 1;
}

/*************************************************************************
*  Procedure:	format
*
*  Description: This procedure processes the "format" command.
*
*  Returns:
**************************************************************************/

format( argc, argv )

int	argc ;		/* number of arguments on the command line */
char	*argv[] ;	/* array of pointers to argument values */

{
	if (!command_warning(formatwarn)) return;

        /*
         * Check if there is vendor flaw data on the disk.
         */

	/* TEMP until read_vendor_data is implemented in hw
	track.cylinder  = cyl;
	track.head      = head;
	if ( !ioctl(fd,DSKIOCVENDORFLAW,&track) ) {
 		printf("Vendor data is present on this disk.\n");
 		printf("Are you sure you want to continue?");
 		if (!yes()) return;
 	}
	*/

	/*
	 * If no filepath was specified, read the flaw map from disk;
         * otherwise read the flaw map from the backup files.
	 */

	if (argc==1){
		if (!read_hflaw())
			return;
	}
	else {
        	strcpy(hflaw_backup_file, argv[1]);
		if (!read_hflaw_backup())
			return;
        }

	/*
	 * If the flawmap file is empty, warn the user.
	 */

	if (hflaw_count==0) {
		printf("There are no flaws.\n");
		printf("Are you sure you want to continue?");
		if (!yes()) return;
	}

	/*
	 * Do common format processing, then write out the flawmap
	 * and geometry block.
	 */

        format_proc(TRUE);
	write_hflaw();
        write_geometry_block();
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	format_option
*
*  Description: This procedure processes the "format" command when
*               the command was invoked with an option on the "format"
*               command line; i.e. "format -f [filepath] <disk>"
*
*  Returns:
**************************************************************************/

format_option(argc,argv)

int	argc ;		/* number of arguments on the command line */
char	*argv[] ;	/* array of pointers to argument values */

{
	/*
	 * If no filepath was specified, read the flaw maps from disk;
         * otherwise read the flaw maps from the backup file.
         * Then do common format processing.
	 */

	if (argc==3) {
		if (!read_hflaw()) return;
	}
	else {
        	strcpy(hflaw_backup_file, argv[2]);
		if (!read_hflaw_backup())
			return;
        }
        if (!format_proc(TRUE))
        	terminate(1);

	/*
	 * Write out the flaw map and geometry block.
	 */

	write_hflaw();
        write_geometry_block();
}
#endif

/*************************************************************************
*  Procedure:	format_proc
*
*  Description: This procedure formats a disk or just prints the sectors
*               it would have formatted.
*
*  Returns:     FALSE = error occurred
**************************************************************************/

format_proc(do_format)

int  	do_format ; 		/* TRUE = really format the disk;
                                 * FALSE= just echo the sectors
                                 */
{
	int cyl,head,sec;
	int offset,length,bytes,start_sector,end_sector;
	int i,ff,vf,tf,count,abort,err;
	char *ch, *cf;
	char tag1[80],tag2[80],tag3[80],tag4[80];
	struct formatop track;
	int fflaw_count;
	int bpt; /* physical number of bytes per track for formatting */
	char *operation;

	operation = (do_format ? "Formatting" : "Listing");

        /*
	 * Flaw record format is:
         *     "cyl xxx head xxx offset xxxx length xxxx"
         * The first flaw record may optionally be a comment line.
         *
         * Prepare the harris-format flaw records in the fflaw array.
         * They were read in one piece - newline's delimiting the records.
         * Here they are unpacked into one record per fflaw array entry.
         * The optional comment record and blank records are removed here.
         */

	fflaw_count = hflaw_count;
	if (fflaw_count > 0) {
	 	ch = (char *) &hflaw[0];
		count = sscanf(ch,"%s",tag1);
		if ( strcmp(tag1,"cyl") != 0) {
			fflaw_count--;
			while (*ch != '\n') *ch++;
			*ch = 0;
			printf("Skipping first flaw record: %s\n",hflaw);
			*ch++ = '\n';
		}
	}
 	cf= (char *) &fflaw[0];
 	for (i=0; i < FFLAWSIZE * hflaw_count; i++)
 		cf[i] = 0;
 	for (ff=0; ff<fflaw_count;) {
 		for (cf=ch; *cf==' '; cf++);
 		if (*cf=='\n') {
 			ch = cf+1;
 			fflaw_count--;
 			continue;
 		}
 		cf = (char *) &fflaw[ff];
		while (*ch != '\n')
 			*cf++ = *ch++;
 		ch++;
 		ff++;
	}

        /*
         * A pre-processed version of the flaw map is now in "fflaw".
         * Validate each record of the harris-format (ascii) flaw map.
         * Build an equivalent vflaw version (not ascii).
         */

	if (do_format)
		printf("Validating %d flaws.\n",fflaw_count);
	bpt = (disk_status.phys_sectors + 1) * disk_status.bytes_per_sec;
	abort = FALSE;
	vf = 0;

        for (ff=0; ff<fflaw_count; ff++) {
		count=sscanf(&fflaw[ff],"%s %d %s %d %s %d %s %d",
			tag1,&cyl,tag2,&head,tag3,&offset,tag4,&length);
	        if (count < 8) {
			printf("Flaw %d too few arguments.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
	        else if ( strcmp(tag1,"cyl") != 0 ) {
			printf("Flaw %d could not find 'cyl'.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
	        else if ( strcmp(tag2,"head") != 0 ) {
			printf("Flaw %d could not find 'head'.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
	        else if ( strcmp(tag3,"offset") != 0 ) {
			printf("Flaw %d could not find 'offset'.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
	        else if ( strcmp(tag4,"length") != 0 ) {
			printf("Flaw %d could not find 'length'.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
		else if (cyl >= disk_status.phys_cylinders || cyl < 0) {
			printf("Flaw %d 'cyl' address out of range.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
		else if (head >= disk_status.phys_heads || head < 0) {
			printf("Flaw %d 'head' address out of range.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
		else if (offset >= bpt || offset < 0) {
			printf("Flaw %d 'offset' out of range.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		}
		else if (length >= bpt || length < 0) {
			printf("Flaw %d 'length' out of range.\n",ff+1);
			printf("%s\n", &fflaw[ff]);
			abort = TRUE;
		};
		vflaw[vf].cylinder = cyl;
		vflaw[vf].head     = head;
		vflaw[vf].flaw[0].offset = offset;
		vflaw[vf].flaw[0].length = length;
		vf++;
        }
        if (abort) {
        	printf("%s terminated.\n", operation);
        	return FALSE;
        }
        vflaw_count = vf;

	/*
	 * Each individual flaw record is valid.
	 * Check for duplicate and out-of-order flaw records.
	 */

	for (vf=1; vf<vflaw_count; vf++) {
		if (vflaw[vf].cylinder > vflaw[vf-1].cylinder) continue;
		if (vflaw[vf].cylinder < vflaw[vf-1].cylinder) {
			printf("Flaw %d cyl out of sequence.\n",vf+1);
			printf("%s\n", &fflaw[vf]);
			abort = TRUE;
			continue;
		}
		if (vflaw[vf].head     > vflaw[vf-1].head    ) continue;
		if (vflaw[vf].head     < vflaw[vf-1].head    ) {
			printf("Flaw %d head out of sequence.\n",vf+1);
			printf("%s\n", &fflaw[vf]);
			abort = TRUE;
			continue;
		}
		if (vflaw[vf].flaw[0].offset != vflaw[vf-1].flaw[0].offset)
			continue;
		printf("Flaw %d duplicated.\n",vf+1);
			printf("%s\n", &fflaw[vf]);
		abort = TRUE;
	}
        if (abort) {
        	printf("%s terminated.\n", operation);
        	return FALSE;
        }

        /*
         * We now have an error-free, ordered list of flaws for the disk.
         * The following code does the actual formatting.
         * Print a header every 50 cylinders and a dot after each cylinder.
         */

        vf=0;
        if (do_format)
        	printf("Formatting.\n");
        for ( cyl=0; cyl<disk_status.phys_cylinders; cyl++ ) {
        	if (do_format) {
        		if ( cyl == 50*(cyl/50) )
        			printf("\ncylinder %4d of %4d ",
        				cyl, disk_status.phys_cylinders);
        		printf(".");
        		fflush(stdout);
        	}

		/*
		 * For each track on this cylinder, issue a "format track"
		 * ioctl to the handler. "vflaw"s which are for the same
		 * cyl/head are combined into one format-track operation.
		 *
		 * Note that a vendor flaw consists of a starting byte
		 * offset within the track, and the number of consecutive
		 * flawed bits. The byte before the first flaw byte is also
		 * suspect - as well as the byte after the last flaw byte.
		 * This info has to be converted into sectors.
		 *
		 * track=        structure which defines the track to format;
		 * vf=           pointer to the next flaw record
		 * tf=           track flaw number
		 * offset=       byte offset of flaw within track
		 * bytes=        number of bytes in the flaw
		 * start_sector= first sector of the flaw
		 * end_sector=   last sector of the flaw
		 */

        	for ( head=0; head<disk_status.phys_heads; head++ ) {
        		track.cylinder  = cyl;
        		track.head      = head;
        		track.flaw_count= 0;
        		tf = 0;
			while (vf<vflaw_count) {
				if (cyl  != vflaw[vf].cylinder) break;
				if (head != vflaw[vf].head) break;
				offset = vflaw[vf].flaw[0].offset;
				bytes = (vflaw[vf].flaw[0].length+6)/8 + 1;
				start_sector = (offset-1)
						/ disk_status.bytes_per_sec;
				if (start_sector<0) start_sector = 0;
				end_sector = (offset + bytes)
						/ disk_status.bytes_per_sec;
				if (end_sector >= disk_status.phys_sectors)
					end_sector = disk_status.phys_sectors-1;
				for (sec=start_sector; sec<=end_sector; sec++)
					if (tf<MAXVFLAW) {
						track.flaw_count++;
						track.flaw[tf]=sec;
						tf++;
					}
				vf++;
			}
			if (do_format) {
		 		err = ioctl(fd,DSKIOCFORMAT,&track);
		 		if (err) {
		 			perror("\nformat: cannot format track");
		 			printf("Formatting aborted.\n");
		 			return FALSE;
		 		}
		 	}
		 	else
		 		for (i=0; i<track.flaw_count; i++)
		 		printf("cyl %4d     head %2d     sector %3d\n",
		 			track.cylinder, track.head,
		 			track.flaw[i]);
        	}
        }
        if (do_format)
        	printf("\nFormatting complete.\n");
        return TRUE;
}

/*************************************************************************
*  Procedure:	geometry_compare
*
*  Description: This procedure compares the incore geometry block
*               with the on-disk geometry block.
*
*  Returns:	TRUE =  the geometry blocks are identical
*               FALSE = they are different
**************************************************************************/

int
geo_compare()
{
	char *c1,*c2;
	int i;

	c1 = (char *) &incore_geometry_sector;
	c2 = (char *) &disk_geometry_sector;
	for ( i=0; i<sizeof(geometry_block); i++ )
		if (c1[i] != c2[i]) return (FALSE);
	return (TRUE);
}

/*************************************************************************
*  Procedure:	getcmd
*
*  Description: This procedure looks up commands in the table of
*               valid commands.
*
*  Returns:     A pointer to a structure containing information
*               about the command.
*               0 =  invalid command.
*               -1=  ambiguous command.
**************************************************************************/

cmd *
getcmd( name )

	register char	*name ;

{
	register char	*p, *q;
	register cmd	*c, *found;
	register int nmatches, longest;

  	longest = 0;
  	nmatches = 0;
  	found = 0;
  	for (c = cmdtab; p = c->name; c++) {
  		for (q = name; *q == *p++; q++)
  			if (*q == 0)		/* exact match? */
  				return (c);
  		if (!*q) {			/* the name was a prefix */
  			if (q - name > longest) {
  				longest = q - name;
  				nmatches = 1;
  				found = c;
  			} else if (q - name == longest)
  				nmatches++;
  		}
  	}
  	if (nmatches > 1)
  		return ((cmd *)-1);
  	return (found);
}

/*************************************************************************
*  Procedure:	help
*
*  Description: This procedure processes the "?" command.
*
*  Returns:
**************************************************************************/

help( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "?" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	cmd	*c_ptr ;	/* command pointer */

        /*
         * No arguments means to print general info about all commands.
         */

	if (argc == 1) {
		printf("Commands may be abbreviated.\n");
		printf("Type '? <command>' for detailed information.\n\n");
		for (c_ptr = cmdtab; c_ptr->name; c_ptr++)
			printf("%s  %s\n", c_ptr->name, c_ptr->helpmsg);
		printf("\n");
		return;
	}

	/*
	 * A single argument means to print info about the specified command.
	 */

	while (--argc > 0) {
		register char *arg;
		arg = *++argv;
		c_ptr = getcmd(arg);
		if (c_ptr == (cmd *)-1)
			printf("Ambiguous help command '%s'.\n", arg);
		else if (c_ptr == (cmd *)0)
			printf("Invalid help command '%s'.\n", arg);
		else	{
			printf(" %s\n", c_ptr->cmdformmsg);
			printf(" %s\n", c_ptr->helpmsg);
			if (strlen(c_ptr->warnmsg) > 0)
				printf(" warning:  %s\n", c_ptr->warnmsg);
		}
	}
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	intr
*
*  Description: This procedure processes SIGINT signals; i.e. the user
*               typed the "intr" key.
*
*  Returns:     No returns. This program is terminated.
**************************************************************************/

intr()
{
	printf("\n");
	terminate(1);
}
#endif

/*************************************************************************
*  Procedure:	list
*
*  Description: This procedure processes the "list" command.
*
*  Returns:
**************************************************************************/

list( argc, argv )

int	argc ;		/* number of arguments on the command line */
char	*argv[] ;	/* array of pointers to argument values */

{
	/*
	 * If no filepath was specified, read the flaw map from disk;
         * otherwise read the flaw map from the backup file.
         * Then list the flaws.
	 */

	if (argc==1){
		if (!read_hflaw())
			return;
	}
	else {
        	strcpy(hflaw_backup_file, argv[1]);
		if (!read_hflaw_backup())
			return;
        }
	if (hflaw_count==0) {
		printf("There are no flaws.\n");
		return;
	}
        format_proc(FALSE);
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	list_option
*
*  Description: This procedure processes the "list" command when the
*               command was invoked with an option on the "format" program
*               invocation line; i.e. "format -l [filepath] <disk>"
*
*  Returns:
**************************************************************************/

list_option(argc,argv)

int	argc ;		/* number of arguments on the command line */
char	*argv[] ;	/* array of pointers to argument values */

{
	/*
	 * If no filepath was specified, read the flaw maps from disk;
         * otherwise read the flaw maps from the backup file.
         * Then list the flaws.
	 */

	if (argc==3) {
		if (!read_hflaw()) return;
	}
	else {
        	strcpy(hflaw_backup_file, argv[2]);
		if (!read_hflaw_backup())
			return;
        }
	if (hflaw_count==0) {
		printf("There are no flaws.\n");
		return;
	}
        if (!format_proc(FALSE))
        	terminate(1);
}
#endif

/*************************************************************************
*  Procedure:	makearg
*
*  Description: This procedure processes user command lines to produce
*               an argument count and a table of pointers to the
*               various arguments on the command line.
*
*  Returns:	1) the argument table (passed in as a parameter) is
*	           filled in with pointers to the command arguments
*               2) the number of arguments found
**************************************************************************/

int
makearg(cp,argp)

 	register char   *cp ; 		/* character pointer to start of
                                         * user command line
                                         */
        register char	**argp ;	/* pointer to array of pointers
        				 * to argument values on the
        				 * command line.
					 */
{
	register int argc = 0 ;         /* number of arguments on cmnd line */

	for (;*cp;) {
		while (isspace(*cp))
			cp++;
		if (*cp == '\0')
			break;
		*argp++ = cp;
		argc ++;
		while (*cp != '\0' && !isspace(*cp))
			cp++;
		if (*cp == '\0')
 			break;
		*cp++ = '\0';
	}
	*argp++ = 0;
	return(argc);
}

/*************************************************************************
*  Procedure:	megabytes
*
*  Description: This procedure converts a byte count to a
*               megabyte count - rounding to nearest megabyte.
*
*  Returns:	The equivalent number of megabytes.
**************************************************************************/

int
megabytes(bc)

 	unsigned int    bc;
{
	return((bc+500000) / 1000000);
}
/*************************************************************************
*  Procedure:	partition
*
*  Description: This procedure processes the "partition" command.
*               It sets or prints the partition sizes of the incore
*               version of the geometry block.
*
*  Returns:
**************************************************************************/

partition( argc, argv )

int		argc ;		/* number of arguments in the
				 * "p" command line
				 */
char 		*argv[] ;	/* array of pointers to argument
       				 * values on the command line
				 */
{
	int i,par,size,count,nsectors,ncyl,percent;
	int tcyl;		/* total number of cylinders on the disk */
	int start;		/* start sector of the on-disk partition */
	int total;		/* total number of on-disk sectors so far*/
	int length;		/* sector length of on-disk partition    */

        /*
         * If the "partition" command has no arguments, just print the
         * partition info for the in-core geometry block.
         */

	if (argc==1) {
		printf("Current partitions");
		if (!geo_compare())
			printf(" (not saved on disk)");
		printf(":\n\n");
		for (i=0;i<8;i++) print_partition(i,geo);
		print_partition_total(geo);
		printf("\n");
  		return;
	}

        /*
         * If the command is "partition defaults" then initialize all
         * partitions to default values.
         */

	if (*argv[1] == 'd') {
		partition_defaults();
		print_partition_total(geo);
		geometry_modified = TRUE;
		return;
	}

        /*
         * Scan and validate the "partition" command;
         * command format is: "p <partition> <size>"
         */

	par=MAXINT;
	size=MAXINT;
     	count = sscanf( argv[1], "%d", &par);
	if (count!=1 || par<0 || par>HDC_DEFPART) {
 		printf("Invalid partition.\n");
 		return;
 	}
 	if (par==HDC_DEFPART) {
 		printf("Partition 7 cannot be modified.\n");
 		return;
 	}
 	if (argc<3) {
 		printf("Invalid partition size.\n");
 		return;
 	}
     	count = sscanf( argv[2], "%d", &size );
	if (count!=1 || size<0 ) {
		printf("Invalid partition size.\n");
		return;
	}

	/*
	 * Set the partition size based on "size" and "units".
	 * Must round up to a cylinder boundary.
	 */

	switch(partition_unit) {
		case 'c':	/* cylinders */
				ncyl=size;
				break;

		case 'm':	/* megabytes */
				size= 15625 * size / 8;
				ncyl= (size + spc - 1)/spc;
				break;

		case 'k':	/* kbytes */
				size= 2*size;
				ncyl= (size + spc - 1)/spc;
				break;

		case 's':	/* sectors */
				ncyl= (size + spc - 1)/spc;
				break;

		case 'p':       /* percent of disk */
				if (size>100) size = 100;
        			tcyl = total_sectors / spc;
        			ncyl = (size * tcyl + 50) / 100;
				break;

		default:	printf("Invalid partition unit.\n");
				return;
	}

        /*
         * If the requested size is too large, use remainder.
         *    count = total sectors of all other partitions
         *    nsectors = #sectors in partition being modified
         *    ncyl = #cylinders in partition being modified
         * Add the new size to the incore geometry block.
         * Print out partition totals.
         */

        count=0;
        for (i=0; i<8; i++)
        	if (i != par) count += geo->partition[i].length * HDC_SPB;
        nsectors = ncyl * spc;
        if ( (count+nsectors) > total_sectors )
        	nsectors = total_sectors - count;
        geo->partition[par].length = nsectors / HDC_SPB;
        partition_starts();
	geometry_modified = TRUE;
	print_partition(par,geo);
	print_partition_total(geo);
}

/*************************************************************************
*  Procedure:	partition_defaults
*
*  Description: This procedure initializes the incore-geometry block
*               partitions to default values.
*
*  Returns:
**************************************************************************/

#define DEFAULT_P0_SIZE  24000  /* blocks */
#define DEFAULT_P1_SIZE  24000  /* blocks */
#define DEFAULT_MIN_SIZE 64000  /* blocks */

partition_defaults()
{
	int i;
	int bpc;  /* blocks per cylinder */

       	printf("Initializing a default geometry block.\n");
       	bpc = spc / HDC_SPB;

	/*
	 * Set default partition 0 and 1.  Then set partition 2 to
	 * be all remaining cylinders. Partition info is in blocks.
	 * The size of def partition is defined in the profile prom.
	 */

       	for (i=0; i<HDC_DEFPART; i++) {
       		geo->partition[i].length = 0;
       		geo->partition[i].start  = 0;
       	}
	geo->partition[HDC_DEFPART].length= def_sectors / HDC_SPB;
	geo->partition[HDC_DEFPART].start=
			(total_sectors - def_sectors) / HDC_SPB;
	if ((total_sectors/HDC_SPB) < DEFAULT_MIN_SIZE) {
		printf("Not enough space for default partitions.\n");
		return;
	}
	geo->partition[0].length= (DEFAULT_P0_SIZE / bpc) * bpc;
	geo->partition[1].length= (DEFAULT_P1_SIZE / bpc) * bpc;
	geo->partition[2].length = geo->partition[HDC_DEFPART].start
	    - geo->partition[0].length - geo->partition[1].length;
	partition_starts();
}

/*************************************************************************
*  Procedure:	partition_starts
*
*  Description: This procedure computes then adds the starting address
*               of each partition to the geometry block. Starts are
*               computed from partition sizes which are consecutive.
*               Partition 7 (definition tracks) start is not modified.
*
*  Returns:
**************************************************************************/

partition_starts()
{
	int count,partition;

	count=0;
	for (partition=0; partition<HDC_DEFPART; partition++) {
        	if (geo->partition[partition].length == 0)
        		geo->partition[partition].start = 0;
        	else {
        		geo->partition[partition].start = count;
        		count += geo->partition[partition].length;
        	}
        }
}

/*************************************************************************
*  Procedure:	print_partition
*
*  Description: This procedure prints a partition size and start.
*
*  Returns:
**************************************************************************/

print_partition( par, gb )

 	int             par ; 		/* partition number (0-7)
                                         */
 	geometry_block  *gb ; 		/* pointer to the geometry block
                                         */
{
	int start;	/* start sector of the partition */
	int length;	/* sector length of partition    */
	int ncyl;
	int percent;

	start= gb->partition[par].start * HDC_SPB;
	length= gb->partition[par].length * HDC_SPB;
	printf("partition %d ", par);
	switch (partition_unit) {

	case 's': /* print sector start and size */

		printf("%9d sectors    ", length);
		printf("%9d start sector\n", start);
	        break;

	case 'c': /* print cylinder start and size */

		printf("%7d cylinders    ", length/spc);
		printf("%7d start cylinder\n", start/spc);
	        break;

	case 'k': /* print kbyte start and size */

		printf("%9d kbytes    ", length/2);
		printf("%9d start kbyte\n", start/2);
	        break;

	case 'm': /* print megabyte sizes */

		printf("%7d megabytes    ", megabytes(length*BPS) );
		printf("%7d start megabyte\n", megabytes(start*BPS) );
	        break;

	case 'p':
		/*
		 * print %-of-disk size.
		 * We never print 100% unless all disk space is used;
		 * do not round up to 100% (all other %'s are rounded).
		 */

		ncyl = disk_status.cylinders;
  		percent = (100*(length/spc) + ncyl/2) / ncyl;
		if ((percent==100) && (length != total_sectors)) percent=99;
		printf("percent of disk: %3d%%\n", percent);
		break;
	}
}

/*************************************************************************
*  Procedure:	print_partition_total
*
*  Description: This procedure prints the total size of all used
*               and unused partitions.
*
*  Returns:
**************************************************************************/

print_partition_total( gb )

 	geometry_block  *gb ; 		/* pointer to the geometry block
                                         */
{
	int i;
	int ncyl;	/* total number of cylinders */
	int percent;
	int total;	/* total number of sectors used */

	total=0;
	for (i=0; i<8; i++) total += gb->partition[i].length;
	total *= HDC_SPB;
	switch (partition_unit) {

	case 's': /* print sector totals */

		printf("%d sectors used    ", total);
		printf("%d sectors unused\n", total_sectors - total);
	        break;

	case 'c': /* print cylinder totals */

		printf("%d cylinders used    ", total/spc);
		printf("%d cylinders unused\n", (total_sectors - total)/spc);
	        break;

	case 'k': /* print kbyte totals */

		printf("%d kbytes used    ", total/2);
		printf("%d kbytes unused\n", (total_sectors - total)/2);
	        break;

	case 'm': /* print megabyte totals */

		printf("%d megabytes used    ", megabytes(total*BPS) );
		printf("%d megabytes unused\n",
			megabytes((total_sectors-total)*BPS));
	        break;

	case 'p':
		/*
		 * print %-of-disk used and unused.
		 * We never print 100% unless all disk space is used;
		 * we do not round up to 100% (all other %'s are rounded).
		 */

		ncyl = disk_status.cylinders;
  		percent = (100*(total/spc) + ncyl/2) / ncyl;
		if ((percent==100) && (total != total_sectors)) percent=99;
		printf("%d%% of disk used\n", percent);
		break;
	}
}

/*************************************************************************
*  Procedure:	quit
*
*  Description: This procedure processes the "quit" command.
*               It normally terminates the program.
*
*  Returns:	This procedure returns only if the incore geometry block
*               has been modified and not written to disk.
**************************************************************************/

quit( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "quit" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	if (!geometry_modified) terminate(0);
	if ( !geo_compare() ) {
		printf("The geometry block has been modified.\n");
		printf("You must 'format', 'write', or use 'q!'");
		printf(" to force a program exit.\n");
		return;
	}
	terminate(0);
}

/*************************************************************************
*  Procedure:	readg
*
*  Description: This procedure processes the "read" command.
*               It reads the geometry block from disk to the in-core
*               version of the geometry block. It also optionally reads
*               the flaw map from disk and writes it to a flaw backup file.
*
*  Returns:
**************************************************************************/

readg( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "read" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	/*
	 * Command format is: "read [flawmap]"
	 */

        if (argc==2)
		if (!command_warning(readwarn)) return;

	/*
	 * Read the geometry block and then read the flaw map
	 * and write it to the backup file.
	 */

	read_geometry_block();
	if (argc==2) {
        	strcpy(hflaw_backup_file, argv[1]);
		if (read_hflaw())
			write_hflaw_backup();
	}
}

/*************************************************************************
*  Procedure:	read_geometry_block
*
*  Description: This procedure reads the geometry block. The start of
*               the geo block contains the ascii string "geometry";
*               the end of the block contains a checksum.
*
*  Returns:	"disk_geometry_block" is set TRUE or FALSE to
*               indicate whether a geometry block is present on disk.
*               Disk geometry block is put in "disk_geometry".
*               In-core geometry block is initialized in "geometry".
**************************************************************************/

read_geometry_block( )
{
	int err,i;

	printf("Reading the geometry block.\n");
	disk_geometry_block = FALSE;
	err = disk_def_io( HDC_READ, &disk_geometry_sector, 0,
		sizeof(geometry_sector) );
	bcopy(	&disk_geometry_sector,
		&incore_geometry_sector,
		sizeof(disk_geometry_sector));
	if ( geo->version > 64000  ||  geo->version < 0 )
		err = TRUE;
	if ( !err  &&  strcmp(&geo->id[0],GB_ID) == 0) {

		/*
		 * A geometry block is present.
		 * Test the geometry block checksum.
		 */

		disk_geometry_block = TRUE;
 		GB_CHECKSUM( geo, incore_geometry_sector.checksum );
 		if (incore_geometry_sector.checksum !=
 					disk_geometry_sector.checksum) {
 			printf("The geometry block checksum is incorrect.\n");
 			printf("Check the partitions and then 'write' the ");
 			printf("new geometry block and flaw map.\n");
 		}
       }
       else {
		/*
 		 * There is no geometry block.
 		 * Initialize empty in-core and on-disk geometry blocks;
 		 * then initialize default partitions for the in-core one.
		 */

        	printf("There is no geometry block on this disk.\n");
        	bzero(&incore_geometry_sector, sizeof(incore_geometry_sector));
		strcpy( &geo->id[0], GB_ID);
		geo->version = GB_VERSION;
		bcopy(	&incore_geometry_sector,
			&disk_geometry_sector,
			sizeof(disk_geometry_sector));
		partition_defaults();
	}
}

/*************************************************************************
*  Procedure:	read_hflaw
*
*  Description: This procedure reads the flaw map from disk.
*
*  Returns:     hflaw_count = the number of flaw records.
*               TRUE =   the flaw map was successfully read.
*               FALSE =  there was an error.
**************************************************************************/

read_hflaw( )
{
	char *c;
	int i;

	printf("Reading the flaw map from disk.\n");
	hflaw_count = 0;
	hflaw_bytes = 0;
	if ( geo->flaw_offset == 0  ||  geo->flaw_size == 0 ) {
		printf("There is no flaw map on the disk.\n");
		return TRUE;
	}
	if (disk_def_io( HDC_READ, hflaw, geo->flaw_offset, geo->flaw_size ) )
		return (FALSE);
	hflaw_bytes = geo->flaw_size;
	if ( !checksum( &hflaw[0], geo->flaw_size) == geo->flaw_checksum ) {
		printf("Flaw map checksum error.\n");
		return (FALSE);
	}

	/*
	 * The flaws are in. Count 'em.
	 */

	c = (char *) &hflaw[0];
	for (i=0; i<hflaw_bytes; i++)
		if (c[i]=='\n') hflaw_count++;
	printf("Read %d flaw records.\n", hflaw_count);
	return TRUE;
}

/*************************************************************************
*  Procedure:	read_hflaw_backup
*
*  Description: This procedure reads the harris-format flaw map
*               from the flaw map backup file.
*
*  Returns:     FALSE means there was an error.
*               hflaw = the entire harris-format flaw map;
*               hflaw_count = the number of flaw entries;
**************************************************************************/

read_hflaw_backup( )
{
	char *c;
	char filename[132];
        int  fd,f,i,bytes;

        hflaw_count = 0;
        hflaw_bytes = 0;
	strcpy(filename,hflaw_backup_file);

        /*
         * If the backup file name is "/dev/null", there are no flaws.
         * This is specifically tested for here so that it will work for
         * the stand-alone version also.
         */

	if ( strcmp(filename,"/dev/null") == 0)
		return TRUE;

        /*
         * Read in flaws.
         */

	printf("Opening the flaw map file '%s'\n",filename);
        fd = open(filename,0);
        if (fd<0) {
        	printf("Could not open the flaw map file.\n");
        	return FALSE;
        }
	printf("Reading flaw records.\n");
	do {
		bytes = read(fd, hflaw_input, sizeof(hflaw_input));
		if (bytes < 0) {
			perror("format: cannot read flaw file");
			hflaw_bytes = 0;
			break;
		}
	        if (bytes > 0)
			bcopy( hflaw_input, (int)hflaw + hflaw_bytes, bytes);
		hflaw_bytes += bytes;
	} while (bytes > 0);
	close(fd);

	/*
	 * The flaws are in. Count 'em.
	 */

	c = (char *) &hflaw[0];
	for (i=0; i<hflaw_bytes; i++)
		if (c[i]=='\n') hflaw_count++;
	printf("Read %d records.\n", hflaw_count);
	return TRUE;
}

/*************************************************************************
*  Procedure:	status
*
*  Description: This procedure processes the "status" command, which
*               provides detailed information about the disk. Info
*               is obtained from the profile prom and geometry block.
*
*  Returns:
**************************************************************************/

status( argc, argv )

int  		argc ; 		/* number of arguments in the
                                 * "status" command line
                                 */
char 		*argv[] ;	/* array of pointers to argument
      				 * values on the command line
				 */
{
	int i;

	/*
	 * Print format is:
	 *
	 *      /dev/rdsk/15s7
	 *      disk model 81  (removable)
	 *	megabytes: 552 logical, 690 physical
	 *	cylinders: 836 logical, 842 physical
	 *	heads:     20 logical, 20 physical
	 *	sectors:   44 logical, 45 physical
	 *	bytes:     512 logical, 612 physical
	 *	1 drive description cylinders at cylinder 833
	 *	2 diagnostic cylinders at cylinder 834
	 *	3951 revolutions per minute
	 *	geometry block version 0  (no geometry block)
	 *	flaw map size is 2742 bytes
	 *      partition 0   75 cylinders    0 start cylinder
	 *          <etc>
	 *      partition 7    3 cylinders  833 start cylinder
	 *      836 cylinders used, 0 cylinders unused
	 */

	printf("\n%s\n",dsk_device);
	printf("disk model %d  %s", disk_status.id,
		(disk_status.id >= HDC_REMOVABLE) ? "(removable)\n" : "\n" );
	printf("megabytes: %d logical, %d physical\n",
		megabytes(total_sectors * BPS),
		megabytes(disk_status.phys_heads *
			disk_status.phys_sectors *
		        disk_status.phys_cylinders *
		        disk_status.bytes_per_sec) );
	printf("cylinders: %d logical, %d physical\n",
		disk_status.cylinders, disk_status.phys_cylinders);
	printf("heads:     %d logical, %d physical\n",
		disk_status.heads, disk_status.phys_heads);
	printf("sectors:   %d logical, %d physical\n",
		disk_status.sectors, disk_status.phys_sectors);
	printf("bytes:     %d logical, %d physical\n",
		disk_status.id==0 ? 0 : BPS, disk_status.bytes_per_sec);
	printf("%d drive description cylinders at cylinder %d\n",
		disk_status.def_cylinders, disk_status.def_cyl);
	printf("%d diagnostic cylinders at cylinder %d\n",
		disk_status.diag_cylinders, disk_status.diag_cyl);
	printf("%d revolutions per minute\n", disk_status.rpm);
	printf("geometry block version %d  %s",
		disk_geometry_block ? geo->version : 0,
		disk_geometry_block ? "\n" : "(no geometry block)\n");
	printf("flaw map size is %d bytes\n",
		disk_geometry_block ? geo->flaw_size : 0);
	for (i=0;i<8;i++) print_partition(i,disk_geo);
	print_partition_total(disk_geo);
	printf("\n");
}

/*************************************************************************
*  Procedure:	terminate
*
*  Description: This procedure terminates this program.
*
*  Returns:	n/a
**************************************************************************/

terminate(err)
int err;
{
	ioctl( fd, DSKIOCFORMATCTL, 0);
	close(fd);
	exit(err);
}

/*************************************************************************
*  Procedure:	unit
*
*  Description: This procedure processes the "unit" command.
*
*  Returns:
**************************************************************************/

unit( argc, argv )

int		argc ;		/* number of arguments in the
				 * "unit" command line
				 */
char 		*argv[] ;	/* array of pointers to argument
       				 * values on the command line
				 */
{
	char newunit;
	int  count;

        /*
         * If no parameters, print current unit.
         */

	if (argc==1) {
		printf("partition unit is ");
		switch(partition_unit) {
		case 'c': printf("cylinders");
			  break;
		case 'k': printf("kbytes");
			  break;
		case 's': printf("sectors");
			  break;
		case 'm': printf("megabytes");
			  break;
		case 'p': printf("percent");
			  break;
		}
		printf("\n");
  		return;
  	}

        /*
         * Scan and validate the "unit" command;
         * command format is: "unit <unit>"
         */

     	count = sscanf( argv[1], "%c", &newunit);
	if (count!=1) {
 		printf("Invalid unit.\n");
 		return;
 	}

	/*
	 * Set the partition unit.
	 */

	switch(newunit) {
		case '%':
		        newunit = 'p';
		case 'c':
		case 'k':
		case 's':
		case 'm':
		case 'p':
			partition_unit = newunit;
			break;

		default:
			printf("Invalid partition unit.\n");
	}
}

/*************************************************************************
*  Procedure:	verify
*
*  Description: This procedure processes the "verify" command.
*
*  Returns:
**************************************************************************/

verify( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "verify" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
        verify_proc();
}

#ifndef STANDALONE
/*************************************************************************
*  Procedure:	verify_option
*
*  Description: This procedure processes the "verify" command when the
*               command was invoked via an option on the "format"
*               command line; i.e. "format -v <disk>".
*
*  Returns:
**************************************************************************/

verify_option(argc,argv)

 	int  		argc ; 		/* number of arguments in the
                                         * command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
        if (!verify_proc())
        	terminate(1);
}
#endif

/*************************************************************************
*  Procedure:	verify_proc
*
*  Description: This procedure verifies the disk.
*
*  Returns:     FALSE = error occurred
*************************************************************************/

verify_proc()
{
	int cyl,sect,err,byte,head;
	struct formatop track;

        /*
         * Print a header every 50 cylinders and a dot after each cylinder.
         */

        printf("Verifying.\n");
        for ( cyl=0; cyl<disk_status.cylinders; cyl++ ) {
        	if ( !(cyl % 50) )
        		printf("\ncylinder %4d of %4d ",
        			cyl, disk_status.cylinders);
        	printf(".");
        	fflush(stdout);

		/*
		 * For each track on this cylinder, issue a verify ioctl.
		 */

        	for ( head=0; head<disk_status.heads; head++ ) {
        		track.cylinder  = cyl;
        		track.head      = head;
 			if ( ioctl(fd, DSKIOCVERIFY, &track) < 0 ) {
 				perror("\nformat: cannot verify track");
 				printf("Verification aborted.\n");
 				return FALSE;
 			}
        	}
        }
        printf("\nVerification complete.\n");
        return TRUE;
}

/*************************************************************************
*  Procedure:	writeg
*
*  Description: This procedure processes the "write" command.
*               It writes the geometry block (and optionally the
*               flaw map) to disk.
*
*  Returns:	n/a
*************************************************************************/

writeg( argc, argv )

 	int  		argc ; 		/* number of arguments in the
                                         * "write" command line
                                         */
        char 		*argv[] ;	/* array of pointers to argument
        				 * values on the command line
					 */
{
	if (!command_warning(writewarn)) return;

	/*
	 * Write the geometry block to disk.
	 * If a filepath was specified, also write the flaw map.
	 */

        if (argc==2) {
        	strcpy(hflaw_backup_file, argv[1]);
        	if (!read_hflaw_backup()) {
        		printf("Aborting the 'write' command.\n");
        		return;
        	}
        	write_hflaw();
        }
	write_geometry_block();
}

/*************************************************************************
*  Procedure:	write_geometry_block
*
*  Description: This procedure writes the geometry block.
*
*  Returns:     The on-disk geometry block structure is set equal
*               to the in-core geometry block structure.
**************************************************************************/

write_geometry_block( )
{
	int err,i;
	geometry_status new_geo;

	/*
	 * Add the checksum to the geometry block.
	 * Copy the in-core version of the geometry block to the
	 * on-disk version, then write out the on-disk version.
	 */

	geo->version  = GB_VERSION;
	GB_CHECKSUM( geo, incore_geometry_sector.checksum );
	printf("Writing the geometry block to disk.\n");
	err = disk_def_io( HDC_WRITE, &incore_geometry_sector, 0,
		sizeof(geometry_sector) );
	if (err) {
		printf("Could not write the geometry block.\n");
		return;
	}
	bcopy(	&incore_geometry_sector,
		&disk_geometry_sector,
		sizeof(disk_geometry_sector));
	disk_geometry_block = TRUE;

	/*
	 * Inform the handler about the new disk geometry so that
	 * a reboot won't be required.
	 */

	for (i=0; i<GB_MAXPART; i++) {
		new_geo.partition[i].length = geo->partition[i].length;
		new_geo.partition[i].start  = geo->partition[i].start;
	}
 	if ( ioctl(fd,DSKIOCSETGEOMETRY,&new_geo) < 0 )
 		printf("New geometry won't take affect until next reboot.\n");
}

/*************************************************************************
*  Procedure:	write_hflaw
*
*  Description: This procedure writes the harris flaw map to disk.
*               The flaw map is written immediately after the geometry
*               block (which resides on the first sector).
*
*  Returns:	The in-core geometry block is updated with flaw map info.
**************************************************************************/

write_hflaw( )
{
	if (hflaw_count==0) {
		printf("There is no flawmap.\n");
		return;
	}
	printf("Writing the flaw map to disk.\n");
	geo->flaw_offset = BPS*HDC_SPB;
	geo->flaw_size = hflaw_bytes;
	geo->flaw_checksum = checksum(&hflaw[0],geo->flaw_size);
	disk_def_io( HDC_WRITE, hflaw, geo->flaw_offset, geo->flaw_size );
}

/*************************************************************************
*  Procedure:	write_hflaw_backup
*
*  Description: This procedure writes the flaw map to the flaw map
*               backup file.
*               It never gets called from standalone mode.
*
*               hflaw = the entire harris-format flaw map;
*               hflaw_count = the number of flaw entries in hflaw
*               hflaw_bytes = the number of bytes in hflaw
*
*  Returns:
**************************************************************************/

write_hflaw_backup( )
{
        int  f,fd;
        char filename[132];

	if (hflaw_count==0) {
		printf("There is no flawmap.\n");
		return;
	}
	strcpy(filename,hflaw_backup_file);
	fd = open( filename, O_RDWR+O_CREAT+O_TRUNC, 0444 );
        if (fd<0) {
        	printf("Could not open the flaw map file '%s'.\n",
        		hflaw_backup_file);
        	return;
        }
        printf("Writing %d flaw records to '%s'.\n",
        	hflaw_count,hflaw_backup_file);
	if (write(fd,hflaw,hflaw_bytes) < 0)
		perror("format: cannot write flaw file");
	close(fd);
}

/*************************************************************************
*  Procedure:	yes
*
*  Description: This procedure reads a "yes" or "no" response.
*
*  Returns:     TRUE means the user typed "yes".
**************************************************************************/

int
yes( )
{
	char line[132];

	printf(" (y|n) ");
	if (gets(line) == 0)
		if (feof(stdin))
		       	clearerr(stdin);
	return (line[0]=='y' ?  TRUE : FALSE);
}
