/*	qt.c	Tahoe version 	May 1986.	*/
/*	qt.c	HCX version 	May 1987.	*/

/*
** Quarter inch tape driver. Stand alone version.
*/

#define	NQT	1

#include "../uts/machine/sys/types.h"
#define KERNEL
#include "../uts/machine/sys/param.h"
#undef KERNEL
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../uts/machine/ml/pte.h"
#include "../uts/machine/ml/mtpr.h"
#include "../stand/saio.h"
#include "../uts/machine/sys/scsi.h"
#include "../uts/machine/sys/qt.h"

#define BLK_FACTOR	(DEV_BSIZE / QT_BSIZE)
static qt_tab		qt_unit;
static int		retrying = 0;
static sc_com		com;
static struct iob	lio;


/*
 * Reset the controller.
 */

qtopen(io)
register struct iob *io;
{
	register int	lun;

	qt_unit.qt_open = 1;
	/*
	**	If the load fails the first time and the drive was found
	** then it might be a power up or a new cartidge in drive.
	*/
	if(qtcmd(io, QT_LOAD) < 0)
		if(qtcmd(io, QT_LOAD)) {
			qt_unit.qt_open = 0;
			return false;
		}
	qtcmd(io, QT_MODE_SELECT);
	if(qtcmd(io, QT_REWIND) < 0)
		_stop("qt: Open rewind failed!\n");
	
	io->i_bn = 0;
	if(qtcmd(io, QT_SPACE))
		_stop("qt: File not found.");

	qt_unit.qt_filenum = io->i_boff;
	qt_unit.qt_blknum = 0;
	return true;
}


/*
**	If tape was open for writing write two file marks,
** then rewind tape.
*/

qtclose(io)
register struct iob *io;
{
	if (io->i_flgs & F_WRITE) {
		io->i_cc = 2;
		qtcmd(io, QT_WRITE_FM);
	}
	qtcmd(io, QT_REWIND);
	qtcmd(io, QT_UNLOAD);
	qt_unit.qt_filenum = 0;
	qt_unit.qt_blknum = 0;
}


qtstrategy(io,func)
register struct iob *io;
register long	func;
{
	register	count;


	/* all transfers must be a multiple of 512 bytes */
	if(io->i_cc % QT_BSIZE)
		return -1;

	/* seek to correct block */
	if(qtseek(io))
		return -1;

	/* do command */
	count = qtcmd(io, (func == READ) ? QT_READ : QT_WRITE);

	/* reset retrying flag if a retry occured */
	retrying = 0;
	return count;
}


/*
**
*/

qtseek(io)
struct iob	*io;
{
	lio = *io;
	lio.i_bn *= BLK_FACTOR;
	if ((lio.i_bn<qt_unit.qt_blknum) || (lio.i_boff<qt_unit.qt_filenum)) {
		(void)qtcmd(&lio, QT_REWIND);
		qt_unit.qt_filenum = 0;
		qt_unit.qt_blknum = 0;
	}
	if ((lio.i_bn>qt_unit.qt_blknum) || (lio.i_boff>qt_unit.qt_filenum))
		return qtcmd(&lio, QT_SPACE);
	return 0;
}


/*
**
*/

qtcmd(io,func)
register struct iob *io;
long	func;
{
	register long	count;
	register int	temp_status;

	switch (func) {
	case QT_READ:
	case QT_WRITE:
		load_sc_cmd(&com, func, 0, 0, QT_RW | (io->i_cc / QT_BSIZE));
		qt_unit.qt_blknum += io->i_cc / QT_BSIZE;
		break;
	case QT_SPACE:
		if (io->i_boff > qt_unit.qt_filenum) {
			io->i_cc = io->i_boff - qt_unit.qt_filenum;
			load_sc_cmd(&com,QT_SPACE,0,0,(QT_FILES | io->i_cc));
			if(scstrategy(io, &com) != 0)
				return -1;
			qt_unit.qt_blknum = 0;
		}
		if (io->i_bn > qt_unit.qt_blknum) {
			io->i_cc = io->i_bn - qt_unit.qt_blknum;
			load_sc_cmd(&com,QT_SPACE,0,0,(QT_BLOCKS | io->i_cc));
			if(scstrategy(io, &com) != 0)
				return -1;
			qt_unit.qt_blknum = io->i_bn;
		}
		return 0;
	case QT_REWIND:
		load_sc_cmd(&com, QT_REWIND, 0, 0, 0);
		qt_unit.qt_filenum = 0;
		qt_unit.qt_blknum = 0;
		qt_unit.qt_eof = 0;
		break;
	case QT_UNLOAD :
		load_sc_cmd(&com, QT_LOAD, 0, 0, QT_UNLOADIT);
		qt_unit.qt_filenum = 0;
		qt_unit.qt_blknum = 0;
		qt_unit.qt_eof = 0;
		break;
	case QT_MODE_SELECT :
		{
			qt_mode		mode_params;

			bzero(&mode_params, sizeof(qt_mode));
			load_sc_cmd(&com, QT_MODE_SELECT, 0, 0, 4);
			mode_params.bufmode = 1;
			mode_params.disc_len = 0;
			io->i_ma = (char *)&mode_params;
			io->i_cc = 4;
			return qtexecute(io, &com);
		}
	case QT_LOAD :
		load_sc_cmd(&com, QT_LOAD, 0, 0, QT_LOADIT);
		qt_unit.qt_filenum = 0;
		qt_unit.qt_blknum = 0;
		qt_unit.qt_eof = 0;
		break;
	case QT_RETENSION :
		load_sc_cmd(&com, QT_LOAD, 0, 0, QT_RET);
		qt_unit.qt_filenum = 0;
		qt_unit.qt_blknum = 0;
		qt_unit.qt_eof = 0;
		break;
	case QT_WRITE_FM :
		load_sc_cmd(&com, QT_WRITE_FM, 0, 0, io->i_cc);
		qt_unit.qt_filenum += io->i_cc;
		qt_unit.qt_blknum = 0;
		qt_unit.qt_eof = 0;
		break;
	}
	return qtexecute(io, &com);
}


qtexecute(io, cmd)
struct iob	*io;
cpb		*cmd;
{
	register int cmd_status = scstrategy(io, cmd);
	
	if(cmd_status == -1)
		return qt_decode_error(io, cmd);
	return cmd_status;
}


qtprint_error(message, val0, val1, val2)
register char	*message;
register long	val0, val1, val2;
{
	printf("qt0: ");
	printf(message, val0, val1, val2);
	printf("!\n");
}

/*
*/

qt_decode_error(io, func)
register struct iob	*io;
{
	register int	 i;
	register char	 *j;

	switch (com.scack.ack_sense.s_class) {
	case S_DRIVE :
		switch(com.scack.ack_sense.s_code) {
		case SC_NS :
			return io->i_cc;
		case SC_HE :
			qtprint_error("Unrecoverable hardware error");
			break;
		case SC_DNR :
			qtprint_error("Drive not ready");
			break;
		case SC_MNL :
			qtprint_error("Media not loaded");
			break;
		case SC_EOT :
			qtprint_error("End of tape");
			break;
		case SC_DT :
			qtprint_error("Drive timeout");
			break;
		default:
			qtprint_error("Unknown drive code '0x%x'",
			     com.scack.ack_sense.s_code);
			_stop("");
			break;
		}
		break;
	case S_TARGET :
		switch(com.scack.ack_sense.s_code) {
		case SC_UDE :
		case SC_BBF :
			if(qtretry(io, func) == -1) {
				qtprint_error("Uncorrectable data error");
				return -1;
			}
			return io->i_cc;
		case SC_BNF :
			qtprint_error("Block not found");
			break;
		case SC_DTE :
			qtprint_error("DMA timeout");
			break;
		case SC_WP :
			qtprint_error("Drive write protected");
			break;
		case SC_CDC :
			return	io->i_cc;
			break;
		case SC_FMD :
			qt_unit.qt_eof = 1;
			break;
		case SC_CE :
			qtprint_error("Data compare error");
			break;
		default:
			qtprint_error("Unknown target code '0x%x'",
			     com.scack.ack_sense.s_code);
			_stop("");
			break;
		}
		break;
	case S_SYS :
		switch(com.scack.ack_sense.s_code) {
		case SC_IC :
			qtprint_error("Unauthorized media or invalid command");
			break;
		default:
			qtprint_error("Unknown system code '0x%x'",
			     com.scack.ack_sense.s_code);
			_stop("");
			break;
		}
		break;
	case S_VENDOR :
		switch(com.scack.ack_sense.s_code) {
		case SC_UA :
			break;
		case SC_CT :
			qtprint_error("Command timeout");
			break;
		case SC_AE :
			qtprint_error("Can only write at end of media");
			break;
		case SC_REOM :
			qtprint_error("Can't read past end of media");
			break;
		default:
			qtprint_error("Unknown vendor code '0x%x'",
			     com.scack.ack_sense.s_code);
			_stop("");
			break;
		}
		break;
	default :
		qtprint_error("Unknown error class '0x%x'",
		     com.scack.ack_sense.s_class);
		_stop("");
		break;
	}
	return -1;
}


/*
**
*/

qtretry(io, func)
struct iob	*io;
{
	if(!retrying) {
		retrying = 1;
		lio = *io;
		lio.i_cc = 0;
		qtcmd(&lio, QT_RETENSION);
		return qtstrategy(io, func);
	}
	return -1;
}
