/*
**  Title:	SC.C - Plessey SCSI-1 SCSI interface controller driver.
**
**  Author:	John R. Franks
**  Date:	24-Apr-84
**
**	This driver drives the scsi buss adapter manfuactured by Plessey.
**  The driver accepts commands from other pseudo drivers which know all about
**  the individual scsi periperal adapters.
**
**	Each Plessey adapter looks like 7 seperate scsi buss controllers.
**  
** Adapted for the HCX
**  
**  Author:	Terry Treder  
**  Date:	21-May-87
*/

#include "../uts/machine/sys/types.h"
#define KERNEL
#include "../uts/machine/sys/param.h"
#undef KERNEL
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include	"../uts/machine/ufs/inode.h"
#include	"../uts/machine/ufs/fs.h"
#include	"../uts/machine/ml/pte.h"
#include	"../uts/machine/ml/mtpr.h"
#include	"../stand/saio.h"
#include	"../uts/machine/sys/scsi.h"

#define TRUE	1
#define FALSE	0

static cpb	CPB;

static rcb 	*ctlr_addr0 = (rcb *)(0xe0010000 + 0xff000000);
static rcb 	*ctlr_addr1 = (rcb *)(0xa0010000 + 0xbf000000);

/*
**
*/

scstrategy(io, request)
register struct iob *io;
register sc_com *request;
{
	register int	time;
	register cpb	*Cpb;

	Cpb = &CPB;


	/* clean out old junk out of command block to start */
	bzero(Cpb, sizeof(cpb));

	/* Load command block */
	Cpb->cpb_command_block = request->sccmd;

	/* load data pointer and length (none for now) */
	Cpb->cpb_data_address.sc_mod = ADDR_MODIFIER;
	Cpb->cpb_data_address.sc_addr = io->i_ma;
	Cpb->cpb_data_length = io->i_cc;

	/* load link address (none) */
	Cpb->cpb_link_address.sc_mod = ADDR_MODIFIER;
	Cpb->cpb_link_address.sc_addr = 0;

	/* load protocol block */
	Cpb->cpb_protocol.auto_sense = TRUE;
	Cpb->cpb_protocol.int_disable = TRUE;
	Cpb->cpb_protocol.xfer_32_bit = TRUE;
	Cpb->cpb_protocol.int_level = SCSI_LEVEL;
	Cpb->cpb_protocol.int_vector = (char)0;

	Cpb->cpb_target = io->i_drive;
	Cpb->cpb_lun = 0;

	time = 15000000;

	/* and start command */
	if (io->i_bus == 0) {
		SCSI_GO(ctlr_addr0, Cpb);
	}
	else {
		SCSI_GO(ctlr_addr1, Cpb);
	}

	/* now wait for completion */
	while(!(Cpb->cpb_status.cmd_complete) && time) {
		DELAY(20);
		uncache(&Cpb->cpb_status);
		time--;
	}
	mtpr(0, PADC);

	/* load the return status for pseudo driver to interpret */
	bcopy(Cpb->cpb_sense, &request->scack.ack_sense, 12);
	request->scack.ack_status = Cpb->cpb_status;
	if(Cpb->cpb_status.bad_free || Cpb->cpb_status.bad_phase ||
	    Cpb->cpb_status.sc_timeout || Cpb->cpb_status.bad_reset ||
	    Cpb->cpb_status.long_status || Cpb->cpb_status.long_command ||
	    Cpb->cpb_status.bus_error || Cpb->cpb_status.check_status) {
		return -1;
	}
	if(Cpb->cpb_status.underflow)
		return (io->i_cc - Cpb->cpb_data_length);
	return io->i_cc;
}


print_cpb(ptr)
register unsigned char	*ptr;
{
	register int i,j,delay;

	printf("Dump of CPB:\n");
	printf("  SCSI command block:");
	for(i=0; i<2; i++) {
		for(delay=0;delay<500000;delay++);
		printf("\n\t");
		for(j=0; j<8; j++) {
			uncache(ptr);
			printf("        0x%x  ", *(ptr++) & 0xff);
		}
	}
	printf("\n  Data address:\n\t");
	for(i=0; i<8; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  Transfer length:\n\t");
	for(i=0; i<4; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	for(delay=0;delay<500000;delay++);
	printf("\n  Link address:\n\t");
	for(i=0; i<8; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  protocol block:\n\t");
	for(i=0; i<4; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  SCSI status:\n\t");
	for(i=0; i<2; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	for(delay=0;delay<500000;delay++);
	printf("\n  Host status\n\t");
	for(i=0; i<2; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	uncache(ptr);
	printf("\n  Target identfier:\n\t0x%x", *(ptr++) & 0xff);
	uncache(ptr);
	for(delay=0;delay<500000;delay++);
	printf("\n  LUN identfier:\n\t0x%x", *(ptr++) & 0xff);
	printf("\n  SCSI sense block:");
	for(i=0; i<2; i++) {
		printf("\n\t");
		for(j=0; j<6; j++) {
			uncache(ptr);
			printf("0x%x  ", *(ptr++) & 0xff);
		}
	}
	for(delay=0;delay<500000;delay++);
	printf("\n\n");
}


/*
**
*/

load_sc_cmd(ptr, cmd, group_code, lun, depend)
register unsigned char	*ptr;
unsigned char		cmd;
unsigned char		group_code;
unsigned char		lun;
register unsigned long	depend;
{
	*(ptr++) = ((group_code & 0x07) << 5) | (cmd & 0x1f);
	*(ptr++) = (unsigned char)((lun & 0x07)<<5)|((depend & 0x1f000000)>>24);
	*(ptr++) = (unsigned char)((depend & 0xff0000) >> 16);
	*(ptr++) = (unsigned char)((depend & 0xff00) >> 8);
	*(ptr++) = (unsigned char)(depend & 0xff);
	*ptr = 0;
}

