/*	sys.c	6.2	83/09/23	*/

#define KERNEL
#include "../uts/machine/sys/param.h"
#undef  KERNEL
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../uts/machine/sys/dir.h"
#include "../uts/machine/ml/mtpr.h"
#include "../stand/saio.h"
#include "../uts/machine/sys/reboot.h"

extern int buses[ ] ;

#define MAX_BUS		 2
#define MAX_PART	 8
#define MAX_DRIVE	16
#define MAX_VERSA_CTLR	16
#define MAX_HVME_CTLR	21

#define SKIP_BLANKS( _ptr ) while ( *_ptr == ' ' ) _ptr ++
#define NO_ERROR   0

ino_t	dlook();

struct dirstuff {
	int loc;
	struct iob *io;
};

static
openi(n, io)
	register struct iob *io;
{
	register struct dinode *dp;
	int cc;

	io->i_offset = 0;
	io->i_bn = fsbtodb(&io->i_fs, itod(&io->i_fs, n)) + io->i_boff;
	io->i_cc = io->i_fs.fs_bsize;
	io->i_ma = io->i_buf;
	cc = devread(io);
	dp = (struct dinode *)io->i_buf;
	io->i_ino.i_ic = dp[itoo(&io->i_fs, n)].di_ic;
	return (cc);
}

static
find(path, file)
	register char *path;
	struct iob *file;
{
	register char *q;
	char c;
	int n;

	if (path==NULL || *path=='\0') {
		printf("null path\n");
		return (0);
	}

	if (openi((ino_t) ROOTINO, file) < 0) {
		printf("can't read root inode\n");
		return (0);
	}
	while (*path) {
		while (*path == '/')
			path++;
		q = path;
		while(*q != '/' && *q != '\0')
			q++;
		c = *q;
		*q = '\0';

		if ((n = dlook(path, file)) != 0) {
			if (c == '\0')
				break;
			if (openi(n, file) < 0)
				return (0);
			*q = c;
			path = q;
			continue;
		} else {
			printf("%s not found\n", path);
			return (0);
		}
	}
	return (n);
}

static daddr_t
sbmap(io, bn)
	register struct iob *io;
	daddr_t bn;
{
	register struct inode *ip;
	int i, j, sh;
	daddr_t nb, *bap;

	ip = &io->i_ino;
	if (bn < 0) {
		printf("bn negative\n");
		return ((daddr_t)0);
	}

	/*
	 * blocks 0..NDADDR are direct blocks
	 */
	if(bn < NDADDR) {
		nb = ip->i_db[bn];
		return (nb);
	}

	/*
	 * addresses NIADDR have single and double indirect blocks.
	 * the first step is to determine how many levels of indirection.
	 */
	sh = 1;
	bn -= NDADDR;
	for (j = NIADDR; j > 0; j--) {
		sh *= NINDIR(&io->i_fs);
		if (bn < sh)
			break;
		bn -= sh;
	}
	if (j == 0) {
		printf("bn ovf %D\n", bn);
		return ((daddr_t)0);
	}

	/*
	 * fetch the first indirect block address from the inode
	 */
	nb = ip->i_ib[NIADDR - j];
	if (nb == 0) {
		printf("bn void %D\n",bn);
		return ((daddr_t)0);
	}

	/*
	 * fetch through the indirect blocks
	 */
	for (; j <= NIADDR; j++) {
		if (blknos[j] != nb) {
			io->i_bn = fsbtodb(&io->i_fs, nb) + io->i_boff;
			io->i_ma = b[j];
			io->i_cc = io->i_fs.fs_bsize;
			if (devread(io) != io->i_fs.fs_bsize) {
				if (io->i_error)
					errno = io->i_error;
				printf("bn %D: read error\n", io->i_bn);
				return ((daddr_t)0);
			}
			blknos[j] = nb;
		}
		bap = (daddr_t *)b[j];
		sh /= NINDIR(&io->i_fs);
		i = (bn / sh) % NINDIR(&io->i_fs);
		nb = bap[i];
		if(nb == 0) {
			printf("bn void %D\n",bn);
			return ((daddr_t)0);
		}
	}
	return (nb);
}

static ino_t
dlook(s, io)
	char *s;
	register struct iob *io;
{
	register struct direct *dp;
	register struct inode *ip;
	struct dirstuff dirp;
	int len;

	if (s == NULL || *s == '\0')
		return (0);
	ip = &io->i_ino;
	if ((ip->i_mode&IFMT) != IFDIR) {
		printf("not a directory\n");
		return (0);
	}
	if (ip->i_size == 0) {
		printf("zero length directory\n");
		return (0);
	}
	len = strlen(s);
	dirp.loc = 0;
	dirp.io = io;
	for (dp = readdir(&dirp); dp != NULL; dp = readdir(&dirp)) {
		if(dp->d_ino == 0)
			continue;
		if (dp->d_namlen == len && !strcmp(s, dp->d_name))
			return (dp->d_ino);
	}
	return (0);
}

/*
 * get next entry in a directory.
 */
struct direct *
readdir(dirp)
	register struct dirstuff *dirp;
{
	register struct direct *dp;
	register struct iob *io;
	daddr_t lbn, d;
	int off;

	io = dirp->io;
	for(;;) {
		if (dirp->loc >= io->i_ino.i_size)
			return (NULL);
		off = blkoff(&io->i_fs, dirp->loc);
		if (off == 0) {
			lbn = lblkno(&io->i_fs, dirp->loc);
			d = sbmap(io, lbn);
			if(d == 0)
				return NULL;
			io->i_bn = fsbtodb(&io->i_fs, d) + io->i_boff;
			io->i_ma = io->i_buf;
			io->i_cc = blksize(&io->i_fs, &io->i_ino, lbn);
			if (devread(io) < 0) {
				errno = io->i_error;
				printf("bn %D: read error\n", io->i_bn);
				return (NULL);
			}
		}
		dp = (struct direct *)(io->i_buf + off);
		dirp->loc += dp->d_reclen;
		if (dp->d_ino == 0)
			continue;
		return (dp);
	}
}

lseek(fdesc, addr, ptr)
	int fdesc, ptr;
	off_t addr;
{
	register struct iob *io;

	if (ptr != 0) {
		printf("Seek not from beginning of file\n");
		errno = EOFFSET;
		return (-1);
	}
	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((io = &iob[fdesc])->i_flgs & F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	io->i_offset = addr;
	io->i_bn = addr / DEV_BSIZE;
	io->i_cc = 0;
	return (0);
}

getc(fdesc)
	int fdesc;
{
	register struct iob *io;
	register struct fs *fs;
	register char *p;
	int c, lbn, off, size, diff;


	if (fdesc >= 0 && fdesc <= 2)
		return (getchar());
	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((io = &iob[fdesc])->i_flgs&F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	p = io->i_ma;
	if (io->i_cc <= 0) {
		if ((io->i_flgs & F_FILE) != 0) {
			diff = io->i_ino.i_size - io->i_offset;
			if (diff <= 0)
				return (-1);
			fs = &io->i_fs;
			lbn = lblkno(fs, io->i_offset);
			io->i_bn = fsbtodb(fs, sbmap(io, lbn)) + io->i_boff;
			off = blkoff(fs, io->i_offset);
			size = blksize(fs, &io->i_ino, lbn);
		} else {
			io->i_bn = io->i_offset / DEV_BSIZE;
			off = 0;
			size = DEV_BSIZE;
		}
		io->i_ma = io->i_buf;
		io->i_cc = size;
		if (devread(io) < 0) {
			errno = io->i_error;
			return (-1);
		}
		if ((io->i_flgs & F_FILE) != 0) {
			if (io->i_offset - off + size >= io->i_ino.i_size)
				io->i_cc = diff + off;
			io->i_cc -= off;
		}
		p = &io->i_buf[off];
	}
	io->i_cc--;
	io->i_offset++;
	c = (unsigned)*p++;
	io->i_ma = p;
	return (c);
}

/* does this port?
getw(fdesc)
	int fdesc;
{
	register w ;
	register i ;
	int val;

	for (i = 0, val = 0; i < sizeof(val); i++) {
		w = getc(fdesc);
		if (w < 0) {
			if (i == 0)
				return (-1);
			else
				return (val);
		}
		val = (val << 8) | (w & 0xff);
	}
	return (val);
}
*/
int	errno;

read(fdesc, buf, count)
	int fdesc, count;
	char *buf;
{
	register i;
	register struct iob *file;

	errno = 0;
	if (fdesc >= 0 && fdesc <= 2) {
		i = count;
		do {
			*buf = getchar();
		} while (--i && *buf++ != '\n');
		return (count - i);
	}
	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((file = &iob[fdesc])->i_flgs&F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	if ((file->i_flgs&F_READ) == 0) {
		errno = EBADF;
		return (-1);
	}
	if ((file->i_flgs & F_FILE) == 0) {
		file->i_cc = count;
		file->i_ma = buf;
		file->i_bn = file->i_boff + (file->i_offset / DEV_BSIZE);
		i = devread(file);
		file->i_offset += count;
		if (i < 0)
			errno = file->i_error;
		return (i);
	} else {
		if (file->i_offset+count > file->i_ino.i_size)
			count = file->i_ino.i_size - file->i_offset;
		if ((i = count) <= 0)
			return (0);
		do {
			*buf++ = getc(fdesc+3);
		} while (--i);
		return (count);
	}
}

write(fdesc, buf, count)
	int fdesc, count;
	char *buf;
{
	register i;
	register struct iob *file;

	errno = 0;
	if (fdesc >= 0 && fdesc <= 2) {
		i = count;
		while (i--)
			putchar(*buf++);
		return (count);
	}
	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((file = &iob[fdesc])->i_flgs&F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	if ((file->i_flgs&F_WRITE) == 0) {
		errno = EBADF;
		return (-1);
	}
	file->i_cc = count;
	file->i_ma = buf;
	file->i_bn = file->i_boff + (file->i_offset / DEV_BSIZE);
	i = devwrite(file);
	file->i_offset += count;
	if (i < 0)
		errno = file->i_error;
	return (i);
}

ioctl(fdesc, command, param)
	int fdesc, command, param;
{
	register i;
	register struct iob *file;

	errno = 0;
	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((file = &iob[fdesc])->i_flgs&F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	i = devioctl(file,command,param);
	if (i < 0)
		errno = file->i_error;
	return (i);
}

char *
get_number( str, value_ptr )
	char * str ;
	int *  value_ptr ;
{
	register int	value ;
	register int	base ; 
	int		over_ride = 0 ;

	/* default is the value passed in */
	value = *value_ptr ;
	for ( ;; ) {
		switch ( *str ) {
		case '0' :
		case '1' :
		case '2' :
		case '3' :
		case '4' :
		case '5' :
		case '6' :
		case '7' :
		case '8' :
		case '9' :
			base = '0' ;
			break ;
		case 'a' :
		case 'b' :
		case 'c' :
		case 'd' :
		case 'e' :
		case 'f' :
			base = 'a' - 10 ;
			break ;
		case 'A' :
		case 'B' :
		case 'C' :
		case 'D' :
		case 'E' :
		case 'F' :
			base = 'A' - 10 ;
			break ;
		default  :
                        goto done ;
		}
		if ( !over_ride ) {
			over_ride = 1 ;
			value = 0 ;
		}
		value = (value * 16) + (*str - base) ;
		str ++ ;
	}
done:
	*value_ptr = value ;
	return str ;
}

int
check_scan( file )
	register struct iob * file;
{

	if (file -> i_bus < 0 || file -> i_bus > MAX_BUS) {
		printf("error: bus must be less than %d\n", MAX_BUS );
		return EUNIT ;
	}
	switch ( buses[ file -> i_bus ] ) {
	case RB_NO_BUS:
		printf("error: bus %x does not exist", file -> i_bus );
		return EUNIT ;
	case RB_VERSA_BUS:
		if ( file -> i_ctlr == DEFAULT_CTLR )
		        break ;
		if ( file -> i_ctlr < 0 || file -> i_ctlr > MAX_VERSA_CTLR ) {
			printf("error: controller must be less than %d\n",
				MAX_VERSA_CTLR );
			return  EUNIT ;
		}
		break ;
	case RB_VME_BUS:
		if ( file -> i_ctlr == DEFAULT_CTLR )
		        break ;
		if ( file -> i_ctlr < 0 || file -> i_ctlr > MAX_HVME_CTLR ) {
			printf("error: controller must be less than %d\n",
				MAX_HVME_CTLR );
			return  EUNIT ;
		}
		break ;
	}
	if (file -> i_drive < 0 || file -> i_drive > MAX_DRIVE) {
		printf("error: drive must be less than %d\n", MAX_DRIVE );
		return  EUNIT ;
	}
	if (file -> i_part < 0 || file -> i_part > MAX_PART) {
		printf("error: partition must be less than %d\n", MAX_PART );
		return EUNIT ;
	}

	return NO_ERROR ;
}

char *
scan( str, file )
	char * str;
	register struct iob * file;
{
	register	char *		cp;
	register	struct devsw *	dp;
	extern		int		version ;
	extern		int		boot_device ;

        SKIP_BLANKS( str ) ;
	for ( cp = str ; *cp && *cp != '(' && *cp != ' ' ; cp ++ ) ;
	if ( *cp != 0 ) {
		*cp ++ = '\0' ;
		SKIP_BLANKS( cp ) ;
	}
	if ( *cp == '(' ) {
		cp ++ ;
		SKIP_BLANKS( cp ) ;
	}
	/*
	 * The default device values are passed from the CP.
	 */
	file -> i_drive		= (boot_device)       & 0x0f ;
	if ( version < 2 ) {
		file -> i_ctlr	= (boot_device >>  4) & 0x0f ;
		file -> i_bus	= (boot_device >>  8) & 0x0f ;
	} else {
		file -> i_ctlr	= (boot_device >>  4) & 0x3f ;
		file -> i_bus	= (boot_device >> 10) & 0x03 ;
	}
	file -> i_dev_type	= (boot_device >> 12) & 0x0f ;
	file -> i_ino.i_dev	= (boot_device >> 12) & 0x0f ;
	file -> i_part		= (boot_device >> 20) & 0x0f ;

	/*
	 * Check if there is a specified device.
	 */
	for (dp = devsw; dp->dv_name; dp++) {
		if ( strcmp( str, dp -> dv_name ) == 0 ) {
			file -> i_dev_type  = dp - devsw ;
			file -> i_ino.i_dev = dp - devsw ;
			goto scan_more ;
		}
	}
	/*
	 * Default device is assumed, the string is the file name.
	 */
	return str ;

scan_more:
      /* Get the controller number */
	cp = get_number( cp, & file -> i_ctlr ) ;
	SKIP_BLANKS( cp ) ;
	if ( *cp == ',' ) {
		cp ++ ;
		SKIP_BLANKS( cp ) ;
	}

      /* Get the drive number */
	cp = get_number( cp, & file -> i_drive ) ;
	SKIP_BLANKS( cp ) ;
	if ( *cp == ',' ) {
		cp ++ ;
		SKIP_BLANKS( cp ) ;
	}

      /* Get the partition number */
	cp = get_number( cp, & file -> i_part ) ;
	SKIP_BLANKS( cp ) ;
	if ( *cp == ',' ) {
		cp ++ ;
		SKIP_BLANKS( cp ) ;
	}

      /* Get the bus number */
	cp = get_number( cp, & file -> i_bus ) ;
	SKIP_BLANKS( cp ) ;
	if ( *cp == ')' ) {
		cp ++ ;
		SKIP_BLANKS( cp ) ;
	}

	errno = check_scan( file ) ;
	if ( errno == NO_ERROR ) {
		return cp ;
	} else {
	 	file->i_flgs = 0;
		return (char *)0 ;
	}
}

int	openfirst = 1;

open( str, how )
	char * str;
	int how;
{
	register char *cp;
	register struct iob *file;
	int i ;
	int fdesc;

	if (openfirst) {
		for (i = 0; i < NFILES; i++)
			iob[i].i_flgs = 0;
		openfirst = 0;
	}

	for (fdesc = 0; fdesc < NFILES; fdesc++)
		if (iob[fdesc].i_flgs == 0)
			goto gotfile;
	_stop("No more file slots");
gotfile:
	(file = &iob[fdesc])->i_flgs |= F_ALLOC;

	cp = scan( str, file ) ;

        if ( !cp )
        	return (-1) ;

	file->i_flgs |= how+1;
 	if ( ! devopen(file) )
		return( -1 ) ;
	if (*cp == '\0') {
		file->i_cc = 0;
		file->i_offset = 0;
		return( fdesc+3 ) ;
	}
	file->i_flgs |= F_FILE;
	file->i_ma = (char *)(&file->i_fs);
	file->i_cc = SBSIZE;
	file->i_bn = SBLOCK + file->i_boff;
	file->i_offset = 0;
	if (devread(file) < 0) {
		errno = file->i_error;
		printf("super block read error\n");
		file->i_flgs = 0;
		return (-1);
	}
	if ((i = find(cp, file)) == 0) {
		file->i_flgs = 0;
		errno = ESRCH;
		return (-1);
	}
	if (how != 0) {
		printf("Can't write files yet.. Sorry\n");
		file->i_flgs = 0;
		errno = EIO;
		return (-1);
	}
	if (openi(i, file) < 0) {
		errno = file->i_error;
		file->i_flgs = 0;
		return (-1);
	}
	file->i_offset = 0;
	file->i_cc = 0;
	return (fdesc+3);
}

close(fdesc)
	int fdesc;
{
	struct iob *file;

	fdesc -= 3;
	if (fdesc < 0 || fdesc >= NFILES ||
	    ((file = &iob[fdesc])->i_flgs&F_ALLOC) == 0) {
		errno = EBADF;
		return (-1);
	}
	if ((file->i_flgs&F_FILE) == 0)
		devclose(file);
	file->i_flgs = 0;
	return (0);
}


exit()
{
	_stop("Exit called");
}

_stop(s)
	char *s;
{
	int i;

	for (i = 0; i < NFILES; i++)
		if (iob[i].i_flgs != 0)
			close(i);
	printf("%s\n", s);
	_rtt();
}

trap(ps)
	int ps;
{
	printf("Trap %o\n", ps);
	for (;;)
		;
}

uncache (addr)
	char *addr;
{
	/* Return *(addr-0x4000); DIRTY assumes this address is valid */
	mtpr (addr, PDCS);
}


/*
** get_yes_no is used to ask simple yes or no questions.  The question
** prompt is supplied by the caller,  The question mark, possible responses,
** and the default response is printed at the end of the prompt.  The routine
** then reads the answer and returns a 1 if a 'y' is typed or no response was
** given, otherwise, a zero is returned.
*/

int
get_yes_no(str)
	register char	*str;
{
	char	answer[ 80 ] ;
	char *  ptr ;

	for(;;) {
		printf("%s? [y/n] (n): ", str);
		gets( answer );
	      	for ( ptr = answer ; *ptr == ' ' ; ptr ++ ) ;
		if ( (*ptr == 'Y') || (*ptr == 'y') )
			return( 1 ) ;
		if ( (*ptr == 'N') || (*ptr == 'n') )
			return( 0 ) ;
		if ( *ptr == 0 )
			return( 0 ) ;
		printf("\nA 'Y' (yes) or 'N' (no) must be typed!\n\n");
	}
}

