/*
** Stand alone driver for the VDDC controller 
**	HCX Version, Oct 1983.
**
*/

#include "../uts/machine/ml/mtpr.h"
#define KERNEL
#include "../uts/machine/sys/param.h"
#undef  KERNEL
#include "../uts/machine/sys/buf.h"
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../uts/machine/sys/vbavar.h"
#include "../uts/machine/sys/dsk.h"
#define VDGENDATA       1
#include "../uts/machine/sys/vddc.h"
#undef  VDGENDATA
#include "../stand/saio.h"

#define NVD		4			/* Max number of controllers */
#define NDR             4                       /* Max drives per controller */

#define TRUE		1
#define	FALSE		0

/*
**	MDCB
*/

fmt_mdcb	mdcb;

/*
**	DCB
*/

fmt_dcb		dcb;

/*
** Unit specific information.
*/

struct {
	char	configured;
	fs_tab	info;
}unit_info[NVD][16];

/*
** Controller specific information.
*/

struct {
	unsigned char	ctlr_type;
	char		*ctlr_name;
	unsigned char	initialized;
	unsigned char	ctlr_started;
} ctlr_info[NVD];

static char		junk[1024];
static geometry_sector	gcb ; /* geometry control block (gcb) */


/*
**
*/

vdopen(io)
register struct iob	*io;
{
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;
	register int	i, j;

	/* Make sure controller number is in range */
	if(ctlr >= NVD) {
		printf("vddc: Controller number can't be greater than %x!\n",
		    NVD );
		return( false ) ;
	}
       	io -> i_ctlr_addr = vddcaddr[ ctlr ] & 0x00ffffff ;
        if ( io -> i_bus == PRIMARY )
        	io -> i_ctlr_addr |= PRIMARY_IOBASE ;
        else
        	io -> i_ctlr_addr |= SECONDARY_IOBASE ;

	/* Make sure drive number is in range */
	if(drive >= NDR) {
		printf("vddc: Drive number can't be greater than %x!\n",
		    NDR );
		return( false ) ;
	}
	/* check file system for validity */
	if((io->i_part < 0) || (io->i_part > 5)) {
		printf("vddc: File system #%d, should be less than #6.\n",
		    io->i_part);
		return( false ) ;
	}
	if(!ctlr_info[ctlr].initialized) {
		vdinit(io);	/* initialize controller/drive */
		ctlr_info[ctlr].initialized = TRUE;
		for(i=0; i<NVD; i++)
			for(j=0; j<16; j++)
				unit_info[i][j].configured = FALSE;
	}
	if(!unit_info[ ctlr ][ drive ].configured) {
		vdconfigure_drive(io);
		unit_info[ ctlr ][ drive ].configured = TRUE;
	}
	io->i_boff=unit_info[ctlr][drive].info.partition[io->i_part].start;

	return( true ) ;
}


/*
**
*/

vdinit(io)
register struct iob	*io;
{
	register int	ctlr       = io -> i_ctlr ;
	register int	drive      = io -> i_drive ;
	register cdr	*ctlr_addr = (cdr *)io -> i_ctlr_addr ;
	register char	*ctlr_type;

	/* Check to see if the controller is there. */
	if(badaddr(ctlr_addr, 2, 0)) {
		printf("vddc: Controller at %x is non-existant!\n", ctlr_addr);
		_stop("");
	}

	/* Probe further to find what kind of controller it is */
	ctlr_addr->cdr_reset = 0xffffffff;
	DELAY(1000000);
	/* Probe further to find what kind of controller it is */
	if(ctlr_addr->cdr_reset != 0xffffffff) {
		ctlr_info[ctlr].ctlr_type = SMDCTLR;
		ctlr_info[ctlr].ctlr_name = "SMD";
		DELAY(1000000);
	}
	else {
		ctlr_info[ctlr].ctlr_type = SMD_ECTLR;
		ctlr_info[ctlr].ctlr_name = "SMD/E";
		ctlr_addr->cdr_reserved = 0x0;
		DELAY(3000000);
	}
	if(ctlr_info[ctlr].ctlr_type == SMD_ECTLR) {
		ctlr_addr->cdr_csr =  0;
		ctlr_addr->mdcb_tcf = AM_ENPDA;
		ctlr_addr->dcb_tcf = AM_ENPDA;
		ctlr_addr->trail_tcf = AM_ENPDA;
		ctlr_addr->data_tcf = AM_ENPDA;
		ctlr_addr->cdr_ccf = CCF_STS | XMD_32BIT | BSZ_16WRD |
		    CCF_ENP | CCF_EPE /* | CCF_EDE */ | CCF_ECE | CCF_ERR;
	}
	if(vdaccess_with_no_trailer(io, INIT, 10) & HRDERR) {
		vdprint_error( io, "Initialization error",
		    dcb.operrsta,dcb.err_code);
		_stop("");
	}
	if(vdaccess_with_no_trailer(io, DIAG, 10) & HRDERR) {
		vdprint_error( io, "Diagnostic error",
		    dcb.operrsta, dcb.err_code);
		_stop("");
	}
}


/*
**
*/

vdconfigure_drive(io)
register struct iob	*io;
{
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;
	register fs_tab	*file_sys;
	dskadr		daddr;
	register int	i, j ;
	int             status = 0 ;

	for(i=0; i<nvddrv; i++) {
		unit_info[ ctlr ][ drive ].info = vdst[i];

	        /*
	        ** Clear old partition information to make room for geometry
	        ** blocks.
	        */
	        for (j = 0; j < GB_MAXPART; j ++) {
			unit_info [ctlr][drive].info.partition [j].start  = 0;
			unit_info [ctlr][drive].info.partition [j].length = 0;
		}

		if(ctlr_info[ ctlr ].ctlr_type == SMDCTLR)
			if(unit_info[ ctlr ][ drive ].info.nsec != 32)
				continue;
		vdconfigure(io, 0);
		daddr.cylinder = unit_info[ ctlr ][ drive ].info.ncyl - 2;
		daddr.track = unit_info[ ctlr ][ drive ].info.ntrak - 1;
		daddr.sector = unit_info[ ctlr ][ drive ].info.nsec - 1;
		io->i_ma = junk;
		io->i_drive_type = i ;  /* drive type */
		io->i_cc = unit_info[ ctlr ][ drive ].info.secsize;

	        /*
	        ** Validate the type of drive by cylinder seeks.
	        ** Next, read the geometry block and validate.
	        ** If both these conditions are successful, then exit with
	        ** the drive configured.
	        */
	        if (!(vdaccess( io, & daddr, RD ) & HRDERR))
 		        if (!((status = vdgeometry( io )) & HRDERR))
				return ;
			else
			        break ;
	}

	if (status != 0) {
	        printf("vddc: Invalid geometry block on ") ;
	        printf("controller %d, drive %d, drive type %s.\n",
	        	ctlr, drive, vdst[i].type_name ) ;
	}
	else
		printf("vddc: Unrecognizable drive; controller %d, drive %d!\n",
		       ctlr, drive );
	_stop("");
}

/*
** vdgeometry
**
** This routine is responsible for reading and validating the geometry block
** located at sector 0 on the disk.  If the geometry block is valid then
** the partition info structure is updated and successful return status.
*/

vdgeometry( io )
register struct iob *	io ;
{
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;
	dskadr		daddr;
	register int	i, j ;
	int             status ;
	int		checksum ;

	/*
	** Read the geometry block from sector 0 on the disk.  Note, this
	** only applies to vddc disk drives.  Setup information required
	** by vdaccess.
	*/
	daddr.cylinder = 0 ;
	daddr.track    = 0 ;
	daddr.sector   = 0 ;
	io->i_ma = (char *) (& gcb) ; /* geometry control block buffer */
	io->i_cc = sizeof( gcb ) ;

        /*
        ** Read the geometry block and validate.  If successful, then copy
        ** the partition parameters into the info structure.
        ** The data cache must be purged after successful read of the geometry
        ** block inorder to read the data from memory. You know!
        */
	status = vdaccess( io, & daddr, RD ) ;
	if (status & HRDERR) return( status ) ;

	mtpr(0,PADC);

  	GB_CHECKSUM( & gcb, checksum ) ;
	if (gcb . checksum != checksum) return( -1 ) ;

	/*
	** Update the drive info structure and return.
	*/
	for (i = 0 ; i < GB_MAXPART ; i ++) {
		unit_info[ ctlr ][ drive ] . info . partition [i] =
			gcb . geometry_block . partition [i] ;
	} ;

	return( 0 ) ;

}

/*
**
*/

vdstart_drive(io)
register struct iob	*io;
{
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;

	if(ctlr_info[ctlr].ctlr_started) {
		DELAY(5500000);
		return TRUE;
	}
	if(vdaccess_with_no_trailer(io, VDSTART, ((drive * 6) + 62)) & HRDERR) {
		vdprint_error( io, "Start error",
		    dcb.operrsta, dcb.err_code);
		_stop("");
	}
	ctlr_info[ ctlr ].ctlr_started = TRUE;
	DELAY((drive * 5500000) + 62000000);
	return TRUE;
}


/*
**  This routine actually configures a particular drive.
**
**  If the controller is an SMD/E controller then the number of sectors per
**  track is loaded into the appropriate register, otherwise it is left 
**  alone because the old SMD controller requires a constant 32 sectors
**  per track for it's drives. (an error would be returned if the value is
**  loaded.)
**
**  In the stand-alone spirit of things the system is halted if an error
**  occurs during this operation.
*/

vdconfigure(io, pass)
register struct iob	*io;
int	pass;
{
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;
	register cdr	*ctlr_addr = (cdr *)io -> i_ctlr_addr ;

	dcb.opcode = RSTCFG;		/* command */
	dcb.intflg = NOINT;
	dcb.nxtdcb = (fmt_dcb *)0;	/* end of chain */
	dcb.operrsta  = 0;
	dcb.devselect = (char)drive ;
	dcb.trail.rstrail.ncyl = unit_info[ ctlr ][ drive ].info.ncyl;
	dcb.trail.rstrail.nsurfaces = unit_info[ ctlr ][ drive ].info.ntrak;
	if(ctlr_info[ ctlr ].ctlr_type == SMD_ECTLR) {
		dcb.trailcnt = (char)4;
		dcb.trail.rstrail.nsectors = unit_info[ctlr][drive].info.nsec;
		dcb.trail.rstrail.slip_sec = unit_info[ctlr][drive].info.nslip;
	}
	else
		dcb.trailcnt = (char)2;
	mdcb.firstdcb = &dcb;
	mdcb.vddcstat = 0;
	VDDC_ATTENTION(ctlr_addr, &mdcb, ctlr_info[ctlr].ctlr_type);
	POLLTILLDONE(ctlr_addr,&dcb,5,ctlr_info[ctlr].ctlr_type);
	if(vdtimeout <= 0)
		_stop(" during drive configuration.\n");
	if(dcb.operrsta & (NOTCYLERR | DRVNRDY))
		if(!pass) {
			vdstart_drive(io);
			vdconfigure(io, 1);
		}
	if(dcb.operrsta & HRDERR) {
		vdprint_error( io, "Configuration error",
		    dcb.operrsta, dcb.err_code);
		_stop("");
	}
}


/*
**  Strategy is called to the actual I/O to the disk drives.
**
**  Some simple checks are made to make sure we don't do anything rediculus,
**  If everything is sane then the request is issued.
**
**  If no errors occured then the original byte count is returned,
**  otherwise -1 is returned to indicate an error occured.
*/

vdstrategy(io, func)
register struct iob	*io;
register int func;
{
	dskadr		daddr;
	register int	ctlr  = io -> i_ctlr ;
	register int	drive = io -> i_drive ;
	register fs_tab	*u_info = &unit_info[ctlr][drive].info;
	register int	op = (func == READ) ? RD : WD;
	register int	blk;
	int		doskipsec ; /* Boolean.  True if trying to write 
					geometry block; false otherwise */
	int		partstart =
		unit_info[ctlr][drive].info.partition[io->i_part].start ;
	int		partlen =
		unit_info[ctlr][drive].info.partition[io->i_part].length ;
	int		secsize =
		unit_info[ctlr][drive].info.secsize ;

	if(io->i_cc == 0)
		_stop("vddc: Can't transfer zero length records!\n");
	if(io->i_cc > 0xffff)
		_stop("vddc: Can't transfer greater than 2 to the 16th bytes!\n");


	if ( (io->i_bn < partstart) || (io->i_bn >= partstart+partlen) )
		return( -1 ) ;

	io->i_cc = MIN(io->i_cc, (partstart+partlen-io->i_bn) * DEV_BSIZE ) ;
	blk = io->i_bn * DEV_BSIZE / u_info->secsize;

	/*
	** Prevent fastcopy and other programs from overwriting 
	** the geometry block.  The upgrade and format programs
	** don't use this driver, so they ARE able to modify the
	** geometry block.
	*/
	if ( (op == WD) && (blk == 0) )
	{
		doskipsec = TRUE ;
		if ( io->i_cc > secsize )
		{
			io->i_cc -= secsize ;
			io->i_ma += secsize ;
			blk++ ;
		}
		else
			return( io->i_cc ) ;
	}
	else
		doskipsec = FALSE ;

	daddr.sector = blk % u_info->nsec;
	daddr.track = (blk / u_info->nsec) % u_info->ntrak;
	daddr.cylinder = (blk/u_info->nsec) / u_info->ntrak;
	if(vdaccess(io, &daddr, op) & HRDERR) {
		vdprint_error( io,"I/O error",dcb.operrsta,dcb.err_code);
		return(-1);
	}
	mtpr(0,PADC);
	return(io->i_cc + (doskipsec ? secsize : 0) );
}



/*
**
*/

vdprint_error( io, str, status, smde_status)
register struct iob	*io;
char		*str;
unsigned long	status;
unsigned long	smde_status;
{
	printf("vddc: ctlr = %d, drive = %d, bus = %d, addr = %x,%s\n",
	      io -> i_ctlr, io -> i_drive, io -> i_bus, io -> i_ctlr_addr, str);
	if(status & DRVNRDY)
		printf("Drive is not ready");
	else if(status & INVDADR)
		printf("Invalid disk address issued");
	else if(status & DNEMEM)
		printf("Non-existent memory error");
	else if(status & PARERR)
		printf("Main memory parity error");
	else if(status & OPABRT) 
		printf("Program aborted operation");
	else if(status & WPTERR)
		printf("Drive is write protected");
	else if(status & DSEEKERR)
		printf("Disk seek error");
	else if(status & UCDATERR)
		printf("Uncorrectable data error");
	else if(status & CTLRERR)
		printf("Controller faulted");
	else if(status & NOTCYLERR)
		printf("Not on cylinder error");
	else if(status & INVCMD)
		printf("Invalid command issued to controller");
	else 
		printf("Controller error");
	printf("!  Status = 0x%x", status);
	if(smde_status)
		printf("  Error code = %x", smde_status);
	printf("\n");
}


/*
**
*/

vdaccess_with_no_trailer(io, function, time)
register struct iob	*io;
register int		function, time;
{
	register int	ctlr = io -> i_ctlr ;
	register cdr	*ctlr_addr = (cdr *)io -> i_ctlr_addr ;

	dcb.opcode = function;		/* command */
	dcb.intflg = NOINT;
	dcb.nxtdcb = (fmt_dcb *)0;	/* end of chain */
	dcb.operrsta  = 0;
	dcb.devselect = (char)(io -> i_drive);
	dcb.trailcnt = (char)0;
	mdcb.firstdcb = &dcb;
	mdcb.vddcstat = 0;
	VDDC_ATTENTION(ctlr_addr, &mdcb, ctlr_info[ctlr].ctlr_type);
	POLLTILLDONE(ctlr_addr,&dcb,time,ctlr_info[ctlr].ctlr_type);
	if(vdtimeout <= 0)
		_stop(" during initialization operation.\n");
	return dcb.operrsta;
}


/*
**
*/

vdaccess(io, daddr, func)
register struct iob	*io;
dskadr			*daddr;
int			func;
{
	register int	ctlr = io -> i_ctlr ;
	register cdr	*ctlr_addr = (cdr *)io -> i_ctlr_addr ;

	dcb.opcode = (short)func;		/* format sector command */
	dcb.intflg = NOINT;
	dcb.nxtdcb = (fmt_dcb *)0;	/* end of chain */
	dcb.operrsta  = 0;
	dcb.devselect = (char)(io->i_drive);
	dcb.trailcnt = (char)(sizeof(trrw) / 4);
	dcb.trail.rwtrail.memadr = io->i_ma; 
	dcb.trail.rwtrail.wcount = ((io->i_cc + 1) / sizeof(short));
	dcb.trail.rwtrail.disk.cylinder = daddr->cylinder;
	dcb.trail.rwtrail.disk.track = daddr->track;
	dcb.trail.rwtrail.disk.sector = daddr->sector;
	mdcb.firstdcb = &dcb;
	mdcb.vddcstat = 0;
	VDDC_ATTENTION(ctlr_addr, &mdcb, ctlr_info[ctlr].ctlr_type);
	POLLTILLDONE(ctlr_addr, &dcb, 60, ctlr_info[ctlr].ctlr_type);
	if(vdtimeout <= 0)
		_stop(" during I/O operation.\n");
	return dcb.operrsta;
}

/*
**	Print_dcb() dumps the MDCB and DCB for diagnostic purposes.  This
** routine is called whenever a fatal error is encountered.
*/

vdprintdcb(ptr)
register long	*ptr;
{
	register long	i;
	register long	trailer_count;

	printf("Dump of MDCB: ");
	for(i=0; i<4; i++)
		printf("  %lx", *(ptr+i));
	if(ptr = (long *)*ptr) {
		printf(" and DCB:");
		trailer_count = *(ptr+3) & 0xff;
		for(i=0; i<7+trailer_count; i++) {
			uncache(ptr+i);
			printf("  %lx", *(ptr+i));
		}
	}
	printf("\n");
	for(i=0; i<5000000; i++) ;
}

