
/*
** vdupdate:	update disks on a VDDC/SMD_E controller for geometry block
**		support
**
** Author:	L. Quik
**
**	This program is used to update disks on the VDDC/SMD_E controller to
** support geometry blocks.  This conversion involves locating all such disks
** on the system and writing a geomtry block for that particular drive out to
** sector zero.  No other modification is required.  A disk updated to
** geometry blocks will still run under any previous operating system earlier
** than 2.2.  However, a disk not updated for geometry blocks will not be
** configured under a 2.2 or later operating system.  This program exist as
** a standalone and online version.  The standalone version facilitates ease
** of upgrade to the 2.2 release.  This should be run as the first part of
** the gen process prior to copying the os onto disk.
*/

#include "../uts/machine/ml/mtpr.h"
#define	KERNEL
#include "../uts/machine/sys/param.h"
#undef	KERNEL
#include "../uts/machine/sys/buf.h"
#include "../uts/machine/sys/time.h"
#include "../uts/machine/sys/vnode.h"
#include "../uts/machine/ufs/inode.h"
#include "../uts/machine/ufs/fs.h"
#include "../uts/machine/sys/vbavar.h"
#include "../uts/machine/sys/dsk.h"
#include "../uts/machine/sys/vddc.h"
#include "../uts/machine/sys/reboot.h"
#include "../stand/saio.h"
#include "../stand/vdupdate.h"

int	boot_device ;
int	version ;
int     buses[ 2 ] ;


/*
**	Main, of course, is the entry point into vdupdate.  It is responsible
** for identifying all VDDC/SMD_E disks on the system and reporting these to
** the user.  This is the main interface for interacting with the user on
** disk conversion.  Once the information is input from the user, conversion
** of all selected drives proceeds.
*/

main()
{
	register int	r12 ;
	register int	r11 ;
	register int	r10 ;
	register int	controller, drive ;     /* controller/drive to access */
	register int    max_drive ;             /* max drive on controller */
	register int    status ;                /* operation status */

	boot_device = r10 ;
	version = (boot_device & 0xf000000) >> 24 ;
	buses[ 0 ] = RB_PRIMARY_IO( r10 ) ;
	if ( buses[ 0 ] == RB_NO_BUS )
		buses[ 0 ] = RB_VERSA_BUS ;
	buses[ 1 ] = RB_SECONDARY_IO( r10 ) ;

	printf("\nVDUPDATE  Version 1.0\n\n") ;


process_loop:

	status = 0 ;

	/*
	** Initialize global data structures.
	*/
   	for (controller = 0; controller < MAXCONTROLLER; controller++) {

           	ctlr_type [controller] = UNKNOWN ;
              	ctlr_name [controller] = "UNKNOWN" ;
                ctlr_addr [controller] = 0 ;

		for (drive = 0; drive < MAXDRIVE; drive++) {
			drive_type [controller][drive] = UNKNOWN ;
			drive_name [controller][drive] = "UNKNOWN" ;
			drive_list [controller][drive] = FALSE ;
		}
	}

	/*
	** Identify which controllers are present and what type they are.
	** Also, determine which drives are present and what type they are.
	** This must all be performed within this module and not the driver
	** due to the simple nature of the driver.  If a drive does not exist,
	** the driver halts.
	*/
	printf( "<::: Configure the Disk Subsystem :::>\n\n" ) ;
	for (controller = 0; controller < MAXCONTROLLER; controller++) {

		if (! probe_controller( controller, RB_SECONDARY_IO( r10 ) ) )
			continue ;

		printf( "Controller %d is an %s disk controller.\n",
		        controller, ctlr_name [controller]);

	        /*
	        ** Next, scan for drives on this controller.  Determine
	        ** which drives exist and the drive types.
	        */
		max_drive = (ctlr_type [controller] == SMDCTLR) ? 4 : 4 ;

		for (drive = 0; drive < max_drive; drive ++) {

			if (! probe_drive( controller, drive )) continue ;

			printf( "        Drive %d : %s\n",
			        drive,
				drive_name [controller][drive] );

	        } ;

	} ;

	/*
	** Fire wall.  Check to insure that at least one controller and drive
	** were found on the system.  I saw a case once where this didn't
	** happen.
	*/
	max_drive = 0 ;
   	for (controller = 0; controller < MAXCONTROLLER; controller++) {
		for (drive = 0; drive < MAXDRIVE; drive++) {
			if (drive_type [controller][drive] != UNKNOWN) {
				max_drive ++ ;
			}
		}
	}
	if (max_drive == 0) {
		printf( "\n\n" ) ;
		printf( "VDUPDATE failed to find any controllers or drives " ) ;
		printf( "on this system.\n" ) ;
		if (get_yes_no("Do you want to restart VDUPDATE")) {
		        printf( "\n" ) ;
		        goto process_loop ;
		}
		printf( "\nVDDC Disk Drive Update Complete.\n" ) ;
		asm( "halt" ) ;
	}

	printf( "\n\n" ) ;

	/*
	** Query the user on which of these drives he would like us to
	** build a geometry block.  Options include all, none, or some.
	** From this, a list is built consisting of controller number/drive
	** number pairs.
	*/

	if (!get_yes_no("Do you want to continue with VDDC disk update")) {
		if (get_yes_no("Do you want to restart VDUPDATE")) {
		        printf( "\n" ) ;
		        goto process_loop ;
		}
		printf( "\nVDDC Disk Drive Update Complete.\n" ) ;
		asm( "halt" ) ;
	}

	if (!get_yes_no("Do you want to update all VDDC drives located")) {
                get_disk_parameters() ;
        }
        else
        {
   		for (controller = 0; controller < MAXCONTROLLER; controller++) {
			for (drive = 0; drive < MAXDRIVE; drive++) {
				if (drive_type [controller][drive] != UNKNOWN) {
					drive_list [controller][drive] = TRUE ;
				}
			}
		}
	}
	printf( "\n\n" ) ;


 	/*
	** Process the drive_type table writing a geometry block on
	** each drive.  Note, VDDC geometry blocks only exist at
	** sector zero of the drive.  This location is hard coded in
	** the driver routines to prevent writing to any other locations
	** on disk.
	**
	** The geometry block written is a default block of the drive
	** type.  The partition information is taken from the VDST
	** structure for VDDC drives.
	*/

	printf( "<::: Updating the following disk drives :::>\n\n" ) ;
   	for (controller = 0; controller < MAXCONTROLLER; controller++) {
		for (drive = 0; drive < MAXDRIVE; drive++) {
			if (drive_list [controller][drive]) {
				printf( "   controller %d", controller ) ;
				printf( "  drive %d", drive ) ;
				printf( "   [%s/%s]\n", ctlr_name [controller],
				        drive_name [controller][drive] ) ;
				status += write_geometry_block( controller,
								drive ) ;
			}
		}
	}

	if (status != 0) {
	        printf( "\nErrors occurred during drive update.\n\n" ) ;
		if (get_yes_no("Do you want to restart VDUPDATE"))
		        goto process_loop ;
	}

	printf( "\nVDDC Disk Drive Update Complete.\n" ) ;
	asm( "halt" ) ;

}



/*
** access:
**
** This routine is used by other routines to do reads and writes to the disk.
** The status of the read / write is returned to the caller for processing.
*/

int
access( controller, drive, function, daddr_ptr, buf_ptr, byte_count )

	int     controller, drive ;     /* controller/drive number to access */
	int     function ;              /* disk drive function to execute */
	dskadr * daddr_ptr ;            /* sector address on disk */
	char *  buf_ptr ;               /* I/O buffer address */
	int     byte_count ;            /* I/O byte count */

{

	/*
	** Validate disk address.  Writes are only allowed to sector 0 of
	** the disk and no larger than 512 bytes.
	*/
	if ((function == WD) && ((daddr_ptr -> cylinder != 0) ||
	                         (daddr_ptr -> track != 0) ||
	                         (daddr_ptr -> sector != 0) ||
	                         (byte_count > 512))) {
		return( -1 ) ;
	}

	/*
	** Setup the mdcb and dcb for the I/O transfer.
	*/
	mdcb . firstdcb = & dcb ;
	mdcb . vddcstat = 0 ;

	dcb . opcode = function ;	/* I/O function command */
	dcb . intflg = NOINT;
	dcb . nxtdcb = (fmt_dcb *)0;	/* end of chain */

	dcb . operrsta  = 0;
	dcb . devselect = (char) drive ;
	dcb . trailcnt  = (char) (sizeof( trrw ) / 4) ;

	dcb . trail . rwtrail . memadr = buf_ptr ;
	dcb . trail . rwtrail . wcount = (byte_count / sizeof( short )) ;

	dcb . trail . rwtrail . disk . cylinder = daddr_ptr -> cylinder ;
	dcb . trail . rwtrail . disk . track    = daddr_ptr -> track ;
	dcb . trail . rwtrail . disk . sector   = daddr_ptr -> sector ;

	/*
	** Initiate the I/O operation and wait for completion.  Return
	** I/O status to caller.
	*/
	VDDC_ATTENTION( ctlr_addr [controller],
			& mdcb,
			ctlr_type [controller] ) ;
	POLLTILLDONE( ctlr_addr [controller],
	 	      & dcb, 2*60,
	 	      ctlr_type [controller] ) ;
	if (vdtimeout <= 0) {
		printf(" in access.\n");
		return( -1 ) ;
	}

	return( dcb.operrsta ) ;

}


/*
** access_with_no_trailer:
**
** This routine is used to perform controller functions which
** require no data movement.
*/

access_with_no_trailer( controller, drive, function, wait_time )

	int     controller, drive ;
	int	function, wait_time ;

{
	/*
	** Initialize the MDCB and DCB for the non-data movement function.
	*/
	mdcb . firstdcb = & dcb ;
	mdcb . vddcstat = 0 ;

	dcb . opcode    = function ;		/* command */
	dcb . intflg    = NOINT ;
	dcb . nxtdcb    = (fmt_dcb *) 0 ;	/* end of chain */
	dcb . operrsta  = 0 ;
	dcb . devselect = (function == VDSTART) ? 0 : (char) drive ;
	dcb . trailcnt  = (char) 0 ;

	/*
	** Initiate the function and poll.  Status is returned.
	*/
	VDDC_ATTENTION( ctlr_addr[controller], & mdcb, ctlr_type[controller] ) ;
	POLLTILLDONE( ctlr_addr[controller], & dcb,
		      wait_time, ctlr_type[controller] ) ;

	if (vdtimeout <= 0) {
		printf(" during startup operation.\n");
		return( -1 ) ;
	}

	return( dcb . operrsta ) ;

}


/*
** configure_drive:
**
** This routinere_drive tells the controller what kind of drive is at
** on a particular line.  A boolean value is returned indicating the
** success or failure of the drive configuration operation.
*/

int
configure_drive( controller, drive, type, pass )

	int	controller, drive ;     /* controller/drive to configure */
	int     type ;                  /* drive type to configure */
	int	pass ;                  /* configuration pass */

{

	int     status ;                /* I/O status */

	/*
	** Setup the mdcb and dcb for defining the drive attached to the
	** controller a the specified location.
	*/
	mdcb . firstdcb = &dcb;
	mdcb . vddcstat = 0;

	dcb . opcode = RSTCFG;		/* command */
	dcb . intflg = NOINT;
	dcb . nxtdcb = (fmt_dcb *)0;	/* end of chain */

	dcb . operrsta  = 0;
	dcb . devselect = (char) drive ;

	dcb . trail . rstrail . ncyl      = vdst [type] . ncyl ;
	dcb . trail . rstrail . nsurfaces = vdst [type] . ntrak ;

	if (ctlr_type [controller] == SMDCTLR)
		dcb . trailcnt = (char) 2 ;
	else {
		dcb . trailcnt = (char) 4 ;
		dcb . trail . rstrail . nsectors = vdst [type] . nsec ;
		dcb . trail . rstrail . slip_sec = vdst [type] . nslip ;
	}

	/*
	** Initiate the operation.  A retry operation is attempted if the
	** first pass fails.
	*/
	VDDC_ATTENTION( ctlr_addr [controller], & mdcb,
			ctlr_type [controller] ) ;
	POLLTILLDONE( ctlr_addr [controller], & dcb,
		      5, ctlr_type [controller] ) ;

	if (vdtimeout <= 0) {
		printf(" during drive configuration.\n");
		return( FALSE ) ;
	}

	if (dcb.operrsta & HRDERR) {
	        if (! (dcb.operrsta & (NOTCYLERR | DRVNRDY))) {
	                printf("\nVDDC: Configuration error.\n") ;
	                return( FALSE ) ;
	        }
		if (pass) {
			return( FALSE ) ;
		}

		status = configure_drive( controller, drive, type, 1 ) ;
		return( status ) ;
	}

	return( TRUE ) ;

}


/*
** Get_controller_number reads, validates and returns a controller number.
*/

get_controller_number()
{
	char	controller [20];
	int	controller_number;
	int	i, controller_printed = FALSE;

	/*
	** The following loop reads a controller number from the user and
	** validates.
	*/
	for (;;) {

	        /*
	        ** Prompt for controller number.
	        */
		printf( "\nController number? [" ) ;
		for (i = 0; i < MAXCONTROLLER; i++)
			if (ctlr_type [i] != UNKNOWN) {
				if (controller_printed) printf(", ");
				controller_printed = TRUE;
				printf("%d", i);
			}
		printf( "]: " );
		gets( controller ) ;

		/*
		** Extract controller number and validate.
		*/
		if (!controller[0]) return( -1 ) ;

		if (!sscanf( controller, "%d", & controller_number )) {
			printf( "\nResponse must start with a digit!\n" );
			continue;
		}

		if ((controller_number >= 0) && (controller_number <= 3))
			if (ctlr_type [controller_number] != UNKNOWN)
				break;

		printf( "\nController number %d is illegal!\n",
		        controller_number);
	}
	return( controller_number ) ;
}


/*
** This routine ask the user the controller number and unit number of the drive
** to be updated for geometry blocks.  When all information is gathered and
** correct, then the drive is initialized for writing geometry blocks, what
** ever that means.
*/

get_disk_parameters()
{
	register int	at_least_one_drive_specified = 0;
	register int	controller, drive ;

	/*
	** The following loop queries the user for controller and drive
	** numbers until a return is hit in response to the controller
	** number question.
	*/

	printf("\nEnter controller number and drive number for each disk.\n");
	printf("Key <RETURN> to controller number prompt when through.\n");

	for (;;) {

		controller = get_controller_number() ;

		if (controller != -1) {

			drive = get_drive_number( controller ) ;

			/*
			** Verify that the information is correct via the user.
			*/
			printf( "Is controller %d, unit %d ok",
			        controller, drive);

			if (!get_yes_no("")) {
				printf("\nEntry ignored.\n");
			}
			else
			{
				drive_list [controller][drive] = TRUE ;
				at_least_one_drive_specified = 1;
			}

			continue;
		}

		if (at_least_one_drive_specified) break;
		printf( "You must specify at least one controller number!\n" );

	}
}


/*
** Get_drive_number reads, validates and returns a drive number.
*/

get_drive_number( controller )
int	controller;
{
	char	drive_string [20];
	int	drive, max_drive ;

	/*
	** The following loop queries the user for a drive number associated
	** with the controller number just entered.
	*/
	for (;;) {

	        /*
	        ** Prompt the user for drive number.
	        */
		max_drive = (ctlr_type [controller] == SMDCTLR) ? 3 : 15;
		printf( "Unit number? [0-%d]: ", max_drive ) ;
		gets( drive_string );

		if (!drive_string[0]) {
			printf( "\nA drive number must be specified!\n\n" );
			continue;
		}

		if (!sscanf(drive_string, "%d", &drive)) {
			printf( "\nResponse must start with a digit!\n\n" ) ;
			continue;
		}

	        /*
	        ** Validate the drive number.
	        */
		if ((drive >= 0) && (drive <= max_drive))
		        if (drive_type [controller][drive] != UNKNOWN)
				break;

		printf( "\nDrive number %d is illegal!\n\n", drive );
	}
	return( drive ) ;
}


/*
** printdcb:
**
** This routine dumps the MDCB and DCB for diagnostic purposes.  This
** routine is called whenever a fatal error is encountered.
*/

printdcb( ptr )
	register long *		ptr ;
{
	register long	i;
	register long	trailer_count;

	printf("Dump of MDCB: ");
	for (i=0; i<4; i++)
		printf("  %lx", *(ptr+i));
	if (ptr = (long *)*ptr) {
		printf(" and DCB:");
		trailer_count = *(ptr+3) & 0xff;
		for (i=0; i<7+trailer_count; i++) {
			uncache(ptr+i);
			printf("  %lx", *(ptr+i));
		}
	}
	printf("\n");
	for (i=0; i<5000000; i++) ;
}


/*
** probe_controller:
**
** This routine is responsible for probing a disk controller and validating
** its existence.  If a controller is found, then global data structures
** are updated to reflect the address, type, and name of the specific
** controller.  This function returns a boolean based upon the existence
** of the controller.
*/

int
probe_controller( controller, secondary )

	int     controller ;    /* controller number to probe */
	int	secondary ;

{

	cdr *   ctlr_address ;  /* valid controller address */
	cdr *   ctlr_addr1 ;    /* address on primary bus */
	cdr *   ctlr_addr2 ;    /* address on secondary bus */

	/*
	** Determine if a controller exist on either bus.
	*/
	ctlr_addr1 = (cdr *)(vddcaddr [controller] | PRIMARY_IOBASE) ;
  	ctlr_addr2 = (cdr *)(vddcaddr [controller] | SECONDARY_IOBASE) ;

	if (! badaddr( ctlr_addr1, 2, 0 ))
		ctlr_address = ctlr_addr1 ;
  	else if ( secondary && ! badaddr( ctlr_addr2, 2, 0 ))
		ctlr_address = ctlr_addr2 ;
  	else
	        return( FALSE ) ;

        /*
        ** Determine controller type and intialize structures.
        */
	ctlr_addr [controller] = ctlr_address ;
	ctlr_address->cdr_reset = 0xffffffff;
	DELAY(1000000);

	if (ctlr_address->cdr_reset != 0xffffffff) {
		ctlr_type [controller] = SMDCTLR;
		ctlr_name [controller] = "SMD";
		DELAY(1000000);
	}
	else
	{
		ctlr_type [controller] = SMD_ECTLR;
		ctlr_name [controller] = "SMD/E";
		ctlr_address->cdr_reserved = 0x0;
		DELAY(3000000);
	}

	/*
	** Reset the controller and initialize.
	*/
	VDDC_RESET( ctlr_addr [controller], ctlr_type [controller] ) ;

	if (ctlr_type [controller] == SMD_ECTLR) {
		(ctlr_addr [controller]) -> cdr_csr   =  0;
		(ctlr_addr [controller]) -> mdcb_tcf  =  AM_ENPDA;
		(ctlr_addr [controller]) -> dcb_tcf   =  AM_ENPDA;
		(ctlr_addr [controller]) -> trail_tcf =  AM_ENPDA;
		(ctlr_addr [controller]) -> data_tcf  =  AM_ENPDA;
		(ctlr_addr [controller]) -> cdr_ccf   =  CCF_STS   | XMD_32BIT |
							 BSZ_16WRD | CCF_ERR   |
					                 CCF_ENP   | CCF_EPE   |
			                                 CCF_EDE   | CCF_ECE ;
	}

	access_with_no_trailer( controller, 0, INIT, 10 ) ;
	access_with_no_trailer( controller, 0, DIAG, 20 ) ;

        return( TRUE ) ;

}


/*
** probe_drive:
**
** This routine is reponsible for probing for disk drives on a specified
** controller.  The method of probing for a drive involves cylinder
** seeks.  If a drive is located, the type and name of the drive are updated
** within the global data structures.
**
** This function returns a boolean value based upon the existence of
** a drive.
*/

int
probe_drive( controller, drive )

	int     controller, drive ;     /* controller & drive to probe */

{

	int     i ;     	/* local loop control variable */
	dskadr  daddr ; 	/* sector address on disk */
	int     status ;        /* I/O status from disk access */

	/*
	** Cylinder Seeks:  The following loop performs a set of cylinder
	** seek operations on the disk drive to determine the type of
	** drive.
	*/
	for (i = 0; i < nvddrv; i ++) {

		/*
		** Validate that the drive controller combination makes
		** sense.  If not, try next drive type.
		*/
		if ((ctlr_type [controller] == SMDCTLR) &&
	            (vdst [i] . nsec != 32))
			continue;

 	        /*
 	        ** Configure the drive for this type.
 	        */
		if (! configure_drive( controller, drive, i, 0 )) break ;

		/*
		** Access the last sector on the disk given the current drive
		** type being evaluated.  If successful, then this is the drive
		** type returned.
		*/
		daddr . cylinder = (vdst [i] . ncyl)  - 2;
		daddr . track    = (vdst [i] . ntrak) - 1;
		daddr . sector   = (vdst [i] . nsec)  - 1;

		status = access( controller, drive,
				 RD, & daddr,
		                 scratch, 512 ) ;

	        /*
	        ** If the access was successful, initialize global structures
	        ** related to this drive and return status.
	        */
		if (!(status & HRDERR)) {
		        drive_type [controller][drive] = i ;
		        drive_name [controller][drive] = vdst [i] . type_name ;
			return( TRUE ) ;
		}
	}

	return( FALSE ) ;
}


/*
** write_geometry_block:
**
** This routine is reponsible for writing a default geometry block to the
** specified disk drive.  The default geometry block is based on the drive
** type such that the partitioning matches the current configuration of
** the drive.
**
** This function returns the status of the write and read/verify operation.
*/

int
write_geometry_block( controller, drive )

	int     controller, drive ;     /* controller & drive to probe */

{

	int                     checksum ;      /* gb checksum value */
	dskadr                  daddr ;         /* I/O disk address */
	geometry_block *	gb_ptr ;        /* geometry block pointer */
	geometry_sector *	gs_ptr ;        /* geometry sector pointer */
	int                     status ;        /* disk I/O status */

	/*
	** Initialize local pointer variables.
	*/
	gs_ptr = & geometry_sector_buffer ;
	gb_ptr = & (geometry_sector_buffer . geometry_block) ;

	/*
	** Generate the default geometry block for the specific disk drive
	** and checksum the sector buffer.
	*/
	GB_DEFAULT( gb_ptr, drive_type [controller][drive] ) ;
	GB_CHECKSUM( gs_ptr, gs_ptr -> checksum ) ;

	/*
	** Write the geometry block to sector 0 of the disk.  If this is
	** successful, the block is read and validated from the drive.
	*/
	daddr . cylinder = 0 ;
	daddr . track    = 0 ;
	daddr . sector   = 0 ;

	status = access( controller, drive,
			 WD, & daddr,
		         gs_ptr, sizeof( geometry_sector ) ) ;

	if (status & HRDERR) {
	        printf( "Error:   controller %d drive %d.",
	                controller, drive ) ;
	        printf( "  Failed to write geometry block.\n" ) ;
		return( status ) ;
	}

	/*
	** Read geometry block just written and validate.
	*/
	gs_ptr = (geometry_sector *) scratch ;
	status = access( controller, drive,
	                 RD, & daddr,
	                 gs_ptr, sizeof( geometry_sector ) ) ;

	if (status & HRDERR) {
	        printf( "Error:   controller %d drive %d.",
	                controller, drive ) ;
	        printf( "  Failed to write/read geometry block.\n" ) ;
		return( status ) ;
	}


	GB_CHECKSUM( gs_ptr, checksum ) ;
	if (gs_ptr -> checksum != checksum) {
	        printf( "Error:   controller %d drive %d.",
	                controller, drive ) ;
	        printf( "  Failed to write/read geometry block.\n" ) ;
		return( -1 ) ;
	}

	return( 0 ) ;

}
