#include 	"../fpe/fp.h"
#include 	"../fpe/fp_in_krnl.h"

 #
 # Multiply two doubles
 #
 # double
 # Kmuld(acc_most, acc_least, op_most, op_least, hfs)
 # long acc_most, acc_least, op_most, op_least;
 # long *hfs;
 #

/* temporary result is a 4 long word */
#define	tmp0		r0		/* must be an even register */
#define	tmp1		r1		/* must follow tmp0 */
#define	tmp2		r2		/* must be an even register */
#define	tmp3		r3		/* must follow tmp1 */
#define	svtmp3		r11

#define	mpA		r4		/* most significant part of operand A */
#define	lpA		r5		/* least significant part of operand A*/
#define	mpB		r6		/* most significant part of operand B */
#define	lpB		r7		/* least significant part of operand B*/

#define	expA		r8		/* exponent of operand A */
#define	expB		r9		/* exponent of operand B */
#define	exp		r8		/* exponent of multiplication result */

#define	i		r10		/* scratch */

#define	savehigh	r12		/* see below for values */

	.set	HIGH0, 0		# operand A had negative bit
	.set	HIGH1, 1		# operand B had negative bit

	.set	BIT31, 0x80000000
	.set	SAVEHIGH0, (1<HIGH0)
	.set	SAVEHIGH1, (1<HIGH1)

/*
 * Since 'emul' is a signed multiplication, we will have to 
 * remove the negative bit before 'emul'ing and compensate for that
 * after the multiplication.
 * 
 * Suppose a and b both have negative bits, then:
 *
 *	a * b = (pos_a  + 0x80000000) * (pos_b + 0x80000000)
 *
 *	      = pos_a * pos_b +      pos_a * 0x80000000 +
 *		pos_b * 0x80000000 + 0x80000000 * 0x80000000
 *
 *      pos_a and pos_b are a and b stripped off the negative bit.
 *
 * We save (pos_a * 0x80000000) and (pos_b * 0x80000000) on stack for
 * frequent reuse.
 */

#define	l_lpAs	-56(fp)		/* (h_lpAs,l_lpAs) saves the value of     */
#define	h_lpAs	-60(fp)		/*     (pos_lpA * 0x80000000) 		  */

#define	l_lpBs	-64(fp)		/* (h_lpBs,l_lpBs) saves the value of     */
#define	h_lpBs	-68(fp)		/*     (pos_lpB * 0x80000000) 		  */

#define	l_mpAs	-72(fp)		/* (h_mpAs,l_mpAs) saves the value of     */
#define	h_mpAs	-76(fp)		/*     (mpA * 0x80000000) 		  */

#define	l_mpBs	-80(fp)		/* (h_mpBs,l_mpBs) saves the value of     */
#define	h_mpBs	-84(fp)		/*     (mpB * 0x80000000) 		  */

/*
 * SHLQUADS	shift left 2 quads.
 */

#define	SHLQUADS(L0,L1,L2,SHIFTSZ)					\
	shlq	$SHIFTSZ,L0,L0;		/* (L0,L1) << SHIFTSZ */	\
	shrl	$(32-SHIFTSZ),L2,i;	/* get leftmost SHIFTSZ bits */	\
	orl2	i,L1;			/*   and shift them to L1 */	\
	shlq	$SHIFTSZ,L2,L2;		/* (L0,L1,L2,L3) << SHIFTSZ */

	.text
	.globl	_Kmuld	    

_Kmuld:				
	.word	0x1ffc			# we use all registers!

	#
	# adjust stack for local variables
	#
	subl3	$88,fp,sp

	#
	# get operands and co.
	#
	movl	4(fp),mpA		# get most significant part of A
	andl2	$~SIGNBIT,mpA		# turn off the sign bit.
	shrl	$EXPSHIFT,mpA,expA	# get the exponent

	movl	12(fp),mpB		# get most significant part of B
	andl2	$~SIGNBIT,mpB		# turn off the sign bit.
	shrl	$EXPSHIFT,mpB,expB	# get the exponent

	movl	8(fp),lpA		# get least significant part of A
	movl	16(fp),lpB		# get least significant part of B

	orl3	mpA,lpA,i		# A is zero?
	jeql	retzero
	orl3	mpB,lpB,i		# B is zero?
	jeql	retzero

	andl2	$~EXPMASK,mpA		# turn off the exponent.
	orl2	$~CLEARHID,mpA		# turn on the hidden bit

	andl2	$~EXPMASK,mpB		# turn off the exponent.
	orl2	$~CLEARHID,mpB		# turn on the hidden bit
	
 #
 # 	Now perform the multiplication   
 #
mul:
	addl3	expA,expB,exp
	subl2	$BIASP1,exp		# exp is the resultant exponent

	jleq	underflow		# exp is too small
	cmpl	exp,$256
	jgeq	overflow 		# exp is too large

	# We have to turn off the high order bit since
 	# the 'emul' instruction does not do unsigned multiplication (sigh!!).
	#
	clrl 	tmp0
	clrl	tmp1
	clrl	tmp2
	clrl	tmp3

	#
	# Calculate compensation for stripped-off negative bit.
	#
	clrl	savehigh		# clear flag of saving negative bits.

	bitl	$SIGNBIT,lpA		# if (lpA has a negative bit)
	jeql	.c0
	andl2	$~BIT31,lpA		#   turn off negative bit before 'emul'
	orl2	$SAVEHIGH0,savehigh	#   remember we had a negative bit

	clrl	h_lpBs			#   (h_lpBs,l_lpBs)= pos_lpB*0x80000000
	andl3	$~BIT31,lpB,l_lpBs
	shlq	$31,h_lpBs,h_lpBs

	clrl	h_mpBs			#   (h_mpBs,l_mpBs) = mpB*0x80000000
	andl3	$~BIT31,mpB,l_mpBs
	shlq	$31,h_mpBs,h_mpBs	#   mpB never has sign bit!
.c0:					# endif

	bitl	$SIGNBIT,lpB		# if (lpB has a negative bit)
	jeql	.c1
	andl2	$~BIT31,lpB		#   turn off negative bit before 'emul'
	orl2	$SAVEHIGH1,savehigh	#   remember we had a negative bit

	clrl	h_lpAs			#   (h_lpAs,l_lpAs)= pos_lpA*0x80000000
	andl3	$~BIT31,lpA,l_lpAs
	shlq	$31,h_lpAs,h_lpAs

	clrl	h_mpAs			#   (h_mpAs,l_mpAs)= pos_mpA*0x80000000
	andl3	$~BIT31,mpA,l_mpAs
	shlq	$31,h_mpAs,h_mpAs	#   mpA never has sign bit!
.c1:					# endif

	#
	# (tmp2,tmp3) == lpB * lpA
	#
	tstl	8(fp)			# lpA == 0?
	jeql	.m1
	tstl	16(fp)			# lpB == 0?
	jeql	.m1

	emul	lpB,lpA,$0,tmp2		# (tmp2,tmp3) = pos_lpB * pos_lpA
	casel	savehigh,$0,$3		# compensate for negative bits?
	.align	2
.L1:	.word	.m1-.L1			# no compensation
	.word	.c_lpA-.L1		# compensate for neg bit in lpA
	.word	.c_lpB-.L1		# compensate for neg bit in lpB
	.word	.c_lpA1-.L1		# compensate for neg bit in lpA/lpB
	halt				# should never happen
.c_lpA:	addl2	l_lpBs,tmp3		# (tmp2,tmp3) += 0x80000000*lpB
	adwc	h_lpBs,tmp2
	jmp	.m1
.c_lpB:	addl2	l_lpAs,tmp3		# (tmp2,tmp3) += 0x80000000*lpA
	adwc	h_lpAs,tmp2
	jmp	.m1
.c_lpA1: addl2	l_lpBs,tmp3		# (tmp2,tmp3) += 0x80000000*lpB
	adwc	h_lpBs,tmp2
	addl2	l_lpAs,tmp3		# (tmp2,tmp3) += 0x80000000*lpA
	adwc	h_lpAs,tmp2
	addl2	$0x40000000,tmp2	# (tmp2,tmp3) += 0x80000000*0x80000000
.m1:

	#
	# (tmp2,tmp3) == lpB * (mpA,lpA); mpA is non-zero.
	#
	movl	tmp3,svtmp3		# save tmp3
	tstl	16(fp)			# lpB == 0?
	jeql	.m2
					# due to sign extension, we cannot use 
					#	emul lpB,mpA,tmp2,tmp2 
	emul	lpB,mpA,$0,tmp0		# (tmp0,tmp1) = pos_lpB * mpA
	bbc	$HIGH1,savehigh,.m2	# compensate for the negative bit?
	addl2	l_mpAs,tmp1		# (tmp0,tmp1) += 0x80000000*mpA
	adwc	h_mpAs,tmp0
.m2:

	addl3	tmp1,tmp2,tmp3
	adwc	$0,tmp0
	movl	tmp0,tmp2		# (tmp2,tmp3) = pos_lpB * (mpA,lpA)

	#
	# (tmp0,tmp1) == mpB * lpA; mpB is non-zero
	#
	tstl	8(fp)			# lpA == 0?
	jneq	.m4
	clrl	tmp0
	clrl	tmp1
	jmp	.m3
.m4:
	emul	mpB,lpA,$0,tmp0		# (tmp0,tmp1) = mpB * pos_lpA
	bbc	$HIGH0,savehigh,.m3	# compensate for the negative bit?
	addl2	l_mpBs,tmp1		# (tmp0,tmp1) += 0x80000000*mpB
	adwc	h_mpBs,tmp0
.m3:

	#
	# (tmp2,tmp3) = mpB*lpA +  lpB*(mpA,lpA)
	#
	addl2	tmp1,tmp3
	adwc	tmp0,tmp2

	#
	# (tmp0, tmp1) == mpB * mpA; mpA and mpB are non-zero.
	#
	emul	mpB,mpA,$0,tmp0		# (tmp0,tmp1) = mpB * mpA

	#
	# (tmp0,tmp1,tmp2,tmp3) = (mpB,lpB) * (mpA,lpA)
	#
	addl2	tmp2,tmp1
	adwc	$0,tmp0
	movl	tmp3,tmp2
	movl	svtmp3,tmp3		# DONE WITH THE MULTIPLICATION!

	#
	# Align the fraction (tmp0,tmp1,tmp2,tmp3) by shifting left
	# from bit 15 to bit 23. Note that the result could have two bits
	# at the left of the decimal point; therefore, we compute number of
	# bits to be shifted left accordingly.  After the shift, excess
	# bits remain in tmp2.  The most significant bit in tmp2 is tested
	# to determine if rounding is necessary.
	#
	bbc	$15,tmp0,.l0		# if (2 bits at left of decimal point)
	SHLQUADS(tmp0,tmp1,tmp2,8)	# then  (tmp0, tmp1, tmp2, tmp3) << 11
	incl	exp
	cmpl	exp,$256
	jgeq	overflow 		# exp is too large
	jmp	.l1
.l0:	SHLQUADS(tmp0,tmp1,tmp2,9)	# else  (tmp0, tmp1, tmp2, tmp3) << 12
.l1:

	bbc 	$31,tmp2,.l2		# is 1st bit leftover after shift a 1?
	incl	tmp1			# round up
	adwc	$0,tmp0			# propagate the carry bit
	bbc	$24,tmp0,.l2		# has fraction overflowed to exp bits?
	clrl	tmp0			# renormalize
	incl	exp			# and adjust exponent

.l2:
	andl2	$CLEARHID,tmp0		# clear hidden bit
	shll	$EXPSHIFT,exp,exp	# shift exp to proper position
	orl2	exp,tmp0		# insert exponent
	xorl3	4(fp),12(fp),i		# see if result is negative
	bbc	$31,i,done
	orl2	$SIGNBIT,tmp0		# set sign bit
done:
	ret

retzero:
	clrl	tmp0			# return 0
	clrl	tmp1
	ret

overflow:
	orl2	$HFS_OVF,*20(fp)	# set overflow indicator
	ret

underflow:
	orl2	$HFS_UNDF,*20(fp)	# set underflow indicator
	ret
