/*
 * ACC v/eiu Ethernet Communications Controller interface
 */

#include "../ml/pte.h"
#include "../ml/mtpr.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/mbuf.h"
#include "../sys/buf.h"
#include "../sys/protosw.h"
#include "../sys/socket.h"
#include "../sys/vmmac.h"
#include "../sys/ioctl.h"
#include "../sys/errno.h"
#include "../sys/vbavar.h"
 
#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"

#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#include "../netinet/if_ether.h"
 
#include "../hcxif/if_com.h"
#include "../hcxif/if_en.h"


/*
 * The following procedures are called by autoconfigure.
 * (enprobe, enslave, enattach)
 */

/*
 * Probe determines whether the controller board actually enists.
 * The um structure is passed in. The address of the controller board
 * is in the um structure. If the controller board needs to be initialized
 * it will be done here. The um structure provides all the information
 * necessary to do this.
 */
int
enprobe( um )
	struct vba_ctlr *		um ;
{
	register struct en_reg *	en_reg_ptr ;
	int				present ;
	short				csr ;
	short                           ivct ;
	extern	int			Xenxintr0, Xenrintr0 ;
	extern	int			Xenxintr1, Xenrintr1 ;
	static	int			proc[ ] = {
						(int)& Xenxintr0,
						(int)& Xenrintr0,
	                        		(int)& Xenxintr1,
	                        		(int)& Xenrintr1
	                		} ;

	en_reg_ptr = (struct en_reg *) um -> um_addr ;
	if ( badaddr( & en_reg_ptr -> csr, 2, 0 ) )
		return( FALSE ) ;
	movow( CSR_GO,
	     & en_reg_ptr -> csr ) ;
	DELAY( 50000 ) ;
	csr = en_reg_ptr -> csr ;
	present = csr & CSR_ACTV ;
	movow( CSR_RESET,
	     & en_reg_ptr -> csr ) ;
	if ( ! present )
	        return( FALSE ) ;

	for ( ivct = 0 ; ; ) {
	        ivct = get_ivct( ivct, 2 ) ;
	        if ( ivct == (-1) )
	                return( FALSE ) ;
	        if ( (ivct & 0x7) == 0 )
	                break ;
	        return_ivct( ivct ++ ) ;
	        return_ivct( ivct ) ;
	}
	movow( ivct,
		& en_reg_ptr -> ivct ) ;
	um -> um_ivct = ivct ;
	init_ivct( ivct,     proc[ (um -> um_ctlr * 2) + 0 ] ) ;
	init_ivct( ivct + 1, proc[ (um -> um_ctlr * 2) + 1 ] ) ;
	return( TRUE ) ;
}

/*
 * Slave determines if the controller has a 'drive' associated with it.
 * There is always exactly one 'drive' associated with this controller.
 */
/*ARGSUSED*/
int
enslave( ui, reg )
	struct vba_device *	ui ;
	caddr_t			reg ;
{
	return( TRUE ) ;
}

/*
 * Attach initializes the 'drive'
 */
int
enattach( ui )
	struct vba_device *		ui;
{
	register int			unit   = ui -> ui_unit ;
	register struct en_softc *	en_ptr = & en_softc[ unit ] ;
	register struct ifnet  *	if_ptr = & en_ptr -> is_if ;
	int				i ;

 
	/*
	 * Figure out where the DPM memory is in the logical address space.
	 */
	if ( ui -> ui_vbanum == 0 )
		endpm[ unit] |= 0xff000000 ;
	else
		endpm[ unit] |= 0xbf000000 ;

	en_ptr -> frame_ptr = (struct frame *) set_up_pte( endpm[ unit ], 32 ) ;
 
	/*
	 * Initialize the I/O registers of the v/eiu.
	 */
	en_ptr -> en_reg_ptr = (struct en_reg *) ui -> ui_addr ;
 
	/*
	 * Load the segment boundary register.
	 */
	movow( (TSEG << 11),
	     &  en_ptr -> en_reg_ptr -> segb ) ;
 
	/*
	 * Load the Broadcast address.
	 */
	for ( i = 0 ; i < SZ_BROADCAST_ADDR ; ++ i )
		movow( ~0xffff,
		     &  en_ptr -> en_reg_ptr -> broadcast_addr[ i ] ) ;
 
	/*
	 * Load the multicast address.
	 */
	for ( i = 0 ; i < SZ_MULTICAST_ADDR ; ++ i )
		movow( ~0x0f,
		     &  en_ptr -> en_reg_ptr -> multicast_addr[ i ] ) ;

	/*
	 * Load the receive address to an initial value.
	 */
	en_ptr -> is_addr[ 0 ] = 0xF0 ;
	en_ptr -> is_addr[ 1 ] = 0xF0 ;
	en_ptr -> is_addr[ 2 ] = 0xF0 ;
	en_ptr -> is_addr[ 3 ] = 0xB0 ;
	en_ptr -> is_addr[ 4 ] = 0xB0 ;
	en_ptr -> is_addr[ 5 ] = (0xB0 | (unit & 0x0F)) ;
 
	/*
	 * Initialize the socket address and interface pointer.
	 */
	if_ptr -> if_unit = unit;
	if_ptr -> if_name = "en";
	if_ptr -> if_flags = IFF_BROADCAST;
	if_ptr -> if_addrlist = 0;
	if_ptr -> if_mtu = ETHERMTU;
 
	if_ptr -> if_init     = eninit;
	if_ptr -> if_output   = enoutput;
	if_ptr -> if_ioctl    = enioctl;
	if_ptr -> if_reset    = enreset;
	if_attach( if_ptr ) ;
}

/* end of autoconfigure procedures */

/*
 * The following procedures are external interfaces.
 * (enreset, eninit, enoutput, enioctl)
 */

/*
 * Reset the controller board.
 */
int
enreset( unit, vban )
	int				unit ;
	int				vban ;
{
	register struct vba_device *	ui;

	if (unit < nen &&
	   ((ui = eninfo[ unit ]) != 0) &&
	     ui -> ui_alive       == TRUE &&
	     ui -> ui_vbanum      == vban &&
	     eninit( unit ) )
		return( TRUE ) ;

	return( FALSE ) ;
}

/*
 * Initialization of interface.
 */
int
eninit( unit )
	int				unit;
{
	register struct en_softc *	en_ptr ;
	register struct ifnet *		if_ptr ;
	short *				ptr ;
	struct frame *			frame_ptr ;
	int				s ;
	int				i ;
	int				j ;
	int				k ;

	/* Initialize local variables */
	en_ptr  = & en_softc[ unit ] ;
	if_ptr = & en_ptr -> is_if ;

	if ( if_ptr->if_addrlist == 0 )
		return( FALSE ) ;
 
	if ( ! (if_ptr -> if_flags & IFF_RUNNING) ) {
		s = splimp( ) ;
		/*
		 * Reset the board.
		 */
		movow( CSR_RESET,
			  & en_ptr -> en_reg_ptr -> csr ) ;

		/*
		 * Load the receive address.
		 */
		for ( i = 0 ; i < 6 ; i ++ )
			movow( ~(en_ptr -> is_addr[ i ]),
			       & en_ptr -> en_reg_ptr -> station_addr[ i ] ) ;
 
		/*
		 * Zero the dual ported memory.
		 */
		ptr = (short *) en_ptr -> frame_ptr ;
		for ( i = 0 ; i < 16384 ; ++ i )
			*(ptr ++) = 0 ;
 
		/*
		 * Load the transmit frames with collision back-off tables.
		 */
		for ( i = TSEG ; i < NUMBER_FRAMES ; ++ i ) {
			frame_ptr = & (en_ptr -> frame_ptr[ i ]) ;
			for ( j = 1, k = 0x0080 ; j < 11 ; ++ j, k <<= 1 )
				frame_ptr -> random[ j ] = k - 2 ;
			for ( ; j < 16 ; ++ j )
				frame_ptr -> random[ j ] = k ;
		}
 
		/*
		 * Initialize the controller structure.
		 */
		en_ptr -> is_if.if_opackets   = 0 ;
		en_ptr -> is_if.if_ipackets   = 0 ;
		en_ptr -> is_if.if_collisions = 0 ;
		en_ptr -> is_if.if_oerrors    = 0 ;
		en_ptr -> is_if.if_ierrors    = 0 ;
		en_ptr -> xmit_avail          = TRUE ;
		en_ptr -> xmit_avail_head     = TSEG ;
		en_ptr -> xmit_avail_tail     = TSEG ;
		en_ptr -> recv_buffer         = 0 ;
		en_ptr -> error_odd           = 0 ;
		en_ptr -> error_crc           = 0 ;
		en_ptr -> error_over_run      = 0 ;
		en_ptr -> error_too_short     = 0 ;
		en_ptr -> output_head_ptr     = NULL ;
		en_ptr -> output_tail_ptr     = NULL ;
		en_ptr -> queue_depth         = 0 ;
		en_ptr -> max_queue_depth     = 0 ;
		en_ptr -> queued_packets      = 0 ;
 
		/*
		 * This is debugging code.
		 */
#ifdef EN_MONITOR
		en_ptr -> monitor         = TRUE ;
		en_ptr -> counter         = 0 ;
		en_ptr -> debug_ptr       = en_debug_buffer ;
		en_ptr -> start_debug_ptr = en_debug_buffer ;
		en_ptr -> end_debug_ptr   = en_debug_buffer + SZ_DEBUG_BUFFER ;
#endif EN_MONITOR
 
		/*
		 * Turn on the v/eiu board for receiving frames.
		 */
		movow( (CSR_GO | CSR_ENINT),
		     &  en_ptr -> en_reg_ptr -> csr ) ;
 
		if_ptr -> if_flags |= IFF_UP | IFF_RUNNING ;
		splx( s ) ;
 
	}
	return( TRUE ) ;
}

/*
 * Ethernet output routine.
 * Encapsulate a packet of type family for the local net.
 * Use trailer local net encapsulation if enough data in first
 * packet leaves a multiple of 512 bytes of data in remainder.
 * If destination is this address or broadcast, send packet to
 * loop device to kludge around the fact that ACC interfaces can't
 * talk to themselves.
 */
int
enoutput( if_ptr, mbuf_ptr, dst )
	struct ifnet *			if_ptr ;
	struct mbuf *			mbuf_ptr ;
	struct sockaddr *		dst ;
{
	int				type ;
	int				s ;
	int				error = 0 ;
	char				edst[ 6 ] ;
	struct in_addr			idst;
	int				unit = if_ptr -> if_unit ;
	register struct en_softc *	en_ptr = & en_softc[ unit ] ;
	register struct mbuf *		m_ptr = mbuf_ptr ;
	register struct ether_header *	en_hdr_ptr ;
	register int			off;
	int				length ;
	char *				data_ptr ;
	struct frame *			frame_ptr ;
	unsigned short *		ptr ;
	extern   struct ifnet		loif;
	int				usetrailers;
#ifdef EN_MONITOR
	int				i ;
	struct en_debug *		debug_ptr ;
	unsigned short *		en_to_ptr ;
	unsigned short *		en_from_ptr ;
#endif EN_MONITOR

	switch ( dst -> sa_family ) {
#ifdef INET
	case AF_INET:
		idst = ((struct sockaddr_in *) dst) -> sin_addr;
		if (!arpresolve(&en_ptr->is_ac,m_ptr,&idst,edst,&usetrailers))
			goto end_it ;      /* if not yet resolved */
		if (in_broadcast(idst)) {
			struct mbuf *mcopy ;
			mcopy = m_copy(m_ptr, 0, (int)M_COPYALL);
	                error = looutput(&loif, mcopy, dst) ;
	                if ( error )
	                	goto end_it ;
		}
		off = ntohs( (unsigned short)mtod( m_ptr,
			struct ip *)->ip_len)-m_ptr->m_len;
		/* need per host negotiation */
		if (usetrailers && off > 0 && (off & 0x1ff) == 0 &&
		      m_ptr->m_off >= MMINOFF + 2 * sizeof(unsigned short)){
			type = ETHERTYPE_TRAIL + (off>>9);
			m_ptr -> m_off -= 2 * sizeof( unsigned short ) ;
			m_ptr -> m_len += 2 * sizeof( unsigned short ) ;
			*mtod( m_ptr, unsigned short * ) =
				htons( (unsigned short)ETHERTYPE_IP ) ;
			*(mtod( m_ptr, unsigned short * ) + 1) =
				htons( (unsigned short)m_ptr -> m_len ) ;

			/*
			 * Packet to be sent as trailer: move first packet
			 * (control information) to end of chain.
			 */
			while ( m_ptr -> m_next )
				m_ptr = m_ptr -> m_next;
			m_ptr -> m_next = mbuf_ptr ;
			m_ptr = mbuf_ptr -> m_next ;
			mbuf_ptr -> m_next = 0 ;
			mbuf_ptr = m_ptr ;
			break ;
		}
		type = ETHERTYPE_IP;
		off = 0;
		break ;
#endif

	case AF_UNSPEC:
		en_hdr_ptr = (struct ether_header *)dst -> sa_data;
 		bcopy( (caddr_t)en_hdr_ptr -> ether_dhost,

		       (caddr_t)edst,
			sizeof (edst));
		type = en_hdr_ptr -> ether_type;
		break ;

	default:
		printf( "en %d: can't handle af%x\n", unit, dst -> sa_family ) ;
		error = EAFNOSUPPORT;
		goto end_it ;
	}

	/*
	 * Check if the queue is empty.
	 */
	s = splimp( ) ;
	if ( en_ptr -> output_head_ptr != NULL ||
		  ! en_ptr -> xmit_avail ) {
		/*
		 * Get an mbuf to hold the ethernet header.
		 */
		MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
		if ( m_ptr == NULL ) {
			error = ENOBUFS ;
			goto end_it ;
		}
		m_ptr -> m_next = mbuf_ptr ;

		ptr = mtod( m_ptr, unsigned short * ) ;
		ptr[ 0 ] = * (unsigned short *)(& edst[ 0 ]) ;
		ptr[ 1 ] = * (unsigned short *)(& edst[ 2 ]) ;
		ptr[ 2 ] = * (unsigned short *)(& edst[ 4 ]) ;
		ptr[ 3 ] = * (unsigned short *)(& en_ptr -> is_addr[ 0 ]) ;
		ptr[ 4 ] = * (unsigned short *)(& en_ptr -> is_addr[ 2 ]) ;
		ptr[ 5 ] = * (unsigned short *)(& en_ptr -> is_addr[ 4 ]) ;
		ptr[ 6 ] = htons( (unsigned short) type ) ;
		m_ptr -> m_len = SZ_ETHER_HDR ;
		/*
		 * Queue this onto the output queue.
		 */
		if ( en_ptr -> output_head_ptr == NULL )
			en_ptr -> output_head_ptr = m_ptr ;
		else
			en_ptr -> output_tail_ptr -> m_act = m_ptr ;
		en_ptr -> output_tail_ptr = m_ptr ;
		m_ptr -> m_act = NULL ;
		/*
		 * Determine queue depth.
		 */
		if ( ++ en_ptr -> queue_depth > en_ptr -> max_queue_depth )
			en_ptr -> max_queue_depth = en_ptr -> queue_depth ;
		++ en_ptr -> queued_packets ;
	} else {
		/*
		 * Mark the frame buffer in use.
		 */
		frame_ptr = (struct frame *)
			& en_ptr -> frame_ptr[ en_ptr -> xmit_avail_head ] ;
		if ( ++ en_ptr -> xmit_avail_head >= NUMBER_FRAMES )
			en_ptr -> xmit_avail_head = TSEG ;
		if (en_ptr -> xmit_avail_head == en_ptr -> xmit_avail_tail)
			en_ptr -> xmit_avail = FALSE ;

		/*
		 * Write the ethernet header directly into the frame.
		 */
		frame_ptr -> hdr_dhost[ 0 ] = *(unsigned short *)(& edst[ 0 ]) ;
		frame_ptr -> hdr_dhost[ 1 ] = *(unsigned short *)(& edst[ 2 ]) ;
		frame_ptr -> hdr_dhost[ 2 ] = *(unsigned short *)(& edst[ 4 ]) ;
		frame_ptr -> hdr_shost[ 0 ] = *(unsigned short *)(& en_ptr->is_addr[0]) ;
		frame_ptr -> hdr_shost[ 1 ] = *(unsigned short *)(& en_ptr->is_addr[2]) ;
		frame_ptr -> hdr_shost[ 2 ] = *(unsigned short *)(& en_ptr->is_addr[4]) ;
		frame_ptr -> hdr_type = htons( (unsigned short) type ) ;

		/*
		 * Copy the frame data from the mbuf to the v/eui board.
		 */
		length = SZ_ETHER_HDR ;
		data_ptr = (char *) frame_ptr -> frame_data ;
		m_ptr = mbuf_ptr ;
		while ( mbuf_ptr ) {
			enput( mtod( mbuf_ptr, caddr_t ),
				data_ptr,(unsigned)mbuf_ptr->m_len );
			data_ptr += mbuf_ptr -> m_len ;
			length   += mbuf_ptr -> m_len ;
			mbuf_ptr = mbuf_ptr -> m_next ;
		}
		m_freem( m_ptr ) ;

		/*
		 * Ensure the length is at least the minimum.
		 */
		if ( length < (SZ_MIN_FRAME + SZ_ETHER_HDR) )
			length = (SZ_MIN_FRAME + SZ_ETHER_HDR) ;

		/*
		 * This is debugging code.
		 */
#ifdef EN_MONITOR
		if ( en_ptr -> monitor ) {
			debug_ptr = en_ptr -> debug_ptr ;
			if (++en_ptr -> debug_ptr >= en_ptr->end_debug_ptr)
				en_ptr -> debug_ptr = en_ptr -> start_debug_ptr;
			debug_ptr -> counter = en_ptr -> counter ++ ;
			debug_ptr -> direction = EN_XMIT ;
			debug_ptr -> cmnd_status = CS_FULL | length ;
			en_to_ptr   = (unsigned short *) debug_ptr -> hdr_dhost ;
			en_from_ptr = (unsigned short *) frame_ptr -> hdr_dhost ;
			for ( i = (SZ_DEBUG_DATA + 7) ; i ; -- i )
				*(en_to_ptr ++) = *(en_from_ptr ++) ;
		}
#endif EN_MONITOR

		/*
		 * Mark the frame buffer full and ready to xmit.
		 */
		frame_ptr -> cmnd_status = CS_FULL | length ;
		++ en_ptr -> is_if.if_opackets ;
	}
	splx( s ) ;

end_it :
	if ( error != 0 )
		m_freem( mbuf_ptr ) ;
	return( error ) ;
}

/*
 * Process an ioctl request.
 */
int
enioctl( if_ptr, cmd, data )
	register struct ifnet *	if_ptr;
	int			cmd;
	caddr_t			data;
{
	register struct ifreq * ifr_ptr = (struct ifreq *) data;
	register struct ifaddr* ifa_ptr = (struct ifaddr *) data;
	int			s = splimp( ) ;
	int			error = 0;
	int			unit = if_ptr -> if_unit ;
	struct en_softc *	en_ptr ;

	switch ( cmd ) {

	case SIOCSIFADDR:
		if_ptr->if_flags |= IFF_UP;
		switch (ifa_ptr->ifa_addr.sa_family) {
#ifdef INET
		case AF_INET:
			eninit(if_ptr->if_unit);   /* before arpwhohas */
			((struct arpcom *)if_ptr)->ac_ipaddr =
				IA_SIN(ifa_ptr)->sin_addr;
			arpwhohas((struct arpcom *)if_ptr,&IA_SIN(ifa_ptr)->sin_addr);
			break;
#endif
#ifdef NS
		case AF_NS:
#endif
		default:
			return (EADDRNOTAVAIL);
		}
		break;

	case SIOCSETETADDR:
		if_ptr -> if_flags &= ~(IFF_RUNNING | IFF_UP) ;
		en_ptr = & en_softc[ unit ] ;
		movow( CSR_RESET,
			  & en_ptr -> en_reg_ptr -> csr ) ;
		bcopy( (caddr_t) ifr_ptr -> ifr_addr.sa_data,
		       (caddr_t) en_ptr -> is_addr,
			sizeof( en_ptr -> is_addr ) ) ;
		eninit( unit ) ;
		break;

	default:
		error = EINVAL;
	}
	splx( s ) ;
	return( error ) ;
}

/* end of external interfaces */

/*
 * V/EUI xmit complete interrupt.
 */
int
enxintr( unit )
	int				unit;
{
	register struct en_softc *	en_ptr ;
	short				csr;
	struct frame *			frame_ptr ;
	unsigned int			j ;
	unsigned int			k ;
	struct mbuf *			mbuf_ptr ;
	unsigned int			length ;
	char *				data_ptr ;
	struct mbuf *			m_ptr ;
#ifdef EN_MONITOR
	unsigned int			i ;
	struct en_debug *		debug_ptr ;
	unsigned short *		en_to_ptr ;
	unsigned short *		en_from_ptr ;
#endif EN_MONITOR

	/*
	 * Initialize the local structures.
	 */
	en_ptr = & en_softc[ unit ] ;
 
	for ( ;; ) {
		frame_ptr =
		     (struct frame *)&en_ptr->frame_ptr[en_ptr->xmit_avail_tail ];
		csr = frame_ptr -> cmnd_status ;
		if ( csr & CS_FULL )
			break ;
		if ( csr & CS_TRTC ) {
			en_ptr -> is_if.if_collisions += csr & CS_TRTC ;
			if ( csr & CS_TRTFAIL )
				en_ptr -> is_if.if_oerrors ++ ;
			for ( j = 1, k = 0x0080 ; j < 11 ; ++ j, k <<= 1 )
				frame_ptr -> random[ j ] = k - 2 ;
			for ( ; j < 16 ; ++ j )
				frame_ptr -> random[ j ] = k ;
		}
		if ( ++ en_ptr -> xmit_avail_tail >= NUMBER_FRAMES )
			en_ptr -> xmit_avail_tail = TSEG ;
		if ( en_ptr -> output_head_ptr != NULL ) {
			/*
			 * Dequeue the mbuf chain.
			 */
			mbuf_ptr = en_ptr -> output_head_ptr ;
			en_ptr -> output_head_ptr = mbuf_ptr -> m_act ;
			mbuf_ptr -> m_act = NULL ;
			/*
			 * Decrement the queue depth.
			 */
			en_ptr -> queue_depth -- ;
 
			/*
			 * Copy the frame data from the mbuf to the v/eui board.
			 */
			frame_ptr = (struct frame *)
			    & en_ptr -> frame_ptr[ en_ptr -> xmit_avail_head ] ;
			length = 0 ;
			data_ptr = (char *) frame_ptr -> hdr_dhost ;
			m_ptr = mbuf_ptr ;
			while ( mbuf_ptr ) {
				enput( mtod( mbuf_ptr, caddr_t ),
					data_ptr,
					(unsigned)mbuf_ptr -> m_len );
				data_ptr += mbuf_ptr -> m_len ;
				length   += mbuf_ptr -> m_len ;
				mbuf_ptr = mbuf_ptr -> m_next ;
			}
			m_freem( m_ptr ) ;
 
			/*
			 * Ensure the length is at least the minimum.
			 */
			if ( length < (SZ_MIN_FRAME + SZ_ETHER_HDR) )
				length = (SZ_MIN_FRAME + SZ_ETHER_HDR) ;
			/*
			 * This is debugging code.
			 */
#ifdef EN_MONITOR
			if ( en_ptr -> monitor ) {
				debug_ptr = en_ptr -> debug_ptr ;
				if ( ++ en_ptr -> debug_ptr >=
						en_ptr -> end_debug_ptr )
					en_ptr -> debug_ptr =
						en_ptr -> start_debug_ptr ;
				debug_ptr -> counter = en_ptr -> counter ++ ;
				debug_ptr -> direction = EN_XMIT ;
				debug_ptr -> cmnd_status = CS_FULL | length ;
				en_to_ptr   = (unsigned short *) debug_ptr -> hdr_dhost ;
				en_from_ptr = (unsigned short *) frame_ptr -> hdr_dhost ;
				for ( i = (SZ_DEBUG_DATA + 7) ; i ; i -- )
					*(en_to_ptr ++) = *(en_from_ptr ++) ;
			}
#endif EN_MONITOR
 
			/*
			 * Mark the frame buffer full and ready to xmit.
			 */
			frame_ptr -> cmnd_status = CS_FULL | length ;
			++ en_ptr -> is_if.if_opackets ;

			if ( ++ en_ptr -> xmit_avail_head >= NUMBER_FRAMES)
				en_ptr -> xmit_avail_head = TSEG ;
			if ( en_ptr -> xmit_avail_head ==
					en_ptr -> xmit_avail_tail )
				en_ptr -> xmit_avail = FALSE ;
		} else
			en_ptr -> xmit_avail = TRUE ;
		if (en_ptr -> xmit_avail_tail == en_ptr -> xmit_avail_head)
			break ;
	}
 
}

/*
 * V/EUI receive complete interrupt.
 */
int
enrintr( unit )
	int				unit;
{
	register struct en_softc *	en_ptr ;
	short				csr;
	struct frame *			frame_ptr ;
#ifdef EN_MONITOR
	struct en_debug *		debug_ptr ;
	unsigned short *		en_to_ptr ;
	unsigned short*			en_from_ptr ;
	unsigned int			i ;
#endif EN_MONITOR

	/*
	 * Initialize the local structures.
	 */
	en_ptr = & en_softc[ unit ] ;

	/*
	 * Ignore interrupts prior to en board initialization.
	 */
	if ( ! en_ptr -> frame_ptr )
		return ;
 
	for ( ;; ) {
		frame_ptr =
		  (struct frame *) & en_ptr -> frame_ptr[ en_ptr -> recv_buffer ];
		csr = frame_ptr -> cmnd_status ;
		if ( ! (csr & CS_FULL) )
			break ;
 
		/*
	 	 * The following is debugging code.
		 */
#ifdef EN_MONITOR
		if ( en_ptr -> monitor ) {
			debug_ptr = en_ptr -> debug_ptr ;
			if (++en_ptr -> debug_ptr >= en_ptr->end_debug_ptr)
				en_ptr -> debug_ptr =
					en_ptr -> start_debug_ptr ;
			debug_ptr -> counter = en_ptr -> counter ++ ;
			debug_ptr -> direction = EN_RECV ;
			en_to_ptr   = (unsigned short *) & debug_ptr -> cmnd_status ;
			en_from_ptr = (unsigned short *) frame_ptr;
			for ( i = (SZ_DEBUG_DATA + 8) ; i ; -- i )
				*(en_to_ptr ++) = *(en_from_ptr ++) ;
		}
#endif EN_MONITOR
 
		enrecv( unit, frame_ptr ) ;
		frame_ptr -> cmnd_status = 0 ;
		if ( ++ en_ptr -> recv_buffer == TSEG )
			en_ptr -> recv_buffer = 0 ;
	}
 
}
 
/*
 * Ethernet interface receive frame procedure.
 * If input error just drop packet, otherwise enamine packet to
 * determine type.  If can't determine length
 * from type, then have to drop packet.  Otherwise decapsulate
 * packet based on type and pass to type specific higher-level
 * input routine.
 */
int
enrecv( unit, frame_ptr )
	int				unit;
	struct frame *			frame_ptr ;
{
	register struct en_softc *	en_ptr = & en_softc[ unit ] ;
	struct mbuf*			m      = 0 ;
	short				csr ;
	int				len ;/* length of frame received */
	int				dlen ;   /* Data length */
	int				hlen ;   /* header length */
	int				s ;
	register struct ifqueue *	inq ;

	en_ptr -> is_if.if_ipackets ++ ;
	csr = (frame_ptr -> cmnd_status & 0xffff) ;
	len = (csr & CS_BYTE_CNT) - SZ_ETHER_HDR - SZ_ETHER_FCS ;
	if ( (csr & (CS_RODD | CS_RCRC | CS_ROVRN)) || len < SZ_MIN_FRAME ) {
		if ( csr & CS_RODD )
			en_ptr -> error_odd ++ ;
		if ( csr & CS_RCRC )
			en_ptr -> error_crc ++ ;
		if ( csr & CS_ROVRN )
			en_ptr -> error_over_run ++ ;
		if ( len < SZ_MIN_FRAME )
			en_ptr -> error_too_short ++ ;
		en_ptr -> is_if.if_ierrors ++ ;
		goto error_return ;
	}
 
	/*
	 * Deal with trailer protocol: if type is PUP trailer
	 * get true type from first 16-bit word past data.
	 * Remember that type was trailer by setting off.
	 */
	frame_ptr -> hdr_type = ntohs( frame_ptr -> hdr_type ) ;
#define   endataaddr( en, off, type ) \
		((type)(((caddr_t)(en) + (off))))
	if ( frame_ptr -> hdr_type >= ETHERTYPE_TRAIL &&
	     frame_ptr -> hdr_type < ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER ) {
		dlen = (frame_ptr -> hdr_type - ETHERTYPE_TRAIL) * 512;
		if ( dlen >= ETHERMTU )
			goto error_return ;
		frame_ptr -> hdr_type =
			ntohs( *endataaddr( frame_ptr -> frame_data,
				dlen, unsigned short * ) ) ;
		hlen = ntohs( *(endataaddr( frame_ptr -> frame_data,
				dlen+2, unsigned short * )));
		if ( dlen + hlen > len )
			goto error_return ;         /* sanity */
		len = dlen ;
	} else
		hlen = 0;
	if ( len == 0 )
		goto error_return ;
 
	/*
	 * Pull packet off interface.  Off is nonzero if packet
	 * has trailing header; enget will then force this header
	 * information to be at the front, but we still have to drop
	 * the type and length which are at the front of any trailer data.
	 */
	m = enget( frame_ptr -> frame_data, len, hlen, &en_ptr->is_if ) ;
	if ( m == 0 )
		goto error_return ;
	if ( hlen ) {
		m -> m_off += 2 * sizeof( unsigned short ) ;
		m -> m_len -= 2 * sizeof( unsigned short ) ;
	}
	switch ( frame_ptr -> hdr_type ) {
 
#ifdef INET
	case ETHERTYPE_IP:
		schednetisr( NETISR_IP ) ;
		inq = & ipintrq ;
		if ( IF_QFULL( inq ) ) {
			IF_DROP( inq ) ;
			goto error_return ;
		}
		s = splimp( ) ;
		IF_ENQUEUE( inq, m ) ;
		splx( s ) ;
		break ;
 
	case ETHERTYPE_ARP:
		arpinput( & en_ptr -> is_ac, m ) ;
		break ;
#endif
	default:
error_return :
		if ( m )
			m_freem( m ) ;
		return ;
	}
 
}
 
/*
 * This procedure will copy from the frame in the controller board's memory
 * into mbufs. hlen is the trailer length, it must be moved to the
 * front of the mbufs.
 */
struct mbuf *
enget( ptr, length, hlen, if_ptr )
	register short *	ptr ;
	register int		length ;
	register int		hlen ;
	struct ifnet *		if_ptr ;
{
	register short *	m_data_ptr ;
	register int		size ;
	register struct mbuf *	m_ptr ;        /* mbuf    */
	register struct mbuf *	tail_m_ptr ;
	struct mbuf *		head_m_ptr ;

	MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
	if ( m_ptr == 0 ) return( 0 ) ;
	head_m_ptr = tail_m_ptr = m_ptr ;
	while ( length > 0 ) {
		tail_m_ptr -> m_next = m_ptr ;
		if ( m_ptr == 0 ) {
			m_freem( head_m_ptr ) ;
			return( 0 ) ;
		}
		tail_m_ptr = m_ptr ;
		if (if_ptr)
			length += sizeof(if_ptr);
		size = MIN( length, MLEN ) ;
		if (length >= MCLBYTES) {
			MCLGET( m_ptr );
			if (m_ptr->m_len == MCLBYTES)
				size = MCLBYTES ;
		}
		m_ptr -> m_len = size ;
		length -= size ;
		m_data_ptr = mtod( m_ptr, short * ) ;
		if (if_ptr) {
			*(mtod(m_ptr, struct ifnet **)) = if_ptr;
			m_data_ptr += sizeof(if_ptr)/sizeof(* m_data_ptr);
			size -= sizeof (if_ptr);
			if_ptr = (struct ifnet *)0;
		}
		while ( size > 0 ) {
			*m_data_ptr = *ptr ;
			m_data_ptr ++ ;
			ptr ++ ;
			size -= 2 ;
		}
		if ( length )
			MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
	}
	tail_m_ptr -> m_next = 0 ;
 
	if ( hlen ) {
		MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
		if ( m_ptr == 0 ) {
			m_freem( head_m_ptr ) ;
			return( 0 ) ;
		}
		m_ptr -> m_len = hlen ;
		m_data_ptr = mtod( m_ptr, short * ) ;
		while ( hlen > 0 ) {
			*m_data_ptr = *ptr ;
			m_data_ptr ++ ;
			ptr ++ ;
			hlen -= 2 ;
		}
		m_ptr -> m_next = head_m_ptr ;
		head_m_ptr = m_ptr ;
	}
	return( head_m_ptr ) ;
}
 
/*
 * Copy from an mbuf data pointer to the dpm.
 */
int
enput( from_ptr, to_ptr, length )
	register	char * from_ptr ;       /* r12 */
	register	char * to_ptr ;         /* r11 */
	register	int   length ;         /* r10 */
{
	register	short * short_from_ptr ;  /* r9 */
	register	short * short_to_ptr ;    /* r8 */
	register	int   temp ;           /* r7 */


	/* There are four conditions depending on which pointers are odd. */
#define BOTH_EVEN 0
#define TO_ODD    1
#define FROM_ODD  2
#define BOTH_ODD  3
	temp = (((int) from_ptr & 1) << 1) | ((int) to_ptr & 1) ;
	switch ( temp ) {
	case TO_ODD :
		short_to_ptr = (short *) ((int) to_ptr - 1) ;	
		to_ptr ++ ;
		temp =  *short_to_ptr ;
		temp |= (unsigned char)*(from_ptr ++) ;
		*short_to_ptr = temp ;
		length -- ;
		/* DROP THROUGH */
	case FROM_ODD :
		short_to_ptr = (short *) to_ptr ;
		for ( ; length > 1 ; length -= 2 ) {
			temp = *(from_ptr ++) ;
			temp <<= 8 ;
			temp |= (unsigned char)*(from_ptr ++) ;
			asm( "movow r7,(r8)" ) ; /* *(short_to_ptr) = temp */
			short_to_ptr ++ ;
		}
		break ;
	case BOTH_ODD :
		short_to_ptr = (short *) ((int) to_ptr - 1) ;
		to_ptr ++ ;
		temp =  *short_to_ptr ;
		temp |= (unsigned char)*(from_ptr ++) ;
		*short_to_ptr = temp ;
		length -- ;
		/* DROP THROUGH */
	case BOTH_EVEN :
		short_to_ptr   = (short *) to_ptr ;
		short_from_ptr = (short *) from_ptr ;
		for ( ; length > 1 ; length -= 2 ) {
			asm( "movow (r9),(r8)" );
			short_to_ptr ++ ;
			short_from_ptr ++ ;
		}
		from_ptr = (char *) short_from_ptr ;
		break ;
	}

	/*
	* Check for an odd ending boundary.
	*/
	if ( length ) {
		temp = (unsigned char)*from_ptr ;
		temp <<= 8 ;
		*short_to_ptr = temp ;
	}
}
