/*
 * Excelan EXOS 202(VME) Link Level Ethernet Interface Driver
 */

#include "../ml/pte.h"
#include "../ml/mtpr.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/mbuf.h"
#include "../sys/buf.h"
#include "../sys/protosw.h"
#include "../sys/socket.h"
#include "../sys/vmmac.h"
#include "../sys/ioctl.h"
#include "../sys/errno.h"
#include "../sys/vbavar.h"
#include "../sys/reboot.h"

#include "../net/if.h"
#include "../net/netisr.h"
#include "../net/route.h"

#include "../netinet/in.h"
#include "../netinet/in_systm.h"
#include "../netinet/in_var.h"
#include "../netinet/ip.h"
#include "../netinet/ip_var.h"
#include "../netinet/if_ether.h"

#include "../hcxif/if_com.h"
#include "../hcxif/if_ex.h"

extern struct ex_msg * exgetcbuf( ) ;

/*
 * The following procedures are called by autoconfigure.
 * (exprobe, exslave, exattach)
 */

/*
 * Probe determines whether the controller board actually exists.
 * The um structure is passed in. The address of the controller board
 * is in the um structure. If the controller board needs to be initialized
 * it will be done here. The um structure provides all the information
 * necessary to do this.
 */
int
exprobe( um )
	struct vba_ctlr *	um ;
{
 	register struct ex_reg *	ex_reg_ptr ;
	int				ivcnt ;
 	register struct ex_softc *	ex_ptr ;
	extern  int                     Xexintr0, Xexintr1 ;
	static	int			proc[ ] = {
						(int)& Xexintr0, (int)& Xexintr1
					} ;


	ex_reg_ptr = (struct ex_reg *) um -> um_addr ;
	if ( badaddr( (caddr_t) & ex_reg_ptr -> ex_portb, 2, 0 ) )
		return( FALSE ) ;
	/*
	 * Reset EXOS and run self-test (should complete within 2 seconds).
	 */
  	movow( EX_RESET,
	     & ex_reg_ptr -> ex_porta ) ;
	if ( ex_reg_ptr -> ex_portb & EX_TESTOK )
		return( FALSE ) ;
	DELAY( 3000000 ) ;  /* Wait for several seconds. */

	if ( ! ex_reg_ptr -> ex_portb & EX_TESTOK )
		return( FALSE ) ;

	/* Allocate the interrupt vector */
        um -> um_ivct = get_ivct( 0, 1 ) ;
        if ( um -> um_ivct == (-1) )
                return( FALSE ) ;

        init_ivct( um -> um_ivct, proc[ um -> um_ctlr ] ) ;
	return( TRUE ) ;
}

/*
 * Slave determines if the controller has a 'drive' associated with it.
 * There is always exactly one 'drive' associated with this controller.
 */
/*ARGSUSED*/
int
exslave( ui, reg )
	struct vba_device *	ui ;
	caddr_t			reg ;
{
	return( TRUE ) ;
}

/*
 * Attach initializes the 'drive'
 */
int
exattach( ui )
	struct vba_device *	ui ;
{
	register int			unit   = ui -> ui_unit ;
	register struct ex_softc *	ex_ptr = & ex_softc[ unit ] ;
	register struct ifnet *		if_ptr = & ex_ptr -> is_if ;
	register struct ex_reg *	ex_reg_ptr ;
	register struct ex_msg *	ex_msg_ptr ;
	extern unsigned int		exwaitbuf() ;
	extern unsigned int		excopyaddr() ;

	ex_reg_ptr = (struct ex_reg *) ui -> ui_addr ;
	ex_ptr -> ex_reg_ptr = ex_reg_ptr ;
	ex_ptr -> unit = unit ;
	/*
	 * Send the configure message to the EXOS 202
	 */
	if ( ! exconfig( ui, 0 ) ) 	/* without interrupts */
		return( FALSE ) ;	/* bad conf */

	/*
	 * Read the receive address from the exos 202 board.
	 */
	ex_msg_ptr = exgetcbuf( ex_ptr ) ;	
	ex_msg_ptr -> mb_na.na_request = LLNET_ADDRS ;
	ex_msg_ptr -> mb_na.na_mask    = READ_OBJ ;
	ex_msg_ptr -> mb_na.na_slot    = PHYSSLOT ;
	ex_msg_ptr -> mb_status        = MH_EXOS ;
	movow( EX_INTERRUPT,
	     & ex_reg_ptr -> ex_portb ) ;

	exwaitbuf( ex_ptr, excopyaddr ) ;


	/*
	 * Initialize the socket address and interface pointer.
	 */
	if_ptr -> if_unit = unit ;
	if_ptr -> if_name = "ex" ;
	if_ptr -> if_flags = IFF_BROADCAST ;
	if_ptr -> if_addrlist = 0 ;
	if_ptr -> if_mtu = ETHERMTU ;

	if_ptr -> if_init     = exinit ;
	if_ptr -> if_output   = exoutput ;
	if_ptr -> if_ioctl    = exioctl ;
	if_ptr -> if_reset    = exreset ;
	if_attach( if_ptr ) ;
	return( TRUE ) ;
}

/* end of autoconfigure procedures */

/*
 * The following procedures are external interfaces.
 * (exreset, exinit, exoutput, exioctl)
 */

/*
 * Reset the controller board.
 */
int
exreset( unit, vban )
	int				unit ;
	int				vban ;
{
	register struct vba_device *	ui ;

	if (unit < nex &&
	   ((ui = exinfo[ unit ]) != 0) &&
	     ui -> ui_alive       == TRUE &&
	     ui -> ui_vbanum      == vban &&
	     exinit( unit ) )
		return( TRUE ) ;

	return( FALSE ) ;
}

/*
 * Initialization of interface ; clear recorded pending operations, and
 * reinitialize BUS usage. Called at boot time, and at ifconfig time via 
 * exioctl, with interrupts disabled.
 */
int
exinit( unit )
	int				unit ;
{
	register struct ex_softc *	ex_ptr ;
	register struct ifnet *		if_ptr ;
	register struct vba_device *	ui ;
	register struct ex_reg *	ex_reg_ptr ;
	register struct ex_msg *	ex_msg_ptr ;
	int	 			s ;

	/* Initialize local variables */
	ex_ptr  = & ex_softc[ unit ] ;
	ui = exinfo[ unit ] ;
	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
	if_ptr = & ex_ptr -> is_if ;

	if ( if_ptr -> if_addrlist == 0 )
		return( FALSE ) ;

	if ( ! (if_ptr -> if_flags & IFF_RUNNING) ) {
		s = splimp( ) ;
		exconfig( ui, 4 ) ;	/* with vectored interrupts */

		/*
		 * Initialize the controller structure.
		 */
		ex_ptr -> is_if.if_opackets   = 0 ;
		ex_ptr -> is_if.if_ipackets   = 0 ;
		ex_ptr -> is_if.if_collisions = 0 ;
		ex_ptr -> is_if.if_oerrors    = 0 ;
		ex_ptr -> is_if.if_ierrors    = 0 ;

		/*
		 * Load the receive address.
		 */
		ex_msg_ptr = exgetcbuf( ex_ptr ) ;
		bcopy( (caddr_t) ex_ptr -> is_addr,
		       (caddr_t) & ex_msg_ptr -> mb_na.na_addrs[ 0 ],
			sizeof( ex_ptr -> is_addr ) ) ;
		ex_msg_ptr -> mb_na.na_request = LLNET_ADDRS ;
		ex_msg_ptr -> mb_na.na_mask    = WRITE_OBJ ;
		ex_msg_ptr -> mb_na.na_slot    = PHYSSLOT ;
		ex_msg_ptr -> mb_status        = MH_EXOS ;
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		exwaitbuf( ex_ptr, 0 ) ;

		/*
		 * Enable receiving on that address slot.
		 */
		ex_msg_ptr = exgetcbuf( ex_ptr ) ;
		ex_msg_ptr -> mb_nr.nr_request = LLNET_RECV ;
		ex_msg_ptr -> mb_nr.nr_mask    = WRITE_OBJ | ENABLE_RCV ;
		ex_msg_ptr -> mb_nr.nr_slot    = PHYSSLOT ;
		ex_msg_ptr -> mb_status        = MH_EXOS ;
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		exwaitbuf( ex_ptr, 0 ) ;

	        /*
	         * Enable receiving on the broadcast slot
	         */
	        ex_msg_ptr = exgetcbuf( ex_ptr ) ;
	        ex_msg_ptr -> mb_nr.nr_request = LLNET_RECV ;
	        ex_msg_ptr -> mb_nr.nr_mask    = WRITE_OBJ | ENABLE_RCV ;
	        ex_msg_ptr -> mb_nr.nr_slot    = BROADSLOT ;
	        ex_msg_ptr -> mb_status        = MH_EXOS ;
	        movow( EX_INTERRUPT,
	             & ex_reg_ptr -> ex_portb ) ;
	        exwaitbuf( ex_ptr, 0 ) ;

#ifdef EX_MONITOR
		ex_ptr -> counter         = 0 ;
		ex_ptr -> debug_ptr       = ex_debug_buffer ;
		ex_ptr -> start_debug_ptr = ex_debug_buffer ;
		ex_ptr -> end_debug_ptr   = ex_debug_buffer + SZ_DEBUG_BUFFER ;
#endif EX_MONITOR

		/*
		 * Put EXOS on the Ethernet, using NET_MODE command
		 */
		ex_msg_ptr = exgetcbuf( ex_ptr ) ;
		ex_msg_ptr -> mb_nm.nm_request = LLNET_MODE ;
		ex_msg_ptr -> mb_nm.nm_mask    = WRITE_OBJ ;
		ex_msg_ptr -> mb_nm.nm_optn    = 0x30 ;
		ex_msg_ptr -> mb_nm.nm_mode    = MODE_PERF ;
		ex_msg_ptr -> mb_status        = MH_EXOS ;
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		exwaitbuf( ex_ptr, 0 ) ;

		exhangrcv( unit ) ;		/* hang receive request */
		exstart( unit ) ;		/* start transmits */
		if_ptr -> if_flags |= IFF_UP | IFF_RUNNING ;
		splx( s ) ;
	}

	return( TRUE ) ;
}

/*
 * Ethernet output routine. Encapsulate a packet of type family for the local 
 * net.  Use trailer local net encapsulation if enough data in first packet 
 * leaves a multiple of 512 bytes of data in remainder.
 */
int
exoutput( if_ptr, mbuf_ptr, dst )
	register struct ifnet *		if_ptr ;
	struct mbuf *			mbuf_ptr ;
	struct sockaddr *		dst ;
{
	int				error = 0 ;
	int				type ;
	int				s ;
	char				edst[ 6 ] ;
	struct in_addr			idst ;
	register struct ex_softc *	ex_ptr = & ex_softc[if_ptr -> if_unit] ;
	register struct mbuf *		m_ptr = mbuf_ptr ;
	register struct ether_header *	ex_hdr_ptr ;
	register int			off ;
	extern   struct ifnet		loif;
	int				usetrailers;

	switch ( dst -> sa_family ) {
#ifdef INET
	case AF_INET:
		idst = ((struct sockaddr_in *) dst) -> sin_addr ;
		if ( ! arpresolve( &ex_ptr -> is_ac, m_ptr, &idst, edst, &usetrailers))
			goto end_it ;	/* if not yet resolved */
		if (in_broadcast(idst)) {
			struct mbuf *	mcopy ;
			mcopy = m_copy( m_ptr, 0, (int) M_COPYALL ) ;
	                error = looutput( & loif, mcopy, dst ) ;
	                if ( error )
	                	goto end_it ;
		}
		off = ntohs((unsigned short)mtod( m_ptr,
                        	struct ip *) -> ip_len) - m_ptr -> m_len ;
		/* need per host negotiation */
		if ( usetrailers && off > 0 && (off & 0x1ff) == 0 &&
		      m_ptr -> m_off >= MMINOFF + 2 * sizeof(unsigned short)) {
			type = ETHERTYPE_TRAIL + (off>>9) ;
			m_ptr -> m_off -= 2 * sizeof( unsigned short ) ;
			m_ptr -> m_len += 2 * sizeof (unsigned short ) ;
			*mtod( m_ptr, unsigned short * ) =
				htons( (unsigned short) ETHERTYPE_IP ) ;
			*(mtod( m_ptr, unsigned short * ) + 1 ) =
				htons( (unsigned short) m_ptr -> m_len ) ;
			/*
			 * Packet to be sent as trailer: move first packet
			 * (control information) to end of chain.
			 */
			while ( m_ptr -> m_next )
				m_ptr = m_ptr -> m_next;
			m_ptr -> m_next = mbuf_ptr ;
			m_ptr = mbuf_ptr -> m_next ;
			mbuf_ptr -> m_next = 0 ;
			mbuf_ptr = m_ptr ;
			break ;
		} else {
			type = ETHERTYPE_IP ;
			off = 0 ;
		}
		break ;
#endif

	case AF_UNSPEC:
		ex_hdr_ptr = (struct ether_header *) dst -> sa_data ;
		bcopy( (caddr_t) ex_hdr_ptr -> ether_dhost,
		       (caddr_t) edst,
		        sizeof( edst ) ) ;
		type = ex_hdr_ptr -> ether_type ;
		off = 0 ;
		break ;

	default:
		printf("ex%d: can't handle af%d\n", if_ptr -> if_unit,
			dst -> sa_family) ;
		error = EAFNOSUPPORT ;
		goto end_it ;
	}

	/*
	 * Add local net header.  If no space in first mbuf, allocate another.
	 */
	if ( m_ptr -> m_off > MMAXOFF ||
	    MMINOFF + sizeof( struct ether_header ) > m_ptr -> m_off ) {
		m_ptr = m_get( M_DONTWAIT, MT_HEADER ) ;
		if ( m_ptr == 0 ) {
			error = ENOBUFS ;
			goto end_it ;
		}
		m_ptr -> m_next = mbuf_ptr ;
		m_ptr -> m_off = MMINOFF ;
		m_ptr -> m_len = SZ_ETHER_HDR ;
	} else {
		m_ptr -> m_off -= SZ_ETHER_HDR ;
		m_ptr -> m_len += SZ_ETHER_HDR ;
	}
	ex_hdr_ptr = mtod( m_ptr, struct ether_header * ) ;
	ex_hdr_ptr -> ether_type = htons( (unsigned short) type ) ;
	bcopy( (caddr_t) edst,
	       (caddr_t) ex_hdr_ptr -> ether_dhost,
	        sizeof( edst ) ) ;
	bcopy( (caddr_t) ex_ptr -> is_addr,
	       (caddr_t) ex_hdr_ptr -> ether_shost,
	        sizeof( ex_ptr -> is_addr ) ) ;

	/*
	 * Queue message on interface, and start output if interface not yet 
	 * active.
	 */
	s = splimp( ) ;
	if ( IF_QFULL( & if_ptr -> if_snd ) ) {
		IF_DROP( & if_ptr -> if_snd ) ;
		error = ENOBUFS ;
		goto end_it ;
	}
	IF_ENQUEUE( & if_ptr -> if_snd, m_ptr ) ;

	/*
	 * If transmit request not already pending, then kick the back end.
	 */
	if ( (ex_ptr -> is_flags & EX_XPENDING) == 0 )
		exstart( if_ptr -> if_unit ) ;
	splx( s ) ;

end_it:
	if ( error != 0 )
		m_freem( mbuf_ptr ) ;
	return( error ) ;
}

/*
 * Process an ioctl request.
 */
int
exioctl( if_ptr, cmd, data )
	register struct ifnet *		if_ptr ;
	int				cmd ;
	caddr_t				data ;
{
	register struct ifreq *		ifr_ptr = (struct ifreq *) data ;
	register struct ifaddr *	ifa_ptr = (struct ifaddr *) data ;
	int				s = splimp( ) ;
	int				error = 0 ;
	int				unit = if_ptr -> if_unit ;
	register struct ex_softc *	ex_ptr ;
	register struct ex_reg *	ex_reg_ptr ;
	struct ex_msg *			ex_msg_ptr ;

	switch ( cmd ) {

	case SIOCSIFADDR:
		if_ptr -> if_flags |= IFF_UP;
		switch (ifa_ptr->ifa_addr.sa_family) {
#ifdef INET
		case AF_INET:
			exinit(if_ptr->if_unit);   /* before arpwhohas */
			((struct arpcom *)if_ptr)->ac_ipaddr =
				IA_SIN(ifa_ptr)->sin_addr;
			arpwhohas((struct arpcom *)if_ptr,&IA_SIN(ifa_ptr)->sin_addr);
			break;
#endif
#ifdef NS
		case AF_NS:
#endif
		default:
			return (EADDRNOTAVAIL);
		}
		break ;

	case SIOCSETETADDR:
		if_ptr -> if_flags &= ~(IFF_RUNNING | IFF_UP) ;
		ex_ptr = & ex_softc[ unit ] ;
		ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
		/* Reset the controller */
		movow( EX_RESET,
		     & ex_reg_ptr -> ex_porta ) ;
		DELAY( 3000000 ) ;

		bcopy( (caddr_t) ifr_ptr -> ifr_addr.sa_data,
		       (caddr_t) ex_ptr -> is_addr,
			sizeof( ex_ptr -> is_addr ) ) ;
		exinit( unit ) ;
		break;

	/* gather statistics from the excelan board */
	case SIOCGSTAT:
		ex_ptr = & ex_softc[ unit ] ;
		ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
		ex_msg_ptr = exgetcbuf( ex_ptr ) ;
		ex_msg_ptr -> mb_ns.ns_request = LLNET_STSTCS ;
		ex_msg_ptr -> mb_ns.ns_mask    = READ_OBJ ;
		ex_msg_ptr -> mb_ns.ns_rsrv    = 0 ;
                ex_msg_ptr -> mb_ns.ns_nobj    = 8 ;
                ex_msg_ptr -> mb_ns.ns_xobj    = 0 ;
                ex_msg_ptr -> mb_ns.ns_bufp    =
                	PHYS_ADDR( &ex_ptr -> ex_qb_ptr -> sa ) +
                	(VME_ADDRESS_MODIFIER << 24 ) ;
		ex_msg_ptr -> mb_status        = MH_EXOS ;
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		exwaitbuf( ex_ptr, 0 ) ;
	        break ;

	/* reset statistics on the excelan board */
	case SIOCSSTAT:
		ex_ptr = & ex_softc[ unit ] ;
		ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
		ex_msg_ptr = exgetcbuf( ex_ptr ) ;
		ex_msg_ptr -> mb_ns.ns_request = LLNET_STSTCS ;
		ex_msg_ptr -> mb_ns.ns_mask    = WRITE_OBJ ;
		ex_msg_ptr -> mb_ns.ns_rsrv    = 0 ;
                ex_msg_ptr -> mb_ns.ns_nobj    = 8 ;
                ex_msg_ptr -> mb_ns.ns_xobj    = 0 ;
                ex_msg_ptr -> mb_ns.ns_bufp    = 0 ;
		ex_msg_ptr -> mb_status        = MH_EXOS ;
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		exwaitbuf( ex_ptr, 0 ) ;
	        break ;

	default:
		error = EINVAL ;
		break ;
	}
	splx( s ) ;
	return ( error ) ;
}

/* end of external interfaces */

unsigned int
excopyaddr( ex_ptr )
	struct ex_softc *	ex_ptr ;
{
	struct ex_msg *		ex_msg_ptr = ex_ptr -> is_x2hnext ;

	uncache( & ex_msg_ptr -> mb_na.na_addrs[ 0 ] ) ;
	uncache( & ex_msg_ptr -> mb_na.na_addrs[ 5 ] ) ;
	bcopy( (caddr_t) & ex_msg_ptr -> mb_na.na_addrs[ 0 ],
	       (caddr_t) ex_ptr -> is_addr,
		sizeof( ex_ptr -> is_addr ) ) ;
}

exput( ifu, m0 )
	register u_char *ifu;
	register struct mbuf *m0;
{
	register struct mbuf *m = m0;
	register u_char *cp = ifu;

	while (m) {
		bcopy(mtod(m, u_char *), cp, m->m_len);
		cp += m->m_len;
		MFREE(m, m0);
		m = m0;
	}
	if ((int)cp & 1)
		*cp++ = 0;
	return (cp - ifu);
}

/*
 * This procedure will copy from the buffer in memory
 * into mbufs. hlen is the trailer length, it must be moved to the
 * front of the mbufs.
 */
struct mbuf *
exget( ptr, length, hlen, if_ptr )
	char *			ptr ;
	register int		length ;
	int			hlen ;
	struct ifnet *		if_ptr ;
{
	register char *		m_data_ptr ;
	register struct mbuf *	head_m_ptr ;
	register struct mbuf *	tail_m_ptr ;
	register struct mbuf *	m_ptr ;        /* mbuf    */
	register struct mbuf *	c_ptr ;        /* cluster */
	register int		size ;
	char *                  d_ptr ;

	mtpr( ptr, P1DC ) ;
	d_ptr = (char *)(((int)ptr & 0xfffffc00) + NBPG) ;
	while ( d_ptr < (char *)((int)ptr + length) ) {
	        mtpr( d_ptr, P1DC ) ;
	        d_ptr += NBPG ;
	}
	MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
	if (m_ptr == 0) return (0);
	head_m_ptr = tail_m_ptr = m_ptr ;
	while ( length > 0 ) {
		tail_m_ptr -> m_next = m_ptr ;
		if ( m_ptr == 0 ) {
			m_freem( head_m_ptr ) ;
			return( 0 ) ;
		}
		tail_m_ptr = m_ptr ;
		if (if_ptr)
			length += sizeof(if_ptr);
		size = MIN( length, MLEN ) ;
		if (length >= MCLBYTES) {
			MCLGET( m_ptr );
			if (m_ptr->m_len == MCLBYTES)
				size = MCLBYTES ;
		}
		m_ptr -> m_len = size ;
		length -= size ;
		m_data_ptr = mtod( m_ptr, char * ) ;
		if (if_ptr) {
			*(mtod(m_ptr, struct ifnet **)) = if_ptr;
			m_data_ptr += sizeof(if_ptr)/sizeof(*m_data_ptr);
			size -= sizeof(if_ptr);
			if_ptr = (struct ifnet *)0;
		}
		bcopy( ptr, m_data_ptr, size ) ;
		ptr += size ;
		if ( length )
			MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
	}
	tail_m_ptr -> m_next = 0 ;

	if ( hlen ) {
		MGET( m_ptr, M_DONTWAIT, MT_DATA ) ;
		if ( m_ptr == 0 ) {
			m_freem( head_m_ptr ) ;
			return( 0 ) ;
		}
		m_ptr -> m_len = hlen ;
		m_data_ptr = mtod( m_ptr, char * ) ;
		bcopy( ptr, m_data_ptr, size ) ;
		ptr += size ;
		m_ptr -> m_next = head_m_ptr ;
		head_m_ptr = m_ptr ;
	}
	return( head_m_ptr ) ;
}

/*
 * Print an error msg for error statuses.
 */
unsigned int
exmsgerr( request, reply, unit )
	unsigned int	request ;
	unsigned int	reply ;
	unsigned int	unit ;
{
	printf( "ex%x: request = 0x%x, ", unit, request ) ;
	switch ( reply ) {
	case LLX_MODE:
		printf( "EXOS not in link level mode\n" ) ;
		break ;
	case LLNA_BADSLOT:
		printf( "slot doesn't exist or can't be accessed\n");
		break ;
        case LLNR_BADADDR:
        	printf( "designated slot was empty\n" ) ;
        	break ;
       	case LLNA_BADADDR:
       		printf( "invalid address for designated slot\n" ) ;
       		break ;
       	default:
       		printf( "unknown reply %x\n", reply ) ;
       		break ;
	}
}

/*
 * Get a request buffer, fill in standard values, advance pointer.
 */
struct ex_msg *
exgetcbuf( ex_ptr )
	struct ex_softc *		ex_ptr ;
{
	register struct ex_msg *	ex_msg_ptr = ex_ptr -> is_h2xnext ;

	uncache( & ex_msg_ptr -> mb_status ) ;
	while ( (ex_msg_ptr -> mb_status & MH_OWNER) == MH_EXOS )
	        uncache( & ex_msg_ptr -> mb_status ) ;
	ex_msg_ptr -> mb_1rsrv  = 0 ;
	ex_msg_ptr -> mb_length = MBDATALEN ;
	ex_ptr -> is_h2xnext    = ex_msg_ptr -> mb_next ;
	return( ex_msg_ptr ) ;
}

/*
 * Wait for a buffer, clean up after it arrives.
 */
unsigned int
exwaitbuf( ex_ptr, proc_ptr )
	register struct ex_softc *	ex_ptr ;
	int				(*proc_ptr)( ) ;
{
	register struct ex_msg *	ex_msg_ptr = ex_ptr -> is_x2hnext ;
	register struct ex_reg *	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;

	uncache( & ex_msg_ptr -> mb_status ) ;
	while ( (ex_msg_ptr -> mb_status & MH_OWNER) == MH_EXOS )
	        uncache( & ex_msg_ptr -> mb_status ) ;
	uncache( & ex_msg_ptr -> mb_na.na_reply ) ;
	if ( ex_msg_ptr -> mb_na.na_reply != LL_OK ) {
		uncache( & ex_msg_ptr -> mb_na.na_request ) ;
		exmsgerr( ex_msg_ptr -> mb_na.na_request,
		          ex_msg_ptr -> mb_na.na_reply,
		          ex_ptr -> unit ) ;
	}
	if ( proc_ptr )
		(*proc_ptr)( ex_ptr ) ;
	ex_msg_ptr -> mb_length = MBDATALEN ;
	ex_msg_ptr -> mb_status = MH_EXOS ;	/* free up buffer */
	movow( EX_INTERRUPT,
	     & ex_reg_ptr -> ex_portb ) ;
	ex_ptr -> is_x2hnext = ex_msg_ptr -> mb_next ;
	return( TRUE ) ;
}

/*
 * Reset, test, and configure EXOS.  It is called by exinit, and exattach.
 * Returns 0 if successful, 1 if self-test failed.
 */
int
exconfig( ui, itype )
	struct	vba_device *		ui ;
	int				itype ;
{
	register int			unit = ui -> ui_unit ;
	register struct ex_softc *	ex_ptr = & ex_softc[ unit ] ;
	register struct ex_reg *	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
	register struct ex_msg *	ex_msg_ptr ;
	struct ex_queue_base *		ex_qb_ptr ;
	struct ex_config *		ex_cm_ptr ;
	int				i ;
	unsigned int  			shiftreg ;
	static	unsigned int 		cmaddr[ 8 ] =
		{ 0xFF, 0xFF, 0, 0, 0, 0, 0, VME_ADDRESS_MODIFIER } ;

	ex_ptr -> is_flags = 0 ;
	/*
	 * Reset EXOS, wait for self-test to complete
	 */
  	movow( EX_RESET,
	     & ex_reg_ptr -> ex_porta ) ;
	while ( (ex_reg_ptr -> ex_portb & EX_TESTOK) == 0 )
	        DELAY( 200000 ) ;

	/*
	 * Set up the queue base pointer.
	 */
	ex_qb_ptr = (struct ex_queue_base *)
		((unsigned int)& ex_ptr -> queue_base & (unsigned int)0xfffffff0) ;
	ex_ptr -> ex_qb_ptr = ex_qb_ptr ;

	/*
	 * Set up configuration message.
	 */
	ex_cm_ptr = & ex_qb_ptr -> cm ;
	ex_cm_ptr -> cm_0rsrv       = 0x0001 ;
	ex_cm_ptr -> cm_cc          = 0xFF ;
	ex_cm_ptr -> cm_opmode      = 0 ;	/* link-level controller mode */
	ex_cm_ptr -> cm_dfo         = 0x0101 ;
	ex_cm_ptr -> cm_1rsrv       = 0 ;
	ex_cm_ptr -> cm_context     = 0 ;
	ex_cm_ptr -> cm_2rsrv       = 0 ;
	ex_cm_ptr -> cm_ham         = 3 ;	/* absolute address mode */
	ex_cm_ptr -> cm_3rsrv       = 0 ;
	ex_cm_ptr -> cm_mapsiz      = 0 ;
	ex_cm_ptr -> cm_byteptrn[0] = 0x01 ;
	ex_cm_ptr -> cm_byteptrn[1] = 0x03 ;
	ex_cm_ptr -> cm_byteptrn[2] = 0x07 ;
	ex_cm_ptr -> cm_byteptrn[3] = 0x0F ;
	ex_cm_ptr -> cm_wordptrn[0] = 0x0103 ;
	ex_cm_ptr -> cm_wordptrn[1] = 0x070F ;
	ex_cm_ptr -> cm_lwordptrn   = 0x0103070F ;
	for ( i = 0 ; i < 20 ; i ++ )
		ex_cm_ptr -> cm_rsrvd[ i ] = 0 ;
	ex_cm_ptr -> cm_mba = 0xFFFFFFFF ;
	ex_cm_ptr -> cm_nproc = 0xFF ;
	ex_cm_ptr -> cm_nmbox = 0xFF ;
	ex_cm_ptr -> cm_nmcast = 0 ;
	ex_cm_ptr -> cm_nhost = 1 ;
	ex_cm_ptr -> cm_h2xba = PHYS_ADDR( ex_qb_ptr ) +
				(VME_ADDRESS_MODIFIER << 24 ) ;
	ex_cm_ptr -> cm_h2xhdr = fldoff( ex_queue_base, h2xhdr ) ;
	ex_cm_ptr -> cm_h2xtyp = 0 ;
	ex_cm_ptr -> cm_x2hba = ex_cm_ptr -> cm_h2xba ;
	ex_cm_ptr -> cm_x2hhdr = fldoff( ex_queue_base, x2hhdr ) ;
	ex_cm_ptr -> cm_x2htyp = itype ;	/* 0 for none, 4 for vectored */
	ex_cm_ptr -> cm_x2haddr = ui -> ui_mi -> um_ivct ;

	/*
	 * Set up message queues and headers: first the request queue
	 */
	ex_ptr -> is_h2xnext = & ex_qb_ptr -> h2xmsg[ 0 ] ;
	ex_qb_ptr -> h2xhdr =  (unsigned short)((unsigned int)(ex_ptr -> is_h2xnext) -
				       (unsigned int)(ex_qb_ptr)) ;
	for ( ex_msg_ptr = ex_ptr -> is_h2xnext ;
              ex_msg_ptr < & ex_qb_ptr -> h2xmsg[ NH2X ] ;
              ex_msg_ptr ++ ) {	
		ex_msg_ptr -> mb_link =
			(unsigned short)((unsigned int)(ex_msg_ptr+1)-(unsigned int)ex_qb_ptr) ;
		ex_msg_ptr -> mb_rsrv = 0 ;	
		ex_msg_ptr -> mb_length = MBDATALEN ;	
		ex_msg_ptr -> mb_status = MH_HOST ;	
		ex_msg_ptr -> mb_next = ex_msg_ptr + 1 ;
	}
	ex_qb_ptr -> h2xmsg[ NH2X - 1 ].mb_next = ex_ptr -> is_h2xnext ;
	ex_qb_ptr -> h2xmsg[ NH2X - 1 ].mb_link = ex_qb_ptr -> h2xhdr ;

	/* now the reply queue */
	ex_ptr -> is_x2hnext = & ex_qb_ptr -> x2hmsg[ 0 ] ;
	ex_qb_ptr -> x2hhdr =  (unsigned short)((unsigned int)(ex_ptr -> is_x2hnext) -
				       (unsigned int)(ex_qb_ptr)) ;
	for ( ex_msg_ptr = ex_ptr -> is_x2hnext ;
              ex_msg_ptr < & ex_qb_ptr -> x2hmsg[ NX2H ] ;
              ex_msg_ptr ++ ) {
		ex_msg_ptr -> mb_link =
			(unsigned short)((char *)(ex_msg_ptr+1)-(unsigned int)ex_qb_ptr) ;
		ex_msg_ptr -> mb_rsrv = 0 ;	
		ex_msg_ptr -> mb_length = MBDATALEN ;	
		ex_msg_ptr -> mb_status = MH_EXOS ;	
		ex_msg_ptr -> mb_next = ex_msg_ptr + 1 ;
	}
	ex_qb_ptr -> x2hmsg[ NX2H - 1 ].mb_next = ex_ptr -> is_x2hnext ;
	ex_qb_ptr -> x2hmsg[ NX2H - 1 ].mb_link = ex_qb_ptr -> x2hhdr ;
	/*
	 * Write config msg address to EXOS and wait for configuration to 
	 * complete (guaranteed response within 2 seconds).
	 */
	shiftreg = PHYS_ADDR( & ex_qb_ptr -> cm ) ;
	for ( i = 4 ; i < 7 ; i ++ ) {
		cmaddr[ i ] = (u_char) (shiftreg & 0xFF) ;
		shiftreg >>= 8 ;
	}
	for ( i = 0 ; i < 8 ; i ++ ) {
		while ( ex_reg_ptr -> ex_portb & EX_UNREADY )
		        DELAY( 200000 ) ;
		DELAY( 2000 ) ;
		movow( cmaddr[ i ],
		     & ex_reg_ptr -> ex_portb ) ;
	}
	DELAY( 1000000 ) ;
	uncache( & ex_cm_ptr -> cm_cc ) ;
	if ( ex_cm_ptr -> cm_cc ) {
		printf( "ex%d: configuration failed ; cc = %x\n\n",
			unit, ex_cm_ptr -> cm_cc ) ;
		return(  FALSE ) ;
	}
	return( TRUE ) ;
}

/*
 * Start or re-start output on interface. Get another datagram to send off of 
 * the interface queue, and map it to the interface before starting the output.
 * This routine is called by exinit(), exoutput(), and excdint().  In all cases,
 * interrupts by EXOS are disabled.
 */
int
exstart( unit )
	int				unit ;
{
	struct vba_device *		ui     = exinfo[ unit ] ;
	register struct ex_softc *	ex_ptr = & ex_softc[ unit ] ;
	register struct ex_reg *	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
	register struct ex_msg *	ex_msg_ptr ;
	register struct mbuf *		m_ptr ;
        int				len ;
	struct mbuf *			mbuf_ptr = (struct mbuf *)0 ;
	register int			nb   = 0 ;
	register int			tlen = 0 ;
	unsigned int			addr = 0 ;
	unsigned int			temp ;

	IF_DEQUEUE( & ex_ptr -> is_if.if_snd, m_ptr ) ;
	if ( m_ptr == 0 )
		return ;
	/*
	 * Get a transmit request.
	 */
	ex_msg_ptr = exgetcbuf( ex_ptr ) ;	
	ex_msg_ptr -> mb_et.et_request = LLTRANSMIT ;

	/*
	 * point directly to the first group of mbufs to be transmitted. The
	 * hardware can only support NFRAGMENTS descriptors.
	 */
	ex_msg_ptr -> mb_mid = 0 ;
	while ( m_ptr && !((unsigned int)m_ptr & (unsigned int)0x3f000000) &&
			((nb < NFRAGMENTS-1) || (m_ptr -> m_next == 0)) ) {
		addr = mtod( m_ptr, int ) ;
		ex_msg_ptr -> mb_et.et_blks[nb].bb_len = (unsigned short)m_ptr -> m_len ;
		temp = PHYS_ADDR( addr ) + (VME_ADDRESS_MODIFIER << 24) ;
		ex_msg_ptr -> mb_et.et_blks[nb].bb_addr[0]=(unsigned short) (temp >> 16);
		ex_msg_ptr -> mb_et.et_blks[nb].bb_addr[1]=(unsigned short)  temp ;
		tlen += m_ptr -> m_len ;
		mbuf_ptr = m_ptr ;
		if ( ex_msg_ptr -> mb_mid == 0 )
			ex_msg_ptr -> mb_mid = (int) m_ptr ;
		m_ptr = m_ptr -> m_next ;
		nb ++ ;
	}

	/* null end of chain pointed to by mb_mid, to be freed when xmit done */
	if ( mbuf_ptr )
		mbuf_ptr -> m_next = 0 ;

	/*
	 * if not all of the descriptors would fit then merge remaining data
	 * into the transmit buffer, and point to it.  Note: the mbufs are freed
	 * during the merge, they do not have to be freed when we get the 
	 * transmit interrupt.
	 */
	if ( m_ptr ) {
		len = exput( & ex_ptr -> ex_tbuf[ 0 ], m_ptr ) ;
		ex_msg_ptr -> mb_et.et_blks[ nb ].bb_len = (unsigned short) len ;
		temp = PHYS_ADDR( & ex_ptr -> ex_tbuf[ 0 ] ) +
		       (VME_ADDRESS_MODIFIER << 24) ;
		ex_msg_ptr -> mb_et.et_blks[nb].bb_addr[0]=(unsigned short) (temp >> 16);
		ex_msg_ptr -> mb_et.et_blks[nb].bb_addr[1]=(unsigned short)  temp ;
		tlen += len ;
		nb ++ ;
	}

	/*
	 * If the total length of the packet is too small, pad the last frag
	 */
	if ( tlen - sizeof( struct ether_header ) < ETHERMIN ) {
		len = (ETHERMIN + sizeof( struct ether_header )) - tlen ;
		ex_msg_ptr -> mb_et.et_blks[ nb - 1 ].bb_len += (unsigned short) len ;
		tlen += len ;
	}

#ifdef EX_MONITOR
	{
	register unsigned int   ptr ;
	ex_ptr -> debug_ptr -> counter     = ex_ptr -> counter ++ ;
	ex_ptr -> debug_ptr -> direction   = EX_XMIT ;
	ex_ptr -> debug_ptr -> cmnd_status = tlen ;
	ptr  = (unsigned int)(ex_msg_ptr -> mb_et.et_blks[ 0 ].bb_addr[0]) << 16 ;
	ptr |= (unsigned int)(ex_msg_ptr -> mb_et.et_blks[ 0 ].bb_addr[1]) ;
	ptr &= 0x00ffffff ;
	ptr |= 0xc0000000 ;
	bcopy( ptr,
	       ex_ptr -> debug_ptr -> hdr_dhost,
	       min( ex_msg_ptr -> mb_et.et_blks[ 0 ].bb_len, 74 ) ) ;
	if ( ++ ex_ptr -> debug_ptr >= ex_ptr -> end_debug_ptr )
		ex_ptr -> debug_ptr = ex_ptr -> start_debug_ptr ;
#endif EX_MONITOR

	/* set number of fragments in descriptor */
	ex_msg_ptr -> mb_et.et_nblock = nb ;
	ex_ptr -> is_flags |= EX_XPENDING ;
	ex_msg_ptr -> mb_status = MH_EXOS ;
	movow( EX_INTERRUPT,
	     & ex_reg_ptr -> ex_portb ) ;
}

/*
 * interrupt service routine.
 */
int
exintr( unit )
	int				unit ;
{
	register struct ex_softc *	ex_ptr = & ex_softc[ unit ] ;
	register struct ex_msg *	ex_msg_ptr = ex_ptr -> is_x2hnext ;
	struct vba_device *		ui = exinfo[ unit ] ;
	register struct ex_reg *	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;

	uncache( & ex_msg_ptr -> mb_status ) ;
	while ( (ex_msg_ptr -> mb_status & MH_OWNER) == MH_HOST ) {
		uncache( & ex_msg_ptr -> mb_na.na_request ) ;
		switch ( ex_msg_ptr -> mb_na.na_request ) {
		case LLRECEIVE:
			if ( (ex_ptr -> is_flags & EX_RPENDING) == 0 ) {
				printf( "ex%x: exintr: no recv pending\n",unit);
	                        break ;
			}
			ex_ptr -> is_flags &= ~EX_RPENDING ;
			exrecv( unit, ex_msg_ptr ) ;
			exhangrcv( unit ) ;
			break ;

		case LLTRANSMIT:
		case LLRTRANSMIT:
			if ( (ex_ptr -> is_flags & EX_XPENDING) == 0 ) {
				printf( "ex%x: exintr: no xmit pending\n",unit);
	                        break ;
			}
			ex_ptr -> is_flags &= ~EX_XPENDING ;
			ex_ptr -> is_if.if_opackets ++ ;
			uncache( & ex_msg_ptr -> mb_et.et_reply ) ;
			if (ex_msg_ptr -> mb_et.et_reply == LL_OK) ;
			else if ( ex_msg_ptr -> mb_et.et_reply & LLXM_1RTRY )
				ex_ptr -> is_if.if_collisions ++ ;
			else if ( ex_msg_ptr -> mb_et.et_reply & LLXM_RTRYS )
				ex_ptr -> is_if.if_collisions += 2 ;
			else if ( ex_msg_ptr -> mb_et.et_reply & LLXM_ERROR )
				ex_ptr -> is_if.if_oerrors ++ ;
			uncache( & ex_msg_ptr -> mb_mid ) ;
			if ( ex_msg_ptr -> mb_mid )
				m_freem( (struct mbuf *) ex_msg_ptr -> mb_mid) ;
			exstart( unit ) ;
			break ;

		default:
			uncache( & ex_msg_ptr -> mb_na.na_request ) ;
			printf("ex%d: unknown reply 0x%x\n",
				unit, ex_msg_ptr -> mb_na.na_request) ;
		}
		ex_msg_ptr -> mb_length = MBDATALEN ;	
		ex_msg_ptr -> mb_status = MH_EXOS ;	/* free up buffer */
		movow( EX_INTERRUPT,
		     & ex_reg_ptr -> ex_portb ) ;
		ex_msg_ptr = ex_msg_ptr -> mb_next ;
		uncache( & ex_msg_ptr -> mb_status ) ;
	}
	ex_ptr -> is_x2hnext = ex_msg_ptr ;
}

/*
 * Process Ethernet receive completion:  If input error just drop packet, 
 * otherwise examine packet to determine type.  If can't determine length from 
 * type, then have to drop packet, otherwise decapsulate packet based on type
 * and pass to type-specific higher-level input routine.
 */
int
exrecv( unit, ex_msg_ptr )
	int 					unit ;
	register struct ex_msg * 	ex_msg_ptr ;
{
	register struct ex_softc *	ex_ptr = & ex_softc[ unit ] ;
	register struct ether_header *		ex_hdr_ptr ;
    	register struct mbuf *			m_ptr = 0 ;
	int					len ;
	int					off ;
	int					resid ;
	register struct ifqueue *		inq ;
	int					s ;

	ex_ptr -> is_if.if_ipackets ++ ;
	uncache( & ex_msg_ptr -> mb_er.er_blks[ 0 ] ) ;
	len = ex_msg_ptr -> mb_er.er_blks[ 0 ].bb_len -
		sizeof( SZ_ETHER_HDR ) - 4 ;
	uncache( & ex_msg_ptr -> mb_er.er_reply ) ;
	if ( ex_msg_ptr -> mb_er.er_reply != LL_OK )
		goto error_return ;
	ex_hdr_ptr = (struct ether_header *) (ex_ptr -> ex_rbuf) ;

	/*
	 * Deal with trailer protocol: if type is PUP trailer get true type from
	 * first 16-bit word past data.  Remember that type was trailer by 
	 * setting off.
	 */
	ex_hdr_ptr -> ether_type = ntohs( (unsigned short) ex_hdr_ptr -> ether_type ) ;

#define	exdataaddr(ex, off, type)	((type)(((caddr_t)((ex)+1)+(off))))

	if (ex_hdr_ptr -> ether_type >= ETHERTYPE_TRAIL &&
	    ex_hdr_ptr -> ether_type < ETHERTYPE_TRAIL+ETHERTYPE_NTRAILER) {
		off = (ex_hdr_ptr -> ether_type - ETHERTYPE_TRAIL) * 512 ;
		if ( off >= ETHERMTU )
			goto error_return ;	/* sanity */
		ex_hdr_ptr -> ether_type =
			ntohs( *exdataaddr( ex_hdr_ptr, off, unsigned short * ) ) ;
		resid = ntohs( *(exdataaddr( ex_hdr_ptr, off + 2, unsigned short * ) ) ) ;
		if ( off + resid > len )
			goto error_return ;	/* sanity */
		len = off + resid ;
	} else
		off = 0 ;
	if ( len == 0 )
		goto error_return ;

	/*
	 * Pull packet off interface.  Off is nonzero if packet has trailing 
	 * header ; exget will then force this header information to be at
	 * the front, but we still have to drop the type and length which are 
	 * at the front of any trailer data.
 	 */
	m_ptr = exget( (int)ex_ptr -> ex_rbuf + SZ_ETHER_HDR,
	               len, off, &ex_ptr->is_if ) ;
	if ( m_ptr == 0 )
		goto error_return ;
	if ( off ) {
		m_ptr -> m_off += 2 * sizeof (unsigned short) ;
		m_ptr -> m_len -= 2 * sizeof (unsigned short) ;
	}

#ifdef EX_MONITOR
	ex_ptr -> debug_ptr -> counter     = ex_ptr -> counter ++ ;
	ex_ptr -> debug_ptr -> direction   = EX_RECV ;
  	ex_ptr -> debug_ptr -> cmnd_status = len ;
  	bcopy( ex_ptr -> ex_rbuf,
	       ex_ptr -> debug_ptr -> hdr_dhost,
	       74 ) ;
 	if ( ++ ex_ptr -> debug_ptr >= ex_ptr -> end_debug_ptr )
		ex_ptr -> debug_ptr = ex_ptr -> start_debug_ptr ;
#endif EX_MONITOR

	switch ( ex_hdr_ptr -> ether_type ) {

#ifdef INET
	case ETHERTYPE_IP:
		schednetisr( NETISR_IP ) ;
		inq = & ipintrq ;
		if ( IF_QFULL( inq ) ) {
			IF_DROP( inq ) ;
			goto error_return ;
		}
		s = splimp( ) ;
		IF_ENQUEUE( inq, m_ptr ) ;
		splx( s ) ;
		break ;

	case ETHERTYPE_ARP:
		arpinput( & ex_ptr -> is_ac, m_ptr ) ;
		break ;
#endif
	default:
error_return:
		ex_ptr -> is_if.if_ierrors ++ ;
		if ( m_ptr )
			m_freem( m_ptr ) ;
	}
}

/*
 * Hang a receive request. This routine is called by exinit and excdint,
 * with interrupts disabled in both cases.
 */
int
exhangrcv( unit )
	int				unit ;
{
	register struct ex_softc *	ex_ptr     = & ex_softc[ unit ] ;
	register struct ex_msg *	ex_msg_ptr ;
	register struct ex_reg *	ex_reg_ptr = ex_ptr -> ex_reg_ptr ;
	register int                   temp ;

	if ( ex_ptr -> is_flags & EX_RPENDING )
		return ;
	ex_ptr -> is_flags |= EX_RPENDING ;
	ex_msg_ptr = exgetcbuf( ex_ptr ) ;
	ex_msg_ptr -> mb_er.er_request = LLRECEIVE ;
	ex_msg_ptr -> mb_er.er_nblock = 1 ;
	ex_msg_ptr -> mb_er.er_blks[0].bb_len = EXMAXRBUF ;
	temp = PHYS_ADDR(ex_ptr -> ex_rbuf) + (VME_ADDRESS_MODIFIER << 24 ) ;
	ex_msg_ptr -> mb_er.er_blks[ 0 ].bb_addr[ 0 ] = (unsigned short) (temp >> 16) ;
	ex_msg_ptr -> mb_er.er_blks[ 0 ].bb_addr[ 1 ] = (unsigned short)  temp ;
	ex_msg_ptr -> mb_status = MH_EXOS ;
	movow( EX_INTERRUPT,
	     & ex_reg_ptr -> ex_portb ) ;
}
