/*
 * Copyright (c) 1984 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char sccsid[] = "@(#)machdep.c	1.7 (Berkeley) 8/22/85";
#endif not lint

#include <stdio.h>
#include <ctype.h>
#include "inline.h"

extern char *strcpy();
extern char *strcat();
extern char *index();

/*
 * The routines and tables in this file must be rewritten
 * for each new machine that this program is ported to.
 */

/*
 * Instruction stop table.
 * All instructions that implicitly modify any of the temporary
 * registers, change control flow, or implicitly loop must be
 * listed in this table. It is used to find the end of a basic
 * block when scanning backwards through the instruction stream
 * trying to merge the inline expansion.
 */
struct inststoptbl inststoptable[] = {
	{ "bbc" }, { "bbs" },
	{ "brb" }, { "brw" }, { "jmp" }, { "bbssi" }, { "bcc" }, { "bcs" },
	{ "beql" }, { "beqlu" }, { "bgeq" }, { "bgequ" },
	{ "bgtr" }, { "bgtru" }, { "bleq" }, { "blequ" },
	{ "blss" }, { "blssu" }, { "bneq" }, { "bnequ" },
	{ "bvc" }, { "bvs" },
	{ "casel" }, { "aoblss" }, { "aobleq" },
	{ "calls" }, { "callf" }, { "ret" },
	{ "movblk" }, { "movs2" }, { "movs3" }, { "cmps2" }, { "cmps3" },
	{ "kcall" }, { "rei" }, { "ldpctx" }, { "svpctx" },
	{ "halt" }, { "bpt" }, { "storer" }, { "loadr" },
/* some of the following might not be needed, but just in case ... - HO */
	{ "ediv" }, { "emul" }, { "shlq" }, { "shrq" }, { "std" },
	{ "" }
};

/*
 * Invalid instruction table.
 * All instructions that should not be generated must be listed in this
 * table.  For the HCX-7, this table contains any instructions which might
 * manipulate the accumulator or set/clear the DBL bit in the PSL.
 */
struct instinvtbl instinvtable[] = {
	{ "addd" }, { "addf" }, { "atanf" }, { "bicpsw" }, { "bispsw" },
	{ "btcs" }, { "cmpd" }, { "cmpf" }, { "cmpd2" }, { "cmpf2" },
	{ "cosf" }, { "cvdf" }, { "cvdl" }, { "cvfl" }, { "cvld" },
	{ "cvlf" }, { "divd" }, { "divf" }, { "expf" }, { "ldd" }, { "ldf" },
	{ "ldfd" }, { "lnd" }, { "lnf" }, { "logf" }, { "movpsl" },
	{ "muld" }, { "mulf" }, { "negd" }, { "negf" }, { "pushd" },
	{ "sinf" }, { "sqrtf" }, { "std" }, { "stf" }, { "subd" },
	{ "subf" }, { "tstd" }, { "tstf" },
	{ "" }
};

/*
 * Check to see if a line is a candidate for replacement.
 * Return pointer to name to be looked up in pattern table.
 * This function also ensures that the instruction is valid.
 */
char *
doreplaceon(cp)
	register char *cp;
{

	if (isinvinst(cp)) {
		(void) fprintf(stderr, "%s %d\n",
			"invalid instruction in input file near line", lineno);
		exit (1);
	}

	if (bcmp(cp, "callf\t", 6) != 0)
		return (0);
	if ((cp = index(cp + 6, ',')) == 0)
		return (0);
	return (++cp);
}

/*
 * Check for instructions that should not exist (invalid instructions).
 */
isinvinst(cp)
	register char *cp;
{
	char buf[100];		/* the instruction, NULL terminated */
	char *bufptr;		/* next location in 'buf' */
	register char *bufend;	/* last location in 'buf' which is useable */
	register struct instinvtbl *iip;
	int size;

	/*
	 * copy the instruction to a local buffer
	 */
	bufptr = buf;
	while ((*cp != '\t') && (*cp != '\n') && (*cp != '\0')) {
		if (bufptr >= &buf[99])
			return (1);
		*bufptr++ = *cp++;
	}
	*bufptr = '\0';

	/*
	 * check to see if the instruction is valid
	 */
	for (iip = instinvtblhdr[hash(buf, &size)]; iip; iip = iip->next)
		if (iip->size == size && bcmp(buf, iip->name, size) == 0)
			return (1);
	return (0);
}

/*
 * Find out how many arguments the function is being called with.
 * A return value of -1 indicates that the count can't be determined.
 */
int
countargs(cp)
	register char *cp;
{
	int tmp;

	if ((cp = index(cp, '$')) == 0)
		return (-1);
	if (!isdigit(*++cp))
		return (-1);
	tmp = atoi(cp);
	return ((tmp-4)/4);
}

/*
 * Find the next argument to the function being expanded.
 * See 'modifies()' for registers that we care about.
 */
nextarg(argc, argv)
	int argc;
	char *argv[];
{
	register char *lastarg = argv[2];

	if (argc == 3 &&
	    bcmp(argv[0], "mov", 3) == 0 &&
	    bcmp(argv[1], "(sp)+", 6) == 0 &&
	    lastarg[0] == 'r' && isdigit(lastarg[1]) && lastarg[2] == '\0')
		return (lastarg[1] - '0');
	return (-1);
}

/*
 * Determine whether the current line pushes an argument.
 */
ispusharg(argc, argv)
	int argc;
	char *argv[];
{

	if (argc < 2)
		return (0);
	if (argc == 2 && bcmp(argv[0], "push", 4) == 0)
		return (1);
	if (bcmp(argv[argc - 1], "-(sp)", 6) == 0)
		return (1);
	return (0);
}

/*
 * Determine which (if any) registers are modified
 * Return register number that is modified, -1 if none are modified.
 * This algorithm only worries about registers 0 to 9.  This is sufficient
 * for the HCX, as none of the patterns are expeceted to use so many registers.
 */
modifies(argc, argv)
	int argc;
	char *argv[];
{
	register char *lastarg = argv[argc - 1];

	if (lastarg[0] == 'r' && isdigit(lastarg[1]) && lastarg[2] == '\0')
		return (lastarg[1] - '0');
	return (-1);
}

/*
 * Rewrite the instruction in (argc, argv) to store its
 * contents into arg instead of onto the stack. The new
 * instruction is placed in the buffer that is provided.
 */
rewrite(instbuf, argc, argv, target)
	char *instbuf;
	int argc;
	char *argv[];
	int target;
{

	switch (argc) {
	case 0:
		instbuf[0] = '\0';
		(void) fprintf(stderr, "blank line to rewrite?\n");
		return;
	case 1:
		(void) sprintf(instbuf, "\t%s\n", argv[0]);
		(void) fprintf(stderr, "rewrite?-> %s", instbuf);
		return;
	case 2:
		if (bcmp(argv[0], "push", 4) == 0) {
			(void) sprintf(instbuf, "\tmov%s\t%s,r%d\n",
				&argv[0][4], argv[1], target);
			return;
		}
		(void) sprintf(instbuf, "\t%s\tr%d\n", argv[0], target);
		return;
	case 3:
		(void) sprintf(instbuf, "\t%s\t%s,r%d\n", argv[0], argv[1], target);
		return;
	case 4:
		(void) sprintf(instbuf, "\t%s\t%s,%s,r%d\n",
			argv[0], argv[1], argv[2], target);
		return;
	case 5:
		(void) sprintf(instbuf, "\t%s\t%s,%s,%s,r%d\n",
			argv[0], argv[1], argv[2], argv[3], target);
		return;
	default:
		(void) sprintf(instbuf, "\t%s\t%s", argv[0], argv[1]);
		argc -= 2, argv += 2;
		while (argc-- > 0) {
			(void) strcat(instbuf, ",");
			(void) strcat(instbuf, *argv++);
		}
		(void) strcat(instbuf, "\n");
		(void) fprintf(stderr, "rewrite?-> %s", instbuf);
		return;
	}
}

/*
 * Do any necessary post expansion cleanup.
 */
/*ARGSUSED*/
cleanup(numargs)
	int numargs;
{

	return;
}
