#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/dir.h"
#include "../sys/user.h"
#include "../sys/map.h"
#include "../sys/ioctl.h"
#include "../sys/tty.h"
#include "../sys/proc.h"
#include "../sys/file.h"
#include "../sys/conf.h"
#include "../sys/buf.h"
#include "../sys/dk.h"
#include "../sys/uio.h"
#include "../sys/kernel.h"
#include "../sys/alm.h"
#include "../bop/bopdebug.h"

#ifdef BOP_DEBUG
#include "../bop/sdlcbuf.h"
#endif

/* Defines for almflg */
#define ALM_OPEN	01
#define ALM_SLEEP	02
#define ALM_OFLOW	04
#define ALM_READ	010
#define ALM_STOP	020
#define ALM_EXIT	040

struct alm alm[SZALMQ];

int almcnt = 0;	/* next alarm blk to use */
int almrd  = 0;	/* next alarm blk to read */
int almflg = ALM_STOP;

alm_open(dev)
dev_t dev;
{
	short s;

	if( minor(dev) != 0)
		return 0;
	s = spl8();
	if(almflg & ALM_OPEN)
		{
		splx(s);
		return EBUSY;
		}
	almflg |= ALM_OPEN;
	splx(s);
	return 0;
}

alm_close(dev)
dev_t dev;
{
	short s;

	if( minor(dev) != 0)
		return 0;
	s = spl8();
	almflg &= ~ALM_OPEN;
	splx(s);
	return 0;
}

alm_read(dev,uio)
dev_t dev;
struct uio *uio;
{
	short s;
	register struct alm *al;

	if( minor(dev) != 0) 
		return EINVAL;
	s = spl8();
	if(almflg & ALM_EXIT) 
		{
		almflg &= ~ALM_EXIT;
		splx(s);
		return EXDEV;
		}
	while(((almcnt==almrd) && !(almflg&ALM_OFLOW)) || (almflg&ALM_STOP)) {
		almflg |= ALM_SLEEP;
		while(almflg & ALM_SLEEP)
			slumber((caddr_t)alm,INTERRUPTABLE,selboost);
		if(almflg & ALM_EXIT) {
			almflg &= ~ALM_EXIT;
			splx(s);
			return EXDEV;
		}
	}
	almflg |= ALM_READ;
	al = &alm[almrd++];
	if(almrd == SZALMQ)
		almrd = 0;
	splx(s);
	uiomove((caddr_t)al,sizeof(struct alm),UIO_READ,uio);
	/* Zero alm blk */
	al->a_type = 0;
	al->a_sev = 0;
	al->a_missed = 0;
	al->a_time = 0L;
	al->a_string[0] = '\0';
	almflg &= ~(ALM_OFLOW|ALM_READ);
return 0;
}

alm_ioctl(dev, cmd, arg, mode)
dev_t dev;
caddr_t arg; 
int cmd, mode;
{
	short s;

	cmd &= 0xf;
	if( minor(dev) == 0) { 
		return EINVAL;
	}
	s = spl8();
	switch (cmd) {
	case ALMON:
		almflg &= ~ALM_STOP;
		if(almflg & ALM_SLEEP) {
			almflg &= ~ALM_SLEEP;
			wakeup((caddr_t)alm);
		}
		break;

	case ALMOFF:
		almflg |= ALM_STOP;
		break;

	case ALMEXIT:
		almflg |= ALM_EXIT;
		if(almflg & ALM_SLEEP) {
			almflg &= ~ALM_SLEEP;
			wakeup((caddr_t)alm);
		}
		break;

	default:
		u.u_error = EINVAL;
		break;
	}
	splx(s);
return 0;
}

alarmk(type,sev,string,val)
int type, sev;
register char *string;
unsigned long *val;
{
	register struct alm *al;
	register char *ptr;
	register int i;
	short s;

	s = spl8();
	if((almcnt == (almrd-1)) || ((almrd==0) && (almcnt == (SZALMQ-1)))) {
		if(almflg & ALM_READ)
			goto tossit;
		else
			almflg |= ALM_OFLOW;
	}
	if((almcnt == almrd) && (almflg & ALM_OFLOW))
		goto tossit;
	al = &alm[almcnt++];
	if(almcnt == SZALMQ)
		almcnt = 0;

	/* copy data into alarm blk */
	ptr = al->a_string;
	while(*string && ptr < al->a_string + sizeof(al->a_string)-1)
		*ptr++ = *string++;
	*ptr = '\0';
	al->a_sev = sev;
	al->a_type = type;
	al->a_missed = 0;
	al->a_time = time.tv_sec;
	if(val == (unsigned long *)0)
		al->a_type |= AL_NOARGS;
	else
		for(i=0;i<SZALMVAL;i++)
			al->a_val[i] = *val++;
	if((almflg & ALM_SLEEP) && !(almflg & ALM_STOP)) {
		almflg &= ~ALM_SLEEP;
		wakeup((caddr_t)alm);
	}
	splx(s);
	return 0;
tossit:
	if(almcnt == 0)
		alm[SZALMQ-1].a_missed++;
	else
		alm[almcnt-1].a_missed++;
	splx(s);
}
