/* 	        REAL-TIME CLOCK DEVICE DRIVER			*/

/*	rt.c		HCX version 1.0		6/30/87		*/


#include	"../sys/types.h"
#include	"../sys/param.h"
#include	"../sys/file.h"
#include	"../sys/buf.h"
#include	"../sys/dir.h"
#include	"../sys/user.h"
#include	"../sys/uio.h"
#include	"../sys/vbavar.h"
#include	"../sys/kernel.h"
#include	"../sys/errno.h"

#include	"../sys/rtc.h"
#include	"../sys/rtcmacros.h"

/* #define	RTC_DEBUG 	*/

extern	rtcprobe();
extern	rtcslave();
extern	rtcattach();

extern int	nrtc;
extern int	nclock;

extern struct vba_ctlr    *rtc_vba_ctlr[];
extern struct vba_device  *rtc_vba_device[];
extern struct rtclock	  rtc_tab_info[];


struct vba_driver  rtcdriver =
{
	rtcprobe,
	rtcslave,
	rtcattach,
	0,			/* dgo routine (unused) */
	0,			/* device addresses (unused) */
	"real-time clock ",
	rtc_vba_device,
	"rtc ",
	rtc_vba_ctlr,
	0,			/* HVME board id (unused) */
	0			/* exclusive use of bdp's (unused) */
};	  



/*
 * 	Rtcattach is called by configure() whenever rtcslave() returns TRUE.
 *	All of the device initialization code is done in rtcslave(), therefore,
 *	return immediately.
 */

int	
rtcattach( dev_info )
     struct vba_device  *dev_info;
	
{
	return;
}



/*
 *	Rtclose stops the clock from counting and cleans up the data
 *	structure associated with this clock.
 */

int
rtcclose( dev, flags )
     register dev_t  dev;
     int  flags;

{
	register struct rtclock  *clock =
		&rtc_tab_info[RTC_UNIT(dev)];
	register struct rtc_regs  *dev_regs =
		(struct rtc_regs *) rtc_vba_ctlr[RTC_CTLR(dev)]->um_addr;

	
	STOP_CLOCK( dev_regs, RTC_CLOCK(dev) );
	
	clock->c_flags &= RTC_CONFIG;		/* reset flags */
	spl7();
	clock->c_pwait = 0;
	clock->c_counting = 0;
	clock->c_repeat = 0;
	spl0();

	bzero((caddr_t) &clock->c_rtc, sizeof (struct rtc));

	return ( SUCCESS );
}



/*
 *	Rtcintr wakes up any processes sleeping on the clock that
 *	generated the interrupt. It also marks the clock as not
 *	counting if the clock is not in repetitive count mode.  
 */

void
rtcintr( mdev )
     register int  mdev;	/* minor device number */

{
	register struct rtclock  *clock = &rtc_tab_info[mdev];

#ifdef RTC_DEBUG
	printf("rtcintr: minor dev #%d\n", mdev);
#endif	
	
	/*
	 * Wakeup any processes that are sleeping on this clock.
	 */
	if (clock->c_pwait) {
		wakeup ( &clock->c_pwait );
		clock->c_pwait = 0;
	}

	/*
	 * Mark clock as not counting if clock is not in repetitive 
	 * count mode.
	 */
	if ( !(clock->c_repeat) )
		clock->c_counting = FALSE;
}



/*
 *	Rtcioctl provides commands to initialize the clock, start and stop the
 *	clock counting, (re)set the clock count, get the current clock count,
 *	get the current modes/setting of the clock, and wait for the clock to
 *	reach its terminal count. All of these commands require the process to
 *	have read and write access to the device (clock).
 *
 *	The RTIOCGETCNT, RTCIOCSETCNT, and RTCIOCRES commands require the data
 *	parameter to be a pointer to an unsigned short, while the RTCIOCGET and
 *	RTCIOCSET require it to be a pointer to a rtclock structure. All other
 *	commands ignore the data parameter.
 *
 *	Note that most of the variables are initialized in each of the ioctl
 *	commands, instead of when they are declared. This was done because of
 *	speed, that is, to make RTCIOCWAIT faster (this is the only ioctl
 *	command where time is of the essence).
 */

int
rtcioctl( dev, command, data, flags )
     dev_t  dev;
     int  command;
     register union {
	     struct rtc  *rtc;
	     u_short  *val;
     } data;
#define	rtcp	data.rtc  
#define	valp	data.val
     int  flags;

{

	int  i;
	int  ctlr_num;
	int  clk_num;
	u_short  master_reg, counter_reg;
	register struct rtclock  *ctlr;
	register struct rtc_regs *dev_regs;
	register struct rtclock  *clock = &rtc_tab_info[RTC_UNIT(dev)];


	if ( (flags & (FREAD | FWRITE)) != (FREAD | FWRITE) )
		return ( EBADF );

	
	switch ( command ) {

		
	      case RTCIOCWAIT :		/* wait on clock to wake process up */
		/*
	         * Clock must be counting or process would never get woken up.
		 */
		spl7();
		if ( !(clock->c_counting) ) {
			spl0();
			return ( EIO );
		}
			
		clock->c_pwait++;     /* mark that someone's waiting */
		if ( slumber(&clock->c_pwait, INTERRUPTABLE | PCATCH, iocomboost) ) {
			clock->c_pwait--;
			spl0();
			return ( EINTR );
		}
		spl0();
		break;

		
              case RTCIOCSTART :	/* start clock counting */
		/*
		 * Clock must be initialized via a RTCIOCSET command.
		 */
		if ( !(clock->c_flags & RTC_INIT) )
			return ( EIO );

		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		spl7();
		START_CLOCK( dev_regs, clk_num );
		clock->c_counting = TRUE;
		spl0();
		break;

		
	      case RTCIOCSTOP :		/* stop clock counting */   
		/*
	         * Clock must be initialized via a RTCIOCSET command.
		 */
		if ( !(clock->c_flags & RTC_INIT) )
			return ( EIO );

		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		spl7();
		STOP_CLOCK( dev_regs, clk_num );
		clock->c_counting = FALSE;
		spl0();
		break;

		
	      case RTCIOCGETCNT :	/* get current clock count */
		/*
		 * Clock must be initialized via a RTCIOCSET command.
		 */
		if ( !(clock->c_flags & RTC_INIT) )
			return ( EIO );

		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		GET_CLOCK_COUNT( dev_regs, clk_num, *(valp) );
		break;

		
	      case RTCIOCSETCNT :	/* set current clock count */
		/*
		 * Clock must be initialized via a RTCIOCSET command
		 * and not currently counting.
		 */
		spl7();
		if ( !(clock->c_flags & RTC_INIT) || (clock->c_counting) ) {
			spl0();	
			return ( EIO );
		}
		spl0();

		/*
		 * clock count must be greater than zero in default mode
		 */
		if (*valp == 0)
		  	return( EINVAL );

		/*
		 * If repeat mode is also selected, the clock count must
		 * must be greater than one. If this was not the case, i
		 * interrupts would not be generated repetetively.
		 */ 
		if ( (clock->c_repeat) && (*valp == 1) )
		  	return( EINVAL );
		
		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		SET_LOAD_REG( dev_regs, clk_num, *(valp) );
		LOAD_CLOCK_COUNT( dev_regs, clk_num );

		/* save the clock count value in the appropriate field */
		if (clock->c_rtc.r_modes & RTC_DIRECT)
		  	clock->c_rtc.r_load_reg = *(valp);
		else
		  	clock->c_rtc.r_clkcnt = *(valp);
		break;

		
	      case RTCIOCGET :		/* get current clock settings */
		/*
		 * Clock must be initialized via a RTCIOCSET command.
		 */
		if ( !(clock->c_flags & RTC_INIT) )
		  	return( EIO );
		
		bcopy((caddr_t) &clock->c_rtc, (caddr_t) rtcp, sizeof(struct rtc));
		break;

		
	      case RTCIOCSET :		/* set clock modes and count */
		/*
	 	 * Clock must not be currently counting.
		 */
		spl7();
		if ( clock->c_counting ) {
			spl0();	
			return ( EIO );
		}
		spl0();
				
#ifdef RTC_DEBUG
	printf("RTCIOCSET input parameters:\n");	
	printf("flags   = %o\n", rtcp->r_modes);
	printf("clkcnt  = %d\n", rtcp->r_clkcnt);
	printf("res     = %d\n", rtcp->r_res);
	printf("master  = %x\n", rtcp->r_master_reg);
	printf("counter = %x\n", rtcp->r_counter_reg);
	printf("load    = %x\n", rtcp->r_load_reg);
	printf("hold    = %x\n", rtcp->r_hold_reg);
#endif

		/* Validate input parameters:
		 *
		 * Direct Mode return error if the master or counter mode
		 * register have bits set which are not supported by this
		 * driver  (i.e., time of day mode)
		 *
		 * Default Mode return error if the resolution per clock
		 * count value is not supported or the clock count value
		 * is equal to 0 in any mode or 1 in repeat mode.
		 *
		 * Also, store the master mode register value in master_reg.
		 */

		if ( rtcp->r_modes & RTC_DIRECT ) {
			if ((rtcp->r_master_reg & (MM_TOD | MM_CMPAR)) ||
			    ((rtcp->r_counter_reg & CM_SRC) == CM_TCN) )
				return ( EINVAL );  /* unsupported bits */
			master_reg = rtcp->r_master_reg;

			/*
			 * If source line 1 is selected with pattern 0000,
			 * use bit pattern 1011 instead. This is done so
			 * master mode register checking (see below) will
			 * always work.
			 */
			if ((rtcp->r_master_reg & MM_FSRC) == 0)
			  	master_reg |= MM_FSRC1;
			
		} else { 	/* default mode */
			
			switch ( rtcp->r_res ) {
				
			      case USEC:	/* 1 microsec. resolution */
				master_reg = MM_USEC;
				break;

			      case USEC10:	/* 10 microsec. resolution */
				master_reg = MM_USEC10;
				break;

			      case USEC100:	/* 100 microsec. resolution */
				master_reg = MM_USEC100;
				break;
					
			      case MSEC:	/* 1 millisec. resolution */
				master_reg = MM_MSEC;
				break;

			      case MSEC10:	/* 10 millisec. resolution */
				master_reg = MM_MSEC10;
				break;

			      default:		/* unsupported resolution */
				return ( EINVAL );
			}

			/*
			 * clock count must be greater than zero in all modes
			 */
			if (rtcp->r_clkcnt == 0)
		  		return( EINVAL );

			/*
		 	 * In repetitive mode, it must also be greater than 
		 	 * one. If this was not the case, interrupts would
			 * not be generated repetetively.
		 	 */ 
			if ( (rtcp->r_modes & RTC_REPEAT) && (rtcp->r_clkcnt == 1) )
		  		return( EINVAL );
		}
	

		/*
		 * Now check that the resolution requested by this clock
		 * does not effect other active clocks on the same controller.
		 */

		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		ctlr = &rtc_tab_info[ctlr_num * CLKS_PER_CTLR];
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		for (i = 0; i < CLKS_PER_CTLR; i++, ctlr++)
			if ( (i != clk_num) &&
			     (ctlr->c_flags & RTC_OPEN) &&
			     (ctlr->c_flags & RTC_INIT) ) {
				GET_MASTER_REG(dev_regs, clock->c_rtc.r_master_reg); 
				if ( (clock->c_rtc.r_master_reg & MM_RES) !=
				     (master_reg & MM_RES) )
					return ( EIO );
				break;	/* they are the same, so continue */ 
			}

		/*
		 * Input parameters are valid, now save ioctl parameters
		 * in driver for later retrieval.
		 */
		bcopy ((caddr_t) rtcp, (caddr_t) &clock->c_rtc, sizeof(struct rtc));

		/*
		 * Determine counter mode register value based on mode of clock
		 */
		if ( rtcp->r_modes & RTC_DIRECT ) {
			counter_reg = rtcp->r_counter_reg;

		} else { 	/* default mode */

			counter_reg = CM_SRC1 | CM_AHTCP;
			if ( rtcp->r_modes & RTC_CNTUP ) {
				counter_reg |= CM_CNTUP;
			}
			if ( rtcp->r_modes & RTC_REPEAT ) {
				counter_reg |= CM_REPEAT;
	                	clock->c_repeat = TRUE;
			} else {
				clock->c_repeat = FALSE;
			}
		}

		master_reg |= MM_BUS16;		/* must be in 16 bus mode */
#ifdef RTC_DEBUG
	printf("RTCIOCSET master/counter mode register\n");
	printf("master	= 0x0%x\n", master_reg);
	printf("counter = 0x0%x\n", counter_reg);
#endif 

		/*
		 * Program the clock's registers and initialize internal
		 * data structures.
		 */
		SET_MASTER_REG( dev_regs, master_reg );
		SET_COUNTER_REG( dev_regs, clk_num, counter_reg );
		if (rtcp->r_modes & RTC_DIRECT) {
			SET_LOAD_REG( dev_regs, clk_num, rtcp->r_load_reg );
			SET_HOLD_REG( dev_regs, clk_num, rtcp->r_hold_reg );

		} else {  			/* default mode */
			SET_LOAD_REG( dev_regs, clk_num, rtcp->r_clkcnt );
			SET_HOLD_REG( dev_regs, clk_num, 0 );
		}
		LOAD_CLOCK_COUNT(dev_regs, clk_num);

		clock->c_flags |= RTC_INIT;
		
#ifdef RTC_DEBUG
	printf("RTCIOCSET driver data struct values:\n");	
	printf("c_flags   = 0o0%o\n", clock->c_flags);
	printf("c_pwait   = %d\n"   , clock->c_pwait);
	printf("c_counting= %d\n"   , clock->c_counting);
	printf("c_repeat  = %d\n"   , clock->c_repeat);
	printf("r_modes   = %d\n"   , clock->c_rtc.r_modes);
	printf("r_clckcnt = %d\n"   , clock->c_rtc.r_clkcnt);
	printf("r_res     = %d\n"   , clock->c_rtc.r_res);
	printf("r_master  = 0x0%x\n", clock->c_rtc.r_master_reg);
	printf("r_counter = 0x0%x\n", clock->c_rtc.r_counter_reg);
	printf("r_load    = 0x0%x\n", clock->c_rtc.r_load_reg);
	printf("r_hold    = 0x0%x\n", clock->c_rtc.r_hold_reg);
#endif 
				
		break;

		
	      case RTCIOCRES :		/* get valid resolution setting(s) */ 
		/*
		 * Check if another clock on this controller is open and
		 * has set the resolution. 
		 */
		
		ctlr_num = RTC_CTLR(dev);
		clk_num = RTC_CLOCK(dev);
		ctlr = &rtc_tab_info[ctlr_num * CLKS_PER_CTLR];
		dev_regs = (struct rtc_regs *) rtc_vba_ctlr[ctlr_num]->um_addr;

		for (i = 0; i < CLKS_PER_CTLR; i++, ctlr++)
		  if ( (i != clk_num) && (ctlr->c_flags & RTC_OPEN) &&
		       (ctlr->c_flags & RTC_INIT) )
		    if (ctlr->c_rtc.r_modes & RTC_DIRECT) {
			    GET_MASTER_REG( dev_regs, master_reg );
			    switch ( master_reg & MM_RES ) {

				  case MM_USEC :
				    *(valp) = USEC;
				    break;

				  case MM_USEC10 :
				    *(valp) = USEC10;
				    break;

				  case MM_USEC100 :
				    *(valp) = USEC100;
				    break;

				  case MM_MSEC :
				    *(valp) = MSEC;
				    break;

				  case MM_MSEC10 :
				    *(valp) = MSEC10;
				    break;

				  default :
				    *(valp) = ctlr->c_rtc.r_master_reg;
			    	    return( EIO );
			    }

			    return( SUCCESS );
			    
		    } else { 	/* default mode */

			    *(valp) = ctlr->c_rtc.r_res;
			    return( SUCCESS );
		    }
			    
		*(valp) = 0;		/* can be set to any supported value */
		break;

		
	      default :		/* invalid command */
		return ( ENXIO );
	}
	
	return ( SUCCESS );
}



/*
 *	Rtcopen insures that the device exists and is not currently open.
 *	If this is true, the clock is marked as opened and SUCCESS is
 *	returned.
 */

int
rtcopen( dev, flags )
	register dev_t  dev;
	int  flags;

{
	register int  mdev = RTC_UNIT(dev);	
	register struct rtclock  *clock = &rtc_tab_info[mdev];

	if ( (mdev >= nclock) || (mdev >= RTCMAXCLOCKS) )
		return ( ENXIO );

	if ( !(clock->c_flags & RTC_CONFIG) )
		return ( ENXIO );

	if ( clock->c_flags & RTC_OPEN )
		return ( EBUSY );

	clock->c_flags |= RTC_OPEN;
	return ( SUCCESS ); 
}



/*
 *	Rtcprobe checks for a real-time clock controller at the specified
 *	address by writing a count value into counter register 1 and reading
 *	it back. If they are the same, TRUE is returned, otherwise FALSE is
 *	returned. This routine also resets the controller.
 */

int
rtcprobe( ctlr_info )
	register struct vba_ctlr  *ctlr_info;

{
	u_short  rval;
	register struct rtc_regs  *dev_regs =
		(struct rtc_regs *) ctlr_info->um_addr;

#ifdef RTC_DEBUG
	printf("Entering rtcprobe ...\n");
#endif	
	
	if ( badaddr(&dev_regs->cmd_reg1, sizeof(dev_regs->cmd_reg1), 0) )
		return( FALSE );

	rtc_reset_ctlr( dev_regs );    /* reset controller */

	SET_LOAD_REG( dev_regs, 0, 0xf5a7 );
	LOAD_CLOCK_COUNT( dev_regs, 0 ); 
	GET_CLOCK_COUNT( dev_regs, 0, rval );

#ifdef RTC_DEBUG
	printf("Exiting rtcprobe with rval=%x (0xf5a7) ...\n", rval);
#endif	

	if ( rval == 0xf5a7 )
		return( TRUE );
	else
		return( FALSE );
}



/*
 *	Rtc_reset_ctlr insures that the clock is in a known state at
 *	start-up time.
 *
 *	IMPORTANT: Chip's 2 FOUT line is the frequency, more precisely
 *                 the F1 line, for chip 1. FOUT on chip 2 MUST output
 *		   a frequency of 4MHz or the frequencies provided in
 *                 default mode will not work as advertised without changes
 *		   to this driver.
 *	          
 *	           
 */ 

int
rtc_reset_ctlr ( dev_regs )
     register struct rtc_regs  *dev_regs;

{
	
#ifdef RTC_DEBUG
	printf("Entering rtc_reset_ctlr ...\n");
#endif
	
	/*
	 * Reset chip 2. This sets FOUT to a frequency of 4MHz.
	 */
	movob(MASTER_RESET, &dev_regs->cmd_reg2);
	movob(LOAD_DP_REG | (MASTER_ELEM_PTR << 3) | MASTER_GRP_PTR,
	      	&dev_regs->cmd_reg2);
	movob(BUS_16_BITS, &dev_regs->cmd_reg2); 
	movow(MM_FDIV1, &dev_regs->data_reg2);

	/*
	 * Initialize chip 1, the one used by this driver.
	 */
	movob(MASTER_RESET, &dev_regs->cmd_reg1);	/* reset chip 1 */
	movob((LOAD | 0xf), &dev_regs->cmd_reg1);	/* init all counters */
	movob((LOAD_DP_REG | MASTER_ELEM_PTR | MASTER_GRP_PTR), &dev_regs->cmd_reg1);
	movob(BUS_16_BITS, &dev_regs->cmd_reg1);	/* use 16 bus */

	
#ifdef RTC_DEBUG
	printf("Exiting rtc_reset_ctlr ...\n");
#endif

}



/*
 *	Rtcslave initializes the 4 clocks on the controller. 
 */

int
rtcslave( dev_info, ctlr_addr )
	struct vba_device *dev_info;
	caddr_t ctlr_addr;

{
	struct rtc_regs *dev_regs = (struct rtc_regs *) ctlr_addr;
	int clock;		/* starting clock number */
	int ctlr;		/* controller index into device data struct */
	int status = FALSE;	/* flag indicating at least one clock is configured */
	u_short rval;		/* read clock count */
	int vec_addr;		/* interrupt vector address on the board */

	int  	ivct;		/* interrupt vector number */

	extern	int	Xrtcintr0,   Xrtcintr1,   Xrtcintr2,   Xrtcintr3,
			Xrtcintr4,   Xrtcintr5,   Xrtcintr6,   Xrtcintr7,
			Xrtcintr8,   Xrtcintr9,   Xrtcintr10,  Xrtcintr11,
			Xrtcintr12,  Xrtcintr13,  Xrtcintr14,  Xrtcintr15,
			Xrtcintr16,  Xrtcintr17,  Xrtcintr18,  Xrtcintr19,
			Xrtcintr20,  Xrtcintr21,  Xrtcintr22,  Xrtcintr23,
			Xrtcintr24,  Xrtcintr25,  Xrtcintr26,  Xrtcintr27,
			Xrtcintr28,  Xrtcintr29,  Xrtcintr30,  Xrtcintr31;

	static	int  rtc_proc[] = {
		(int)& Xrtcintr0,   (int)& Xrtcintr1,   (int)& Xrtcintr2,
		(int)& Xrtcintr3,   (int)& Xrtcintr4,   (int)& Xrtcintr5,
		(int)& Xrtcintr6,   (int)& Xrtcintr7,   (int)& Xrtcintr8,
		(int)& Xrtcintr9,   (int)& Xrtcintr10,  (int)& Xrtcintr11,
		(int)& Xrtcintr12,  (int)& Xrtcintr13,  (int)& Xrtcintr14,
		(int)& Xrtcintr15,  (int)& Xrtcintr16,  (int)& Xrtcintr17,
		(int)& Xrtcintr18,  (int)& Xrtcintr19,  (int)& Xrtcintr20,
		(int)& Xrtcintr21,  (int)& Xrtcintr22,  (int)& Xrtcintr23,
		(int)& Xrtcintr24,  (int)& Xrtcintr25,  (int)& Xrtcintr26,
		(int)& Xrtcintr27,  (int)& Xrtcintr28,  (int)& Xrtcintr29,
		(int)& Xrtcintr30,  (int)& Xrtcintr31
	  };


#ifdef RTC_DEBUG
	printf("Entering rtcslave ...\n");
#endif

	ctlr = dev_info->ui_ctlr * CLKS_PER_CTLR; 

	for ( clock = 0; clock < CLKS_PER_CTLR; clock++ ) {
		SET_LOAD_REG( dev_regs, clock, 0xf5a7 );
		LOAD_CLOCK_COUNT( dev_regs, clock );
 		GET_CLOCK_COUNT( dev_regs, clock, rval );
#ifdef RTC_DEBUG
	printf("Clock's %d rval=%x (0xf5a7)\n", clock, rval);
#endif
		if ( rval == 0xf5a7 ) {
			if ( (ivct = get_ivct( 0, 1)) != -1) {
				init_ivct( ivct, rtc_proc[ctlr + clock] );
			      	vec_addr = (int) &(dev_regs->vec1) +
			                       ((VEC2-VEC1) * clock);
				SET_INTR_VEC_NUM( vec_addr, ivct );
				rtc_tab_info[ ctlr + clock ].c_flags =
				  				RTC_CONFIG;
				status = TRUE;
		      	} else
				printf("SCB table full\n");
		} else
		  	printf("Real-time clock %d on rtc%d is bad.\n",
			       clock, dev_info->ui_ctlr);
	}			

#ifdef RTC_DEBUG
	printf("Exiting rtcslave with status of %d ...\n", status);
#endif

	return( status );
}
