/*
**  Title:	SC.C - Plessey SCSI-1 SCSI interface controller driver.
**
**  Author:	John R. Franks
**  Date:	24-Apr-84
**
**	This driver drives the scsi buss adapter manfuactured by Plessey.
**  The driver accepts commands from other pseudo drivers which know all about
**  the individual scsi periperal adapters.
**
**	Each Plessey adapter looks like 7 seperate scsi buss controllers.
** It can handle up to 16 concurrent tasks, therefore it is limited to
** 16 units on one board.  Currently, since this driver has just one
** queue per target (7), if two devices are on the same controller, they
** can't overflow anyway.
** 
**  
**  Adapted for HCX
**  Author: Terry Treder
**  Date:   21-May-87
**
**  The Plessey SCSI board looks like 7 seperate SCSI bus controllers,
**  in the 8th slot it has the VME card.
**
**  This controller violates some common assumptions, so look out.
**  You may have up to 7 controllers on one board, each of these controllers
**  may drive different devices.
** 
**  The SCSI isn't really a controller, it's more like a bus.
**
**  The unit number determines the target on the SCSI bus.
**
**  Controllers have 7 units.  Unit 0 is on board 0 and is target 0,
**  unit 1 is on board 0 and is target 1, on up to unit 6 is on board 0
**  and target 6.  Unit 7 is on board 1 and is target 0.
**
**  Warning: There are some subtle changes here from the original CCI driver.
**           They are mainly in extra parameters passed and differences in
**           passing controller vs. unit numbers.
**
**  For more information on the board read the PME SCSI-1 Manual,
**  publication 421/HH/23622, from Plessey Microsystems Limited.
**  Its fairly good from a software point of view.
*/

#include	"../sys/param.h"
#include	"../sys/systm.h"
#include	"../ml/mtpr.h"
#include	"../sys/vm.h"
#include	"../sys/buf.h"
#include	"../ml/pte.h"
#include	"../sys/dir.h"
#include	"../sys/user.h"
#include	"../sys/file.h"
#include	"../sys/uio.h"
#include	"../sys/ioctl.h"
#include	"../sys/mtio.h"
#include	"../sys/errno.h"
#include	"../sys/cmap.h"
#include	"../sys/vbavar.h"
#include	"../sys/scsi.h"
#include	"../sys/elog.h"		/* for system activity */
#include	"../sys/iobuf.h"	/* for system activity */
#include	"../sys/kernel.h"	/* for system activity */

#define	SCOM		3		/* sending command */


#define TRUE	1
#define FALSE	0

extern int	nsc, nscdevs;
extern int	blkacty;
extern struct	iotime	scstat[];	/* for system activity */
extern struct	iobuf	sctab[];	/* for error logging */
extern struct	vba_ctlr	*scminfo[];
extern struct	vba_device	*scdinfo[];
extern sc_tab	sc_info[];
extern unsigned long sc_intvect_proc[];    /* device info */

int	scprobe();		/* See if controller is really there */
int	scslave();		/* See if unit is really there */
int	scattach();		/* Enter a unit in internal tables */
int	scdgo();		/* Old VAX function, not used. */


/* for error logging */
struct screg {
	sc_status	status;
} sc_reg;


struct	vba_driver scdriver = {
	scprobe, scslave, scattach, scdgo, (long *)&sc_reg,
	"", scdinfo, "SCSI controller #", scminfo, 0
};


/*
**	See if the controller is really there. 
*/

scprobe(um)
	struct vba_ctlr * um ;
{
	register caddr_t ctlr_vaddr = (caddr_t) um -> um_addr ;

	if ( badaddr( ctlr_vaddr, 2 ) )
		return FALSE;

	um->um_ivct = get_ivct(0,1);	
	if ((um->um_ivct < 0) ||
	 (init_ivct(um->um_ivct, sc_intvect_proc[um->um_ctlr]) < 0 )) {
		printf("\nInterrupt vector for scsi%d not set up.\n",um->um_ctlr);
		return( FALSE );
	}

	return TRUE;
}


/*
** See if a unit is really there
*/

#define	TEST	0

scslave(vba_dev_info, ctlr_addr)
register struct vba_device	*vba_dev_info;
register rcb			*ctlr_addr;
{
	register int		time;
	register cpb		*Cpb = &sc_info[0].sc_cpb;
	register int               ctlr = vba_dev_info->ui_unit;
	register int               target; 
	register int               ret = 0;

	/* Check drive status */
	/* Load command block */
	/* This always uses the 1st command block since no one else is around */
        

	for (target = 0; target < CTLR_PER_SCSI; target++) {

		load_sc_cmd(&Cpb->cpb_command_block, TEST, 0, 0, 0);

		/* load data pointer and length (none for now) */
		Cpb->cpb_data_address.sc_mod = ADDR_MODIFIER;
		Cpb->cpb_data_address.sc_addr = 0;
		Cpb->cpb_data_length = 0;

		/* load link address (none) */
		Cpb->cpb_link_address.sc_mod = ADDR_MODIFIER;
		Cpb->cpb_link_address.sc_addr = 0;

		/* load protocol block */
		Cpb->cpb_protocol.auto_sense = FALSE;
		Cpb->cpb_protocol.int_disable = TRUE;
		Cpb->cpb_protocol.xfer_32_bit = TRUE;
		Cpb->cpb_protocol.int_level = SCSI_LEVEL;
		Cpb->cpb_protocol.int_vector = 0;

		Cpb->cpb_target = target;
		Cpb->cpb_lun = 0;
		time = 250000;

		/* and start command */
		SCSI_GO(ctlr_addr, vtoph(0, Cpb));

		/* now wait for completion */
		while(!(Cpb->cpb_status.cmd_complete) && time) {
			DELAY(20);
			uncache(&Cpb->cpb_status);
			time--;
		}
		mtpr(0, PADC);

		if(!time || (Cpb->cpb_status.sc_timeout) || (Cpb->cpb_status.scsi_ned)){
		}
		else {
			sc_info[ctlr*CTLR_PER_SCSI + target].sc_alive = 1;
			ret = 1;
			printf("sc%d target %d unit %d \n",ctlr,target,
					  ctlr*CTLR_PER_SCSI+target);
		}
	}
	return ret;
}


/*
**
*/

scattach(vba_dev_info)
struct vba_device *vba_dev_info;
{
}


/*
** Historic VAX routine.
*/

scdgo(um)
struct vba_ctlr *um;
{
}


/*
**
*/

scopen(unit)
register int	unit;
{
	register struct vba_device	*dev_info;
	
	if((unit < nsc*CTLR_PER_SCSI) &&
	  ((dev_info = scdinfo[unit/CTLR_PER_SCSI]) != NULL) &&
	  (dev_info->ui_alive) &&
	  (sc_info[unit].sc_alive)) {
		if(sc_info[unit].sc_open == 0) {
			sctab[unit].io_nreg =
			    (sizeof(struct screg) / sizeof(u_int));
			sctab[unit].io_stp =  &scstat[unit].ios;
			sc_info[unit].sc_open = 1;
			return 0;
		}
		return EBUSY;
	}
	return ENXIO;
}


/*
**
*/

scclose(unit)
register int	unit;
{
	
	sc_info[unit].sc_open = 0;
	return 1;
}


/*
**	The scsi controller command is passed in the 'request' buf structure 
**  using the av_back as a pointer to the scsi command block.
**
**        request
**  +------------------+
**  |      .....       |
**  +------------------+
**  | (sccom *)av_back | -> Pointer to sccom structure which contains both the
**  +------------------+    sc_cmd structure for passing commands and the sc_ack
**  |      .....       |    structure for passing back the status and sense.
**  +------------------+
**  |     b_bcount     | -> Number of data bytes to transfer.
**  +------------------+
**  |      .....       |
**  +------------------+
**  |   b_un.b_addr    | -> Start address of data transfer.
**  +------------------+
*/

scstrategy(request)
register struct buf *request;
{
	register struct buf	 *ctlr_queue;
	register int		 priority;
	register sc_com		 *cmd_buf;
	register int	unit = minor(request->b_dev);
	

	if(sc_info[unit].sc_open != 0) {
		ctlr_queue = &sc_info[unit].sc_tab;
		request->b_start = time.tv_sec;
		scqueue(request, ctlr_queue);
		priority = spl/**/SCSI_LEVEL();
		if(!ctlr_queue->b_active++)
			scstart(unit);
		/*
		** Goto sleep.
		** If we wakeup early then remove request from queue and exit.
		*/
		if(slumber(&sc_info[unit], PRIBIO, iocomboost)) {
			/* if request is active then sleep until done */
			if((ctlr_queue->av_forw==request)&&ctlr_queue->b_active)
				slumber(&sc_info[unit], PRIBIO, iocomboost);
			else if(ctlr_queue->b_active) {
				ctlr_queue->b_active--;
				scremqueue(ctlr_queue, request);
			}
			request->b_resid = request->b_bcount;
			request->b_error = EINTR;
			request->b_flags |= B_ERROR;
		}
		splx(priority);
		return;
	}
	request->b_resid = request->b_bcount;
	request->b_error = ENXIO;
	request->b_flags |= B_ERROR;
}

/*
**
*/

scsi_timeout(unit)
register int	unit;
{
	sc_info[unit].sc_cpb.cpb_status.sc_timeout = TRUE;
	scintr(unit);
}


/*
** Start up a transfer on a unit.
*/

scstart(unit)
register int unit;
{
	register rcb		*ctlr_addr = (rcb *)(scminfo[unit/CTLR_PER_SCSI]->um_addr);
	register struct buf	*request;
	register cpb		*Cpb;
	register int		priority;

	if((request = sc_info[unit].sc_tab.av_forw) != NULL) {

		Cpb = &sc_info[unit].sc_cpb;

		/* clean out old junk out of command block to start */
		bzero(Cpb, sizeof(cpb));

		/* Load command block */
		Cpb->cpb_command_block = ((sc_com *)(request->av_back))->sccmd;

		/* load data pointer and length */
		Cpb->cpb_data_address.sc_mod = ADDR_MODIFIER;
		Cpb->cpb_data_address.sc_addr = request->b_un.b_addr;
		Cpb->cpb_data_length = request->b_bcount;

		/* load link address (none) */
		Cpb->cpb_link_address.sc_mod = ADDR_MODIFIER;
		Cpb->cpb_link_address.sc_addr = 0;

		/* load protocol block */
		Cpb->cpb_protocol.auto_sense = TRUE;
		Cpb->cpb_protocol.int_disable = FALSE;
		Cpb->cpb_protocol.xfer_32_bit = TRUE;
		Cpb->cpb_protocol.int_level = SCSI_LEVEL;
		Cpb->cpb_protocol.int_vector = scminfo[unit/CTLR_PER_SCSI]->um_ivct+unit;

		Cpb->cpb_target = unit % CTLR_PER_SCSI;
		Cpb->cpb_lun = 0;

		scstat[unit].io_bcnt += request->b_bcount / 512;
		scstat[unit].ios.io_misc++;
		sctab[unit].b_active = SCOM;

		priority = spl8();
		timeout(scsi_timeout, unit, 5*60*60);
		/* and start command */
		SCSI_GO(ctlr_addr, vtoph(0, Cpb));
		splx(priority);
	}
}


/*
 * Handle a scsi controller interrupt.
 */
scintr(unit)
register int	unit;
{
	register sc_tab			*c_info = &sc_info[unit];
	register cpb			*Cpb = &c_info->sc_cpb;
	register struct buf		*ctlr_queue = &c_info->sc_tab;
	register struct buf		*request = ctlr_queue->av_forw;
	register sc_com			*cmd_buf = (sc_com *)(request->av_back);
	int				priority, index, purge_ptr;

	/* Cancel lost interupt timer */
	untimeout(scsi_timeout, unit);

	/* Purge old command data in cache */
	for(index=0; index<((sizeof(cpb) + 3) / sizeof(long)); index++)
		uncache(((long *)Cpb)+index);

	/* load the return status for pseudo driver to interpret */
	bcopy(Cpb->cpb_sense, &cmd_buf->scack.ack_sense, 12);
	cmd_buf->scack.ack_status = Cpb->cpb_status;

	if(Cpb->cpb_status.bad_free)
		printf("Bad SCSI buss free\n");
	else if(Cpb->cpb_status.bad_phase)
		printf("Unknown SCSI buss phase unit\n");
	else if(Cpb->cpb_status.sc_timeout)
		printf("SCSI buss selection timeout\n");
	else if(Cpb->cpb_status.bad_reset)
		printf("SCSI buss received unexpected buss free\n");
	else if(Cpb->cpb_status.long_status)
		printf("SCSI status error\n");
	else if(Cpb->cpb_status.long_command)
		printf("SCSI command error\n");
	else if(Cpb->cpb_status.bus_error)
		printf("VME DMA error\n");
	else
		goto ok;
	printf("sc%d target %d unit %d\n",unit/CTLR_PER_SCSI,unit%CTLR_PER_SCSI,
	       unit);
	request->b_resid = request->b_bcount;
	request->b_error = EIO;
	request->b_flags |= B_ERROR;
	sc_reg.status = Cpb->cpb_status;
	sctab[unit].b_dev = (SCSI << 8) | unit;
	sctab[unit].b_actf = request;
	sctab[unit].io_addr = request->b_un.b_addr;
	fmtberr(&sctab[unit], (u_int)0, (u_int *)&sc_reg);
ok:
	if(Cpb->cpb_status.underflow)
		request->b_resid = Cpb->cpb_data_length;
	else
		request->b_resid = 0;

	/* begin system activity and error logging */
	if (sctab[unit].b_active) {
		if (sctab[unit].io_erec)
			logberr(&sctab[unit], request->b_flags & B_ERROR);
		sctab[unit].b_errcnt = 0;
		sctab[unit].b_active = 0;
		sctab[unit].b_actf = request->av_forw;
		scstat[unit].io_resp += time.tv_sec - request->b_start;
		scstat[unit].io_act += time.tv_sec - sctab[unit].io_start;
		sctab[unit].io_start = 0;
	}
	ctlr_queue->b_active--;
	scdequeue(ctlr_queue);

	/* Notify strategy routine of completion */
	wakeup(&sc_info[unit]);

	/* start up next command if any before returning */
	scstart(unit);
}

/*
**
*/

sc_print_cpb(ptr)
register unsigned char	*ptr;
{
	register int i,j,delay;

	printf("Dump of CPB:\n");
	printf("  SCSI command block:");
	for(i=0; i<2; i++) {
		for(delay=0;delay<500000;delay++);
		printf("\n\t");
		for(j=0; j<8; j++) {
			uncache(ptr);
			printf("        0x%x  ", *(ptr++) & 0xff);
		}
	}
	printf("\n  Data address:\t");
	for(i=0; i<8; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  Transfer length:\t");
	for(i=0; i<4; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	for(delay=0;delay<500000;delay++);
	printf("\n  Link address:\t");
	for(i=0; i<8; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  protocol block:\t");
	for(i=0; i<4; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	printf("\n  SCSI status:\t");
	for(i=0; i<2; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	for(delay=0;delay<500000;delay++);
	printf("\n  Host status\t");
	for(i=0; i<2; i++) {
		uncache(ptr);
		printf("0x%x  ", *(ptr++) & 0xff);
	}
	uncache(ptr);
	printf("\n  Target identfier:\t0x%x", *(ptr++) & 0xff);
	uncache(ptr);
	for(delay=0;delay<500000;delay++);
	printf("\n  LUN identfier:\t0x%x", *(ptr++) & 0xff);
	printf("\n  SCSI sense block:");
	for(i=0; i<2; i++) {
		printf("\n\t");
		for(j=0; j<6; j++) {
			uncache(ptr);
			printf("0x%x  ", *(ptr++) & 0xff);
		}
	}
	for(delay=0;delay<500000;delay++);
	printf("\n\n");
}



/*
 *  This is very device specific, but here it goes.
 * 
 *  SCSI commands are included as the first 16 bytes of the Command 
 *  Parameter Block with the rest of the CPB comprising the (mainly)
 *  VMEbus address pointers and parameters.
 *
 *  The SCSI commands are device dependent and the board will not
 *  perform any decode of these commands.  It merely transfers the
 *  entire block.
 *
 *  Currently in SCSI land there are three Command Block lengths
 *  6, 10, and 12 bytes.   This corresponds to Group Codes
 *  0, 1, and 5.
 *  
 *  The code below only handles 6 byte blocks so the group code
 *  should always be 0.
 * 
 *  Reguardless of the group type byte 0 has the Group code in the
 *  first 3 bits and a 5 bit command code in the rest.
 *
 *  Bits 5, 6, and 7 of the second byte of a Command Block define the
 *  Logical Unit Number (LUN) to address one of 8 devices attached to
 *  a target.
 *  
 *  The last byte of a block is a control byte.  Bit 0 is a link flag
 *  set if you wish to automatically link to the next command.
 *  Bit 1 is set if you desire an interrupt on linked commands.
 * 
 *  In this case we only use the 6 byte format:
 * 
 *  0 - Operation Code
 *  1 - LUN, Logical Block Address bits (MSB)
 *  2 - Logical Block Address
 *  3 - Logical Block Address (LSB)
 *  4 - Transfer Length
 *  5 - Control byte
 * 
 *  Refer to the manual for more information. 
 */

load_sc_cmd(ptr, cmd, group_code, lun, depend)
register unsigned char	*ptr;
unsigned char		cmd, group_code, lun;
register unsigned long	depend;
{
	*(ptr++) = ((group_code & 0x07) << 5) | (cmd & 0x1f);
	*(ptr++) = (unsigned char)((lun & 0x07)<<5)|((depend & 0x1f000000)>>24);
	*(ptr++) = (unsigned char)((depend & 0xff0000) >> 16);
	*(ptr++) = (unsigned char)((depend & 0xff00) >> 8);
	*(ptr++) = (unsigned char)(depend & 0xff);
	*ptr = 0;
}


/*
**	The routines below are used to handle a unit's request queue.
**  The queue is a linked list of buf structures.  The linkage is as follows:
**
**    +------------------------------------------------+
**    |                                                v
**    | +-----------+       +-----------+        +-----------+
**    | |  av_forw  |------>|  av_forw  |--~ ~-->|  av_forw  |-->NULL
**    | +-----------+       +-----------+        +-----------+
**    +-|  av_back  |       | ......... |        | ......... |
**      +-----------+       +-----------+        +-----------+
**      | ......... |        First queue           Last queue
**      +-----------+          element               element
**    head of unit queue
**  (unit_tab[unit].queue)
*/

/*
**	Queue_request places a queue element on the end of a unit's
**  request queue.
*/

scqueue(request, que)
register struct buf	*request;
register struct buf	*que;
{
	register int	priority = spl/**/SCSI_LEVEL();

	request->av_forw = NULL;
	if (que->av_forw == NULL) 
		que->av_forw = request;
	else
		que->av_back->av_forw = request;
	que->av_back = request;
	splx(priority);
}


/*
**	Dequeue_request removes the first element in a unit's request queue.
*/

scdequeue(que)
register struct buf	*que;
{
	register int	priority = spl/**/SCSI_LEVEL();

	if ((que->av_forw = que->av_forw->av_forw) == NULL) 
		que->av_back = NULL;
	splx(priority);
}


/*
**	scremqueue removes a queue entry that might be in the middle of
**  the activity queue.
*/

scremqueue(que, request)
register struct buf 	*que, *request;
{
	register struct buf 	*next;
	register int		priority = spl/**/SCSI_LEVEL();

	next = que;
	while(next->av_forw != NULL) {
		if(next->av_forw == request) {
			if ((next->av_forw = next->av_forw->av_forw) == NULL) 
				que->av_back = NULL;
			break;
		}
	}
	splx(priority);
}


