/* @(#)tty_tt0.c	6.3 */
/*
 * Line discipline 0
 * No Virtual Terminal Handling
 */

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/tty.h"
#include "../sys/proc.h"
#include "../sys/file.h"
#include "../sys/conf.h"
#include "../sys/buf.h"
#include "../sys/dk.h"
#include "../sys/uio.h"
#include "../sys/kernel.h"
#include "../sys/termio.h"
#include "../sys/ioctl.h"
#include "../sys/sysinfo.h"

extern char partab[];

extern char maptab[];

/*
 * routine called on first teletype open.
 * establishes a process group for distribution
 * of quits and interrupts from the tty.
 */
ttopen(tp)
register struct tty *tp;
{
	register struct proc *pp;

	pp = u.u_procp;

	/* If this process did a TIOCNOTTY, play by the UCB rules. */
	if (pp->p_pgrp == 0) {
		u.u_ttyp = tp;
		u.u_ttyd = tp->t_dev;
		if (tp->t_pgrp == 0)
			tp->t_pgrp = pp->p_pid;
		pp->p_pgrp = tp->t_pgrp;
	}

	/* If thie process is a process group leader, play by the ATT rules */
 	if ((pp->p_pid == pp->p_pgrp)
	 && (u.u_ttyp == NULL)
	 && (tp->t_pgrp == 0)) {
		u.u_ttyp = tp;
		u.u_ttyd = tp->t_dev;
		tp->t_pgrp = pp->p_pgrp;
	}

	(void) ttioctl(tp, LDOPEN, (caddr_t)0, 0);
	tp->t_state &= ~WOPEN;
	tp->t_state |= ISOPEN;
	return(0);
}

ttclose(tp)
register struct tty *tp;
{
	if ((tp->t_state&ISOPEN) == 0)
		return(0);
	tp->t_state &= ~(ISOPEN|XCLUDE|NBIO|ASYNC);
	if (tp->t_open_cnt)	/* someone else is waiting on open */
		tp->t_state |= WOPEN;
	tp->t_pgrp = 0;
	(void) ttioctl(tp, LDCLOSE, (caddr_t)0, 0);
	return(0);
}

/*
 * Called from device's read routine after it has
 * calculated the tty-structure given as argument.
 */
ttread(tp, uio)
register struct tty *tp;
register struct uio *uio;
{
	register struct clist *tq;
	register int first;
	register int error = 0;

loop:
#define bit(a) (1<<(a-1))
	while ((tp->t_xflag&NEWTTY) &&
	   u.u_procp->p_pgrp != tp->t_pgrp && tp == u.u_ttyp) {
		if ((u.u_procp->p_flag&SVFORK) == 0 &&
		   !(u.u_procp->p_sigignore & bit(SIGTTIN)) &&
		   !(u.u_procp->p_sigmask & bit(SIGTTIN))) {
			gsignal(u.u_procp->p_pgrp, SIGTTIN);
			slumber((caddr_t)&lbolt, INTERRUPTABLE, delayboost);
		} else
			return(EIO);
	}
#undef	bit

	tq = &tp->t_canq;
	if (tq->c_cc == 0) {
		if ((error = canon(tp)) != 0)
			return(error);
	}
	first = 1;
	while (uio->uio_resid > 0 && error == 0) {
		if (uio->uio_resid >= CBSIZE && tp->t_dscnt == 0) {
			register n;
			register struct cblock *cp;

			if ((cp = getcb(tq)) == NULL)
				break;
			n = min(uio->uio_resid, cp->c_last - cp->c_first);
			error = uiomove(&cp->c_data[cp->c_first], n, UIO_READ,
					uio);
			putcf(cp);
		} else {
			register c;

			if ((c = getc(tq)) < 0)
				break;
			/*
			 * Check for delayed suspend character.
			 */
			if ((tp->t_xflag&NEWTTY) && (tp->t_lflag&ISIG)
			    && c == tp->t_cc[VDSUSP]) {
				tp->t_dscnt--;
				gsignal(tp->t_pgrp, SIGTSTP);
				if (first) {
					slumber((caddr_t)&lbolt, INTERRUPTABLE,
						delayboost);
					goto loop;
				}
				break;
			}
			error = ureadc(c, uio);
			first = 0;
		}
	}
	if (tp->t_state&TBLOCK) {
		if (tp->t_rawq.c_cc<TTXOLO) {
			(*tp->t_proc)(tp, T_UNBLOCK);
		}
	}
	return(error);
}

/*
 * Called from device's write routine after it has
 * calculated the tty-structure given as argument.
 */
ttwrite(tp, uio)
register struct tty *tp;
register struct uio *uio;
{
	register int error = 0;
	register s;
	int cnt;

	if (!(tp->t_state&CARR_ON))
		return(0);

#define bit(a) (1<<(a-1))
	while ((tp->t_xflag&NEWTTY) &&
	   u.u_procp->p_pgrp != tp->t_pgrp && tp == u.u_ttyp &&
	   (tp->t_xflag&TOSTOP) && (u.u_procp->p_flag&SVFORK) == 0 &&
	   !(u.u_procp->p_sigignore & bit(SIGTTOU)) &&
	   !(u.u_procp->p_sigmask & bit(SIGTTOU))) {
		gsignal(u.u_procp->p_pgrp, SIGTTOU);
		slumber((caddr_t)&lbolt, INTERRUPTABLE, delayboost);
	}
#undef	bit

	cnt = uio->uio_resid;
	while (uio->uio_resid) {
		if (tp->t_outq.c_cc > tthiwat[tp->t_cflag&CBAUD]) {
			s = spl8();
			(*tp->t_proc)(tp, T_OUTPUT);
			while (tp->t_outq.c_cc > tthiwat[tp->t_cflag&CBAUD]) {
				if (tp->t_state&NBIO) {
					if (uio->uio_resid == cnt) {
						splx(s);
						return (EWOULDBLOCK);
					}
					splx(s);
					return (0);
				}
				tp->t_state |= OASLP;
				slumber((caddr_t)&tp->t_outq, INTERRUPTABLE,
					ttoboost);
			}
			splx(s);
		}

		/* -------------------------------------------------- */
		s = spl8();
		if ((!(tp->t_oflag & OPOST)) && (tp->t_state & BUSY) &&
			(tp->t_tbuf.c_ptr == 0) &&
			(tp->t_outq.c_cc >= CBSIZE<<1)) {
				ttout(tp);	/* Fill ahead */
			}
		splx(s);
		/* -------------------------------------------------- */

		if (uio->uio_resid >= (CBSIZE/2)) {
			register n;
			register struct cblock *cp;

			if ((cp = getcf()) == NULL)
				break;
			n = MIN(uio->uio_resid, cp->c_last);
			if (error = uiomove(cp->c_data, n, UIO_WRITE, uio)){
				putcf(cp);
				break;
			}
			cp->c_last = n;
			if (tp->t_xflag&FLUSHO)
				putcf(cp);
			else {
				ttxput(tp, cp, n);
				tp->t_rocount = 0;
			}
		} else {
			char c;

			if (error = uiomove(&c, 1, UIO_WRITE, uio))
				break;
			if ((tp->t_xflag&FLUSHO) == 0) {
				ttxput(tp, c, 0);
				tp->t_rocount = 0;
			}
		}
	}
	s = spl8();
	(*tp->t_proc)(tp, T_OUTPUT);
	splx(s);
	return(error);
}

#define	lobyte(X)	(((unsigned char *)&X)[1])
#define	hibyte(X)	(((unsigned char *)&X)[0])

/*
 * Place a character on raw TTY input queue, putting in delimiters
 * and waking up top half as needed.
 * Also echo if required.
 */
ttin(tp, code)
register struct tty *tp;
{
	register c;
	register iflg, lflg;
	register char *cp;
	ushort nchar;

	if (code == L_BREAK){
		gsignal(tp->t_pgrp, SIGINT);
		ttyflush(tp, (FREAD|FWRITE));
		return;
	}
	nchar = tp->t_rbuf.c_size - tp->t_rbuf.c_count;
	/* reinit rx control block */
	tp->t_rbuf.c_count = tp->t_rbuf.c_size;
	if (nchar==0)
		return;
	iflg = tp->t_iflag;
	lflg = tp->t_lflag;
	/* KMC does all but IXOFF */
	if (tp->t_state&EXTPROC)
		iflg &= IXOFF;
	cp = tp->t_rbuf.c_ptr;
	if (nchar < cfreelist.c_size || (iflg&(INLCR|IGNCR|ICRNL|IUCLC))
	    || (lflg&(ISIG|ICANON|IBRK|XCASE|ECHO|ECHONL))) {
			/* must do per character processing */
		for ( ;nchar--; cp++) {
			c = *cp;
			if (tp->t_state & LNCH) {
				if (putc(c, &tp->t_rawq))
					continue;
				tp->t_state &= ~LNCH;
				if (lflg&ICANON)
					goto setbol;
				else
					goto echoit;
			}
			if (c == '\n' && iflg&INLCR)
				c = '\r';
			else if (c == '\r')
				if (iflg&IGNCR)
					continue;
				else if (iflg&ICRNL)
					c = '\n';
			if (iflg&IUCLC && 'A' <= c && c <= 'Z')
				c += 'a' - 'A';
			if (tp->t_xflag&NEWTTY && tp->t_lflag&(ICANON|ISIG)) {
				if (c == tp->t_cc[VLNEXT]) {
					if (putc(c, &tp->t_rawq))
						continue;
					tp->t_state |= LNCH;
					if (lflg&ICANON)
						goto setbol;
					else
						goto echoit;
				}
				if (c == tp->t_cc[VFLUSH]){
					if (tp->t_xflag&FLUSHO)
						tp->t_xflag &= ~FLUSHO;
					else {
						ttyflush(tp, FWRITE);
						if (lflg&ECHO) {
							if (tp->t_xflag&CTLECH)
								ttecho(tp, c);
							if (tp->t_rawq.c_cc + tp->t_canq.c_cc)
								ttretype(tp);
						}
						tp->t_xflag |= FLUSHO;
					}
					(*tp->t_proc)(tp, T_OUTPUT);
					tk_nin++;
					sysinfo.rawch++;
					continue;
				} else
				        tp->t_xflag &= ~FLUSHO;
			}
			if (lflg&ISIG) {
				if (c == tp->t_cc[VINTR]) {
					gsignal(tp->t_pgrp, SIGINT);
					if (!(lflg&NOFLSH))
						ttyflush(tp, (FREAD|FWRITE));
					if (tp->t_xflag&CTLECH)
						goto echoit;
					else {
						tk_nin++;
						sysinfo.rawch++;
					}
					continue;
				}
				if (c == tp->t_cc[VQUIT]) {
					gsignal(tp->t_pgrp, SIGQUIT);
					if (!(lflg&NOFLSH))
						ttyflush(tp, (FREAD|FWRITE));
					if (tp->t_xflag&CTLECH)
						goto echoit;
					else {
						tk_nin++;
						sysinfo.rawch++;
					}
					continue;
				}
				if (c == tp->t_cc[VSWTCH]) {
					if (!(lflg&NOFLSH))
						ttyflush(tp, FREAD);
					(*tp->t_proc)(tp, T_SWTCH);
					if (tp->t_xflag&CTLECH)
						goto echoit;
					else {
						tk_nin++;
						sysinfo.rawch++;
					}
					continue;
				}
				if (tp->t_xflag&NEWTTY) {
					if (c == tp->t_cc[VSUSP]) {
						gsignal(tp->t_pgrp, SIGTSTP);
						if (!(lflg&NOFLSH))
							ttyflush(tp, FREAD);
						if (tp->t_xflag&CTLECH)
							goto echoit;
						else {
							tk_nin++;
							sysinfo.rawch++;
						}
						continue;
					}
				}
			}
			if (lflg&ICANON) {
				if (c == '\n') {
					if (lflg&ECHONL)
						lflg |= ECHO;
					if (putc(c, &tp->t_rawq))
						continue;
					tp->t_delct++;
					tp->t_lncount = 0;
					tp->t_rocount = 0;
					goto echoit;
				} else if (c == tp->t_cc[VEOL] || c == tp->t_cc[VEOL2]) {
					if (putc(c, &tp->t_rawq))
						continue;
					tp->t_delct++;
					tp->t_lncount = 0;
					tp->t_rocount = 0;
					goto echoit;
				} else {
					if (!(tp->t_state&CLESC)) {
						if (c == '\\') {
							if (putc(c, &tp->t_rawq))
								continue;
							tp->t_state |= CLESC;
							goto setbol;
						} else if (c == tp->t_cc[VERASE]) {
							if (tp->t_lncount) {
								tterase(tp, unputc(&tp->t_rawq));
								(*tp->t_proc)(tp, T_OUTPUT);
							}
							tk_nin++;
							sysinfo.rawch++;
							continue;
						} else if (c == tp->t_cc[VKILL]) {
							if (tp->t_xflag&CRTKIL &&
							    tp->t_lncount == tp->t_rocount) {
								while (tp->t_lncount)
									tterase(tp, unputc(&tp->t_rawq));
							} else {
								if (lflg&ECHO)
									ttecho(tp, c);
								if (lflg&ECHOK)
									ttxput(tp, '\n', 0);
								while (tp->t_lncount) {
									(void) unputc(&tp->t_rawq);
									tp->t_lncount--;
								}
								tp->t_rocount = 0;
							}
							tp->t_state &= ~LOCAL;
							(*tp->t_proc)(tp, T_OUTPUT);
							continue;
						} else if (c == tp->t_cc[VEOF]) {
							if (putc(c, &tp->t_rawq))
								continue;
							if (tp->t_state&ERASE) {
								tp->t_state &= ~ERASE;
								ttxput(tp, '/', 0);
							}
							if ((lflg&ECHO) && (tp->t_xflag&CTLECH)) {
								register int i;

								i = tp->t_col;
								ttecho(tp, c);
								i = MIN(2, tp->t_col - i);
								while (i > 0) {
									ttxput(tp, '\b', 0);
									i--;
								}
								(*tp->t_proc)(tp, T_OUTPUT);
							}
							tp->t_delct++;
							tp->t_lncount = 0;
							tp->t_rocount = 0;
							tk_nin++;
							sysinfo.rawch++;
							continue;
						} else {
							if (tp->t_xflag&NEWTTY) {
								if (c == tp->t_cc[VWERASE]) {
									tk_nin++;
									sysinfo.rawch++;
									if (tp->t_lncount != 0) {
										ttwerase(tp);
										if (lflg&ECHO)
											(*tp->t_proc)(tp, T_OUTPUT);
									}
									continue;
								}
								if (c == tp->t_cc[VRPRNT]) {
									tk_nin++;
									sysinfo.rawch++;
									if (lflg&ECHO) {
										ttretype(tp);
										(*tp->t_proc)(tp, T_OUTPUT);
									}
									continue;
								}
							}
						}
					} else {
						if (lflg&XCASE) {
							tp->t_state &= ~CLESC;
							if ((c < 0200) && maptab[c]) {
								register struct cblock *bp = tp->t_rawq.c_cl;

								/*
								 * Replace the \
								 * at the end of
								 * the raw queue
								 * with the
								 * mapped character.
								 */
								bp->c_data[bp->c_last - 1] = maptab[c];
								goto setbol;
							} else if (c == '\\')
								goto echoit;
						} else {
							if (putc(c, &tp->t_rawq))
								continue;
							if (c != '\\')
								tp->t_state &= ~CLESC;
							goto setbol;
						}
					}
					if (putc(c, &tp->t_rawq))
						continue;
setbol:
					/*
					 * Set up stuff indicating the beginning
					 * of a line.
					 */
					if (tp->t_lncount == 0) {
						register struct cblock *bp = tp->t_rawq.c_cf;

						tp->t_beglin.cr_cb = bp;
						tp->t_beglin.cr_ndx = bp->c_last - 1;
					}
					tp->t_lncount++;
					if (tp->t_rocount == 0)
						tp->t_rocol = tp->t_col;
					tp->t_rocount++;
				}
			} else if ((lflg&IBRK) && c == tp->t_cc[VBRK]) {
				if (putc(c, &tp->t_rawq))
					continue;
				tp->t_delct = 1;
			} else {
				if (putc(c, &tp->t_rawq))
					continue;
			}
echoit:
			if (tp->t_state&ERASE) {
				tp->t_state &= ~ERASE;
				ttxput(tp, '/', 0);
			}
			if (lflg&ECHO) {
				ttecho(tp, c);
 				(*tp->t_proc)(tp, T_OUTPUT);
			}
			tk_nin++;
			sysinfo.rawch++;
		}
	} else {
		/* may do block processing */
		putcb(CMATCH((struct cblock *)cp), &tp->t_rawq);
		tk_nin += nchar;
		sysinfo.rawch += nchar;
		/* allocate new rx buffer */
		if ((tp->t_rbuf.c_ptr = getcf()->c_data)
			== ((struct cblock *)NULL)->c_data) {
			tp->t_rbuf.c_ptr = NULL;
			return;
		}
		tp->t_rbuf.c_count = cfreelist.c_size;
		tp->t_rbuf.c_size = cfreelist.c_size;
	}

	if (tp->t_rawq.c_cc > TTXOHI) {
		if (iflg&IXOFF && !(tp->t_state&TBLOCK))
			(*tp->t_proc)(tp, T_BLOCK);
		if (tp->t_rawq.c_cc > TTYHOG) {
			if (tp->t_outq.c_cc < tthiwat[tp->t_cflag&CBAUD] &&
			    (tp->t_xflag&NEWTTY))
				ttxput(tp, CTRL(g), 0);
			else
				ttyflush(tp, FREAD);
			return;
		}
	}
	if (!(lflg&ICANON)) {
		tp->t_state &= ~RTO;
		if (tp->t_rawq.c_cc >= tp->t_cc[VMIN])
			tp->t_delct = 1;
		else if (tp->t_cc[VTIME]) {
			if (!(tp->t_state&TACT))
				tttimeo(tp);
		}
	}
	if (tp->t_delct) {
		if (tp->t_state&IASLP) {
			tp->t_state &= ~IASLP;
			tp->t_state &= ~CLESC;
			wakeup((caddr_t)&tp->t_rawq);
		}
		if (tp->t_rsel) {
			selwakeup(tp->t_rsel, tp->t_state&RCOLL);
			tp->t_state &= ~RCOLL;
			tp->t_rsel = 0;
		}
		if (tp->t_state&ASYNC)
			gsignal(tp->t_pgrp, SIGIO);
	}
}

/*
 * Erase a character from the current line.
 */
tterase(tp, c)
	register struct tty *tp;
	register int c;
{
	tp->t_lncount--;
	ttrub(tp, c);

	/*
	 * If character is one of the special characters, the only way it
	 * could have gotten into the clist was if it was escaped by \ or
	 * ^V.  If so, remove the \ or ^V.
	 */
	if (c == tp->t_cc[VINTR] ||
	    c == tp->t_cc[VQUIT] ||
	    c == tp->t_cc[VERASE] ||
	    c == tp->t_cc[VKILL] ||
	    c == tp->t_cc[VEOF] ||
	    c == tp->t_cc[VSWTCH] ||
	    c == tp->t_cc[VSUSP] ||
	    c == tp->t_cc[VDSUSP] ||
	    c == tp->t_cc[VRPRNT] ||
	    c == tp->t_cc[VFLUSH] ||
	    c == tp->t_cc[VWERASE] ||
	    c == tp->t_cc[VLNEXT]) {
		c = unputc(&tp->t_rawq);
		/*
		 * the above comment is wrong in certain cases. A literal
		 * next character (^V) CAN be in the input queue all by 
		 * itself when we reach this point. For example, if the
		 * user types a ^V followed by the letter 'A' (or any other
		 * character that did NOT require a literal next to precede
		 * it), and then he does a kill line, we will get to the
		 * ^V, rub it out (at the top of this routine), then see
		 * that it is a literal next and try to remove the preceding
		 * character. However, there is none, so we make lncount
		 * negative and call ttrub (with humorous consequences).
		 * Anyways, the solution is to make damn sure that this
		 * character was preceded by a literal next, delete it if
		 * so, otherwise put it back on the queue. The -1 check
		 * is there because unputc returns -1 if the queue was
		 * empty. We don't want to put a -1 into onto the rawq.
		 * (How's that for well documented code???)
		 */
		if ((c == (int) '\\') || (c == (int) tp->t_cc[VLNEXT])) {
			tp->t_lncount--;
			ttrub(tp, c);
		}
		else if (c != -1)
			putc (c, &tp->t_rawq);	
	}
}

/*
 * Erase a "word" from the current line.
 */
ttwerase(tp)
	register struct tty *tp;
{
	register int c;

	do {
		c = unputc(&tp->t_rawq);
		if (c != ' ' && c != '\t')
			goto erasenb;
		tterase(tp, c);
	} while (tp->t_lncount);
	return;

erasenb:
	do {
		tterase(tp, c);
		if (tp->t_lncount == 0)
			return;
		c = unputc(&tp->t_rawq);
	} while (c != ' ' && c != '\t');
	(void) putc(c, &tp->t_rawq);
}

/*
 * Rubout one character from the input line
 * as cleanly as possible.
 */
ttrub(tp, c)
	register struct tty *tp;
	register c;
{
	register int col;
	struct cur_clist ccl;
	int s;

	if ((tp->t_lflag&ECHO) == 0)
		return;
	tp->t_xflag &= ~FLUSHO;
	c &= 0377;
	if ((tp->t_lflag&ECHOE) || (tp->t_xflag&CRTBS)) {
		if (tp->t_rocount == 0) {
			/*
			 * Screwed by ttwrite; retype
			 */
			ttretype(tp);
			return;
		}
		switch (partab[c]&077) {

		case 9:	/* XCASE character */
			if (tp->t_lflag&XCASE)
				ttbs(tp, 2);
			else
				ttbs(tp, 1);
			break;

		case 0:	/* ordinary */
			ttbs(tp, 1);
			break;

		case 2:	/* backspace */
			if (tp->t_xflag&CTLECH)
				ttbs(tp, 2);
			else
				ttxput(tp, ' ', 0);
			break;

		case 1:	/* non-printing */
		case 3:	/* line feed */
		case 5:	/* vertical tab */
		case 6:	/* carriage return */
		case 7:	/* form feed */
		case 8:	/* QESC character */
			if (tp->t_xflag&CTLECH)
				ttbs(tp, 2);
			break;

		case 4:	/* tab */
			if (tp->t_rocount < tp->t_lncount) {
				ttretype(tp);
				return;
			}
			s = spl8();
			col = tp->t_rocol;
			if (tp->t_lncount != 0) {
				ccl = tp->t_beglin;
				while ((c = nextc(&ccl)) != -1) {
					switch (partab[c]&077) {

					case 0:	/* ordinary */
					case 9:	/* XCASE character */
						col++;

					case 1:	/* non-printing */
					case 5:	/* vertical tab */
					case 7:	/* form feed */
					case 8:	/* QESC character */
						break;

					case 2:	/* backspace */
						if (col)
							col--;
						break;

					case 3:	/* line feed */
					case 6:	/* carriage return */
						col = 0;
						break;

					case 4:	/* tab */
						col += 8 - (col&07);
						break;
					}
				}
			}
			splx(s);
			/*
			 * col will now be length of the tab
			 */
			col = tp->t_col - col;
			if (col > 8)
				col = 8;		/* overflow screw */
			while (--col >= 0)
				ttxput(tp, '\b', 0);
			break;

		default:
			panic("ttrub");
		}
	} else if (tp->t_xflag&PRTERA) {
		if ((tp->t_state&ERASE) == 0) {
			ttxput(tp, '\\', 0);
			tp->t_state |= ERASE;
		}
		ttecho(tp, c);
	} else
		ttecho(tp, tp->t_cc[VERASE]);
	tp->t_rocount--;
}

/*
 * Crt back over cnt chars.  If ECHOE is set, erase the chars when backing
 * over them, otherwise just backspace without erasing.  CRTBS or
 * ECHOE must be set for this routine to be called.
 */
ttbs(tp, cnt)
	register struct tty *tp;
	int cnt;
{
	while (--cnt >= 0) {
	        if (tp->t_lflag&ECHOE) {
			ttxput(tp, '\b', 0);
			ttxput(tp, ' ', 0);
			ttxput(tp, '\b', 0);
		} else                          /* CRTBS set */
			ttxput(tp, '\b', 0);
	}
}

/*
 * Reprint the rawq line.
 * We assume c_cc has already been checked.
 */
ttretype(tp)
	register struct tty *tp;
{
	struct cur_clist ccl;
	register int c;
	int s;

	if (tp->t_cc[VRPRNT] != 0377)
		ttecho(tp, tp->t_cc[VRPRNT]);
	ttxput(tp, '\n', 0);
	s = spl8();
	setcclist(&ccl, &tp->t_canq);
	while ((c = nextc(&ccl)) != -1)
		ttecho(tp, c);
	setcclist(&ccl, &tp->t_rawq);
	while ((c = nextc(&ccl)) != -1)
		ttecho(tp, c);
	tp->t_state &= ~ERASE;
	splx(s);
	tp->t_rocount = tp->t_lncount;
	tp->t_rocol = 0;
}

/*
 * Echo a typed character to the terminal
 */
ttecho(tp, c)
	register struct tty *tp;
	register c;
{
	c &= 0377;
	if (tp->t_xflag&CTLECH) {
		if (c <= 037 && c != '\t' && c != '\n' || c == 0177) {
			ttxput(tp, '^', 0);
			if (c == 0177)
				c = '?';
			else if (tp->t_oflag&OLCUC)
				c += 'a' - 1;
			else
				c += 'A' - 1;
		}
	}
	ttxput(tp, c, 0);
}

#if !defined(vax) || defined(lint)
/*
 * Put character(s) on TTY output queue, adding delays,
 * expanding tabs, and handling the CR/NL bit.
 * It is called both from the base level for output, and from
 * interrupt level for echoing.
 */
ttxput(tp, ucp, ncode)
register struct tty *tp;
union {
	uint	ch;
	struct cblock *ptr;
} ucp;
{
	register c;
	register flg;
	register unsigned char *cp;
	register char *colp;
	register struct clist *tq;
	int ctype;
	int cs;
	struct cblock *scf;
	char c1;

	/*
	 * the following is to make sure we do not consume all of the free
	 * clists. this can happen if the user types a XOFF and then holds
	 * down a key. the output queue will grow until the system dies.
	 * this can also happen if a modem and the vioc get into a shouting
	 * match. the following checks to see if we are too large and, if so,
	 * we return. the number should insure that writes will complete 
	 * even if the user is a fast typist.
	 */
	if (tp->t_outq.c_cc > (tthiwat[tp->t_cflag&CBAUD] + 2*CBSIZE + TTYHOG))
		return ;

	if (tp->t_xflag&FLUSHO) {
		if (ncode != 0)
			putcf(ucp.ptr);
		return;
	}
	tq = &tp->t_outq;
	if (tp->t_state&EXTPROC) {
		if (tp->t_lflag&XCASE)
			flg = OPOST;
		else
			flg = 0;
	} else
		flg = tp->t_oflag;
	if (ncode == 0) {
		ncode++;
		if (!(flg&OPOST)) {
			tk_nout++;
			sysinfo.outch++;
			putc(ucp.ch, tq);
			return;
		}
		c1 = (char)ucp.ch;
		cp = (unsigned char *)&c1;
		scf = NULL;
	} else {
		if (!(flg&OPOST)) {
			tk_nout += ncode;
			sysinfo.outch += ncode;
			putcb(ucp.ptr, tq);
			return;
		}
		cp = (unsigned char *)&ucp.ptr->c_data[ucp.ptr->c_first];
		scf = ucp.ptr;
	}
	while (ncode--) {
		c = *cp++;
		cs = c;
		/*
		 * Calculate delays.
		 * The numbers here represent clock ticks
		 * and are not necessarily optimal for all terminals.
		 * The delays are indicated by characters above 0200.
		 */
		ctype = partab[c];
		colp = &tp->t_col;
		c = 0;
		switch (ctype&077) {

		case 0:	/* ordinary */
			if (flg&OLCUC && 'a' <= cs && cs <= 'z')
				cs += 'A' - 'a';
			(*colp)++;

		case 1:	/* non-printing */
			break;

		case 2:	/* backspace */
			if (flg&BSDLY)
				c = 2;
			if (*colp)
				(*colp)--;
			break;

		case 3:	/* line feed */
			if (flg&ONLRET)
				goto cr;
			if (flg&ONLCR) {
				if (!(flg&ONOCR && *colp==0)) {
					tk_nout++;
					sysinfo.outch++;
					putc('\r', tq);
				}
				goto cr;
			}
		nl:
			if (flg&NLDLY)
				c = 5;
			break;

		case 4:	/* tab */
			c = 8 - ((*colp)&07);
			*colp += c;
			ctype = flg&TABDLY;
			if (ctype == TAB0) {
				c = 0;
			} else if (ctype == TAB1) {
				if (c < 5)
					c = 0;
			} else if (ctype == TAB2) {
				c = 2;
			} else if (ctype == TAB3) {
				tk_nout += c;
				sysinfo.outch += c;
				do
					putc(' ', tq);
				while (--c);
				continue;
			}
			break;

		case 5:	/* vertical tab */
			if (flg&VTDLY)
				c = 0177;
			break;

		case 6:	/* carriage return */
			if (flg&OCRNL) {
				cs = '\n';
				goto nl;
			}
			if (flg&ONOCR && *colp == 0)
				continue;
		cr:
			ctype = flg&CRDLY;
			if (ctype == CR1) {
				if (*colp)
					c = max((*colp>>4) + 3, 6);
			} else if (ctype == CR2) {
				c = 5;
			} else if (ctype == CR3) {
				c = 9;
			}
			*colp = 0;
			break;

		case 7:	/* form feed */
			if (flg&FFDLY)
				c = 0177;
			break;

		case 8:	/* QESC character */
			if ((flg&(NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY)) == 0)
				/*
				 * Not doing delay processing, no need to
				 * escape the escape character.
				 */
				break;
			/*
			 * Escape the escape character.
			 */
			putc(QESC, tq);
			tk_nout++;
			sysinfo.outch++;
			break;

		case 9:	/* XCASE character */
			if ((tp->t_lflag&XCASE) == 0) {
				/*
				 * Not doing XCASE processing.
				 */
				if (tp->t_xflag&TILDE && cs == '~')
			        	cs = '`';
				(*colp)++;	/* XCASE are printable */
				break;
			}
			/*
			 * Generate escapes for upper-case-only terminals.
			 */
			if ('A' <= cs && cs <= 'Z') {
				putc('\\', tq);
				*colp += 2;
				tk_nout++;
				sysinfo.outch++;
				break;
			}

			colp = "({)}!|^~'`\\\\";
			while(*colp++)
				if (cs == *colp++) {
					putc('\\', tq);
					tp->t_col += 2;
					tk_nout++;
					sysinfo.outch++;
					cs = colp[-2];
					break;
				}
			break;
		}
		tk_nout++;
		sysinfo.outch++;
		putc(cs, tq);
		if (c) {
			if ((c < 32) && flg&OFILL) {
				if (flg&OFDEL)
					cs = 0177;
				else
					cs = 0;
				putc(cs, tq);
				if (c > 3)
					putc(cs, tq);
			} else {
				putc(QESC, tq);
				putc(c|0200, tq);
			}
		}

	}
	if (scf != NULL)
		putcf(scf);
}
#endif

/*
 * Get next packet from output queue.
 * Called from xmit interrupt complete.
 */

ttout(tp)
register struct tty *tp;
{
	register struct ccblock *tbuf;
	register c;
	register char *cptr;
	register retval;
	extern ttrstrt();

	if (tp->t_state&TTIOW && tp->t_outq.c_cc==0) {
		tp->t_state &= ~TTIOW;
		wakeup((caddr_t)&tp->t_oflag);
	}
	tbuf = &tp->t_tbuf;
delay :	if (hibyte(tp->t_lflag)) {
		if (tbuf->c_ptr) {
			putcf(CMATCH((struct cblock *)tbuf->c_ptr));
			tbuf->c_ptr = NULL;
		}
		tp->t_state |= TIMEOUT;
		timeout(ttrstrt, (caddr_t)tp, (hibyte(tp->t_lflag)&0177)+6);
		hibyte(tp->t_lflag) = 0;
		return(0);
	}
	retval = 0;

	c = tp->t_oflag;
	if ((tp->t_state&EXTPROC) || !(c&OPOST) ||
	    !(c&(NLDLY|CRDLY|TABDLY|BSDLY|VTDLY|FFDLY))) {
		/*
		 * Can do block processing because:
		 *    KMC is doing delay processing for us or
		 *    we're not doing any output processing at all or
		 *    no delays are required for this line.
		 */
		if (tbuf->c_ptr)
			putcf(CMATCH((struct cblock *)tbuf->c_ptr));
		if ((tbuf->c_ptr = (char *)getcb(&tp->t_outq)) == NULL)
			return(0);
		tbuf->c_count = ((struct cblock *)tbuf->c_ptr)->c_last -
				((struct cblock *)tbuf->c_ptr)->c_first;
		tbuf->c_size = tbuf->c_count;
		tbuf->c_ptr = &((struct cblock *)tbuf->c_ptr)->c_data
				[((struct cblock *)tbuf->c_ptr)->c_first];
		retval = CPRES;
	} else {		/* watch for timing	*/
		if (tbuf->c_ptr == NULL) {
			if ((tbuf->c_ptr = getcf()->c_data)
				== ((struct cblock *)NULL)->c_data) {
				tbuf->c_ptr = NULL;
				return(0);	/* Add restart? */
			}
		}
		tbuf->c_count = 0;
		cptr = tbuf->c_ptr;
		while ((c=getc(&tp->t_outq)) >= 0) {
			if (c == QESC) {
				if ((c = getc(&tp->t_outq)) < 0)
					break;
				if (c > 0200 && !(tp->t_state&EXTPROC)) {
					hibyte(tp->t_lflag) = c;
					if (!retval)
						goto delay;
					break;
				}
			}
			retval = CPRES;
			*cptr++ = c;
			tbuf->c_count++;
			if (tbuf->c_count >= cfreelist.c_size)
				break;
		}
		tbuf->c_size = tbuf->c_count;
	}

	if (tp->t_outq.c_cc<=ttlowat[tp->t_cflag&CBAUD]) {
		if (tp->t_state&OASLP) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}
		if (tp->t_wsel) {
			selwakeup(tp->t_wsel, tp->t_state & WCOLL);
			tp->t_wsel = 0;
			tp->t_state &= ~WCOLL;
		}
	}
	return(retval);
}

tttimeo(tp)
register struct tty *tp;
{
	tp->t_state &= ~TACT;
	if (tp->t_lflag&ICANON || tp->t_cc[VTIME] == 0)
		return;
	if (tp->t_rawq.c_cc == 0 && tp->t_cc[VMIN])
		return;
	if (tp->t_state&RTO) {
		tp->t_delct = 1;
		if (tp->t_state&IASLP) {
			tp->t_state &= ~IASLP;
			wakeup((caddr_t)&tp->t_rawq);
		}
		if (tp->t_rsel) {
			selwakeup(tp->t_rsel, tp->t_state&RCOLL);
			tp->t_state &= ~RCOLL;
			tp->t_rsel = 0;
		}
		if (tp->t_state&ASYNC)
			gsignal(tp->t_pgrp, SIGIO);
	} else {
		tp->t_state |= RTO|TACT;
		timeout(tttimeo, (caddr_t)tp, tp->t_lflag&IBRK? tp->t_cc[VTIME] : tp->t_cc[VTIME]*(hz/10));
	}
}

/*
 * I/O control interface
 */
/*ARGSUSED*/
ttioctl(tp, cmd, arg, mode)
register struct tty *tp;
caddr_t arg;
{
	register s;
	ushort	chg;

	switch(cmd) {
	case LDOPEN:
		if (tp->t_rbuf.c_ptr == NULL) {
			/* allocate RX buffer */
			while((tp->t_rbuf.c_ptr = getcf()->c_data)
				== ((struct cblock *)NULL)->c_data) {
				tp->t_rbuf.c_ptr = NULL;
				cfreelist.c_flag = 1;
				slumber((caddr_t)&cfreelist, INTERRUPTABLE,
					resboost);
			}
			tp->t_rbuf.c_count = cfreelist.c_size;
			tp->t_rbuf.c_size  = cfreelist.c_size;
			(*tp->t_proc)(tp, T_INPUT);
		}
		break;

	case LDCLOSE:
		s = spl8();
		(*tp->t_proc)(tp, T_RESUME);
		splx(s);
		ttywait(tp);
		ttyflush(tp, (FREAD|FWRITE));
		if (tp->t_tbuf.c_ptr) {
			putcf(CMATCH((struct cblock *)tp->t_tbuf.c_ptr));
			tp->t_tbuf.c_ptr = NULL;
			tp->t_tbuf.c_count = 0;
			tp->t_tbuf.c_size = 0;
		}
		if (tp->t_rbuf.c_ptr) {
			putcf(CMATCH((struct cblock *)tp->t_rbuf.c_ptr));
			tp->t_rbuf.c_ptr = NULL;
			tp->t_rbuf.c_count = 0;
			tp->t_rbuf.c_size = 0;
		}

	case LDCHG:
		chg = tp->t_lflag^(ushort)arg;
		if (!(chg&ICANON))
			break;
		s = spl8();
		if (tp->t_canq.c_cc) {
			if (tp->t_rawq.c_cc) {
				tp->t_canq.c_cc += tp->t_rawq.c_cc;
				tp->t_canq.c_cl->c_next = tp->t_rawq.c_cf;
				tp->t_canq.c_cl = tp->t_rawq.c_cl;
			}
			tp->t_rawq = tp->t_canq;
			tp->t_canq = ttnulq;
		}
		tp->t_delct = tp->t_rawq.c_cc;
		splx(s);
		break;

	default:
		break;
	}

	return(0);
}
