/*	tty_pty.c	6.3	83/10/01	*/

/*
 * Pseudo-teletype Driver
 * (Actually two drivers, requiring two entries in 'cdevsw')
 */
#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/tty.h"
#include "../sys/termio.h"
#include "../sys/ioctl.h"
#include "../sys/user.h"
#include "../sys/conf.h"
#include "../sys/file.h"
#include "../sys/proc.h"
#include "../sys/uio.h"
#include "../sys/kernel.h"
#include "../sys/tty_pty.h"

/*
 * pts == /dev/tty[pP]?
 * ptc == /dev/ptp[pP]?
 */

#define	PF_RCOLL	0x0001	/* multiple processes wanting to read master */
#define	PF_WCOLL	0x0002	/* multiple processes wanting to write master */
#define	PF_OASLP	0x0004	/* sleeping waiting to write */
#define	PF_IASLP	0x0008	/* sleeping waiting to read */
#define	PF_PKT		0x0010	/* packet mode */
#define	PF_STOPPED	0x0020	/* user told stopped */
#define	PF_REMOTE	0x0040	/* remote and flow controlled input */
#define	PF_NOSTOP	0x0080  /* no more ^s, ^q */
#define PF_NBIO		0x0100	/* non-blocking i/o (nr) */

int	ptsproc();

/*
 * External variables.
 */
extern	int		npty;
extern	struct tty	pt_tty[];
extern	struct pt_ioctl	pt_ioctl[];


/*
 * Open the slave device of the pair.  The open will only show
 * "carrier" when the master device has been opened read/write.
 */
/*ARGSUSED*/
ptsopen(dev, flag)
	dev_t dev;
{
	register struct tty *tp;

	if (minor(dev) >= npty)
		/*
		 * No such device.
		 */
		return (ENXIO);
	tp = &pt_tty[minor(dev)];
	/*
	 * If this is the first open on the line, initialize
	 * the modes.
	 */
	if ((tp->t_state&(ISOPEN|WOPEN)) == 0) {
		ttinit(tp);
		tp->t_proc = ptsproc;
		tp->t_dev = dev;
	} else if (tp->t_state&XCLUDE && u.u_uid!=0)
		return EBUSY;	/* device is busy, sorry */

	/*
	 * Wait for carrier.
	 */
	if ((tp->t_cflag&CLOCAL) || (tp->t_state&OPROC))
		tp->t_state |= CARR_ON;
	if (!(flag&FNDELAY))
		while ((tp->t_state&CARR_ON)==0) {
			tp->t_state |= WOPEN;
			slumber((caddr_t)&tp->t_canq, INTERRUPTABLE, ttiboost);
		}
	if ((tp->t_state & ISOPEN) == 0) {
		tp->t_iflag = IXON|ICRNL|ISTRIP;
		tp->t_oflag = ONLCR|OPOST;
		tp->t_cflag = HUPCL|SSPEED|CS8|CREAD;
/*		tp->t_lflag = ECHO|ECHOK|ICANON|ISIG;*/
		tp->t_lflag = ICANON|ISIG;
	}
	return((*linesw[tp->t_line].l_open)(tp));
}

/*
 * Close the slave device.
 */
ptsclose(dev)
	dev_t dev;
{
	register struct tty *tp;
	int	error = 0;

	tp = &pt_tty[minor(dev)];
	error = (*linesw[tp->t_line].l_close)(tp);
	/*
	 * Tell the master reader the party's over.
	 */
	ptcwakeup(tp, &pt_ioctl[minor(dev)], 1);
	tp->t_state &= OPROC;
	return(error);
}

/*
 * Read from the slave device.
 */
ptsread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp = &pt_tty[minor(dev)];
	register struct pt_ioctl *pti = &pt_ioctl[minor(dev)];
	register struct clist *tq;
	int error = 0;

again:
	if (pti->pt_state&PF_REMOTE) {
		if (tp->t_canq.c_cc == 0) {
			if (tp->t_rawq.c_cc == 0)
				tp->t_delct = 0;
			if (tp->t_delct == 0) {
				if (u.u_fmode&FNDELAY)
					return (0);

				if (tp->t_state & NBIO)		/* (nr) */
					return(EWOULDBLOCK);

				tp->t_state |= IASLP;
				slumber((caddr_t)&tp->t_rawq, INTERRUPTABLE,
					ttiboost);
				goto again;
			}
			tp->t_canq = tp->t_rawq;
			tp->t_rawq = ttnulq;
			tp->t_delct = 0;
		}
		tq = &tp->t_canq;
		while (uio->uio_resid > 0 && error == 0) {
			if (uio->uio_resid >= CBSIZE) {
				register n;
				register struct cblock *cp;

				if ((cp = getcb(tq)) == NULL)
					break;
				n = min(uio->uio_resid, cp->c_last - cp->c_first);
				error = uiomove(&cp->c_data[cp->c_first], n,
				    UIO_READ, uio);
				putcf(cp);
			} else {
				register c;

				if ((c = getc(tq)) < 0)
					break;
				error = ureadc(c, uio);
			}
		}
	} else {
		error = (*linesw[tp->t_line].l_read)(tp, uio);
		if (tp->t_rawq.c_cc <= TTYHOG/2 || tp->t_delct == 0) {
			/*
			 * OK to place input onto the slave device.
			 */
			if (pti->pt_state&PF_OASLP) {
				pti->pt_state &= ~PF_OASLP;
				wakeup((caddr_t)&pti->pt_outq);
			}
			if (pti->pt_wsel) {
				selwakeup(pti->pt_wsel, pti->pt_state&PF_WCOLL);
				pti->pt_wsel = 0;
				pti->pt_state &= ~WCOLL;
			}
		}
	}
	return(error);
}

/*
 * Write to the slave device.  Wakeups of the controlling tty will happen
 * when "ptsproc(tp, T_OUTPUT)" is called.
 */
ptswrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp;
	int error;

	tp = &pt_tty[minor(dev)];
	error = (*linesw[tp->t_line].l_write)(tp, uio);
	return(error);
}

/*
 * Start output on pseudo-tty.
 * Wake up process selecting or sleeping for input from controlling tty.
 */
ptsproc(tp, cmd)
	struct tty *tp;
{
	register struct pt_ioctl *pti = &pt_ioctl[minor(tp->t_dev)];
	register struct cblock *cp;

	switch (cmd) {
	case T_WFLUSH:
	case T_RESUME:
		tp->t_state &= ~TTSTOP;
		goto start;

	case T_TIME:
	case T_OUTPUT:
	start:
		if (tp->t_state&TTSTOP)
			break;
		if (pti->pt_state & PF_STOPPED) {
			pti->pt_state &= ~PF_STOPPED;
			pti->pt_send = TIOCPKT_START;
		}
		if (tp->t_state&OPROC) {
			/*
			 * Jab the master reader.
			 */
			ptcwakeup(tp, pti, 1);
			return;
		}
		/*
		 * Master reader went away, not likely to
		 * have the output queue drained -- flush it ourselves.
		 */
		while ((cp = getcb(&tp->t_outq)) != NULL)
			putcf(cp);
		if (tp->t_state&OASLP) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}
		if (tp->t_state&TTIOW) {
			tp->t_state &= ~TTIOW;
			wakeup((caddr_t)&tp->t_oflag);
		}
		break;

	case T_SUSPEND:
		tp->t_state |= TTSTOP;
		pti->pt_state |= PF_STOPPED;
		pti->pt_send |= TIOCPKT_STOP;
		ptcwakeup(tp, pti, 1);
		break;

	default:
		break;
	}
}

/*
 * Wake up the master reader if there are characters to be read and
 * force == 0 or if force == 1.
 */
ptcwakeup(tp, pti, force)
	register struct tty *tp;
	register struct pt_ioctl *pti;
{
	/*
	 * Only wake up the master reader if there's something
	 * to be read, or there was a close or flush.
	 */
	if (force || tp->t_outq.c_cc > 0) {
		/*
		 * Wake up anybody wanting to read from the slave
		 * device.
		 */
		if (pti->pt_state&PF_IASLP) {
			pti->pt_state &= ~PF_IASLP;
			wakeup((caddr_t)&pti->pt_canq);
		}
		/*
		 * Jab those wanting to read from the slave device.
		 */
		if (pti->pt_rsel) {
			selwakeup(pti->pt_rsel, pti->pt_state&PF_RCOLL);
			pti->pt_rsel = 0;
			pti->pt_state &= ~RCOLL;
		}
	}
}

/*
 * Open the master device of the master/slave pair.  This implies
 * carrier has come on.
 */
/*ARGSUSED*/
ptcopen(dev, flag)
	dev_t dev;
	int flag;
{
	register struct tty *tp;
	struct pt_ioctl *pti;

	if (minor(dev) >= npty)
		return(ENXIO);
	tp = &pt_tty[minor(dev)];
	if (tp->t_state&OPROC)
		/*
		 * The master has already been opened.
		 */
		return(EBUSY);
	/*
	 * Wake up slave openers waiting for carrier.
	 */
	if (tp->t_state&WOPEN)
		wakeup((caddr_t)&tp->t_canq);
	tp->t_state |= CARR_ON|OPROC;
	tp->t_proc = ptsproc;
	/*
	 * Clear the state word and the "send" buffer.
	 */
	pti = &pt_ioctl[minor(dev)];
	pti->pt_state = 0;
	pti->pt_send = 0;
	/*
 	 * Clear terminal flags in case slave side from previous
	 * open still opens.
	 */
	ttinit(tp);
	return(0);
}

/*
 * Close the master device, signaling loss of carrier.
 */
ptcclose(dev)
	dev_t dev;
{
	register struct tty *tp;

	/*
	 * Signal hangup.
	 */
	tp = &pt_tty[minor(dev)];
	if ((tp->t_state&ISOPEN) || (tp->t_outq.c_cc)) {
		gsignal(tp->t_pgrp, SIGHUP);
		ttyflush(tp, FREAD|FWRITE);
	}
	/*
	 * Turn carrier off.
	 */
	tp->t_state &= ~(CARR_ON|OPROC);
	return(0);
}

/*
 * Read from the output queue of the slave device.
 */
ptcread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp = &pt_tty[minor(dev)];
	struct pt_ioctl *pti = &pt_ioctl[minor(dev)];
	register struct clist *tq = &tp->t_outq;
	int error = 0;

	/*
	 * Make sure that either the master or the slave are open.
	 */
	if ((tp->t_state & (CARR_ON|ISOPEN)) == 0) {
		error = EIO;
		goto out;
	}

	/*
	 * If we're in packet mode, and a flow control signal is pending,
	 * pass it back; else, pass back a 0.
	 */
	if (pti->pt_state & PF_PKT) {
		if (pti->pt_send) {
			error = ureadc(pti->pt_send, uio);
			if (error)
				return(error);
			pti->pt_send = 0;
			return(0);
		}
		error = ureadc(0, uio);
	}

	/*
	 * Wait for the slave writer to place output onto
	 * the output queue.  Read semantics are just like a tty
	 * in -icanon mode -- the read is satisfied when there are
	 * no more characters and we've got at least one character.
	 */
	while (tq->c_cc == 0) {
		if (u.u_fmode & FNDELAY || (tp->t_state & CARR_ON) == 0)
			goto out;

		if (pti->pt_state & PF_NBIO)	/* (nr) */
			return(EWOULDBLOCK);

		pti->pt_state |= PF_IASLP;
		slumber((caddr_t)&pti->pt_canq, INTERRUPTABLE, ttiboost);
	}

	/*
	 * Drain the output queue, doing the necessary wakeups
	 * afterwards exactly like ttout (we don't call ttout
	 * because it would attempt to do delay processing).
	 */
	while (uio->uio_resid > 0 && tq->c_cc > 0 && error == 0) {

		if (uio->uio_resid >= CBSIZE) {
			register n;
			register struct cblock *cp;

			if ((cp = getcb(tq)) == NULL)
				continue;
			n = min(uio->uio_resid, cp->c_last - cp->c_first);
			error = uiomove(&cp->c_data[cp->c_first], n, UIO_READ,
					uio);
			putcf(cp);
		} else {
			register c;

			if ((c = getc(tq)) < 0)
				continue;
			error = ureadc(c, uio);
		}

		/* flow-control is at the slave side. master should
 		   also stop reading */
		while (tp->t_state&TTSTOP) {
			pti->pt_state |= PF_IASLP;
			slumber((caddr_t)&pti->pt_canq, INTERRUPTABLE,
				ttiboost);
		}

	}
out:
	/*
	 * Wake up the slave writer when we've drained the output
	 * a nontrivial amount (in hopes of reducing context switching).
	 */
	if (tq->c_cc <= ttlowat[B9600]) {	/* a bit ugly */
		if (tp->t_state&TTIOW && tq->c_cc == 0) {
			tp->t_state &= ~TTIOW;
			wakeup((caddr_t)&tp->t_oflag);
		}
		if (tp->t_state&OASLP) {
			tp->t_state &= ~OASLP;
			wakeup((caddr_t)&tp->t_outq);
		}
		if (tp->t_wsel) {
			selwakeup(tp->t_wsel, tp->t_state&WCOLL);
			tp->t_wsel = 0;
			tp->t_state &= ~WCOLL;
		}
	}
	return (error);
}

/*
 * A process is doing synchronous, non-blocking I/O on the master device
 * and wants to be informed of any interesting activity.
 */
ptcselect(dev, rw)
	dev_t dev;
	int rw;
{
	register struct tty *tp = &pt_tty[minor(dev)];
	struct pt_ioctl *pti = &pt_ioctl[minor(dev)];
	struct proc *p;

	if ((tp->t_state&ISOPEN) == 0)
		return(1);
	switch (rw) {
	case FREAD:
		if (tp->t_outq.c_cc > 0 && (tp->t_state&TTSTOP) == 0)
			return(1);
		if ((p = pti->pt_rsel) && p->p_wchan == (caddr_t)&selwait)
			pti->pt_state |= PF_RCOLL;
		else
			pti->pt_rsel = u.u_procp;
		break;

	case FWRITE:
		if (tp->t_rawq.c_cc < TTYHOG/2 || tp->t_delct == 0)
			return(1);
		if ((p = pti->pt_wsel) && p->p_wchan == (caddr_t)&selwait)
			pti->pt_state |= PF_WCOLL;
		else
			pti->pt_wsel = u.u_procp;
		break;
	}
	return(0);
}

/*
 * Write to the input side of the slave device.
 * To avoid lockup, we must not sleep until we are guaranteed
 * the slave reader will be able to complete his read.
 */
ptcwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp = &pt_tty[minor(dev)];
	register struct pt_ioctl *pti = &pt_ioctl[minor(dev)];
	register struct clist *tq = &tp->t_rawq;
	register struct ccblock *tbuf;
	register int n;
	register cnt = uio->uio_resid;	/* (nr) */
	int error = 0;

	static	char	tb[CBSIZE];	/* SYS5: temporary place to process
				 	   input for pttys */

	/* make sure that cfreelist.c_size <= CBSIZE */
	if (cfreelist.c_size > CBSIZE)
		panic("ptcwrite: tb buf too small");

	if ((tp->t_state&ISOPEN) == 0)
		return(EIO);

	while (uio->uio_resid) {
		/*
		 * If tp->t_delct is 0, then the slave reader's
		 * read would not complete and we would not be
		 * able to put any more input onto the queue.
		 * Thus, since c_cc[VMIN] <= TTYHOG, putting
		 * TTYHOG characters onto the input queue will
		 * satisfy the read if tp->t_lflag&ICANON == 0;
		 * If tp->t_lflag&ICANON != 0, then we may have
		 * to put characters onto the queue until it gets
		 * flushed in order to complete the read (equivalent
		 * to a user typing 256 non-EOL characters and an
		 * EOL to input an EOL).
		 */
		if (tq->c_cc >= TTYHOG && tp->t_delct != 0) {
			if (u.u_fmode&FNDELAY)
				return(0);

			if (tp->t_state & NBIO) 	/* (nr) */
				return((uio->uio_resid == cnt) ? EWOULDBLOCK : 0);

			/*
			 * Ok to sleep, the slave reader's read will
			 * complete.
			 */
			pti->pt_state |= PF_OASLP;
			slumber((caddr_t)&pti->pt_outq, INTERRUPTABLE,
				ttoboost);
			/*
			 * Slave may have closed while we were asleep.
			 */
			if ((tp->t_state&ISOPEN) == 0)
				return(0);
		}
		if ((tbuf = &tp->t_rbuf)->c_ptr == NULL)
			/*
			 * No receive buffer -- punt.
			 */
			return(EIO);
		/*
		 * Figure out how many characters to move into the
		 * receive buffer.
		 */
		if (tq->c_cc < TTYHOG) {
			n = min(uio->uio_resid, TTYHOG-tq->c_cc);
			n = min(n, (int)tbuf->c_count);
		} else
			/*
			 * Add a character at a time until they flush the
			 * input queue.
			 */
			n = min(1, (int)tbuf->c_count);

		if (error = uiomove(tb, n, UIO_WRITE, uio))
			break;
		if (pti->pt_state & PF_REMOTE) {
			bcopy(tb, tbuf->c_ptr, n);
			tp->t_rbuf.c_count = tp->t_rbuf.c_size;
			putcb(CMATCH((struct cblock *)tp->t_rbuf.c_ptr),
			    &tp->t_rawq);
			/* allocate new rx buffer */
			if ((tp->t_rbuf.c_ptr = getcf()->c_data)
				== ((struct cblock *)NULL)->c_data) {
				tp->t_rbuf.c_ptr = NULL;
				return(0);
			}
			tp->t_rbuf.c_count = cfreelist.c_size;
			tp->t_rbuf.c_size = cfreelist.c_size;
			(void) putc(0, &tp->t_rawq);
			wakeup((caddr_t)&tp->t_rawq);
			return(0);
		}
		{
			register int j;
			register char	*p = tb;
			int	setparms;
			int	err;
	
			for (j=0; j< n; j++) {
				/*
				 * Queue up input on the slave.  Give the
				 * master writer read and write access to
				 * the slave device so that the call to
				 * ttiocom will not fail.
				 */
				err=ttiocom(tp, TIOCSTI, p++, (FREAD|FWRITE), &setparms);
				if (err)
					return(err);
			}
		}
	}

	return(error);
}

ptyioctl(dev, cmd, data, flag)
	caddr_t data;
	dev_t	dev;
{
	register struct tty *tp = &pt_tty[minor(dev)];
	register struct pt_ioctl *pti = &pt_ioctl[minor(dev)];
	int error;
	int setparms;

	if (cdevsw[major(dev)].d_open == ptcopen)
		switch (cmd) {

		case TIOCPKT:	/* turn on packet flow control */
			if (*(int *)data)
				pti->pt_state |= PF_PKT;
			else
				pti->pt_state &= ~PF_PKT;
			return(0);

		case TIOCREMOTE:	/* turn on remote editing */
			if (*(int *)data)
				pti->pt_state |= PF_REMOTE;
			else
				pti->pt_state &= ~PF_REMOTE;
			return(0);

		case FIONBIO:	/* (nr) */
			if (*(int *)data)
				pti->pt_state |= PF_NBIO;
			else
				pti->pt_state &= ~PF_NBIO;
			return(0);
		}
	error = ttiocom(tp, cmd, data, flag, &setparms);
	if (pti->pt_state & PF_NOSTOP) {
		if (tp->t_iflag & IXON) {
			pti->pt_send &= ~TIOCPKT_NOSTOP;
			pti->pt_send |= TIOCPKT_DOSTOP;
			pti->pt_state &= ~PF_NOSTOP;
			ptcwakeup(tp, pti, 1);
		}
	} else {
		if (!(tp->t_iflag & IXON)) {
			pti->pt_send &= ~TIOCPKT_DOSTOP;
			pti->pt_send |= TIOCPKT_NOSTOP;
			pti->pt_state |= PF_NOSTOP;
			ptcwakeup(tp, pti, 1);
		}
	}
	return(error);
}
