/*
 *	VIOC-X Modem control
 */

#include "../sys/param.h"
#include "../sys/file.h"
#include "../sys/termio.h"
#include "../sys/ioctl.h"
#include "../sys/tty.h"
#include "../sys/conf.h"
#include "../sys/sysinfo.h"
#include "../sys/vioc.h"
#include "../sys/time.h"
#include "../sys/kernel.h"
#include "../sys/buf.h"
#include "../sys/vbavar.h"

/*	BSC header
*/
#include "../sys/bsc/bscio.h"
#include "../sys/bsc/bsc.h"
extern char bscport[];
extern int nvbsc;
/*
	===========
*/

extern	caddr_t		VIOCBAS[];
extern	struct	vcx	vcx[] ;
extern	struct	tty	vx_tty[];
extern	struct	vcmds	v_cmds[] ;

extern	int	vxproc() ;
extern	struct	vxcmd	*vobtain() ;
extern	struct	vxcmd	*nextcmd() ;

extern  struct  vba_device *vxinfo[];

vcmodem(dev,flag)
dev_t dev ;
{
	struct tty *tp ;
	register struct vxcmd *cp ;
	register struct vcx *xp ;
	register struct vblok *kp ;
	register port ;
	struct   vba_ctlr *um;
	unsigned char *addr;

	port = minor(dev) ;
	tp = &vx_tty[port] ;
	xp = (struct vcx *)&vcx[port>>4];
	port &= 017 ;
	cp = vobtain(xp) ;
	kp = VBAS(xp->v_nbr) ;

        /*
         * Issue MODEM command
         *      key off of CLOCAL flag to see what mode the vioc should
         *      be in
         */
        if (tp->t_cflag & CLOCAL) {
		/* turn on receivers on board (solves VIOC firmware bug) */
		um = vxinfo[xp->v_nbr]->ui_mi;
		addr = (unsigned char *) (um->um_addr + 0x082a + port);
		*addr = 0xc1;

                /* turn on or off manual mode */
                cp->cmd = MDMCTL ;
                cp->par[0] = (flag == VMOD_ON) ? V_MENAB : V_MDISAB ;
                cp->par[1] = port;
                vcmd(xp->v_nbr, &cp->cmd) ;
        }
        else {
                /* turn on or off auto mode */
                cp->cmd = MDMCTL ;
                cp->par[0] = (flag == VMOD_ON) ? V_ENAB : V_DISAB ;
                cp->par[1] = port;
                vcmd(xp->v_nbr, &cp->cmd) ;
        }
	port -= xp->v_loport ;

#ifdef VX_DEBUG
	if (flag == VMOD_ON)
		printf("\nvcmodem:ON, port%d,dcd(%lx)",
			port,kp->v_dcd&0xffff);
	else
		printf("\nvcmodem:OFF, port%d,dcd(%lx)",
			port,kp->v_dcd&0xffff);
#endif

	if (flag == VMOD_OFF)
	        tp->t_state &= ~DTRON;
	else tp->t_state |= DTRON;

	if((kp->v_dcd >> port) & 1) {
		if(flag == VMOD_ON)
			tp->t_state |= CARR_ON;
		return(1) ;
	}
	return(0) ;
}

/*
 * VCMINTR called when an unsolicited interrup occurs signaling
 * some change of modem control state.
 */
vcmintr(n)
register n ;	/* viocx number */
{
	register struct vblok *kp ;
	register struct tty *tp ;
	register port ;

	sysinfo.mdmint++;
	kp = VBAS( n ) ;
	port = kp->v_usdata[0] & 017 ;
	tp = &vx_tty[port+n*16] ;

	/*
	 * 	Check for change in DSR for BISYNC port.
	 */
	if ((kp->v_ustat & DSR_CHG) && (bscport[port+n*16] & BISYNC)) {
		register struct bsc *bp ;
		extern	 struct	bsc bsc[] ;

		bp = &bsc[(minor(tp->t_dev) >> 4) & 0xf] ;
		bp->b_hlflgs &= ~BSC_DSR ;
		if (kp->v_ustat & DSR_ON)
			bp->b_hlflgs |= BSC_DSR ;
	}
	if (bscport[port+n*16] & BISYNC) return;
	/*	==========================================
	*/

	if (kp->v_ustat&DCD_ON) {

#ifdef VX_DEBUG
		printf("\nvcmintr: CARR_ON,VIOC%d,port%d,state(%lx),cflag(%lx)",
				n,port,tp->t_state,tp->t_cflag&0xffff);
#endif

		if ((tp->t_state&CARR_ON) == 0 &&
		    tp->t_state&(WOPEN|ISOPEN) &&
		    (tp->t_cflag&CLOCAL) == 0) {
			/*
			 * Carrier transition from low to high and
			 * we care about carrier.
			 */
			tp->t_state |= CARR_ON;
			if (tp->t_state&WOPEN)
				wakeup((caddr_t)&tp->t_canq) ;
		}
		return;
	}

	if (kp->v_ustat&DCD_OFF) {
		if (tp->t_state&CARR_ON &&
		    tp->t_state&(WOPEN|ISOPEN) &&
		    (tp->t_cflag&CLOCAL) == 0) {
			/*
			 * Carrier transition from high low and
			 * we care about carrier.
			 */
			ttyflush(tp, FREAD|FWRITE);
			vxflush(tp);
			if(tp->t_state&ISOPEN) {
				gsignal(tp->t_pgrp, SIGHUP);
				gsignal(tp->t_pgrp, SIGCONT);
			}
			tp->t_state &= ~CARR_ON;
		}
		return;
	}

	if (kp->v_ustat&BRK_CHR && tp->t_state&ISOPEN) {
		/*
		 * #@!$% SCC chip!!!
		 * The goddamned thing gives the VIOC an interrupt
		 * PLUS a perfectly good "0" character.  Thus we
		 * have to turn the next good "\0" character into
		 * a break.
		 */
		vx_break[n] |= 1<<port;
		return ;
	}
}

/*
 * Stop further output on the line.
 * Must be called from spltty or higher.
 */
vxflush(tp)
register struct tty *tp;
{
	register int n = minor(tp->t_dev)>>4;
	register int port = minor(tp->t_dev)&0xf;
	register struct vcx *xp;
	register struct vcmds *cp = &v_cmds[n];
	register struct vxcmd *cmdp;
	register int i;

	xp = &vcx[n];
	if ((tp->t_state&BUSY) == 0)
		/*
		 * Output already stopped.
		 */
		return;

	/*
	 * Either output is in progress for this port
	 * or an output command is queued up for it.
	 * Check the queue and see if it has gotten to
	 * the vioc yet.
	 */
	for (i = cp->v_empty; i!=cp->v_fill;) {
		cmdp = (struct vxcmd *)((long *)cp->cmdbuf[i]-1);
		if ((cmdp->cmd==XMITDTA || cmdp->cmd==XMITIMM) &&
		   ((struct vxmit *)cmdp->par)->line == port) {
			/*
			 * Found the command on the input queue --
			 * turn it into something harmless.
			 */
			cmdp->cmd = FDTATOX ;
			cmdp->par[1] = port ;
			tp->t_state &= ~BUSY;
			return;
		}
		if(++i >= VC_CMDBUFL)
			i = 0;
	}

	/*
	 * Command has already gotten to the VIOC --
	 * tell the VIOC to get rid of it.
	 */
	cmdp = vobtain(xp);
	cmdp->cmd = FDTATOX ;
	cmdp->par[1] = port ;
	vcmd(n, &cmdp->cmd);
}
