/*
**	VDDC Driver - Versabus to SMD direct interface version.
**		Written for HCX vmunix, 9/1/83.		
**		Modified for Disk Geometries, 03/14/86.
*/

#include "../sys/param.h"
#include "../sys/buf.h"
#include "../sys/cmap.h"
#include "../sys/conf.h"
#include "../sys/dk.h"
#include "../sys/map.h"
#include "../ml/mtpr.h"
#include "../ml/pte.h"
#include "../sys/systm.h"
#include "../sys/vbavar.h"
#include "../sys/user.h"
#include "../sys/vmmac.h"
#include "../sys/proc.h"
#include "../sys/uio.h"
#define DSKGENDATA
#include "../sys/dsk.h"                 /* common disk structures */
#undef DSKGENDATA
#include "../sys/ioctl.h"               /* common ioctl structures */
#include "../sys/dskio.h"               /* disk ioctl structures */
#include "../sys/elog.h"		/* for system activity */
#include "../sys/iobuf.h"		/* for system activity */
#include "../sys/kernel.h"
#define	VDGENDATA
#include "../sys/vddc.h"
#undef	VDGENDATA

/*
** Layout of minor number assignments for the VDDC devices.
**
**	  7	       3 2   0
**	 +--------------+-----+
**	 |  Unit number | FLS |
**	 +--------------+-----+
**		   |    |_____ File system # ( 0-7 )
**		   |__________ Unit # in the system
*/

/*
 * External variables
 */

extern	int			ndsk;
extern 	struct	vba_ctlr	*vdminfo[];
extern 	struct  vba_device	*vddinfo[];
extern 	vdunit_tab		vdunit_info[];
extern 	vdctlr_tab		vdctlr_info[];
extern	struct	iotime		vdstat[];	/* for system activity */
extern	struct	iobuf		vdtab[];	/* for system activity */
extern				blkacty;	/* for error logging */

int	vdprobe();		/* See if VDDC is really there */
int	vdslave();		/* See if drive is really there */
int	vdattach();		/* Enter a drive in internal tables */
int	vddgo();		/* Old VAX function, not used. */

struct	vba_driver vddriver = {
	vdprobe, vdslave, vdattach, vddgo, vddcaddr,
	"", vddinfo, "", vdminfo, 0
};

/* for error logging */
struct vddcreg {
	u_int status;
	u_int track;
	u_int sector;
	u_int cylinder;
} vddcreg;


/*
**	See if the controller is really there. 
**  if TRUE - initialize the controller.
*/

vdprobe( um )
	struct vba_ctlr * um ;
{
	cdr	*ctlr_addr;

	ctlr_addr = (cdr *) um -> um_addr ;
	if(!badaddr(ctlr_addr, 2, 0)) {
		ctlr_addr->cdr_reset = 0xffffffff;
		DELAY(1000000);
		if(ctlr_addr->cdr_reset != 0xffffffff) {
			vddriver.ud_mname = "VSMD Controller #";
			DELAY(1000000);
		}
		else {
			vddriver.ud_mname = "SMD/E Controller #";
			ctlr_addr->cdr_reserved = 0x0;
			DELAY(3000000);
		}
		return (TRUE);
	}
	return (FALSE);				/* no controller */	
}


/*
** See if a drive is really there
** Try to Reset/Configure the drive, then test its status.
*/

vdslave(vba_dev_info, ctlr_addr)
register struct vba_device	*vba_dev_info;
register cdr			*ctlr_addr;
{
	register int		ctlr = vba_dev_info->ui_ctlr;
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];
	register int		unit = vba_dev_info->ui_unit;
	register vdunit_tab	*u_info = &vdunit_info[unit];
	register int		slave = vba_dev_info->ui_slave;
	register int		type;
	register fmt_mdcb	*mdcb = &c_info->ctlr_mdcb;
	register fmt_dcb	*dcb = &c_info->ctlr_dcb;

	if(c_info->ctlr_type == UNKNOWN) {
		ctlr_addr->cdr_reset = 0xffffffff;
		DELAY(1000000);
		if(ctlr_addr->cdr_reset != 0xffffffff) {
			c_info->ctlr_type = SMDCTLR;
			c_info->overlap_seeks = FALSE;
			DELAY(1000000);
		}
		else {
			c_info->ctlr_type = SMD_ECTLR;
			c_info->overlap_seeks = TRUE;
			ctlr_addr->cdr_reserved = 0x0;
			DELAY(3000000);
			ctlr_addr->cdr_csr = 0;
			ctlr_addr->mdcb_tcf = AM_ENPDA;
			ctlr_addr->dcb_tcf = AM_ENPDA;
			ctlr_addr->trail_tcf = AM_ENPDA;
			ctlr_addr->data_tcf = AM_ENPDA;
			ctlr_addr->cdr_ccf = CCF_STS | XMD_32BIT | BSZ_16WRD |
			    CCF_ENP | CCF_EPE | CCF_EDE | CCF_ECE | CCF_ERR;
		}
		if(vdaccess_with_no_trailer(ctlr_addr,ctlr,slave,INIT,10) & 
		    HRDERR) {
			printf("\nVDDC: Controller #%d init error.\n", ctlr);
			return (FALSE);
		}
		if(vdaccess_with_no_trailer(ctlr_addr,ctlr,slave,DIAG,10) & 
		    HRDERR) {
			printf("\nVDDC: Controller #%d diagnostic error.\n",ctlr);
			return (FALSE);
		}
	}
	/*
	 * Seek on all drive types starting from the largest one.
	 * a successful seek to the last sector/cylinder/track verifies 
	 * the drive type connected to this port. 
	 */
	for (type = 0; type < nvddrv; type++) {	
		if((c_info->ctlr_type == SMDCTLR) && (vdst[type].nsec != 32))
			continue;
		if(vdconfigure_drive(ctlr_addr, ctlr, slave, type, 0)) {
			dcb->opcode = (short)RD;
			dcb->intflg = NOINT;
			dcb->nxtdcb = (fmt_dcb *)0;	/* end of chain */
			dcb->operrsta  = 0;
			dcb->devselect = (char)slave;
			dcb->trailcnt = (char)(sizeof(trrw) / sizeof(long));
			dcb->trail.rwtrail.memadr=(char *)PHYS(c_info->rawbuf); 
			dcb->trail.rwtrail.wcount =
			    vdst[type].secsize/sizeof(short);
			dcb->trail.rwtrail.disk.cylinder = vdst[type].ncyl - 2;
			dcb->trail.rwtrail.disk.track = vdst[type].ntrak - 1;
			dcb->trail.rwtrail.disk.sector=vdst[type].nsec - 1;
			mdcb->firstdcb = (fmt_dcb *)(PHYS(dcb));
			mdcb->vddcstat = 0;
			VDDC_ATTENTION(ctlr_addr, (fmt_mdcb *)(PHYS(mdcb)),
			    c_info->ctlr_type);
			POLLTILLDONE(ctlr_addr, dcb, 60, c_info->ctlr_type);
			if(vdtimeout <= 0)
				printf(" during probe operation.\n");
			if(!(dcb->operrsta & HRDERR))
				break;
		}
		else
			return (FALSE);
	}
	if (type >= nvddrv) {
		/*
		 * If reached here, a drive which is not defined in the
		 * 'vdst' tables is connected. Cannot set it's type.
		 */
		printf("\nVDDC: Could not identify controller #%d, drive #%d!\n",
		    ctlr, slave);
		return (FALSE);
	}

	/*
	** The following statements initialize the u_info structure with
	** device specific information.  The call to vdinstall_geometry sets
	** up the partition information from the geometry block on the drive.
	** For vddc, only partition information comes from the geometry
	** block on disk.  Other drive characteristics are defined in static
	** tables.  If the geometry block can not be read from disk, then
	** the drive will not be configured and false is returned to the caller.
	*/
	u_info->drive_type = type;
	u_info -> info = vdst [type] ; /* establish initial defaults */
	u_info->sec_per_blk = DEV_BSIZE / u_info->info.secsize;

	if (! vdinstall_geometry( vba_dev_info, ctlr_addr )) {
		printf( "\nVDDC: Invalid geometry block on " ) ;
		printf( "controller #%d, drive #%d!\n", ctlr, slave ) ;
		return( FALSE ) ;
	} ;

	vba_dev_info->ui_type = type;
	vba_dev_info->ui_dk = TRUE;
	vddriver.ud_dname = u_info->info.type_name;
	return( TRUE ) ;
}



/*
**
*/

vdconfigure_drive(ctlr_addr, ctlr, slave, type, pass)
cdr		*ctlr_addr;
register int	ctlr, slave, type, pass;
{
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];

	c_info->ctlr_dcb.opcode = RSTCFG;		/* command */
	c_info->ctlr_dcb.intflg = NOINT;
	c_info->ctlr_dcb.nxtdcb = (fmt_dcb *)0;	/* end of chain */
	c_info->ctlr_dcb.operrsta  = 0;
	c_info->ctlr_dcb.devselect = (char)slave;
	c_info->ctlr_dcb.trail.rstrail.ncyl = vdst[type].ncyl;
	c_info->ctlr_dcb.trail.rstrail.nsurfaces = vdst[type].ntrak;
	if(c_info->ctlr_type == SMD_ECTLR) {
		c_info->ctlr_dcb.trailcnt = (char)4;
		c_info->ctlr_dcb.trail.rstrail.nsectors = vdst[type].nsec;
		c_info->ctlr_dcb.trail.rstrail.slip_sec = vdst[type].nslip;
	}
	else
		c_info->ctlr_dcb.trailcnt = (char)2;
	c_info->ctlr_mdcb.firstdcb = (fmt_dcb *)(PHYS(&c_info->ctlr_dcb));
	c_info->ctlr_mdcb.vddcstat = 0;
	VDDC_ATTENTION(ctlr_addr, (fmt_mdcb *)(PHYS(&c_info->ctlr_mdcb)),
	    c_info->ctlr_type);
	POLLTILLDONE(ctlr_addr, &c_info->ctlr_dcb, 5, c_info->ctlr_type);
	if(vdtimeout <= 0) {
		printf(" during drive configuration.\n");
		return (FALSE);
	}
	if(c_info->ctlr_dcb.operrsta & HRDERR) {
		if(!(c_info->ctlr_dcb.operrsta & (NOTCYLERR | DRVNRDY)))
			printf("\nVDDC: Configuration error.\n");
		else if(pass == 0) {
			vdstart_drive(ctlr_addr, ctlr);
			return (vdconfigure_drive(ctlr_addr,ctlr,slave,type,1));
		}
		else if(pass == 2)
			return (vdconfigure_drive(ctlr_addr,ctlr,slave,type,3));
		return (FALSE);
	}
	return (TRUE);
}


/*
**
*/

vdstart_drive(ctlr_addr, ctlr)
cdr		*ctlr_addr;
register int	ctlr;
{
	vdctlr_info[ctlr].ctlr_started = TRUE;
	return (TRUE);
/*
	If there was some way to do a VDSTOP it would be worth doing a
	VDSTART here.  We choose to assume that if a drive is found on
	a controller that it is also spinning.  That is, if a drive is
	not spinning, it should be shut off!

	printf("\nVDDC: Attempting to start drive #%d.  Please wait...", slave);
	if(vdctlr_info[ctlr].ctlr_started) {
		DELAY(5500000);
		printf("\n");
		return (TRUE);
	}
	vdctlr_info[ctlr].ctlr_started = TRUE;
	if(vdaccess_with_no_trailer(ctlr_addr, ctlr, 0, VDSTART, (slave*6)+62)
	    & HRDERR) {
		printf("\n");
		return (FALSE);
	}
	DELAY((slave * 5500000) + 62000000);
	printf("\n");
	return (TRUE);
*/
}


/*
**
*/

vdaccess_with_no_trailer(ctlr_addr, ctlr, unit, function, time)
register cdr		*ctlr_addr;
register int		ctlr, unit, function, time;
{
	fmt_mdcb	*mdcb = &vdctlr_info[ctlr].ctlr_mdcb;
	fmt_dcb		*dcb = &vdctlr_info[ctlr].ctlr_dcb;

	dcb->opcode = function;		/* command */
	dcb->intflg = NOINT;
	dcb->nxtdcb = (fmt_dcb *)0;	/* end of chain */
	dcb->operrsta  = 0;
	dcb->devselect = (char)unit;
	dcb->trailcnt = (char)0;
	mdcb->firstdcb = (fmt_dcb *)(PHYS(dcb));
	mdcb->vddcstat = 0;
	VDDC_ATTENTION(ctlr_addr, (fmt_mdcb *)(PHYS(mdcb)),
	    vdctlr_info[ctlr].ctlr_type);
	POLLTILLDONE(ctlr_addr, dcb, time, vdctlr_info[ctlr].ctlr_type);
	if(vdtimeout <= 0) {
		printf(" during initialization operation.\n");
		return (DCBCMP | ANYERR | HRDERR | OPABRT);
	}
	return (dcb->operrsta);
}


/*
**
*/

vdattach(vba_dev_info)
struct vba_device *vba_dev_info;
{
	register int		unit = vba_dev_info->ui_unit;
	register vdunit_tab	*u_info = &vdunit_info[unit];
	register vdctlr_tab	*c_info = &vdctlr_info[vba_dev_info->ui_ctlr];
	register struct buf	*c_queue = &vba_dev_info->ui_mi->um_tab;
	register struct buf	*u_queue = c_queue->b_forw;
	register struct buf	*start_queue = u_queue;
	register int		type = vba_dev_info->ui_type;
	register fs_tab		*fs = &u_info->info;


	/*
	** Setup the default drive info parameters and then check for
	** a geometry block.  For the vddc, only partition information is
	** taken from the geometry block.  Note, it is assumed that at this
	** point the geometry block has already been read from disk in the
	** slave routine.
	*/
	u_info->info = vdst[type] ; /* establish defaults */
	vdinstall_geometry( vba_dev_info, ((cdr *) vba_dev_info -> ui_addr) ) ;

	u_info->sec_per_blk = DEV_BSIZE / u_info->info.secsize;
	u_info->sec_per_cyl = u_info->info.nsec * u_info->info.ntrak;
	u_info->xfer_queue.b_dev = vba_dev_info->ui_slave;
	c_info->unit_type[vba_dev_info->ui_slave] = type;
	/* load unit into controller's active unit list */
	if(u_queue == NULL) {
		c_queue->b_forw = &u_info->xfer_queue;
		u_info->xfer_queue.b_forw = &u_info->xfer_queue;
		u_info->xfer_queue.b_back = &u_info->xfer_queue;
	}
	else {
		while(u_queue->b_forw != start_queue)
			u_queue = u_queue->b_forw;
		u_info->xfer_queue.b_forw = start_queue;
		u_info->xfer_queue.b_back = u_queue;
		u_queue->b_forw = &u_info->xfer_queue;
		start_queue->b_back = &u_info->xfer_queue;
	}



	/*
	 * bytes per second:
	 *      ((number of sectors per track) * (bytes per sector)) *(rpm/60)
 	 */
	dk_bps[unit] = fs->nsec * fs->secsize * (fs->rpm / 60) ;
}

/*
** Install Geometry:
**
** This procedure is responsible for reading the disk geometry block from the
** drive defined by the vba_device structure.  If the geometry block is
** unavailable for any reason, FALSE is returned from the routine indicating
** that the drive is not configurable.  If the geometry block is read
** successfully, then the partition information is copied into the u_info
** structure and TRUE is returned.
*/

int
vdinstall_geometry( vba_device_ptr, ctlr_ptr )

register struct vba_device * 	vba_device_ptr ;	/* device control blk */
register cdr *			ctlr_ptr ;		/* controller address */

{

	vdctlr_tab * 		c_info_ptr ;          	/* controller block   */
	register geometry_sector * geometry_sector_ptr ;/* disk geometry      */
	register geometry_block * geometry_ptr ;	/* geometry block ptr */
	register  		i ;			/* loop variable      */
	register vdunit_tab *	u_info_ptr ;		/* drive block	      */

	/*
	** Initialize some local variables.
	*/
	c_info_ptr          = & (vdctlr_info [vba_device_ptr -> ui_ctlr]) ;
	u_info_ptr          = & (vdunit_info [vba_device_ptr -> ui_unit]) ;
	geometry_sector_ptr = (geometry_sector *) (c_info_ptr -> rawbuf) ;
	geometry_ptr        = & (geometry_sector_ptr -> geometry_block) ;

	/*
	** Attempt to read in the geometry sector.  If it is available, then
	** the partition information is copied into the unit info structure.
	*/
	for (i = 0 ; i < GB_MAXPART ; i ++) {
		u_info_ptr -> info . partition [i] . start  = 0 ;
		u_info_ptr -> info . partition [i] . length = 0 ;
	} ;

	if (vdread_geometry( vba_device_ptr, ctlr_ptr, geometry_sector_ptr)) {
		for (i = 0 ; i < GB_MAXPART ; i ++) {
			u_info_ptr -> info . partition [i] =
				geometry_ptr -> partition [i] ;
		} ;
		return( TRUE ) ;
	} ;

	return( FALSE ) ;

}



/*
** Read Geometry Block:
**
** This procedure is responsible for reading and validation of the disk
** geometry block.  This block is read from sector 0 of the target disk.
** The geometry block checksum is validated and header id.
*/

int
vdread_geometry( vba_device_ptr, ctlr_ptr, geometry_sector_ptr )

register struct vba_device * 	vba_device_ptr ;	/* controller info    */
register cdr *			ctlr_ptr ;		/* controller address */
register geometry_sector *	geometry_sector_ptr ;	/* disk geometry      */

{

	register vdctlr_tab *	c_info_ptr ;          	/* controller block   */
	register                checksum ;		/* geometry checksum  */
	register fmt_dcb *	dcb_ptr ;		/* drive control blk  */
	register geometry_block * geometry_ptr ;	/* geometry block ptr */
	register fmt_mdcb *	mdcb_ptr ;		/* master control blk */
	register trrw *         rwt_ptr ;		/* read/write trailer */

	/*
	** Initialize some local variables.
	*/
	c_info_ptr   = & (vdctlr_info [vba_device_ptr -> ui_ctlr]) ;
	dcb_ptr      = & (c_info_ptr -> ctlr_dcb) ;
	geometry_ptr = & (geometry_sector_ptr -> geometry_block) ;
	mdcb_ptr     = & (c_info_ptr -> ctlr_mdcb) ;

	/*
	** Initialize the dcb for reading the geometry block.
	*/
	dcb_ptr -> opcode    = (short) RD ;
	dcb_ptr -> intflg    = NOINT ;
	dcb_ptr -> nxtdcb    = (fmt_dcb *) 0 ; 	/* end of chain */
	dcb_ptr -> operrsta  = 0 ;
	dcb_ptr -> devselect = (char) (vba_device_ptr -> ui_slave) ;
	dcb_ptr -> trailcnt  = (char) (sizeof(trrw) / sizeof(long)) ;

	rwt_ptr                  = & (dcb_ptr -> trail . rwtrail) ;
	rwt_ptr -> memadr        = (char *) PHYS( geometry_sector_ptr ) ;
	rwt_ptr -> wcount        = sizeof(geometry_sector) / sizeof(short) ;
	rwt_ptr -> disk.cylinder = 0 ;
	rwt_ptr -> disk.track    = 0 ;
	rwt_ptr -> disk.sector   = 0 ;

	/*
	** Initialize the master dcb.
	*/
	mdcb_ptr -> firstdcb = (fmt_dcb *) (PHYS( dcb_ptr )) ;
	mdcb_ptr -> vddcstat = 0 ;

	/*
	** Initiate the I/O operation and wait for completion.
	*/
	VDDC_ATTENTION( ctlr_ptr,				/* controller */
			(fmt_mdcb *) (PHYS( mdcb_ptr )),	/* mdcb       */
			c_info_ptr -> ctlr_type ) ;		/* ctlr type  */

	POLLTILLDONE( ctlr_ptr,				/* controller address */
		      dcb_ptr,				/* dcb address	      */
		      60,				/* timeout period     */
		      c_info_ptr -> ctlr_type ) ;	/* controller type    */

	/*
	** Evaluate error conditions.  If successful, purge the data page from
	** the cache.
	*/
	if (vdtimeout <= 0) {
		printf( " during geometry read.\n" ) ;
		return( FALSE ) ;
	} ;

	if ((dcb_ptr -> operrsta) & HRDERR) {

		printf( "\nVDDC: Hard error during geometry read.  " ) ;
		printf( "Status = %x  ", dcb_ptr -> operrsta ) ;

		if (c_info_ptr -> ctlr_type == SMD_ECTLR) {
			uncache( & (dcb_ptr -> err_code) ) ;
			printf( "Error code = %x", dcb_ptr -> err_code ) ;
		} ;

		printf( "\n" ) ;
		return( FALSE ) ;

	} ;

	mtpr ((int) geometry_sector_ptr, P1DC);
	if (((int) geometry_sector_ptr & PGOFSET) >
	    (NBPG - sizeof( geometry_sector )))
	        mtpr ((int) geometry_sector_ptr + NBPG, P1DC) ;

	/*
	** Validate the block checksum and version and return.
	*/
	GB_CHECKSUM( geometry_sector_ptr, checksum ) ;
	if ((checksum != geometry_sector_ptr -> checksum) ||
	    (bcmp(geometry_ptr -> id, GB_ID, GB_ID_LEN)) ||
	    (geometry_ptr -> version != GB_VERSION))
		return( FALSE ) ;

	return( TRUE ) ;

}



/*
** Historic VAX routine.
*/

/*ARGSUSED*/
vddgo(um)
struct vba_ctlr *um;
{
}


/*
**
*/

vdstrategy(request)
register struct buf *request;
{
	register int			unit = VDUNIT(request->b_dev);
	register struct vba_device	*vba_dev_info = vddinfo[unit];
	register par_tab		*par;
	register vdunit_tab		*u_info = &vdunit_info[unit];
	register vdctlr_tab		*c_info;
	register int			blk;
	int				priority, slave;

	par = &u_info->info.partition[FILSYS(request->b_dev)];
	if((request->b_bcount != 0) && (vba_dev_info != NULL) &&
	    (vba_dev_info->ui_alive != 0)&&(par->length != 0)) {
		slave = vba_dev_info->ui_slave;
		c_info = &vdctlr_info[vba_dev_info->ui_ctlr];
		blk = (request->b_bcount + DEV_BSIZE - 1) >> DEV_BSHIFT;
		if(par->length < (request->b_blkno + blk)) {
			blk = par->length - request->b_blkno;
			if(blk > 0)
				request->b_bcount = blk * DEV_BSIZE;
			else
				goto error;
		}

		/* begin system activity */
		request->b_start = time.tv_sec;
		vdstat[unit].io_cnt++;
		vdstat[unit].io_bcnt += blk * 2;	/* in terms of 512B bl*/
		/* end system activity */

		blk = request->b_blkno + par->start;
		blk *= u_info->sec_per_blk;
		request->b_daddr = (blk/u_info->info.nsec)%u_info->info.ntrak;
		request->b_cylin = blk/u_info->sec_per_cyl;
		buf_setup (request, u_info->info.secsize);
		priority = spl7();
		if(u_info->xfer_queue.av_forw == NULL)
			if((request->b_cylin != c_info->cur_cyl[slave]) ||
			    (request->b_daddr != c_info->cur_trk[slave]))
				c_info->off_cylinder |= 1 << slave;
		request->b_daddr |= (blk % u_info->info.nsec) << 8;
		disksort(&u_info->xfer_queue, request);
		if(!vba_dev_info->ui_mi->um_tab.b_active++) {
			splx(priority);
			vdstart(vba_dev_info->ui_mi);
		}
		else
			splx(priority);
		return;
	}
 	harderr(request, u_info->info.type_name);
 	printf("- Strategy was passed bad information: " );
 	printf("b_bcount=%d, dev_info=%x, alive=%x, and partition length=%d.\n",
 	    request->b_bcount,vba_dev_info,vba_dev_info->ui_alive,par->length);
error:	request->b_flags |= B_ERROR;
	request->b_error = ENXIO;
	request->b_resid = request->b_bcount;
	iodone(request);
}


/*
**  vdioctl:
**
**  Vdioctl is called by UNIX every time an ioctl call is made by a user
**  program.  The ioctl commands that are processed here are common to all
**  disk on the system.  For the vddc, most of these commands are nops or they
**  return an error condition.  Currently, only the disk status and partition
**  information commands return any information for the vddc.
*/

vdioctl(dev, command, data_ptr, flag)

	register int		command ;	/* ioctl command word	      */
	register char *         data_ptr ;	/* structure for parameters   */
	register dev_t		dev ;		/* action device	      */
	register int		flag ;		/* flag word                  */

{

	register int			ctlr ;		/* controller number  */
	register fmt_dcb *		dcb_ptr ;	/* drive dcb pointer  */
	register dsk_status *		dsk_status_ptr ;/* disk status struct */
	register fs_tab *		fs_ptr ;	/* file system struct */
	register geometry_status *      gb_status_ptr ; /* geometry status    */
	register int			i ;		/* local loop var     */


	/*
	** Switch on the command and process.
	*/

	switch (command) {

	/*
	** The following commands have no meaning for VDDC and should never
	** be called.  If these commands are executed, an error condition is
	** returned.  I have coded them here for completeness.
	*/
	case DSKIOCFORMAT :
	case DSKIOCFORMATCTL :
	case DSKIOCVENDORFLAW :
		return( EACCES ) ;

	/*
	** Return the status information for the specified disk device.
	** This consist of configuration information for the drive.
	*/
	case DSKIOCSTATUS :
		fs_ptr         = & (vdunit_info [VDUNIT(dev)] . info) ;
		ctlr           = vddinfo [VDUNIT(dev)] -> ui_ctlr ;
		dcb_ptr        = & (vdctlr_info [ctlr] . ctlr_dcb) ;
		uncache( & (dcb_ptr -> operrsta) ) ;
		dsk_status_ptr = (dsk_status *) data_ptr ;

	        dsk_status_ptr -> id = vdunit_info [VDUNIT(dev)] . drive_type ;
	        dsk_status_ptr -> drive_status	      =	dcb_ptr -> operrsta ;
	        dsk_status_ptr -> cylinders   	      =	fs_ptr -> ncyl ;
	        dsk_status_ptr -> heads     	      =	fs_ptr -> ntrak ;
	        dsk_status_ptr -> sectors    	      =	fs_ptr -> nsec ;
	        dsk_status_ptr -> rpm                 = fs_ptr -> rpm ;
	        dsk_status_ptr -> phys_cylinders      =	-1 ;
	        dsk_status_ptr -> phys_heads    	      =	-1 ;
	        dsk_status_ptr -> phys_sectors        = -1 ;
	        dsk_status_ptr -> def_cyl             = -1 ;
	        dsk_status_ptr -> def_cylinders       = -1 ;
	        dsk_status_ptr -> diag_cyl     	      =	-1 ;
	        dsk_status_ptr -> diag_cylinders      = -1 ;
	        break ;

	/*
	** Return geometry information for the disk device.  This describes
	** the partition configuration defined within the geometry block.
	*/
	case DSKIOCGEOMETRY :
		gb_status_ptr = (geometry_status *) data_ptr ;
		fs_ptr        = & (vdunit_info [VDUNIT(dev)] . info) ;

	        for (i = 0 ; i < GB_MAXPART ; i ++) {
	        	gb_status_ptr -> partition [i] =
	  		      	fs_ptr -> partition [i];
	        } ;
	        break ;

	/*
	** Invalid command.
	*/
	default :
		return( ENXIO ) ;

	}

	return( NOERROR ) ;

}


/*
** Start up a transfer on a drive.
*/
vdstart(controller_info)
register struct vba_ctlr *controller_info;
{
	register struct buf	*c_queue = &controller_info->um_tab;
	register struct buf	*u_queue = c_queue->b_forw;

	/* Search for next ready unit */
	c_queue->b_forw = c_queue->b_forw->b_forw;
	u_queue = c_queue->b_forw;
	do {
		if(u_queue->av_forw != NULL) {
			c_queue->b_forw = u_queue;
			vdexecute(controller_info, u_queue);
			return;
		}
		u_queue = u_queue->b_forw;
	} while(u_queue != c_queue->b_forw);
}


/*
**
*/

vdload_seeks(c_info, u_queue)
register vdctlr_tab	*c_info;
register struct buf	*u_queue;
{
	register fmt_dcb	*dcb;
	register struct buf	*request;
	register struct buf	*start_queue = u_queue;
	register int		unit, slave, seek_count = 0;

	do {
		if((request = u_queue->av_forw) != NULL) {
			unit = VDUNIT(request->b_dev);
			slave = vddinfo[unit]->ui_slave;
			if(c_info->off_cylinder & (1 << slave)) {
				c_info->off_cylinder &= ~(1 << slave);
				if(c_info->cur_cyl[slave] != request->b_cylin) {
					c_info->cur_cyl[slave]=request->b_cylin;
					dk_seek[unit]++;
				}
				c_info->cur_trk[slave] = request->b_daddr&0xff;
				dcb = &c_info->seek_dcb[seek_count++];
				dcb->opcode = SEEK;
				dcb->intflg = NOINT | INT_PBA;
				dcb->operrsta  = 0;
				dcb->devselect = (char)slave;
				dcb->trailcnt = (char)1;
				dcb->trail.sktrail.skaddr.cylinder =
				    request->b_cylin;
				dcb->trail.sktrail.skaddr.track =
				    request->b_daddr & 0xff;
				dcb->trail.sktrail.skaddr.sector = 0;
			}
		}
		u_queue = u_queue->b_forw;
	} while((u_queue != start_queue) && (seek_count < 4));
	return (seek_count);
}



vdexecute(controller_info, u_queue)
struct vba_ctlr *controller_info;
struct buf	 *u_queue;
{
	extern			vd_int_timeout();
	register struct	buf	*request = u_queue->av_forw;
	int			ctlr = controller_info->um_ctlr;
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];
	register int		unit = VDUNIT(request->b_dev);
	register int		slave = vddinfo[unit]->ui_slave;
	register fmt_mdcb	*mdcb = &c_info->ctlr_mdcb; 
	register fmt_dcb	*dcb = &c_info->ctlr_dcb; 
	struct iobuf		*iobp;	/* for system activity */
	register int		i;
	
	if((c_info->off_cylinder & ~(1<<slave)) && c_info->overlap_seeks) {
		int		seek_count = vdload_seeks(c_info, u_queue);

		mdcb->firstdcb = (fmt_dcb *)PHYS(&c_info->seek_dcb[0]);
		for(i = 1; i < seek_count; i++)
			c_info->seek_dcb[i-1].nxtdcb = 
			    (fmt_dcb *)PHYS(&c_info->seek_dcb[i]);
		c_info->seek_dcb[seek_count-1].nxtdcb = (fmt_dcb *)PHYS(dcb);
	}
	else {
		if(request->b_cylin != c_info->cur_cyl[slave]) {
			c_info->cur_cyl[slave] = request->b_cylin;
			dk_seek[unit]++;
		}
		c_info->cur_trk[slave] = request->b_daddr & 0xff;
		c_info->off_cylinder = 0;
		mdcb->firstdcb = (fmt_dcb *)(PHYS(dcb));
	}
	/* begin system activity */
	iobp = &vdtab[unit];
	if (iobp->b_active == 0)
		iobp->io_start = time.tv_sec;
	iobp->b_active++;
	/* end system activity */

	blkacty |= (1 << major(request->b_dev));	/* for error logging */

	dcb->opcode = (request->b_flags & B_READ) ? RD : WD;
	dcb->intflg = INTDONE;
	dcb->nxtdcb = (fmt_dcb *)0;	/* end of chain */
	dcb->operrsta  = 0;
	dcb->devselect = (char)slave;
	dcb->trailcnt = (char)(sizeof(trrw) / sizeof(long));
	dcb->trail.rwtrail.memadr = (char *)get_ioadr(request, c_info->rawbuf,
	    c_info->map, c_info->utl);
	dcb->trail.rwtrail.wcount=(short)((request->b_bcount+1)/sizeof(short));
	dcb->trail.rwtrail.disk.cylinder = request->b_cylin;
	dcb->trail.rwtrail.disk.track = request->b_daddr & 0xff;
	dcb->trail.rwtrail.disk.sector = request->b_daddr >> 8;
	mdcb->vddcstat = 0;
   	dk_wds[unit] += request->b_bcount / 32;
	/*
	 * The following data structures are updated
	 * for sadp(1).
	 * 'lastcyl[unit]' is the cylinder number of the
	 * last I/O operation of drive 'unit'.
	 */
	i = dcb->trail.rwtrail.disk.cylinder;
	dkprf[unit][i >> 3]++;
	i -= lastcyl[unit];
	if (i < 0)
		i = -i;
	skprf[unit][i >> 3]++;
	lastcyl[unit] = dcb->trail.rwtrail.disk.cylinder;
	c_info->int_expected = TRUE;

	timeout(vd_int_timeout, ctlr, 20*60);
  	dk_busy |= 1 << unit;
	VDDC_ATTENTION((cdr *)(vdminfo[ctlr]->um_addr),(fmt_mdcb *)(PHYS(mdcb)),
	    c_info->ctlr_type);
}


/*
**
*/

vd_int_timeout(ctlr)
register int	ctlr;
{
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];
	register fmt_dcb	*dcb = &c_info->ctlr_dcb;

	uncache(&dcb->operrsta);
	printf("\nVDDC: Controller #%d didn't interrupt!", ctlr);
	printf("  Status = %x", dcb->operrsta);
	if(c_info->ctlr_type == SMD_ECTLR) {
		uncache(&dcb->err_code);
		printf("  Error code = %x", dcb->err_code);
	}
	printf("\n");
	if(!(dcb->operrsta & DCBCMP)) { 
		VDDC_ABORT((cdr *)(vdminfo[ctlr]->um_addr), c_info->ctlr_type);
		dcb->operrsta |= DCBUSC | DCBABT | ANYERR | HRDERR | CTLRERR;
	}
	vdintr(ctlr);
}


/*
 * Handle a disk interrupt.
 */
vdintr(ctlr)
register int	ctlr;
{
	extern			vd_int_timeout();
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];

	untimeout(vd_int_timeout, ctlr);
	if(c_info->int_expected) {
		register struct buf	*u_queue=vdminfo[ctlr]->um_tab.b_forw;
		register struct buf	*request = u_queue->av_forw;
		register int		slave = u_queue->b_dev;
		register int		unit = VDUNIT(request->b_dev);
		fmt_mdcb 		*mdcb = &c_info->ctlr_mdcb;
		register fmt_dcb 	*dcb = &c_info->ctlr_dcb;
		int			error;

   		dk_busy &= ~(1 << unit);
   		dk_xfer[unit]++;
		c_info->int_expected = FALSE;

		blkacty &= ~(1 << major(request->b_dev)); /* error logging */

		uncache(&mdcb->intdcb);
		uncache(&dcb->operrsta);
		if(c_info->ctlr_type == SMD_ECTLR)
			uncache(&dcb->err_code);

		if (error = vddecode_error(request, dcb)) {

			/* begin error logging */
			vddcreg.status = (u_int)dcb->operrsta;
			vddcreg.track = (u_int)dcb->trail.rwtrail.disk.track;
			vddcreg.sector = (u_int)dcb->trail.rwtrail.disk.sector;
			vddcreg.cylinder = (u_int)dcb->trail.rwtrail.disk.cylinder;
			vdtab[unit].b_dev = request->b_dev;
			vdtab[unit].b_actf = request;
			vdtab[unit].io_addr = vddinfo[unit]->ui_addr;
			fmtberr(&vdtab[unit], vddcreg.cylinder, (u_int *)&vddcreg);
			/* end error logging */

			switch (error) {
			case CTLR_ERROR :
				vdhard_error(c_info, request, dcb,
				    (vdminfo[ctlr]->um_tab.b_errcnt < 2),
				    "Disk controller");
				vdreset_ctlr(vdminfo[ctlr]->um_addr,ctlr);
				vdstat[unit].ios.io_misc++; /* error logging */
				if(vdminfo[ctlr]->um_tab.b_errcnt++ < 2) {
					vdminfo[ctlr]->um_tab.b_forw = u_queue->b_back;
					vdstart(vdminfo[ctlr]);
					return;
				}
				request->b_resid = request->b_bcount;
				break;
			case DRIVE_ERROR :
				vdhard_error(c_info, request, dcb,
				    (vdminfo[ctlr]->um_tab.b_errcnt < 2), "Disk drive");
				reset_drive(vdminfo[ctlr]->um_addr, ctlr, slave, 2);
				vdstat[unit].ios.io_misc++; /* error logging */
				if(vdminfo[ctlr]->um_tab.b_errcnt++ < 2) {
					vdminfo[ctlr]->um_tab.b_forw = u_queue->b_back;
					vdstart(vdminfo[ctlr]);
					return;
				}
				request->b_resid = request->b_bcount;
				break;
			case HARD_DATA_ERROR :
				vdhard_error(c_info, request, dcb,
				    (vdminfo[ctlr]->um_tab.b_errcnt < 3),
				    "Uncorrected data");
				reset_drive(vdminfo[ctlr]->um_addr, ctlr, slave, 2);
				vdstat[unit].ios.io_misc++; /* error logging */
				if(vdminfo[ctlr]->um_tab.b_errcnt++ < 3) {
					vdminfo[ctlr]->um_tab.b_forw = u_queue->b_back;
					vdstart(vdminfo[ctlr]);
					return;
				}
				request->b_resid = 0;
				break;
			case SOFT_DATA_ERROR :
				vdsoft_error(c_info, request, dcb);
			default :
				if(vdminfo[ctlr]->um_tab.b_errcnt)
					printf("  VDDC: Retried successfully.\n");
				request->b_error = 0;
				request->b_resid = 0;
				break;
			}
		} else {		/* no error */
			request->b_error = 0;
			request->b_resid = 0;
		}
		end_transfer(request, c_info->rawbuf, c_info->map, c_info->utl);
		u_queue->av_forw = request->av_forw;
		if(u_queue->av_back == request)
			u_queue->av_back = NULL;
		else {
			struct buf		*next;

			unit = VDUNIT(u_queue->av_forw->b_dev);
			slave = vddinfo[unit]->ui_slave;
			next = u_queue->av_forw;
			if((next->b_cylin != c_info->cur_cyl[slave]) ||
			    ((next->b_daddr & 0xff) != c_info->cur_trk[slave]))
				c_info->off_cylinder |= 1 << slave;
		}
		vdminfo[ctlr]->um_tab.b_errcnt = 0;
		vdminfo[ctlr]->um_tab.b_active--;
		/* begin error logging  and system activity */
		if (vdtab[unit].b_active) {
			if (vdtab[unit].io_erec)
				logberr(&vdtab[unit], request->b_flags&B_ERROR);
			vdtab[unit].b_active = 0;
			vdtab[unit].b_errcnt = 0;
			vdtab[unit].b_actf = request->av_forw;
			vdstat[unit].io_resp += time.tv_sec - request->b_start;
			vdstat[unit].io_act += time.tv_sec - vdtab[unit].io_start;
		}
		/* end error logging and system activity */

		if(request->b_flags & B_ERROR)
			request->b_error = EIO;
		iodone(request);
		vdstart(vdminfo[ctlr]);
		return;
	}
	printf("\nVDDC: Unexpected interrupt from controller #%d!\n", ctlr);
}



/*
**
*/

vdhard_error(c_info, request, dcb, retry, type)
vdctlr_tab	*c_info; 
struct buf	*request;
fmt_dcb		*dcb;
int		retry;
char		*type;
{
	vdreport(c_info, request, dcb, type);
	if(retry)
		printf("  VDDC: Retrying failed operation...\n");
	else {
		request->b_flags |= B_ERROR;
		printf("  VDDC: Retries failed to correct error!\n");
	}
}


/*
**
*/

vdsoft_error(c_info, request, dcb)
vdctlr_tab	*c_info; 
struct buf	*request;
fmt_dcb		*dcb;
{
	vdreport(c_info, request, dcb, "Corrected data");
}


/*
*/

vdreport(c_info, request, dcb, type)
vdctlr_tab		*c_info; 
register struct buf	*request;
register fmt_dcb	*dcb;
char			*type;
{
	register int		delta;
	register int		unit = VDUNIT(request->b_dev);
	register vdunit_tab	*u_info = &vdunit_info[unit];
	register int		blk = request->b_blkno * u_info->sec_per_blk;
	register int		par = FILSYS(request->b_dev);

	harderr(request, u_info->info.type_name);
	printf("- %s error.  Status = 0x%x", type, dcb->operrsta);
	if(c_info->ctlr_type == SMD_ECTLR)
		printf(", Error code = 0x%x", dcb->err_code);
	printf("\n");
	delta = (request->b_bcount - 1) / u_info->info.secsize;
	blk = request->b_blkno + u_info->info.partition[par].start;
	blk *= u_info->sec_per_blk;
	if(delta == 0)
		printf("  VDDC: Error occured on absolute sector %d.\n", blk);
	else {
		printf("  VDDC: Error occured between");
		printf(" absolute sectors %d and %d.\n", blk, (blk + delta));
	}
}


/*
**
*/

/*ARGSUSED*/
vdopen(dev, flag)
register dev_t dev;
int flag;
{
	register unit = VDUNIT(dev);
	register struct vba_device *vba_dev_info = vddinfo[unit];

	if((vba_dev_info == 0) ||
	    (vba_dev_info->ui_alive == 0) ||
	    (vba_dev_info->ui_type >= nvddrv))
		return (ENXIO);
	if(vdunit_info[unit].info.partition[FILSYS(dev)].length == 0)
		return (ENXIO);

	/* for error logging */
	vdtab[unit].io_nreg = (sizeof(struct vddcreg) / sizeof(u_int));
	vdtab[unit].io_stp = &vdstat[unit].ios;

	return (NOERROR);
}


/*
**
*/

vdread(dev, uio)
dev_t dev;
struct uio *uio;
{
	extern			vdstrategy();
	register int		unit = VDUNIT(dev);
	register vdunit_tab	*u_info = &vdunit_info[unit];

	if (unit >= ndsk)
		return (ENXIO);
	return (physio(vdstrategy, &u_info->raw_q_element, dev, B_READ,
	    minphys, uio));
}


/*
**
*/

vdwrite(dev, uio)
dev_t dev;
struct uio *uio;
{
	extern			vdstrategy();
	register int		unit = VDUNIT(dev);
	register vdunit_tab	*u_info = &vdunit_info[unit];

	if (unit >= ndsk)
		return (ENXIO);
	return (physio(vdstrategy, &u_info->raw_q_element, dev, B_WRITE,
	    minphys, uio));
}


/*
**
*/

vddump(dev)
dev_t	dev;
{
	register int		 unit = VDUNIT(dev);
	register vdunit_tab	 *u_info = &vdunit_info[unit];
	register fs_tab		 *fs = &u_info->info;
	register int		 ctlr = vddinfo[unit]->ui_ctlr;
	register struct vba_ctlr *vba_vdctlr_info = vdminfo[ctlr];
	register int		 filsys = FILSYS(dev);
	register cdr		 *ctlr_addr = (cdr *)(vba_vdctlr_info->um_addr);
	register int		 cur_blk, blkcount, blocks;
	int			 dump_short = FALSE;
	int			 memaddr = 0x0;

	vdreset_ctlr(ctlr_addr, ctlr);
	blkcount = dumpsize;		/* In 1k byte pages */
	if ((dumplo + blkcount) > fs->partition[filsys].length) {
		blkcount = fs->partition[filsys].length - dumplo;
		dumpsize = blkcount ; /* let savecore know about shortage */
		printf("\nVDDC: Only dumping first %dmb of memory!\n",
		    blkcount / 1024);
		dump_short = TRUE;
	}
	cur_blk = fs->partition[filsys].start + dumplo;
	while (blkcount > 0) {
		blocks = MIN(blkcount, DUMPSIZE);
		if(!vdwrite_block(ctlr_addr,ctlr,unit,memaddr,cur_blk,blocks))
			return (EIO);
		blkcount -= blocks;
		memaddr += blocks * NBPG;
		cur_blk += blocks;
	}
	return (dump_short ? ENXIO : NOERROR);
}


/*
**
*/

vdwrite_block(ctlr_addr, ctlr, unit, addr, block, blocks)
register cdr	 *ctlr_addr;
register int	ctlr, unit;
register char	*addr;
register int	block, blocks;
{
	register fmt_mdcb	*mdcb = &vdctlr_info[ctlr].ctlr_mdcb;
	register fmt_dcb	*dcb = &vdctlr_info[ctlr].ctlr_dcb;
	register vdunit_tab	*u_info = &vdunit_info[unit];
	register fs_tab		*fs = &u_info->info;

	block *= (int)u_info->sec_per_blk;
	blocks *= (int)u_info->sec_per_blk;
	mdcb->firstdcb = (fmt_dcb *)(PHYS(dcb));
	dcb->intflg = NOINT;
	dcb->opcode = WD;
	dcb->operrsta = 0;
	dcb->devselect = (char)(vddinfo[unit])->ui_slave;
	dcb->trailcnt = (char)(sizeof(trrw) / sizeof(long));
	dcb->trail.rwtrail.memadr = addr;
	dcb->trail.rwtrail.wcount = (short)((blocks*fs->secsize)/sizeof(short));
	dcb->trail.rwtrail.disk.cylinder= (short)(block / u_info->sec_per_cyl);
	dcb->trail.rwtrail.disk.track = (char)((block / fs->nsec) % fs->ntrak);
	dcb->trail.rwtrail.disk.sector = (char)(block % fs->nsec);
	VDDC_ATTENTION(ctlr_addr, (fmt_mdcb *)(PHYS(mdcb)),
	    vdctlr_info[ctlr].ctlr_type);
	POLLTILLDONE(ctlr_addr, dcb, 5, vdctlr_info[ctlr].ctlr_type);
	if(vdtimeout <= 0) {
		printf(" during dump operation!\n");
		return (FALSE);
	}
	if (dcb->operrsta & HRDERR) {
		printf("\nVDDC: Dump error.  Status = %x\n", dcb->operrsta);
		return (FALSE);
	}
	return (TRUE);
}


/*
**
*/

vdsize(dev)
register dev_t dev;
{
	register unit = VDUNIT(dev);
	register struct vba_device *vba_dev_info = vddinfo[unit];

	if((vba_dev_info == 0) || 
	    (vba_dev_info->ui_alive == 0) ||
	    (vba_dev_info->ui_type >= nvddrv))
		return (-1);
	return (vdunit_info[unit].info.partition[FILSYS(dev)].length);
}


/*
**
*/

/*ARGSUSED*/
vdreset_ctlr(ctlr_addr, ctlr)
register cdr	*ctlr_addr;
register int	ctlr;
{
	register struct buf	*c_queue = &vdminfo[ctlr]->um_tab;
	register struct buf	*u_queue = c_queue->b_forw;
	register vdctlr_tab	*c_info = &vdctlr_info[ctlr];
	
	printf("\nVDDC: Resetting controller #%d.  Please wait...\n", ctlr);
	VDDC_RESET(ctlr_addr, c_info -> ctlr_type);
	c_info->ctlr_started = FALSE;
	if(c_info->ctlr_type == SMD_ECTLR) {
		ctlr_addr->cdr_csr = 0;
		ctlr_addr->mdcb_tcf = AM_ENPDA;
		ctlr_addr->dcb_tcf = AM_ENPDA;
		ctlr_addr->trail_tcf = AM_ENPDA;
		ctlr_addr->data_tcf = AM_ENPDA;
		ctlr_addr->cdr_ccf = CCF_STS | XMD_32BIT | BSZ_16WRD |
		    CCF_ENP | CCF_EPE | CCF_EDE | CCF_ECE | CCF_ERR;
	}
	if(vdaccess_with_no_trailer(ctlr_addr,ctlr,0,INIT,10) & HRDERR) {
		printf("\nVDDC: Controller #%d failed to initialize!\n", ctlr);
		return (FALSE);
	}
	if(vdaccess_with_no_trailer(ctlr_addr,ctlr,0,DIAG,10) & HRDERR) {
		printf("\nVDDC: Controller #%d diagnostic error!\n",ctlr);
		return (FALSE);
	}
	/*  reset all units attached to controller */
	u_queue = c_queue->b_forw;
	do {
		reset_drive(ctlr_addr, ctlr, u_queue->b_dev, 0);
		u_queue = u_queue->b_forw;
	} while(u_queue != c_queue->b_forw);
	printf("VDDC: Controller #%d was reset successfully.\n", ctlr);
	return (TRUE);
}


/*
**
*/

reset_drive(ctlr_addr, ctlr, slave, start)
register cdr	*ctlr_addr;
register int	ctlr, slave, start;
{
	register int	type = vdctlr_info[ctlr].unit_type[slave];

	if(type != UNKNOWN)
		if(!vdconfigure_drive(ctlr_addr, ctlr, slave, type, start))
			printf("\nVDDC: Couldn't reset unit #%d, controller #%d!\n",
			    slave, ctlr);
}


/*
	Print_dcb() dumps the mdcb and DCB for diagnostic purposes.  This
 routine is called whenever a fatal error is encountered.
*/

vdprintdcb(ptr)
register long	*ptr;
{
	register long	i;
	register long	trailer_count;

	if(ptr != (long *)NULL) {
		printf(" Dump of DCB:");
		trailer_count = *(ptr+3) & 0xff;
		for(i=0; i<7+trailer_count; i++)
			printf("  %lx", *(ptr+i));
	}
	printf("\n");
	for(i=0; i<5000000; i++) ;
}
