/*
 * This is the driver for the IKON 10088 hardcopy interface for the VMEbus 
 * It supports Versatek plotters, Centronix compatible devices, and
 * Tektronix color copiers.  Although only Versatek plotters are useable
 * with our hardware configuration.
 */

/* If you have a real slumber call (ATT 5.2) remove this stuff */
/* #define INTERRUPTABLE PZERO+1
   #define slumber(a,b,c) sleep(a,b)
   #define iocomboost 0 */

#include "../sys/types.h"
#include "../sys/param.h"
#include "../sys/systm.h"
#include "../ml/mtpr.h"
#include "../sys/vm.h"
#include "../sys/buf.h"
#include "../ml/pte.h"
#include "../sys/user.h"
#include "../sys/file.h"
#include "../sys/uio.h"
#include "../sys/ioctl.h"
#include "../sys/mtio.h"
#include "../sys/errno.h"
#include "../sys/cmap.h"
#include "../sys/kernel.h"
#include "../sys/vbavar.h"
#include "../sys/proc.h"
#include "../sys/vp.h" 
#include "../sys/vcmd.h"

/* Error control and trace messages */

#undef VP_TRCM 1  /* If defined, driver will output      */
                  /* trace messages on the console.      */

#undef VP_ERRM 1  /* If defined, driver will output      */
                  /* error messages on the console.      */

#undef VP_ANCM 1  /* If defined, initialize routine      */
                  /* announces presence of interface on  */
                  /* console at boot time.               */

#undef VP_DUMP 1 /* If defined, the board's registers    */
                 /* are dumped to the console at various */
                 /* points in the driver.                */

/* System Structures */

extern struct vba_device *vp_vba_device[]; /* vba device info */ 
extern struct vba_ctlr *vp_vba_ctlr[];     /* vba controller info */
extern vp_aux_type vp_aux[];               /* device info */
extern unsigned long vp_intvect_proc[];    /* device info */
extern int nvp;                            /* number of vp devices */

extern vpattach() ;
extern vpprobe() ; 
extern vpslave() ; 

struct vba_driver vpdriver = 
{
	vpprobe, 
	vpslave, 
	vpattach, 
	0, /* go routine */
	0, /* possible address */
	"vp",
	vp_vba_device, 
	"Versatec Plotter", 
	vp_vba_ctlr,
	0
} ;
/*
 * Check for presence of the board.
 * If it is present initialize it with a master clear,
 * set the interrupt vector, clear the command and status
 * elements of vp_aux, and return TRUE.
 */
vpprobe( vba_ctlr )
struct vba_ctlr *vba_ctlr;
{
	register vp_registers *addr;      /* pointer to board */
	register ushort status,status2;   /* used during reads of reigisters */
	register vp_aux_type *vp_aux_ptr; /* vp controller info    */
	register int ctlr;                /* the controller number    */

	ctlr = vba_ctlr->um_ctlr;
	vp_aux_ptr = &vp_aux[ctlr];

	addr = (vp_registers *) vba_ctlr->um_addr;

	if ( badaddr(&addr->vp_lfis, sizeof(&addr->vp_lfis), 0 ) ) {
		vp_aux_ptr->vp_flags = 0;   /* not present */
		return( FALSE );
	}
	else {
  		/* setup kernel buf */
		vp_aux_ptr->vp_phys = (u_long) vtoph( 0, vp_aux_ptr->vp_space);

		/* set present flag and clear all others */

		vp_aux_ptr->vp_flags = VP_PRES;
		vp_aux_ptr->vp_cmd = 0;   /* clear command */
		vp_aux_ptr->vp_istat = 0; /* clear device interrupt status */
		vp_aux_ptr->vp_dstat = 0; /* clear device interrupt status */
		vp_aux_ptr->vp_time = 0;  /* clear time out flag */
		vp_aux_ptr->vp_addr = addr; /* save address */

		vba_ctlr->um_ivct = get_ivct(0,1);	
		if ((vba_ctlr->um_ivct < 0) ||
		 (init_ivct(vba_ctlr->um_ivct, vp_intvect_proc[ctlr]) < 0 )) {
			printf("\nInterrupt vector for vp%d not set up.\n",ctlr);
			vp_aux_ptr->vp_flags = 0;   /* not present */
			return( FALSE );
		}
		addr->vp_intvect = vba_ctlr->um_ivct;

		addr->vp_pfds = VP_MCLR; /* pulse master clear in cmd reg */

/* Read initial device and interface status. Use dummy labels to make
 * sure that the optimizer doesn't delete any device register reads.
 */

#ifdef VP_ANCM
		status = addr->vp_lfis; /* get initial interface status */
		status2 = addr->vp_pfds; /* get initial device status */

		printf("\nvp board present, int stat =0x%x, dev stat = 0x%x\n",
		       status, status2);
#endif
		return TRUE;
	}
}

vpslave( drive_info, ctlr_addr )
struct vba_device *drive_info ;
caddr_t ctlr_addr ;
{
		return TRUE;
}

/* void */
vpattach( device_info )
register struct vba_device *device_info ;
{
}

/*
 * vptime() is the local watchdog timer.
 *
 * This is not started unless vp_time > 0 which is initially
 * set with VP_TOCK.  This is basically a diagnostic option
 * since I can't think of a device where this is useful.
 * Don't use it with a versatec, since long rewinds in color
 * plots may take up to a minute.
 *
 * It is called by the system clock routine approximately every
 * 10 seconds.  It checks to see if the hardcopy driver is waiting
 * for dma end of range, or a command done interrput.  If it is,
 * the routine checks to see if the driver was waiting for the 
 * same event (by comparing sequence numbers) the last time that
 * vptime() was called.  If it was, vp_toc is incremented, and
 * if vp_toc execcds the defined VP_TOCK value, a time-out has 
 * occurred.  If it was a time-out waiting for a non-data command
 * to complete, a wakeup(&vp_aux) is done to wake up the waiting
 * vpioctl() routine that issued the command.  If the driver was
 * waiting for and end-of-range interrupt, a master clear is done
 * to terminate the hung dma operation, and an iodone(bp) is issued
 * to wakeup physio().  In either case interrupts are inhibited.
 * Whether there was a timeout or not, before exiting, vptime()
 * calls timeout to arrange for another call to itself in 10 seconds.
 * This continues as long as the hardcopy driver is open.
 */
static vptime(vp_aux_ptr)
vp_aux_type * vp_aux_ptr;   /* Will be a pointer to the         */
                            /* structure on which the sleeping  */
                            /* routine is waiting.              */
{
	register vp_registers *addr;   /* pointer to board         */
	int intlevel;                  /* will save interrupt level*/
	ushort temp;                   /* write to dev through this*/

#ifdef VP_TRCM
	printf("\nvptime() called\n");
#endif

/* get pointer to the board */

	addr = vp_aux_ptr->vp_addr;

/* check for unchanged sequence number */

	if (vp_aux_ptr->vp_cseq != vp_aux_ptr->vp_lseq) {
		vp_aux_ptr->vp_tock = 0; /* clear counter */
	}
	else {

/* determine which sleep - if any - timed out */

		if (vp_aux_ptr->vp_flags & VP_ACTV) { 
			/* waiting for end of range */
			vp_aux_ptr->vp_tock++; /* increment tock counter   */

			if (vp_aux_ptr->vp_tock >= vp_aux_ptr->vp_time) {
				/* dma timeout */

#ifdef VP_ERRM
	printf("\ntimeout waiting for dma end-of-range\n");
#endif

/* Issue master clear to terminate dma and disable interrupts
 * This will also reset the various mode control bits and
 * may cause unexpected results if the calling program fails
 * to check for a timeout condition before re-starting dma.
 */

		           addr->vp_pfds = VP_MCLR;/* pulse master clear */

/* set various flags to their appropriate values and wake up physio() */

			   /* clear active flag */
		           vp_aux_ptr->vp_flags &= ~VP_ACTV;
		           /* set dma time-out flag */
		           vp_aux_ptr->vp_flags |= VP_TMDM;
			   /* flag error */
		           vp_aux_ptr->vp_bstat |= VP_BERROR;
			   /* wake up write */
	    	           wakeup(vp_aux_ptr->vp_space);
			}
		}

		if (vp_aux_ptr->vp_flags & VP_CMDW){ /*waiting for command end*/
			vp_aux_ptr->vp_tock++;            /*increment tock counter */

			if (vp_aux_ptr->vp_tock >= vp_aux_ptr->vp_time) {
			/* time-out */

#ifdef VP_ERRM
	printf("\ntimeout wainting for non-data command end\n");
#endif

/* Issue master clear to disable interrputs.
 **** In some application it may not be appropriate to issue master clear.
 * It may be necessary to implement a more graceful recovery from a dma
 * or command time-out.
 */

				addr->vp_pfds = VP_MCLR;/* issue master clear */

/* update flags and wake up vpioctl() which is waiting for cmd completion */

				/* clear waiting flag */
				vp_aux_ptr->vp_flags &= ~VP_CMDW;
				/* set timeout flag   */
				vp_aux_ptr->vp_flags |= VP_TCMD;
				/* wake up vpioctl()  */
				wakeup((caddr_t)vp_aux_ptr);

			}
		}
	}

/* call timeout to arrange for another call to vptime in 10 seconds */

	if ((vp_aux_ptr->vp_flags & VP_OPEN) && (vp_aux_ptr->vp_time > 0)) {
                /* only call while open */
		timeout(vptime,(caddr_t)vp_aux_ptr,VP_TICK);
	}
	else
		vp_aux_ptr->vp_time = 0;
	
/* update sequence variable */

	vp_aux_ptr->vp_lseq = vp_aux_ptr->vp_cseq;
	return;

}

/* vpopen() opens the hardcopy interface and device for use by
 * the calling program.  It checks to make sure that the unit
 * hasn't been opened already - if it has an error indication
 * is returned.  If it hasn't it is flagged as open.  The open
 * routine also clears the flags in vp_aux and writes a 0 to
 * the interface latched function register to disable interrputs.
 * The last thing that vpopen() does is call timeout() to start 
 * the local watchdog timer.
 */

int
vpopen( dev, flag )
dev_t dev ;
int flag ;
{
	register int unit = VPUNIT(dev);                    /* which vp device */
	register vp_aux_type *vp_aux_ptr = &vp_aux[unit];   /* dev info */
	register vp_registers *addr = vp_aux_ptr->vp_addr;  /* board    */
	register ushort temp;            /* used for writing to board    */
	

#ifdef VP_TRCM
	printf("\nvpopen() called\n");
#endif
	
	if (!vp_vba_device[unit] || (unit >= nvp)) {
 	/* make sure it is present */

#ifdef VP_ERRM
	printf("\nopen attempted on non-existent hardcopy interface\n");
#endif
		return ENXIO; /* flag as error */
	}

	if (vp_aux_ptr->vp_flags & VP_OPEN) { /* check for already open */

#ifdef VP_ERRM
	printf("\nopen attempted when hardcopy interface already open\n");
#endif
		return EACCES;
	}

/* proceed to open the hardcopy interface */

	switch (VPTYPE(dev)) {

		/* type versatec */

		case VP_TVP:
		   addr->vp_lfis = VP_INIT;   /* set latched functions */
			                           /* to initialized value  */
			                           /* also inhibits interrupts */
		   vp_aux_ptr->vp_lmsk = VP_INIT; /* init latched function */
		                           	/* register mask         */
		   break;

		/* type centronix */

		case VP_TCN:
		   addr->vp_lfis = VP_INIT | VP_SELO;
		   vp_aux_ptr->vp_lmsk = VP_INIT | VP_SELO;
		   break;

		/* type centronix - long reset */

		case VP_TCNL:
	  	   addr->vp_lfis = VP_INIT | VP_SELO | VP_ALRS;
		   vp_aux_ptr->vp_lmsk = VP_INIT | VP_SELO | VP_ALRS;
		   break;

		/* type tektronix */

		case VP_TTK:
		   addr->vp_lfis = VP_INIT | VP_SELO | VP_SDST;
		   vp_aux_ptr->vp_lmsk = VP_INIT | VP_SELO | VP_SDST;
		   break;

		/* type tektronix - long reset */

		case VP_TTKL:
		   addr->vp_lfis = VP_INIT | VP_SELO | VP_SDST | VP_ALRS;
		   vp_aux_ptr->vp_lmsk = VP_INIT | VP_SELO | VP_SDST | VP_ALRS;
		   break;

		/* no type */

		default:	
		   return EIO;
		   break;
	}

	vp_aux_ptr->vp_flags |= VP_OPEN;	/* flag as open */
	addr->vp_pfds = VP_RINT;   /* clear the interrupt flag */

/* start the local software watchdog timer.  This call to timeout
 * will start it.  It will continue to call itself as long as the
 * hardcopy interface is open.
 */
	vp_aux_ptr->vp_cseq = 1;      /* init current seq number */
	vp_aux_ptr->vp_lseq = 0;      /* init last seq number */
	if (vp_aux_ptr->vp_time > 0)        /* timeout already started */
		vp_aux_ptr->vp_time = VP_TOCK;   /* default timeout count  */
	else {
		vp_aux_ptr->vp_time = VP_TOCK;   /* default timeout count  */
		if (vp_aux_ptr->vp_time > 0)     
 			/* start timer */
			timeout(vptime,(caddr_t)vp_aux_ptr,VP_TICK);
	}
	return 0;
}

/* vpclose() makes sure that the device hasn't been closed
 * already, then makes sure that the interrput enable latch and
 * interrput flag are off, clears the open flag in vp_aux[].vp_flags
 * and returns.
 */

/* void */
vpclose( dev, flag )
register dev_t dev ;
register int flag ;
{
	register vp_aux_type *vp_aux_ptr = &vp_aux[VPUNIT(dev)];
	register vp_registers *addr = vp_aux_ptr->vp_addr; /* board */
	register ushort temp;                  /* used for writing to board */
	register int intlevel;                 /* used to save int level */

#ifdef VP_TRCM
	printf("\nvpclose() called\n");
#endif

	if(!(vp_aux_ptr->vp_flags & VP_OPEN)) { /* if already closed */

#ifdef VP_ERRM
	printf("\nattempt to close hardcopy device when already closed\n");
#endif

		return EIO; /* flag as error */
	}
	vp_aux_ptr->vp_flags &= ~VP_OPEN; /* clear open flag */

	return 0;
}

/* vpintr() is the interrput code for the hardcopy interface driver.
 * The controler number is passed.  The interrupt code wakes up 
 * vpwrite() if we were doing a dma transfer, and wakes up vpioctl()
 * if we were waiting for the transfer at the end of a non-data command.
 * One should note that the hardcopy interface resets its onboard
 * interrupt enable mask during the interrupt ack sequence.
 */

/* void */
vpintr( devnum )
int devnum ;
{
	register vp_aux_type * vp_aux_ptr = &vp_aux[VPUNIT(devnum)];
	register vp_registers * addr = vp_aux_ptr->vp_addr;  /* board */
	register ushort temp;         /* used to buffer writes to board */
	register ushort status;       /* used when reading from board */

#ifdef VP_TRCM
	printf("\nvpinitr() called\n");
#endif

	vp_aux_ptr->vp_istat = addr->vp_lfis;  /* get interface status */
	vp_aux_ptr->vp_dstat = addr->vp_pfds;  /* get device status */

/* Clear the interrupt flag */

	addr->vp_pfds = VP_RINT;

#ifdef VP_TRCM
	printf("\nat interrupt: int status = 0x%x, dev status = 0x%x\n",
			 vp_aux_ptr->vp_istat, vp_aux_ptr->vp_dstat);
#endif

/* Make sure that the hardcopy board is really interrupting.  If not,
 * this is a really serious hardware or software error!  Remember
 * that by the time we get to this place in the code, the hardcopy
 * interrupt enable latch has been cleared.
 */

	if (!(vp_aux_ptr->vp_istat & VP_IFLG)) { /* we didn't interrupt */

#ifdef VP_ERRM
	printf("\nhardcopy int code called when in flag not on\n");
#endif
		/* flag error in buffer header */
		vp_aux_ptr->vp_bstat |= VP_BERROR;
		return;
	}

/* Now figure out what kind of interrupt we were expecting.
 * If we fall through both tests, assume that we have some kind
 * of error.
 */

	if(vp_aux_ptr->vp_flags & VP_ACTV) {  /* waiting for e-o-r */

#ifdef VP_TRCM
	printf("\nend of range interrupt received\n");
#endif

		vp_aux_ptr->vp_flags &= ~VP_ACTV; /* clear active flag */
		vp_aux_ptr->vp_bstat &= ~VP_BINUSE; /* free up buffer    */
		wakeup(vp_aux_ptr->vp_space); /* wake up write     */

		return;
	}

	if (vp_aux_ptr->vp_flags & VP_CMDW) { /* waiting for command end */

#ifdef VP_TRCM
	printf("\nend of command interrupt recieved\n");
#endif

		vp_aux_ptr->vp_flags &= ~VP_CMDW; /* clear waiting flag */
		wakeup(vp_aux_ptr); /* wakeup vpioctl()   */
		return;
	}

/* Assume an error since we got an interrupt when we weren't 
 * waiting for anything.  (The interrupt enable latch shouldn't
 * have been set.
 */

	vp_aux_ptr->vp_bstat |= VP_BERROR; /* flag error */
	return;
}

/* vpstrategy() is the routine that sets the device up for data
 * transfer and start the dma.  It is called by vpwrite() which
 * has interpreted the users calling parameters, including the data
 * buffer address and the transfer byte count.  Vpwrite has also
 * copied the user's data buffer to a kernel buffer to make sure 
 * that it is not paged out while that data is being transferred
 * to the device.
 */

static int vpstrategy( dev, length )
dev_t dev ;
long length;
{
	register int unit = VPUNIT(dev);
	register vp_aux_type * vp_aux_ptr = &vp_aux[unit];
	register vp_registers * addr = vp_aux_ptr->vp_addr; /* board */
	register ushort temp;   /* will write to board thru this */
	register ushort status; /* read from board through this */
	int intlevel;           /* will save interrupt level here */

#ifdef VP_TRCM
	printf("\nvpstrategy() called\n");
#endif

/* Make sure that hardcopy board not already active with dma or waiting
 * for command completion when strategy called.  This is impossible 
 * unless major hardware or software failure.
 */

	if (vp_aux_ptr->vp_flags & (VP_ACTV | VP_CMDW)) { /* if error */

#ifdef VP_ERRM
	printf("\nhardcopy driver already busy when vpstrategy() called\n");
#endif

		return EIO;
	}

/* Things seem to be OK.  Start dma set-up sequence and parameter checking */

	addr->vp_addlo = (ushort)vp_aux_ptr->vp_phys; /* write low 16 bits */
	addr->vp_addhi = (char) (vp_aux_ptr->vp_phys >> 16); /* write high 8 bits */

/* Check the requested transfer count to make sure that it will fit in
 * the board's 16 bit counter.
 */
#ifdef VP_TRCM
	printf("\ntransfer count = 0x%x\n",length);
#endif

	if (length > 0x10000) { /* range count too large */

#ifdef VP_ERRM
	printf("\ntransfer count too large\n");
#endif
		return EIO;
	}

	/* Write count to board. A count of 0 = 64K bytes. */

	addr->vp_range = (ushort) length;

/* Write address modifier to board.  This could have been done in
 * vpopen(), since the address modifier register doesn't change in
 * this driver.  A more complex system might allow dumping data from
 * more than one address maps.
 */

	addr->vp_addmod = (char) VP_ADMD;

/* allow for dumping the board's registers as a diagnostic tool. */

#ifdef VP_DUMP

	printf("\ninterface status = 0x%x\n",addr->vp_lfis);
	printf("device status = 0x%x\n",addr->vp_pfds);
	printf("diagnostic data input = 0x%x\n",addr->vp_podd);
	printf("dma address modifier = 0x%x\n",addr->vp_addmod);
	printf("interrupt vector = 0x%x\n",addr->vp_intvect);
	printf("high dma address = 0x%x\n",addr->vp_addhi);
	printf("low dma address = 0x%x\n",addr->vp_addlo);
	printf("dma range count = 0x%x\n",addr->vp_range);
#endif

/* Increment sequence counter for the watchdog timer. */

	vp_aux_ptr->vp_cseq++;

/* Set dma active flag in vp_aux[].vp_flags so interrput routine
 * will know what is going on.
 */

	vp_aux_ptr->vp_flags |= VP_ACTV;

/* Enable interrupts before starting dma, since a very short block
 * might cause end of range before we got the enable latch turned
 * on.  The interrput flag (and request) will not set if end-of-range
 * occurrs while the enable latch is reset--and we would not get
 * an interrput when we did set the enable.
 */

/* First make sure that the int flag is off. */

	addr->vp_pfds = VP_RINT;

	vp_aux_ptr->vp_lmsk |= VP_IENB; /* update mask and set int enb */
	addr->vp_lfis = vp_aux_ptr->vp_lmsk;

	addr->vp_pfds = VP_SDMA; /* start dma */

	return 0;
}

/*
 * NUMBER OF CHARACTERS TO COPY
 *
 * Determines the number of characters in user's buffer to request ( < max ).
 *
 */
static int vpnumber(uio, max_size)
       struct uio *uio;
       int max_size;
{
 register int cc = 0;
 
   while (uio->uio_resid > 0) 
   {
      cc = uio->uio_iov->iov_len;
      if (cc == 0)
      { 
         uio->uio_iovcnt--;
         uio->uio_iov++;
         if (uio->uio_iovcnt < 0)
	 {
            printf ("tknumber uio error");
            cc = 0;
            continue;
         };
      }
      else if (cc > max_size) 
      {
         cc = max_size;
         break;
      }
      else
         break;
   };

#ifdef VP_TRCM
	printf("\nnumber to transfer %d\n",cc);
#endif

   return cc;
}

/* vpwrite() chops up the buffer into large chunks and copies it
 * one chunk at a time into a huge kernel buffer.  It then calls
 * vpstrategy() to transfer the buffer.  Waits are done on the 
 * buffer for dma completion.  Note that a timeout may occur here.
 */

int
vpwrite( dev, uio )
dev_t dev ;
struct uio *uio ;
{
	register vp_aux_type *vp_aux_ptr = &vp_aux[VPUNIT(dev)];
	register long cc;
	register int error;
	register int intlevel;   /* will save processor priority */

#ifdef VP_TRCM
	printf("\nvpwrite() called\n");
#endif

	error = 0;

	cc = vpnumber(uio, (int) sizeof(vp_aux_ptr->vp_space));

	while ((cc > 0) && (!error)) {

		/* Hold everything and wait for the last DMA */

		intlevel = spl5();
		while ((vp_aux_ptr->vp_bstat & VP_BINUSE)
              && (!(vp_aux_ptr->vp_bstat & VP_BERROR))) { 
			if (slumber(vp_aux_ptr->vp_space, 
			    PCATCH|INTERRUPTABLE, iocomboost)) {
				vpreset(vp_aux_ptr);
				return EIO;
			}
		}
		splx(intlevel);

		if (vp_aux_ptr->vp_bstat & VP_BERROR) {
			vpreset(vp_aux_ptr);
			 return EIO; 
		}

		vp_aux_ptr->vp_bstat |= VP_BINUSE;
		uiomove(vp_aux_ptr->vp_space,cc, UIO_WRITE, uio);
		error = vpstrategy(dev, cc);
		cc = vpnumber(uio, (int) sizeof(vp_aux_ptr->vp_space));
	}

	if (vp_aux_ptr->vp_time > 0) { 

		/* Hold everything and wait for the last DMA */

		intlevel = spl5();
		while ((vp_aux_ptr->vp_bstat & VP_BINUSE)
     		      && (!(vp_aux_ptr->vp_bstat & VP_BERROR))) { 
			if (slumber(vp_aux_ptr->vp_space, 
			    PCATCH|INTERRUPTABLE, iocomboost)) {
				vpreset(vp_aux_ptr);
				return EIO;
			}
		}
		splx(intlevel);
		if (vp_aux_ptr->vp_bstat & VP_BERROR) {
			vpreset(vp_aux_ptr);
			 return EIO; 
		}
	}
	if (error) {
		vpreset(vp_aux_ptr);
		return error;
	}
}

/* vpioctl() in the i/o control routine for the hardcopy interface
 * driver.  It is used to issue commands and mode changes to the
 * interface, and to wait for command completion (using interrupts).
 *
 * The cmd argument contains bits which specify the function
 * to be performed.  data points to a user array that contains
 * additional agruments - the number of clock "tocks" to use
 * as a timeout value, and programmed i/o data to be sent to
 * the device - and is used to return status to the calling 
 * program.
 *
 * There are two kinds of commands Versatec only and diagnostic.
 *
 * The Versatec only commands are all of those which are practically
 * used on the Versatec.  They roughly correspond to the normal
 * Unix ioctls for a Versatec in both ucb and att.
 *
 * Versatec
 * --------
 *          VPGETSTATE:  Returns a mask of the following:
 *                          VPRINTPLOT    - simultaineous print/plot mode
 *                          VPLOT         - plot mode
 *                          VPRINT        - print mode
 *                          VOPEN         - open
 *
 *          VPSETSTATE:  Sets the plotter into one of the following modes:
 *                          VPRINTPLOT    - simultaneous print/plot mode
 *                          VPLOT         - plot mode
 *                          VPRINT        - print mode
 *                          VPTERMCOM     - send terminate signal
 *                          VPFFCOM       - send form feed signal
 *                          VPEOTCOM      - send eot signal
 *                          VPCLRCOM      - send clear signal
 *                          VPRESET       - send reset to plotter and board
 *
 * The diagnostic commands let you do just about anything with the 
 * board.  These commands came with the sample IKON driver.
 *
 * Diagnostic
 * ----------
 *
 * When used as calling arguments, the contents of the user's array are:
 *
 *     arg[0] value to write to p-i/o output register
 *     arg[1] number of "tocks" until timeout
 *            (currently one "tock" = 10 seconds)
 * 
 * The status values returned to the user's array are:
 *
 *     arg[0] status flags (contents of vp_aux[].vp_flags)
 *     arg[1] interface status at last interrupt
 *     arg[2] device status at last interrupt
 *     arg[3] current interface status
 *     arg[4] current device status
 *     arg[5] diagnostic input register
 *     arg[6] address modifier + interrupt vector register
 *     arg[7] high dma address register
 *     arg[8] low dma address register
 *     arg[9] dma range count register
 */

int
vpioctl( dev, command, data, flag )
dev_t dev;              /* device number */
int command;            /* command from calling program */
caddr_t data;           /* address of argument array in user's space */
int flag;               /* not currently used */
{
	register vp_aux_type * vp_aux_ptr = &vp_aux[VPUNIT(dev)];
	register vp_registers *addr = vp_aux_ptr->vp_addr;  /* board */
	register ushort temp;   /* write to board thru this */
	register ushort status; /* read through this */
	int intlevel;           /* will save processor priority */
	vp_op *diag;            /* diagnostic arguments */
	int cmd;                /* ioctl command */

/* Save the command in the aux structure */

	vp_aux_ptr->vp_cmd = command;

/* coerce arguments */
	diag = (vp_op *) data;

/* Dump the arguements if trace is on */

#ifdef VP_TRCM
	printf("\nvpioctl() called\n");
	printf("command = 0x%x\n",vp_aux_ptr->vp_cmd);
	printf("arg[0] = 0x%x, arg[1] = 0x%x\n",diag->arg0,diag->arg1);
#endif

/* Versatec compatiblity calls */

	if (vp_aux_ptr->vp_cmd == VGETSTATE) {

#ifdef VP_TRCM
	printf("\nGet versatec state\n");
#endif

		vpwait((int)VPWAT,dev); 

		*(int *)data = 0;
		status = addr->vp_pfds;

		if (status & VP_VSPP) *(int *)data |= VPRINTPLOT;
		if (status & VP_VPLT) *(int *)data |= VPLOT;
		if (!(status & VP_VPLT)) *(int *)data |= VPRINT;
		*(int *)data |= VPOPEN;

		vpwait((int)VPWAT,dev);
		vp_aux_ptr->vp_flags &= ~(VP_TMDM | VP_TCMD | VP_TNRY);
		return 0;
	}

	if (vp_aux_ptr->vp_cmd == VSETSTATE) {

#ifdef VP_TRCM
	printf("\nSet versatec state\n");
#endif

		vpwait((int)VPWAT,dev);

		if (*(int *)data & VPRINTPLOT) {

#ifdef VP_TRCM
	printf("\nSet versatec print-plot mode\n");
#endif

			temp = vp_aux_ptr->vp_lmsk | VP_SSPP;
			vp_aux_ptr->vp_lmsk = temp;
			addr->vp_lfis = temp;
			vpwait((int)VPWAT,dev);
		};	
		if (*(int *)data & VPLOT) {

#ifdef VP_TRCM
	printf("\nSet versatec plot mode\n");
#endif

			temp = vp_aux_ptr->vp_lmsk | VP_SPLT;
			vp_aux_ptr->vp_lmsk = temp;
			addr->vp_lfis = temp;
			vpwait((int)VPWAT,dev);
		};	
		if (*(int *)data & VPRINT) {

#ifdef VP_TRCM
	printf("\nSet versatec print mode\n");
#endif

			temp = vp_aux_ptr->vp_lmsk & ~VP_SPLT;
			vp_aux_ptr->vp_lmsk = temp;
			addr->vp_lfis = temp;
			vpwait((int)VPWAT,dev);
		};	
		if (*(int *)data & VPTERMCOM) {

#ifdef VP_TRCM
	printf("\nSend versatec terminate\n");
#endif

			addr->vp_pfds = VP_VLTR;
			vpwait((int)VPWAT,dev);
		};
		if (*(int *)data & VPFFCOM) {

#ifdef VP_TRCM
	printf("\nSend versatec form feed\n");
#endif

			addr->vp_pfds = VP_VFFD;
			vpwait((int)VPWAT,dev);
		};
		if (*(int *)data & VPEOTCOM) {

#ifdef VP_TRCM
	printf("\nSend versatec EOT\n");
#endif

			addr->vp_pfds = VP_VEOT;
			vpwait((int)VPWAT,dev);
		};
		if (*(int *)data & VPCLRCOM) {

#ifdef VP_TRCM
	printf("\nSend versatec clear\n");
#endif

			addr->vp_pfds = VP_VCLR;
			vpwait((int)VPWAT,dev);
		};
		if (*(int *)data & VPRESET) {

#ifdef VP_TRCM
	printf("\nSend versatec reset\n");
#endif

			addr->vp_pfds = VP_MCLR;
			vpwait((int)VPWAT,dev);
		};
		
		vp_aux_ptr->vp_flags &= ~(VP_TMDM | VP_TCMD | VP_TNRY);
		return 0;
	}

/* IKON diagnostic modes */

	cmd = vp_aux_ptr->vp_cmd & ~(VMMSK & (VPWAT | VWAIT));

#ifdef VP_TRCM
	printf("\ncmd = 0x%x\n",cmd);
#endif

	if (cmd == VSTIM) {
				/* set time-out value */
#ifdef VP_TRCM
	printf("\nsetting timeout, value = 0x%x\n",diag->arg1);
#endif
		vp_aux_ptr->vp_time = diag->arg1;
	}

	if (cmd == VSWAK) { /* software ack pulse */

#ifdef VP_TRCM
	printf("\nissuing a software ack\n");
#endif
		addr->vp_pfds = VP_SACK; /* issue pulse */
	}

/* If we are supposed to wait before issuing command or mode change,
 * call the wait subroutine.  If it returns 0, the wait timed out,
 * and the commands are not issued.
 */

	if (!((vp_aux_ptr->vp_cmd & VMMSK) & VPWAT) || vpwait((int)VPWAT,dev)) {
		/* check command list */	
		if (cmd == VSMOD) { /* write to latched register */

#ifdef VP_TRCM
	printf("\nwriting to latched function register\n");
#endif

			temp = diag->arg0 & VMMSK;
			vp_aux_ptr->vp_lmsk = temp; /* save copy for strategy */
			addr->vp_lfis = temp;
		}
	
		if (cmd == VPIOW) { /* p-i/o output */

#ifdef VP_TRCM
	printf("\nwriting to p-i/o output register\n");
#endif
	
			addr->vp_podd = diag->arg0;
		}
	
		if (cmd == VSMLR) { /* reset device and interface */

#ifdef VP_TRCM
	printf("\nresetting device and interface\n");
#endif

			addr->vp_pfds = VP_MCLR;
		}

		if (cmd == VSCLR) { /* send clear to versatec */

#ifdef VP_TRCM
	printf("\nsending versatec clear\n");
#endif

			addr->vp_pfds = VP_VCLR;
		}

		if (cmd == VSFFD) { /* send versatec form feed */

#ifdef VP_TRCM
	printf("\nsending versatec form feed\n");
#endif

			addr->vp_pfds = VP_VFFD;
		}

		if (cmd == VSEOT) { /* send eot to versatec */

#ifdef VP_TRCM
	printf("\nsending versatec e-o-t\n");
#endif

			addr->vp_pfds = VP_VEOT;
		}
	
		if (cmd == VSLTR) { /* line terminate to v-tec */

#ifdef VP_TRCM
	printf("\nsending versatec line terminate\n");
#endif

			addr->vp_pfds = VP_VLTR;
		}
	} /* end of ready wait if */

/* See if we are supposed to wait for ready after the command */

	if ((vp_aux_ptr->vp_cmd & VMMSK) & VWAIT) { /* wait for ready */
		vpwait((int)VWAIT,dev);
	}

#ifdef VP_TRCM
	if (cmd == VSTAT) { /* status */
		printf("\nReturning status information\n");
	}
#endif

/* Now read the board's registers and report to the calling program. */
/* First grab the flags and clear a few. */

	diag->arg0 = vp_aux_ptr->vp_flags;
	vp_aux_ptr->vp_flags &= ~(VP_TMDM | VP_TCMD | VP_TNRY);
	diag->arg1 = vp_aux_ptr->vp_istat;
	diag->arg2 = vp_aux_ptr->vp_dstat;
	diag->arg3 = addr->vp_lfis;
	diag->arg4 = addr->vp_pfds;
	diag->arg5 = addr->vp_podd;
	diag->arg6 = (addr->vp_addmod << 8) | addr->vp_intvect;
	diag->arg7 = addr->vp_addhi;
	diag->arg8 = addr->vp_addlo;
	diag->arg9 = addr->vp_range;

	return 0;
}

/* vpwait() is called from the vp i/o control routine to wait
 * for device ready before or after issuing a command or mode
 * change to the device.  Its argument tells it whether it is
 * waiting before or after a command.  If it is waiting prior 
 * to issuing the command, and it times out, it will flag the
 * timeout and return a 0 value to the calling program.  If
 * it is waiting after issuing a command, and it times out,m it
 * will flag the error as occurring after a command and return
 * a value of 0.  If the device is ready when it is called, or
 * becomes ready within the timeout period, it will return a
 * value of 1.
 */
static int vpwait(wait_flag, dev)
int wait_flag;
dev_t dev;
{
	register vp_aux_type * vp_aux_ptr = &vp_aux[VPUNIT(dev)];
	register vp_registers *addr = vp_aux_ptr->vp_addr;  /* board */
	register ushort status;
	register ushort temp;
	int intlevel;
	int retval = 1;

#ifdef VP_TRCM
	printf("\nvpwait() called\n");
#endif

/* First raise processor priority level to make sure that we 
 * don't get an interrupt from the board while we are fiddleing
 * with bits.
 */

	intlevel = spl5();
	while ((vp_aux_ptr->vp_bstat & VP_BINUSE)
     	      && (!(vp_aux_ptr->vp_bstat & VP_BERROR))) { 
		if (slumber(vp_aux_ptr->vp_space, 
		    PCATCH|INTERRUPTABLE, iocomboost)) {
			vpreset(vp_aux_ptr);
			return 0;
		}
	}
     	if (vp_aux_ptr->vp_bstat & VP_BERROR) { 
		vpreset(vp_aux_ptr);
		return 0;
	}

	splx(intlevel); 

	status = addr->vp_lfis; /* Must avoid the optimizer */
	if (!(status & VP_DIRY)) { /* not ready - do the wait */

		intlevel = spl5(); /* save processor level */

/* Now turn on board's interrupt enable latch to allow the interrupt
 * flag to set - so if the board comes ready just after we check it
 * we won't miss the interrupt.
 */

		temp = (vp_aux_ptr->vp_lmsk |= VP_ENBI);
		addr->vp_lfis = temp;

/* If not ready, do the wait. */

		status = addr->vp_lfis; /* Must avoid the optimizer */

		while (!(status & VP_DIRY)) { /* not ready - do the wait */

		   vp_aux_ptr->vp_cseq++;   /* inc seq count for watch-dog */
		   vp_aux_ptr->vp_flags |= VP_CMDW;   /* set waiting flag */
		   if (slumber(vp_aux_ptr, 
		       PCATCH|INTERRUPTABLE, iocomboost)) {
			   vpreset(vp_aux_ptr);
			   return 0;
		   }
		 	/* sleep lowers priority so board */
	 		/* can interrupt */
			status = addr->vp_lfis;/* Must avoid the optimizer */
		}

/* Make sure that interrupt enable latch is off ( it will still be
 * on if the sleep wasn't done.
 */

		temp = (vp_aux_ptr->vp_lmsk &= ~VP_ENBI);
		addr->vp_lfis = temp;

/* Restore interrupt level. */

		splx(intlevel);
	}
/* If there was a time-out and we were waiting for ready prior to
 * issuing a command, adjust the flags to indicate accordingly.
 */

	if (vp_aux_ptr->vp_flags & VP_TCMD) { /* there was a timeout */

		retval = 0; /* 0 return value */

		if (wait_flag & VPWAT) { /* wait prior to command */

			vp_aux_ptr->vp_flags &= ~VP_TCMD;
			vp_aux_ptr->vp_flags |= ~VP_TNRY;

		}
		else { /* no time out */
			retval = 1;
		}
	}
	return retval;
}

/* Issue master clear to terminate dma and disable interrupts
 * This will also reset the various mode control bits and
 * may cause unexpected results if the calling program fails
 * to check for a timeout condition before re-starting dma.
 */
static vpreset(vp_aux_ptr)
vp_aux_type * vp_aux_ptr;	
{
	register vp_registers *addr = vp_aux_ptr->vp_addr;
	int intlevel;   /* will save interrupt level*/
	ushort temp;    /* write to dev through this*/

#ifdef VP_TRCM
	printf("\nvpreset() called\n");
#endif


	/* pulse master clear */
 	addr->vp_pfds = VP_MCLR;

	/* clear active flag */
   vp_aux_ptr->vp_flags &= ~VP_ACTV;

   /* clear dma time-out flag */
   vp_aux_ptr->vp_flags &= ~VP_TMDM;

	/* clear error */
   vp_aux_ptr->vp_bstat &= ~VP_BERROR;

	/* clear dma */
	vp_aux_ptr->vp_bstat &= ~VP_BINUSE;

	/* clear waiting flag */
	vp_aux_ptr->vp_flags &= ~VP_CMDW;

	/* clear timeout flag   */
	vp_aux_ptr->vp_flags &= ~VP_TCMD;

}
