#include "../sys/types.h"
#include "../sys/param.h"
#include "../sys/systm.h"
#include "../ml/mtpr.h"
#include "../sys/vm.h"
#include "../sys/buf.h"
#include "../ml/pte.h"
#include "../sys/user.h"
#include "../sys/file.h"
#include "../sys/uio.h"
#include "../sys/ioctl.h"
#include "../sys/mtio.h"
#include "../sys/errno.h"
#include "../sys/cmap.h"
#include "../sys/vbavar.h"
#include "../sys/proc.h"
#include "../sys/xymacro.h"
#include "../sys/xydata.h"
#include "../sys/elog.h"	/* for system activity */
#include "../sys/iobuf.h"	/* for system activity */
#include "../sys/kernel.h"	/* for system activity */



extern xyattach() ;
extern xyctlr_timeout() ;
extern xyminsize() ;
extern xyprobe() ; 
extern xyslave() ; 
extern xystrategy() ;
extern struct ctlr_parms_iopb xy_ctlr_setup_struct[] ;
extern struct drive_parms_iopb xy_drive_setup_struct[] ;
extern struct vba_ctlr *xy_vba_ctlr[] ;
extern struct vba_device *xy_vba_device[] ;
extern struct unit_tab xy_unit_info[] ;
extern struct ctlr_tab xy_ctlr_info[] ;
extern struct iopb xy_free_iopb_list_arr[] ;
#ifndef NOSG
extern struct sg_block xy_free_sg_list_arr[] ;
#endif
extern struct buf xy_free_buf_list_arr[] ;
extern xybuf ;

extern		drivename[];    /* drive type strings */

/*
** For system activity reporting.
*/
extern blkacty ;
extern struct iotime xystat[] ;
extern struct iobuf xytab[] ;
extern xynumiopb ;
extern nxyd ;
/*
** #defining NOSG makes the driver avoid scatter/gather.
** When using s/g, odd size records are not handled, and
** the maximum block size is 31k.
*/
#ifdef NOSG
extern char xybuffer[][XYASYNCBUFSZ*NBPG] ; 
#endif NOSG

struct vba_driver xydriver = 
{
	xyprobe, 
	xyslave, 
	xyattach, 
	0, /* go routine */
	xyaddr,
	"mt",
	xy_vba_device, 
	"xylogics", 
	xy_vba_ctlr,
	0
} ;


struct xyreg
{
	u_int status ;
	u_int error ;
} xyreg ;



/*
	Xyattach adds the specified drive to the list  of  drives
	on  the  same controller.  Also, it resets and configures
	the controller, if this has not yet been done, and  init-
	ializes the unit_tab stucture for this drive.
*/
/* void */
xyattach( device_info )
register struct vba_device *device_info ;
{
	register struct unit_tab *my_drive = 
		&xy_unit_info[device_info->ui_unit] ;

	register struct ctlr_tab *ctlr_info = 
		&xy_ctlr_info[device_info->ui_ctlr] ;

	register int i ;

	/* 
	** Initialize unit_tab structure for this drive.
	*/
	my_drive->block_num = 0 ;
	my_drive->file_num = 0 ;
	my_drive->last_op = XYNOP ;
	my_drive->last_resid = 0 ;
	my_drive->last_status = 0 ;
	my_drive->next_drive = ctlr_info->unit_list ;
	ctlr_info->unit_list = my_drive ;
	my_drive->can_read = TRUE ;
	my_drive->open = 0 ;
	my_drive->eof = 0 ;
	my_drive->bot = 0 ;
	my_drive->eot = 0 ;
	my_drive->wanted = 0 ;
	my_drive->vba_unit = device_info ;
	for (i=0; i<xynumiopb; i++)
	{
		my_drive->ondrivehead = NULL ;
		my_drive->ondrivetail = NULL ;
	}
}




/*
	Xyclose  checks  if  the  drive  is  online  and open for 
	writing. 
	If this is the case, then write
	two filemarks.  If this is a rewinding device, call
	xycmd  to  rewind;  otherwise,  position  before the last
	filemark in ianticipation of further  writes.   Mark  the
	drive as no longer open.
*/
/* void */
xyclose( dev, flags )
register dev_t dev ;
register int flags ;
{
	int unit = XYUNIT(dev) ;
	register struct unit_tab *my_drive = &xy_unit_info[unit] ;
	int priority ;

	if ( ! my_drive->settype )
	{
		xycmd( dev, DO_SYNC, 1 ) ;
#ifndef NOSG
		if ( my_drive->async )
		{
			memfree( (int)XYmap+unit*XYASYNCBUFSZ, XYASYNCBUFSZ, 1);
		}
#endif NOSG
	
		if ( my_drive->last_status & STATE_ONLINEANDREADY )
		{
			xyhandle_filemarks(dev, 1) ;
	
			if ( REWINDING_VARIETY(dev) && (! my_drive->bot) )
				xycmd(dev, MTREW, 1) ;
		}
		xycmd( dev, DO_SYNC, 1 ) ;
	}

	my_drive->open = FALSE ;
}
	



/*
	Xycmd allows internal and ioctl commands to use  xystart,
	and  therefore  the  same  command synchonization methods
	that blocked reads/writes use through physio.   It  fills
	in  an  appropriate buf structure with the requested com-
	mand and calls xystart.  When xystart returns, the opera-
	tion   is   done   (either  successfully,  or  with  non-
	recoverable errors).  Xycmd can then free the buf  struc-
	ture used for this ioctl, and wakeup processes waiting on
	the free_buf_list.  Note that since a limited  number  of
	buf structures are shared by all process accessing
	accessing xylogics drives, a process may have  to  sleep,
	waiting  for  a buf structure to be freed by the previous
	command.

	Returns: Residual count -- the number of operations requested
		minus the number actually performed.
*/
int
xycmd( dev, command, count )
register dev_t dev ;
register long command ;
register int count ;
{
	int unit = XYUNIT(dev) ;
	register struct unit_tab *my_drive= &xy_unit_info[unit] ;
	register struct buf *my_buf ;
	int residual ;
	int priority ;


	if ( command == DO_WAIT )
	{
		int waitmore = 4*60 ;

		xycmd( dev, DO_STAT, 1 ) ;
		priority = spl5() ;
		while ( ((my_drive->last_status & STATE_ONLINEANDREADY) != 
			STATE_ONLINEANDREADY) && --waitmore ) 
		{
			slumber( &lbolt, 0, delayboost ) ;
			splx(priority) ;
			xycmd( dev, DO_STAT, 1 ) ;
			priority = spl5() ;
		}
		if ( ! waitmore )
			xyerrmsg( XYUNIT(dev), XYWAIT_TIMEOUT, 0 ) ;
		return( 0 ) ;
	}
	else if ( command == DO_SYNC )
	{
		priority = spl5() ;

		if ( my_drive->doingsync == TRUE )
		{
			while ( my_drive->doingsync )
				slumber( &my_drive->doingsync, 0, iocomboost ) ;
			splx(priority) ;

			return( 0 ) ;
		}
		
		my_drive->doingsync = TRUE ;
		while ( my_drive->ondrivehead )
			slumber( &my_drive->doingsync, 0, iocomboost ) ;

		my_drive->doingsync = FALSE ;
		splx(priority) ;
		wakeup( &my_drive->doingsync ) ;
		return( 0 ) ;
	}


	/* 
	** Get an available buf structure.
	*/
	priority = spl5() ;
	xy_want_buf++ ;
	while ( ! xy_free_buf_list )
		slumber( &xy_free_buf_list, 0, iocomboost ) ;
	xy_want_buf-- ;
	splx(priority) ;

	my_buf = xy_free_buf_list ;
	xy_free_buf_list = xy_free_buf_list->b_forw ;
	

	/* 
	** Fill in buf structure with appropriate request.
	*/
	my_buf->b_flags = B_PHYS ;
	my_buf->b_forw = NULL ;
	my_buf->b_cmd = (struct buf *)command ;
	my_buf->b_bcount = count ;
	my_buf->b_bufsize = 0 ;
	my_buf->b_error = 0 ;
	my_buf->b_dev = dev ;
	my_buf->b_un.b_addr = NULL ;
	my_buf->b_blkno = my_drive->block_num ;
	my_buf->b_resid = 0 ;
	my_buf->b_start = 0 ;
	my_buf->b_proc = u.u_procp ;
	/* my_buf->b_iodone = xydone ;*/

	/* 
	** Start this request.
	*/
	xystart( my_buf ) ;
	residual = my_buf->b_resid ;



	/* 
	** Free buf structure.
	*/
	my_buf->b_forw = xy_free_buf_list ;
	xy_free_buf_list = my_buf ;
	if ( xy_want_buf )
		wakeup( &xy_free_buf_list ) ;
	return(residual) ;
}





/*
	Xyexecute places this procedure in the Unix  call  queue.
	When  time passes, this routine will issue
	a controller reset, and wakeup all processes sleeping  on
	buf  structures  or  iopb  structures on (drives on) this
	controller, after setting error bits in the  request  buf
	structures.
*/
/* void */
xyctlr_timeout( my_iopb )
register struct iopb *my_iopb ;
{
	int priority = spl5() ;
	int unit_num = my_iopb->unit_num ;
	register struct unit_tab *my_drive = &xy_unit_info[unit_num] ;
	register int i ;
	struct iopb *tmpiopb ;
	struct iopb *nextiopb ;


	printf("xylogics drive %x timed out (iopb @%x).\n",
		unit_num, my_iopb ) ;

/*			START DEBUG
	printf("handle drive %x\n", my_drive ) ;
			END DEBUG */

	/* 
	** Handle processes waiting for operations to complete
	*/
	for ( tmpiopb=my_drive->ondrivehead; tmpiopb; 
		tmpiopb = nextiopb)
		
	{
		if ( tmpiopb->iopb_ptr.request )
		{
/*				START DEBUG
			printf("timeout %x\n", tmpiopb->iopb_ptr.request) ;
				END DEBUG */

			if ( ! tmpiopb->async )
				tmpiopb->iopb_ptr.request->b_flags |= B_ERROR ;

			wakeup( tmpiopb->iopb_ptr.request ) ;
			nextiopb = tmpiopb->ondrivenext ;
			tmpiopb->ondrivenext = NULL ;
		}
	}
	my_drive->ondrivehead = NULL ;
	my_drive->ondrivetail = NULL ;

	/* 
	** Handle processes waiting for buf structures.
	*/
	wakeup( &xy_free_buf_list ) ;

	/* 
	**  Handle processes waiting for iopb structures.
	while ( xy_wait_iopb_list )
	{
		register struct buf *tmp_request = xy_wait_iopb_list ;

		wakeup(xy_wait_iopb_list) ;
		xy_wait_iopb_list = xy_wait_iopb_list->b_forw ;
		tmp_request->b_forw = NULL ;
	}
	?????
	*/

	/* 
	** Handle processes waiting for drive to become free.
	wakeup( my_drive ) ;
	?????
	*/


	/*
	** Handle processes waiting for sync.
	*/
	wakeup( &my_drive->doingsync ) ;

	/*
	** Reset controller (soft reset).
	*/
	{
		caddr_t controller = xy_vba_device[XYUNIT(my_drive->device)]->
			ui_mi->um_addr ;
		SET_CONTROL_REG(controller,  XYCONTROLLER_RESET) ;
	}

	splx( priority ) ;
}




/*
	Xydump dumps all of main memory to  the  specified  tape.
	Since  this  routine  is called only during a panic, it's
	best to not use interrupts (they may be  broken);  hence,
	xydump  polls for command completion.  The information is
	written in blocks of XYMAX_BLOCKSIZE bytes.
*/
int
xydump( device )
register dev_t device ;
{
	int start = 0x800 ;
	register int numbytes = maxfree ;
	int blocksize ;
	int waitmore ;
	static struct iopb dumpiopb ;
	caddr_t controller = XYCTLRADDRS(device) ;
	unsigned char ctlr_status ;
	

	if ( XYUNIT(device) > XYMAXDRIVES )
		return( ENXIO ) ;


	while ( numbytes > 0 )
	{
		blocksize = ( numbytes > XYMAX_BLOCKSIZE ) ? 
			XYMAX_BLOCKSIZE : numbytes ;

		dumpiopb.error_detect = 0 ;
		dumpiopb.done = 0 ;
		dumpiopb.chain = 0 ;
		dumpiopb.scat_gath = 0 ;
		dumpiopb.cmd = XYWRITE ;
		dumpiopb.completion_code = 0 ;
		dumpiopb.status = 0 ;
		dumpiopb.current_state = 0 ;
		dumpiopb.subfunction = XYNOSUBCODE ;
		dumpiopb.slave_num = XYSLAVE( device ) ;
		dumpiopb.list_length = 0 ;
		dumpiopb.interrupt_level = 0 ;
		dumpiopb.byte_count = blocksize ;
		dumpiopb.last_error = 0 ;
		dumpiopb.error_count = 0 ;
		dumpiopb.actual_count = 0 ;
		dumpiopb.data_addrs_mod = XYADDRS_MOD ;
		dumpiopb.priority = 0 ;
		dumpiopb.next_iopb_addrs_mod = 0 ;
		dumpiopb.data_addrs = start ;
		dumpiopb.next_iopb_addrs = 0 ;
		dumpiopb.iopb_chksum = 0 ;


		/* 
		** Wait till controller is ready.
		*/
		waitmore = WAITCOUNT ;
		volatile(controller) ;
		ctlr_status = GET_STATUS_REG(controller) ;
		while ( (ctlr_status &  (XYAIO | XYBUSY)) && --waitmore )
		{
			DELAY(1000) ;
			volatile(controller) ;
			ctlr_status = GET_STATUS_REG(controller) ;
		}

		if ( ! waitmore )
		{
			printf("xylogics dump device not ready.  Dump aborted.\n") ;
			return(ENXIO) ;
		}
	
		/* 
		** Send a command to controller.
		*/
		SET_IOPB_ADDRS(controller, &dumpiopb) ;
		SET_IOPB_ADDRS_MOD(controller,XYADDRS_MOD) ;
		SET_CONTROL_REG(controller,XYAIO) ;
	
	
		/* 
		** Wait till controller is done.
		*/
		waitmore = WAITCOUNT ;
		volatile(controller) ;
		ctlr_status = GET_STATUS_REG(controller) ;

		while ( ((ctlr_status & XYBUSY) || 
			!(ctlr_status & XYREMOVE_IOPB)) &&
			--waitmore ) 
		{
			DELAY(1000);
			volatile( controller ) ;
			ctlr_status = GET_STATUS_REG(controller) ;
		}


		if ( ! waitmore )
		{
			printf("xylogics: Controller %x at %x timed out during dump.\n", XYCTLRNUM(device), controller ) ;
			return( ENXIO ) ;
		}
	
		numbytes -= blocksize ;
		start += blocksize ;
	}
	return( SUCCESS ) ;
}





/* void */
xyerrmsg( unit, errnum, fromintr )
register int unit ;
register int errnum ;
int fromintr ;
{
	int mesgnum ;

	for ( mesgnum = 0; (xy_error_msg[mesgnum].errnum != errnum) && 
		(mesgnum < XYNUMERRMSG); mesgnum++ ) ;

	if ( xy_unit_info[unit].async || fromintr )
	{
		printf("xylogics tape unit %d: %s (err #%x)\n", 
			unit, xy_error_msg[mesgnum].errmsg, errnum ) ;
	}
	else
	{
		uprintf("xylogics tape unit %d: %s (err #%x)\n", 
			unit, xy_error_msg[mesgnum].errmsg, errnum ) ;
	}
/*
	xy_unit_info[unit].message = xy_error_msg[mesgnum].errmsg ;
*/
}






/*
	Xyexecute first queues a call to  xyctlr_timeout  on  the
	Unix  call  queue.   Then,  wait for the controller to be
	able to accept a command.  When the controller is  ready,
	pass  it  my_iopb by writing to the mapped registers, and
	return.

	Returns: FALSE when timeout occurs and in POLLED mode,
		TRUE otherwise.
*/
xyexecute( my_iopb, controller )
register struct iopb *my_iopb ;
register caddr_t controller ;
{
	register unsigned char ctlr_status ;
	register int ctlrnum = 
		xy_unit_info[my_iopb->unit_num].vba_unit->ui_ctlr ;

	register struct unit_tab *my_drive= &xy_unit_info[my_iopb->unit_num] ;
	struct buf *my_buf = my_iopb->iopb_ptr.request ;
	int priority ;

	/*
	** If doing a synchonous read or write, timeout it a few seconds;
	** otherwise, timeout in a few minutes.
	*/
	timeout(xyctlr_timeout, my_iopb, 
		((!my_drive->async)&&((my_iopb->cmd==XYREAD)||(my_iopb->cmd==XYWRITE))) ?
			hz*20: 60*hz*5) ; 

	priority = spl5() ;

	/* 
	** Wait till controller is ready.
	*/
	xy_ctlr_info[ctlrnum].wanted++ ;
	volatile(controller) ;
	ctlr_status = GET_STATUS_REG(controller) ;
	while ( ctlr_status &  XYAIO)
	{
		splx(priority) ;
		DELAY(10) ;
		volatile(controller) ;
		priority = spl5() ;
		ctlr_status = GET_STATUS_REG(controller) ;
	}
	xy_ctlr_info[ctlrnum].wanted-- ;


	/* 
	** Send a command to controller.
	*/
	SET_IOPB_ADDRS(controller, my_iopb) ;
	SET_IOPB_ADDRS_MOD(controller,XYADDRS_MOD) ;

	if ( my_drive->ondrivehead == NULL )
	{
		my_drive->ondrivehead = my_iopb ;
		my_drive->ondrivetail = my_iopb ;

		/*
		** For system activity reporting.
		*/
		xytab[my_iopb->unit_num].io_start = time.tv_sec ;
	}
	else
	{
		my_drive->ondrivetail->ondrivenext = my_iopb ;
		my_drive->ondrivetail = my_iopb ;
	}

	/*
	** For system activity reporting.
	*/
	blkacty |= ( 1 << major(my_drive->device)) ;

	SET_CONTROL_REG(controller,XYAIO) ;
	splx(priority) ;

	return ;
}






/* When doing asynchronous writes, allocate a space out of the async buffer
   (first fit) for specified drive of the specified length.  Update the 
   bidirectional linked list (threading iopbs) that indicates which ranges 
   of the buffer are currently in use.  
*/
char *
xygetbuf( my_drive, length, my_iopb )
register struct unit_tab *my_drive ;
int length ;
struct iopb *my_iopb ;
{
	struct iopb *tmpasync = my_drive->asynchead ;

	if (tmpasync == NULL) 
	{
		my_drive->asynchead = my_iopb ;
		my_iopb->asyncnext = NULL ;
		my_iopb->asynclast = NULL ;
		my_iopb->asyncstart = my_drive->asyncstart ;
		my_iopb->asyncend = my_iopb->asyncstart + length - 1 ;
		return( my_drive->asyncstart ) ;
	}
	if (tmpasync->asyncstart - my_drive->asyncstart >= length) 
	{
		tmpasync->asynclast = my_iopb ;
		my_iopb->asyncnext = tmpasync ;
		my_iopb->asynclast = NULL ;
		my_drive->asynchead = my_iopb ;
		my_iopb->asyncstart = my_drive->asyncstart ;
		my_iopb->asyncend = my_iopb->asyncstart + length - 1;
		return( my_drive->asyncstart ) ;
	}

	while ( (tmpasync->asyncnext != NULL) && 
		(length > (int)tmpasync->asyncnext->asyncstart  -
			(int)tmpasync->asyncend))
	{
		tmpasync = tmpasync->asyncnext ;
	}
	if ( tmpasync->asyncnext != NULL )
	{
		my_iopb->asyncnext = tmpasync->asyncnext ;
		my_iopb->asynclast = tmpasync ;
		tmpasync->asyncnext->asynclast = my_iopb ;
		tmpasync->asyncnext = my_iopb ;
		my_iopb->asyncstart = tmpasync->asyncend+1 ;
		my_iopb->asyncend = my_iopb->asyncstart + length - 1;
		return( (char *)my_iopb->asyncstart ) ;
	}
		
	if ( my_drive->asyncend - tmpasync->asyncend >= length )
	{
		my_iopb->asynclast = tmpasync ;
		my_iopb->asyncnext = NULL ;
		tmpasync->asyncnext = my_iopb ;
		my_iopb->asyncstart = tmpasync->asyncend+1 ;
		my_iopb->asyncend = my_iopb->asyncstart + length - 1;
		return( (char *)my_iopb->asyncstart ) ;
	}
	
	return( NULL ) ;
}






/* void */
xyhandle_filemarks(dev, moveback)
register dev_t dev ;
int moveback ;
{
	register struct unit_tab *my_drive = &xy_unit_info[XYUNIT(dev)] ;
	register int fmneeded ;
	int save_blocknum ;

	if ( ! my_drive->can_read )
	/*
	** !can_read implies the unit was opened for writing, and that
	** some form of write took place (either data or filemarks),
	** and that we have not already written logical end-of-tape filemarks.
	** The current block number is preserved across this operation.
	*/
	{
		save_blocknum = my_drive->block_num ;
		if ( ! my_drive->eot )
		{
			xycmd( dev, MTWEOF, 2 ) ;
			xycmd( dev, DO_SYNC, 1 ) ;
			xycmd( dev, MTBSF, moveback ) ;
		}
		my_drive->block_num = save_blocknum ;
	}
	my_drive->can_read = TRUE ;
}

	





/*
	Xyinit initializes global structures.  It is called only once,
	during autoconfiguration.
*/
/* void */
xyinit()
{
	register struct iopb *tmp_iopb, *last_iopb ;
	register struct buf *tmp_buf, *last_buf ;
	register int count ;
#ifndef NOSG
	register struct sg_block *tmp_sgblock, *last_sgblock ;

	xy_free_sg_list = &(xy_free_sg_list_arr[0]) ;
#else
	struct pte *bufpte;
	int i ;
#endif
	xy_wait_iopb_list = NULL ;
	xy_free_iopb_list = &(xy_free_iopb_list_arr[0]) ;
	xy_free_buf_list = &(xy_free_buf_list_arr[0]) ;


	/* 
	** Initialize list of free iopbs.
	*/
	tmp_iopb = xy_free_iopb_list ;

	for (count = 0; count < xynumiopb; count++ ) 
	{
		tmp_iopb->chain = 0 ;
		tmp_iopb->interrupt_level = INTERRUPT_LEVEL ;
		tmp_iopb->next_iopb_addrs = NULL ;
		tmp_iopb->data_addrs_mod = XYADDRS_MOD ;
		tmp_iopb->next_iopb_addrs_mod = NULL ;
		tmp_iopb->priority = FALSE ;
		tmp_iopb->iopb_chksum = FALSE ;

		tmp_iopb->iopb_ptr.iopb_next = 
			(struct iopb *)((unsigned int)tmp_iopb +
				sizeof( struct iopb )) ;

		tmp_iopb->asyncstart = NULL ;
		tmp_iopb->asyncend = NULL ;
		tmp_iopb->asyncnext = NULL ;
		tmp_iopb->asynclast = NULL ;

		tmp_iopb->ondrivenext = NULL ;
		tmp_iopb->frompool = TRUE ;

		last_iopb = tmp_iopb ;
		tmp_iopb = tmp_iopb->iopb_ptr.iopb_next ;
	}
	last_iopb->iopb_ptr.iopb_next = NULL ;
	
	


#ifdef NOSG
	/*
	** Make pages of xybuffer non-cached.
	*/
	bufpte = vtopte( (struct proc *)0, btop(xybuffer) ) ;
	for ( i=1; i<nxyd*XYASYNCBUFSZ; i++)
	{
		bufpte->pg_nc = 1 ;
		bufpte++ ;
	}
#else
	/* 
	** Initialize list of free scatter/gather blocks.
	*/
	tmp_sgblock = xy_free_sg_list ;
	
	for ( count=0; count<xynumiopb; count++ ) 
	{
		register int i ;

		for (i=0; i<NUM_SGIOPB_PER_IOPB; i++ )
			tmp_sgblock->sg_iopb[i].data_addrs_mod = XYADDRS_MOD ;

		tmp_sgblock->sg_next_block = 
			(struct sg_block *)((unsigned int)tmp_sgblock + 
				sizeof( struct sg_block )) ;

		last_sgblock = tmp_sgblock ;
		tmp_sgblock = tmp_sgblock->sg_next_block ;
	}
	last_sgblock->sg_next_block = NULL ;
#endif



	/* 
	** Initialize list of free buf structures.
	*/
	tmp_buf = xy_free_buf_list ;

	for ( count=0; count<xynumiopb; count++ ) 
	{
		tmp_buf->b_forw =
			(struct buf *)((unsigned int)tmp_buf +
				sizeof( struct buf )) ;

		last_buf = tmp_buf ;
		tmp_buf = tmp_buf->b_forw ;
	}
	last_buf->b_forw = NULL ;

}






/*
	Xyintr uses the mapped controller registers to  determine
	which  IOPB  has  completed.   It  examines this IOPB and
	determines its completion status.  If the operation  com-
	pleted  successfully,  or if there were unrecoverable er-
	rors, update the request status fields and wake the  pro-
	cess sleeping on this request (by issuing a wakeup on the
	associated request address).  If the
	completion status indicates that  the  drive  timed  out,
	reset  the drive, before waking any processes.  Since the
	controller is obviously alive,  xyintr  will  remove  the
	timeout  entry  (enqueued  in  xyexecute)  from  the call
	queue.
*/
/* void */
xyintr( ctlrnum )
int ctlrnum ;
{
	register caddr_t controller = xy_vba_ctlr[ctlrnum]->um_addr ;
	register struct iopb   *my_iopb ;
	unsigned char completion_code ;
	unsigned char completion_status ;
	unsigned char ctlr_status ;
	register struct unit_tab *my_drive ;
	register struct buf *my_buf ;
	register int unit ;
	int count ;
	int residual ;

	/* 
	** Verify that an iopb has completed.
	*/
	/*
		Could check for controller != 0 here.  This would avoid
		a panic on reboot without hard reset when an interrupt was
		pending from a configured xylogics board.  
	*/
	ctlr_status = GET_STATUS_REG(controller) ;
	if ( ! (ctlr_status & XYREMOVE_IOPB) )
	{
		printf("xylogics: spurious tape interrupt.  controller=%x\n", 
			controller) ;
		return ;
	}
	

	/* 
	** Fetch completed iopb, and check completion status.
	*/
	my_iopb = (struct iopb *)GET_IOPB_ADDRS(controller) ;	
	SET_CONTROL_REG(controller, XYREMOVE_IOPB) ;

	unit = my_iopb->unit_num ;
	my_drive = &xy_unit_info[unit] ;

	if ( my_iopb != my_drive->ondrivehead )
	{
		xyerrmsg( unit, XYOUTOFORDER, 1 ) ;
		printf("mydrive = %x    my_iopb=%x   expected=%x\n\n", 
			my_drive, my_iopb, my_drive->ondrivehead ) ;
	}
	else
	{
		my_drive->ondrivehead = my_drive->ondrivehead->ondrivenext ;
		if ( my_drive->ondrivehead == NULL )
		{
			my_drive->ondrivetail = NULL ;

			/*
			** For system activity reporting.
			*/
			xystat[unit].io_act += 
				time.tv_sec - xytab[unit].io_start ;
			blkacty &= ~(1 << major(my_drive->device)) ;
		}
		my_iopb->ondrivenext = NULL ;
	}

	uncache( &my_iopb->completion_code ) ;
	uncache( &my_iopb->status ) ;
	uncache( &my_iopb->current_state ) ;
	uncache( &my_iopb->actual_count ) ;
	uncache( &my_iopb->byte_count ) ;

	my_buf = my_iopb->iopb_ptr.request ;
	completion_code = my_iopb->completion_code ;
	completion_status = my_iopb->status ;


	if ( my_buf == NULL )
	{
		if ( (my_iopb->cmd == XYWRITE_PARAMS) || 
			(my_iopb->cmd == XYSET_DRIVE_PARAMS) )
		{
			wakeup( SLEEP_RESET ) ;
			return ;
		}
	}

	/* 
	** Update unit_tab structure for this drive.
	*/
	my_drive->last_op = (int)my_iopb->user_cmd ;
	my_drive->last_resid = residual ;  
	/*
	** The board does not handle status bits in any obvious way --
	** commands that don't actually go out to the drive (i.e. NOP)
	** latch certain status lines (bits), and zero other bits.  Therefore,
	** if this is a NOP, preserve status (except for online/ready bits).
	*/
#define CHANGE_STATUS (STATE_WRITE_PROTECT<<8)
#define CHANGE_STATE (STATE_ONLINEANDREADY | STATE_REWINDING)
	if ( (my_iopb->cmd != XYNOP) && (my_iopb->cmd != XYWRITE_PARAMS) )
		my_drive->last_status = 
			(completion_status<<8)|my_iopb->current_state ;
	else
		my_drive->last_status = 
			(my_drive->last_status& ~(CHANGE_STATUS|CHANGE_STATE)) |
			(my_iopb->status & CHANGE_STATUS)  |
			(my_iopb->current_state & CHANGE_STATE) ;
			
	my_drive->last_err = completion_code ;
	my_drive->bot = (my_drive->last_status & STATE_HIT_BOT) ? TRUE : FALSE ;
	my_drive->eot = (my_drive->last_status & (STATE_HIT_EOT<<8))?TRUE:FALSE;


	/* 
	** Remove timeout entry from call queue.
	*/
	untimeout( xyctlr_timeout, my_iopb ) ;

/* 			START DEBUG 
	if ( completion_code != 0 )
	{
		printf("NON-ZERO COMPLETION CODE=%x for iopb=%x\n", 
			completion_code, my_iopb) ;
		printf("halting...\n\n") ;
		asm(" halt") ;
	}
			END DEBUG */

	/* 
	** Respond to completion status.
	*/
	if ( ((completion_code&RECOVERY_CODE) == XYNOACTION) ||
		((completion_code&RECOVERY_CODE) == XYRECOVERED) ||
		( ! (completion_status & STATE_HARD_TAPE_ERROR) &&
			( (completion_code == LONG_RECORD) ||
			  (completion_code == SHORT_RECORD) ) ) ||
		(completion_code == FM_ON_READ) )
	{
		count = my_iopb->actual_count ;
		residual = my_iopb->byte_count;

		if ( my_drive->eot )
			printf("xylogics: Warning -- command completed after eot\n" ) ;

		switch (my_drive->last_op)
		{  
			case MTWEOF:  
				my_drive->block_num = 0 ;  
				my_drive->file_num += count ;  
				my_drive->eof = TRUE ;  
				break ;  
			case MTFSF:  
				my_drive->block_num = 0 ;  
				my_drive->file_num += count ;
				my_drive->eof = FALSE ;  
				break ;  
			case MTBSF:  
				my_drive->block_num = 0 ;  
				my_drive->file_num -= count ;  
				my_drive->eof = FALSE ;  
				break ;  
			case MTFSR:  
				my_drive->block_num += count ;
				my_drive->eof = 
					(my_drive->last_status & STATE_HIT_FM) ?
						 TRUE : FALSE ;
				break ;
			case MTBSR:  
				my_drive->block_num -= count ;
				my_drive->eof = FALSE ;
				break ;
			case DO_RWOV:
			case MTREW:  
				my_drive->block_num = 0 ;
				my_drive->file_num = 0 ;  
				my_drive->eof = FALSE ;
				break ;  
			case MTOFFL:  
				break ;  
			case MTNOP /* MTSTAT */:  
				break ;
			case DO_RRD:
			case DO_RWT:
			case DO_BRD:
			case DO_BWT:
				my_drive->block_num++ ;
				break ;
		}  
	}
	else /* Abnormal command completion */
	{
		switch ( completion_code & RECOVERY_CODE )
		{
			case XYRETRY:
					xyerrmsg(unit, completion_code, 1 ) ;
					if ( (! my_iopb->async) && my_buf )
					{
						my_buf->b_flags |= B_ERROR ;
						my_buf->b_error = EIO ;
					}
					break ;
			case XYPROG_ERR:
				xyerrmsg(unit, completion_code, 1 ) ;
				printf("Controller Error! \n") ;
				/* asm(" halt" ) ; */
				if ( (! my_iopb->async) && my_buf )
				{
					my_buf->b_flags |= B_ERROR ;
					my_buf->b_error = EIO ;
				}
				break ;
			case XYRESET_RETRY:
			{
				int density = 
					XYDENSITY(unit) ;
	
				xyerrmsg(unit, completion_code, 1 ) ;
				if ( (! my_iopb->async) && my_buf )
				/*xyreset_drive(density,unit, DO_POLL);*/
				{
					my_buf->b_flags |= B_ERROR ;
					my_buf->b_error = EIO ;
				}
				break ;
			}
			case XYNORETRY:
				xyerrmsg(unit, completion_code, 1 ) ;
				if ( (! my_iopb->async) && my_buf )
				{
					my_buf->b_flags |= B_ERROR ;
					my_buf->b_error = EIO ;
				}
				break ;
			case XYMISC:
			case XYMANUAL:
			case XYLOST:
				xyerrmsg(unit, completion_code, 1 ) ;
				if ( (! my_iopb->async) && my_buf )
				{
					my_buf->b_flags |= B_ERROR ;
					my_buf->b_error = EIO ;
				}
				break ;
			default:
				if ( (! my_iopb->async) && my_buf )
				{
					my_buf->b_flags |= B_ERROR ;
					my_buf->b_error = EIO ;
				}
				printf("xyintr: undefined recovery code(%x). controller=%x\n", completion_code, controller) ;
		}

		/* 
		** begin error logging 
		*/
		xytab[unit].b_dev = my_drive->device ;
		xytab[unit].b_actf = my_drive->async ? NULL : my_buf ;
		xytab[unit].io_addr = xy_vba_device[unit]->ui_addr ;
		xyreg.status = my_drive->last_status ;
		xyreg.error = my_drive->last_err ;
		fmtberr( &xytab[unit], 0, &xyreg ) ;
		logberr( &xytab[unit], 0 ) ;
		/* 
		** end error logging 
		*/
	}

	if ( (! my_iopb->async) && my_buf )
	{
		/*
		** Update buf structure residual from controller-updated 
		** residual.
		*/
		my_buf->b_resid = my_iopb->byte_count; 

		/* 
		** Set B_DONE and wakeup process waiting for this 
		** request.
		*/
		iodone( my_buf ) ;

		wakeup( my_buf ) ;


	}
	else /* ASYNCHRONOUS */
	{
		/* 
		** Release iopb.
		*/
		if ( (my_iopb->cmd == XYWRITE) || 
			(my_iopb->cmd == XYREAD) )
		{
			xyreadwrite_release(my_iopb, my_drive ) ;
		}

		if ( my_iopb->frompool )
		{
			my_iopb->iopb_ptr.iopb_next = xy_free_iopb_list;
			xy_free_iopb_list = my_iopb ;
			if ( xy_wait_iopb_list )
				wakeup( &xy_wait_iopb_list ) ;
		}
	}

	if ( my_drive->doingsync && (my_drive->ondrivehead == NULL) ) 
		wakeup( &my_drive->doingsync ) ;

	if ( xy_ctlr_info[ctlrnum].wanted )
		wakeup( controller ) ; 
}








/*
	If command is MTIOCTOP (Magnetic Tape  Input/Output  Con-
	trol -- Tape OPeration), and the operation is valid, call
	xycmd to handle things. If a  command  involves  backward
	movement  of tape, and filemarks are needed,
	use xycmd to write two filemarks
	before executing the user's command.  If  command
	is  MTIOCGET, fill in data in the appropriate values from
	the corresponding xy_unit_info structure.

	Returns: Residual count (= count requested - actual count done)
*/
int
xyioctl( dev, command, data, flag )
dev_t dev ;
int command ;
register struct mtop *data ;
int flag ;
{
	struct unit_tab *my_drive = 
		&xy_unit_info[XYUNIT(dev)] ;

	if ( (my_drive->eot) && ( ! back_movement(data->mt_op)) )
	{
		printf("attempt to move forward after hitting eot\n") ;
		return( ENXIO ) ;
	}
	if ( (!my_drive->can_read) && ((command == DO_SFMF) || 
		(command == DO_SPF)) )
	{
		return( ENXIO ) ;
	}

	switch (command)
	{
		case MTIOCTOP:
		{
			if ( data->mt_op <= USER_PERMITTED_IOCTLS )
			{
				/*
				** For system activity reporting.
				*/
				xytab[XYUNIT(dev)].io_stp->io_misc++ ;

				if (back_movement(data->mt_op))
					xyhandle_filemarks(dev, 2) ;
				if ( (xycmd(dev, data->mt_op, 
					data->mt_count)==0) ||
					(data->mt_op==DO_RWTA)) /*ctlr bug?*/
				{
					return( SUCCESS ) ;
				}
				else
					return ( EIO ) ;
			}
			else
			{
				if ( (data->mt_op == DO_SETTYPE) && 
					suser() && 
					((short)data->mt_drivetype <
						NUMDRIVETYPE))
				{
					int unit = XYUNIT(dev) ;

					xy_unit_info[unit].vba_unit->
						ui_type=(short)data->mt_drivetype;
					printf("xyioctl: drive %d switched to type %s\n", 
						unit, drivename[(short)data->mt_drivetype] ) ;
					return( SUCCESS ) ;
				}
				else
					return(EIO) ;
			}
		}

		case MTIOCGET:
		{
			register struct unit_tab *unit_info = 
				&xy_unit_info[XYUNIT(dev)];

			
			xycmd( dev, DO_WAIT, 1 ) ;
			((struct mtget *)data)->mt_type=MT_ISXYLOGICS;
			((struct mtget *)data)->mt_dsreg=unit_info->last_status;
			((struct mtget *)data)->mt_erreg=unit_info->last_err;
			((struct mtget *)data)->mt_resid=unit_info->last_resid;
			((struct mtget *)data)->mt_fileno=unit_info->file_num;
			((struct mtget *)data)->mt_blkno=unit_info->block_num;
			return( SUCCESS ) ;
		}

		default:
			return( ENXIO ) ; 
	}
}






/*
	If the  b_bcount  field  of  this  request  is  too  large
	(greater  than  XYMAX_BLOCKSIZE),  reduce it to the max-
	imum.
*/
/* void */
xyminsize( request )
register struct buf *request ;
{
	if ( request->b_bcount > XYMAX_BLOCKSIZE )
		request->b_bcount = XYMAX_BLOCKSIZE ;
}






/*
	Xyopen insures that this unit  is  valid:  "exists",  not
	currently  open, on-line, and ready for operations speci-
	fied by flag.  If all is ok, configure the drive  to  the
	appropriate    speed/density,    and    initialize    the
	xy_unit_info structure for this drive.
*/
int
xyopen( dev, flags )
dev_t dev ;
int flags ;
{
	int density = XYDENSITY(dev) ;
	int unit = XYUNIT(dev) ;
	register struct unit_tab *my_drive= &xy_unit_info[unit] ;
	register int i ;

	/* 
	** Insure unit is valid, and set density.
	*/
	if ( ! xy_vba_device[unit] ) 
		return ( ENXIO ) ;
	if ( my_drive->open )
		return ( EBUSY ) ;

	my_drive->open = TRUE ;

	/*
	** O_SETTYPE (SET drive TYPE) is provided so that the mt command
	** can open a drive and issue an ioctl MTIOCTOP DO_SETTYPE command.
	** Without this flag, we would have to set the density on a drive
	** without knowing what type it is.
	*/ 
	if ( ! (flags & FSETTYPE) )
	{

		my_drive->settype = FALSE ;
		my_drive->ondrivehead = NULL ;
		my_drive->ondrivetail = NULL ;
		my_drive->doingsync = FALSE ;
		my_drive->device = dev ;

		/*
		** begin error logging
		*/
		xytab[unit].io_nreg = (sizeof(struct xyreg) /
			sizeof(u_int)) ;
		xytab[unit].io_stp = &xystat[unit].ios ;
		/*
		** end error logging
		*/

		xycmd( dev, DO_STAT, 1 ) ;

		/*
		** Wait for drive to finish rewinding.
		*/
		if ( my_drive->last_status & STATE_REWINDING )
			xycmd( dev, DO_WAIT, 1 ) ;

		my_drive->async = FALSE ;

		if ( !(my_drive->last_status & STATE_ONLINEANDREADY) )
		{
			xyerrmsg( unit, XYNOTREADY, 0 ) ;
			my_drive->open = FALSE ;
			return( EIO ) ;
		}

		if ( (flags & FWRITE) && 
			(my_drive->last_status & (STATE_WRITE_PROTECT<<8)) )
		{
			xyerrmsg( unit, REEL_WRITE_PROTECTED, 0 ) ;
			my_drive->open = FALSE ;
			return( EIO ) ;
		}

		if ( XYSPEED(dev) && (flags & FWRITE) )
		{
			int offset ;
	
			my_drive->async = TRUE ;
			my_drive->xy_wait_buffer_list = NULL ;
	
#ifdef NOSG
			my_drive->asyncstart = &(xybuffer[unit][0]) ;
#else
			offset = unit * XYASYNCBUFSZ ;
			if (memall((int)XYmap+offset, XYASYNCBUFSZ,
				&proc[0], CSYS) == 0 )
			{
 				return( ENOMEM ) ;
			}
			vmaccess((int)XYmap+offset, (int)&xybuf+offset, 
				XYASYNCBUFSZ ) ;
	
			my_drive->asyncstart = 
				(char *)((int)(&xybuf)+unit*XYASYNCBUFSZ*NBPG) ;
#endif NOSG

			my_drive->asyncend = 
				(char *)((int)my_drive->asyncstart+
					XYASYNCBUFSZ*NBPG-1);

		}
		else
		{
			my_drive->async = FALSE ;
#ifdef NOSG
			my_drive->asyncstart = &(xybuffer[unit][0]) ;
			my_drive->asyncend = 
				(char *)((int)my_drive->asyncstart+
					XYASYNCBUFSZ*NBPG-1);
#else
			my_drive->asyncstart = NULL ;
			my_drive->asyncend = NULL ;
#endif
		}
		my_drive->asynchead = NULL ;
		my_drive->density_is_ok = FALSE ;
		my_drive->speed_is_ok = FALSE ;

		if ( ! xyreset_drive( density, dev, DO_NOT_POLL ) )
		{
			my_drive->open = FALSE ;
			return( EIO ) ;
		}
	}
	else
		my_drive->settype = TRUE ;




	/* 
	** Initialize drive structure.
	*/
	my_drive->open_flags = flags ;
	my_drive->block_num = 0 ;
	my_drive->file_num = 0 ;
	my_drive->last_op = XYNOP ;
	my_drive->last_resid = 0 ;
	/* last_status, last_err, bot, eot remain from last operation */
	my_drive->can_read = TRUE ;
	my_drive->eof = FALSE ;
	my_drive->message = NULL ;
	return( SUCCESS ) ;
}








/*
	Xyprobe checks for a Xylogics controller at the specified
	address. If one exists, a message is printed, and TRUE is
	returned; otherwise, false is returned.

	Returns: TRUE if probed and reset,
		FALSE otherwise
*/
xyprobe( vba_controller )
struct vba_ctlr *vba_controller ;
{
	static didinit = FALSE ;
	struct xylogics_registers *controller = 
		(struct xylogics_registers *)vba_controller->um_addr ;
	register struct ctlr_tab *ctlr_info = 
		&xy_ctlr_info[vba_controller->um_ctlr] ;
	int     ivct ;
	extern	int	Xxyintr0, Xxyintr1 ;
	static	int	xy_proc[] = { (int)& Xxyintr0, (int)& Xxyintr1 } ;

	ctlr_info->wanted = 0 ;
	/* On the very first probe, initialiaze global structures.
	   A better way to implement this would be to have a standard
	   "init" routine called from autoconfig, but we don't have
	   this yet. */    
	if ( ! didinit )
	{
		xyinit() ;
		didinit = TRUE ;
	}

	if ( badaddr(&controller->csr_reg, sizeof(controller->csr_reg), 0 ) )
		return( FALSE ) ;

	vba_controller->um_ivct = get_ivct( 0, 1 ) ;
	if ( vba_controller->um_ivct == (-1) )
		return( FALSE ) ;
	init_ivct( vba_controller->um_ivct, xy_proc[ vba_controller->um_ctlr ]);

	return( xyreset_ctlr(vba_controller, DO_POLL) );
}







/*
	Xyread allocates a buf structure for this drive, and sim-
	ply  calls  physio  with B_READ as the command to be pro-
	cessed.

	Returns: Count of bytes read.
*/
int
xyread( dev, uio )
dev_t dev ;
register struct uio *uio ;
{
	register struct buf *my_buf ;
	int error ;
	register struct unit_tab *my_drive = 
		&xy_unit_info[XYUNIT(dev)] ;
	int priority ;

	if ( my_drive->eot )
	{
		printf("attempt to read past eot\n") ;
		return( ENXIO ) ;
	}

	/* 
	** Get an available buf structure.
	*/
	priority = spl5() ;
	xy_want_buf++ ;
	while ( ! xy_free_buf_list )
		slumber( &xy_free_buf_list, 0, iocomboost ) ;
	xy_want_buf-- ;
	splx(priority) ;

	my_buf = xy_free_buf_list ;
	xy_free_buf_list = xy_free_buf_list->b_forw ;


	error = physio(xystrategy, my_buf, dev, B_READ, xyminsize, uio) ;

	/* 
	** Free buf structure.
	*/
	my_buf->b_forw = xy_free_buf_list ;
	xy_free_buf_list = my_buf ;
	if ( xy_want_buf )
		wakeup( &xy_free_buf_list ) ;

	if (! my_drive->speed_is_ok)
		xyreset_drive( XYDENSITY(dev), dev, DO_NOT_POLL ) ;
	        
	return( error ) ;
}






/*
	xyreadwrite_release releases the scatter_gather block used
	for this operation, and purges the dma buffers from the data cache
	(on a read operation).
*/
/* void */
xyreadwrite_release( my_iopb, my_drive )
struct iopb *my_iopb ;
struct unit_tab *my_drive ;
{
	struct sg_block *my_sg_block ;
	int i ;
	int priority = spl5() ;
#ifdef NOSG
	char *source_address ;
#endif
/*
release
*/


#ifndef NOSG
	if ( my_iopb->async && my_iopb->asyncend )
#endif
	{
		if ( my_iopb->asynclast == NULL )
			my_drive->asynchead = my_iopb->asyncnext ;
		else
			my_iopb->asynclast->asyncnext = my_iopb->asyncnext ;

		if ( my_iopb->asyncnext != NULL )
			my_iopb->asyncnext->asynclast = my_iopb->asynclast ;

#ifdef NOSG
		source_address = (char *)my_iopb->asyncstart ;
#endif NOSG
		my_iopb->asyncstart = NULL ;
		my_iopb->asyncend = NULL ;
		my_iopb->asyncnext = NULL ;
		my_iopb->asynclast = NULL ;

		if ( my_drive->xy_wait_buffer_list )
			wakeup( &my_drive->xy_wait_buffer_list ) ;
	}

#ifndef NOSG
	if ( my_iopb->scat_gath )
	{
		my_sg_block = my_iopb->scat_gath_block ;
		if ( my_iopb->cmd == XYREAD )
		{
			for ( i=0; i<my_iopb->list_length; i++ ) 
				mtpr( my_sg_block->uncache_address[i], P1DC );
		}
		my_sg_block->sg_next_block = xy_free_sg_list ;
		xy_free_sg_list = my_sg_block ;
	}
	else
#endif
		if ( my_iopb->cmd == XYREAD )
		{
#ifdef NOSG
			/*
			** If scatter/gather is disabled,
			** this was a transfer to a (non-cached) kernel buffer.
			*/
			/* mtpr( 1, PADC ) ; */
			copyout( source_address, 
				my_iopb->iopb_ptr.request->b_un.b_addr,
				my_iopb->actual_count ) ;
#else
			/*
			** If not a scatter/gather read, and s/g is enabled,
			** this must be a small transfer (all on one page)
			** directly into user's buffer.
			*/
			mtpr( my_iopb->iopb_ptr.request->b_un.b_addr, P1DC ) ;
#endif
		}

	splx(priority) ;
}








/*
	xyreadwrite_setup establishes scatter_gather iopbs to point to 
	each page of this buffer, and fills in the scat_gath, list_length, 
	data_addrs_mod, and data_addrs fields of the iopb.  The buffer
	has already been locked in memory, either by the buffered I/O
	system or by physio in the case of raw I/O.
*/
/* void */
xyreadwrite_setup( my_iopb, request, async )
register struct iopb *my_iopb ;
register struct buf *request ;
int async ;
{
	register struct sg_block *my_sg_block ;
	register caddr_t start_address  ;
	long length = request->b_bcount ;
	register struct proc *proc = u.u_procp ;
	register long physaddr ;
	register int length_onpage ;
	unsigned char curr_list_len ;
	int unit = XYUNIT(request->b_dev) ;
	struct unit_tab *my_drive = &xy_unit_info[unit] ;
	int priority ;

	
/*
setup
*/
#ifndef NOSG
	if ( async )
#endif NOSG
	{
		priority = spl5() ;
		if ( my_drive->xy_wait_buffer_list != NULL )
		/* enqueue on wait list, and sleep */
		{
			struct buf *tmp_request = my_drive->xy_wait_buffer_list;

			while (tmp_request->b_forw != NULL)
				tmp_request = tmp_request->b_forw ;
			tmp_request->b_forw = request ;
			request->b_forw = NULL ;
		}
		else
		{
			start_address = xygetbuf( my_drive, length, my_iopb ) ;
			if ( start_address == NULL )
			{
				my_drive->xy_wait_buffer_list = request ;
				request->b_forw = NULL ;
			}
		}


		if ( my_drive->xy_wait_buffer_list != NULL )
		{
			while (	(my_drive->xy_wait_buffer_list != request) ||
				( (start_address = 
				xygetbuf(my_drive, length, my_iopb))==NULL))
			{
				slumber(&my_drive->xy_wait_buffer_list, 0, iocomboost);
			}

			my_drive->xy_wait_buffer_list = 
				my_drive->xy_wait_buffer_list->b_forw ;
		}

		splx(priority) ;
#ifdef NOSG
		if (my_iopb->cmd == XYWRITE)
#endif NOSG
		{
			copyin( request->b_un.b_addr, start_address, 
				request->b_bcount);
		}
	}
#ifndef NOSG
	else
		start_address = request->b_un.b_addr ;
#endif NOSG

	physaddr =vtoph( proc, start_address ) ;
	length_onpage = NBPG - ((int)start_address & PGOFSET) ;

	/* 
	** (No scatter/gather iopbs needed)
	*/
#ifndef NOSG
	if ( length_onpage >= length )
#endif NOSG
	{
		my_iopb->scat_gath = FALSE ;
		my_iopb->list_length = 0 ;
		my_iopb->data_addrs = physaddr ;
		return ;
	}


#ifndef NOSG
	/* 
	** Get a block of scatter/gather iopbs.
	*/
	my_sg_block = xy_free_sg_list ;
	xy_free_sg_list = xy_free_sg_list->sg_next_block ;

	/* 
	** Setup initial scatter/gather block.
	*/
	my_sg_block->sg_iopb[0].data_addrs = physaddr ;
	my_sg_block->sg_iopb[0].byte_count = length_onpage ;
	my_sg_block->uncache_address[0] = start_address ;
	length -= length_onpage ;
	start_address += length_onpage ;


	/* 
	** Setup middle scatter/gather blocks.
	*/
	for ( curr_list_len = 1;
		(curr_list_len < NUM_SGIOPB_PER_IOPB) && (length > NBPG) ; 
		curr_list_len++ )
	{
		physaddr = vtoph( proc, start_address ) ;
		my_sg_block->sg_iopb[curr_list_len].data_addrs = physaddr ;
		my_sg_block->sg_iopb[curr_list_len].byte_count = NBPG ;
		my_sg_block->uncache_address[curr_list_len] = start_address ;
		length -= NBPG ;
		start_address += NBPG ;
	}

	/* 
	** Setup last scatter/gather block.
	*/
	if ( length > 0 )
	{
		physaddr = vtoph( proc, start_address ) ;
		my_sg_block->sg_iopb[curr_list_len].data_addrs = physaddr ;
		my_sg_block->sg_iopb[curr_list_len].byte_count = length ;
		my_sg_block->uncache_address[curr_list_len] = start_address ;
		curr_list_len++ ;
		if (curr_list_len == NUM_SGIOPB_PER_IOPB )
			curr_list_len = 0 ; 
	}
	
	
	my_iopb->scat_gath = TRUE ;
 	my_iopb->list_length = curr_list_len ;
	my_iopb->data_addrs = virt_to_phys( my_sg_block) ;
	/* Note: By the end of the operation, the controller will corrupt the
	   data_addrs field */
	my_iopb->scat_gath_block = my_sg_block ;
#endif
}








/* void */
xyreset(vba)
int vba ;
{
	printf("xylogics: xyreset NOT IMPLEMENTED.\n") ;
}





/*

	Returns: TRUE if reset successful
		FALSE otherwise

*/
xyreset_ctlr( vba_controller, poll )
register struct vba_ctlr *vba_controller ;
int poll ;
{
	register struct ctlr_parms_iopb *xy_ctlr_setup = 
		&(xy_ctlr_setup_struct[vba_controller->um_ctlr]) ;
	register caddr_t controller = vba_controller->um_addr ;
	unsigned char ctlr_status ;
	int waitmore ;

	/* 
	** Make sure this is a xylogics board.
	*/
	ctlr_status = GET_STATUS_REG(controller) ;
	SET_CONTROL_REG(controller,  XYCONTROLLER_RESET) ;
	waitmore = WAITCOUNT ;
	ctlr_status = GET_STATUS_REG(controller) ;
	while( ( ctlr_status & XYCONTROLLER_RESET ) && --waitmore )
	{
		DELAY(1000);
		ctlr_status = GET_STATUS_REG(controller) ;
	}


	if ( waitmore == 0 )
	{
		printf("xylogics controller timed out during soft reset.\n") ;
		return(FALSE) ;
	}

	/* Get status twice.  First time, should be all zero, or possibly
	** REG_BUSY_SEM (bit 0) set.  Second time, REG_BUSY_SEM should
	** definitely be set.
	*/
	volatile(controller) ;
	ctlr_status = GET_STATUS_REG(controller) ;
	volatile(controller) ;
	ctlr_status = GET_STATUS_REG(controller) ;

	if ( ctlr_status != XYREG_BUSY_SEM )
	{
/*
	WAITING FOR CONTROLLER FIX: REG_BUSY BIT IS CURRENTLY NON-FUNTIONAL
		printf("	xylogics soft reset failed.  value=%x\n", 
			ctlr_status);
		return( FALSE ) ;
*/
	}



	/* 
	** Form controller initialization iopb.
	*/
	xy_ctlr_setup->cmd = XYWRITE_PARAMS ;
	xy_ctlr_setup->scat_gath = 0 ;
	xy_ctlr_setup->chain = 0 ;
	xy_ctlr_setup->done = 0 ;
	xy_ctlr_setup->error_detect = 0 ;
	xy_ctlr_setup->completion_code = 0 ;
	xy_ctlr_setup->status = 0 ;
	xy_ctlr_setup->current_state = 0 ;
	xy_ctlr_setup->subfunction = XYWRITE_CTLR_PARMS ;
	xy_ctlr_setup->interrupt_level = 
		poll ? 0 : INTERRUPT_LEVEL ; 
	xy_ctlr_setup->interrupt_vector = 
		poll ? 0 : vba_controller->um_ivct;
	xy_ctlr_setup->AIO_response_time = 0 ; /* TUNABLE */
	xy_ctlr_setup->non_priv_reg = 1 ; /* TRUE */
	xy_ctlr_setup->iopb_cksum = 0 ; /* FALSE */ 
	xy_ctlr_setup->overlapped_rewind = 0 ;
	xy_ctlr_setup->write_wait_dma = 1 ; /* TRUE */
	xy_ctlr_setup->transfer_mode = 1 ; /* LONGWORD MODE */
	xy_ctlr_setup->auto_update = 1 ; /* TRUE */
	xy_ctlr_setup->release_on_request = 0 ; /* FALSE */
	xy_ctlr_setup->throttle_dead_time = 0 ;
	xy_ctlr_setup->throttle = XYTHROTTLE ;
	xy_ctlr_setup->next_iopb_addrs_mod = 0 ;
	xy_ctlr_setup->priority = 0 ;
	xy_ctlr_setup->next_iopb_addrs = 0 ;
	xy_ctlr_setup->iopb_chksum = 0 ;


	/* 
	** Initialize controller.
	*/
	SET_IOPB_ADDRS(controller,xy_ctlr_setup) ;
	SET_IOPB_ADDRS_MOD(controller, XYADDRS_MOD) ;
	SET_CONTROL_REG(controller, XYAIO) ;

	if ( poll )
	{
		/*
		** Poll till completion.
		*/
		waitmore = WAITCOUNT ;
		volatile(controller) ;
		ctlr_status = GET_STATUS_REG(controller) ;
		while ( ((ctlr_status & XYBUSY) || 
			!(ctlr_status & XYREMOVE_IOPB)) && 
			--waitmore )
		{
			DELAY(1000);
			volatile(controller) ;
			ctlr_status = GET_STATUS_REG(controller) ;
		}
		
		if ( waitmore )
		{
			SET_CONTROL_REG(controller, XYREMOVE_IOPB);
		}
		else
		{
			printf("xylogics controller reset failed (polled command timeout)\n") ;
			return (FALSE) ;
		}
	}
	else /* ...if using interrupts */
	{
		int priority = spl5() ;
		while ( ! xy_ctlr_setup->done )
			slumber( SLEEP_RESET, 0, iocomboost ) ;
		splx(priority) ;
	}
	
	if ( xy_ctlr_setup->completion_code == 0 )
	{
		return (TRUE) ;
	}
	else
	{
		printf("xylogics controller reset failed (bad completion code)\n") ;
		return (FALSE) ;
	}
}






/*
	Reset a drive.

	Returns: TRUE if drive reset is successful,
		FALSE otherwise
*/
xyreset_drive( density, device, poll )
int density ;
register dev_t device ;
int poll ;
{
	register caddr_t controller = XYCTLRADDRS(device) ;
	int slave = XYSLAVE(device) ;
	int drivetype = XYDRIVETYPE(device) ;
	int waitmore ;
	unsigned char ctlr_status ;
	register unit = XYUNIT(device) ;
	register int i ;
	register struct drive_parms_iopb *xy_drive_setup = 
		&(xy_drive_setup_struct[unit]);
	register struct iopb *xy_density_setup = 
		(struct iopb *)xy_drive_setup ;
	int priority ;
	struct unit_tab *my_drive = &xy_unit_info[unit] ;

	xy_drive_setup->cmd = XYWRITE_PARAMS ;
	xy_drive_setup->scat_gath = 0 ;
	xy_drive_setup->chain = 0 ;
	xy_drive_setup->done = 0 ;
	xy_drive_setup->error_detect = 0 ;
	xy_drive_setup->completion_code = 0 ;
	xy_drive_setup->status = 0 ;
	xy_drive_setup->current_state = 0 ;
	xy_drive_setup->subfunction = XYWRITE_DRIVE_PARMS ;
	xy_drive_setup->slave_num = slave ;
	xy_drive_setup->interrupt_level = 
		poll ? 0 : INTERRUPT_LEVEL ;
	xy_drive_setup->interrupt_vector = 
		poll ? 0 :
		       (xy_unit_info[XYUNIT(device)].vba_unit->ui_mi)->um_ivct ;
	xy_drive_setup->swap_bytes = 0 ;
	xy_drive_setup->swap_words = 0 ;
	xy_drive_setup->density_select = 0 ; 
	xy_drive_setup->gap_retry = 1 ;
	xy_drive_setup->write_retry = 1 ;
	xy_drive_setup->read_retry = 1 ;
	xy_drive_setup->ignore_write_error = 0 ;
	xy_drive_setup->ignore_read_error = 0 ;
	xy_drive_setup->density = 0x4 | density ;
	/* (select drive density via the five standard tape interface 
	   command lines) */

	/* NOTE: The STC2920 DOES NOT ALLOW DENSITY CHANGE WHEN AT
	   HIGH SPEED.  If we want to change density and be at high
	   speed, we will change density for now, and after the first
	   read or write, we will go to high speed.
	*/
	if ( my_drive->density_is_ok  ||
	    ( xy_density_command[drivetype][density] == 0 ) ||
	        ( XYSPEED( device ) == 0 ) )
	{
		xy_drive_setup->speed = XYSPEED( device ) ;
		my_drive->speed_is_ok = TRUE ;
	}
	else
	{
	        xy_drive_setup->speed = 0 ; /* Just for first read/write */
		/* my_drive->speed_is_ok = FALSE ; */
      	}
        my_drive->density_is_ok = TRUE ;

	xy_drive_setup->inhibit_parity_on_write = 1 ; 
	xy_drive_setup->extended_gap = 1 ;

	for (i=0; i<NUMDENSITYS; i++)
		xy_drive_setup->density_command[i] = 
			 xy_density_command[drivetype][i] ;

	xy_drive_setup->data_busy_timeout = DATA_BUSY_TIMEOUT ;
	xy_drive_setup->variable_gaps = 0 ;
	xy_drive_setup->next_iopb_addrs_mod = 0 ;
	xy_drive_setup->priority = 0 ;
	xy_drive_setup->next_iopb_addrs = 0 ;
	xy_drive_setup->iopb_chksum = 0 ;
	xy_drive_setup->unit_num = unit ;

	priority = spl5() ;

	/* 
	** WAIT till controller is ready.
	*/
	waitmore = WAITCOUNT ;
	volatile(controller) ;
	ctlr_status = GET_STATUS_REG(controller) ;
	while ( (ctlr_status & XYAIO)  && --waitmore ) 
	{
		splx(priority) ;
		DELAY(10);
		volatile(controller) ;
		priority = spl5() ;
		ctlr_status = GET_STATUS_REG(controller) ;
	}

	if ( waitmore == 0 )
	{
		printf("xylogics drive %d on controller at %x reset failed (controller timeout)\n", slave, controller ) ;
		return( FALSE ) ;
	}

	/* 
	** Initialize drive.
	*/
	SET_IOPB_ADDRS(controller,xy_drive_setup) ;
	SET_IOPB_ADDRS_MOD(controller, XYADDRS_MOD) ;

	if ( ! poll )
	{
		if ( my_drive->ondrivehead == NULL )
		{
			my_drive->ondrivehead = (struct iopb *)xy_drive_setup ;
			my_drive->ondrivetail = (struct iopb *)xy_drive_setup ;
		}
		else
		{
			my_drive->ondrivetail->ondrivenext = 
				(struct iopb *)xy_drive_setup ;
			my_drive->ondrivetail = (struct iopb *)xy_drive_setup ;
		}

		xy_drive_setup->ondrivenext = NULL ;

		SET_CONTROL_REG(controller, XYAIO) ;

		while ( ! xy_drive_setup->done )
			slumber( SLEEP_RESET, 0, iocomboost ) ;
		splx(priority) ;
	}
	else
	{
		SET_CONTROL_REG(controller, XYAIO) ;
		splx(priority) ;

		/*
		** Poll till completion.
		*/
		waitmore = WAITCOUNT ;
		volatile(controller) ;
		ctlr_status = GET_STATUS_REG(controller) ;
		while ( ((ctlr_status & XYBUSY) || 
			!(ctlr_status & XYREMOVE_IOPB)) && 
			--waitmore )
		{
			DELAY(1000);
			volatile(controller) ;
			ctlr_status = GET_STATUS_REG(controller) ;
		}
		
		if ( waitmore )
		{
			SET_CONTROL_REG(controller, XYREMOVE_IOPB);
		}
		else
		{
			printf("xylogics drive reset failed (polled command timeout)\n") ;
			return (FALSE) ;
		}
	}




	if ( (xy_unit_info[unit].bot) &&     /* set density will fail if !BOT */
		(xy_drive_setup->completion_code != 0) )
	{
		printf("xylogics drive %d on controller at %x reset failed (completion code)\n", slave, controller ) ;
		return( FALSE ) ;
	}
}







/*
	Xyseek positions the tape at the correct record in  order
	to  do  a  blocked  read/write.   This is done by calling
	xycmd  with  the DO_SPF,  or DO_SPB  commands,  with  the 
	appropriate counts.
*/
int
xyseek( request )
register struct buf *request ;
{
	register struct unit_tab *unit_info = 
		&xy_unit_info[XYUNIT(request->b_dev)] ;

	/* 
	** Seek backwards
	*/
	if ( request->b_blkno < unit_info->block_num )
	{
		xyhandle_filemarks(request->b_dev, 2) ;
		if ( unit_info->bot )
			return( ENXIO ) ;
		else
			xycmd(request->b_dev, MTBSR, unit_info->block_num-request->b_blkno) ;
	}
	/* 
	** Seek forwards
	*/
	else if ( request->b_blkno > unit_info->block_num )
	{
		if ( unit_info->eof || unit_info->eot )
			return( ENXIO ) ;
		else
			xycmd(request->b_dev, MTFSR, request->b_blkno-unit_info->block_num) ;
	}
	request->b_blkno = unit_info->block_num + 1 ;
	return( SUCCESS ) ;
}







/*
	Xyslave assumes a drive is present,  and  simply  returns
	TRUE.   Thus, the os will think that all drives specified
	in the configuration file (when the OS  was  linked)  are
	available.

	Returns: TRUE if drive is present (currently, always the case)
		FALSE otherwise
*/
xyslave( drive_info, ctlr_addr )
struct vba_device *drive_info ;
caddr_t ctlr_addr ;
{
	return (TRUE) ;
}







/*
	Xystart forms an IOPB for this request, then calls xyexe-
	cute to execute this IOPB. After calling xyexecute, sleep
	(on the address of the request) until we are notified  of
	command  completion  (via a wakeup from either the inter-
	rupt or timeout routines).  
	Once we give up trying to execute
	the command, release the IOPB structure used for this re-
	quest  and  return.   Note that since all IOPB structures
	are preallocated (during system initialization), the pro-
	cess may have to sleep in xystart until an IOPB structure
	is released by completion of a previous  tape  operation.
	Also,  note that forming an iopb will involve locking the
	data buffer in memory, and forming scatter/gather  iopbs,
	each pointing to 1k chunks of the buffer.
*/
/* void */
xystart( request )
register struct buf *request ;
{
	register struct iopb *my_iopb ;
	register struct unit_tab *my_drive = 
		&xy_unit_info[XYUNIT(request->b_dev)] ;
	int xycommand  = logical_to_xycmd[(int)request->b_cmd] ;
	int asyncmode ;
	int priority ;

	if ( xycommand == XYHANDLER )
	{
		switch ((int)request->b_cmd)
		{
			default:
				printf("xylogics: unknown command(%x).\n", 
					(int)request->b_cmd ) ;
		}
		return ;
	}

	/* 
	** Fetch iopb.
	*/
	priority = spl5() ;
  	if ( ! xy_free_iopb_list )
	{
		/* enqueue on wait list, and sleep */
		if ( xy_wait_iopb_list != NULL )
		{
			struct buf *tmp_request = xy_wait_iopb_list ;

			while (tmp_request->b_forw != NULL)
				tmp_request = tmp_request->b_forw ;
			tmp_request->b_forw = request ;
		}
		else
			xy_wait_iopb_list = request ;

		request->b_forw = NULL ;

		do
		{
			slumber( &xy_wait_iopb_list, 0, iocomboost) ;
		}
		while ((!xy_free_iopb_list) || (xy_wait_iopb_list != request)) ;

		/* dequeue from wait list */
		xy_wait_iopb_list = xy_wait_iopb_list->b_forw ;
	}
	splx(priority) ;
			
	/* We have a free iopb. */
	my_iopb = xy_free_iopb_list ;
	xy_free_iopb_list = xy_free_iopb_list->iopb_ptr.iopb_next;


	/* 
	** Fill in iopb.
	*/
	my_iopb->cmd = xycommand ; 
	my_iopb->subfunction = logical_to_xysub[(int)request->b_cmd] ;

	/*
	** READ and STAT commands always use the synchronous interface.  
	** Other commands may use the asynchronous interface, if the drive 
	** was opened for asynchronous activity.
	*/
	if ( (int)request->b_cmd == DO_STAT )
	{
		asyncmode = FALSE ;
	}
	else
	{
		asyncmode = my_drive->async ;
	}


	if ( xycommand == XYWRITE )
	{
		xyreadwrite_setup(my_iopb, request, my_drive->async ) ;
		my_drive->can_read = FALSE ;
	}
	else if ( xycommand == XYREAD )
	{
		asyncmode = FALSE ;
		xyreadwrite_setup(my_iopb, request, FALSE ) ;
	}
	else
	{
		my_iopb->scat_gath = 0 ;
		my_iopb->list_length = 0 ;
		my_iopb->data_addrs = 0 ;
		if ( (int)request->b_cmd == MTWEOF )
			my_drive->can_read = FALSE ;
	}
	my_iopb->async = asyncmode ;
	my_iopb->slave_num = XYSLAVE(request->b_dev) ;
	my_iopb->done = 0 ;
	my_iopb->error_detect = 0 ;
	my_iopb->completion_code = 0 ;
	my_iopb->status = 0 ;
	my_iopb->current_state = 0 ;
	my_iopb->interrupt_vector = my_drive->vba_unit->ui_mi->um_ivct ;
	my_iopb->byte_count = request->b_bcount ;
	my_iopb->last_error = 0 ;
	my_iopb->error_count = 0 ;
	my_iopb->actual_count = 0 ;
	my_iopb->iopb_ptr.request = request ;
	my_iopb->user_cmd = (int)request->b_cmd ;

	priority = spl5() ;
	while( my_drive->doingsync )
		slumber( &my_drive->doingsync, 0, iocomboost ) ;
	splx(priority) ;

	my_iopb->unit_num = XYUNIT(request->b_dev) ;

	/* 
	** Execute iopb.
	*/
	xyexecute( my_iopb, XYCTLRADDRS(request->b_dev) ) ;
	if ( asyncmode )
	{
		request->b_resid = 0 ;
		iodone( request ) ;
		return ;
	}

	/* IF SYNCHRONOUS */
	priority = spl5() ;
	while ( ! (request->b_flags & (B_DONE|B_ERROR)) )
		slumber( request, 0, iocomboost ) ;
	splx(priority) ;

	if ( (my_iopb->cmd == XYWRITE) || (my_iopb->cmd == XYREAD) )
	{
		xyreadwrite_release(my_iopb, my_drive ) ;
	}

	if ( my_iopb->frompool )
	{
		my_iopb->iopb_ptr.iopb_next = xy_free_iopb_list ;
		xy_free_iopb_list = my_iopb ;
		if ( xy_wait_iopb_list )
			wakeup( &xy_wait_iopb_list ) ;

	}

/*
	if ( my_drive->message )
	{
		uprintf("xylogics tape unit %d: %s\n", XYUNIT(my_drive->device),
			my_drive->message ) ;
		my_drive->message = NULL ;
	}
*/

}








/*
	Xystrategy verifies that this read/write request is on  a
	sufficiently small buffer.  If not, set fields in the re-
	quest structure to indicate an error and set the residual
	count  to the number of bytes requested, and call iodone.
	If the request  is  valid,  fill  in  the  command  field
	(b_cmd)  of  the  request  structure with the appropriate
	read/write command, and call xystart to process the  com-
	mand.   When  xystart returns, the operation is done (ei-
	ther successfully, or with non-recoverable errors).
*/
/* void */
xystrategy( request )
register struct buf* request ;
{
	register int unit = XYUNIT( request->b_dev ) ;
	register struct unit_tab *my_drive = 
		&xy_unit_info[unit] ;
	/* 
	** Verify that buffer is small enough to do in a single transfer.
	*/
	if ( request->b_bcount > XYMAX_BLOCKSIZE )
	{
		xyerrmsg( XYUNIT(request->b_dev), XYTOOLARGE, 0 ) ;
		request->b_resid = request->b_bcount ;
		request->b_flags |= B_ERROR ;
		request->b_error = EIO ;
		iodone( request ) ;
		return ;
	}


	if ( (! my_drive->can_read) && (request->b_flags&B_READ))
	{
		request->b_resid = request->b_bcount ;
		request->b_flags |= B_ERROR ;
		request->b_error = ENXIO ;
		iodone(request ) ;
		return ;
	}
	

	/* 
	** Fill in b_cmd field for xystart.
	*/
	if(request->b_flags & B_PHYS)
		if(request->b_flags & B_READ)
			request->b_cmd = (struct buf *)DO_RRD;
		else
			request->b_cmd = (struct buf *)DO_RWT;
	else
		if(request->b_flags & B_READ)
			request->b_cmd = (struct buf *)DO_BRD;
		else
			request->b_cmd = (struct buf *)DO_BWT;

	/*
	** For system activity reporting.
	*/
	xytab[unit].io_stp->io_ops++ ;
	xystat[unit].io_bcnt += btoc(request->b_bcount) * 2 ; /* 512B bl */
	request->b_start = time.tv_sec ;

	/* 
	** Start request.
	*/
	xystart(request) ;

	/*
	** For system activity reporting.
	*/
	xystat[unit].io_resp += time.tv_sec - request->b_start ;
}






/*
	Xywrite allocates a buf structure  for  this  drive,  and
	simply call physio with B_WRITE as the command to be pro-
	cessed.

	Returns: Number of bytes written.
*/
int
xywrite( dev, uio )
dev_t dev ;
struct uio *uio ;
{
	register struct buf *my_buf ;
	int error ;
	register struct unit_tab *my_drive = 
		&xy_unit_info[XYUNIT(dev)] ;
	int priority ;

	if ( my_drive->eot )
	{
		printf("attempt to write past eot.\n") ;
		return( ENXIO ) ;
	}
	
	/* 
	** Get an available buf structure.
	*/
	priority = spl5() ;
	xy_want_buf++ ;
	while ( ! xy_free_buf_list )
		slumber( &xy_free_buf_list, 0, iocomboost ) ;
	xy_want_buf-- ;
	splx(priority) ;

	my_buf = xy_free_buf_list ;
	xy_free_buf_list = xy_free_buf_list->b_forw ;


	error = physio(xystrategy, my_buf, dev, B_WRITE, xyminsize, uio) ;

	/* 
	** Free buf structure.
	*/
	my_buf->b_forw = xy_free_buf_list ;
	xy_free_buf_list = my_buf ;
	if ( xy_want_buf )
		wakeup( &xy_free_buf_list ) ;

	if (! my_drive->speed_is_ok)
		xyreset_drive( XYDENSITY(dev), dev, DO_NOT_POLL ) ;

	return( error ) ;
}









/*
	This is a dummy procedure which is called whenever there is danger
	of the optimizer optimizing out loop invariants that are actually
	volatile (that is, memory-mapped locations on controller boards).
*/
/* void */
volatile(address)
caddr_t address ;
{} ;
