
/*	autoconf.c	4.31	81/07/05	*/
/*  	HCX version, 	  1/13/83 		*/

/*
 * Setup the system to run on the current machine.
 *
 * Configure() is called at boot time and initializes the vba 
 * device tables and the memory controller monitoring.  Available
 * devices are determined (from possibilities mentioned in ioconf.c),
 * and the drivers are initialized.
 *
 */

#include "../ml/pte.h"
#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/map.h"
#include "../sys/buf.h"
#include "../sys/dk.h"
#include "../sys/vm.h"
#include "../sys/conf.h"
#include "../sys/dmap.h"
#include "../ml/mem.h"
#include "../ml/mtpr.h"
#include "../sys/vbavar.h"
#include "../sys/reboot.h"

#define false 0
#define true  1

/*
 * The following several variables are related to
 * the configuration process, and are used in initializing
 * the machine.
 */
	struct pte *	pte_ptr		= (struct pte *)PRIMmap ;
	int             ptes            = IOSIZE ;

	/* The following are used by the set_up_pte procedure. */
	char *  	virt_addr	= Pmem ;
	unsigned int	page_no ;

/*
 * Determine mass storage configuration for a machine.
 */
configure()
{
	register int *ip;
	extern char Sysbase[];
	extern int nvba ;
	extern int ignore_strays ;

        switch ( RB_PRIMARY_IO( boot_device ) ) {
        case RB_NO_BUS:
        case RB_VERSA_BUS:
        default:
		printf("\nControllers found on Primary VERSAbus:\n\n");
		bus_find( 0, "VERSA" ) ;
		break ;
        case RB_VME_BUS:
		printf("\nControllers found on Primary VMEbus:\n\n");
		bus_find( 0, "VME" ) ;
		break ;
	}

        if ( nvba > 1 ) {
        	switch ( RB_SECONDARY_IO( boot_device ) ) {
        	case RB_NO_BUS:
        	default:
        		printf( "There is no secondary bus.\n" ) ;
        		break ;
        	case RB_VERSA_BUS:
			printf("\nControllers found on Secondary VERSAbus:\n\n");
			bus_find( 1, "VERSA" ) ;
			break ;
        	case RB_VME_BUS:
			printf("\nControllers found on Secondary VMEbus:\n\n");
			bus_find( 1, "VME" ) ;
			break ;
		}
	}
	ignore_strays = false ;

	/*
	 * Write protect the scb.  It is strange
	 * that this code is here, but this is as soon
	 * as we are done mucking with it, and the
	 * write-enable was done in assembly language
	 * to which we will never return.
	 */
	ip = (int *)&Sysmap[2]; *ip &= ~PG_PROT; *ip |= PG_KR;
	mtpr(Sysbase+0x800, TBIS);
    	if (boothowto & RB_GENERIC)
		setconf();
	else
	        if ( !defaultconf( ) ) {
	                printf( "default configuration failed\n" ) ;
	        	setconf( ) ;
	        }
}

/*
 * Find devices on the BUS.
 * Uses per-driver routine to see who is on the bus
 * and then fills in the tables, with help from a per-driver
 * slave initialization routine.
 */
int
bus_find( numvba, name_ptr )
	int	numvba ;
	char *	name_ptr ;
{
	register struct vba_device *ui;
	register struct vba_ctlr *um;
	struct vba_ctlr *um_ptr;
	u_short * virt_addr ;
	int	phys_addr;
	int	old_addr ;
	struct vba_driver *udp;
	int     slot ;
	int     hvme_id ;
	int	ctlr ;
	int	found ;
	int	alive ;
	int	end ;

	/*
	 * Check each controller.
	 * For each one which is potentially on this vba,
	 * see if it is really there, and if it is record it and
	 * then go looking for slaves.
	 */
	for (um = vbminit; udp = um->um_driver; um++) {
		if (um->um_vbanum != numvba && um->um_vbanum != '?')
			continue;
		if (um->um_alive == 1)
			continue ;
		/*
		 * Save user supplied value in case the controller
		 * does not respond.
		 */
		old_addr = (int)um->um_addr ;
		if ( old_addr == '?' ) {
			ctlr = HVME_MIN_SLOT ;
			end  = HVME_MAX_SLOT ;
		} else {
			ctlr = end = old_addr ;
		}

		for ( ; ctlr <= end ; ctlr ++ ) {

			/*
			 * Set up phys_addr as the physical address of the board.
			 */
			phys_addr = get_physical( ctlr, udp->hvme_id, numvba );

	                /*
	                 * Check too be sure this address is not alive.
	                 */
	                alive = false ;
			for (ui = vbdinit; ui->ui_driver; ui++)
				if ( (ui->ui_physaddr == (caddr_t)phys_addr) &&
				     (ui->ui_alive) ) {
				     	alive = true ;
					break ;
				}
			if ( alive )
				continue ;

			/*
			 * Now, set up virt_addr as the virtual address.
			 */
			virt_addr = (u_short *)set_up_pte( phys_addr,
							   1 ) ;
			um->um_addr = (caddr_t)virt_addr ;

			/*
			 * The interface to probe has been changed.
			 * It used to have the virtual address as
			 * the parameter.  This is not enough
			 * information to initialize the controller.
			 * So, the parameter is the pointer to the
			 * vba_ctlr structure.
			 */
			found = (*udp->ud_probe)( um ) ;
			if ( found )
				break ;
			remove_pte( 1 ) ;
		} /* end_for */
		if ( ! found ) {
			um->um_addr = (caddr_t)old_addr ;
			continue ;
		}

		um->um_vbanum = numvba;
		printf("%s%d at %sbus address %x \n",
		    udp->ud_mname, um->um_ctlr, name_ptr, phys_addr);
		um->um_alive = 1;
		udp->ud_minfo[um->um_ctlr] = um;
		for (ui = vbdinit; ui->ui_driver; ui++) {
			if (ui->ui_driver != udp || ui->ui_alive ||
			    ui->ui_ctlr != um->um_ctlr && ui->ui_ctlr != '?' ||
			    ui->ui_vbanum != numvba && ui->ui_vbanum != '?')
				continue;
			if ((*udp->ud_slave)(ui, virt_addr)) {
				ui->ui_alive = 1;
				ui->ui_ctlr = um->um_ctlr;
				if ( ctlr > HVME_MAX_SLOT )
					ui->ui_flags = um->um_ctlr;
				else
					ui->ui_flags = (short)ctlr ;
				ui->ui_vbanum = numvba;
				ui->ui_addr = (caddr_t)virt_addr ;
				ui->ui_physaddr = (caddr_t)phys_addr ;
				ui->ui_mi = um;
				/* ui_type comes from driver */
				udp->ud_dinfo[ui->ui_unit] = ui;
				if ( ui -> ui_slave != -1 )
					printf("        Drive %d : %s\n",
					    ui->ui_slave, udp->ud_dname);
				(*udp->ud_attach)(ui);
			}
		}
	}
}

int
get_physical( addr, hvme_id, numvba )
	int	addr ;
	int	hvme_id ;
	int	numvba ;
{
	int	phys_addr ;

	if ( (addr >= HVME_MIN_SLOT) &&
	     (addr <= HVME_MAX_SLOT) )
		phys_addr = (addr << 24) | (hvme_id << 16) ;
	else if ( ((addr & 0xff000000) == 0) ||
	          ((addr & 0xff000000) == 0xff000000) ) {
		phys_addr  = addr & 0x00ffffff ;
		phys_addr |= 0x3f000000 ;
	} else
	       	phys_addr  = addr & 0x3fffffff ;
	phys_addr |= 0x80000000 ;
	if ( numvba == 0 )
		phys_addr |= 0x40000000 ;
	return( phys_addr ) ;
}

/*
 *  Return the unit number of the specified dsk device.
 */
int
get_dsk_unit( bus, ctlr, drive )
	int     bus ;
	int     ctlr ;
	int     drive ;
{
	register struct vba_device *ui;
	for (ui = vbdinit; ui->ui_driver; ui++) {
	        if ( ui -> ui_alive &&
	             ui -> ui_dk &&
	             ui -> ui_vbanum == bus &&
	             ui -> ui_flags  == ctlr &&
	             ui -> ui_slave  == drive )
	        	return( ui -> ui_unit ) ;
	}
	return( -1 ) ;
}

/*
 *  Return the major number of the specified dsk unit.
 */
int
get_dsk_major( unit )
	int     unit ;
{
#define MAJOR_VD        1
#define MAJOR_HD        6
	register struct vba_device *ui;
	for (ui = vbdinit; ui->ui_driver; ui++)
  	        if ( ui->ui_alive &&
  	             ui->ui_dk &&
  	             ui->ui_unit == unit )
  	                if ( ui->ui_dk == DK_VDDC )
	                        return( MAJOR_VD ) ;
	                else
	                        return( MAJOR_HD ) ;
	return( -1 ) ;
}

/*
 * Make the controllers accessible at physical address phys
 * by mapping kernel ptes starting at pte.
 */

int
set_up_pte( phys_addr, nptes )
	caddr_t phys_addr ;
	int     nptes ;
{
	register int            start_virt_addr ;

	if ( phys_addr != 0 )
		page_no	= btop( phys_addr );
	start_virt_addr = (int)virt_addr + ((int)phys_addr & 0x3ff) ;
	do {
		virt_addr += 1024 ;
		*(int *)pte_ptr ++ = PG_V | PG_KW | PG_N | page_no ++ ;
		if ( -- ptes == 0 )
			panic( "No more IO ptes\n" ) ;
	}
	while ( -- nptes > 0 ) ;
	mtpr( 0, TBIA ) ;
	return( start_virt_addr ) ;
}

int
remove_pte( nptes )
	int     nptes ;
{
	ptes += nptes ;
	virt_addr -= (nptes * 1024) ;
	pte_ptr -= nptes ;
	page_no -= nptes ;
}

/*
 * The following procedures (get_ivct, return_ivct, init_ivct)
 * are used to find unused interrupt vectors.
 *
 * get_ivct returns an interrupt vector if there are 'need' vectors
 * in a row to allocate.  Start is the starting ivct, call with 0 the
 * first time.  If there are no ivcts available it returns (-1).
 */

#define MIN_IVCT	0x40
#define MAX_IVCT	0xff

int
get_ivct( start, need )
	int     	start ;
	int     	need ;
{
	extern  int     scb ;
	int *	   	scb_ptr = & scb ;
	int             index ;
	extern  int	Xstray ;
	extern  int	Xused ;

	need -- ;
	for ( start = MAX( start, MIN_IVCT ) ;
	      start <= MAX_IVCT - need ;
	      start ++ ) {
		if ( scb_ptr[ start ] == (int)& Xstray ) {
			for ( index = need + start ; index > start ; index -- ){
				if ( scb_ptr[ index ] != (int)& Xstray )
					break ;
			}
			if ( start == index ) {
				for ( index = start + need ;
				      index >= start ;
				      index -- )
					scb_ptr[ index ] = (int)& Xused ;
				return( start ) ;
			}
		}
	}
	printf( "get_ivct: could not allocate an interrupt vector\n" ) ;
	return( -1 ) ;

}

int
return_ivct( ivct )
	int     ivct ;
{
	int *   scb_ptr = & scb ;

	if ( (ivct < MIN_IVCT) || (ivct > MAX_IVCT) )
		return ;
	scb_ptr[ ivct ] = (int) & Xstray ;
}

int
init_ivct( ivct, proc_ptr )
	int     ivct ;
	int     proc_ptr ;
{
	int *   scb_ptr = & scb ;

	if ( (ivct < MIN_IVCT) || (ivct > MAX_IVCT) )
		return ;
	scb_ptr[ ivct ] = proc_ptr ;
}

#define	DMMIN	32
#define	DMMAX	4096
#define	DMTEXT	1024
#define	MAXDUMP	(8*1024)
/*
 * Configure swap space and related parameters.
 */
swapconf()
{
	register struct swdevt *swp;
	register int nblks;

	for (swp = swdevt; swp->sw_dev; swp++) {
		if (bdevsw[major(swp->sw_dev)].d_psize)
			nblks =
			  (*bdevsw[major(swp->sw_dev)].d_psize)(swp->sw_dev);
		if (swp->sw_nblks == 0 || swp->sw_nblks > nblks)
			swp->sw_nblks = nblks;
	}
	if (dumplo == 0 && bdevsw[major(dumpdev)].d_psize)
		dumplo = (*bdevsw[major(dumpdev)].d_psize)(dumpdev) - physmem;
	if (dumplo < 0)
		dumplo = 0;
	if (dmmin == 0)
		dmmin = DMMIN;
	if (dmmin < KLMAX*CLSIZE)
	        dmmin = KLMAX*CLSIZE;
	if (dmmax == 0)
		dmmax = DMMAX;
	if (dmtext == 0)
		dmtext = DMTEXT;
	if (dmtext > dmmax)
		dmtext = dmmax;
}
