/*	cons.c	4.10	81/05/09	*/
/*	HCX version Dec 1982		*/
/*	Minor device 0 is the CP itself.
/*	  No real read/writes can be done to him.
/*	Minor 1 is the console terminal.
/*	Minor 2 is the remote line trminal.
/**/

/* #define CNDEBUG 65536 */
#ifndef CNDEBUG
#define CNTRACE( a, b, c, d )
#else
int cntrace[ CNDEBUG ] ;
int cnindex = 0 ;
int cnsindex = 0 ;
int cnenable = 1 ;
#define CNTRACE( a, b, c, d ) \
	if ( cnenable ) { \
	        if ( cnindex + 4 >= CNDEBUG ) {\
	        	if ( cnsindex ) { \
	        		cnenable = 0 ; \
	        	} \
	                cnindex = 0 ;\
	        }\
	        cntrace[cnindex++] = (int)(a) ;\
	        cntrace[cnindex++] = (int)(b) ;\
	        cntrace[cnindex++] = (int)(c) ;\
	        cntrace[cnindex++] = (int)(d) ;\
	}\


#endif

/*
 * HCX console processor driver
 *
 */
#include "../sys/param.h"
#include "../sys/conf.h"
#include "../sys/user.h"
#include "../sys/tty.h"
#include "../sys/termio.h"
#include "../sys/uio.h"
#include "../sys/dk.h"
#include "../sys/callout.h"
#include "../sys/systm.h"
#include "../sys/sysinfo.h"
#include "../sys/reboot.h"
#include "../sys/msgbuf.h"
#include "../sys/kernel.h"
#include "../ml/cp.h"
#include "../ml/mtpr.h"

#define FALSE 0
#define TRUE  1

int	cnrestart() ;
int	timeout() ;

struct	tty cons[3];	/* One for each unit on CP */
struct	cons_info {
	char	last_one;	/* Last char sent - possibly repeat
				 * If zero, timeout has nothing to do
				 */
	int	active_timeout;	/* There is an active timeout for this line
				 * Set by 'cnputc' when a timeout is initiated.
				 * Reset by the routine  called at timeout.
			 	 */
	int	try_later;	/* If true, timeout has nothing to do.
				 * Set by 'cnputc' every time a char is sent.
				 * Reset by the timeout arrival. If next time
				 * the timeout comes it's zero, then it may
				 * have something to do.
				 */
	} cons_info[3];
struct	cpdcb_o consout[3] = { 
			/* 	unit,		cmd,count, buf */
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 }
			};
struct	cpdcb_i consstty[3] = {
			/* 	unit,		cmd,count, buf */
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 }
			};
struct	cpdcb_i consin[3] = {
			/* 	unit,		cmd,count, buf */
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 },
			{(char)(CPTAKE | CPDONE),0,   0 }
			};
struct	cphdr *lasthdr;

int	cnproc();
int	ttrstrt();
extern  char	partab[];
extern int rebooting ;

cnopen(dev, flag)
dev_t dev;
{
	register struct cpdcb_i *cin;
	register struct tty *tp;
	register int timo;

#ifdef lint
	flag = flag ;
#endif

	if (minor(dev) > CPREMOT) 
		return EEXIST;
	tp = &cons[minor(dev)];
	if (tp->t_state&XCLUDE && u.u_uid != 0)
		return EBUSY;
	if (lasthdr != (struct cphdr *)0) {
		timo = 10000;
		uncache((char *)&lasthdr->cp_unit);
		while ((lasthdr->cp_unit & CPTAKE)==0 && --timo )
			uncache((char *)&lasthdr->cp_unit);
	}	/* Here we give up waiting */
	cin = &consin[minor(dev)];
	cin->cp_hdr.cp_unit = (char)(minor(dev));
	cin->cp_hdr.cp_comm = (char)CPREAD;
	cin->cp_hdr.cp_count = 1;	/* Get ready for input */
	mtpr (cin, CPMDCB);
	lasthdr = (struct cphdr *)cin;
	tp->t_proc = cnproc;
	tp->t_dev = dev;
	if ((tp->t_state&ISOPEN) == 0) {
		ttinit(tp);
		if ( RB_BAUD(boot_device) != B0 ) {
		        tp->t_cflag =
		        	(tp->t_cflag & ~CBAUD) | RB_BAUD(boot_device);
		}
		tp->t_state |= CARR_ON;
		(*linesw[tp->t_line].l_open)(tp);
		tp->t_iflag |= ICRNL|IXON|IXANY|ISTRIP;
		tp->t_oflag |= OPOST|ONLCR|TAB3;
		tp->t_lflag |= ISIG|ICANON|ECHO|ECHOK;
		return(0);
	} else
		return((*linesw[tp->t_line].l_open)(tp));
}

cnclose(dev)
dev_t dev;
{
	register struct tty *tp = &cons[minor(dev)];

	return((*linesw[tp->t_line].l_close)(tp));
}

/*ARGSUSED*/
cnread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp = &cons[minor(dev)];

	return ((*linesw[tp->t_line].l_read)(tp, uio));
}

/*ARGSUSED*/
cnwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct tty *tp = &cons[minor(dev)];

	return ((*linesw[tp->t_line].l_write)(tp, uio));
}

/*
 * Got a console receive interrupt -
 * the console processor wants to give us a character.
 * Catch the character, and see who it goes to.
 */
cnrint(dev)
dev_t dev;
{
	register int c, timo;
	register struct tty *tp;
	register int ch;

	if (intenable == 0) return;
	sysinfo.rcvint++;
	/* make sure we dont take it from cache */
	uncache((char *)&consin[minor(dev)].cpi_buf[0]);
	c = consin[minor(dev)].cpi_buf[0];
	CNTRACE( 5, 0, c                                  , 0 ) ;
/*
/* Wait about 5 milli for last CPMDCB to be read by CP,
/* otherwise give up
/**/
	timo = 10000;
	uncache((char *)&lasthdr->cp_unit);
	while ((lasthdr->cp_unit & CPTAKE)==0  && --timo  )
		uncache((char *)&lasthdr->cp_unit);
	uncache((char *)&lasthdr->cp_unit);
	if (lasthdr->cp_unit & CPTAKE)
	{
		consin[minor(dev)].cp_hdr.cp_unit = (char)(minor(dev));
			/* This resets status bits */
		mtpr (&consin[minor(dev)], CPMDCB); /* Ready for new character */
		lasthdr = (struct cphdr *)&consin[minor(dev)];
		tp = &cons[minor(dev)];
		tk_rcvint++;
		ch = c & 0177;
		if (!(tp->t_state & ISOPEN))
			return;
		if (tp->t_rbuf.c_ptr == NULL)
			return;
		if (tp->t_iflag&IXON) {
			if (tp->t_state&TTSTOP) {
				if (ch == tp->t_cc[VSTART] || tp->t_iflag&IXANY)
					(*tp->t_proc)(tp, T_RESUME);
			} else {
				if (ch == tp->t_cc[VSTOP])
					(*tp->t_proc)(tp, T_SUSPEND);
			}
			if (ch == tp->t_cc[VSTART] || ch == tp->t_cc[VSTOP])
				return;
		}
		if (tp->t_iflag&ISTRIP)
			c &= 0177;
		else
			c &= 0377;
		*tp->t_rbuf.c_ptr = c;
		tp->t_rbuf.c_count--;
		(*linesw[tp->t_line].l_input)(tp, L_BUF);
	}
}

cnioctl(dev, cmd, addr, flag)
dev_t dev;
caddr_t addr;
{
	register struct tty *tp = &cons[minor(dev)];
	register error;
	int	setparms;

	error = ttiocom(tp, cmd, addr, flag, &setparms);
	if (error)
		return(error);
	if (setparms)
		cnparams(tp);
	return(error);
}

int cn_interrupt_pending = 0 ; /* handshake with printf */
/*
 * Got a console transmission interrupt -
 * the console processor wants another character.
 */
cnxint(dev)
dev_t dev;
{
	register struct tty *tp;
	register int line_no;

	if (intenable == 0) return;
	sysinfo.xmtint++;
	line_no = minor(dev);
	if ( line_no = CPCONS )
	        cn_interrupt_pending = 0 ;
	tp = &cons[line_no];
	tp->t_state &= ~BUSY;
	CNTRACE( 1, tp, cons_info[line_no].last_one & 0x7f , 0 ) ;
	cons_info[line_no].last_one = (char)0;
	cnproc(tp,T_OUTPUT);
}

cnproc(tp, cmd)
	register struct tty *tp;
{
	switch (cmd) {

	case T_TIME:
		tp->t_state &= ~TIMEOUT;
		goto start;

	case T_WFLUSH:
	case T_RESUME:
		tp->t_state &= ~TTSTOP;
		goto start;

	case T_OUTPUT:
	start:
		{
		register struct ccblock *tbuf;
		register c;
		register s;

		s = spl8();
		tbuf = &tp->t_tbuf;
		if (tp->t_state & (BUSY|TIMEOUT|TTSTOP))
			goto out;

		if ( msgbuff.msg_bufw != msgbuff.msg_bufx &&
			tp == &cons[CPCONS] ) {
		        cnputc( 0, 0 ) ;
		        goto out ;
		}

		if (tbuf->c_ptr == 0 || tbuf->c_count == 0) {
			if (tbuf->c_ptr)
				tbuf->c_ptr -= tbuf->c_size - tbuf->c_count;
			if (!(CPRES & (*linesw[tp->t_line].l_output)(tp)))
				goto out;
		}
		c = *tbuf->c_ptr++;
		if (c<=0177)
			cnputc ((c | (partab[c]&0200))&0xff,tp);
		else {
			timeout(ttrstrt, (caddr_t)tp, (c&0177));
			tp->t_state |= TIMEOUT;
			goto out;
		}
		tp->t_state |= BUSY;
		tbuf->c_count--;
	out:
		splx(s);
		}
		break;

	case T_SUSPEND:
		tp->t_state |= TTSTOP;
		break;

	case T_BLOCK:
		tp->t_state |= TBLOCK;
		break;

	case T_RFLUSH:
	case T_UNBLOCK:
		tp->t_state &= ~TBLOCK;
		break;
	}
}

/*
 * Print a character on console.
 */
cnputc(c,tp)
register char c;
register struct tty *tp;
{
	register timo , line_no;
	register struct cpdcb_o *current;

loop:
	/* tp == 0 only in system error messages */
	if (tp == 0) {
		if ( msgbuff.msg_bufw == msgbuff.msg_bufx ||
			( cn_interrupt_pending && !rebooting) ) {
			return ;
		}
		current = &consout[CPCONS];
		line_no = CPCONS;
		if(lasthdr == 0)	/* not done anythig yet */
			lasthdr = (struct cphdr *)current;
		c = msgbuff.msg_bufc[msgbuff.msg_bufw++];
		c |= partab[c&0177]&0200;
		if (msgbuff.msg_bufw >= MSG_BSIZE)
			msgbuff.msg_bufw = 0;
	}
	else {
		current = &consout[minor(tp->t_dev)];
		line_no = minor(tp->t_dev);
	}
   	CNTRACE( 2, tp, cons_info[line_no].last_one & 0x7f, c & 0x7f ) ;
	/*
	 * Try waiting for the console tty to come ready,
	 * otherwise give up after a reasonable time.
	 * make sure we dont test this bit in cache!
	 *
	 * If we are outputting reboot information, never give up.
	 * This allows an XOFF while outputting a panic sequence
	 * to have effect forever.
	 */
	timo = 30000;
	uncache((char *)&current->cp_hdr.cp_unit);
	while ((current->cp_hdr.cp_unit & CPDONE) == 0 &&
		( --timo || rebooting) )
		uncache((char *)&current->cp_hdr.cp_unit);

#ifdef CNDEBUG
	if ((current->cp_hdr.cp_unit & CPDONE) == 0 ) {
   		CNTRACE( 4, tp, cons_info[line_no].last_one & 0x7f, c & 0x7f ) ;
	        if ( cnenable )
	        	cnsindex = cnindex ;
	}
#endif

	current->cp_hdr.cp_comm = (char)CPWRITE;
	current->cp_hdr.cp_count = 1;
	current->cp_buf[0] = (char)(c & 0xff);
	timo = 10000;
	/*
	 * Try waiting for the console tty to come ready,
	 * otherwise give up after a reasonable time.
	 */
	uncache((char *)&lasthdr->cp_unit);
	while ((lasthdr->cp_unit & CPTAKE)==0 && --timo )
		uncache((char *)&lasthdr->cp_unit);
	/* Reset done bit */
	current->cp_hdr.cp_unit = (char)line_no; 
	lasthdr = (struct cphdr *)current;
	cons_info[line_no].last_one = c;
	if ( !cons_info[line_no].active_timeout && clk_enable) {
		cons_info[line_no].active_timeout = TRUE;
		timeout (cnrestart, (caddr_t)tp, 5*hz);
	}
	cons_info[line_no].try_later = TRUE; /* For timeout-means wait some more */
	mtpr (current, CPMDCB);
	if ( line_no == CPCONS ) {
		cn_interrupt_pending = intenable ;
		if ( tp == 0 ) {
			goto loop ;
		}
	}

}

/*
 * Restart (if necessary) transfer to CP line.
 * This way, lost 'transmit' interrupts don't break the chain.
 */
cnrestart (tp)
struct tty *tp;
{
	register line_no;

	if (tp == 0) {
		line_no = CPCONS;
	} else
		line_no = minor(tp->t_dev);
	if (cons_info[line_no].try_later) {
		cons_info[line_no].try_later = FALSE;
		timeout (cnrestart, (caddr_t)tp, 5*hz);
	}
	else {
		cons_info[line_no].active_timeout = FALSE;
		if (cons_info[line_no].last_one != (char)0) {
		        if ( line_no == CPCONS ) {
		                cn_interrupt_pending = 0 ;
		        }
		        /* Never pass tp == 0 on a restart */
			cnputc (cons_info[line_no].last_one, &cons[line_no]);
		}
	}
}

/*
 * Set line parameters
 */
cnparams(tp)
register struct tty *tp;
{
	register timo ;
	register struct cpdcb_i *current;
	register struct cpdcb_i *cin;

	current = &consstty[minor(tp->t_dev)];
	timo = 30000;
	/*
	 * Try waiting for the console tty to come ready,
	 * otherwise give up after a reasonable time.
	 * make sure we dont test this bit in cache!
	 */
	uncache((char *)&current->cp_hdr.cp_unit);
	while ((current->cp_hdr.cp_unit & CPDONE) == 0 && --timo )
		uncache((char *)&current->cp_hdr.cp_unit);
	current->cp_hdr.cp_comm = (char)CPSTTY;
	current->cp_hdr.cp_count = 4;
	current->cpi_buf[0] = tp->t_cflag&CBAUD;
	/* the rest are defaults */
	current->cpi_buf[1] = 0;	/* no parity */
	current->cpi_buf[2] = 0; /* stop bits */
	current->cpi_buf[3] = 8; /* data bits */
	timo = 10000;
	/*
	 * Try waiting for the console tty to come ready,
	 * otherwise give up after a reasonable time.
	 */
	uncache((char *)&lasthdr->cp_unit);
	while ((lasthdr->cp_unit & CPTAKE)==0 && --timo )
		uncache((char *)&lasthdr->cp_unit);
	/* Reset done bit */
	current->cp_hdr.cp_unit = (char)minor(tp->t_dev); 
	lasthdr = (struct cphdr *)current;
	mtpr (current, CPMDCB);

	timo = 10000;
	uncache((char *)&lasthdr->cp_unit);
	while ((lasthdr->cp_unit & CPTAKE)==0 && --timo )
		uncache((char *)&lasthdr->cp_unit);
	cin = &consin[minor(tp->t_dev)];
	cin->cp_hdr.cp_unit = (char)(minor(tp->t_dev));
	cin->cp_hdr.cp_comm = (char)CPREAD;
	cin->cp_hdr.cp_count = 1;	/* Get ready for input */
	mtpr (cin, CPMDCB);
	lasthdr = (struct cphdr *)cin;

}
