/*	trap.c	4.10	84/02/09	*/

#include "../ml/psl.h"
#include "../ml/reg.h"
#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/seg.h"
#include "../ml/trap.h"
#include "../sys/acct.h"
#include "../sys/kernel.h"
#include "../sys/sysinfo.h"
#include "../ml/mtpr.h"

#define	USER	040		/* user-mode flag added to type */

struct	sysent	sysent[];
int nsysent;

/*
 * Called from the trap handler when a processor trap occurs.
 */
/*ARGSUSED*/
trap(sp, type, hfs, accmst, acclst, dbl, code, pc, psl)
unsigned code;
{
	/* Next 2 dummy variables MUST BE the first local */
	/* variables; leaving place for registers 0 and 1 */
	/* which are not preserved by the 'cct' */

	int	dumm1;		/* register 1 */
	int	dumm0;		/* register 0 */

	register int *locr0 = ((int *)&psl)-PS;
	register int i;
	register struct proc *p;
	register char *typename;
#ifndef HCXRT
	struct	timeval syst;
#endif

#ifdef lint
	dumm1 = 0 ;
	dumm0 = 0 ;
	dumm1 = dumm1 ;
	dumm0 = dumm0 ;
#endif lint

#ifndef HCXRT
	syst = u.u_ru.ru_stime;
#endif

	if (type == T_ASTFLT) {
		astoff();
		u.u_ar0 = locr0;
		goto out;
	}

	if (USERMODE(locr0[PS])) {
		type |= USER;
		u.u_ar0 = locr0;
	}
	switch (type) {

	default: switch (type) {
		case T_RESADFLT:
			typename = "reserved addressing mode";break;
		case T_PRIVINFLT:
			typename = "illegal opcode";break;
		case T_RESOPFLT:
			typename = "reserved operand";break;
		case T_BPTFLT:
			typename = "breakpoint";break;
		case T_SYSCALL:
			typename = "kernel call";break;
		case T_ARITHTRAP:
			typename = "arithmetic exception";break;
#ifdef notdef /* handled before switch statement */
		case T_ASTFLT:
			typename = "system forced exception";break;
#endif
		case T_SEGFLT:
			typename = "limit fault";break;
		case T_PROTFLT:
			typename = "illegal access type";break;
		case T_TRCTRAP:
			typename = "trace trap";break;
#ifdef notdef /* handled in outer switch */
		case T_PAGEFLT:
			typename = "page fault";break;
#endif
		case T_TABLEFLT:
		case T_TABLEFLT + USER:
			typename = "page table fault";break;
		case T_ALIGNFLT:
			typename = "alignment fault";break;
		case T_KSPNOTVAL:
			typename = "kernel stack not valid";break;
		default:
			typename = "unknown trap";break;
		}
		printf("System trap (%s), code = %x, pc = %x\n", 
				typename, code, pc);
		panic("trap");

	case T_PROTFLT + USER:	/* protection fault */
		i = SIGBUS;
		break;

	case T_PRIVINFLT + USER:	/* privileged instruction fault */
	case T_RESADFLT + USER:	/* reserved addressing fault */
	case T_RESOPFLT + USER:	/* resereved operand fault */
	case T_ALIGNFLT + USER:	/* unaligned data fault */
		u.u_code = type &~ USER;
		i = SIGILL;
		break;

	case T_ARITHTRAP + USER:
		u.u_code = code;
		i = SIGFPE;
		break;

	/*
	 * If the user SP is above the stack segment,
	 * grow the stack automatically.
	 */
	case T_SEGFLT + USER:
		if (grow((unsigned)mfpr(USP)) || grow(code))
			goto out;
		i = SIGSEGV;
		break;

	case T_PAGEFLT:		/* allow page faults in kernel mode */
	case T_PAGEFLT + USER:	/* page fault */
		i = u.u_error;
		pagein(code, 0);
		u.u_error = i;
		if (type == T_PAGEFLT)
			return;
		goto out;

	case T_BPTFLT + USER:	/* bpt instruction fault */
	case T_TRCTRAP + USER:	/* trace trap */
		locr0[PS] &= ~PSL_T;
		i = SIGTRAP;
		break;

	case T_KSPNOTVAL + USER:
		panic("trap: KSP not valid");

	}
	psignal(u.u_procp, i);
out:
	p = u.u_procp;
	if (p->p_cursig || ISSIG(p))
		psig();
#ifndef HCXRT
	p->p_prior = p->p_usrpri;
#endif
	if (runrun) {
		/*
		 * Since we are u.u_procp, clock will normally just change
		 * our priority without moving us from one queue to another
		 * (since the running process is not on a queue.)
		 * If that happened after we setrq ourselves but before we
		 * swtch()'ed, we might not be on the queue indicated by
		 * our priority.
		 */
		u.u_ru.ru_nivcsw++;
		setrq_and_swtch();
	}
#ifndef HCXRT
	if (u.u_prof.pr_scale) {
		int ticks;
		struct timeval *tv = &u.u_ru.ru_stime;

		ticks = ((tv->tv_sec - syst.tv_sec) * hz) +
			((tv->tv_usec - syst.tv_usec) / tick) ;
		if (ticks)
			addupc(locr0[PC], &u.u_prof, ticks);
	}
#endif
}

#ifdef SYSCALLTRACE
int syscalltrace = 0;
#endif

/*
 * Called from the trap handler when a system call occurs
 */
/*ARGSUSED*/
syscall(sp, type, hfs, accmst, acclst, dbl, code, pc, psl)
register unsigned code;	/* r12 */
{
	/* Next 2 dummy variables MUST BE the first local */
	/* variables; leaving place for registers 0 and 1 */
	/* which are not preserved by the 'cct' */

	int	dumm1;		/* register 1 */
	int	dumm0;		/* register 0 */

	register caddr_t params;		/* known to be r11 below */
	register int i;				/* known to be r10 below */
	register int *locr0 = ((int *)&psl)-PS;
	register struct sysent *callp;
	register struct proc *p;
#ifndef HCXRT
	struct	timeval syst;
#endif

#ifdef lint
	dumm1 = 0 ;
	dumm1 = dumm1 ;
	dumm0 = 0 ;
	dumm0 = dumm0 ;
#endif lint

#ifndef HCXRT
	syst = u.u_ru.ru_stime;
	sysinfo.syscall++;
#endif

	u.u_ar0 = locr0;

	if (u.u_madesyscalls == 0) {
		u.u_madesyscalls = 1;
	        u.u_ap = u.u_arg;
		if (setjmp(&u.u_qsave)) {
			/* got here not necessarily during first system call */
			if (u.u_error == 0)
				u.u_error = EINTR;
	        	u.u_ap = u.u_arg;
			goto retsyscall;
		}
	}

	u.u_error = 0;

#ifdef HCXRT
	callp = &sysent[code];
#else
	callp = (code >= nsysent) ? &sysent[0] /* nosys */ : &sysent[code];
#endif
	if (i = callp->sy_narg) {
		i *= sizeof (int);
		params = (caddr_t)locr0[FP] + NBPW;
#ifndef HCXRT
		asm("prober $1,(r11),$1");		/* GROT */
		asm("beqlu cantread");			/* GROT */
		asm("prober $1,-1(r11)[r10],$1");	/* GROT */
		asm("bnequ ok");			/* GROT */
asm("cantread:");					/* GROT */
		u.u_error = EFAULT;			/* GROT */
		locr0[R0] = u.u_error;			/* GROT */
		locr0[PS] |= PSL_C;	/* carry bit */	/* GROT */
		goto done;				/* GROT */
asm("ok:");						/* GROT */
#ifndef lint
label_to_get_rid_of_warning:
#endif lint
/*NOTREACHED*/
#endif
		switch ( i ) {		/* typical case: 1-3 parameters */
		case 12 :
		        u.u_arg[2] = ((int*)params)[2] ;
		        /* fall through */
		case 8 :
		        u.u_arg[1] = ((int*)params)[1] ;
		        /* fall through */
		case 4 :
		        u.u_arg[0] = ((int*)params)[0] ;
		        break ;
		default:
			switch ( i ) {	/* non-typical case: 4-8 parameters */
			case 32 :
				u.u_arg[7] = ((int*)params)[7] ;
				/* fall through */
			case 28 :
				u.u_arg[6] = ((int*)params)[6] ;
				/* fall through */
			case 24 :
				u.u_arg[5] = ((int*)params)[5] ;
				/* fall through */
			case 20 :
				u.u_arg[4] = ((int*)params)[4] ;
				/* fall through */
			case 16 :
				u.u_arg[3] = ((int*)params)[3] ;
				u.u_arg[2] = ((int*)params)[2] ;
				u.u_arg[1] = ((int*)params)[1] ;
				u.u_arg[0] = ((int*)params)[0] ;
				break;
			}
		}
	}

	u.u_r.r_val1 = 0;

#ifdef SYSCALLTRACE
	if (syscalltrace) {
		char *cp;

		printf("0x%x", code);
		cp = "(";
		for (i= 0; i < callp->sy_narg; i++) {
			printf("%s%x", cp, u.u_ap[i]);
			cp = ", ";
		}
		if (i)
			putchar(')', 0);
		putchar('\n', 0);
	}
#endif

	(*(callp->sy_call))();

retsyscall:
	if (u.u_error) {
		if (u.u_error == ERESTARTSYS) {
			/*------ DIRTY CODE !!!!!!!!!---------*/
			/* reconstruct pc, assuming code is an immed constant */
			pc -= 2;		/* short literal */
			if (code > 0x3f) {
				pc--;	/* byte immed */
				if (code > 0x7f) {
					pc--;	/* word immed */
					if (code > 0x7fff)
						pc -= 2;	/* long immed */
				}
			}
			/*------------------------------------*/
		} else if (u.u_error != EJUSTRETURN) {
			locr0[R0] = u.u_error;
			locr0[PS] |= PSL_C;	/* carry bit */
		}
	} else {
		locr0[PS] &= ~PSL_C;	/* clear carry bit */
		locr0[R0] = u.u_r.r_val1;
	}
done:
	p = u.u_procp;
	if (p->p_cursig || ISSIG(p))
		psig();
#ifndef HCXRT	
	p->p_prior = p->p_usrpri;
#endif
	if (runrun) {
		/*
		 * Since we are u.u_procp, clock will normally just change
		 * our priority without moving us from one queue to another
		 * (since the running process is not on a queue.)
		 * If that happened after we setrq ourselves but before we
		 * swtch()'ed, we might not be on the queue indicated by
		 * our priority.
		 */
		u.u_ru.ru_nivcsw++;
		astoff();
		setrq_and_swtch();
	}

#ifndef HCXRT
	if (u.u_prof.pr_scale) {

		int ticks;
		struct timeval *tv = &u.u_ru.ru_stime;

		ticks = ((tv->tv_sec - syst.tv_sec) * hz) +
			((tv->tv_usec - syst.tv_usec) / tick) ;
		if (ticks)
			addupc(locr0[PC], &u.u_prof, ticks);
	}
#endif
}

/*
 * nonexistent system call-- signal process (may want to handle it)
 * flag error if process won't see signal immediately
 * Q: should we do that all the time ??
 */
nosys()
{
	if (u.u_signal[SIGSYS] == SIG_IGN || u.u_signal[SIGSYS] == SIG_HOLD)
		u.u_error = EINVAL;
	psignal(u.u_procp, SIGSYS);
}

/*
 * Ignored system call
 */
nullsys()
{
}

/*
 * Null system calls. Not invalid, just not configured.
 */
errsys()
{
	u.u_error = EINVAL;
}
