/*	vm_machdep.c	6.1	83/07/29	*/

#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/cmap.h"
#include "../sys/vm.h"
#include "../sys/text.h"
#include "../sys/vfs.h"
#include "../ufs/mount.h"
#include "../ml/mtpr.h"

/*
 * Set a red zone in the kernel stack after the u. area.
 */
setredzone(pte, vaddr)
	register struct pte *pte;
	caddr_t vaddr;
{

	pte += (sizeof (struct user) + NBPG - 1) / NBPG;
	*(int *)pte &= ~PG_PROT;
	*(int *)pte |= PG_URKR;
	if (vaddr)
		mtpr(vaddr + sizeof (struct user) + NBPG - 1, TBIS);
}

#ifndef mapin
mapin(pte, v, pfnum, count, prot)
	struct pte *pte;
	u_int v, pfnum;
	int count, prot;
{

	while (count > 0) {
		*(int *)pte++ = pfnum | prot;
		mtpr(ptob(v), TBIS);
		v++;
		pfnum++;
		count--;
	}
}
#endif

#ifdef notdef
/*ARGSUSED*/
mapout(pte, size)
	register struct pte *pte;
	int size;
{

	panic("mapout");
}
#endif

/*
 * Check for valid program size
 */
chksize(ts, ds, ss)
	register unsigned ts, ds, ss;
{
	static int maxdmap = 0;

	if (ts > MAXTSIZ || ds > MAXDSIZ || ss > MAXSSIZ) {
		u.u_error = ENOMEM;
		return (1);
	}
	/* check for swap map overflow */
	if (maxdmap == 0) {
		register int i, blk;

		blk = dmmin;
		for (i = 0; i < NDMAP; i++) {
			maxdmap += blk;
			if (blk < dmmax)
				blk *= 2;
		}
	}
	if (ctod(ts) > NXDAD * dmtext ||
	    ctod(ds) > maxdmap || ctod(ss) > maxdmap) {
		u.u_error = ENOMEM;
		return (1);
	}
	/*
	 * Make sure the process isn't bigger than our
	 * virtual memory limit.
	 *
	 * THERE SHOULD BE A CONSTANT FOR THIS.
	 */
	if (ts + ds + ss + LOWPAGES + HIGHPAGES > btoc(USRSTACK)) {
		u.u_error = ENOMEM;
		return (1);
	}
	return (0);
}

#ifdef notdef
/*ARGSUSED*/
newptes(pte, v, size)
	struct pte *pte;
	u_int v;
	register int size;
{
	register caddr_t a = ptob(v);

#ifdef lint
	pte = pte;
#endif
	if (size >= 30) {
		mtpr(0, TBIA);
		return;
	}
	while (size > 0) {
		mtpr(a, TBIS);
		a += NBPG;
		size--;
	}
}
#endif

/*
 * Change protection codes of text segment.
 * Have to flush translation buffer since this
 * affect virtual memory mapping of current process.
 */
chgprot(addr, tprot)
	caddr_t addr;
	long tprot;
{
	unsigned v;
	int tp;
	register struct pte *pte;
	register struct cmap *c;
	int s ;

	v = clbase(btop(addr));
	if (!isatsv(u.u_procp, v)) {
		u.u_error = EFAULT;
		return (0);
	}
	tp = vtotp(u.u_procp, v);
	pte = tptopte(u.u_procp, tp);
	if (pte->pg_fod == 0 && pte->pg_pfnum) {
		c = &cmap[pgtocm(pte->pg_pfnum)];
		if (c->c_blkno && c->c_vp != swapdev_vp) {
		        s = splimp() ;
			munhash(c->c_vp, (daddr_t)(u_long)c->c_blkno);
			splx(s) ;
		}
	}
	*(int *)pte &= ~PG_PROT;
	*(int *)pte |= tprot;
	distcl(pte);
	tbiscl(v);
	return (1);
}

settprot(tprot)
	long tprot;
{
	register int *ptaddr, i;

	ptaddr = (int *)mfpr(P0BR);
	for (i = 0; i < u.u_tsize; i++) {
		ptaddr[i] &= ~PG_PROT;
		ptaddr[i] |= tprot;
	}
	mtpr(0, TBIA);
}

/*
 * Rest are machine-dependent
 */

getmemc(addr)
	caddr_t addr;
{
	register int c;
	struct pte savemap;

	savemap = mmap[0];
	*(int *)mmap = PG_V | PG_KR | btop(addr);
	mtpr(vmmap, TBIS);
	uncache (&vmmap[(int)addr & PGOFSET]);
	c = *(char *)&vmmap[(int)addr & PGOFSET];
	mmap[0] = savemap;
	mtpr(vmmap, TBIS);
	return (c & 0377);
}

#ifdef notdef
putmemc(addr, val)
	caddr_t addr;
{
	struct pte savemap;

	savemap = mmap[0];
	*(int *)mmap = PG_V | PG_KW | btop(addr);
	mtpr(vmmap, TBIS);
	*(char *)&vmmap[(int)addr & PGOFSET] = val;

	mtpr (0, PADC);
	mtpr (0, PACC);

	mmap[0] = savemap;
	mtpr(vmmap, TBIS);
}
#endif

/*
 * Move pages from one kernel virtual address to another.
 * Both addresses are assumed to reside in the Sysmap,
 * and size must be a multiple of CLSIZE.
 */
pagemove(from, to, size)
	register caddr_t from, to;
	register int size;
{
	register struct pte *fpte, *tpte;

	if (size % CLBYTES)
		panic("pagemove");
	fpte = &Sysmap[btop(from - 0xC0000000)];
	tpte = &Sysmap[btop(to - 0xC0000000)];
	while (size > 0) {
		*tpte++ = *fpte;
		*(int *)fpte++ = 0;
		mtpr(from, TBIS);
		mtpr(to, TBIS);
		mtpr(to, P1DC);		/* purge !! */
		from += NBPG;
		to += NBPG;
		size -= NBPG;
	}
}

/* General (includes system) virtual address to physical */
vtoph(p, v)
register struct proc *p;
register unsigned v;
{
	register struct pte *thispte;
	register unsigned vp = btop(v);

	thispte = vtopte(p, vp);
	return ( (thispte->pg_pfnum << PGSHIFT) + (v & PGOFSET));
}


/*
 * Some code and data key management routines.
 * The arrays ckey_cnt and ckey_cache are allways kept in such a way
 * 	that the following invariant holds:
 *	(ckey_cnt > 0) ==> (ckey_cache == 1)
 * meaning as long as a code key is used by at least one process, it's
 * marked as being 'in the cache'. Of course, the following invariant
 * also holds:
 *	(ckey_cache==0) ==> (ckey_cnt==0)
 * which is just the reciprocal of the 1'st invariant.
 */

struct dkey_info {
	struct dkey_info *next;	/* next in empty or dirty list */
} dkey_info[NOFDKEY], *empty_dkey, *dirty_dkey;

/* 
 * ckeyrelease -- release a code key.
 */
ckeyrelease (p)
struct proc *p;
{
	register int key;

	key = p->p_ckey;
	p->p_ckey = 0;  /* so we don't try and use it in getcodekey */

	if (--ckey_cnt[key] < 0) {
		printf ("ckeyrelease: key = %d\n", key);
		ckey_cnt[key] = 0;
	}
}

/* 
 * dkey_init -- initialize data structures for data key management.
 * Key 0 is not used.  Key MAXDKEY is reserved for process 0.
 */
dkey_init ()
{

	register int dk;

	for (dk = 1; dk < MAXDKEY-1; dk++)
		dkey_info[dk].next = &dkey_info[dk+1];
	dkey_info[MAXDKEY-1].next = 0;
	empty_dkey = &dkey_info[1];
	dirty_dkey = 0;
}

/* 
 * dkeyrelease -- release a data key.
 */
dkeyrelease (p)
struct proc *p;
{
	register int key;

	key = p->p_dkey;
	p->p_dkey = 0;  /* so we don't use it in getdatakey */

	dkey_info[key].next = dirty_dkey;
	dirty_dkey = &dkey_info[key];
}

/* 
 * getcodekey -- get a code key.
 */
getcodekey()
{
	register int i,
		allocated,	/* number of non-zero ckey_cnt's */
		return_key,
		usable_key;
	register struct proc *p;

	allocated = 0;

	for (i = 1; i <= MAXCKEY; i++) {
		if ((int) ckey_cache[i] == 0) {		/* Bingo */
			ckey_cache[i] = 1;
			ckey_cnt[i] = 1;
			return (i);
		}

		if (ckey_cnt[i] != 0)
			allocated++;
	}

	/*
	 * If we are here, all code keys were marked as being in cache.
	 *
	 * Two cases: some of them are free for re-allocation, or all of
	 * them are currently allocated. In this (second) case, a more
	 * drastic procedure will follow - i.e. we strip some processes of
	 * their keys and let them get new ones whenever they need it.
	 */
	if (allocated < MAXCKEY) {
		/*
		 * This is the easy case.
		 */
		for (i = 1; i <= MAXCKEY; i++) {
			if (ckey_cnt[i] == 0) {
				ckey_cache[i] = 0;
				return_key = i;
			}
		}

		ckey_cnt[return_key] = 1;
		ckey_cache[return_key] = 1;
		mtpr (0, PACC);
		return (return_key);
	}

	/*
	 * Now we have to get nasty.
	 * Strip some of them of the code key. First time,
	 * 1) Try hard not to do that to kernel processes !!
	 * 2) Try hard NOT to strip shared text processes of
	 *    their (shared) key, because then they'll run
	 *    with different keys from now on, i.e. less efficient
	 *    cache utilization.
	 */
	for (p = proc; p < procNPROC; p++) {
		/*
		 * Look for a meaningful key but not used and not shared
		 * text. Also we can not take the key of a running process
		 * or a child process in the middle of a fork or vfork (see
		 * 'getdatakey' code for details).
		 */
		if (p->p_ckey && p->p_ckey!=MAXCKEY && ckey_cnt[p->p_ckey]<2 &&
		    p->p_stat!=SRUN && p->p_stat!=SIDL) {
			i = p->p_ckey;
			p->p_ckey = 0;
			ckey_cnt[i] = 1;
			ckey_cache[i] = 1;
			mtpr (0, PACC);
			return (i);
		}
	}

	/*
	 * Second time around!
	 * Highly unlikely situation. It means that all keys are
	 * allocated AND shared (i.e. we have at least 510 active
	 * processes).
	 * Strip some of them. We pick some key (known to be shared
	 * by several processes) and strip the poor process group.
	 * At least 2 processes will loose but we gain one key to be reused.
	 * The way this code is written virtually assures us that this
	 * key isn't the 'init' group key (1) nor the 'shell' group key
	 * (2 or 3). It's probably something like 254.
	 */
	for (i = MAXCKEY; (i >= 1) && (ckey_cnt[i] > 1); i --)  {
		usable_key = 1;
		for (p = proc; p < procNPROC; p++) {
			if (p->p_ckey==i) {
				if ((p->p_stat==SRUN) || (p->p_stat==SIDL)) {
					usable_key = 0;
				}
			}
		}
		if (usable_key) {
			for (p = proc; p < procNPROC; p++) {
				if (p->p_ckey==i) {
					p->p_ckey = 0;
					ckey_cnt[i]--;
				}
			}
			if (ckey_cnt[i] != 0)
				printf("getcodekey: key = %d cnt = %d\n",
					i, ckey_cnt[i]);
			ckey_cnt[i] = 1;
			ckey_cache[i] = 1;
			mtpr (0, PACC);
			return (i);
		}
	}

	panic ("getcodekey");
/*NOTREACHED*/
}

/* 
 * getdatakey -- get a data key.
 *
 * General strategy:
 * 1) Try to find a data key that isn't in the cache. Allocate it.
 * 2) If all data keys are in the cache, find one which isn't
 *    allocated. Clear all status and allocate this one.
 * 3) If all of them are allocated, strip the keys of many processes,
 *    and use one of the stripped keys.  We strip several processes at
 *    one time so that in the future conditions 1) and 2) are much more
 *    likely to occur.
 */
getdatakey()
{
	register int i;
	register struct proc *p;

	if (empty_dkey) {
		/*
		 * Case 1. The best case.
		 */
		i = empty_dkey - &dkey_info[0];
		empty_dkey = empty_dkey->next;
		return (i);
	}

	if (dirty_dkey) {
		/*
		 * Case 2. This is the easy case.
		 */
		empty_dkey = dirty_dkey;
		dirty_dkey = 0;
		i = empty_dkey - &dkey_info[0];
		empty_dkey = empty_dkey->next;
		mtpr (0, PADC);
		return (i);
	}
		
	/*
	 * Now, we have to take a key from someone.
	 * We should first try to not take the key from a running process
	 * or from a child process in the middle of a fork or vfork
	 * (because of the longjmp code in the context switcher).
	 */
	for (p = allproc; p != NULL; p = p->p_nxt) {
		if ((p->p_dkey!=0) && (p->p_stat!=SRUN) && (p->p_stat!=SIDL)) {
			i = p->p_dkey;
			p->p_dkey = 0;
			dkey_info[i].next = empty_dkey;
			empty_dkey = &dkey_info[i];
		}
	}
	if (empty_dkey) {
		mtpr (0, PADC);
		i = empty_dkey - &dkey_info[0];
		empty_dkey = empty_dkey->next;
		return (i);
	}
	panic ("getdatakey");
/*NOTREACHED*/
}
