#include "../sys/param.h"
#include "../sys/ioctl.h"
#include "../sys/tty.h"
#include "../sys/errno.h"
#include "../sys/uio.h"
#include "../sys/conf.h"
#include "../sys/file.h"
#include "../sys/mp_host.h"
#include "../sys/mp_dl.h"
#include "../sys/mp_ahst.h"
#include "../sys/mp_adef.h"
#include "../sys/pool.h"

extern struct tty mpa_16tty[];
extern struct tty mpa_32tty[];
extern struct vbmem vbmpcc;
extern struct clist xclist[];
extern ttrstrt();
extern short nolevel1;

#ifdef DEBUG
extern int mpatrace;
#endif
extern int mpahis;

mpaincmp(board,list) /*# Inbound Completion Routine */
short board;	     /*# MPCC Board Number */ 
register char *list; /*# List of Ports With Inbounds Complete */
/*
*#	This routine is called by the MPCC interrupt handler with a 
*#	list of ASYNC ports which had previously sent Host events to
*#	the MPCC, and which the MPCC is now acknowledging as done.
*#
*#	Process:
*#
*/
{
	register MPCCPRT *mp;
	register MPCCINENT *inadr;
	register MPCCOUTENT *out;
	register struct tty *tp;
	register struct cblock *cbp;
	register char *ptr;
	register ALLPRAM *ap;
	register short *size;
	short prtidx,count,port,i,s,j;

	j=0;
#ifdef DEBUG
	mplog(1);
#endif

	/*# while there are ports in the list with */
	/*# completed inbound events to process    */
	while((port=list[j++]) != (short)EOL) {
#ifdef DEBUG
		mplog(2);
		mplog(board);
		mplog(port);
		if(mpatrace)
			printf("incmp %d ",port);
#endif
		if(j > MAXASYPRT)
			break;
		/*# set up ptr to Host port structure */
		mp = &vbmpcc.mbloks[board]->mpccport[port];

		prtidx = (board*MAXASYPRT)+port;
		/*# set up tty ptr */
		if(port < MAXASYPRT/2)
			tp = &mpa_16tty[(board*MAXASYPRT/2) + port];
		else
			tp = &mpa_32tty[(board*MAXASYPRT/2) + (port-(MAXASYPRT/2))];
		
		/*# Get pointer to next done MPCC Inbound Queue Entry */
		inadr = &mp->Inents[mp->off];

		/*# For all entries that are done */
		while(inadr->go_stat & MP_DONE) {
			/*# set up pointer to inbound entry params */
			ap = (ALLPRAM *)inadr->un.async.params;
			mppurge(ap,sizeof(ALLPRAM));
			/*# switch on completed event type */
			switch(inadr->op_code) {
			case MP_OPEN: 	/*# OPEN completion */
#ifdef DEBUG
				mplog(3);
				if(mpatrace)
					printf("OPEN ");
#endif
				/*# if this is port was closed  */
				/*# on the host due to a signal */
				if (mp->protocol != MP_ASYNC) 
					/*# break out of switch */
					break;
				/*# make MPCC Outbound Queue  */
				/*# entries available for use */
				for(i=0;i<MP_OUTSET;i++) 
					mp->Outents[i].go_stat = MP_GO;

				if (!nolevel1) {
					/*# if carrier is present */
					if (ap->mdmctls.dcd == ASSERT) {
						/*# show it in tty state */
						tp->t_state |= CARR_ON;
						wakeup((caddr_t)&tp->t_canq);
					}
					/*# else */
					else
						/*# clear it in tty state */
						tp->t_state &= ~CARR_ON;
				}
				if(mpahis >= 0)
				    mpccrecord((long)tp->t_dev,(long)(MP_OPEN+100),(long)inadr,(long)tp->t_state);
				/*# free up inbound transmit entry */
				mpa_freein(inadr);

				/*# point to next done inbound entry */
				adjptr(mp->off,MP_INSET);

				/*# break */
				break;
			case MP_CLOSE: /*# CLOSE completion */
#ifdef DEBUG
				mplog(4);
				if(mpatrace)
					printf("CLOSE ");
#endif
				/*# make Outbound entrys unavailable     */
				for(i=0;i<MP_OUTSET;i++) {
					mp->Outents[i].go_stat = MP_FREE;
					ptr = mp->Outents[i].un.async.params;
					putcf(CMATCH((struct cblock *)ptr));
 					mp->Outents[i].un.async.un.rcvblk = (unsigned char *)0;
					mp->Outents[i].un.async.params = (char *)0;
 					pool_dealloc(PL_ASYPRT,mp->Inents[i].un.async.params);
					mp->Inents[i].un.async.params = (char *)0;
				}

				if(mpahis >= 0)
				    mpccrecord((long)tp->t_dev,(long)(MP_CLOSE+100),(long)inadr,(long)tp->t_state);
				/*# free up inbound transmit entry */
				mpa_freein(inadr);

				/*# point to next done inbound entry */
				adjptr(mp->off,MP_INSET);

				/*# clear carrier from tty state   */
				/*# so must be set up by next open */
				tp->t_state &= ~(CARR_ON | BUSY);
				
				/*# init mpcc host port structures for  */
				/*# next open - must be done before we  */
				/*# set protocol to unused for the mpcc */
				/*# interrupt routine			*/
				mp->on = mp->off = mp->nextrcv = 0;

				/*# show close no longer in progress */
				mp->flags &= ~MP_PROGRESS;

				/*# show mpcc port no longer open for async */
				mp->protocol = MP_UNUSED;

				/*# wake up proc waiting for close completion */
				/*# - kludge saying CARR_ON for remote tty    */
				wakeup((caddr_t)&tp->t_canq);

				/*# no more for this port */
				break;
			case MP_IOCTL: 	/*# IOCTL completion */
#ifdef DEBUG
				mplog(5);
				if(mpatrace)
					printf("IOCTL ");
#endif
				if(mpahis >= 0)
				    mpccrecord((long)tp->t_dev,(long)(MP_IOCTL+100),(long)inadr,(long)tp->t_state);
				/*# free up inbound transmit entry */
				mpa_freein(inadr);

				/*# point to next done inbound entry */
				adjptr(mp->off,MP_INSET);

				/*# break */
				break;
			case MP_WRITE: 	/*# WRITE (transmit) completion */
#ifdef DEBUG
				mplog(6);
				if(mpatrace)
					printf("WRITE ");
#endif
				/*# if this was a not a flush command */
				if(inadr->flags != (short)A_FLUSH) {
					/*# if next char to output is */
					/*# a delay char              */
					if (tp->t_state & TIMEOUT) 
						/*# set up timeout required */
						/*# by the delay char value */
						timeout(ttrstrt,(caddr_t)tp,
						((xclist[prtidx].c_cl->c_data[2]&0177)+6));
					/*# else */
					else 
						/*# return cblocks in transmit */
						/*# list to free cblock list   */
						while ((cbp = getcb(&xclist[prtidx])) != (struct cblock *)0)
							putcf(cbp);
				}
				if(mpahis >= 0)
				    mpccrecord((long)tp->t_dev,(long)(MP_WRITE+100),(long)inadr,(long)tp->t_state);
				/*# show no xmt in process */
				tp->t_state &= ~BUSY;

				/*# if an error occurred */
				switch(inadr->un.async.error) {

				case A_SIZERR:	/*# error in xmt data size */
					/*# print error msg on console */
					mpconsole(board,port,A_XSIZE,0);
					break;
				case A_NXBERR:	/*# no more xmt evt buffers */
					/*# print error msg on console */
					mpconsole(board,port,A_NOXBUF,0);
				}

				/*# free up inbound transmit entry */
				mpa_freein(inadr);

				/*# point to next done inbound entry */
				adjptr(mp->off,MP_INSET);

				/*# start up next transmit */
				mpaproc(tp,T_OUTPUT);

				/*# break */
				break;
			default:	/*# unknown event type */
				/*# print msg on console */
				mpconsole(board,port,A_INVCMD,inadr->op_code);	

				/*# free up inbound transmit entry */
				mpa_freein(inadr);

				/*# point to next done inbound entry */
				adjptr(mp->off,MP_INSET);
			}
			/*# get pointer to next done Inbound Entry */
			inadr = &mp->Inents[mp->off];
		}
loop:	/*# go to next port with completed inbound entries */
	;
	}
}
mpaoutcmp(board,list)	/*# Outbound Completion Routine */
short board;	        /*# MPCC board number */
register char *list;	/*# list of ports */
/*
*#	This routine is called from the MPCC interrupt handler with
*#	a list of ASYNC ports with events for the Host.  These events
*#	may either be receive data events or status events.
*#
*#	Process:
*#
*/
{
	register struct ccblock *rbuf;
	register struct tty *tp;
	register char *endptr,*dataptr;
	register MPCCOUTENT *out;
	register MPCCPRT *mp;
	unsigned char c;
	short cc,count,j,i, port,prtidx;
	char *rcverr;

	j = 0;
#ifdef DEBUG
	mplog(16);
#endif
	while((port=list[j++]) != EOL) {
#ifdef DEBUG
		mplog(17);
		mplog(board);
		mplog(port);
		if(mpatrace)
			printf("outcmp %d ",port);
#endif
		if(j > MAXASYPRT)
			break;
		prtidx = (board*MAXASYPRT)+port;
		if(port < MAXASYPRT/2)
			tp = &mpa_16tty[(board*MAXASYPRT/2) + port];
		else
			tp = &mpa_32tty[(board*MAXASYPRT/2) + (port-(MAXASYPRT/2))];
		mp = &vbmpcc.mbloks[board]->mpccport[port];
		/*# get pointer to MPCC Outbound Queue */
		out = &mp->Outents[mp->nextrcv];
		/*# for all entries that are done */
		while(out->go_stat & MP_DONE) {
			/*# switch on type of event */
			switch(out->op_code) {
			case MP_READ:	/*# Receive Event */
				/*# if this is not an open ASYNC port */
#ifdef DEBUG
				mplog(18);
				if(mpatrace)
					printf("READ ");
#endif
				if(mpahis >= 0)
					mpccrecord((long)tp->t_dev,(long)MP_READ,(long)(out->un.async.count|((long)out->un.async.error<<16)),(long)tp->t_state);

				if((mp->protocol != MP_ASYNC) || 
				   (((tp->t_state & ISOPEN) == 0) &&
				    (tp->t_state & WOPEN)== 0)) {
					/*# discard this receive data */
					/*# break */
					break;
				}
				/*# if no data chars */
				if((cc = count = out->un.async.count) == 0) {
					/*# break */
					break;
					}
				rbuf = &tp->t_rbuf;
				if (rbuf->c_ptr == NULL) {
					if ((rbuf->c_ptr = (getcf()->c_data))
					  == ((struct cblock *)NULL)->c_data)
						panic("mpaoutcmp:config more cblocks");
					rbuf->c_count = cfreelist.c_size;
					rbuf->c_size = cfreelist.c_size;
				} else {
					if (rbuf->c_count != rbuf->c_size) {
						printf("read c_count %d c_size %d\n",
					 	rbuf->c_count,rbuf->c_size);
						panic("mpaoutcmp:2");
					}
				}


				/*# point into receive data block */
				dataptr = out->un.async.params;
				mppurge(dataptr,CBSIZE);

				/* set up for next read */
				out->un.async.params = (char *)((struct cblock *)CMATCH((struct cblock *)rbuf->c_ptr))->c_data;
				out->un.async.un.rcvblk = (unsigned char *)vtoph(0, out->un.async.params);

				/* set up tty structure of port for read */
				rbuf->c_ptr = dataptr;
				rbuf->c_size = cfreelist.c_size;
				rbuf->c_count = rbuf->c_size - count;

				/*# if there is a break char embedded w/in data */
				if (out->un.async.error == BRKASRT) {
					/*# dec count to use as index */
					cc--;
					/*# init ptr to last rcv char */
					endptr = dataptr + cc;
					/*# get index to break char */
					while ((cc>=0) && (*endptr != 0x00)) {
						endptr--;
						cc--;
					}
					/*# send all chars up to break */
					/*# char individually to ttin  */
					if(cc>0) {
						rbuf->c_count = rbuf->c_size-cc;
						(*linesw[tp->t_line].l_input)									(tp, L_BUF);
						count -= cc;
					}
					/* send break char if found */
					if ((count > 0) &&
						(*(dataptr+cc) == 0x00)) {
						(*linesw[tp->t_line].l_input) 									(tp, L_BREAK);
						rbuf->c_count = rbuf->c_size;
						count--;
					}
					/* send any chars found after break */
					if (count > 0) {
						rbuf->c_ptr = dataptr + cc + 1;
						rbuf->c_size = cfreelist.c_size;
						rbuf->c_count = rbuf->c_size - 
									count;
						(*linesw[tp->t_line].l_input)									(tp, L_BUF);
					}
					break;
				}
				/* send all but last char */
				count--;
				if (count > 0) {
					rbuf->c_count += 1;
					(*linesw[tp->t_line].l_input)(tp,L_BUF);
					/* set up tty to send last char */
					rbuf->c_ptr = dataptr + count;
					rbuf->c_size = cfreelist.c_size;
					rbuf->c_count = rbuf->c_size - 1;
				}
						
				/*# switch on type of receive event */
				switch(out->un.async.error) {

				case RCVDTA:	/* Normal (good) rcv data */
					rcverr = (char *)0;
					break;
				case PARERR:	/* parity error */
					rcverr = "PARITY ERROR";
					break;
				case FRAMERR:	/* frame error */
					rcverr = "FRAME ERROR";
					break;
				case OVRNERR:	/* Overrun error */
					rcverr = "OVERRUN ERROR";
					break;
				case OVFERR:	/* Overflow error */
					rcverr = "OVERFLOW ERROR";
					break;
				default:
					rcverr = "UNDEFINED RCV ERROR";
				}
				if (rcverr != (char *)0)
					mpconsole(board,port,rcverr,0);
				(*linesw[tp->t_line].l_input)(tp,L_BUF);
				break;
			case MP_STATUS:	/*# Status Event */
#ifdef DEBUG
				mplog(19);
				if(mpatrace)
					printf("STATUS ");
#endif
				if(mpahis >= 0)
					mpccrecord((long)tp->t_dev,(long)MP_STATUS,(long)out->flags,(long)tp->t_state);
				/*# switch on type of status event */
				switch (out->flags) {
				case DCDASRT:	/*# carrier detect  asserted */
					/*# show carrier asserted in tty state */
					tp->t_state |= CARR_ON;

					/*# wakeup anyone waiting for carrier  */
					wakeup((caddr_t)&tp->t_canq);
					/*# break */
					break;
				case DCDDROP:	/*# carrier detect dropped */
					/*# show carrier dropped in tty state */
					tp->t_state &= ~CARR_ON;

					/*# wake up ttys for loss of carrier */
					wakeup((caddr_t)&tp->t_canq);

					/*# flush I/O queues and send hangup signals */
					ttyflush(tp,(FREAD|FWRITE));
					if (tp->t_state&ISOPEN) {
						gsignal(tp->t_pgrp, SIGHUP);
						gsignal(tp->t_pgrp, SIGCONT);
					}

					/*# break */
					break;
				case NORBUF:	/*# no more rcv buffers   */
				case NOEBUF:	/*# no more event buffers */
					/*# print msg on console */
					mpconsole(board,port,out->flags,0);
					/*# break */
					break;
			    /*# The following events have no affect on */
			    /*# on async operation, (and they should   */
			    /*# never be enabled), so they are ignored */ 
				case DSRASRT:	/*# data set ready or ring indicator asserted */
				case MPDSRDROP:	/*# data set ready or ring indicator dropped */
				case CTSASRT:	/*# clear to send asserted */
				case CTSDROP:	/*# clear to send dropped */
				case SRXASRT:	/*# secondary xmit asserted */
				case SRXDROP:	/*# secondary xmit dropped */
				case RATEASRT:	/*# rate signal asserted */
				case RATEDROP:	/*# rate signal dropped  */
					/*# break */
					break;
				default: /*# unknown status event type */
					/*# print error msg on console */
					mpconsole(board,port,A_INVSTS,(char)out->flags); 
				} 
				/*# break */
				break;
			default:	/*# unknown event type */
				/*# print error msg on console */
				mpconsole(board,port,A_INVCMD,out->op_code);	
			}
			/*# initialize entry for re-use */
			out->go_stat = MP_GO;
			out->op_code = 0;
			out->flags = 0;
			out->un.async.error   = 0;
			out->un.async.flags   = 0;
			/*# bump pointer to next done entry on Outbound Queue */
			adjptr(mp->nextrcv,MP_OUTSET);

			/*# get addr of next potentially done entry */
			out = &mp->Outents[mp->nextrcv];
		}
	}
}
mpconsole(board,port,msgtype,flags) /*# Print Msg To Console */
short board,port;	          /*# mpcc board & port #'s */
register char *msgtype;		  /*# error msg to print    */
char flags;
/*
*#	This routine will print an error message and an optional
*#	flags field to the console.
*#
*#	Process:
*#
*/
{
/*
	printf("MPCC: Board #%d Port #%d - %s",board,port,msgtype); */
	if(flags != 0)
		printf(" (%d)",flags); 
}
mpa_freein(inadr)	   /*# Free Up Inbound Entry */
register MPCCINENT *inadr; /*# addr of inbound entry */
/*
*#	This routine is called from within the mpcc async driver to
*#	reinitialize the fields in a certain MPCC Inbound Queue entry
*#	and flag it as being available for re-use.
*#
*#	Process:
*#
*/
{
	/*# re-init all values in this entry */
	inadr->op_code = 0;
	inadr->flags   = 0;
	inadr->un.async.error = 0;
	inadr->un.async.flags = 0;
	inadr->un.async.count = 0;
	/*# show this entry is available for use */
	inadr->go_stat = MP_FREE;
}
#ifdef A_OPTIMIZE 
qdata(cbp,data,count,flag)   /*# move rcv data to cblock */
register struct cblock *cbp; /*# cblock to move data into */
register char *data;	     /*# receive data */
short count;		     /*# count of receive chars */
short flag;		     /*# input control flags */
/*
*#	This routine is used to try and speed up the moving
*#	of chars to the raw input queue.  It moves them from
*#	the inbound entry to a cblock performing any required
*#	terminal control operations. The case statement allows
*#	the data to be processed as fast as possible since only
*#	the requested control operations are checked for each case.
*#	Also, it eliminates xon/xoff checks which are now being
*#	done on the mpcc board. Moving 'good' chars to a cblock
*#	will allow us to bypass much of the ttin code which checks
*#	each individual char before putting it on the rawq.
*#
*#	The possible control operations which concern us here are:
*#		
*#		ISTRIP 	- strip 8th data bit
*#		INLCR  	- map NL to CR 
*#		IGNCR	- ignore CR's
*#		ICRNL	- map CR to NL
*#		IUCLC	- map upper-case to lower-case
*#
*#	Process:
*#
*/
{
	register char *cbdata;
	char c;
	int i,x;

	cbdata = cbp->c_data;
	cbp->c_first = 0;

	/*# switch on input control processing flags */
	switch(flag & (ISTRIP|IUCLC|INLCR|IGNCR|ICRNL)) {

	case ISTRIP:	/*#*/
		for (i=0;i<count;i++)
			*cbdata++ = (*data++)&0177;
		cbp->c_last = --count;
		/*# break */
		break;
	case IUCLC:	/*#*/	
		for (i=0;i<count;i++) {
			c = *data++;
			if (('A' <= c) && (c <= 'Z'))
				c += 'a' - 'A';
			*cbdata++ = c;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	case IGNCR:	/*#*/
		for (x=0,i=0;i<count;i++) {
			if ((*cbdata = *data++) != '\r') {
				cbdata++;
				x++;
			}
		}
		cbp->c_last = --x;
		/*# break */
		break;
	case ICRNL:	/*#*/
		for (i=0;i<count;i++) {
			if ((*cbdata = *data++) == '\r')
				*cbdata++ = '\n';
			else
				cbdata++;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	case INLCR:	/*#*/
		for (i=0;i<count;i++)
			if ((*cbdata = *data++) == '\n')
				*cbdata++ = '\r';
			else
				cbdata++;
		cbp->c_last = --count;
		/*# break */
		break;
	case (ISTRIP|IUCLC):	/*#*/
		for (i=0;i<count;i++) {
			c = (*data++)&0177;
			if (('A' <= c) && (c <= 'Z'))
				c += 'a' - 'A';
			*cbdata++ = c;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	case (IUCLC|INLCR):	/*#*/
		for (i=0;i<count;i++) {
			if ((c = *data++) == '\n')
				c = '\r';
			else
				if (('A' <= c) && (c <= 'Z'))
					c += 'a' - 'A';
			*cbdata++ = c;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	case (INLCR|IGNCR):	/*#*/
		for (i=0,x=0;i<count;i++) {
			if ((c = *data++) != '\r') {
				if (c == '\n')
					c = '\r';
				*cbdata++ = c;
				x++;
			}
		}
		cbp->c_last = --x;
		/*# break */
		break;
	case (INLCR|ICRNL):	/*#*/
		for (i=0;i<count;i++) {
			if ((c = *data++) == '\n')
				*cbdata++ = '\r';
			else
				if (c == '\r')
					*cbdata++ = '\n';
				else
					*cbdata++ = c;
		}
		cbp->c_last = --x;
		/*# break */
		break;
	case (ISTRIP|IUCLC|INLCR):	/*#*/
		for (i=0;i<count;i++) {
			c = (*data++)&0177;
			if (c == '\n')
				c = '\r';
			else
				if (('A' <= c) && (c <= 'Z'))
					c += 'a'-'A';
			*cbdata++ = c;
		}
		cbp->c_last = --x;
		/*# break */
		break;
	case (IUCLC|INLCR|IGNCR):	/*#*/
		for (i=0,x=0;i<count;i++)
			if ((c = *data++) != '\r') {
				if (c == '\n')
					c = '\r';
				else
					if (('A' <= c) && (c <= 'Z'))
						c += 'a'-'A';
				*cbdata++ = c;
				x++;
			}
		cbp->c_last = --x;
		/*# break */
		break;
	case (IUCLC|INLCR|ICRNL):	/*#*/
		for (i=0;i<count;i++) {
			if ((c = *data++) == '\n')
				c = '\r';
			else
				if (c == '\r')
					c = '\n';
				else
					if (('A' <= c) && (c <= 'Z'))
						c += 'a'-'A';
			*cbdata++ = c;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	case (ISTRIP|IUCLC|INLCR|IGNCR):	/*#*/
		for (x=0,i=0;i<count;i++) {
			c = (*data++)&0177;
			if (c != '\r') {
				if (c == '\n')
					c = '\r';
				else
					if (('A' <= c) && (c <= 'Z'))
						c += 'a'-'A';
				*cbdata++ = c;
				x++;
			}
		}
		cbp->c_last = --x;
		/*# break */
		break;
	case (ISTRIP|IUCLC|INLCR|ICRNL):	/*#*/ 
		for (i=0;i<count;i++) {
			c = (*data++)&0177;
			if (c == '\r')
				c = '\n';
			else
				if (c == '\n')
					c = '\r';
				else
					if (('A' <= c) && (c <= 'Z'))
						c += 'a'-'A';
			*cbdata++ = c;
		}
		cbp->c_last = --count;
		/*# break */
		break;
	default:	/*# no input char processing */
		for (i=0;i<count;i++)
			*cbdata++ = *data++;
		cbp->c_last = --count;
	}
}
#endif A_OPTIMIZE
