#include "../sys/param.h"
#include "../sys/uio.h"
#include "../sys/tty.h"
#include "../sys/errno.h"
#include "../sys/dir.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/ioctl.h"
#include "../sys/kernel.h"
#include "../sys/buf.h"
#include "../sys/vbavar.h"
#include "../sys/mp_host.h"
#include "../sys/mp_dl.h"
#include "../sys/mp_ahst.h"
#include "../sys/mp_adef.h"
#include "../sys/pool.h"
#include "../sys/x25conf.h"
#include "../sys/alm.h"
#include "../bop/sdlcbuf.h"


extern	short spl8();
extern short *mpbrdaddr[];
extern	struct vbmem vbmpcc;
extern struct mblok *bscmblk;
extern struct mblok *snamblk;
extern struct mblok *x25mb[];
extern int numx25mb;
extern struct vba_ctlr *mpminfo[];
char	dlbuffer[DLBUFSIZE];
char	mpdlerr;

/*
#page
*/

mpdlopen(board)
dev_t board;
{
	board = minor(board);
	if((board < 0) || (board >= MAXMPCC))
		return(ENODEV);
	return (0);
}

/*
#page
*/

mpdlwrite(board, uio)
dev_t	board;
struct uio *uio;
{
register struct dl *dl;		/* The download interface	*/
register int ret;

	board = minor(board);
	if ((vbmpcc.mbloks[board] == (MBLOK *)0) ||
	    (vbmpcc.mbloks[board]->status != MP_DLOPEN))
		return(EFAULT);
	dl = &vbmpcc.mbloks[board]->dlcnf;
	u.u_procp->p_flag |= SULOCK;	/* lock in core */
	dl->count = uio->uio_iov->iov_len;
	dl->data = dlbuffer;
	/* Copy data into kernel space */
	if(ret=uiomove(dlbuffer,dl->count,UIO_WRITE,uio))
		return(ret);
	uio->uio_resid -= dl->count;	/* set up return from write */
	dl->op_code = NORMAL;
	ret = wt4bd(dl);
	u.u_procp->p_flag &= ~SULOCK;	/* unlock from core */
	return (ret);
}

mpdlread(board,uio)
dev_t board;
struct uio *uio;
{
	register struct dl *dl;		/* The download interface	*/
	register struct mblok *mb;

	board = minor(board);
	if((mb = vbmpcc.mbloks[board]) == (MBLOK *)0)
		return(EEXIST);
	dl = &mb->dlcnf;
	if(!(mb->corestat & MP_COREWAIT))
		return(EINVAL);
	mb->corestat |= MP_READCORE;
	slumber(&mb->corestat, 0, iocomboost);
	uiomove((caddr_t)dlbuffer,1024,UIO_READ,uio);
	return(0);
}
/*
#page
*/


mpdlclose(board)
dev_t	board	;
{

	board = minor(board);
	if ((vbmpcc.mbloks[board] == (MBLOK *)0) ||
	    (vbmpcc.mbloks[board]->status != MP_DLDONE)) {
		vbmpcc.vbinit.status = 0;
		return(EEXIST);
	}

	vbmpcc.mbloks[board]->status = MP_OPOPEN;
	vbmpcc.vbinit.status = 0;
	/* set to dead, for board handshake */
	vbmpcc.mbloks[board]->hostint.imok = MP_DEAD;

	return(0);
}

/*
#page
*/


mpdlioctl(board, cmd, addr, flag)
dev_t	board;
int	flag, cmd;
register char	*addr;
{
	register struct mblok *mb;
	register struct dl *dl;
	register int	ret,i,j;
	int	dltime();
	short s;
	struct x25ptcf *l1;
	struct x25bdcf *ptr;
	register long *lptr;
	extern int bscbase;

	board = minor(board);
	if((mb = vbmpcc.mbloks[board]) == (MBLOK *)0)
		return(EEXIST);
	dl = &mb->dlcnf;
	ret = 0;
	u.u_procp->p_flag |= SULOCK;	/* lock in core */

	switch (cmd) {

	case	PORTMAP:
		bcopy(addr, mb->protoport,MAXMPCCPRT);
		break;

	case	FCCS:
		/*  addr+0 used to be bit mask of fccs ports */
		/*  Now we assume if FCCS if specified, ALL ports are FCCS */
		dlbuffer[0] = *(addr +1);  /* should be 1 for FCCS on */
		dlbuffer[1] = *(addr +2);  /* Bit mask of ports for SOC */
		dl->count = 2;
		dl->data = dlbuffer;
		dl->op_code = FCCS & IOCPARM_MASK;
		ret = wt4bd(dl);
		break;

	case	HILO:
		bcopy(addr,(char *)&mb->hiport,2);
		break;

	case	ENDDL:
		dl->count = 0;
		dl->data = (char *) 0;
		dl->op_code = ENDDL & IOCPARM_MASK;
		ret = wt4bd(dl);
		mpccinit(board);
		mb->status = MP_DLDONE;

		/* remember first port on this board that is MP_BISYNC */
		for(i=0;i<MAXMPCCPRT;i++)
			if (mb->protoport[i] == MP_BISYNC) 
				{
				bscbase = i;
				break;
				}

		for(i=0;i<MAXMPCCPRT;i++)
			if(mb->protoport[i] == MP_SNA) 
				snamblk = mb;

		for(i=0;i<MAXMPCCPRT;i++)
			if(mb->protoport[i] == MP_X25) {
				for(j=0;j<numx25mb;j++)
					if(x25mb[j] == mb)
						break;
/* 9/27/87  JFC
 *  Modified to allow X.25 ports on boards other than board 0
 */
				if (board <= numx25mb)
				{
					mb->hostint.brdnum = board;
					x25mb[board] = mb;
					
				}
				break;
			}
		break;

	case	ENDCODE:
		dl->count = 0;
		dl->data = (char *) 0;
		dl->op_code = ENDCODE & IOCPARM_MASK;
		ret = wt4bd(dl);
		break;

	case	X25BDCF:
		bcopy(addr, dlbuffer, sizeof(struct x25bdcf) );
		ptr = (struct x25bdcf *)dlbuffer;
		dl->count = sizeof(struct x25bdcf);
		dl->data = dlbuffer;
		dl->op_code = X25BDCF & IOCPARM_MASK;
		ret = wt4bd(dl);
		break;

	case	X25PTCF:
		bcopy(addr, dlbuffer, sizeof(struct x25ptcf) );
		dl->data = dlbuffer;
		dl->count = sizeof(struct x25ptcf);
		dl->op_code = X25PTCF & IOCPARM_MASK;
		ret = wt4bd(dl);
		break;

	case	ASYNCNF:
		bcopy(addr, dlbuffer, sizeof(struct abdcf) );
		dl->count = sizeof(struct abdcf);
		dl->data = dlbuffer;
		dl->op_code = ASYNCNF & IOCPARM_MASK;
		ret = wt4bd(dl);
		break;

	case	BSCCNF:
		/* put in MPCC board addr for the bisync board */
		bscmblk = mb;
		break;

	case	SNACNF:
		break;

	case	STARTDL:
		if (!mpbrdaddr[board])
		{
			/* board doesn't exist - not okay */
			ret = ENODEV;
			break;
		}

		/* if the initialization area is in use */
		if(vbmpcc.vbinit.status)
		{
			/* Don't allow access */
			ret = EACCES;
			/* if it is this board that is being downloaded */
			if(mb->status == MP_DLOPEN)
			{
				ret = ENOSPC;
			}
			break;
		}

		/* if the board is in use */
		if(mb->status == MP_OPOPEN)
		{
			/* Don't allow access */
			ret = EBUSY;
			break;
		}

		/* board some other status */
		/* initialize the downloading interface */

		vbmpcc.vbinit.magic = MPMAGIC;
		vbmpcc.vbinit.mb = mb;
		vbmpcc.vbinit.status = 1;
		dl->go_stat = MP_FREE;
		dl->count = 0;
		dl->op_code = 0;
		dl->data = (char *) 0;
		mpdlerr = 0;
		mb->magic = MPMAGIC;
		mb->ivect = mpminfo[board]->um_ivct & 0x0000ff; /*interrupt*/
		mb->status = MP_DLPEND;
		mb->di_swtch[0] = 'A';
		mb->di_swtch[1] = 'P';
		s = spl8();
		*(mpbrdaddr[board]) = 2;
		timeout(dltime, mb, 900); /* approx 15 seconds */
		slumber(&mb->status, 0, iocomboost);
		splx(s);
		if(mb->status == MP_DLOPEN)
		{
			untimeout(dltime, mb);
		}
		else if(mb->status == MP_DLTIME)
		{
			vbmpcc.vbinit.status = 0;
			ret = EXTMOUT;
		}
		else
		{
			vbmpcc.vbinit.status = 0;
			ret = ENXIO;
			printf("MPCC %d - startdl: unknown status",board);
		}
		/* Zero mpccport structure */
		j = sizeof(struct mpccprt) * MAXMPCCPRT / sizeof(long);
		lptr = (long *)mb->mpccport;
		for(i=0;i<j;i++)
			*lptr++ = 0L;
		break;

	case	RESETBOARD:
		if (!mpbrdaddr[board])
		{
			/* board doesn't exist - not okay */
			ret = EEXIST;
			break;
		}
		s = spl8();
		*(mpbrdaddr[board]) = 0x100;
		if(mb->status == MP_DLOPEN || mb->status == MP_DLDONE) {
			mpdlerr = MP_DLERROR;
			dl->go_stat = MP_FREE;
			wakeup(&dl->go_stat);
			vbmpcc.vbinit.status = 0;
		}
		for(i=0;i<MAXMPCCPRT;i++) {
			if(mb->herr || mb->serr)
				mpccalarm(mb,i);
			if(mb->protoport[i] == MP_SNA)
				xl_cleanprt(mb,i);
			if(mb->protoport[i] == MP_X25)
				xl_cleanprt(mb,i);
			if(mb->protoport[i] == MP_ASYNC)
				a3_cleanprt(mb,i);
			mb->protoport[i] = MP_UNUSED;
		}
		mb->status = 0;
		splx(s);
		break;

	case MPCOREDMP:
		if(mb->corestat & MP_COREWAIT)
			ret = EINVAL;
		else {
			ret = 0;
			mb->corestat |= MP_COREWAIT;
		}
		break;

	default:
		ret = EINVAL;	/* Bad Command */
		break;
	}
	u.u_procp->p_flag &= ~SULOCK;	/* unlock from core */
	return (ret);
}

/*
#page
*/

mpdlint(board)
short	board;
{
	register struct mblok *mb;
	register struct dl *dl;
	long alarmb[3];

	mb = vbmpcc.mbloks[board];
	dl = &mb->dlcnf;

	switch(mb->status) {
	case MP_DLOPEN:
		if(dl->go_stat != MP_DONE)
			mpdlerr = MP_DLERROR;

		dl->go_stat = MP_FREE;
		wakeup(&dl->go_stat);
		return;

	case MP_DLPEND:
		mb->status = MP_DLOPEN;
		wakeup(&mb->status);

	case MP_DLTIME:
		return;

	case MP_OPOPEN:
		/*****
		printf("MPCC core dump code\n");
		if(!(mb->corestat & MP_READCORE)) {
			dl->count = -1;
			dl->go_stat = DONE;
		}
		else
			wakeup(&mb->corestat);
		*****/
		break;

	default:
		printf("MPCC %ld: Random download interupt - ignored\n",board);
	}
}

/*
#page
*/
dltime(mp)
register struct mblok *mp;
{
	mp->status = MP_DLTIME;
	wakeup(&mp->status);
}

/*
#page
*/

/*	generic routine which sleeps until a
	transfer to the board is completed.
	indicated by an interrupt, or a timeout.
*/

wt4bd(dl)
register struct dl *dl;
{
	register int s;
	register int ret;

	s = spl8();
	ret = 0;
	dl->go_stat = MP_GO;

	while( dl->go_stat != MP_FREE )
	{
		slumber(&dl->go_stat, 0, iocomboost);
		if(mpdlerr == MP_DLERROR)
			ret = EIO;
	}
	splx(s);
	return(ret);
}
