#include "../sys/param.h"
#include "../sys/errno.h"
#include "../bop/sdlcbuf.h"
#include "../bop/vcbop.h"
#include "../bop/qmgmt.h"
#include "../ml/mtpr.h"
#include "../sys/mp_host.h"
#include "../sys/alm.h"
#include "../sys/vmmeter.h"

extern struct vbmem vbmpcc;

#ifdef DEBUG
extern int mpatrace;
int mpxtrace = 0;
#endif

extern struct mpccsw mprototbl[]; 
extern	int mphis ;			/* port to trace */
extern	int mpahis ;			/* port to trace */
extern int sl_mpcmpl(), sl_mpframe();

struct oflowq {
	struct sdlcbuf *_f;
	struct sdlcbuf *_b;
}oflowq[MAXMPCC*MAXSYNCPRT];


/*
#page
#doc mpint -  Interrupt from MPCC handler 
*/
/* One interupt per MPCC */
mpint(n)
short n;	/* MPCC number */
/*
#edoc
*/
{
	register int i;
	register HIS *his;
	register MBLOK *mb;

	/** Get pointer to host interface structure */
	mb = vbmpcc.mbloks[n];
	/** If board in nt operational, ignore interupt */
	if(mb->status != MP_OPOPEN)
		return;
	/** mblok is in non-cacheable memory, ITS UP TO THE PROTOCOL COMPLETION
	 ** ROUTINE TO PURGE THE CACHE FOR ITS BUFFERS/STRUCTURES.
	 **/
	his = &mb->hostint;
	/** Semaphore should already be set to not available */
	his->semaphore &= ~MP_AVAILABLE;
#ifdef DEBUG
	mplog(0);
	if(mpatrace || mpxtrace)
		printf("mpint ");
#endif
	if(mb->herr || mb->serr)
		mpccalarm(mb,n);
	for(i=1;i<MAXPROTOCOL;i++) {  /** skip the UNUSED protocol */
		/** if work for this protocol's outbound completion routine */
		if (his->proto[i].outbdone[0] != EOL)
			/** Process all incoming frames/events */
			(*his->proto[i].outcmpl)(n,his->proto[i].outbdone);
		/** if work for this protocol's inbound completion routine */
		if (his->proto[i].inbdone[0] != EOL)
			/** Process all write/command completions */
			(*his->proto[i].incmpl)(n,his->proto[i].inbdone);
	}
	/** Set semaphore to available */
	his->semaphore |= MP_AVAILABLE;
}

mpccalarm(mb,board)
register MBLOK *mb;
short board;
{
	register int i;
	long alarmb[3];

	alarmb[0] = (long)board;
	alarmb[1] = (long)mb->di_swtch[0];
	switch((unsigned)mb->serr) {
	case NOSERR:	/* no error */
		break;
	case SERR1:	/* xmit compl with no active xmit */
		alarmb[2] = (long)"xmit compl with no active xmit";
		break;
	case SERR2:	/* There are no rcv buf headers */
		alarmb[2] = (long)" There are no rcv buf headers ";
		break;
	case SERR3:	/* xl_rxidtst:entry not busy */
		alarmb[2] = (long)" xl_rxidtst:entry not busy ";
		break;
	case SERR4:	/* xl_ses:Unrecognized command */
		alarmb[2] = (long)" xl_ses:Unrecognized command ";
		break;
	case SERR5:	/* xl_seu:Unrecognized command */
		alarmb[2] = (long)" xl_seu:Unrecognized command ";
		break;
	case SERR6:	/* xl_stch:Data error */
		alarmb[2] = (long)" xl_stch:Data error ";
		break;
	case SERR7:	/* xl_xidtst:entry not busy */
		alarmb[2] = (long)" xl_xidtst:entry not busy ";
		break;
	case SERR8:	/* xl_clkstr: Invalid timer */
		alarmb[2] = (long)" xl_clkstr: Invalid timer ";
		break;
	case SERR9:	/* xl_clkstr:no timers in list */
		alarmb[2] = (long)" xl_clkstr:no timers in list ";
		break;
	case SERR10:	/* xl_tmcan: Invalid timer */
		alarmb[2] = (long)" xl_tmcan: Invalid timer ";
		break;
	case SERR11:	/* xl_tmexp: Invalid timer */
		alarmb[2] = (long)" xl_tmexp: Invalid timer ";
		break;
	case SERR12:	/* xl_tmset: Invalid timer */
		alarmb[2] = (long)" xl_tmset: Invalid timer ";
		break;
	case SERR13:	/* xl_tmset: time left and no timers */
		alarmb[2] = (long)" xl_tmset: time left and no timers ";
		break;
	case SERR14:	/* xl_ufrm:FRMR:no entry in ackq */
		alarmb[2] = (long)" xl_ufrm:FRMR:no entry in ackq ";
		break;
	case SERR15:	/* xl_ufrm:FRMR:entry in ackq */
		alarmb[2] = (long)" xl_ufrm:FRMR:entry in ackq ";
		break;
	case SERR16:	/* xl_wrcomp:no entry in ackq */
		alarmb[2] = (long)" xl_wrcomp:no entry in ackq ";
		break;
	case SERR17:	/* xl_wrcomp:entry in ackq */
		alarmb[2] = (long)" xl_wrcomp:entry in ackq ";
		break;
	case SERR18:	/* xl_fixwt: no entry in ackq */
		alarmb[2] = (long)" xl_fixwt: no entry in ackq ";
		break;
	case SERR19:	/* xl_ruframe: not a uframe */
		alarmb[2] = (long)" xl_ruframe: not a uframe ";
		break;
	case SERR20:	/* xl_ses:no buffer headers */
		alarmb[2] = (long)" xl_ses:no buffer headers ";
		break;
	case SERR21:	/* xl_seu:no data buffer */
		alarmb[2] = (long)" xl_seu:no data buffer ";
		break;
	case SERR22:	/* xl_seu:no buffer headers */
		alarmb[2] = (long)" xl_seu:no buffer headers ";
		break;
	case SERR23:	/* xl_1tompcc:unknwn entry type */
		alarmb[2] = (long)"xl_1tompcc: unknown entry type";
		break;
	case SERR25:	/* xl_ercv unknown entry */
		alarmb[2] = (long)"xl_ercv unknown entry";
		break;
	case SERR30: 	/* sl_1tmpcc: unknown entry type */
		alarmb[2] = (long)"sl_1tmpcc: unknown entry type ";
		break;
	case SERR31:	/* sl_ercv: unknown entry type */
		alarmb[2] = (long)"sl_ercv: unknown entry type ";
		break;
	case SERR32:	/* sl_rxidtst:entry not busy */
		alarmb[2] = (long)"sl_rxidtst:entry not busy ";
		break;
	case SERR33:	/* sl_3tompcc:undefined opcode  */
		alarmb[2] = (long)"sl_3tompcc:undefined opcode  ";
		break;
	case SERR34:	/* sl_ses:no buffer headers */
		alarmb[2] = (long)"sl_ses:no buffer headers ";
		break;
	case SERR35:	/* sl_ses:Unrecognized command  */
		alarmb[2] = (long)"sl_ses:Unrecognized command  ";
		break;
	case SERR36:	/* sl_seu:no data buffer */
		alarmb[2] = (long)"sl_seu:no data buffer ";
		break;
	case SERR37:	/* sl_seu:no buffer headers */
		alarmb[2] = (long)"sl_seu:no buffer headers ";
		break;
	case SERR38:	/* sl_seu:Unrecognized command  */
		alarmb[2] = (long)"sl_seu:Unrecognized command  ";
		break;
	case SERR39:	/* sl_stch:Data error  */
		alarmb[2] = (long)"sl_stch:Data error  ";
		break;
	case SERR40:	/* sl_xidtst:entry not busy  */
		alarmb[2] = (long)"sl_xidtst:entry not busy  ";
		break;
	case SERR41:	/* sl_ruframe:not a uframe  */
		alarmb[2] = (long)"sl_ruframe:not a uframe";
		break;
	case SERR42:	/* xl_dec:decision table error  */
		alarmb[2] = (long)"xl_dec:decision table error";
		break;
	default:
		alarmb[2] = (long)"Unknown error";
	}
	if(mb->serr) {
		printf("Soft error: MPCC %d port %d\n%s\n",alarmb[0],
			alarmb[1],alarmb[2]);
		mb->serr = 0;
	}
	switch((unsigned)mb->herr) {
	case NOHERR:	/* no error */
		break;
	case MPBUSERR:	/* bus error */
		alarmb[1] = (long)"Bus error";
		break;
	case ADDRERR:	/* address error */
		alarmb[1] = (long)"Address error";
		break;
	case UNDECC:	/* undefined ecc interrupt */
		alarmb[1] = (long)"Undefined ecc interrupt";
		break;
	case UNDINT:	/* undefined interrupt */
		alarmb[1] = (long)"Undefined interrupt";
		break;
	case PWRFL:		/* power fail occurred */
		alarmb[1] = (long)"Power failure occurred";
		break;
	case NOXENTRY:	/* xdone was enterred without an xmit entry on queue */
		alarmb[1] = (long)"Stray transmit done interrupt";
		break;
	case TWOFTMRS:	/* trying to start two fast timers on one port */
		alarmb[1] = (long)"Two fast timers on one port";
		break;
	case INTQFULL:	/* interupt queue full */
		alarmb[1] = (long)"Interupt queue full";
		break;
	case INTQERR:	/* interupt queue ack error */
		alarmb[1] = (long)"Interupt queue ack error";
		break;
	case ACPDEAD: /* ACAP for 32 port MPCC failed power up */
		alarmb[1] = (long)"32 port ACAP failed power up";
		break;
	case PANIC1:	/* clk_rem: timer not in queue */
		alarmb[1] = (long)"clk_rem: timer not in queue";
		break;
	case PANIC2:	/* clk_res: illegal time value */
		alarmb[1] = (long)"clk_res: illegal time value";
		break;
	case PANIC3:	/* clk_set: illegal tix */
		alarmb[1] = (long)"clk_set: illegal tix";
		break;
	case PANIC4:	/* clkgetb: VRTX allocation error */
		alarmb[1] = (long)"clkgetb: VRTX allocation error";
		break;
	case PANIC5:	/* clkgetb: Too many clocks */
		alarmb[1] = (long)"clkgetb: Too many clocks";
		break;
	case PANIC6:	/* clkrelb: release error */
		alarmb[1] = (long)"clkrelb: release error";
		break;
	case PANIC7:	/* Bad queue */
		alarmb[1] = (long)"Bad queue";
		break;
	case PANIC8:	/* mp_scan: VRTX sc_post error */
		alarmb[1] = (long)"mp_scan: VRTX sc_post ";
		break;
	case PANIC9:	/* mp_ipool: Buffer pools too big\n */
		alarmb[1] = (long)"mp_ipool: Buffer pools too big";
		break;
	case PANIC10:	/* xl_preprocess:VRTX sc_pend error */
		alarmb[1] = (long)"xl_preprocess: VRTX sc_pend error";
		break;
	case PANIC11:	/* xl_1tompcc: unknown entry type\n */
		alarmb[1] = (long)"xl_1tompcc: unknown entry type";
		break;
	case PANIC12:	/* xl_ercv: unknown error status */
		alarmb[1] = (long)"xl_ercv: unknown error status";
		break;
	case PANIC13:	/* xl_rxidtst:entry not busy */
		alarmb[1] = (long)"xl_rxidtst: entry not busy";
		break;
	case PANIC14:	/* xl_3tompcc: undefined op_code */
		alarmb[1] = (long)"xl_3tompcc: undefined op_code";
		break;
	case PANIC15:	/* xl_ses: no buffer headers. */
		alarmb[1] = (long)"xl_ses: no buffer headers";
		break;
	case PANIC16:	/* xl_ses:Unrecognized command */
		alarmb[1] = (long)"xl_ses: Unrecognized command";
		break;
	case PANIC17:	/* xl_seu: no data buffers. */
		alarmb[1] = (long)"xl_seu: no data buffers";
		break;
	case PANIC18:	/* xl_seu: no buffer headers. */
		alarmb[1] = (long)"xl_seu: no buffer headers";
		break;
	case PANIC19:	/* xl_seu: Unrecognized command */
		alarmb[1] = (long)"xl_seu: Unrecognized command";
		break;
	case PANIC20:	/* xl_stch: Data error  */
		alarmb[1] = (long)"xl_stch: Data error";
		break;
	case PANIC21:	/* xl_xidtst:entry not busy */
		alarmb[1] = (long)"xl_xidtst:entry not busy";
		break;
	case PANIC22:	/* xl_guframe: no uframes\n */
		alarmb[1] = (long)"xl_guframe: no uframes";
		break;
	case PANIC23:	/* xl_ruframe: not a uframe\n */
		alarmb[1] = (long)"xl_ruframe: not a uframe";
		break;
	case PANIC24:	/* xl_clkstr: Data error. */
		alarmb[1] = (long)"xl_clkstr: Data error";
		break;
	case PANIC25:	/* xl_clkstr:no timers in list */
		alarmb[1] = (long)"xl_clkstr: no timers in list";
		break;
	case PANIC26:	/* xl_tmcan: no such timer to cancel */
		alarmb[1] = (long)"xl_tmcan: no such timer to cancel";
		break;
	case PANIC27:	/* xl_tmexp: unknown timer\n */
		alarmb[1] = (long)"xl_tmexp: unknown timer";
		break;
	case PANIC28:	/* xl_tmset: no such timer to set */
		alarmb[1] = (long)"xl_tmset: no such timer to set";
		break;
	case PANIC29:	/* xl_tmset: time left with no timers */
		alarmb[1] = (long)"xl_tmset: time left with no timers";
		break;
	default:
		alarmb[1] = (long)"Unknown error";
	}
	if(mb->herr) {
		printf("Hard error: MPCC %d\n%s\n",alarmb[0],alarmb[1]);

		if(mb->status == MP_OPOPEN) {
			for(i=0;i<MAXMPCCPRT;i++) {
				if(mb->mpccport[i].protocol == MP_SNA)
					xl_cleanprt(mb,i);
				if(mb->mpccport[i].protocol == MP_X25)
					xl_cleanprt(mb,i);
				if(mb->mpccport[i].protocol == MP_ASYNC)
					a3_cleanprt(mb,i);
				mb->protoport[i] = MP_UNUSED;
			}
		}
		mb->status = 0;
		mb->herr = 0;
	}
}
mppurge(addr,cnt)
char *addr;
int cnt;
{
	register int i;

	mtpr ((int)addr, P1DC); 
	for (i = cnt ; i >= 0; i -= NBPG)
		mtpr ((int)addr + i-1, P1DC); 
}

xl_cleanprt(mblk,port)
register MBLOK *mblk;
int port;
{
	register int i;
	register MPCCPRT *mp;
	register MPCCINENT *in;
	register MPCCOUTENT *out;
	struct sdlcbuf *nbp;
	char list[2];
	int absport;

	mp = &mblk->mpccport[port];
	i = mp->off;
	while(i != mp->on) {
		/**= Get pointer to MPCC Inbound Queue */
		in = &mp->Inents[i];
		in->un.x25.error = ENXIO;
		in->go_stat = MP_DONE;
		/**= Bump off index */
		if(++i == MP_INSET)
			i = 0;
	}
	list[0] = port;
	list[1] = EOL;
	if (mp->protocol == MP_X25) {
		(*mprototbl[mp->protocol].m_incmpl) ((short)mblk->mpcc_num,list);
		/* check overflowq */
		absport = (mblk->hostint.brdnum*MAXSYNCPRT) + port;
		while(oflowq[absport]._f != (struct sdlcbuf *)&oflowq[absport]) {
			nbp = oflowq[absport]._f;
			deque(nbp);
			nbp->b_error = ENXIO;
			if(mphis >= 0)
		  	mpccrecord((long)port,(long)(MP_WRITE+100),
			  	(long)nbp,(long)nbp->b_error);
			/**= Send bp to level 3 */
			(*nbp->b_compl)(nbp);  /* normally xn_cmpl */
		}
	}
	else if (mp->protocol == MP_SNA)
		(*mprototbl[mp->protocol].m_incmpl) ((short)mblk->mpcc_num,list);

	/** Clear x25 for port */
	mp->protocol = MP_UNUSED;
	mp->flags = 0;
	mp->on = 0;
	mp->off = 0;
	mp->nextrcv = 0;
	for(i=0;i<MP_OUTSET;i++) {
		out = &mp->Outents[i];
		out->go_stat = MP_FREE;
		out->op_code = 0;
		out->un.x25.error = 0;
		if(nbp = (sbuf_t)out->un.x25.hsp) {
			sdlcputb(nbp);
			out->un.x25.hsp = (char *)0;
		}
	}
	for(i=0;i<MP_INSET;i++) {
		in = &mp->Inents[i];
		in->go_stat = MP_FREE;
		in->op_code = 0;
		in->un.x25.error = 0;
		in->un.x25.hsp = (char *)0;
	}
}
