
/**********************************************************************
 *
 *   Multi-Port Communication Controller  (MPCC)
 *
 **********************************************************************/

#include "../sys/mp_host.h"
#include "../sys/pool.h"
#include "../sys/param.h"
#include "../sys/ioctl.h"
#include "../sys/tty.h"
#include "../sys/dir.h"
#include "../sys/user.h"
#include "../sys/map.h"
#include "../ml/pte.h"
#include "../sys/buf.h"
#include "../sys/vbavar.h"
#include "../sys/conf.h"
#include "../sys/file.h"
#include "../sys/uio.h"


/*
 *  Virtual address of MPCCs
 */

short *mpbrdaddr[] = {
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
   (short *)0,
};

/*
 *  External data.
 */

extern struct vba_ctlr		*mpminfo[];	/* vba controller info  */
extern struct vba_device	*mpinfo[];	/* vba device info      */
extern struct vbmem		 vbmpcc;	/* ptr to mbloks	*/

/*
 *  Driver information for auto-configuration 
 */

int	mpfstin = 1; 			 /* flag used in attach */

int  mpprobe();
int  mpslave();
int  mpattach();
int  mprint();

/*
 *  Driver structure.
 */

struct  vba_driver mpdriver = {
   mpprobe,		/* check if controller exists	*/
   mpslave,		/* check if a slave exist	*/
   mpattach,		/* initialize slave		*/
   0,			/* handler go routine		*/
   0,			/* device csr address		*/
   "MPCC",		/* name of the device		*/
   mpinfo,		/* table of unit info		*/
   "MPCC ",		/* name of the controller	*/
   mpminfo,		/* table of ctlr info		*/
   HID_HMPCC,		/* controller's module id	*/
   0			/* no exclusive use of bdp's	*/
};
   

/*************************************************************************
 *  Procedure:	mpprobe
 *
 *  Description: "mpprobe" verifies that an mpcc controller is really
 *               there.  It is called during the configuration phase
 *		of a reboot for each mpcc controller in the configuration
 *
 *  Returns:	TRUE means the controller is ready.
 **************************************************************************/
   
int
mpprobe(vba_ctlr)
   
   register struct vba_ctlr	*vba_ctlr;  /* vba controller information */
   
{
   if ( badaddr(vba_ctlr->um_addr, 2, 0) )
      return(0);
   
   return( mp_get_ivct( vba_ctlr ) ) ;
}


/* 
 *  Allocate two interrrupt vectors
 */
int 
mp_get_ivct( um )
   struct vba_ctlr * um ;
{
   int mpdl_ivct, mp_ivct ;
   extern	int	Xmpdlint0,	Xmpint0,	
			Xmpdlint1,	Xmpint1,	
   			Xmpdlint2,	Xmpint2,	
			Xmpdlint3,	Xmpint3,	
                    	Xmpdlint4,	Xmpint4,	
                    	Xmpdlint5,	Xmpint5,	
                     	Xmpdlint6,	Xmpint6,	
                    	Xmpdlint7,	Xmpint7,	
                    	Xmpdlint8,	Xmpint8,	
                    	Xmpdlint9,	Xmpint9,	
                     	Xmpdlint10,	Xmpint10,	
                     	Xmpdlint11,	Xmpint11,	
                    	Xmpdlint12,	Xmpint12,	
			Xmpdlint13,	Xmpint13,	
			Xmpdlint14,	Xmpint14,	
			Xmpdlint15,	Xmpint15;

	static	int	mp_proc[] = {
	            	(int)& Xmpdlint0,  (int)& Xmpint0, 
	           	(int)& Xmpdlint1,  (int)& Xmpint1, 
                    	(int)& Xmpdlint2,  (int)& Xmpint2, 
                   	(int)& Xmpdlint3,  (int)& Xmpint3, 
	            	(int)& Xmpdlint4,  (int)& Xmpint4, 
	           	(int)& Xmpdlint5,  (int)& Xmpint5, 
                    	(int)& Xmpdlint6,  (int)& Xmpint6, 
                   	(int)& Xmpdlint7,  (int)& Xmpint7, 
	            	(int)& Xmpdlint8,  (int)& Xmpint8, 
	           	(int)& Xmpdlint9,  (int)& Xmpint9, 
                    	(int)& Xmpdlint10, (int)& Xmpint10,
                   	(int)& Xmpdlint11, (int)& Xmpint11,
	            	(int)& Xmpdlint12, (int)& Xmpint12,
	           	(int)& Xmpdlint13, (int)& Xmpint13,
                    	(int)& Xmpdlint14, (int)& Xmpint14,
                    	(int)& Xmpdlint15, (int)& Xmpint15,
		     } ;

   mpdl_ivct = get_ivct( 0, 1 ) ;
   mp_ivct   = get_ivct( 0, 1 ) ;

   if ( (mpdl_ivct==(-1)) || (mp_ivct==(-1)) ) {
      return_ivct( mpdl_ivct ) ;
      return_ivct( mp_ivct ) ;
      return(0) ;
   }

   init_ivct( mpdl_ivct, mp_proc[ (um->um_ctlr * 2) + 0 ] ) ;
   init_ivct(   mp_ivct, mp_proc[ (um->um_ctlr * 2) + 1 ] ) ;
   
   um->um_ivct = mpdl_ivct | (mp_ivct << 8) ;
   return(1);
}


/*************************************************************************
 *  Procedure:	mpslave
 *
 *  Description: "mpslave" always returns true so the "mpattach" routine
 *               will be called
 *
 *  Returns:	always returns TRUE
 **************************************************************************/

int
mpslave(vba_unit,regs)
   
   struct vba_device	*vba_unit;	/* vba drive info */
   caddr_t		regs;		/* io address */
   
{
   return( 1 );
}


/*************************************************************************
 *  Procedure:	mpattach
 *
 *  Description: "mpattach" initializes the mpcc device.  It is called
 * 		 during the configuration phase of a reboot for each mpcc.
 *
 *  Returns:
 **************************************************************************/

mpattach(vba_unit)
   
   register struct vba_device *vba_unit; /* vba device information */
   
{
   register int i;
   
   /*
    *  Initialize the buffer pool (do this only once)
    */
   if(mpfstin) {
      pool_init();
      mpfstin = 0;
   }
   
   /*
    *  Allocate a mblok and put in non cacheable memory
    */
   i = vba_unit->ui_unit;
   vbmpcc.mbloks[i] = (struct mblok *)pool_alloc(PL_MPCC);
   slvnocache(vbmpcc.mbloks[i],sizeof(struct mblok));
   
   /*
    *  Initialize MPCC 
    */
   mpbrdaddr[i] = (short *)vba_unit->ui_addr;
   *(mpbrdaddr[i]) = 0x100;
   return;
}


/*
 * Set the non-cacheable bit of the PTE for
 * the specified virtual page.
 */
slvnocache(v, nbytes)
   register caddr_t v;
   register int nbytes;
{
   register int i;
   register int j;
   
   i = (int) (((unsigned int) v & ~0xc0000000) >> PGSHIFT);
   j = (int) ((((unsigned int) v + nbytes - 1) & ~0xc0000000) >> PGSHIFT);
   while (i <= j)
      *(int *) &Sysmap[i++] |= PG_N;
}
