
#include "../sys/time.h"
#include "../sys/param.h"
#include "../sys/kernel.h"
#include "../sys/pool.h"
#include "../bop/qmgmt.h"
#include "../bop/sdlcbuf.h"

/*
 * operating system internal buffer pool management routines
 */

extern short spl8();
void	pool_empty();

/*	pool_alloc	buffer pool allocation*/

/*
 *	Pool_alloc() returns the address of a block taken from the pool
 *	of the given pool class.  For some classes, it is possible for
 *	NULL to be returned if the pool is empty.  No data structure on
 *	the returned block is set up.
 *
 */

char *
pool_alloc( pool )
int pool ; /* block pool class */
{
	register struct poolque *hdr ;
	register struct slink *element ;
	register int s ;

	if (pool >= npool )
		panic("pool_alloc: illegal argument") ;

	hdr = &poolhdr[pool]; /* adrs of header for this class */

	s = spl8() ;
/*
 *	test for and handle an empty pool
 */
	while ( quempty(hdr) )
		{
		if (hdr->pq_type & PQ_RETNULL )
			{  /* return null on empty pool */
			splx(s) ;
			return ( (char *) 0 );
			}
		pool_empty(hdr) ; /* commom empty pool handling */
		}
/*
 *	pool is not empty, allocate block
 */
	element = (slink_t)hdr->pq_fw ;	/* get top of queue */
	deque(element);
	splx(s) ;
	return ( (char *) element) ;
}

/*	pool_dealloc()	system block pool deallocation*/

/*
 *	Pool_dealloc() is the opposite of pool_alloc().  It is used
 *	to return a block to the block pool system.  The deallocated
 *	block must be returned to the correct pool class, and must
 *	not be accessed after deallocation.
 */

pool_dealloc( pool, block )
int pool ;	/* block pool class */
char *block ;	/* adrs of returned block */
{
	register struct poolque *hdr ;
	register int s ;

	if ( pool >= npool )
		panic("pool_dealloc: illegal argument") ;

	hdr = &poolhdr[pool] ;	/* pool header adrs */

	s = spl8() ;

	enquend( block, hdr );	/* return to end of queue */
/*
 *	check for someone waiting on block deallocate
 */
	if ( hdr->pq_type & PQ_WAITING )
		{
		hdr->pq_type &= ~PQ_WAITING;
		wakeup( (char *)hdr) ;
		}

	splx(s) ;
	return ;
}

/*	pool_empty()	common empty pool handler*/

/*
 *	Pool_empty() is a routine used by pool_alloc and its duplicates
 *	for handling empty poolqueues.	If it panics, it won't return.
 *	If is does return, there is no quarantee that the queue is
 *	now non-empty, since multiple pool waits may be in effect.
 *	The test for the 'return-null' option for the pool class is
 *	handled outside of pool_empty.
 *	This routine must be called within a critical section
 */

void
pool_empty( hdr )
register struct poolque *hdr ;	/* pointer to pool header */
{

/*
 *	if pool class specifies wait-for more blocks, and if its a process
 *	context calling, wait and return when woken up.
 */
	if ( (hdr->pq_type & PQ_WAIT))
		{
		hdr->pq_type |= PQ_WAITING ;
 		if ( hdr->pq_pri <= PZERO ) 
		   slumber( (char *)hdr, 0, devcntrlboost);
		else
		   slumber( (char *)hdr, INTERRUPTABLE, devcntrlboost);
		return ;
		}
	panic ("pool_empty: system pool exhausted") ;
}
/*#page*/
/*#doc	pool_init()	pool initialization*/
/*
 *	pool initialize routine
 */
void
pool_init()

{
	register long poolcnt;
	register long cnt;
	register char *storage;
	register struct poolque *hdr ;	/* pointer to pool header */

	for (poolcnt = 0; poolcnt < npool; poolcnt++) {
		hdr = &poolhdr[poolcnt];
		if ((hdr->pq_cnt == 0) || (hdr->pq_size == 0))
			continue;
		if ((storage=hdr->pq_start) == (char *)0)
			panic("pool_init: no memory");

		for (cnt = 0; cnt < hdr->pq_cnt; cnt++, storage = storage 
				+ hdr->pq_size) {
			enquend(storage, hdr);
		}
	}

}
