/* $Header: in_cksum.c,v 1.0.0.0 85/06/03 03:24:24 release Exp $ */
/*	in_cksum.c	6.1	83/07/29	*/

#include "../sys/types.h"
#include "../sys/mbuf.h"
#include "../netinet/in.h"
#include "../netinet/in_systm.h"

/*
 * Checksum routine for Internet Protocol family headers.
 *
 * This routine is very heavily used in the network
 * code and should be modified for each CPU to be as fast as possible.
 * 
 * This implementation is HCX version.
 */

/*
 * ADDCARRY folds the upper 16 bits of the sum onto the lower 16 bits.
 * This is done periodically to ensure that there is no carry from the
 * 32 bit sum.
 *
 * Note: ADDCARRY itself might produce a carry into the upper 16 bits.
 * Call ADDCARRY a couple of extra times at the end to ensure this
 * doesn't stay broken.
 */
#undef	ADDCARRY
#define ADDCARRY {                                 \
			asm( "shrl $16,r9,r0" ) ;  \
			asm( "andl2 $0xFFFF,r9" ) ;  \
			asm( "addw2 r0,r9" ) ;     \
			asm( "adwc  $0,r9" ) ;     \
		}

in_cksum(m, len)
	register struct mbuf *m;       /* r12 */
	register int len;              /* r11 */
{
	register u_short *w;           /* r10 */
	register int sum = 0;          /* r9 */
	register int mlen = 0;

	for (;m && len; m = m->m_next) {

		if (m->m_len == 0) {
			continue;
		}

		w = mtod(m, u_short *);
		if (mlen == -1) {
			/*
			 * The first byte of this mbuf is the continuation
			 * of a word spanning between this mbuf and the
			 * last mbuf.
			 */

			sum += *(u_char *)w;

			w = (u_short *)((u_char *)w + 1);
			mlen = m->m_len - 1;
			len--;
		} else
			mlen = m->m_len;

		if (len < mlen)
			mlen = len;
		len -= mlen;

		/*
		 * If the words are not aligned,
		 *    take the hit and do it hard way.
		 */
		if ((int)w & 0x1) {
			while ((mlen -= 2) >= 0) {
				sum += (*(u_char *)w << 8) + *((u_char *)w+1);
				w++;
			}
		/*
		 * If the words are word aligned,
		 *    things are much easier.
		 */
		} else {

	                /*
	                 * Align the buffer to a longword boundary.
	                 */

	                if ((int)w & 0x2 && mlen >= 2 ) {
				sum += *w++;
				mlen -= 2 ;
			}

			/*
			 * Sum a bunch of longword aligned data
			 */
	                while ( (mlen -= 32) >= 0 ) {
	                        asm( "addl2 (r10),r9" ) ;
	                        asm( "adwc 4(r10),r9" ) ;
	                        asm( "adwc 8(r10),r9" ) ;
	                        asm( "adwc 12(r10),r9" ) ;
	                        asm( "adwc 16(r10),r9" ) ;
	                        asm( "adwc 20(r10),r9" ) ;
	                        asm( "adwc 24(r10),r9" ) ;
	                        asm( "adwc 28(r10),r9" ) ;
	                        asm( "adwc $0,r9" ) ;
	                        w += 16 ;
	                }
	                mlen += 32 ;
	                ADDCARRY;

		        /*
		         * Pick up any trailing words
		         */
			while ((mlen -= 2) >= 0) {
				sum += *w++;
			}
		}

	        /*
	         * Add the odd byte at the end of the mbuf.
	         */
		if (mlen == -1) {
			sum += *(u_char *)w << 8 ;
		}

		/*
		 * Fold high order 16 bits into low order 16 bits.
		 * Since we don't take care of the carry on each and every
		 * addition, we may run into trouble if we get an mbuf
		 * with more than 64K bytes. (Hardly likely)
		 */
		ADDCARRY;
	}
	/*
	 * Be careful about throwing away the final carry bit.
	 */
	ADDCARRY;
	if (len)
		printf("cksum: out of data\n");
	return (~sum & 0xffff);
}
