/*   audit_sys.c    1/28/85                    */
/*  Documentation block & change history:
   bernie  4/28/85  New for HCX system version.  transfer from VAX
		    Major function include the process of audit by 
		    setting up audit flag in the file descriptor
		    block, the audit status checking of the file,
		    in addition to the actual logging in the audit
		    trail.
*/
#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/file.h"
#include "../sys/vnode.h"
#include "../sys/buf.h"
#include "../sys/uio.h"
#include "../sys/audit.h"
#include "../sys/pathname.h"
#include "../ufs/inode.h"
/*------ debugging tools ------------*/
#ifdef AUDPRINTFS
int	audprintfs;
#define INTTRACE		1
#define RTINTRACE 		2
#define STARTTRACE 		4
#define ACEOTRACE 		8
#define RTINTRACE1 		0x10
#define IOCTLTRACE		0x20
#define ATRACE			0x40
#define SETETTRACE		0x80
#define	ALLTRACE		(INTTRACE | RTINTRACE | STARTTRACE\
				 | ACEOTRACE | RTINTRAUD1 | IOCTLTRACE\
				 | ATRACE | SETETTRACE)

#define AUDTRACE(X,Y)	{if (audprintfs & X) Y}
#else AUDPRINTFS
#define AUDTRACE(X,Y)
#endif AUDPRINTFS



/*
 * Set the current state of the auditing flag for a file.
 */
audit()
{
	struct inode *ip ;
	struct vnode *vp;
	register struct a {
		char	*fname;
		int	mode;
	} *uap = (struct a*)u.u_ap;
	int audchange;
	struct pathname pn ;

	AUDTRACE(ALLTRACE, printf("audit_sys  :  audit  \n");)
	u.u_error=lookuppathname( uap->fname, UIO_USERSPACE, &pn, FOLLOW_LINK,
		(struct vnode *) 0, &vp ) ;
	if (u.u_error) {
	        pn_free(&pn) ;
		return ;
	}

	/*
	 * Auditting is performed on the local system only. Cannot handle
	 * auditting on NFS files.
	 *
	 * Don't allow changing the audit mode on the audit file itself.
	 * That would cause deadlock.
	 */
	if ( ! VNISIN( vp ) || vp == audvp ) {
	        pn_free(&pn) ;
	        VN_RELE( vp ) ;
	        u.u_error = ENOENT ;
	        return ;
	}
	ip = VTOI(vp) ;
	ilock( ip ) ;

	if (uap->mode == QUERY) {
	        pn_free( &pn ) ;
	} else {
		switch (uap->mode) {

		case SECURE:	/* turn auditing on for a file */
			audchange = AUDOFFTON;
			break;

		case UNSECURE:	/* turn auditing off for a file */
			audchange = AUDONTOFF;
			break;

		case MASK:	/* make the auditing match the umask */
			if ((u.u_cmask & AUDMASK) != 0)
				audchange = AUDOFFTON;
			else
				audchange = AUDONTOFF;
			break;

		default:
			auditlog(vp, &pn, AUDREAD, FAILURE);
			u.u_error = EINVAL;
			iput( ip ) ;
			return;
		}

		/* if OWNER returns an error */
		if ( OWNER( u.u_cred, ip )) {
			auditlog(vp, &pn, audchange, FAILURE);
			u.u_error = EACCES;
			iput(ip);
			return;
		}

 		if ((ip->i_mode&IFMT) == IFDIR) {
			auditlog(vp, &pn, audchange, FAILURE);
			iput(ip) ;
			u.u_error = EISDIR;
			return;
		}

		auditlog(vp, &pn, audchange, SUCCESS);
		if (audchange == AUDOFFTON)
			ip->i_ic.ic_flags |= AUDLOG;	/* set audit bit */
		else
			ip->i_ic.ic_flags &= ~AUDLOG;	/* clear audit bit */
		imark( ip, ICHG ) ;
	}

	/*
	 * Return the current state of the audited flag.
	 */
	if ((ip->i_ic.ic_flags & AUDLOG) != 0)
		u.u_r.r_val1 = SECURE;
	else 
		u.u_r.r_val1 = UNSECURE; 
	iput(ip); 

}

/*
 * Get the current state of the auditing flag for a file descriptor.
 */
ckaudit()
{
	register struct file *fp;
	register struct inode *ip;
	register struct a {
		int fd;
	} *uap;

	AUDTRACE(ALLTRACE, printf("audit_sys  :  ckaudit  \n");)
	uap = (struct a *)u.u_ap;
	fp = getf(uap->fd);
	if (fp == NULL)
		return;
	if ( ! INODEFP( fp) ) {
	        u.u_r.r_val1 = UNSECURE ;
	        return ;
	}
	ip = VTOI( (struct vnode *)fp->f_data ) ;
	if ((ip->i_ic.ic_flags & AUDLOG) != 0)
		u.u_r.r_val1 = SECURE;
	else
		u.u_r.r_val1 = UNSECURE;
}

/*
 * Start audit trail logging
 */
audon()
{
	struct vnode * vp ;
	register struct a {
		char	*fname;
	} *uap;
	struct pathname pn ;

	AUDTRACE(ALLTRACE, printf("audit_sys  :  audon  \n");)
	uap = (struct a *)u.u_ap;
	if ( ! suser())
	        return ;

	if (uap->fname==NULL) {
		if (audvp != NULL) {
			VN_RELE( audvp ) ;
			audvp = NULL;
		}
		else
			u.u_error = EINVAL;
		return;
	}
	u.u_error=lookuppathname( uap->fname, UIO_USERSPACE, &pn, FOLLOW_LINK,
		(struct vnode *) 0, &vp ) ;
	if (u.u_error) {
	        pn_free(&pn) ;
		return ;
	}

	if ( vp->v_type == VDIR) {
		u.u_error = EACCES;
	        pn_free(&pn) ;
		VN_RELE( vp ) ;
		return;
	}

	if (audvp != NULL) {
	        VN_RELE( audvp ) ;
	}
	audvp = vp;
}

struct aud_rec outbuf;

auditlog(vp, pnp, mode, result)
register struct vnode *vp;		/* vnode ptr to the file in question */
register struct pathname *pnp ;         /* pathname to log */
int mode;				/* mode - read | write */
int result;				/* result - success | failure */
{
	register struct inode * ip ;
	register struct aud_rec *ap = &outbuf;
	register char *src, *dst;
	int siz, len;
	register struct vnode *lvp;
	struct vattr vattr;
	int s, error;
	int resid ;

	AUDTRACE(ALLTRACE, printf("audit_sys  :  auditlog  \n");)
	/*
	 * If no audit trail file, return.
	 */
	if ( pnp->pn_buf == NULL ) {
		return ;
	}
	if ( (lvp = audvp) == NULL || vp == lvp || vp == 0 || ! VNISIN( vp ) ) {
		pn_free( pnp ) ;
		return;
	}

	ip = VTOI( vp ) ;

	/* is it a change rec? */
	if (mode != AUDOFFTON && mode != AUDONTOFF) {
		if ((ip->i_ic.ic_flags & AUDLOG) == 0) {
			/* file flagged for logging? */
			pn_free( pnp ) ;
			return;
		}
	}

	error = u.u_error;		/* save u_error for return to user */
	u.u_error = 0;
	siz = 0;

/*	Set the flag word	*/

	if (result == SUCCESS)
		ap->a_flag = AUDSUCCESS | mode;
	else
		ap->a_flag = mode;
	siz += sizeof(ap->a_flag);

/*	Set the user id		*/

	ap->a_uid = u.u_ruid;
	siz += sizeof(ap->a_uid);

/*	Set the time 		*/

	s = spl8();
	ap->a_time = time.tv_sec;
	splx(s);
	siz += sizeof(ap->a_time);

/*	Save the command	*/

	dst = ap->a_comm;
	src = u.u_comm;
	do {
		*dst++ = *src;
		siz++;
	} while (*src++ != '\0');

/*	Save the pathname	*/

	if ( pnp->pn_type == ROOTREL) {
		if (u.u_rdir != NULL) {
			len = min( u.u_audrootdir.aud_len, &(ap->a_path[MAXPATHLEN]) - dst ) ;
			bcopy( u.u_audrootdir.aud_buf, dst, len ) ;
			siz += len - 1;	/* less one for the null terminator */
			dst += len - 1;
		}
	} else {
		len = min( u.u_audcwdir.aud_len, &(ap->a_path[MAXPATHLEN]) - dst ) ;
		bcopy( u.u_audcwdir.aud_buf, dst, len ) ;
		siz += len - 1;	/* less one for the null terminator */
		dst += len - 1;
	}
	if (dst < &(ap->a_path[MAXPATHLEN])) {
		*dst++ = '/';
		siz++;
	}
	len = min( strlen(pnp->pn_buf)+1, &(ap->a_path[MAXPATHLEN]) - dst ) ;
	bcopy( pnp->pn_buf, dst, len ) ;
	ap->a_path[MAXPATHLEN] = '\0' ;
	siz += len ;

	pn_free(pnp);

/* post the record for writing to the logfile */

	if (u.u_error = VOP_GETATTR(lvp, &vattr, u.u_cred)) {
		eprintf("Exited auditlog - file not logged: errno %d\n",u.u_error);
		goto bad;
	}

	u.u_error =
	    vn_rdwr(UIO_WRITE, lvp, (caddr_t)ap, siz, vattr.va_size,
	     		UIO_SYSSPACE, IO_UNIT, &resid);
	/* If error, restore file size. */
	if (u.u_error) {
		int oldsize = vattr.va_size ;
		vattr_null(&vattr);
	        vattr.va_size = oldsize ;
	        u.u_error = VOP_SETATTR( vp, &vattr, u.u_cred ) ;
		eprintf("Exited auditlog - file not logged: errno %d\n",u.u_error);
	}
bad:
	u.u_error = error;		/* restore u_error */
} 
/*
 * Pack out "./" and "../" entries from a path name string.
 *
 */
audpack( auddirp )
	struct auddir *auddirp ;
{
	register char *cp, *slashp, *strp;
	int len ;

	AUDTRACE(ALLTRACE, printf("aud_sys:  audpack  \n");)

	cp = strp = auddirp->aud_buf ;
	while (*cp != '\0') {
		if (*cp == '/') {
			*strp++ = *cp++;
			while ( *cp == '/' ) {
				cp ++ ;
			}
			slashp = cp;
			if (*slashp == '.' && *(slashp + 1) == '/') {
				cp += 2;
			} else if (*slashp =='.' && *(slashp + 1) == '.'
			    && ( *(slashp + 2) == '/' || *(slashp+2) == '\0')) {
				cp += 2;
				strp -= 2;
				while (strp > auddirp->aud_buf && *strp != '/')
					strp--;
				if (strp < auddirp->aud_buf)
					strp = auddirp->aud_buf;
			}
		} else
			*strp++ = *cp++;
	}
	*strp = '\0';

	if( strp != cp ) {
		len = strp - auddirp->aud_buf + 1 ;
		if ( auddirlen( len ) != auddirlen( auddirp->aud_len ) ) {
			cp = (char*)kmem_alloc( auddirlen( len )) ;
			bcopy( auddirp->aud_buf, cp , len ) ;
			kmem_free( auddirp->aud_buf, auddirlen( auddirp->aud_len) ) ;
			auddirp->aud_buf = cp ;
		}
		auddirp->aud_len = len ;
	}
}
