/*	init_main.c	6.1	83/07/29	*/
/*   Change history :
   4/15/85  bernie  add audit trail buffer initialization.
*/

#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/vfs.h"
#include "../sys/map.h"
#include "../sys/proc.h"
#include "../sys/seg.h"
#include "../sys/conf.h"
#include "../sys/buf.h"
#include "../sys/vm.h"
#include "../sys/cmap.h"
#include "../sys/text.h"
#include "../sys/tty.h"
#include "../sys/termio.h"
#include "../sys/reboot.h"
#include "../sys/vnode.h"
#include "../ufs/inode.h"
#include "../ufs/quota.h"
#include "../ml/reg.h"

extern	daddr_t		cdlimit;	/* for ulimit */
extern int (*ufs_lockctl_ptr)();
extern int (*flckinit_ptr)();
/*
 * Initialization code.
 * Called from cold start routine as
 * soon as a stack and segmentation
 * have been established.
 * Functions:
 *	clear and free user core
 *	turn on clock
 *	hand craft 0th process
 *	call all initialization routines
 *	fork - process 0 to schedule
 *	     - process 1 execute bootstrap
 *	     - process 2 to page out
 *
 * loop at loc 13 (0xd) in user mode -- /etc/init
 *	cannot be executed.
 */
main(firstaddr)
	int firstaddr;
{
	register int i;
	register struct proc *p;
	int s;

	if ( boothowto & RB_LOAD_ONLY )
	{
		printf("Load-Only option set (VM enabled).\n\n") ;
		asm(" halt") ;
	}

	rqinit();
	startup(firstaddr);

	/*
	 * set up system process 0 (swapper)
	 */
	ckey_cache[MAXCKEY] = 1;	/* Allocated by locore ! */
	ckey_cnt[MAXCKEY] = 1;
	p = &proc[0];
	p->p_ckey = MAXCKEY;	/* Since process 0 is in kernel */
	p->p_dkey = MAXDKEY;	/*  and MAXKEY is a 'system key'*/
	p->p_p0br = u.u_pcb.pcb_p0br;
	p->p_szpt = 1;
	p->p_addr = uaddr(p);
	p->p_stat = SRUN;
	p->p_flag |= SLOAD|SSYS;
	p->p_nice = sys_proc_pri;
	p->p_prior = p->p_basepri = NICE_TO_PRI(p->p_nice);
	p->p_usrpri = p->p_prior;
	p->p_boost = 0;
	p->p_whomp = 0;
	p->p_cpu = 0;
	p->p_compmask = NO_AIO_PENDING;
	p->p_procmask = NO_AIO_PENDING;
	p->p_areq_head = p->p_areq_tail = (struct areq *) &p->p_areq_head;
	p->p_usrareq_head = p->p_usrareq_tail = (struct areq *) &p->p_usrareq_head;
	quantum = p->p_qntm = rt_quantum;
	setredzone(p->p_addr, (caddr_t)&u);
	u.u_procp = p;
	u.u_cmask = CMASK;
	u.u_madesyscalls = 0;
	u.u_lastfile = -1;
	for (i = 0; i < sizeof(u.u_rlimit)/sizeof(u.u_rlimit[0]); i++)
		u.u_rlimit[i].rlim_cur = u.u_rlimit[i].rlim_max = 
		    RLIM_INFINITY;
	u.u_rlimit[RLIMIT_STACK].rlim_cur =
	    u.u_rlimit[RLIMIT_STACK].rlim_max = ctob(MAXSSIZ);
	u.u_rlimit[RLIMIT_DATA].rlim_max =
	    u.u_rlimit[RLIMIT_DATA].rlim_cur = ctob(MAXDSIZ);
	p->p_maxrss = RLIM_INFINITY/NBPG;
	u.u_limit = cdlimit;

	u.u_univ[0] = 'a';	u.u_univ[1] = 't';
	u.u_univ[2] = 't';	u.u_univ[3] = '\0';
	u.u_univlen = 4;

	/*
	 * initialize data key data structures
	 */
	dkey_init();

	/*
	 * initialize kernel memory allocator
	 */
	kmem_init();

	/*
	 * get vnodes for swapdev and argdev
	 */
	swapdev_vp = bdevvp(swapdev);
	argdev_vp = bdevvp(argdev);
	rootvp = bdevvp(rootdev);

	/*
	 * Setup credentials
	 */
	u.u_cred = crget();
	for (i = 1; i < NGROUPS; i++)
		u.u_groups[i] = NOGROUP;

	qtinit();

	/*
	 * Initialize tables, protocols
	 */
	mbinit();
	cinit();			/* needed by dmc-11 driver */

	loattach();			/* XXX */
	/*
	 * Block reception of incoming packets
	 * until protocols have been initialized.
	 */
	s = splimp();
	ifinit();
	domaininit();
	splx(s);

	pqinit();
	ihinit();
	bhinit();
	dnlc_init();
	nfs_init();
	binit();
	bswinit();
#ifdef GPROF
	kmstartup();
#endif

	sdlcinit();	/* VIOC-BOP and sdlc buffer pools */
	x25_init();	/* X25 lun */
	x25init();	/* X25 utility */
	errinit();
	msginit();
	seminit();
	bseminit();
	aioinit();

	/*
	 * Initialize ATT advisory record/file locking.
	 * See comments in ../cf/space.c
	 */
	if ( flckinit_ptr )
		(*flckinit_ptr)();
	ufs_vnodeops.vn_lockctl = ufs_lockctl_ptr ;

/* kick off timeout driven events by calling first time */
	schedcpu();
	/*
	 * mount the root, gets rootdir
	 */
	vfs_mountroot();
	boottime = time;
	p->p_etime = time.tv_sec;

	schedpaging();

	u.u_dmap = zdmap;
	u.u_smap = zdmap;

	/*
	 * Set the scan rate and other parameters of the paging subsystem.
	 */
	setupclock();
	/* Enable clock interrupt */
	clk_enable = 1;

	/*
	 * Make init process.
	 */
	proc[0].p_szpt = CLSIZE;
	if (newproc(0)) {
		p = &proc[1];
		p->p_nice = INIT_NICE;
		p->p_prior = p->p_usrpri = p->p_basepri =NICE_TO_PRI(p->p_nice);
		p->p_pgrp = p->p_pid;
		quantum = p->p_qntm = non_rt_quantum;
		expand(clrnd((int)btoc(szicode)), 0);
		(void) swpexpand(u.u_dsize, 0, &u.u_dmap, &u.u_smap);
		(void) copyout((caddr_t)icode, (caddr_t)0, (unsigned)szicode);
		/*
		 * This code converts 'boothowto' to ascii and passes it
		 * as the first parameter to init. This code depends on the
		 * fact that 'icode' has reserved enough blank filled, trailing
		 * zero space for the entire hex number.
		 */
		{
		        char * ptr = (char*) (szicode-2) ;
		        int n = boothowto ;
			do {
				*ptr-- = "0123456789abcdef"[n%16];
				n /= 16 ;
			} while (n);
			*ptr-- = 'x' ;
			*ptr-- = '0' ;
		}
 		u.u_audcwdir.aud_buf = (char*)kmem_alloc( auddirlen(2));
 		u.u_audcwdir.aud_buf[0] = '/' ;
 		u.u_audcwdir.aud_buf[1] = '\0' ;
 		u.u_audcwdir.aud_len = 2 ;
		/*
		 * Return goes to loc. 0 of user init
		 * code just copied out.
		 */
		return(0);
	}

	/*
	 * Make page-out daemon (process 2).
	 * The daemon has ctopt(nswbuf*CLSIZE*KLMAX) pages of page
	 * table so that it can map dirty pages into
	 * its address space during asychronous pushes.
	 */
	proc[0].p_szpt = clrnd(ctopt(nswbuf*CLSIZE*KLMAX + UPAGES));
	if (newproc(0)) {
		proc[2].p_flag |= SLOAD|SSYS;
		proc[2].p_dsize = u.u_dsize = nswbuf*CLSIZE*KLMAX; 
		pageout();
		/*NOTREACHED*/
	}

	proc[0].p_szpt = 1;

#ifdef HCXRT
	/*
	 * Make periodic event daemon (process 3).
	 */
	if (newproc(0)) {
		proc[3].p_flag |= SLOAD|SSYS;
		softclock();
		/*NOTREACHED*/
	}
#endif

	/*
	 * Enter scheduling loop.
	 */
	sched();
	/*NOTREACHED*/
}

/*
 * Initialize hash links for buffers.
 */
bhinit()
{
	register int i;
	register struct bufhd *bp;

	for (bp = bufhash, i = 0; i < BUFHSZ; i++, bp++)
		bp->b_forw = bp->b_back = (struct buf *)bp;
}

/*
 * Initialize the buffer I/O system by freeing
 * all buffers and setting all device buffer lists to empty.
 */
binit()
{
	register struct buf *bp, *dp;
	register int i;
	struct swdevt *swp;
	int base, residual;

	for (dp = bfreelist; dp < &bfreelist[BQUEUES]; dp++) {
		dp->b_forw = dp->b_back = dp->av_forw = dp->av_back = dp;
		dp->b_flags = B_HEAD;
	}
	base = bufpages / nbuf;
	residual = bufpages % nbuf;
	for (i = 0; i < nbuf; i++) {
		bp = &buf[i];
		bp->b_dev = NODEV;
		bp->b_bcount = 0;
		bp->b_un.b_addr = buffers + i * MAXBSIZE;
		if (i < residual)
			bp->b_bufsize = (base + 1) * CLBYTES;
		else
			bp->b_bufsize = base * CLBYTES;
		binshash(bp, &bfreelist[BQ_AGE]);
		bp->b_flags = B_BUSY|B_INVAL;
		brelse(bp);
	}
	/*
	 * Count swap devices, and adjust total swap space available.
	 * Some of this space will not be available until a vswapon()
	 * system call is issued, usually when the system goes multi-user.
	 */
	swapconf();
	nswdev = 0;
	nswap = 0;
	for (swp = swdevt; swp->sw_dev; swp++) {
		nswdev++;
		if (swp->sw_nblks > nswap)
			nswap = swp->sw_nblks;
	}
	if (nswdev == 0)
		panic("binit");
	if (nswdev > 1)
		nswap = ((nswap + dmmax - 1) / dmmax) * dmmax;
	nswap *= nswdev;
	swfree(0);
}

/*
 * Initialize linked list of free swap
 * headers. These do not actually point
 * to buffers, but rather to pages that
 * are being swapped in and out.
 */
bswinit()
{
	register int i;
	register struct buf *sp = swbuf;

	bswlist.av_forw = sp;
	for (i=0; i<nswbuf-1; i++, sp++)
		sp->av_forw = sp+1;
	sp->av_forw = NULL;
}

/*
 * Initialize clist by freeing all character blocks.
 */
struct chead cfreelist;
cinit()
{
	register n;
	register struct cblock *cp;

	for(n = 0, cp = cfree; n < nclist; n++, cp++) {
		cp->c_next = cfreelist.c_next;
		cfreelist.c_next = cp;
	}
	cfreelist.c_size = CBSIZE;
}
