/*	kern_acct.c	6.1	83/07/29	*/

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/vnode.h"
#include "../sys/vfs.h"
#include "../sys/kernel.h"
#include "../sys/acct.h"
#include "../sys/uio.h"

/*
 * SHOULD REPLACE THIS WITH A DRIVER THAT CAN BE READ TO SIMPLIFY.
 */
struct	vnode *acctp;
struct	vnode *savacctp;

/*
 * Perform process accounting functions.
 */
sysacct()
{
	register struct a {
		char	*fname;
	} *uap = (struct a *)u.u_ap;
	static aclock;
	struct vnode *vp;

	if (aclock || !suser())
		return;
	aclock++;

	if (savacctp) {
		acctp = savacctp;
		savacctp = NULL;
	}

	switch (uap->fname) {
	case NULL:
		if (vp = acctp) {
			VN_RELE(vp) ;
			acctp = NULL;
		}
		break;
	default:
		if (acctp) {
			u.u_error = EBUSY;
			break;
		}
		u.u_error =
		    lookupname(uap->fname, UIO_USERSPACE, FOLLOW_LINK,
			(struct vnode **)0, &vp);
		if (u.u_error)
			break;
		if (vp->v_type != VREG)
			u.u_error = EACCES;
		else
			u.u_error = VOP_ACCESS( vp, VWRITE, u.u_cred );
		if (u.u_error) {
		        int temp = u.u_error;

			VN_RELE(vp);
			u.u_error = temp;
			break ;
		}
		acctp = vp;
	}
	aclock--;
}

int	acctsuspend = 2;	/* stop accounting when < 2% free space left */
int	acctresume = 4;		/* resume when free space risen to > 4% */

struct	acct acctbuf;
/*
 * On exit, write a record on the accounting file.
 */
acct(st)
{
	register int i;
	register struct vnode *vp;
	register struct acct *ap = &acctbuf;
	struct statfs sb;
	struct timeval t;

	if (savacctp) {
		(void)VFS_STATFS(savacctp->v_vfsp, &sb);
		if (sb.f_bavail > (acctresume * sb.f_blocks / 100)) {
			acctp = savacctp;
			savacctp = NULL;
			printf("Accounting resumed\n");
		}
	}
	if ((vp = acctp) == NULL)
		return;
	(void)VFS_STATFS(acctp->v_vfsp, &sb);
	if (sb.f_bavail <= (acctsuspend * sb.f_blocks / 100)) {
		savacctp = acctp;
		acctp = NULL;
		printf("Accounting suspended\n");
		return;
	}
	for (i = 0; i < sizeof (ap->ac_comm); i++)
		ap->ac_comm[i] = u.u_comm[i];

 /*	ap->ac_utime = compress((long)u.u_ru.ru_utime.tv_sec);
	ap->ac_stime = compress((long)u.u_ru.ru_stime.tv_sec);
	ap->ac_etime = compress((long)time.tv_sec - u.u_start);   */
	/*
	 *  Write accounting times in terms of clock ticks
	 */
	ap->ac_utime = compress((long)
		(u.u_ru.ru_utime.tv_sec * hz + u.u_ru.ru_utime.tv_usec/tick));
	ap->ac_stime = compress((long)
		(u.u_ru.ru_stime.tv_sec * hz + u.u_ru.ru_stime.tv_usec/tick));
	t = time;
	timevalsub(&t, &u.u_start);
	ap->ac_etime = compress((long) (t.tv_sec * hz + t.tv_usec/tick));
	ap->ac_btime = u.u_start.tv_sec;
	ap->ac_uid = u.u_ruid;
	ap->ac_gid = u.u_rgid;
	ap->ac_mem = 0;
 	/* 
 	 *  Note that ru_i?rss is in unit of NBPG core tick
 	 */

	/* if (i = (u.u_ru.ru_utime.tv_sec + u.u_ru.ru_stime.tv_sec)  */

	if (i = (u.u_ru.ru_utime.tv_sec + u.u_ru.ru_stime.tv_sec)*hz +
		(u.u_ru.ru_utime.tv_usec + u.u_ru.ru_stime.tv_usec)/tick)
		ap->ac_mem = compress((u.u_ru.ru_ixrss +
			u.u_ru.ru_idrss + u.u_ru.ru_isrss) / i);
	ap->ac_io = compress(u.u_ru.ru_iochr);
	ap->ac_rw = compress((long)(u.u_ru.ru_inblock + u.u_ru.ru_oublock));
	if (u.u_ttyp)
		ap->ac_tty = u.u_ttyd;
	else
		ap->ac_tty = NODEV;
	ap->ac_stat = st;
	ap->ac_flag = u.u_acflag;
	VN_HOLD( vp ) ; /* Prevent file from disappearing */
	u.u_error =
	    vn_rdwr(UIO_WRITE, vp, (caddr_t)ap, sizeof(acctbuf), 0,
		UIO_SYSSPACE, IO_UNIT|IO_APPEND, (int *)0);
	VN_RELE( vp ) ;
}

/*
 * Produce a pseudo-floating point representation
 * with 3 bits base-8 exponent, 13 bits fraction.
 */
compress(t)
	register long t;
{
	register exp = 0, round = 0;

	while (t >= 8192) {
		exp++;
		round = t&04;
		t >>= 3;
	}
	if (round) {
		t++;
		if (t >= 8192) {
			t >>= 3;
			exp++;
		}
	}
	return ((exp<<13) + t);
}
