/*	kern_exec.c	6.2	83/08/23	*/

#include "../ml/reg.h"
#include "../ml/pte.h"
#include "../ml/psl.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/map.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../sys/buf.h"
#include "../sys/vnode.h"
#include "../sys/pathname.h"
#include "../sys/seg.h"
#include "../sys/vm.h"
#include "../sys/text.h"
#include "../sys/file.h"
#include "../sys/uio.h"
#include "../sys/acct.h"
#include "../sys/vfs.h"
#include "../sys/audit.h"
#include "../sys/sysinfo.h"

#include "../ml/mtpr.h"
#include "../fpe/fp_in_krnl.h"

/*
 * exec system call, with and without environments.
 */
struct execa {
	char	*fname;
	char	**argp;
	char	**envp;
};

exec()
{
	((struct execa *)u.u_ap)->envp = NULL;
	exece();
}

exece()
{
	register int nc;
	register char *cp;
	register struct buf *bp = 0;
	register struct execa *uap;
	register int ap;
	int na, ne, ucp, len, cc, ds;
	int indir, uid, gid;
	char *sharg;
	struct vnode *vp;
	struct vattr vattr;
	swblk_t bno;
	struct pathname pn;
	char cfarg[SHSIZE];
	char *execnamep ;
	int resid, error;
	register struct proc *p = u.u_procp;

	struct	naout	{
		short	magic;
		short	vstamp;
		long	tsize,
			dsize,
			bsize,
			entry,
			ts,
			ds;
	};

	struct	filhd	{
		unsigned short	magic;
		unsigned short	nscns;
		long		timdat,
				symptr,
				nsyms;
		unsigned short	opthdr,
				flags;
	} ;

	struct	scnhdr	{
		char		s_name[8];
		long		s_paddr,
				s_vaddr,
				s_size,
				s_scnptr,
				s_relptr,
				s_lnnoptr;
		unsigned short	s_nreloc,
				s_nlnno;
		long		s_flags;
	};

	sysinfo.sysexec++;
	uap = (struct execa *)u.u_ap;
	u.u_error = lookuppathname(uap->fname, UIO_USERSPACE,
		&pn, FOLLOW_LINK, (struct vnode **)0, &vp);
	if (u.u_error) {
		pn_free(&pn);
		return;
	}

	bno = 0;
	indir = 0;
	uid = u.u_uid;
	gid = u.u_gid;
	if (u.u_error = VOP_GETATTR(vp, &vattr, u.u_cred))
		goto bad;
	if ((vp->v_vfsp->vfs_flag & VFS_NOSUID) == 0) {
 		if (vattr.va_mode & VSUID)
 			uid = vattr.va_uid;
 		if (vattr.va_mode & VSGID)
 			gid = vattr.va_gid;
 	} else if (vattr.va_mode & (VSUID|VSGID)) {
 		struct pathname tmppn;
		u.u_error = pn_get(uap->fname, UIO_USERSPACE, &tmppn);
 		if (u.u_error)
 			goto bad;
 		uprintf("%s: Setuid execution not allowed\n", tmppn.pn_buf);
 		pn_free(&tmppn);
 	}

  again:
	/*
	 * XXX should change VOP_ACCESS to not let super user always have it
	 * for exec permission on regular files.
	 */
	if (u.u_error = VOP_ACCESS(vp, VEXEC, u.u_cred))
		goto bad;
	if ((u.u_procp->p_flag&STRC)
	    && (u.u_error = VOP_ACCESS(vp, VREAD, u.u_cred)))
		goto bad;
	if (vp->v_type != VREG ||
	   (vattr.va_mode & (VEXEC|(VEXEC>>3)|(VEXEC>>6))) == 0) {
		u.u_error = EACCES;
		goto bad;
	}

	/*
	 * Read in first few bytes of file for segment sizes, ux_mag:
	 *	407 = plain executable
	 *	410 = RO text
	 *	413 = demand paged RO text
	 *	570 = common object file format
	 * 	575 =  "
	 * Set ux_tstart to start of text portion.
	 * Also an ASCII line beginning with #! is
	 * the file name of a ``shell'' and arguments may be prepended
	 * to the argument list if given here.
	 *
	 * SHELL NAMES ARE LIMITED IN LENGTH.
	 *
	 * ONLY ONE ARGUMENT MAY BE PASSED TO THE SHELL FROM
	 * THE ASCII LINE.
	 */
	u.u_exdata.ux_shell[0] = 0;	/* for zero length files */
	u.u_error =
	    vn_rdwr(UIO_READ, vp, (caddr_t)&u.u_exdata, sizeof (u.u_exdata),
		0, UIO_SYSSPACE, IO_UNIT, &resid);
	if (u.u_error)
		goto bad;
	u.u_count = resid;
#ifndef lint
	if (u.u_count > sizeof(u.u_exdata) - sizeof(u.u_exdata.Ux_A) &&
	    u.u_exdata.ux_shell[0] != '#') {
		u.u_error = ENOEXEC;
		goto bad;
	}
#endif
	if (u.u_exdata.ux_mag == 0570 || u.u_exdata.ux_mag == 0575) {
		ds = u.u_exdata.ux_nscns;
		u.u_exdata.ux_tstart = sizeof(u.u_exdata.Ux_A) +
			(ds * sizeof (struct scnhdr));
		u.u_exdata.ux_mag = u.u_exdata.aux_mag;
	}

	switch (u.u_exdata.ux_mag) {

	case 0407:
		u.u_exdata.ux_dsize += u.u_exdata.ux_tsize;
		u.u_exdata.ux_tsize = 0;
		break;

	case 0413:
		u.u_exdata.ux_tstart = clrnd(btoc(u.u_exdata.ux_tstart));
	case 0410:
		if (u.u_exdata.ux_tsize == 0) {
			u.u_error = ENOEXEC;
			goto bad;
		}
		break;

	default:
		if (u.u_exdata.ux_shell[0] != '#' ||
		    u.u_exdata.ux_shell[1] != '!' ||
		    indir) {
			u.u_error = ENOEXEC;
			goto bad;
		}
		cp = &u.u_exdata.ux_shell[2];		/* skip "#!" */
		while (cp < &u.u_exdata.ux_shell[SHSIZE]) {
			if (*cp == '\t')
				*cp = ' ';
			else if (*cp == '\n') {
				*cp = '\0';
				break;
			}
			cp++;
		}
		if (*cp != '\0') {
			u.u_error = ENOEXEC;
			goto bad;
		}
		cp = &u.u_exdata.ux_shell[2];
		while (*cp == ' ')
			cp++;
		execnamep = cp;
		while (*cp && *cp != ' ')
			cp++;
		cfarg[0] = '\0';
		if (*cp) {
			*cp++ = '\0';
			while (*cp == ' ')
				cp++;
			if (*cp)
				bcopy((caddr_t)cp, (caddr_t)cfarg, SHSIZE);
		}
		indir = 1;
		auditlog( vp, &pn, AUDEXEC, SUCCESS ) ;
		VN_RELE(vp);

		u.u_error = lookuppathname(execnamep, UIO_SYSSPACE,
			&pn, FOLLOW_LINK, (struct vnode **)0, &vp);
		if (u.u_error) {
			vp = (struct vnode *)0;
			goto bad;
		}

		if (u.u_error = VOP_GETATTR(vp, &vattr, u.u_cred))
			goto bad;
		goto again;
	}

	/*
	 * This is a good place to cancel pending asynchronous I/O operations.
	 * There is still a chance of the system call failing, but so what.
	 */
	if (p->p_procmask != NO_AIO_PENDING)
		AIO_cancel(0);

	/*
	 * Collect arguments on "file" in swap space.
	 */
	na = 0;
	ne = 0;
	nc = 0;
	bno = rmalloc(argmap, (long)ctod(clrnd((int)btoc(NCARGS))));
	if (bno == 0) {
		swkill(p, "exece: no swap space for arg list");
		goto bad;
	}
	if (bno % CLSIZE)
		panic("execa: rmalloc");
	cc = CLSIZE*NBPG;
	bp = getblk(argdev_vp, bno + ctod(nc/NBPG), cc);
	cp = bp->b_un.b_addr;
	/*
	 * Copy arguments into file in argdev area.
	 */
	if (uap->argp) for (;;) {
		ap = NULL;
		sharg = NULL;
		if ( indir ) {
			if ( na == 0) {
				sharg = pn.pn_path ;
				ap = (int)sharg;
				uap->argp++;		/* ignore argv[0] */
			} else if ( na == 1 && cfarg[0] ) {
				sharg = cfarg;
				ap = (int)sharg;
			} else if ( na == 1 || na == 2 && cfarg[0] )
				ap = (int)uap->fname;
			else if (uap->argp) {
				ap = fuword((caddr_t)uap->argp);
				uap->argp++;
			}
		} else if (uap->argp) {
			ap = fuword((caddr_t)uap->argp);
			uap->argp++;
		}
		if (ap == NULL && uap->envp) {
			uap->argp = NULL;
			if ((ap = fuword((caddr_t)uap->envp)) != NULL) {
				uap->envp++;
				ne++;
			}
		}
		if (ap == NULL)
			break;
		na++;
		if (u.u_error)
			break;
		do {
			if (cc <= 0) {
				/*
				 * We depend on NCARGS being a multiple of
				 * CLSIZE*NBPG.  This way we need only check
				 * overflow before each buffer allocation.
				 */
				if (nc >= NCARGS-1) {
					error = E2BIG;
					break;
				}
				bdwrite(bp);
				cc = CLSIZE*NBPG;
				bp = getblk(argdev_vp, bno + ctod(nc/NBPG), cc);
				cp = bp->b_un.b_addr;
			}
			if (sharg) {
				error = copystr(sharg, cp, cc, &len);
				sharg += len;
			} else {
				error = copyinstr((caddr_t)ap, cp, cc, &len);
				ap += len;
			}
			cp += len;
			nc += len;
			cc -= len;
		} while (error == ENOENT);
		if (error) {
			u.u_error = error;
			if (bp)
				brelse(bp);
			bp = 0;
			goto badarg;
		}
	}
	if (bp)
		bdwrite(bp);
	nc = (nc + NBPW-1) & ~(NBPW-1);
	getxfile(vp, nc + (na+4)*NBPW, uid, gid);
	if (u.u_error) {
badarg:
		for (cc = 0; cc < nc; cc += CLSIZE*NBPG) {
			bp = baddr(argdev_vp, bno + ctod(cc/NBPG), CLSIZE*NBPG);
			if (bp) {
				bp->b_flags |= B_AGE;		/* throw away */
				bp->b_flags &= ~B_DELWRI;	/* cancel io */
				brelse(bp);
				bp = 0;
			}
		}
		goto bad;
	}

	/*
	 * Copy back arglist.
	 */
	ucp = USRSTACK - nc - NBPW;
	ap = ucp - na*NBPW - 3*NBPW;
	mtpr(ap, USP);
	(void) suword((caddr_t)ap, na-ne);
	nc = 0;
	cc = CLSIZE*NBPG;
	bp = bread(argdev_vp, bno + ctod(nc / NBPG), cc);
	bp->b_flags |= B_AGE;		/* throw away */
	bp->b_flags &= ~B_DELWRI;	/* cancel io */
	cp = bp->b_un.b_addr;
	for (;;) {
		ap += NBPW;
		if (na == ne) {
			(void) suword((caddr_t)ap, 0);
			ap += NBPW;
		}
		if (--na < 0)
			break;
		(void) suword((caddr_t)ap, ucp);
		do {
			if (cc <= 0) {
				brelse(bp);
				cc = CLSIZE*NBPG;
				bp = bread(argdev_vp, bno + ctod(nc / NBPG), cc);
				bp->b_flags |= B_AGE;		/* throw away */
				bp->b_flags &= ~B_DELWRI;	/* cancel io */
				cp = bp->b_un.b_addr;
			}
			error = copyoutstr(cp, (caddr_t)ucp, cc, &len);
			ucp += len;
			cp += len;
			nc += len;
			cc -= len;
		} while (error == ENOENT);
		if (error == EFAULT)
			panic("execa: EFAULT");
	}
	(void) suword((caddr_t)ap, 0);

	/*
	 * Reset caught signals.  Held signals
	 * remain held through p_sigmask.
	 */
	while (p->p_sigcatch) {
		nc = ffs(p->p_sigcatch);
		p->p_sigcatch &= ~sigmask(nc);
		u.u_signal[nc] = SIG_DFL;
	}
	free_usrareq_blocks(p);

	/*
	 * Reset stack state to the user stack.
	 * Clear set of signals caught on the signal stack.
	 */
	u.u_onstack = 0;
	u.u_sigsp = 0;
	u.u_sigonstack = 0;

	u.u_ar0[FP] = 0;	/* bottom of the fp chain */
	u.u_ar0[PC] = u.u_exdata.ux_entloc+2;

	for (nc = u.u_lastfile; nc >= 0; --nc) {
		if (u.u_pofile[nc] & UF_EXCLOSE) {
			closef(u.u_ofile[nc]);
			u.u_ofile[nc] = NULL;
			u.u_pofile[nc] = 0;
		}
		u.u_pofile[nc] &= ~UF_MAPPED;
	}
	while (u.u_lastfile >= 0 && u.u_ofile[u.u_lastfile] == NULL)
		u.u_lastfile--;

	/*
	 * Remember file name for accounting.
	 */
	u.u_acflag &= ~AFORK;
	if (pn.pn_pathlen > MAXCOMLEN)
		pn.pn_pathlen = MAXCOMLEN;
	bcopy((caddr_t)pn.pn_path, (caddr_t)u.u_comm,
	    (unsigned)pn.pn_pathlen);
	u.u_comm[pn.pn_pathlen] = '\0' ;

	p->p_compmask = NO_AIO_PENDING;
	p->p_procmask = NO_AIO_PENDING;
	p->p_areq_head = p->p_areq_tail = (struct areq *) &p->p_areq_head;
	p->p_usrareq_head = p->p_usrareq_tail = (struct areq *) &p->p_usrareq_head;
bad:
  	auditlog(vp, &pn, AUDEXEC, u.u_error ? FAILURE : SUCCESS);
	if (bp)
		brelse(bp);
	if (bno)
		rmfree(argmap, (long)ctod(clrnd((int) btoc(NCARGS))), bno);
	if (vp) {
	        int     temp;

	        temp = u.u_error;
		VN_RELE(vp);
		u.u_error = temp;
	}
}

/*
 * Read in and set up memory for executed file.
 */
getxfile(vp, nargc, uid, gid)
	register struct vnode *vp;
	int nargc, uid, gid;
{
	register size_t ts, ds, ss;
	register int pagi;
	register struct proc *p = u.u_procp;

	if ((p->p_flag & SVFORK) == 0)
		shmexitexec();
		if (p->p_bsemid_ds)
			bsemexitexec();
	punlock();

	/* Only demand page from files and devices which support it */
	if (u.u_exdata.ux_mag == 0413 && vp->v_vfsp->vfs_bsize >= CLBYTES) {
		pagi = SPAGI;
	} else
		pagi = 0;
	if (vp->v_flag & VTEXTMOD) {		/* XXX */
		u.u_error = ETXTBSY;
		goto bad;
	}
	if (u.u_exdata.ux_tsize != 0 && (vp->v_flag&VTEXT) == 0 &&
	    vp->v_count != 1) {
		register struct file *fp;

		for (fp = file; fp < fileNFILE; fp++) {
			if (fp->f_type == DTYPE_VNODE &&
			    fp->f_count > 0 &&
			    (struct vnode *)fp->f_data == vp &&
			    (fp->f_flag&FWRITE)) {
				u.u_error = ETXTBSY;
				goto bad;
			}
		}
	}

	/*
	 * Compute text and data sizes and make sure not too large.
	 * NB - Check data and bss separately as they may overflow 
	 * when summed together.
	 */
	ts = clrnd(btoc(u.u_exdata.ux_tsize));
	ds = clrnd(btoc((u.u_exdata.ux_dsize+u.u_exdata.ux_bsize)));
	ss = clrnd(SSIZE + btoc(nargc));
	if (chksize((unsigned)ts, (unsigned)ds, (unsigned)ss))
		goto bad;

	/*
	 * Make sure enough space to start process.
	 */
	u.u_cdmap = zdmap;
	u.u_csmap = zdmap;
	if (swpexpand(ds, ss, &u.u_cdmap, &u.u_csmap) == NULL)
		goto bad;

	/*
	 * At this point, committed to the new image!
	 * Release virtual memory resources of old process, and
	 * initialize the virtual memory of the new process.
	 * If we resulted from vfork(), instead wakeup our
	 * parent who will set SVFDONE when he has taken back
	 * our resources.
	 */
	if ((p->p_flag & SVFORK) == 0)
		vrelvm();
	else {
		p->p_flag &= ~SVFORK;
		p->p_flag |= SKEEP;
		wakeup((caddr_t)p);
		while ((p->p_flag & SVFDONE) == 0)
			slumber((caddr_t)p, 0, proccomboost);
		p->p_flag &= ~(SVFDONE|SKEEP);
	}
	p->p_flag &= ~(SPAGI|SSEQL|SUANOM);
	p->p_flag |= pagi | SOUSIG ; /* Use ATT signals (till sigvec called) */
	u.u_dmap = u.u_cdmap;
	u.u_smap = u.u_csmap;
	vgetvm(ts, ds, ss);

	if (pagi == 0) {
		u.u_error =
		    vn_rdwr(UIO_READ, vp,
			(char *)ctob(dptov(p, 0)),
			(int)u.u_exdata.ux_dsize,
			(int)(u.u_exdata.ux_tstart + u.u_exdata.ux_tsize),
			UIO_USERSPACE, IO_UNIT, (int *)0);
	}
	xalloc(vp, pagi);

	/*
	 * define new keys 
	 */
	if (!p->p_textp) { /* use existing code key */
		ckeyrelease(p);
		p->p_ckey = getcodekey();
	}
	mtpr(p->p_ckey,CCK);
	dkeyrelease(p);
	p->p_dkey = getdatakey();
	mtpr(p->p_dkey,DCK);

	if (pagi && p->p_textp)
		vinifod((struct fpte *)dptopte(p, 0),
		    PG_FTEXT, p->p_textp->x_vptr,
		    (long)((u.u_exdata.ux_tstart + ts)/CLSIZE),
		    (int)btoc(u.u_exdata.ux_dsize));

	/* THIS SHOULD BE DONE AT A LOWER LEVEL, IF AT ALL */
	mtpr(0, TBIA);

	if (u.u_error)
		swkill(p, "exece: I/O error mapping pages");
	/*
	 * set SUID/SGID protections, if no tracing
	 */
	if ((p->p_flag&STRC)==0) {
		if (uid != u.u_uid || gid != u.u_gid)
			u.u_cred = crcopy(u.u_cred);
		u.u_uid = uid;
		u.u_gid = gid;
		p->p_suid = uid;
		p->p_sgid = gid;
	} else
		psignal(p, SIGTRAP);
	u.u_tsize = ts;
	u.u_dsize = ds;
	u.u_ssize = ss;
	u.u_prof.pr_scale = 0;

bad:
	return;
}
