/*	kern_exit.c	6.1	83/07/29	*/
/*  Change history : 
 *     4/15/85  bernie free the audit trail buffer in the exit
*/

#include "../ml/reg.h"
#include "../ml/psl.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/map.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../sys/buf.h"
#include "../sys/wait.h"
#include "../sys/vm.h"
#include "../sys/file.h"
#include "../sys/mbuf.h"
#include "../sys/vnode.h"
#include "../sys/tty.h"
#include "../fpe/fp_in_krnl.h"

/*
 * Exit system call: pass back caller's arg
 */
rexit()
{
	register struct a {
		int	rval;
	} *uap;

	uap = (struct a *)u.u_ap;
	exit((uap->rval & 0377) << 8);
}

/*
 * Release resources.
 * Save u. area for parent to look at.
 * Enter zombie state.
 * Wake up parent and init processes,
 * and dispose of children.
 */
exit(rv)
	int rv;
{
	register int i;
	register struct proc *p, *q, *nq;
	register int x;
	struct mbuf *m = m_getclr(M_WAIT, MT_ZOMBIE);

#ifdef PGINPROF
	vmsizmon();
#endif
	p = u.u_procp;
#ifndef HCXRT
	BOOST(p, exitboost);
#endif
	if (p->p_compmask != NO_AIO_PENDING)
		AIO_cancel(0);
	p->p_flag &= ~(STRC|SULOCK);
	p->p_flag |= SWEXIT;
	p->p_sigignore = ~0;
	p->p_cpticks = 0;
	p->p_pctcpu = 0;
	for (i = 0; i < NSIG; i++)
		u.u_signal[i] = SIG_IGN;
	free_usrareq_blocks(p);

	untimeout(realitexpire, (caddr_t)p);
	if ((p->p_pid == p->p_pgrp)
	 && (u.u_ttyp != NULL)
	 && (u.u_ttyp->t_pgrp == p->p_pgrp)
	 && (p->p_ppid == 1)) {
		gsignal(u.u_ttyp->t_pgrp, SIGHUP);
		u.u_ttyp->t_pgrp = 0;
	}

	/* release any locked segments */
	punlock();

	/*
	 * Release virtual memory.  If we resulted from
	 * a vfork(), instead give the resources back to
	 * the parent.
	 */
	if ((p->p_flag & SVFORK) == 0) {
		shmexitexec();
		if (p->p_bsemid_ds)
			bsemexitexec();
		vrelvm();
	}
	else {
		p->p_flag &= ~SVFORK;
		wakeup((caddr_t)p);
		while ((p->p_flag & SVFDONE) == 0)
			slumber((caddr_t)p, 0, noboost);
		p->p_flag &= ~SVFDONE;
	}
	p->p_lock = 0;
	for (i = 0; i <= u.u_lastfile; i++) {
		struct file *f;

		f = u.u_ofile[i];
		if (f) {
		        /* Release record locks, if any. */
		        vno_lockrelease(f);

			u.u_ofile[i] = NULL;
			u.u_pofile[i] = 0;
			closef(f);
		}
	}
   	if (u.u_audcwdir.aud_buf != NULL) {
   		kmem_free( u.u_audcwdir.aud_buf, auddirlen( u.u_audcwdir.aud_len));
   		u.u_audcwdir.aud_buf = NULL;
   	}
   	if (u.u_audrootdir.aud_buf != NULL) {
   		kmem_free( u.u_audrootdir.aud_buf, auddirlen( u.u_audrootdir.aud_len) );
   		u.u_audrootdir.aud_buf = NULL;
   	}
	VN_RELE(u.u_cdir);
	if (u.u_rdir) {
		VN_RELE(u.u_rdir);
	}
	u.u_rlimit[RLIMIT_FSIZE].rlim_cur = RLIM_INFINITY;
	semexit();
	acct(rv);
	crfree(u.u_cred);
	vrelpt(p);
	vrelu(p, 0);
	(void) spl8();		/* hack for mem alloc race XXX */
	if (*p->p_prev = p->p_nxt)		/* off allproc queue */
		p->p_nxt->p_prev = p->p_prev;
	if (p->p_nxt = zombproc)		/* onto zombproc */
		p->p_nxt->p_prev = &p->p_nxt;
	p->p_prev = &zombproc;
	zombproc = p;
	multprog--;
	p->p_stat = SZOMB;
	noproc = 1;
	i = PIDHASH(p->p_pid);
	x = p - proc;
	if (pidhash[i] == x)
		pidhash[i] = p->p_idhash;
	else {
		for (i = pidhash[i]; i != 0; i = proc[i].p_idhash)
			if (proc[i].p_idhash == x) {
				proc[i].p_idhash = p->p_idhash;
				goto done;
			}
		panic("exit");
	}
done:
	if (p->p_pid == 1)
		panic("init died");
	p->p_xstat = rv;
	if (m == 0)
		p->p_ru = 0;
	else {
		p->p_ru = mtod(m, struct rusage *);
		*p->p_ru = u.u_ru;
		ruadd(p->p_ru, &u.u_cru);
	}

	for (q = p->p_cptr; q != NULL; q = nq) {
		nq = q->p_osptr;
		if (nq != NULL)
			nq->p_ysptr = NULL;
		if (proc[1].p_cptr)
			proc[1].p_cptr->p_ysptr = q;
		q->p_osptr = proc[1].p_cptr;
		q->p_ysptr = NULL;
		proc[1].p_cptr = q;

		q->p_pptr = &proc[1];
		q->p_ppid = 1;
		if (q->p_stat == SZOMB)
			psignal(&proc[1], SIGCHLD);	/* let "init" know */
		else {
			/*
			 * Traced processes are killed
			 * since their existence means someone is screwing up.
			 * Stopped processes are sent a hangup and a continue.
			 * This is designed to be ``safe'' for setuid
			 * processes since they must be willing to tolerate
			 * hangups anyways.
			 */
			if (q->p_flag&STRC) {
				q->p_flag &= ~STRC;
				psignal(q, SIGKILL);
			} else if (q->p_stat == SSTOP) {
				psignal(q, SIGHUP);
				psignal(q, SIGCONT);
			}
			/*
			 * Protect this process from future
			 * tty signals, clear TSTP/TTIN/TTOU if pending.
			 */
			(void) spgrp(q);
		}
		if (p->p_pid == q->p_pgrp)
			q->p_pgrp = 0;	/* disband the process group */
	}

	p->p_cptr = NULL;
	psignal(p->p_pptr, SIGCHLD);
	dkeyrelease (p);
	ckeyrelease (p);
	swtch();
}

wait()
{
	struct rusage ru, *rup;

	if ((u.u_ar0[PS] & PSL_ALLCC) != PSL_ALLCC) {
		u.u_error = wait1(0, (struct rusage *)0);
		return;
	}
	rup = (struct rusage *)u.u_ar0[R1];
	u.u_error = wait1(u.u_ar0[R0], &ru);
	if (u.u_error)
		return;
	if (rup)
		u.u_error = copyout((caddr_t)&ru, (caddr_t)rup, sizeof (struct rusage));
}

/*
 * Wait system call.
 * Search for a terminated (zombie) child,
 * finally lay it to rest, and collect its status.
 * Look also for stopped (traced) children,
 * and pass back status from them.
 */
wait1(options, ru)
	register int options;
	struct rusage *ru;
{
	register f;
	register struct proc *p;

loop:
	f = 0;	/* If SIGCHLD is ignored, zombie procs are freed in issig */
	for (p = u.u_procp->p_cptr; p; p = p->p_osptr) {
		f++;
		if (p->p_stat == SZOMB) {
			freeprc(p, 1, ru);
			return (0);
		}
		if (p->p_stat == SSTOP && (p->p_flag&SWTED)==0 &&
		    (p->p_flag&STRC || options&WUNTRACED)) {
			p->p_flag |= SWTED;
			u.u_r.r_val1 = p->p_pid;
			u.u_ar0[R1] = (p->p_cursig<<8) | WSTOPPED;
			return (0);
		}
	}
	if (f == 0)
		return (ECHILD);
	if (options&WNOHANG) {
		u.u_r.r_val1 = 0;
		return (0);
	}


	if ((u.u_procp->p_flag&SOUSIG) == 0) {
		if (slumber((caddr_t)u.u_procp, INTERRUPTABLE|PCATCH, resboost))
			return( ERESTARTSYS );
	} else
		slumber((caddr_t)u.u_procp, INTERRUPTABLE, resboost);
	goto loop;
}

/*
 * Remove zombie children from the process table.
 */
freeprc(p, flag, ru)
	register struct proc *p;
	struct rusage *ru;
{
	register struct proc *q;

	if (flag) {
		/*
		 * We're coming here from a "wait" or "wait3".
		 * Fill in the resource structure if provided (i.e., if it's
		 * a "wait3", and return the process ID and status.
		 */
		if (ru && p->p_ru)
			*ru = *p->p_ru;
		u.u_r.r_val1 = p->p_pid;
		u.u_ar0[R1] = p->p_xstat;
	}

	/*
	 * Add its resource usage to the totals for all our children and
	 * release the status buffer.
	 */
	if (p->p_ru) {
		ruadd(&u.u_cru, p->p_ru);
		(void) m_free(dtom(p->p_ru));
		p->p_ru = 0;
	}
	p->p_stat = NULL;
	p->p_pid = 0;
	p->p_ppid = 0;
	p->p_xstat = 0;
	if (*p->p_prev = p->p_nxt)	/* off zombproc */
		p->p_nxt->p_prev = p->p_prev;
	p->p_nxt = freeproc;		/* onto freeproc */
	freeproc = p;

	/*
	 * Update the sibling pointers.
	 */
	if (q = p->p_ysptr)
		q->p_osptr = p->p_osptr;
	if (q = p->p_osptr)
		q->p_ysptr = p->p_ysptr;
	if ((q = p->p_pptr)->p_cptr == p)
		q->p_cptr = p->p_osptr;

	/*
	 * Zero out other parts of the proc structure entry so that it
	 * may be reused.
	 */
	p->p_pptr = 0;
	p->p_ysptr = 0;
	p->p_osptr = 0;
	p->p_cptr = 0;
	p->p_sig = 0;
	p->p_sigcatch = 0;
	p->p_sigignore = 0;
	p->p_sigmask = 0;
	p->p_pgrp = 0;
	p->p_flag = 0;
	p->p_wchan = 0;
	p->p_cursig = 0;
}
