/*	kern_fork.c	6.1	83/07/29	*/
/*  Change history  :
 *  bernie 4/15/85    check audit indicator bit in the inode and set up
 *			flag accordingly.
 */

#include "../ml/reg.h"
#include "../ml/pte.h"
#include "../ml/psl.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/map.h"
#include "../sys/user.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../sys/vnode.h"
#include "../sys/seg.h"
#include "../sys/vm.h"
#include "../sys/text.h"
#include "../sys/file.h"
#include "../sys/acct.h"
#include "../sys/sysinfo.h"

/*
 * fork system call.
 */
fork()
{

	u.u_cdmap = zdmap;
	u.u_csmap = zdmap;
	if (swpexpand(u.u_dsize, u.u_ssize, &u.u_cdmap, &u.u_csmap) == 0) {
		u.u_ar0[R1] = 0;
		return;
	}
	fork1(0);
}

vfork()
{

	fork1(1);
}

extern int maxproc;

fork1(isvfork)
	int isvfork;
{
	register struct proc *p1, *p2;
	register a;

	sysinfo.sysfork++;

	a = 0;
	if (u.u_uid != 0) {
		for (p1 = allproc; p1; p1 = p1->p_nxt)
			if (p1->p_uid == u.u_uid)
				a++;
		for (p1 = zombproc; p1; p1 = p1->p_nxt)
			if (p1->p_uid == u.u_uid)
				a++;
	}
	/*
	 * Disallow if
	 *  No processes at all;
	 *  not su and too many procs owned; or
	 *  not su and would take last slot.
	 */
	p2 = freeproc;
	if (p2==NULL) {
		tablefull("proc");
		syserr.procovf++;
	}
	if (p2==NULL || (u.u_uid!=0 && (p2->p_nxt == NULL || a>maxproc))) {
		u.u_error = EAGAIN;
		if (!isvfork) {
			(void) vsexpand(0, &u.u_cdmap, 1);
			(void) vsexpand(0, &u.u_csmap, 1);
		}
		goto out;
	}
	p1 = u.u_procp;
	if (newproc(isvfork)) {
		u.u_r.r_val1 = p1->p_pid;
		u.u_ar0[R1] = 1;  /* child */
		u.u_start = time;
  		if (u.u_audcwdir.aud_buf != NULL) {
  			char * tmp ;
  			tmp = (char*)kmem_alloc( auddirlen( u.u_audcwdir.aud_len)) ;
  			bcopy( u.u_audcwdir.aud_buf, tmp, u.u_audcwdir.aud_len ) ;
  			u.u_audcwdir.aud_buf = tmp ;
  		}
  		if (u.u_audrootdir.aud_buf != NULL) {
  			char * tmp ;
  			tmp = (char*)kmem_alloc( auddirlen( u.u_audrootdir.aud_len) ) ;
  			bcopy( u.u_audrootdir.aud_buf,tmp,u.u_audrootdir.aud_len ) ;
  			u.u_audrootdir.aud_buf = tmp ;
  		}
		u.u_acflag = AFORK;
		/*
		 * Child must not inherit file-descriptor-oriented locks
		 * (i.e., System-V style record-locking)
		 */
		for (a = 0; a <= u.u_lastfile; a++)
			u.u_pofile[a] &= ~UF_FDLOCK;
		return;
	}
	u.u_r.r_val1 = p2->p_pid;

out:
	u.u_ar0[R1] = 0;
}

/*
 * Create a new process-- the internal version of
 * sys fork.
 * It returns 1 in the new process, 0 in the old.
 */
newproc(isvfork)
	register int isvfork;
{
	register struct proc *rpp, *rip;
	register int n;
	register struct file *fp;
	static int pidchecked = 0;

	/*
	 * First, just locate a slot for a process
	 * and copy the useful info from this process into it.
	 * The panic "cannot happen" because fork has already
	 * checked for the existence of a slot.
	 */
	mpid++;
retry:
	if (mpid >= MAXPID) {
		mpid = 100;
		pidchecked = 0;
	}
	if (mpid >= pidchecked) {
		int doingzomb = 0;

		pidchecked = MAXPID;
		/*
		 * Scan the proc table to check whether this pid
		 * is in use.  Remember the lowest pid that's greater
		 * than mpid, so we can avoid checking for a while.
		 */
		rpp = allproc;
again:
		for (; rpp != NULL; rpp = rpp->p_nxt) {
			if (rpp->p_pid == mpid || rpp->p_pgrp == mpid) {
				mpid++;
				if (mpid >= pidchecked)
					goto retry;
			}
			if (rpp->p_pid > mpid && pidchecked > rpp->p_pid)
				pidchecked = rpp->p_pid;
			if (rpp->p_pgrp > mpid && pidchecked > rpp->p_pgrp)
				pidchecked = rpp->p_pgrp;
		}
		if (!doingzomb) {
			doingzomb = 1;
			rpp = zombproc;
			goto again;
		}
	}
	if ((rpp = freeproc) == NULL)
		panic("no procs");

	freeproc = rpp->p_nxt;			/* off freeproc */
	rpp->p_nxt = allproc;			/* onto allproc */
	rpp->p_nxt->p_prev = &rpp->p_nxt;	/*   (allproc is never NULL) */
	rpp->p_prev = &allproc;
	allproc = rpp;

	/*
	 * Make a proc table entry for the new process.
	 */
	rip = u.u_procp;
	rpp->p_ckey = rip->p_ckey;
	rpp->p_dkey = 0;
	rpp->p_stat = SIDL;
	timerclear(&rpp->p_realtimer.it_value);
	rpp->p_flag = SLOAD | (rip->p_flag & (SPAGI|SOUSIG));
	if (isvfork) {
		rpp->p_flag |= SVFORK;
		rpp->p_ndx = rip->p_ndx;
	} else
		rpp->p_ndx = rpp - proc;
	rpp->p_uid = rip->p_uid;
	rpp->p_suid = rip->p_suid;
	rpp->p_sgid = rip->p_sgid;
	rpp->p_pgrp = rip->p_pgrp;
	rpp->p_nice = rip->p_nice;
	rpp->p_qntm = rip->p_qntm;
	rpp->p_basepri = rip->p_basepri;
	rpp->p_usrpri = rpp->p_basepri;
	rpp->p_prior = rpp->p_basepri;
	rpp->p_cpu = 0;
	rpp->p_boost = 0;
	rpp->p_whomp = 0;
	rpp->p_textp = isvfork ? 0 : rip->p_textp;
	rpp->p_pid = mpid;
	rpp->p_ppid = rip->p_pid;
	rpp->p_pptr = rip;
	rpp->p_osptr = rip->p_cptr;
	if (rip->p_cptr)
		rip->p_cptr->p_ysptr = rpp;
	rpp->p_ysptr = NULL;
	rpp->p_cptr = NULL;
	rip->p_cptr = rpp;
	rpp->p_etime = time.tv_sec;
	rpp->p_bsemid_ds = NULL;
	rpp->p_sigmask = rip->p_sigmask;
	rpp->p_sigcatch = rip->p_sigcatch;
	rpp->p_sigignore = rip->p_sigignore;
	/* take along any pending signals like stops? */
	if (isvfork) {
		rpp->p_tsize = rpp->p_dsize = rpp->p_ssize = 0;
		rpp->p_szpt = clrnd(ctopt(UPAGES));
		forkstat.cntvfork++;
		forkstat.sizvfork += rip->p_dsize + rip->p_ssize;
	} else {
		rpp->p_tsize = rip->p_tsize;
		rpp->p_dsize = rip->p_dsize;
		rpp->p_ssize = rip->p_ssize;
		rpp->p_szpt = rip->p_szpt;
		forkstat.cntfork++;
		forkstat.sizfork += rip->p_dsize + rip->p_ssize;
	}
	rpp->p_rssize = 0;
	rpp->p_maxrss = rip->p_maxrss;
	rpp->p_wchan = 0;
	rpp->p_eslptime = 0;
	rpp->p_pctcpu = 0;
	rpp->p_cpticks = 0;
	rpp->p_compmask = NO_AIO_PENDING;
	rpp->p_procmask = NO_AIO_PENDING;
	rpp->p_areq_head = rpp->p_areq_tail = (struct areq *) &rpp->p_areq_head;
	rpp->p_usrareq_head = rpp->p_usrareq_tail = (struct areq *) &rpp->p_usrareq_head;
	n = PIDHASH(rpp->p_pid);
	rpp->p_idhash = pidhash[n];
	pidhash[n] = rpp - proc;
	multprog++;

	/*
	 * Increase reference counts on shared objects.
	 */
	for (n = 0; n <= u.u_lastfile; n++) {
		fp = u.u_ofile[n];
		if (fp == NULL)
			continue;
		fp->f_count++;
	}
	VN_HOLD(u.u_cdir);
	if (u.u_rdir)
		VN_HOLD(u.u_rdir);
	crhold(u.u_cred) ;

	/*
	 * Partially simulate the environment
	 * of the new process so that when it is actually
	 * created (by copying) it will look right.
	 * This begins the section where we must prevent the parent
	 * from being swapped.
	 */
	rip->p_flag |= SKEEP;
	if (procdup(rpp, isvfork))
		return (1);

	/*
	 * Make child runnable and add to run queue.
	 */
	spl8();
	rpp->p_stat = SRUN;
	setrq(rpp);
	spl0();

	/*
	 * Cause child to take a non-local goto as soon as it runs.
	 * On older systems this was done with SSWAP bit in proc
	 * table; on VAX we use u.u_pcb.pcb_sswap so don't need
	 * to do rpp->p_flag |= SSWAP.  Actually do nothing here.
	 */
	/* rpp->p_flag |= SSWAP; */

	/*
	 * Now can be swapped.
	 */
	rip->p_flag &= ~SKEEP;

	/*
	 * If vfork make chain from parent process to child
	 * (where virtal memory is temporarily).  Wait for
	 * child to finish, steal virtual memory back,
	 * and wakeup child to let it die.
	 */
	if (isvfork) {
		u.u_procp->p_xlink = rpp;
		u.u_procp->p_flag |= SNOVM;
		while (rpp->p_flag & SVFORK)
			slumber((caddr_t)rpp, 0, resboost);
		if ((rpp->p_flag & SLOAD) == 0)
			panic("newproc vfork");
		uaccess(rpp, Vfmap, &vfutl);
		u.u_procp->p_xlink = 0;
		vpassvm(rpp, u.u_procp, &vfutl, &u, Vfmap);
		u.u_procp->p_flag &= ~SNOVM;
		rpp->p_ndx = rpp - proc;
		rpp->p_flag |= SVFDONE;
		wakeup((caddr_t)rpp);
	}

	/*
	 * 0 return means parent.
	 */
	return (0);
}
