/*	kern_synch.c	6.1	83/07/29	*/

#include "../ml/pte.h"
#include "../ml/psl.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/file.h"
#include "../sys/vnode.h"
#include "../sys/vm.h"
#include "../sys/kernel.h"
#include "../sys/buf.h"
#include "../sys/sysinfo.h"
#include "../sys/tty.h"

/*
 * Do operations which need to be done once a second.
 */
schedcpu()
{
#ifndef HCXRT
	register struct proc *p;
	register int s;
	register int temp;
#endif

	wakeup((caddr_t)&lbolt);

#ifndef HCXRT

	/*
	 * Every second this routine recalculates priorities.
	 *
	 * Processes gradually increase in priority up to their base.
	 * When they get cpu cycles this routine notices them and their
	 * priority is penalized.
	 *
	 * Boosts are also added into the priority, but they are decremented
	 * every context switch and so are of temporary duration.
	 *
	 * Of course, you are always charged for your cpu time so boosted
	 * processes will have a tendancy to get severly whomped if they
	 * don't go back to sleep.
	 *
	 * To prevent deadlocks, since the sleep/wakeup mechanism can't
	 * pass time slices, processes sleeping in the kernel will have
	 * a priority higher than anyone else while in the kernel.
	 *
	 * The end result is that nice'd processes recieve a lower percentage
	 * of the cpu.  Processes waiting for terminal I/O with little
	 * processing will be heavily favored.  Processes using less cpu
	 * will be favored over those processes using more cpu.
	 */

	for (p = allproc; p != NULL; p = p->p_nxt) {

		/*
		 * Don't calculate whomp values for
		 * processes that don't use them.
		 */
		if (p->p_basepri < min_whomp_pri)
			continue;

		/* 
		 * Forget a small part (32th) of the cpu ticks recorded while
		 * the process was running.  The minus one is for values
		 * under 32.
		 */
		temp = (long)p->p_cpu - 1;
		temp -= (temp >> 5);

		/*
		 * If the process hasn't run in the last five seconds,
		 * just forget some of the ticks.  In otherwords have
		 * some pity on those processes that haven't run in the
		 * last five seconds.
		 *
		 * Otherwise add the nice value (-20 to 19) to the ticks.
		 * This will penalizes nice'd processes since the nice
		 * value may be greater than the actual cpu ticks recorded
		 * in the last second.
		 * 
		 */
		if (p->p_cpticks) {
			temp += p->p_nice - NZERO;
			if (temp < 0)
				temp = 0;
		}

		/* 
		 * Put a bound (0 to 255) on the cpu ticks.
		 */

		if (temp > 255)
			temp = 255;

		p->p_cpu = temp;

		/*
		 * The whomp of the base priority is in values from 0 to 15.
		 */
		p->p_whomp = temp >> 4;

		/*
		 * Adjust p_usrpri.
		 */

		SETUSRPRI(p)

		/* Providing we are not in the kernel (or otherwise 
		   less than the minimum whomp priority, adjust the
		   priority if it has changed.  */

		if ((p->p_prior > min_whomp_pri) &&
		    (p->p_prior != p->p_usrpri) ) {
			s = spl8();

			/* If it's on the run queue the priority
			   affects what queue it's on. */

			if ((p != u.u_procp || noproc) &&
			    (p->p_stat == SRUN) &&
			    (p->p_flag & SLOAD)) {
				remrq(p);
				p->p_prior = p->p_usrpri;
				setrq(p);
			}
			else
				p->p_prior = p->p_usrpri;
			splx(s);
		}
	}
#endif

	vmmeter();
	if (runin!=0) {
		runin = 0;
		wakeup((caddr_t)&runin);
	}
	if (bclnlist != NULL)
		wakeup((caddr_t)&proc[2]);
	timeout(schedcpu, (caddr_t)0, hz);
}

#define SQSIZE 0100	/* Must be power of 2 */
#define HASH(x)	(( (int) x >> 5) & (SQSIZE-1))
struct proc *slpque[SQSIZE];

/*
 * Give up the processor till a wakeup occurs on chan,
 * at which time the process reenters the scheduling queue.
 * Callers of this routine must be prepared for
 * premature return, and check that the reason for
 * slumbering has gone away.
 */
#ifdef HCXRT
doslumber(chan, flags)
	register caddr_t chan;
	int flags;
#else
doslumber(chan, flags, boost)
	register caddr_t chan;
	int flags, boost;
#endif
{
	register struct proc *rp, **hp;
	register s;

	rp = u.u_procp;

	s = spl8();

	rp->p_eslptime = time.tv_sec;

	if (panicstr) {
		/*
		 * After a panic, just give interrupts a chance,
		 * then just return; don't run any other procs 
		 * or panic below, in case this is the idle process
		 * and already asleep.
		 * The splnet should be spl0 if the network was being used
		 * by the filesystem, but for now avoid network interrupts
		 * that might cause another panic.
		 */
		(void) splnet();
		splx(s);
		return(0);
	}
#ifndef HCXRT
	if (chan==0 || rp->p_stat != SRUN)
		panic("doslumber");
#endif

	hp = &slpque[HASH(chan)];
	rp->p_link = *hp;

	rp->p_wchan = chan;
	*hp = rp;
	if (rp->p_interruptable = (flags & INTERRUPTABLE)) {
		/*
		 * If we stop in issig(), wakeup may already have happened
		 * when we return (rp->p_wchan will then be 0).
		 */
		if (ISSIG(rp)) {
			if (rp->p_wchan)
				unsleep(rp);
			rp->p_stat = SRUN;
			splx(s);
			if (flags & PCATCH) {
				return(1);
			}
			longjmp(&u.u_qsave);
			/*NOTREACHED*/
		}
		if (rp->p_wchan == 0) {
			splx(s);
			return(0);
		}
		rp->p_stat = SSLEEP;
#ifndef HCXRT
		BOOST(rp, boost);
#endif
		u.u_ru.ru_nvcsw++;
		swtch();
		splx(s);
		if (ISSIG(rp)) {
			if (flags & PCATCH) {
				return(1);
			}
			longjmp(&u.u_qsave);
			/*NOTREACHED*/
		} else {
			return(0);
		}
	} else {
		rp->p_stat = SSLEEP;
#ifndef HCXRT
		BOOST(rp, boost);
#endif
		u.u_ru.ru_nvcsw++;
		swtch();
		splx(s);
		return(0);
	}
	/*NOTREACHED*/
}

/*
 * Remove a process from its wait queue
 * Must be called at spl8, and p->p_wchan must not be NULL.
 */
unsleep(p)
	register struct proc *p;
{
	register struct proc **hp;

	hp = &slpque[HASH(p->p_wchan)];
	while (*hp != p)
		hp = &(*hp)->p_link;
	*hp = p->p_link;
	p->p_wchan = 0;
}

/*
 * Wake up all processes sleeping (slumbering) on chan.
 */
wakeup(chan)
	register caddr_t chan;
{
	register struct proc *p, **q;
	int s;

	s = spl8();
restart:
	for (q = &slpque[HASH(chan)]; p = *q; ) {
		if (p->p_wchan==chan) {
			p->p_wchan = 0;
			*q = p->p_link;
			p->p_eslptime = 0;
			if (p->p_stat == SSLEEP) {
				/* OPTIMIZED INLINE EXPANSION OF setrun(p) */
				p->p_stat = SRUN;
				if (p->p_flag & SLOAD) {
					setrq(p);
					if (p->p_prior < u.u_procp->p_prior) {
						runrun++;
						aston();
					}
				} else {
					if (p->p_prior < u.u_procp->p_prior) {
						runrun++;
						aston();
					}
					if (runout != 0) {
						runout = 0;
						wakeup((caddr_t)&runout);
					}
					wantin++;
					goto restart;	/* because of wakeup */
				}
				/* END INLINE EXPANSION */
			}
		} else
			q = &p->p_link;
	}
	splx(s);
}

/*
 * Initialize the (doubly-linked) run queues
 * to be empty.
 */
rqinit()
{
	register int i;

	for (i = 0; i < NQS; i++)
		qs[i].ph_link = qs[i].ph_rlink = (struct proc *)&qs[i];
}

/*
 * Set the process running;
 * arrange for it to be swapped in if necessary.
 * Must be called at spl8.
 */
setrun(p)
	register struct proc *p;
{

	switch (p->p_stat) {

	case 0:
	case SWAIT:
	case SRUN:
	case SZOMB:
	default:
		panic("setrun");

	case SSTOP:
	case SSLEEP:
		if (p->p_wchan)
			unsleep(p);		/* e.g. when sending signals */
		break;

	case SIDL:
		break;
	}
	p->p_stat = SRUN;
	if (p->p_flag & SLOAD) {
		setrq(p);
		if (p->p_prior < u.u_procp->p_prior) {
			runrun++;
			aston();
		}
	} else {
		if (p->p_prior < u.u_procp->p_prior) {
			runrun++;
			aston();
		}
		if (runout != 0) {
			runout = 0;
			wakeup((caddr_t)&runout);
		}
		wantin++;
	}
}
