/*	@(#)raw.c 1.1 86/11/30	*/

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/file.h"
#include "../sys/proc.h"
#include "../sys/uio.h"
#include "../sys/kernel.h"
#include "../sys/vfs.h"
#include "../sys/vnode.h"
#include "../sys/conf.h"
#include "../specfs/snode.h"
#include "../ml/clock.h"
#include "../sys/tty.h"

int raw_rw();
int raw_ioctl();
int raw_select();
int vno_close(); /* Raw devices are really vnode devices anyway */
int raw_asyncrw();
int raw_acancel();
extern int nodev();

struct fileops rawfops = {
	raw_rw,
	raw_ioctl,
	raw_select,
	vno_close,
	raw_asyncrw,
	raw_acancel
};

int
raw_rw(fp, rw, uio)
	register struct file *fp;
	enum uio_rw rw;
	register struct uio *uio;
{
	register struct snode *sp;
	register int dev;
	register struct tty *tp;
	register int mj;
	extern int mem_no;

	sp = VTOS((struct vnode *)fp->f_data);

	dev = (int)(dev_t)sp->s_dev;
	mj = major(dev);

	if (rw == UIO_READ) {
		if ( uio->uio_resid == 0)
			return (0);
		if ( (uio->uio_offset < 0) && (mem_no != mj) )
			return (EINVAL);
		/*
		 * The following is used instead of smark() since
		 * less accuracy is needed.
		 */
		sp->s_flag |= SACC;
		sp->s_atime.tv_sec = time.tv_sec;
		sp->s_atime.tv_usec = 0;
		if (tp = cdevsw[mj].d_ttys) {
			tp += minor(dev);
			return((*linesw[tp->t_line].l_read)(tp, uio));
		}
		else 
			return((*cdevsw[mj].d_read)(dev, uio));
	}
	else {
		if ( (uio->uio_offset < 0) && (mem_no != mj) )
			return (EINVAL);
		/*
		 * The following is used instead of smark(), since
		 * less accuracy is needed.
		 */
		sp->s_flag |= SUPD|SCHG;
		sp->s_ctime.tv_sec = sp->s_mtime.tv_sec = time.tv_sec;
		sp->s_ctime.tv_usec = sp->s_mtime.tv_usec = 0;
		if (tp = cdevsw[mj].d_ttys) {
			tp += minor(dev);
			return((*linesw[tp->t_line].l_write)(tp, uio));
		}
		else
			return((*cdevsw[mj].d_write)(dev, uio));
	}
}

int
raw_ioctl(fp, com, data)
	struct file *fp;
	int com;
	caddr_t data;
{
	register struct snode *sp;
	register int dev;
	int error;
	label_t lqsave;

	sp = VTOS((struct vnode *)fp->f_data);
	dev = (int)(dev_t)sp->s_dev;

	u.u_r.r_val1 = 0;
	if ((u.u_procp->p_flag&SOUSIG) == 0) {
		lqsave = u.u_qsave;
		if (setjmp(&u.u_qsave)) {
			error = ERESTARTSYS;
		} else {
			error = (*cdevsw[major(dev)].d_ioctl)
				(dev, com, data, fp->f_flag);
		}
		u.u_qsave = lqsave;
		return (error);
	} else
		return ((*cdevsw[major(dev)].d_ioctl)
			(dev, com, data, fp->f_flag));
}

int
raw_select(fp, flag)
	struct file *fp;
	int flag;
{
	register struct snode *sp;
	register int dev;

	sp = VTOS((struct vnode *)fp->f_data);
	dev = (int)(dev_t)sp->s_dev;
	return ((*cdevsw[major(dev)].d_select)(dev, flag));
}

int
raw_asyncrw(areq)
	register struct areq *areq;
{
	register struct snode *sp;
	extern int mem_no;
	register int mj;

	sp = VTOS((struct vnode *)areq->fp->f_data);
	areq->dev = sp->s_dev;
	mj = major(areq->dev);

	if (areq->rw == AIO_READ) {
		if (areq->auio.uio_resid == 0)
			return (0);
		if ((areq->auio.uio_offset < 0) && (mem_no != mj))
			return (EINVAL);
		/*
		 * The following is used instead of smark() since
		 * less accuracy is needed.
		 */
		sp->s_flag |= SACC;
		sp->s_atime.tv_sec = time.tv_sec;
		sp->s_atime.tv_usec = 0;
	} else {
		if ((areq->auio.uio_offset < 0) && (mem_no != mj))
			return (EINVAL);
		/*
		 * The following is used instead of smark(), since
		 * less accuracy is needed.
		 */
		sp->s_flag |= SUPD|SCHG;
		sp->s_ctime.tv_sec = sp->s_mtime.tv_sec = time.tv_sec;
		sp->s_ctime.tv_usec = sp->s_mtime.tv_usec = 0;
	}
	return ((*cdevsw[mj].d_asyncio)(areq));
}

/* raw cancel asynchronous I/O routine */
int
raw_acancel(areq)
	struct areq *areq;
{
	register int mj;

	mj = major(areq->dev);
	if (cdevsw[mj].d_acancel == nodev) {
		return (ENODEV);
	}
	return ((*cdevsw[mj].d_acancel)(areq));
}
