/*	subr_prf.c	6.1	83/07/29	*/
/*  Change history :
 *  bernie 4/15/85  New subroutine eprinf for audit trail
 */

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/seg.h"
#include "../sys/buf.h"
#include "../sys/conf.h"
#include "../sys/reboot.h"
#include "../sys/vm.h"
#include "../sys/msgbuf.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/tty.h"
#include "../sys/prf.h"

#include "../ml/mtpr.h"

/*
 * In case console is off,
 * panicstr contains argument to last
 * call to panic.
 */
char	*panicstr;

/*
 * Scaled down version of C Library printf.
 * Used to print diagnostic information directly on console tty.
 * Since it is not interrupt driven, all system activities are
 * suspended.  Printf should not be used for chit-chat.
 *
 * One additional format: %b is supported to decode error registers.
 * Usage is:
 *	printf("reg=%b\n", regval, "<base><arg>*");
 * Where <base> is the output base expressed as a control character,
 * e.g. \10 gives octal; \20 gives hex.  Each arg is a sequence of
 * characters, the first of which gives the bit number to be inspected
 * (origin 1), and the next characters (up to a control character, i.e.
 * a character <= 32), give the name of the register.  Thus
 *	printf("reg=%b\n", 3, "\10\2BITTWO\1BITONE\n");
 * would produce output:
 *	reg=2<BITTWO,BITONE>
 */
/*VARARGS1*/
printf(fmt, x1)
	char *fmt;
	unsigned x1;
{
	prf(fmt, &x1, 0);
}

/*
 * Uprintf prints to the current user's terminal,
 * guarantees not to sleep (so can be called by interrupt routines)
 * and does no watermark checking - (so no verbose messages).
 */
/*VARARGS1*/
uprintf(fmt, x1)
	char *fmt;
	unsigned x1;
{

	prf(fmt, &x1, 2);
}

 /*VARARGS1*/
 eprintf(fmt, x1)
 	char *fmt;
 	unsigned x1;
 {
 
 	prf(fmt, &x1, CONSOLE);

}
prf(fmt, adx, touser)
	register char *fmt;
	register u_int *adx;
{
	register int b, c, i;
	char *s;
	int any;

loop:
	while ((c = *fmt++) != '%') {
		if(c == '\0')
			return;
		putchar(c, touser);
	}
again:
	c = *fmt++;
	/* THIS CODE IS MACHINE DEPENDENT IN HANDLING %l? AND %c */
	switch (c) {

	case 'l':
		goto again;
	case 'x': case 'X':
		b = 16;
		goto number;
	case 'd': case 'D':
		b = -10;
		goto number;
	case 'u':
		b = 10;
		goto number;
	case 'o': case 'O':
		b = 8;
number:
		printn((u_long)*adx, b, touser);
		break;
	case 'c':
		b = *adx;
		if (c = (b & 0x7f))
			putchar(c, touser);
		break;
	case 'b':
		b = *adx++;
		s = (char *)*adx;
		printn((u_long)b, *s++, touser);
		any = 0;
		if (b) {
			putchar('<', touser);
			while (i = *s++) {
				if (b & (1 << (i-1))) {
					if (any)
						putchar(',', touser);
					any = 1;
					for (; (c = *s) > 32; s++)
						putchar(c, touser);
				} else
					for (; *s > 32; s++)
						;
			}
			if (any)
				putchar('>', touser);
		}
		break;

	case 's':
		s = (char *)*adx;
		while (c = *s++)
			putchar(c, touser);
		break;

	case '%':
		putchar('%', touser);
		break;
	}
	adx++;
	goto loop;
}

/*
 * Printn prints a number n in base b.
 * We don't use recursion to avoid deep kernel stacks.
 */
printn(n, b, touser)
	register u_long n;
	register b;
{
	char prbuf[11];
	register char *cp;

	if (b == -10) {
		if((int)n < 0) {
			putchar('-', touser);
			n = (unsigned)(-(int)n);
		}
		b = -b;
	}
	cp = prbuf;
	do {
		*cp++ = "0123456789abcdef"[n%b];
		n /= b;
	} while (n);
	do
		putchar(*--cp, touser);
	while (cp > prbuf);
}

int panic_opt = RB_AUTOBOOT;
/*
 * Panic is called on unresolvable fatal errors.
 * It saves registers, prints "panic: mesg", and then reboots.
 * If we are called twice, then we avoid trying to
 * sync the disks as this often leads to recursive panics.
 */
panic(s)
	char *s;
{

        savreg();
	if (panicstr)
		panic_opt |= RB_NOSYNC;
	else {
		panicstr = s;
	}
	printf("panic: %s\n", s);
	boot(RB_PANIC, panic_opt);
}

/*
 * Warn that a system table is full.
 */
tablefull(tab)
	char *tab;
{

	printf("%s: table is full\n", tab);
}

/*
 * Hard error is the preface to plaintive error messages
 * about failing disk transfers.
 */
harderr(bp, cp)
	struct buf *bp;
	char *cp;
{

	printf("%s%x%c: error sn%d ", cp,
	    dkunit(bp), 'a'+(minor(bp->b_dev)&07), bp->b_blkno);
}

/*
 * Print a character on console or users terminal.
 */
/*ARGSUSED*/
putchar(c, touser)
	register int c, touser;
{
	register struct tty *tp = 0;
	register s = spl8();

	if (touser && (tp = u.u_ttyp) == NULL)
		touser = 0;

	if (touser) {
		if (tp && (tp->t_state&CARR_ON)) {
			if (c == '\n')
				(void) ttxput(tp, '\r', 0);
			(void) ttxput(tp,c,0);
			(*tp->t_proc)(tp, T_OUTPUT);
		}
		splx(s);
		return;
	}

	if (c == 0) {
		splx(s);
		return;
	}

loop:
	if ( c != 0177 && mfpr(MME) ) {
		if (msgbuff.msg_magic != MSG_MAGIC) {
			register int i;

			msgbuff.msg_bufx = 0;
			msgbuff.msg_bufw = 0;
			msgbuff.msg_magic = MSG_MAGIC;
			for (i=0; i < MSG_BSIZE; i++)
				msgbuff.msg_bufc[i] = 0;
		}
		msgbuff.msg_bufc[msgbuff.msg_bufx++] = c;
		if (msgbuff.msg_bufx >= MSG_BSIZE)
			msgbuff.msg_bufx = 0;
		/* If we've overflown the buffer, throw very old chars away */
		if (msgbuff.msg_bufx == msgbuff.msg_bufw) {
			msgbuff.msg_bufw ++ ;
			if (msgbuff.msg_bufw >= MSG_BSIZE)
				msgbuff.msg_bufw = 0;
		}

		if (c == '\n') {
			c = '\r';
			goto loop ;
		}
	}

	cnputc(0, 0);

	splx(s);
}
