#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/buf.h"
#include "../sys/conf.h"
#include "../sys/map.h"
#include "../sys/time.h"
#include "../sys/kernel.h"
#include "../sys/utsname.h"
#include "../sys/err.h"
#include "../sys/elog.h"
#include "../sys/erec.h"
#include "../sys/iobuf.h"

#include "../ml/mem.h"
#include "../ml/mtpr.h"

extern short cputype;

struct	err	err = {
	NESLOT,
};

int	blkacty;

errinit()
{
	if (err.e_nslot)
		rminit(err.e_map, err.e_nslot, 1, "errmap", (err.e_nslot+3)/2);
	err.e_org = err.e_ptrs;
	err.e_nxt = err.e_ptrs;
}

struct errhdr *
geteslot(size)
{
	register ns;
	register struct errhdr *ep;
	int n, sps;

	size += sizeof(struct errhdr);
	ns = (size + sizeof(struct errslot) - 1)/sizeof(struct errslot);
	sps = spl8();
	n = rmalloc(err.e_map, ns);
	splx(sps);
	if (n == 0)
		return(NULL);
	ep = (struct errhdr *)(&err.e_slot[--n]);
	bzero((caddr_t)ep, ns*sizeof(struct errslot));
	ep->e_len = size;
	return(++ep);
}

freeslot(ep)
	register struct errhdr *ep;
{
	register ns, sps;

	ns = (ep->e_len+sizeof(struct errslot)-1)/sizeof(struct errslot);
	sps = spl8();
	rmfree(err.e_map, ns, (((struct errslot *)ep)-err.e_slot)+1);
	splx(sps);
}

struct errhdr *
geterec()
{
	register sps;
	register struct errhdr *ep;

	sps = spl8();
	while(*err.e_org == NULL) {
		err.e_flag |= E_SLP;
		slumber(&err.e_org, INTERRUPTABLE, resboost);
	}
	ep = *err.e_org;
	*err.e_org++ = NULL;
	if (err.e_org >= &err.e_ptrs[err.e_nslot])
		err.e_org = err.e_ptrs;
	splx(sps);
	return(ep);
}

puterec(ep, type)
	register struct errhdr *ep;
{
	register sps;

	(--ep)->e_type = type;
	ep->e_time = (time_t) time.tv_sec;
	sps = spl8();
	*err.e_nxt++ = ep;
	if (err.e_nxt >= &err.e_ptrs[err.e_nslot])
		err.e_nxt = err.e_ptrs;
	splx(sps);
	if (err.e_flag&E_SLP) {
		err.e_flag &= ~E_SLP;
		wakeup(&err.e_org);
	}
}

logstart()
{
	register int sps;

	sps = spl8();
	for(err.e_org = &err.e_ptrs[err.e_nslot-1]; err.e_org >= err.e_ptrs; err.e_org--)
		if (*err.e_org != NULL) {
			freeslot(*err.e_org);
			*err.e_org = NULL;
		}
	err.e_org = err.e_ptrs;
	err.e_nxt = err.e_ptrs;

	{
	register struct estart *ep;

	ep = (struct estart *)geteslot(sizeof(struct estart));
	splx(sps);
	if (ep == NULL)
		return;
	ep->e_cpu = cputype;
	ep->e_name = utsname;
	puterec(ep, E_GOTS);
	}
}

logtchg(nt)
	time_t nt;
{
	register struct etimchg *ep;

	if ((ep = (struct etimchg *)geteslot(sizeof(struct etimchg))) != NULL) {
		ep->e_ntime = nt;
		puterec(ep, E_TCHG);
	}
}

logstray(ipl, addr)
	int ipl;
	physadr addr;
{
	register struct estray *ep;

	if ((ep = (struct estray *)geteslot(sizeof(struct estray))) != NULL) {
		ep->e_ipl = ipl;
		ep->e_sbacty = blkacty;
		ep->e_saddr = addr;
		puterec(ep, E_STRAY);
	}
}

logmemory(memerreg, bus)
	int	memerreg;
	int	bus;
{
	register struct ememory *ep;

	if ((ep = (struct ememory *)geteslot(sizeof(struct ememory))) != NULL) {
		ep->e_bus = bus;
		ep->e_memerreg = memerreg;
		puterec(ep, E_PRTY);
	}
}


fmtberr(dp, cyl, addr)
	register struct iobuf *dp;
	u_int	cyl, *addr;
{
	register struct eblock *ep;
	register struct buf *bp;
	register n;
	register u_int *p;
	struct br {
		struct eblock	eb;
		u_int	cregs[1];
	};
	struct	iostat	*iosp;

	if (dp->io_erec != NULL) {
		dp->io_erec->e_rtry++;
		return;
	}
	if ((iosp = dp->io_stp) == NULL)
		return;

	if (addr == NULL || (ep = (struct eblock *)geteslot(sizeof(struct eblock)+
		(dp->io_nreg*sizeof(u_int)))) == NULL) {
		iosp->io_unlog++;
		return;
	}
	n = major(dp->b_dev);
	bp = dp->b_actf;
	ep->e_dev = makedev(n, (bp==NULL)?minor(dp->b_dev):minor(bp->b_dev));
	ep->e_regloc = dp->io_addr;
	ep->e_bacty = (short)blkacty;
	ep->e_stats.io_ops = iosp->io_ops;
	ep->e_stats.io_misc = iosp->io_misc;
	ep->e_stats.io_unlog = iosp->io_unlog;
	if (bp != NULL) {
		ep->e_bflags = (bp->b_flags&B_READ) ? E_READ : E_WRITE;
		if (bp->b_flags & B_PHYS)
			ep->e_bflags |= E_PHYS;
		ep->e_bnum = bp->b_blkno;
		ep->e_bytes = bp->b_bcount;
		ep->e_memadd = bp->b_un.b_addr;
	}
	else
		ep->e_bflags = E_NOIO;
	ep->e_cyloff = (short)cyl;
	ep->e_nreg = (short)dp->io_nreg;
	p = (u_int *)(&((struct br *)ep)->cregs[0]);

	for (n = dp->io_nreg; n; n--)
		*p++ = *addr++;

	dp->io_erec = ep;
}

logberr(dp, erf)
	register struct iobuf *dp;
{
	register struct eblock *ep;

	if ((ep = dp->io_erec) == NULL)
		return;
	if (erf)
		ep->e_bflags |= E_ERROR;
	puterec(ep, E_BLK);
	dp->io_erec = NULL;
}
