/* @(#)	sys_lock.c	5/14/87 */

#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/dir.h"
#include "../sys/user.h"
#include "../sys/vm.h"
#include "../sys/proc.h"
#include "../sys/text.h"
#include "../sys/cmap.h"
#include "../sys/errno.h"
#include "../sys/lock.h"
#include "../sys/kernel.h"

lock()
{
	struct a {
		long oper;
	};
	
	register struct proc *p = u.u_procp;
		
	switch (((struct a *)u.u_ap)->oper) {

/* Lock text segment */
	case TXTLOCK:
		if (p->p_lock & (PROCLOCK | TXTLOCK))
			goto bad;
		textlock();
		break;

/* Lock data segment */
	case DATLOCK:
 		if (p->p_lock & (PROCLOCK | DATLOCK))
			goto bad;
		datlock();
		break;

/* Lock process */
	case PROCLOCK:
		if (p->p_lock & (PROCLOCK | TXTLOCK | DATLOCK))
			goto bad;
		
		if (textlock() == 0)
			return;

		if (datlock() == 0) {
			tunlock();		/* couldn't 've been locked */
			return;
		}
		
		p->p_lock |= PROCLOCK;
		break;

/* Unlock */
	case UNLOCK:
		if ((p->p_lock & (PROCLOCK | TXTLOCK | DATLOCK)) == 0)
			goto bad;

		punlock();	
		break;

	default:
bad:
		u.u_error = EINVAL;
	}
}

textlock()
{
			int	limit;
	register struct proc 	*p = u.u_procp;
	register struct pte 	*pte;

	if (p->p_textp == NULL)			/* firewall */
		return(0);

	if (p->p_textp->x_lcount == 0) {	/* if no other process has
						   locked this segment*/
		limit = (u.u_uid == 0) ? s_limit : u_limit;
		/*
		** check against limits, update
		*/
		if ((plocked += p->p_tsize) > limit) {
			plocked -= p->p_tsize;
	        	u.u_error = ENOMEM;
	        	return(0);
		}
		p->p_textp->x_lcount++;
		
		/*
		** pagein and set the c_noswap bit in the core map entries 
		** associated with all the text pages for this process
		*/
		for (pte = tptopte(p, 0); pte < tptopte(p, p->p_tsize);
		     pte += CLSIZE) {
			while ( ! pte->pg_v )
	                	pagein(ctob(tptov(p, ptetotp(p, pte))), 0);
			cmap[pgtocm(pte->pg_pfnum)].c_noswap = 1;
		}
	}
	else
		p->p_textp->x_lcount++;		/* just incr locked count */

	p->p_lock |= TXTLOCK;			/* mark proc as text-locked */
	p->p_flag |= SPLOCK;			/* dont swap */
	return(1);
}

tunlock()
{
	register struct proc *p = u.u_procp;
	register struct pte 	*pte;

	if (p->p_textp == NULL)			/* firewall */
		return;
	/*
	** If no other process has locked this text segment,then
	** unlock the text and reset the c_noswap bit in the core map
	** entries associated with all currently valid text pages
	*/
	if (--(p->p_textp->x_lcount) == 0) {
		for (pte = tptopte(p, 0); pte < tptopte(p, p->p_tsize); 
		     pte += CLSIZE)
			cmap[pgtocm(pte->pg_pfnum)].c_noswap = 0;
		plocked -= p->p_tsize;		/* update locked count */
	}
	p->p_lock &= ~TXTLOCK;
	return;
}

/*
** lock down the data and stack pages for the currently
** running process
*/
datlock()
{
			int	limit;
	register struct proc *p = u.u_procp;
	register struct pte 	*pte;
			
	limit = (u.u_uid == 0) ? s_limit : u_limit;
			
	/*
	** check if locking these pages will exceed limit
	** update plocked
	*/
	if ((plocked += (p->p_dsize + p->p_ssize)) > limit) {
		plocked -= (p->p_dsize + p->p_ssize);
        	u.u_error = ENOMEM;
        	return(0);
	}
			
	/*
	** Set the c_noswap bit in the core map entries associated with all 
	** current data and stack pages for this process
	*/
	for (pte = dptopte(p, 0); pte < dptopte(p, p->p_dsize); 
	     pte += CLSIZE) {
		     while ( ! pte->pg_v )
			     pagein(ctob(dptov(p, ptetodp(p, pte))), 0);
		     cmap[pgtocm(pte->pg_pfnum)].c_noswap = 1;
	     }

	for (pte = sptopte(p, (p->p_ssize-1)); pte < sptopte(p, 0);
	     pte += CLSIZE) {
		     while ( ! pte->pg_v )
			     pagein(ctob(sptov(p, ptetosp(p, pte))), 0);
		     cmap[pgtocm(pte->pg_pfnum)].c_noswap = 1;
	     }
			
	p->p_lock |= DATLOCK;
	p->p_flag |= SPLOCK;
}

dunlock()
{
	register struct proc *p = u.u_procp;
	register struct pte 	*pte;

	/*
	** Reset the c_noswap bit in the core map entries associated with all 
	** current data and stack pages for this process
	*/
	for (pte = dptopte(p, 0); pte < dptopte(p, p->p_dsize); pte += CLSIZE)
		cmap[pgtocm(pte->pg_pfnum)].c_noswap = 0;

	for (pte = sptopte(p, (p->p_ssize-1)); pte < sptopte(p, 0); pte += CLSIZE)
		cmap[pgtocm(pte->pg_pfnum)].c_noswap = 0;
			
	plocked -= (p->p_dsize + p->p_ssize);		/* adj # pages locked */
	p->p_lock &= ~DATLOCK;
	return;
}

/*
** punlock - unlock any locked segments (excluding locked shared memory segs)
**           an reset SPLOCK bit if needed
*/
punlock()
{
	register struct proc	*p = u.u_procp;
	register struct shm_pt	*sptp;

	if (p->p_lock & TXTLOCK)		/* unlock text */
	   	tunlock();
	if (p->p_lock & DATLOCK)		/* unlock data */
	   	dunlock();
	p->p_lock &= ~PROCLOCK;			/* turn off PROLOCK flag */
	
	/*
	** Check if we should turn off the SPLOCK flag
	*/
	if (p->p_smbeg)
		for(sptp = p->p_shmpt;
		    sptp < &(p->p_shmpt[shminfo.shmseg]);
		    sptp++)
			if (sptp->shm_shmp &&
			    (sptp->shm_sflg & SHM_MLOCK))
				return;			/* return here */
	p->p_flag &= ~SPLOCK;
}
