/*	sys_process.c	6.1	83/07/29	*/

#include "../ml/reg.h"
#include "../ml/psl.h"
#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/kernel.h"
#include "../sys/vnode.h"
#include "../sys/text.h"
#include "../sys/seg.h"
#include "../sys/vm.h"
#include "../sys/buf.h"
#include "../sys/acct.h"

/*
 * Tracing variables.
 * Used to pass trace command from
 * parent to child being traced.
 * This data base cannot be
 * shared and is locked
 * per user.
 */
struct {
	int	ip_lock;
	int	ip_req;
	int	*ip_addr;
	int	ip_data;
} ipc;

/*
 * sys-trace system call.
 */
ptrace()
{
	register struct proc *p;
	register struct a {
		int	req;
		int	pid;
		int	*addr;
		int	data;
	} *uap;

	uap = (struct a *)u.u_ap;
	if (uap->req <= 0) {
		u.u_procp->p_flag |= STRC;
		return;
	}
	p = pfind(uap->pid);
	if (p == 0 || p->p_stat != SSTOP || p->p_ppid != u.u_procp->p_pid ||
	    !(p->p_flag & STRC)) {
		u.u_error = ESRCH;
		return;
	}
	while (ipc.ip_lock)
		slumber((caddr_t)&ipc, 0, proccomboost);
	ipc.ip_lock = p->p_pid;
	ipc.ip_data = uap->data;
	ipc.ip_addr = uap->addr;
	ipc.ip_req = uap->req;
	p->p_flag &= ~SWTED;
	while (ipc.ip_req > 0) {
		if (p->p_stat==SSTOP) {
			spl8();
			setrun(p);
			spl0();
		}
		slumber((caddr_t)&ipc, 0, proccomboost);
	}
	u.u_r.r_val1 = ipc.ip_data;
	if (ipc.ip_req < 0)
		u.u_error = EIO;
	ipc.ip_lock = 0;
	wakeup((caddr_t)&ipc);
}

#define	NIPCREG 18
int ipcreg[NIPCREG] =
	{R0,R1,R2,R3,R4,R5,R6,R7,R8,R9,R10,R11,R12,FP,SP,PC,RALST,RAMST};

#define	PHYSOFF(p, o) \
	((physadr)(p)+((o)/sizeof(((physadr)0)->r[0])))

/*
 * Code that the child process
 * executes to implement the command
 * of the parent process in tracing.
 */
procxmt()
{
	register int i;
	register *p;
	register struct text *xp;

	if (ipc.ip_lock != u.u_procp->p_pid)
		return (0);
	u.u_procp->p_eslptime = 0;
	i = ipc.ip_req;
	ipc.ip_req = 0;
	if ( u.u_procp -> p_textp == 0 )
		i = -1 ;		/* Invalid - force correct exit out */
	u.u_ar0[SP] = mfpr(USP);
	if (u.u_ar0[PS] & PSL_DBL)
		save_d_acc(&u.u_ar0[RAMST]);
	else
		save_s_acc(&u.u_ar0[RAMST]);
	switch (i) {

	/* read user I */
	case 1:
	/* read user D */
	case 2:
		if (!useracc((caddr_t)ipc.ip_addr, 4, B_READ))
			goto error;
		copyin ((caddr_t)ipc.ip_addr, (caddr_t)&ipc.ip_data, 4);
		break;

	/* read u */
	case 3:
		i = (int)ipc.ip_addr;
		if (i<0 || i >= ctob(UPAGES))
			goto error;
		ipc.ip_data = *(int *)PHYSOFF(&u, i);
		break;

	/* write user I */
	/* Must set up to allow writing */
	case 4:
		/*
		 * If text, must assure exclusive use
		 */
		if (xp = u.u_procp->p_textp) {
			struct vattr vattr;
			struct vnode *vp;

			vp = xp->x_vptr;
			VOP_GETATTR(vp, &vattr, u.u_cred);
			if (xp->x_count != 1 || (vattr.va_mode & VSVTX))
				goto error;
			vp->v_flag |= VTEXTMOD;
		}
		i = -1;
		if (!useracc((caddr_t)ipc.ip_addr, 4, B_WRITE)) {
			if (chgprot((caddr_t)ipc.ip_addr, RW) &&
			    chgprot((caddr_t)ipc.ip_addr+(sizeof(int)-1), RW))
				i = copyout((caddr_t)&ipc.ip_data, ipc.ip_addr, 4);
			(void) chgprot((caddr_t)ipc.ip_addr, RO);
			(void) chgprot((caddr_t)ipc.ip_addr+(sizeof(int)-1), RO);
		} else
			i = copyout((caddr_t)&ipc.ip_data, ipc.ip_addr, 4);
		if (i != 0)
			goto error;
		/* make sure the old value is not in cache */
		ckeyrelease(u.u_procp);
		u.u_procp->p_ckey = getcodekey();
		if (xp) {
			xp->x_ckey = u.u_procp->p_ckey;
			xp->x_flag |= XWRIT;
		}
		break;

	/* write user D */
	case 5:
		if (copyout((caddr_t)&ipc.ip_data, ipc.ip_addr, 4) != 0)
			goto error;
		break;

	/* write u */
	case 6:
		i = (int)ipc.ip_addr;
		p = (int *)PHYSOFF(&u, i);
		for (i=0; i<NIPCREG; i++)
			if (p == &u.u_ar0[ipcreg[i]])
				goto ok;
		if ((p == &u.u_ar0[RAMST]) && (ipc.ip_data == 0x80000000))
			goto error;
		if (p == &u.u_ar0[PS]) {
			ipc.ip_data |= PSL_USERSET;
			ipc.ip_data &=  ~PSL_USERCLR;
			goto ok;
		}
		goto error;

	ok:
		*p = ipc.ip_data;
		mtpr(u.u_ar0[SP], USP);
		if (u.u_ar0[PS] & PSL_DBL)
			rest_d_acc(&u.u_ar0[RAMST]);
		else
			rest_s_acc(&u.u_ar0[RAMST]);
		break;

	/* set signal and continue */
	/* one version causes a trace-trap */
	case 9:
	case 7:
		if ((int)ipc.ip_addr != 1)
			u.u_ar0[PC] = (int)ipc.ip_addr;
		if ((unsigned)ipc.ip_data > NSIG)
			goto error;
		u.u_procp->p_cursig = ipc.ip_data;	/* see issig */
		u.u_procp->p_sig = (ipc.ip_data) ? (1 << (ipc.ip_data - 1)) : 0;
		if (i == 9) 
			u.u_ar0[PS] |= PSL_T;
		wakeup((caddr_t)&ipc);
		return (1);

	/* force exit */
	case 8:
		wakeup((caddr_t)&ipc);
		exit(u.u_procp->p_cursig);

	default:
	error:
		ipc.ip_req = -1;
	}
	wakeup((caddr_t)&ipc);
	return (0);
}
