/*
 *	VAX/UNIX Operating System Profiler
 *
 *	Sorted Kernel text addresses are written into driver.  At each
 *	clock interrupt a binary search locates the counter for the
 *	interval containing the captured PC and increments it.
 *	The last counter is used to hold the User mode counts.
 */

#include	"../sys/param.h"
#include	"../ml/psl.h"
#include	"../sys/uio.h"
#include	"../sys/user.h"
#include	"../sys/buf.h"

# define PRFMAX  4096		/* maximum number of text addresses */
# define PRF_ON    1		/* profiler collecting samples */
# define PRF_VAL   2		/* profiler contains valid text symbols */
# define BPW	   4		/* bytes per word */
# define L2BPW	   2		/* log2(BPW) */

unsigned  prfstat;		/* state of profiler */
unsigned  prfmax;		/* number of loaded text symbols */
unsigned  prfctr[PRFMAX + 1];	/* counters for symbols; last used for User */
unsigned  prfsym[PRFMAX];	/* text symbols */

prfread(dev,uio)
dev_t	dev;
register struct uio	*uio;
{
	unsigned  min();
	register int error;
#ifdef lint
	dev=dev ;
#endif
	if((prfstat & PRF_VAL) == 0)
		return(ENXIO);
 	if (error = uiomove((caddr_t) prfsym, (prfmax * BPW),
	    UIO_READ,uio))
		return(error);
	return(uiomove((caddr_t) prfctr, ((prfmax + 1) * BPW),
	    UIO_READ,uio));
}

prfwrite(dev,uio)
dev_t dev;
register struct uio *uio;
{
	register  unsigned  *ip;
	register  int error;
#ifdef lint
	dev=dev ;
#endif
	if(uio->uio_resid > sizeof prfsym)
		return(ENOSPC);
	if(uio->uio_resid & (BPW - 1) || uio->uio_resid < 3 * BPW)
		return(E2BIG);
	if(prfstat & PRF_ON)
		return(EBUSY);
	for(ip = prfctr; ip != &prfctr[PRFMAX + 1]; )
		*ip++ = 0;
	prfmax = uio->uio_resid >> L2BPW;
	if (error = uiomove((caddr_t) prfsym, sizeof(prfsym), UIO_WRITE,uio)) {
		prfstat = 0;
		return(error);
	}
	for(ip = &prfsym[1]; ip != &prfsym[prfmax]; ip++)
		if(*ip < ip[-1]) {
			prfstat = 0;
			return(EINVAL);
		}
	prfstat = PRF_VAL;
	return(0);
}

prfioctl(dev, cmd, arg, mode)
dev_t	dev;
int 	cmd;
char	*arg;
int 	mode;
{
#ifdef lint
	dev=dev ;
	mode=mode;
#endif
	switch(cmd & 0xff) {
	case 1:
		u.u_r.r_val1 = prfstat;
		break;
	case 2:
		u.u_r.r_val1 = prfmax;
		break;
	case 3:
		if(prfstat & PRF_VAL) {
			prfstat = PRF_VAL | ((int) *arg) & PRF_ON;
			break;
		}
	default:
		return(EINVAL);
	}
	return(0);
}

prfintr(pc, ps)
	register  unsigned  pc;
{
	register  int  h, l, m;

	if(USERMODE(ps))
		prfctr[prfmax]++;
	else {
		l = 0;
		h = prfmax;
		while((m = (l + h) / 2) != l)
			if(pc >= prfsym[m])
				l = m;
			else
				h = m;
		prfctr[m]++;
	}
}
