#include  "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/dir.h"
#include "../sys/user.h"
#include "../sys/vm.h"
#include "../sys/kernel.h"
#include "../sys/proc.h"
#include "../sys/map.h"
#include "../sys/cmap.h"
#include "../sys/buf.h"
#include "../sys/errno.h"
#include "../sys/lock.h"

#include "../ml/mtpr.h"

int	shmtot;		/* total shared memory currently used */

struct	shmid_ds	*ipcget(), *shmconv();

/*
 *	shmat - Shmat system call.
 */

shmat()
{
	register struct a {
		int	shmid;
		char	*addr;
		int	flag;
	}	*uap = (struct a *)u.u_ap;

	register struct proc		*p = u.u_procp;
	register struct shmid_ds	*sp;	/* shared memory header ptr */
	register			shmvp;
	register struct	shm_pt 	*sptp, *nsptp;	/* proc structure info */
	register struct pte	*seg, *segbeg, *segend;

	int		pflag, p0lr, i;


/* Obtain shmem structure address */

	if ((sp = shmconv(uap->shmid, SHM_DEST)) == NULL)
		return;

/* Check for read permission on segment. */

	if (ipcaccess(&sp->shm_perm, SHM_R))
		return;

/* Check for write permission on segment if read only permission is desired. */

	if ((uap->flag & SHM_RDONLY) == 0)
		if (ipcaccess(&sp->shm_perm, SHM_W))
			return;

/* Find the  shm_pt structure to attach sp to */

	for (sptp = p->p_shmpt, i = 0; i < shminfo.shmseg; i++, sptp++)
		if (sptp->shm_shmp == NULL)
			break;

	if (i == shminfo.shmseg) {
		u.u_error = EMFILE;
		return;
	}

/* Determine shared memory  start address when SHM_RND is true */

	if (uap->flag & SHM_RND)
		uap->addr = (char *)((u_int)uap->addr & ~(SHMLBA - 1));

/* Check for page alignment and containment within P0 */

	if ((int)uap->addr & NBPG - 1 ||
	    (((int)(uap->addr + sp->shm_segsz)) & P0MASK) == P0MASK) {
		u.u_error = EINVAL;
		return;
	}

	p0lr = u.u_pcb.pcb_p0lr;

/* An address of 0 places the shared memory into a first fit location */

	if (uap->addr == NULL) {

		segbeg = NULL;

/* Try to find ptes that can be used for this shared memory segment. */

		for (seg = p->p_p0br + u.u_tsize + u.u_dsize + shminfo.shmbrk,
		     segend = p->p_p0br + p0lr; seg < segend; seg++) {

		    if (segbeg && ctob(seg-segbeg+1) >= sp->shm_segsz)
			break;

		    else if (seg->pg_fod == 0 && seg->pg_pfnum) {
			if (cmap[pgtocm(seg->pg_pfnum)].c_gone == 0)
				    segbeg = NULL;
			else if (segbeg == NULL)
			    segbeg = seg;
		    }

		    else if (((*(int *)(seg)) & PG_PROT) == PG_KR) {
			if (segbeg == NULL)
			    segbeg = seg;
		    }

		    else
			segbeg = NULL;

		} /* for */

/* If not enough ptes were found, set segbeg to p0lr.  Add shminfo.shmbrk
 * clicks if this is the only shared memory segment for this process.
 */

		if (segbeg == NULL) {
			segbeg = p->p_p0br + p0lr;
			if (p->p_smbeg == NULL)
				segbeg += shminfo.shmbrk;
		}

	} /* if (uap->addr == NULL) */
	
	else {

/* Check to make sure that this virtual address does not lie in the text
 * or stack regions.  Also check that it does not lie within the bounds of
 * the non-shared data region and that it does not overlay any valid
 * shared memory segments.
 */

		shmvp = btop(uap->addr);

		if (isassv(p, shmvp) || isatsv(p, shmvp)) {
			u.u_error = EINVAL;
			return;
		}

		segbeg = dptopte(p, vtodp(p, shmvp));

		if (segbeg < (segend = p->p_p0br + p0lr)) {

			if (isansdsv(p, shmvp)) {
				u.u_error = ENOMEM;
				return;
			}

			if (segbeg + btoc(sp->shm_segsz) < segend)
				segend = segbeg + btoc(sp->shm_segsz) - 1;

			for (nsptp = p->p_shmpt; nsptp < p->p_shmpt +
			     shminfo.shmseg; nsptp++) {
			    if (nsptp->shm_shmp &&
			       (segbeg >= vtopte(p, nsptp->shm_sva) &&
				segbeg < vtopte(p, nsptp->shm_sva +
				    btoc(nsptp->shm_shmp->shm_segsz)) ||
			        segend >= vtopte(p, nsptp->shm_sva) &&
				segend < vtopte(p, nsptp->shm_sva +
				    btoc(nsptp->shm_shmp->shm_segsz)))) {
					u.u_error = ENOMEM;
					return;
				} 
			}
		} /* if */

	} /* else */

/* Initialize the shm_pt structure, call shmattach(), set shm_atime,
 * shm_lpid, and the return value and return.
 */

	segend = segbeg + btoc(sp->shm_segsz) - 1;

	pflag = (uap->flag & SHM_RDONLY) ? PG_URKW : PG_UW;

	sptp->shm_sva = dptov(p, ptetodp(p, segbeg));
	sptp->shm_shmp = NULL;
	sptp->shm_link = NULL;
	sptp->shm_idx = NULL;
	sptp->shm_spte = segbeg;
	sptp->shm_epte = segend;
	sptp->shm_sflg = pflag;

	if (shmattach(p, sptp, sp) == 0) {
		sptp->shm_sva = 0;
		if (p->p_smbeg == sptp->shm_spte)
			p->p_smbeg = NULL;
		if (p->p_smend == sptp->shm_epte)
			p->p_smend = NULL;
		sptp->shm_epte = sptp->shm_spte = (struct pte *)0;
		sptp->shm_sflg = 0;

		if (u.u_error == 0)
			u.u_error = ENOMEM;
		return;
	}

	ipclock(&sp->shm_perm);
	sp->shm_atime = time.tv_sec;
	sp->shm_lpid = p->p_pid;
	ipcunlock(&sp->shm_perm);

	u.u_r.r_val1 = (int)ptob(sptp->shm_sva);
}


/*
 * 	shmattach - 	Attach proc to linked list of procs sharing a
 *			data segment.  Also initialize ptes.
 *			This function is called by shmat() and shmfork()
 *			
 */

shmattach(p, sptp, sp)
register struct proc 	*p;
register struct shmid_ds *sp;
struct shm_pt 		*sptp;
{

	register 			p0lr = u.u_pcb.pcb_p0lr;
	register 	change;
	register struct pte 		*segend;
	struct 	 pte			*pte;
	struct shm_pt 			*rsptp;
	struct shmid_ds 	 	*rsp;
	register 			s, i;

/* If the index of the last shared memory pte is >= P0LR, check that the size
 * of the data segment plus this shared memory segment is not greater than
 * the allowed maximum size.  Call expand() to initialize the new pte's fo us
 * without increasing the size of the data region.  We don't include the
 * shared memory ptes in p->p_dsize so that shared data pages can be treated
 * as different entities than unshared data pages (in paging, swapping, etc).
 */

	if ((segend = vtopte(p, sptp->shm_sva + btoc(sp->shm_segsz) - 1)) >=
	     p->p_p0br + p0lr) {
		change = clrnd(segend - (p->p_p0br + p0lr) + 1);
		if (chksize((u_int)u.u_tsize,(u_int)(p0lr - u.u_tsize + change),
			     (u_int)u.u_ssize))
			return(0);

/* Set p_smbeg and p_smend, if necessary.  This is needed so that swapin
 * will get enough pages if we must swap as a result of a ptexpand.
 */

		s = spl8();
		if (p->p_smbeg == NULL || p->p_smbeg > vtopte(p, sptp->shm_sva))
			p->p_smbeg = vtopte(p, sptp->shm_sva);

		if (p->p_smend <
		    vtopte(p, sptp->shm_sva + btoc(sp->shm_segsz) - 1))
			p->p_smend = vtopte(p, sptp->shm_sva +
					    btoc(sp->shm_segsz) - 1);
		splx(s);

		expand(change, 2);

/* Reset shared memory start and end ptes for all segments */

		s = spl8();
		p->p_smend = p->p_smbeg = NULL;

		for (rsptp = p->p_shmpt, i = 0; i < shminfo.shmseg; i++, rsptp++) {
			rsp = (rsptp == sptp) ? sp : rsptp->shm_shmp;
			if (rsp) {
				rsptp->shm_spte = vtopte(p, rsptp->shm_sva);
				rsptp->shm_epte = rsptp->shm_spte +
					btoc(rsp->shm_segsz) - 1;
				if (p->p_smbeg) {
					if (p->p_smbeg > rsptp->shm_spte)
						p->p_smbeg = rsptp->shm_spte;
					if (p->p_smend < rsptp->shm_epte)
						p->p_smend = rsptp->shm_epte;
				}
				else {
					p->p_smbeg = rsptp->shm_spte;
					p->p_smend = rsptp->shm_epte;
				}

			}
		}
		splx(s);
	}

/* Set p->p_smbeg and p->p_smemd, if necessary. */

	else {
		s = spl8();
		if (p->p_smbeg == NULL || p->p_smbeg > vtopte(p, sptp->shm_sva))
			p->p_smbeg = vtopte(p, sptp->shm_sva);

		if (p->p_smend <
		    vtopte(p, sptp->shm_sva + btoc(sp->shm_segsz) - 1))
			p->p_smend = vtopte(p, sptp->shm_sva +
					    btoc(sp->shm_segsz) - 1);
		splx(s);
	}

	if (p->p_smend > p->p_addr)
		panic("shmattach p_addr");

/* If this is the first shared memory segment attached, set ptes for the
 * rest of the data segment to PG_KR so that the user will get a fault if
 * the pages are accessed.
 */

	if (p->p_smbeg == vtopte(p, sptp->shm_sva))
		for (pte = dptopte(p, p->p_dsize);
		     pte < vtopte(p, sptp->shm_sva); pte++)
			*(int *)pte = PG_KR;

/* If another process is already attached to this shared memory segment
 * or if the segment was not destroyed on the last detach (if there was one),
 * wait for the segment to be unlocked, then link up to it, complete
 * initialization and return.
 */

	ipclock(&sp->shm_perm);
	if (sp->shm_proc)
		if (u.u_error = shminvalid(p, sptp, sp)) {
			ipcunlock(&sp->shm_perm);
			return(0);
		}

	if (sp->shm_nattch || sp->shm_perm.flag & SHM_SWAP) {
		sp->shm_nattch++;
		s = spl8();
		shmlink(p, sptp, sp);
		sptp->shm_shmp = sp;
		splx(s);
		ipcunlock(&sp->shm_perm);
		return(1);
	}

/* If there was no proc to link to, allocate swap space for this shared
 * memory segment, peform initialization, link this proc and retrun.
 */

	if (vsshmalloc(sp) == NULL) {
		swkill(p, "shmattach");
		ipcunlock(&sp->shm_perm);
		return(0);
	}

	sp->shm_nattch = 1;
	sp->shm_cnattch = 0;
	s = spl8();
	shmlink(p, sptp, sp);
	sptp->shm_shmp = sp;
	splx(s);
	ipcunlock(&sp->shm_perm);
	return(1);

}


/*
 *	shmconv - Convert user supplied shmid into a ptr to the associated
 *		shared memory header.
 */

struct shmid_ds *
shmconv(s, flg)
register int	s;	/* shmid */
int		flg;	/* error if matching bits are set in mode */
{

	register struct shmid_ds	*sp;	/* ptr to associated header */

	sp = &shmem[s % shminfo.shmmni];
	if (!(sp->shm_perm.mode & IPC_ALLOC) || sp->shm_perm.mode & flg
		|| s / shminfo.shmmni != sp->shm_perm.seq) {
		u.u_error = EINVAL;
		return(NULL);
	}
	return(sp);
}

/*
 *	shmctl - Shmctl system call.
 */

shmctl()
{
	register struct a {
		int		shmid,
				cmd;
		struct shmid_ds	*arg;
	}	*uap = (struct a *)u.u_ap;

	register struct shmid_ds *sp;	/* shared memory header ptr */
	register 		i;
	register struct pte	*pte;
	register struct	shm_pt 	*sptp;
			int	limit;	/* # pages locked limit */

	struct 	 shmid_ds	ds;	/* hold area for IPC_SET */

	if ((sp = shmconv(uap->shmid, (uap->cmd == IPC_STAT) ? 0 : SHM_DEST)) ==
		NULL)
		return;
	u.u_r.r_val1 = 0;


	switch(uap->cmd) {

/* Remove shared memory identifier. */
	case IPC_RMID:
		if (u.u_uid != sp->shm_perm.uid && u.u_uid != sp->shm_perm.cuid
			&& !suser())
			return;
		sp->shm_ctime = time.tv_sec;
		sp->shm_perm.mode |= SHM_DEST;

		/* Change key to private so old key can be reused without
		 * waiting for last detach.  Only allowed accesses to
		 * this segment now are shmdt() and shmctl(IPC_STAT).
		 * All others will give bad shmid. 
		 */

		sp->shm_perm.key = IPC_PRIVATE;

		/* If no process has this shared memory segment attached,
		 * free the swap space of the shared memory segment and
		 * its ptes.
		 */

		ipclock(&sp->shm_perm);
		if (sp->shm_nattch == 0) {
			sp->shm_perm.mode = 0;
			shmtot -= btoc(sp->shm_segsz);

			if (sp->shm_perm.flag & SHM_SWAP) {
				vsshmfree(sp, ((long)btoc(sp->shm_segsz)));
				sp->shm_perm.flag &= ~SHM_SWAP;
				if (((int)(++(sp->shm_perm.seq)*shminfo.shmmni +
				    (sp - &shmem[0]))) < 0)
					sp->shm_perm.seq = 0;
			}
		}
		ipcunlock(&sp->shm_perm);

		return;

/* Set ownership and permissions. */
	case IPC_SET:
		if (u.u_uid != sp->shm_perm.uid && u.u_uid != sp->shm_perm.cuid
			&& !suser())
			return;
		u.u_error = copyin((caddr_t)uap->arg, (caddr_t)&ds, sizeof(ds));
		if (u.u_error)
			return;

		sp->shm_perm.uid = ds.shm_perm.uid;
		sp->shm_perm.gid = ds.shm_perm.gid;
		sp->shm_perm.mode = (ds.shm_perm.mode & 0777) |
			(sp->shm_perm.mode & ~0777);
		sp->shm_ctime = time.tv_sec;
		return;

/* Get shared memory data structure. */
	case IPC_STAT:
		if (ipcaccess(&sp->shm_perm, SHM_R))
			return;
		u.u_error = copyout((caddr_t)sp,(caddr_t)uap->arg, sizeof(*sp));
		return;

	/* Lock segment in memory */
	case SHM_LOCK:
		for (sptp = u.u_procp->p_shmpt;
		     sptp < &(u.u_procp->p_shmpt[shminfo.shmseg]);
		     sptp++)
			if (sptp->shm_shmp == sp)
				break;

		if (sptp == &(u.u_procp->p_shmpt[shminfo.shmseg])) {
			u.u_error = EINVAL;
			return;
		}

		if (sptp->shm_sflg & SHM_MLOCK) {	/* check for already locked */
			u.u_error = EINVAL;
			return;
		}

		if (sp->shm_lcount == 0) {		/* nobody else locked */
			limit = (u.u_uid == 0) ? s_limit : u_limit;
			/*
			** check against limits, update
			*/
			if ((plocked += btoc(sp->shm_segsz)) > limit) {
				plocked -= btoc(sp->shm_segsz);
				u.u_error = ENOMEM;
				return;
			}
			sp->shm_lcount++;
			
			sptp->shm_sflg |= SHM_MLOCK;	/* mark locked for pagein */
			
			/*
			** pagein and set the c_noswap bit in the core map entries
			** associated with the pages of this segment
			*/
			for (pte = vtopte(u.u_procp, sptp->shm_sva);
			     pte < vtopte(u.u_procp, sptp->shm_sva+btoc(sp->shm_segsz));
			     pte += CLSIZE) {
				     while ( ! pte->pg_v )
					     pagein(ctob(dptov(u.u_procp,
							       ptetoshmp(u.u_procp,
									 pte))), 0);
				     cmap[pgtocm(pte->pg_pfnum)].c_noswap = 1;
			}
		}
		else
			sp->shm_lcount++;		/* incr glob lck cnt */

		sptp->shm_sflg |= SHM_MLOCK;		/* mark local lcked */
		u.u_procp->p_flag |= SPLOCK;		/* mark proc lcked */
		return;

	/* Unlock segment */
	case SHM_UNLOCK:
		for (sptp = u.u_procp->p_shmpt;
		     sptp < &(u.u_procp->p_shmpt[shminfo.shmseg]);
		     sptp++)
			if (sptp->shm_shmp == sp)
				break;

		if (sptp == &(u.u_procp->p_shmpt[shminfo.shmseg])) {
			u.u_error = EINVAL;
			return;
		}

		if ( ! (sptp->shm_sflg & SHM_MLOCK)) {	/* see if locked */
			u.u_error = EINVAL;
			return;
		}

		/*
		** call common routine to unlock
		*/
		shmpunlock(u.u_procp, sptp, sp);
		return;

	default:
		u.u_error = EINVAL;
		return;
	}
}

/*
 *	shmdt - Shmdt system call.
 */

shmdt()
{
	register struct a {
		char	*addr;
	}	*uap = (struct a *)u.u_ap;

	register struct proc 		*p = u.u_procp;
	register struct shm_pt		*sptp = p->p_shmpt;
	register struct shmid_ds	*sp;
	register struct	pte		*shmepte;
	register u_int		v;
	register struct pte 	*spte, *epte;
	register 		i, s;


/* Check for page alignment */

	if ((int)uap->addr & (NBPG - 1) || (v = btop(uap->addr)) == 0) {
		u.u_error = EINVAL;
		return;
	}

/* Find the segment */

	for (i = 0; i < shminfo.shmseg; i++, sptp++)
		if ( (sp = sptp->shm_shmp) != NULL && v == sptp->shm_sva)
			break;

	if (i == shminfo.shmseg) {
		u.u_error = EINVAL;
		return;
	}
	shmepte = vtopte(p, sptp->shm_sva + btoc(sp->shm_segsz) - 1);

/* Call shmfree() to decrment the counts and free the segment and page tables
 * if neeeded.
 */

	mtpr(0, TBIA);

	shmfree(p, sptp);

/* Reset the shmem structure pointed to by sptp and clear the ptes belonging
 * to this shared memory segment.
 */

	sp->shm_dtime = time.tv_sec;
	sp->shm_lpid = p->p_pid;

	sptp->shm_sva = NULL;
	sptp->shm_link = NULL;
	sptp->shm_idx = NULL;
	sptp->shm_epte = sptp->shm_spte = NULL;
	sptp->shm_sflg = 0;

/* Adjust the p->p_smbeg and p->p_smend pte addresses. */

	s = spl8();
	p->p_smend = p->p_smbeg = NULL;
	
	for (sptp = p->p_shmpt, i = 0; i < shminfo.shmseg; i++, sptp++) {
		if (sptp->shm_shmp == NULL)
			continue;

		if (spte = vtopte(p, sptp->shm_sva)) {
			epte = vtopte(p, sptp->shm_sva +
				      btoc(sptp->shm_shmp->shm_segsz) - 1);

			if (p->p_smbeg) {
				if (p->p_smbeg > spte)
					p->p_smbeg = spte;
				if (p->p_smend < epte)
					p->p_smend = epte;
			}
			else {
				p->p_smbeg = spte;
				p->p_smend = epte;
			}
		}
	}

	splx(s);

	if (p->p_smbeg == NULL)
		expand(clrnd(-(shmepte - dptopte(p, p->p_dsize - 1))), 2);
	else if (shmepte > p->p_smend)
		expand(clrnd(-(shmepte - p->p_smend)), 2);

	u.u_r.r_val1 = 0;

}

/*
 *	shmexitexec - Called by getxfile (kern_exec.c) and exit (kern_exit.c)
 *	to handle shared memory exit and exec processing.
 *
 */

shmexitexec()
{

	register struct proc 		*p = u.u_procp;
	register struct shm_pt 		*sptp = p->p_shmpt;
	register struct pte 		*seg;

	register i;

	if (p->p_smbeg == NULL)
		return;

/* Detach any shared memory segments.  Call shmfree() for each and reset the
 * shm_pt structure for each segment.
 */

	p->p_smend = p->p_smbeg = NULL;

	for (i = 0; i < shminfo.shmseg; sptp++, i++) {

		if (sptp->shm_shmp) {
			shmfree(p, sptp);

			sptp->shm_sva = NULL;
			sptp->shm_link = NULL;
			sptp->shm_idx = NULL;
			sptp->shm_epte = sptp->shm_spte = NULL;
			sptp->shm_sflg = 0;

		}

	}

}

/*
 *	shmfork - Called by newproc(kern_fork.c) to handle shared memory fork
 *		processing.
 */

shmfork(cp, pp)
register struct proc	*cp,	/* ptr to child proc table entry */
			*pp;	/* ptr to parent proc table entry */
{

	register struct shm_pt 	*psptp = pp->p_shmpt;
	register struct shm_pt 	*csptp = cp->p_shmpt;
	register struct shmid_ds 	*psp;
	register 		i;
	register struct pte	*pte;
	register s;

	if (pp->p_smbeg == NULL)
		return;

	forkutl.u_pcb.pcb_p0lr = u.u_pcb.pcb_p0lr;

 /* For each data segment that the parent is sharing, copy the parent's
  * shm_pt structure to the child's.  Also, call shmattach() to copy the
  * parent's ptes and link the child to the list of processes sharing this
  * segment.
  */

	cp->p_smend = cp->p_smbeg = NULL;

	for (i = 0; i < shminfo.shmseg; i++, psptp++, csptp++) {

		if (psp = psptp->shm_shmp) {
			csptp->shm_sva = psptp->shm_sva;
			csptp->shm_link = NULL;
			csptp->shm_idx = NULL;
			csptp->shm_spte = vtopte(cp, csptp->shm_sva);
			csptp->shm_epte = csptp->shm_spte+btoc(psp->shm_segsz)-1;
			csptp->shm_sflg = psptp->shm_sflg;

			if (cp->p_smbeg == NULL||cp->p_smbeg > csptp->shm_spte)
				cp->p_smbeg = csptp->shm_spte;

			if (cp->p_smend  < csptp->shm_epte)
				cp->p_smend = csptp->shm_epte;

			psp->shm_perm.flag |= IPC_LOCK;
			psp->shm_nattch++;
			s = spl8();
			shmlink(cp, csptp, psp);
			csptp->shm_shmp = psp;
			splx(s);
			psp->shm_perm.flag &= ~IPC_LOCK;
		}
	}

	for (pte = dptopte(cp, cp->p_dsize); pte < cp->p_smbeg; pte++)
		*(int *)pte = PG_KR;

}

/*
** shmpunlock - unlock (as in plock etc.) a shared memory segment
**              called by shmctl() and shmfree()
*/
shmpunlock(p, sptp, sp)
register struct proc		*p;
register struct shm_pt		*sptp;
register struct shmid_ds	*sp;
{
	register struct pte	*pte;
	
	if (--(sp->shm_lcount) == 0) {			/* dcr locked count */
		/*
		** reset the c_noswap bit in the core map entries
		** associated with the pages of this segment
		*/
		for (pte = vtopte(p, sptp->shm_sva);
		     pte < vtopte(p, sptp->shm_sva + btoc(sp->shm_segsz));
		     pte += CLSIZE)
			     cmap[pgtocm(pte->pg_pfnum)].c_noswap = 0;

		plocked -= btoc(sp->shm_segsz);		/* adj # page lcked */
	}
	
	sptp->shm_sflg &= ~SHM_MLOCK; 			/* reset specific flag */

	/*
	** Check if we should turn off the SPLOCK flag
	*/
	if ( ! (p->p_lock & (PROCLOCK | TXTLOCK | DATLOCK))) {
		for(sptp = p->p_shmpt;
		    sptp < &(p->p_shmpt[shminfo.shmseg]);
		    sptp++)
			if (sptp->shm_shmp &&
			    (sptp->shm_sflg & SHM_MLOCK))
				return;			/* return here */
		p->p_flag &= ~SPLOCK;			/* reset SPLOCK bit */
	}
}

/*
 *	shmfree - 	Decrement counts.  Free segment and page tables if
 *			indicated.
 */

shmfree(p, sptp)
register struct proc	*p;
register struct shm_pt	*sptp;
{

	register struct shmid_ds 	*sp = sptp->shm_shmp;
	register 			s, i;
	register struct	pte		*pte;
	int 				a;

	if (sp == NULL)
		return;

	if (sptp->shm_sflg & SHM_MLOCK)			/* check for plocked */
		shmpunlock(p, sptp, sp);		/* unlock */

/* If the total number of attaches gets decremented to zero and the segment
 * is to be destroyed, it is unlinked from the linked list of procs (by
 * shmunlink()).  Then vmemfree() is called to free the memory allocated for 
 * the segment.  While page outs are in progress for this segment, sleep.
 * Afterwards, call vsshmfree() to free swap space allocated for the segment.
 */

	ipclock(&sp->shm_perm);

	if (--sp->shm_nattch == 0 && sp->shm_perm.mode & SHM_DEST) {
		sp->shm_cnattch = 0;
		s = spl8();	
		sp->shm_rssize -= vmemfree(vtopte(p, sptp->shm_sva),
					   btoc(sp->shm_segsz));
		if (sp->shm_rssize)
			panic("shmfree rssize");

		while (sp->shm_poip)
	 		slumber( (caddr_t)&sp->shm_poip, 0, resboost);
		shmunlink(p, sptp, sp);
		sptp->shm_shmp = NULL;
		vsshmfree(sp, ((long)btoc(sp->shm_segsz)));
		splx(s);

		sp->shm_perm.mode = 0;
		if (((int)(++(sp->shm_perm.seq)*shminfo.shmmni + 
			   (sp - &shmem[0]))) < 0)
			sp->shm_perm.seq = 0;
		shmtot -= btoc(sp->shm_segsz);
		sp->shm_perm.flag = 0;

	}
	else {
		sp->shm_perm.flag &= ~IPC_LOCK;
		s = spl8();
		shmccdec(p, sptp, sp);
		sptp->shm_shmp = NULL;
		splx(s);
	}

/* Set pages to PG_KR so this process can't access them again. */
	for (pte = vtopte(p, sptp->shm_sva), i = 0; i < btoc(sp->shm_segsz);
	     pte++, i++)
		*(int *)pte = (*(int *)pte & ~PG_PROT) | PG_KR;
}


/*
 *	shmget - Shmget system call.
 */

shmget()
{
	register struct a {
		key_t	key;
		int	size,
			shmflg;
	}	*uap = (struct a *)u.u_ap;

	register struct shmid_ds	*sp;	/* shared memory header ptr */
	int			s;	/* ipcget status */
	int			size;

	if ((sp = ipcget(uap->key, uap->shmflg, shmem, shminfo.shmmni,
	     sizeof(*sp), &s)) == NULL)
		return;
	if (s) {

/* This is a new shared memory segment.  Check that the total amount of 
 * shared memory system wide will not be exceeded when the SHMAT() is done.
 */

		if (uap->size < shminfo.shmmin || uap->size > shminfo.shmmax) {
			u.u_error = EINVAL;
			sp->shm_perm.mode = 0;
			return;
		}
		size = clrnd(btoc(uap->size));

		if (shmtot + size > shminfo.shmall) {
			u.u_error = ENOMEM;
			sp->shm_perm.mode = 0;
			return;
		}

		shmtot += size;

/* Lock the ipc_perm structure and perform initialization. */

		ipclock(&sp->shm_perm);
		sp->shm_segsz = uap->size;
		sp->shm_perm.mode |= SHM_CLEAR;
		sp->shm_proc = NULL;
		sp->shm_pidx = 0;
		sp->shm_cnattch = sp->shm_nattch = 0;
		sp->shm_dtime = sp->shm_atime = (long)0;
		sp->shm_ctime = time.tv_sec;
		sp->shm_lpid = 0;
		sp->shm_cpid = u.u_procp->p_pid;
		sp->shm_rssize = 0;
		ipcunlock(&sp->shm_perm);
	}
	else
		if (uap->size && uap->size > sp->shm_segsz) {
			u.u_error = EINVAL;
			return;
		}

	u.u_r.r_val1 = sp->shm_perm.seq * shminfo.shmmni + (sp - &shmem[0]);
}

/*
 *	shmsys - System entry point for shmat, shmctl, shmdt, and shmget
 *		system calls.
 */

shmsys()
{
	register struct a {
		u_int	id;
	}	*uap = (struct a *)u.u_ap;
	int		shmat(),
			shmctl(),
			shmdt(),
			shmget();
	static int	(*calls[])() = {shmat, shmctl, shmdt, shmget};

	if (uap->id > 3) {
		u.u_error = EINVAL;
		return;
	}
	/* Kludgery, skip past the id in the argument list to point to the real
	   argument list of the user. */
	u.u_ap++;
	(*calls[uap->id])();
	u.u_ap--;
}


/*
 * shmlink - 	Link a proc to the list of processes sharing this data segment
 *		and initialize its shared memory ptes.
 *
 */

shmlink(p, sptp, sp)
register struct proc 	*p;
register struct shm_pt 	*sptp;
register struct shmid_ds *sp;
{

	register s;

	if (sp == NULL)
		return;

/* Call vinitshmpt() to copy the ptes from a linked proc, if one exists,
 * or to initialize the ptes, if none exists.  Then link this proc to
 * the shmem structure, increment the incore attach code and return.
 */

	s = spl8();
	vinitshmpt(p, sptp, sp);

	sptp->shm_link = sp->shm_proc;
	sptp->shm_idx = sp->shm_pidx;

	sp->shm_proc = p;
	sp->shm_pidx = (short)(sptp - p->p_shmpt);
	sp->shm_cnattch++;
	sptp->shm_sflg &= ~SHM_UNLINK;
	splx(s);
}


/*
 * shmunlink -	Unlink a proc from the list of processes sharing a data
 * 		segment.
 *
 */

shmunlink(p, sptp, sp)
register struct proc 	*p;
register struct shm_pt 	*sptp;
register struct shmid_ds *sp;
{

	register struct shm_pt	*qsp;

	if (sp == NULL || sp->shm_proc == NULL || sptp->shm_sflg & SHM_UNLINK)
		return;

	if (sp->shm_proc == p) {
		sp->shm_proc = sptp->shm_link;
		sp->shm_pidx = sptp->shm_idx;
		sptp->shm_link = NULL;
		sptp->shm_idx = NULL;
		sptp->shm_sflg |= SHM_UNLINK;
		return;
	}

	for (qsp = sp->shm_proc->p_shmpt + sp->shm_pidx; qsp->shm_link;
	     qsp = qsp->shm_link->p_shmpt + qsp->shm_idx) {
		
		if (qsp->shm_link == p) {
			qsp->shm_link = sptp->shm_link;
			qsp->shm_idx = sptp->shm_idx;
			sptp->shm_link = NULL;
			sptp->shm_idx = NULL;
			sptp->shm_sflg |= SHM_UNLINK;
			return;
		}

	}

	panic("lost shared-memory segment");
}


/*
 * shmccdec - 	Decrement the incore usage count of a shared memory segment.
 *		When it drops to zero, free the core space.
 *
 */

shmccdec(p, sptp, sp)
register struct proc 	*p;
register struct shm_pt	*sptp;
register struct shmid_ds *sp;
{

	register 		s, i;
	register struct	pte	*pte;

	if (sp == NULL || sp->shm_cnattch == 0)
		return;

	ipclock(&sp->shm_perm);
	if (--sp->shm_cnattch == 0) {
		s = spl8();
		p->p_flag |= SULOCK;
		vsswap(p, vtopte(p, sptp->shm_sva), CSHMEM, sptp->shm_sva,
		       btoc(sp->shm_segsz), (struct dmap *)0);
		swap(p, sp->shm_ptdaddr, (caddr_t)vtopte(p, sptp->shm_sva), 
		     btoc(sp->shm_segsz) * sizeof(struct pte),
		     B_WRITE, B_PAGET, swapdev_vp, 0);
		sp->shm_perm.flag |= SHM_SWAP;
		p->p_flag &= ~SULOCK;

		if (sp->shm_rssize)
			panic("shared memory segment");

		while (sp->shm_poip)
	 		slumber( (caddr_t)&sp->shm_poip, 0, resboost);
		shmunlink(p, sptp, sp);
		splx(s);
	}
	else {
		s = spl8();
		shmunlink(p, sptp, sp);
		splx(s);

/*
 * If any of this shared memory segment's pages have been modified, 
 * distribute the corrsponding pte to all attached processes.
 */

		for (pte = vtopte(p, sptp->shm_sva), i = 0;
		     i < btoc(sp->shm_segsz); pte++, i++)
			if (dirtycl(pte))
                                dishmpte(sptp->shm_shmp, i, pte);
	}

	ipcunlock(&sp->shm_perm);
}


/*
 * shmrepl -	Replace p by q in all its shared memory linked lists.
 *		Used by vfork() internally.
 *
 */

shmrepl(p, q)
register struct proc	*p,
			*q;
{

	register struct shm_pt 	*psptp = p->p_shmpt;
	register struct shm_pt 	*qsptp = q->p_shmpt;
	register struct shmid_ds 	*psp;
	register 		i, s;

	if (p->p_smbeg == NULL)
		return;

	q->p_smbeg = p->p_smbeg;
	q->p_smend = p->p_smend;

 /* For each data segment that the parent is sharing, copy p's shm_pt
  * structure to q's.  Also, call shmunlink() to unlink p from the list of
  * processes sharing this data segment.  Then manually link q to the list.
  * Don't call shmlink() because it will increment the incore attach count.
  * Finally set p's smbeg and smend ptes to NULL.
  */

	for (i = 0; i < shminfo.shmseg; i++, psptp++, qsptp++) {

		if (psp = psptp->shm_shmp) {
			s = spl8();
			qsptp->shm_sva = psptp->shm_sva;
			qsptp->shm_link = NULL;
			qsptp->shm_idx = NULL;
			qsptp->shm_spte = psptp->shm_spte;
			qsptp->shm_epte = psptp->shm_epte;
			qsptp->shm_sflg = psptp->shm_sflg;

			shmunlink(p, psptp, psp);

			psptp->shm_sva  = 0;
			psptp->shm_shmp = NULL;
			psptp->shm_spte = NULL;
			psptp->shm_epte = NULL;
			psptp->shm_sflg = 0;

			qsptp->shm_link = psp->shm_proc;
			qsptp->shm_idx = psp->shm_pidx;

			psp->shm_proc = q;
			psp->shm_pidx = (short)(qsptp - q->p_shmpt);
			qsptp->shm_shmp = psp;
			splx(s);
		}
	}

	p->p_smend = p->p_smbeg = NULL;

}

/* 
 * shminvalid - Determine if a process has already attached a particular
 * 		shared memory segment.
 */

shminvalid(p, sptp, sp)
register struct proc 	*p;
register struct shm_pt 	*sptp;
register struct shmid_ds *sp;
{

	register struct shm_pt	*qsp;

	if (sp == NULL || sp->shm_proc == NULL)
		return(0);
	if (sp->shm_proc == p)
		return(EMFILE);

	for (qsp = sp->shm_proc->p_shmpt + sp->shm_pidx; qsp->shm_link;
	     qsp = qsp->shm_link->p_shmpt + qsp->shm_idx)
		if (qsp->shm_link == p) 
			return(EMFILE);
	
	return(0);
}


/*
 * shmnewptes - reset shared memory pte pointers
 */

shmnewptes(p)
struct proc 	*p;
{

	register struct shm_pt 		*sptp;
	register struct shmid_ds 	*sp;
	register 			i, s;

	p->p_smend = p->p_smbeg = NULL;
	for (sptp = p->p_shmpt, i = 0; i < shminfo.shmseg; i++, sptp++)
		if (sp = sptp->shm_shmp) {
			s = spl8();
			sptp->shm_spte = vtopte(p, sptp->shm_sva);
			sptp->shm_epte = sptp->shm_spte +
			    btoc(sp->shm_segsz) - 1;
			if (p->p_smbeg) {
				if (p->p_smbeg > sptp->shm_spte)
					p->p_smbeg = sptp->shm_spte;
				if (p->p_smend < sptp->shm_epte)
					p->p_smend = sptp->shm_epte;
			}
			else {
				p->p_smbeg = sptp->shm_spte;
				p->p_smend = sptp->shm_epte;
			}
			splx(s);
		}
}

/*
 * determine if any shared memory segments are locked 
 */
shmlocked(p)
register struct proc *p;
{
	register struct shm_pt *sptp;

	for (sptp = p->p_shmpt; sptp < p->p_shmpt + SHMSEG; sptp++)
		if (sptp->shm_shmp && (sptp->shm_shmp->shm_perm.flag & IPC_LOCK))
			return(1);
	return(0);
}


/*
 * determine if any shared memory segments have pageouts in progress
 */
shmpoip(p)
register struct proc *p;
{
	register struct shm_pt *sptp;

	for (sptp = p->p_shmpt; sptp < p->p_shmpt + SHMSEG; sptp++)
		if (sptp->shm_shmp && sptp->shm_shmp->shm_poip)
			return(1);
	return(0);
}


/*
 * find a shm_pt pointer given proc and addr
 */
struct shm_pt *
vtoshmpt(p, v)
	register struct proc *p;
	register v;
{
	register struct shm_pt		*sptp;
	register struct shmid_ds	*sp;

	for (sptp = p->p_shmpt; sptp < &(p->p_shmpt[shminfo.shmseg]); sptp++)
		if ((sp = sptp->shm_shmp)!=NULL &&
		  v >= sptp->shm_sva && v < sptp->shm_sva + btoc(sp->shm_segsz))
			return(sptp);
	return(NULL);
}
