/*	uipc_pipe.c	6.1	83/07/29	*/

#include "../ml/reg.h"
#include "../sys/param.h"
#include "../sys/user.h"
#include "../sys/vnode.h"
#include "../sys/kernel.h"
#include "../sys/file.h"
#include "../sys/mbuf.h"
#include "../sys/protosw.h"
#include "../sys/socket.h"
#include "../sys/socketvar.h"
#include "../sys/unpcb.h"
#include "../sys/audit.h"
#include "../sys/vfs.h"
#include "../ufs/inode.h"
#include "../ufs/fs.h"

extern	struct fileops vnodefops;

/*
 * USG UNIX pipes have a few extra semantic twists that 4.2BSD socket pairs
 * don't have:
 *     A socket pair is unidirectional (it has a read socket and a write
 * socket), while a pipe can be bidirectional (you can write to a pipe and
 * you, or somebody else, can read that data from the same pipe).
 *
 *     A "stat" on a socket gives nothing right now, but is intended to give
 * protocol-dependent data about the socket.  A "stat" on a pipe gives
 * inode-style information about the pipe.
 *
 * This module implements pipes with a inode of type FIFO.  That inode may
 * exist as a real on-disk inode, or may exist as a "dummy" inode which
 * exists in the in-core inode table but not on disk.  That inode contains
 * pointers to two sockets: the read socket and the write socket.
 *
 * This whole thing is more than a bit ugly.  Chalk that off to the fact
 * that pipes were originally inodes, and buffered/read/written just like
 * disk files (well, sort of).  Now they are, in effect, a funny bridge
 * between sockets and inodes.
 */

ino_t	pipecount;		/* sequence number assigned to pipes */

#define	PIPSIZ	4096

/*
 * The pipe vfs defined below is a dummy vfs structure used only for
 * pipes. All vnodes need to point to a vfs structure. Routines like
 * vn_rdwr simply index off the vnode's v_vfsp field to find out if
 * the 'file system' is read only. In the dummy vfs structure below,
 * the VFS_RDONLY flag is not set.
 */
struct vfs pipevfs ;

/*
 * Create an unnamed FIFO.  We allocate a dummy inode slot, etc.
 */
pipe()
{
	struct dinode pipeinode;	/* for initializing the inode */
	register struct inode *ip;
	register struct file *rf, *wf;
	register struct socket *rso, *wso;
	register struct vnode *vp;
	int r, save_R1;

	big_bzero((caddr_t)&pipeinode, sizeof(pipeinode));
	pipeinode.di_mode = (IFIFO|0666) & ~(u.u_cmask & ~AUDMASK);
/* 	pipeinode.di_mode = (IFIFO|0666) & ~u.u_cmask;    */
	pipeinode.di_nlink = 0;		/* pipes have no directory entry */
	pipeinode.di_uid = u.u_uid;	/* created by me */
	pipeinode.di_gid = u.u_gid;
	pipeinode.di_size = 0;
	pipeinode.di_atime = time ;	/* created now */
	pipeinode.di_mtime = time ;
	pipeinode.di_ctime = time ;
	pipeinode.di_ic.ic_flags = 0;

	if ((ip = imake((struct vnode *)0, (struct fs *)NULL,
	    pipecount, &pipevfs, &pipeinode, NULL)) == NULL) {
		return;
	}
	ip->i_flag |= IINCORE;
	vp = ITOV( ip );

	u.u_error = socreate(AF_UNIX, &ip->i_rsocket, SOCK_STREAM, 0);
	if (u.u_error)
		goto free;
	rso = ip->i_rsocket;

	u.u_error = socreate(AF_UNIX, &ip->i_wsocket, SOCK_STREAM, 0);
	if (u.u_error)
		goto free1;
	wso = ip->i_wsocket;

	rf = falloc();
	if (rf == NULL)
		goto free2;
	r = u.u_r.r_val1;
	rf->f_flag = FREAD;
	rf->f_type = DTYPE_VNODE;
	rf->f_ops = &vnodefops;
	rf->f_data = (caddr_t)vp;
	VN_HOLD( vp ) ;

	wf = falloc();
	if (wf == NULL)
		goto free3;
	wf->f_flag = FWRITE;
	wf->f_type = DTYPE_VNODE;
	wf->f_ops = &vnodefops;
	wf->f_data = (caddr_t)vp;
	VN_HOLD( vp ) ;

	save_R1 = u.u_r.r_val1;
	u.u_ar0[R1] = u.u_r.r_val1;
	u.u_r.r_val1 = r;
	if (piconnect(wso, rso) == 0)
		goto free4;
	ip->i_frcnt = 1;
	ip->i_fwcnt = 1;
	iput(ip);
	pipecount++;
	return;

free4:				/* free both file descriptors, both sockets, and inode */
	crfree(wf->f_cred);
	wf->f_count = 0;
	u.u_ofile[save_R1] = 0;
	VN_RELE( vp ) ;
free3:				/* free read file descriptor, both sockets, and inode */
	crfree(rf->f_cred);
	rf->f_count = 0;
	u.u_ofile[r] = 0;
	VN_RELE( vp ) ;
free2:				/* free both sockets and inode */
	soclose(wso);
free1:				/* free read socket and inode */
	soclose(rso);
free:				/* free inode */
	iput(ip);
}

/*
 * Open a FIFO file.  On the first call, this creates two sockets and
 * "piconnect"s them together.  The semantics of pipes on USG UNIX are
 * sufficiently different from the semantics of sockets that this routine
 * has to kludge quite a bit.
 * Check read and write counts, delay as necessary.
 */
openp(ip, mode)
	register struct inode *ip;
	register mode;
{
	register struct socket *rso, *wso;
	int newrsocket = 0, newwsocket = 0;
	register int error = 0;

	/*
	 * If we're opening for writing, and nobody's opened it for
	 * reading yet, and we're not opening for reading, and we're
	 * opening in no-delay mode, return an error indicating nobody's
	 * ready to talk to us yet.
	 */
	if (mode&FWRITE) {
		if (mode&FNDELAY && ip->i_frcnt == 0 && !(mode&FREAD))
			return(ENXIO);
	}

	/*
	 * The open will succeed, so create the sockets that this open
	 * will use if they don't already exist.
	 */
	rso = ip->i_rsocket;
	wso = ip->i_wsocket;
	if (mode&FREAD) {
		if (rso == 0) {
			if (error = socreate(AF_UNIX, &ip->i_rsocket, SOCK_STREAM, 0))
				return(error);
			rso = ip->i_rsocket;
			if (ip->i_fwcnt == 0)
				rso->so_state |= SS_CANTRCVMORE;	/* since there's no writer yet */
			newrsocket++;
		}
		if (ip->i_frcnt++ == 0 && wso != 0) {
			wso->so_state &= ~SS_CANTSENDMORE;	/* now there's a reader */
			wakeup((caddr_t)&ip->i_frcnt);
		}
	}
	if (mode&FWRITE) {
		if (wso == 0) {
			if (error = socreate(AF_UNIX, &ip->i_wsocket, SOCK_STREAM, 0))
				goto free;
			wso = ip->i_wsocket;
			if (ip->i_frcnt == 0)
				wso->so_state |= SS_CANTSENDMORE;	/* since there's no reader yet */
			newwsocket++;
		}
		if (ip->i_fwcnt++ == 0 && rso != 0) {
			rso->so_state &= ~SS_CANTRCVMORE;	/* now there's a writer */
			wakeup((caddr_t)&ip->i_fwcnt);
		}
	}
	if ((newrsocket && wso != 0) || (newwsocket && rso != 0)) {
		if (piconnect(wso, rso) == 0)
			goto free2;
	}
	if (mode&FREAD) {
		while (ip->i_fwcnt == 0) {
			if (mode&FNDELAY || ip->i_size)
				return(0);
			slumber((caddr_t)&ip->i_fwcnt, INTERRUPTABLE, resboost);
		}
	}
	if (mode&FWRITE) {
		while (ip->i_frcnt == 0)
			slumber((caddr_t)&ip->i_frcnt, INTERRUPTABLE, resboost);
	}
	return(0);

free2:
	if (newwsocket) {
		soclose(wso);
	}
free:
	if (newrsocket) {
		soclose(rso);
	}
	return(error);
}

/*
 * Close a pipe.
 * Update counts and cleanup.
 */
closep(ip, mode)
	register struct inode *ip;
	register mode;
{
	register int error = 0, error2 = 0;

	if (mode&FREAD) {
		if (--ip->i_frcnt == 0) {
			error = soclose(ip->i_rsocket);
			ip->i_rsocket = 0;
		}
	}
	if (mode&FWRITE) {
		if (--ip->i_fwcnt == 0) {
			error2 = soclose(ip->i_wsocket);
			ip->i_wsocket = 0;
		}
	}
	if ((ip->i_frcnt == 0) && (ip->i_fwcnt == 0)) {
		ip->i_size = 0;
		imark( ip, IUPD|ICHG );
	}
	if (error == 0)
		u.u_error = error2;
	else
		u.u_error = error;
}

/*
 * Sneakily connect a pipe from wso to rso.
 * It was claimed that "this will get cleaned up when socketpair is added."
 * Well, socketpair *was* added, and it wasn't cleaned up.  Since USG UNIX
 * gives pipes some extra semantics that makes them no longer behave
 * like socket pairs, we won't bother cleaning this up.
 */
piconnect(wso, rso)
	struct socket *wso, *rso;
{

	/* when we reserve memory this routine may fail */
	sotounpcb(wso)->unp_conn = sotounpcb(rso);
	sotounpcb(rso)->unp_conn = sotounpcb(wso);
	wso->so_snd.sb_hiwat = PIPSIZ;
	wso->so_snd.sb_mbmax = 2*PIPSIZ;
	wso->so_state |= SS_ISCONNECTED|SS_CANTRCVMORE;
	rso->so_rcv.sb_hiwat = 0;
	rso->so_rcv.sb_mbmax = 0;
	rso->so_state |= SS_ISCONNECTED|SS_CANTSENDMORE;
	return (1);
}

pselect(ip, which)
	struct inode *ip;
	int which;
{
	register struct socket *so;
	register int s = splnet();

	switch (which) {

	case FREAD:
		if ((so = ip->i_rsocket) == NULL) {
			splx(s);
			return (0);
		}
		if (soreadable(so)) {
			splx(s);
			return (1);
		}
		sbselqueue(&so->so_rcv);
		break;

	case FWRITE:
		if ((so = ip->i_wsocket) == NULL) {
			splx(s);
			return (0);
		}
		if (sowriteable(so)) {
			splx(s);
			return (1);
		}
		sbselqueue(&so->so_snd);
		break;
	}
	splx(s);
	return (0);
}
