/*	vm_mem.c	6.1	83/07/29	*/

#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/cmap.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/text.h"
#include "../sys/vm.h"
#include "../sys/file.h"
#include "../sys/vfs.h"
#include "../sys/vnode.h"
#include "../sys/buf.h"
#include "../sys/trace.h"
#include "../sys/map.h"
#include "../sys/kernel.h"
#include "../sys/lock.h"
#include "../ufs/mount.h"

/*
 * Allocate memory, and always succeed
 * by jolting page-out daemon
 * so as to obtain page frames.
 * To be used in conjunction with vmemfree().
 */
vmemall(pte, size, p, type)
	register struct pte *pte;
	register int size;
	struct proc *p;
{
	register int m;

	if (size <= 0 || size > maxmem)
		panic("vmemall size");
	while (size > 0) {
		if (freemem < desfree)
			outofmem();
		while (freemem == 0)
			slumber((caddr_t)&freemem, 0, resboost);
		m = min(size, freemem);
		(void) memall(pte, m, p, type);
		size -= m;
		pte += m;
	}
	if (freemem < desfree)
		outofmem();
	/*
	 * Always succeeds, but return success for
	 * vgetu and vgetpt (e.g.) which call either
	 * memall or vmemall depending on context.
	 */
	return (1);
}

/*
 * Free valid and reclaimable page frames belonging to the
 * count pages starting at pte.  If a page is valid
 * or reclaimable and locked (but not a system page), then
 * we simply mark the page as c_gone and let the pageout
 * daemon free the page when it is through with it.
 * If a page is reclaimable, and already in the free list, then
 * we mark the page as c_gone, and (of course) don't free it.
 *
 * Determines the largest contiguous cluster of
 * valid pages and frees them in one call to memfree.
 */
vmemfree(pte, count)
	register struct pte *pte;
	register int count;
{
	register struct cmap *c;
	register struct pte *spte;
	register int size, pcnt;
#if CLSIZE > 1
	int j;
#endif

#if CLSIZE > 1
	if (count % CLSIZE)
		panic("vmemfree");
#endif

	for (size = 0, pcnt = 0; count > 0; pte += CLSIZE, count -= CLSIZE) {
		if (pte->pg_fod == 0 && pte->pg_pfnum) {
			c = &cmap[pgtocm(pte->pg_pfnum)];
			pcnt += CLSIZE;
			if (c->c_lock && c->c_type != CSYS) {
#if CLSIZE > 1
				for (j = 0; j < CLSIZE; j++) {
					*(int *)(pte+j) &= PG_PROT;
					if (c->c_type == CSHMEM)
						(pte + j)->pg_nc = 1;
				}
#else
				*(int *)(pte) &= PG_PROT;
				if (c->c_type == CSHMEM)
					(pte)->pg_nc = 1;
#endif
				c->c_gone = 1;
				goto free;
			}
			if (c->c_free) {
				pcnt -= CLSIZE;
#if CLSIZE > 1
				for (j = 0; j < CLSIZE; j++) {
					*(int *)(pte+j) &= PG_PROT;
					if (c->c_type == CSHMEM)
						(pte + j)->pg_nc = 1;
				}
#else
				*(int *)(pte) &= PG_PROT;
				if (c->c_type == CSHMEM)
					(pte)->pg_nc = 1;
#endif
				if (c->c_type == CTEXT)
					distpte(&text[c->c_ndx], (int)c->c_page, pte);
				else if (c->c_type == CSHMEM) {
					dishmpte(&shmem[c->c_ndx], c->c_page, pte);
				}
				c->c_gone = 1;
				goto free;
			}
			if (size == 0)
				spte = pte;
			size += CLSIZE;
			continue;
		}
		if (pte->pg_fod) {
#if CLSIZE > 1
			for (j = 0; j < CLSIZE; j++)
				*(int *)(pte+j) &= PG_PROT;
#else
			*(int *)(pte) &= PG_PROT;
#endif
		}
free:
		if (size) {
			memfree(spte, size, 1);
			size = 0;
		}
	}
	if (size)
		memfree(spte, size, 1);
	return (pcnt);
}

/*
 * Unlink a page frame from the free list -
 *
 * Performed if the page being reclaimed
 * is in the free list.
 */
munlink(c)
	register struct cmap *c;
{
	register int next, prev;

	next = c->c_next;
	prev = c->c_prev;
	cmap[prev].c_next = next;
	cmap[next].c_prev = prev;
	c->c_free = 0;
	if (freemem < minfree)
		outofmem();
	freemem -= CLSIZE;
}

/*
 * Allocate memory -
 *
 * The free list appears as a doubly linked list
 * in the core map with cmap[0] serving as a header.
 */
memall(pte, size, p, type)
	register struct pte *pte;
	int size;
	register struct proc *p;
{
	register struct cmap *c;
	register struct pte *rpte;
	register struct proc *rp;
	int i, j, next;
	register int curpos;
	unsigned pf;
	struct cmap *c1, *c2;
	int s;
	register struct shm_pt *shmpt;

#if CLSIZE > 1
	if (size % CLSIZE)
		panic("memall");
#endif

	s = splimp();
	if (size > freemem) {
		splx(s);
		return (0);
	}
	trace(TR_MALL, size, u.u_procp->p_pid);
	for (i = size; i > 0; i -= CLSIZE) {
		curpos = cmap[CMHEAD].c_next;
		c = &cmap[curpos];
		freemem -= CLSIZE;
		next = c->c_next;
		cmap[CMHEAD].c_next = next;
		cmap[next].c_prev = CMHEAD;
		if (c->c_free == 0)
			panic("dup mem alloc");
		if (cmtopg(curpos) > maxfree)
			panic("bad mem alloc");
		if (c->c_gone == 0 && c->c_type != CSYS) {
			if (c->c_type == CTEXT)
				rp = text[c->c_ndx].x_caddr;
			else if (c->c_type == CSHMEM)
				rp = shmem[c->c_ndx].shm_proc;
			else
				rp = &proc[c->c_ndx];
			while (rp->p_flag & SNOVM)
				rp = rp->p_xlink;
			switch (c->c_type) {

			case CTEXT:
				rpte = tptopte(rp, c->c_page);
				break;

			case CSHMEM:
				shmpt = &(rp->p_shmpt[shmem[c->c_ndx].shm_pidx]);
				rpte = vtopte(rp, shmpt->shm_sva + c->c_page);
				break;

			case CDATA:
				rpte = dptopte(rp, c->c_page);
				break;

			case CSTACK:
				rpte = sptopte(rp, c->c_page);
				break;
			}

			if (rpte->pg_v)
				panic ("memall valid pte");

			zapcl(rpte, pg_pfnum) = 0;
			if (c->c_type == CTEXT)
				distpte(&text[c->c_ndx], (int)c->c_page, rpte);
			else if (c->c_type == CSHMEM)
				dishmpte(&shmem[c->c_ndx], c->c_page, rpte);
		}
		switch (type) {

		case CSYS:
			c->c_ndx = p->p_ndx;
			c->c_noswap = 1;
			break;

		case CTEXT:
			c->c_page = vtotp(p, ptetov(p, pte));
			c->c_ndx = p->p_textp - &text[0];
			c->c_noswap = (p->p_lock & TXTLOCK) ? 1 : 0;
			break;

		case CSHMEM:
			shmpt = vtoshmpt(p, ptetov(p, pte));
			c->c_page = ptetov(p, pte) - shmpt->shm_sva;
			c->c_ndx = shmpt->shm_shmp - &shmem[0];
			c->c_noswap = (shmpt->shm_sflg & SHM_MLOCK) ? 1 : 0;
			break;

		case CDATA:
			c->c_page = vtodp(p, ptetov(p, pte));
			c->c_ndx = p->p_ndx;
			c->c_noswap =  (p->p_lock & DATLOCK) ? 1 : 0;
			break;

		case CSTACK:
			c->c_page = vtosp(p, ptetov(p, pte));
			c->c_ndx = p->p_ndx;
			c->c_noswap =  (p->p_lock & DATLOCK) ? 1 : 0;
			break;
		}
		if (c->c_blkno) {
			munhash(c->c_vp, c->c_blkno);
		}
		pf = cmtopg(curpos);
#if CLSIZE > 1
		for (j = 0; j < CLSIZE; j++)
			*(int *)pte++ = pf++;
#else
		*(int *)pte++ = pf;
#endif
		c->c_free = 0;
		c->c_gone = 0;
		if (c->c_intrans || c->c_want)
			panic("memall intrans|want");
		c->c_lock = 1;
		c->c_type = type;
	}
	splx(s);
	return (size);
}

/*
 * Free memory -
 *
 * The page frames being returned are inserted
 * to the head/tail of the free list depending
 * on whether there is any possible future use of them.
 *
 * If the freemem count had been zero,
 * the processes sleeping for memory
 * are awakened.
 */
memfree(pte, size, detach)
	register struct pte *pte;
	register int size;
{
	register int i;
#if CLSIZE > 1
	int j;
#endif
	register struct cmap *c;
	int s;
	
#if CLSIZE > 1
	if (size % CLSIZE)
		panic("memfree");
#endif

	if (freemem < CLSIZE * KLMAX)
		wakeup((caddr_t)&freemem);
	while (size > 0) {
		size -= CLSIZE;
		i = pte->pg_pfnum;
		if (i < firstfree || i > maxfree)
			panic("bad mem free");
		i = pgtocm(i);
		c = &cmap[i];
		if (c->c_free)
			panic("dup mem free");
		if (detach && c->c_type != CSYS) {
#if CLSIZE > 1
			for (j = 0; j < CLSIZE; j++) {
				*(int *)(pte+j) &= PG_PROT;
				if (c->c_type == CSHMEM)
					(pte + j)->pg_nc = 1;
			}
#else
			*(int *)(pte) &= PG_PROT;
			if (c->c_type == CSHMEM)
				(pte)->pg_nc = 1;
#endif
			c->c_gone = 1;
		}
		s = splimp();
		if (detach && c->c_blkno == 0) {
			register int next = cmap[CMHEAD].c_next;
			cmap[next].c_prev = i;
			c->c_prev = CMHEAD;
			c->c_next = next;
			cmap[CMHEAD].c_next = i;
		} else {
			register int prev = cmap[CMHEAD].c_prev;
			cmap[prev].c_next = i;
			c->c_next = CMHEAD;
			c->c_prev = prev;
			cmap[CMHEAD].c_prev = i;
		}
		if (c->c_type != CSYS) {
			pte->pg_v = 0;
			distcl(pte);
			/*
			 * Since the caller can be the pageout
			 * daemon (cleanup() in specific), make sure
			 * that the PTE of the real owner of this
			 * page as recorded in the cmap entry is
			 * invalidated.
			 */
			if (c->c_gone == 0) {
				register struct proc *pp;
				register struct pte *ppte;
				struct shm_pt *shmpt;

				if (c->c_type == CTEXT)
					pp = text[c->c_ndx].x_caddr;
				else if (c->c_type == CSHMEM)
					pp = shmem[c->c_ndx].shm_proc;
				else
					pp = &proc[c->c_ndx];

				while (pp->p_flag & SNOVM)
					pp = pp->p_xlink;
				
				switch (c->c_type) {

				case CTEXT:
					ppte = tptopte(pp, c->c_page);
					break;

				case CSHMEM:
					shmpt = &(pp->p_shmpt[shmem[c->c_ndx].shm_pidx]);
					ppte = vtopte(pp, shmpt->shm_sva +
						      c->c_page);
					break;

				case CDATA:
					ppte = dptopte(pp, c->c_page);
					break;

				case CSTACK:
					ppte = sptopte(pp, c->c_page);
					break;
				}

				uncache (ppte);
				if (ppte->pg_pfnum != pte->pg_pfnum)
					panic ("memfree pg_pfnum");
				ppte->pg_v = 0;
				distcl (ppte);

				if (c->c_type == CTEXT)
					distpte (&text[c->c_ndx], (int) c->c_page, ppte);
				else if (c->c_type == CSHMEM)
					dishmpte(&shmem[c->c_ndx], c->c_page, ppte);
			}
			
		}
		c->c_free = 1;
		freemem += CLSIZE;
		splx(s);
		pte += CLSIZE;
	}
}


/*
 * Enter clist block c on the hash chains.
 * It contains file system block bn from vnode vp.
 */
mhash(c, vp, bn)
	register struct cmap *c;
	struct vnode *vp;
	daddr_t bn;
{
	register int i = CMHASH(bn);

	c->c_hlink = cmhash[i];
	cmhash[i] = c - cmap;
	c->c_blkno = bn;
	c->c_vp = vp;
	VN_HOLD(vp);
}

/*
 * Pull the clist entry of <vp,bn> off the hash chains.
 * We have checked before calling (using mfind) that the
 * entry really needs to be unhashed, so panic if we can't
 * find it (can't happen).
 * Must be called at splimp.
 */
munhash(vp, bn)
	register struct vnode *vp;
	register daddr_t bn;
{
	register int i = CMHASH(bn);
	register struct cmap *c1, *c2;

	c1 = &cmap[cmhash[i]];
	if (c1 == ecmap)
		panic("munhash ecmap1");
	if (c1->c_blkno == bn && c1->c_vp == vp)
		cmhash[i] = c1->c_hlink;
	else {
		for (;;) {
			c2 = c1;
			c1 = &cmap[c2->c_hlink];
			if (c1 == ecmap)
				panic("munhash ecmap2");
			if (c1->c_blkno == bn && c1->c_vp == vp)
				break;
		}
		c2->c_hlink = c1->c_hlink;
	}
	VN_RELE(vp);
	c1->c_vp = (struct vnode *)0;
	c1->c_blkno = 0;
	c1->c_hlink = 0;
}

/*
 * Look for block bn of vnode vp in the free pool.
 * Currently it should not be possible to find it unless it is
 * c_free and c_gone, although this may later not be true.
 * (This is because active texts are locked against file system
 * writes by the system.)
 * Must be called at splimp.
 */
struct cmap *
mfind(vp, bn)
	register struct vnode *vp;
	register daddr_t bn;
{
	register struct cmap *c1 = &cmap[cmhash[CMHASH(bn)]];

	while (c1 != ecmap) {
		if (c1->c_blkno == bn && c1->c_vp == vp)
			return (c1);
		c1 = &cmap[c1->c_hlink];
	}
	return ((struct cmap *)0);
}

/*
 * Purge blocks from vnode vp from incore cache
 */
mpurge(vp)
	register struct vnode *vp;
{
	register struct cmap *c1, *c2;
	register int i;
	int j = 0;
	int si = splimp();

	for (i = 0; i < CMHSIZ; i++) {
more:
		c1 = &cmap[cmhash[i]];
		if (c1 == ecmap)
			continue;
		if (c1->c_vp == vp)
			cmhash[i] = c1->c_hlink;
		else {
			for (;;) {
				c2 = c1;
				c1 = &cmap[c1->c_hlink];
				if (c1 == ecmap)
					goto cont;
				if (c1->c_vp == vp)
					break;
			}
			c2->c_hlink = c1->c_hlink;
		}
		j++;
		c1->c_vp = (struct vnode *)0;
		c1->c_blkno = 0;
		c1->c_hlink = 0;
		goto more;
cont:
		;
	}
	splx(si);
	while(j--)
		VN_RELE(vp);
}

/*
 * Purge blocks from filesystem vfsp from incore cache
 * before umount().
 */
mpurgevfs(vfsp)
	register struct vfs *vfsp;
{
	register struct cmap *c1, *c2;
	register int i;
	int si = splimp();

more:
	for (i = 0; i < CMHSIZ; i++) {
		c1 = &cmap[cmhash[i]];
		if (c1 == ecmap)
			continue;
		if (c1->c_vp->v_vfsp == vfsp)
			cmhash[i] = c1->c_hlink;
		else {
			for (;;) {
				c2 = c1;
				c1 = &cmap[c1->c_hlink];
				if (c1 == ecmap)
					goto cont;
				if (c1->c_vp->v_vfsp == vfsp)
					break;
			}
			c2->c_hlink = c1->c_hlink;
		}
		splx(si);
		VN_RELE(c1->c_vp);
		si = splimp();
		c1->c_vp = (struct vnode *)0;
		c1->c_blkno = 0;
		c1->c_hlink = 0;
		goto more;
cont:
		;
	}
	splx(si);
}

/*
 * Initialize core map
 */
meminit(first, last)
	int first, last;
{
	register int i;
	register struct cmap *c;

	firstfree = clrnd(first);
	maxfree = clrnd(last - (CLSIZE - 1));
	freemem = maxfree - firstfree;
	ecmx = ecmap - cmap;
	if (ecmx < freemem / CLSIZE)
		freemem = ecmx * CLSIZE;
	for (i = 1; i <= freemem / CLSIZE; i++) {
		cmap[i-1].c_next = i;
		c = &cmap[i];
		c->c_prev = i-1;
		c->c_free = 1;
		c->c_gone = 1;
		c->c_type = CSYS;
		c->c_vp = (struct vnode *)0;
		c->c_blkno = 0;
	}
	cmap[freemem / CLSIZE].c_next = CMHEAD;
	for (i = 0; i < CMHSIZ; i++)
		cmhash[i] = ecmx;
	cmap[CMHEAD].c_prev = freemem / CLSIZE;
	cmap[CMHEAD].c_type = CSYS;
	avefree = freemem;
	hand = 0;
}

/* 
 * Lock a virtual segment.
 *
 * For each cluster of pages, if the cluster is not valid,
 * touch it to fault it in, otherwise just lock page frame.
 * Called from physio to ensure that the pages 
 * participating in raw i/o are valid and locked.
 */
vslock(base, count)
	register caddr_t base;
{
	register unsigned v;
	register int npf;
	register struct pte *pte;
	register struct cmap * c ;
	register struct proc *p = u.u_procp;

	/* HCX I/O drivers may arrive here on raw I/O, 'base' will be */
	/* a system address in this case */

	if ((((int)base >> 30) & 3) == 3) 	/* system addresses */
		return;
	v = btop(base);
	pte = vtopte(p, v);
	npf = btoc(count + ((int)base & CLOFSET));
	while (npf > 0) {
		if (pte->pg_v) {
		/* Fixes swapper/this code bug. AWB */
	                c = &cmap[pgtocm(pte->pg_pfnum)];
			while (c->c_lock) {
				c->c_want = 1;
				slumber((caddr_t)c, 0, resboost);
			}
			uncache(pte) ;
			if (pte->pg_v) /* If valid when wakeup then lock it. */
				c->c_lock = 1;
			else
				pagein(ctob(v), 1);	/* return it locked */
		} else
			pagein(ctob(v), 1);	/* return it locked */
		pte += CLSIZE;
		v += CLSIZE;
		npf -= CLSIZE;
	}
}

/* 
 * Unlock a virtual segment.
 */
vsunlock(base, count, rw)
	register caddr_t base;
{
	register struct pte *pte;
	register struct cmap *c;
	register int npf;
	register struct proc *p = u.u_procp;
	register unsigned v;

	/* HCX I/O drivers may arrive here on raw I/O, 'base' will be */
	/* a system address in this case */

	if ((((int)base >> 30) & 3) == 3) 	/* system addresses */
		return;
	v = btop(base);
	pte = vtopte(p, v);
	npf = btoc(count + ((int)base & CLOFSET));
	while (npf > 0) {
		c = &cmap[pgtocm(pte->pg_pfnum)];
		MUNLOCK(c);
		if (rw == B_READ)	/* Reading from device writes memory */
			pte->pg_m = 1;
		pte->pg_u = 1;		/* In any event, it was referenced */
		pte += CLSIZE;
		npf -= CLSIZE;
	}
}
