/*	vm_proc.c	6.1	83/07/29	*/

#include "../ml/pte.h"
#include "../ml/mtpr.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/map.h"
#include "../sys/cmap.h"
#include "../sys/text.h"
#include "../sys/vm.h"
#include "../sys/kernel.h"
#include "../fpe/fp_in_krnl.h"
#include "../sys/lock.h"


/*
 * Get virtual memory resources for a new process.
 * Called after page tables are allocated, but before they
 * are initialized, we initialize the memory management registers,
 * and then expand the page tables for the data and stack segments
 * creating zero fill pte's there.  Text pte's are set up elsewhere.
 */
vgetvm(ts, ds, ss)
	size_t ts, ds, ss;
{

	setp0lr(ts);
	setp1lr(0);
	setp2lr(P2PAGES - HIGHPAGES);
	u.u_procp->p_tsize = ts;
	u.u_tsize = ts;
	expand((int)ss, 1);
	expand((int)ds, 0);
}

/*
 * Release the virtual memory resources (memory
 * pages, and swap area) associated with the current process.
 * Caller must not be swappable.  Used at exit or execl.
 */
vrelvm()
{
	register struct proc *p = u.u_procp;
	
	/*
	 * Release memory; text first, then data and stack pages.
	 */
	xfree();
	p->p_rssize -= vmemfree(dptopte(p, 0), p->p_dsize);
	p->p_rssize -= vmemfree(sptopte(p, p->p_ssize - 1), p->p_ssize);
	if (p->p_rssize != 0)
		panic("vrelvm rss");
	/*
	 * Wait for all page outs to complete, then
	 * release swap space.
	 */
	p->p_swrss = 0;
	while (p->p_poip)
		slumber((caddr_t)&p->p_poip, 0, resboost);
	(void) vsexpand((size_t)0, &u.u_dmap, 1);
	(void) vsexpand((size_t)0, &u.u_smap, 1);
	p->p_tsize = 0;
	p->p_dsize = 0;
	p->p_ssize = 0;
	u.u_tsize = 0;
	u.u_dsize = 0;
	u.u_ssize = 0;
}

/*
 * Pass virtual memory resources from p to q.
 * P's u. area is up, q's is uq.  Used internally
 * when starting/ending a vfork().
 */
vpassvm(p, q, up, uq, umap)
	register struct proc *p, *q;
	register struct user *up, *uq;
	struct pte *umap;
{

	/*
	 * Pass fields related to vm sizes.
	 */
	uq->u_tsize = q->p_tsize = p->p_tsize; up->u_tsize = p->p_tsize = 0;
	uq->u_dsize = q->p_dsize = p->p_dsize; up->u_dsize = p->p_dsize = 0;
	uq->u_ssize = q->p_ssize = p->p_ssize; up->u_ssize = p->p_ssize = 0;

	/*
	 * Pass proc table paging statistics.
	 */
	q->p_swrss = p->p_swrss; p->p_swrss = 0;
	q->p_rssize = p->p_rssize; p->p_rssize = 0;
	q->p_poip = p->p_poip; p->p_poip = 0;

	/*
	 * Relink text segment.
	 */
	q->p_textp = p->p_textp;
	xrepl(p, q);
	p->p_textp = 0;

	/*
	 * Pass swap space maps.
	 */
	uq->u_dmap = up->u_dmap; up->u_dmap = zdmap;
	uq->u_smap = up->u_smap; up->u_smap = zdmap;

	/*
	 * Pass u. paging statistics.
	 */
	uq->u_outime = up->u_outime; up->u_outime = 0;
	uq->u_ru = up->u_ru;
	big_bzero((caddr_t)&up->u_ru, sizeof (struct rusage));
	uq->u_cru = up->u_cru;
	big_bzero((caddr_t)&up->u_cru, sizeof (struct rusage));

	/*
	 * And finally, pass the page tables themselves.
	 * On return we are running on the other set of
	 * page tables, but still with the same u. area.
	 */
	vpasspt(p, q, up, uq, umap);

	/*
	 * Pass shared memory segments.
	 */

	shmrepl(p, q);
	
}

/*
 * Change the size of the data+stack regions of the process.
 * If the size is shrinking, it's easy-- just release virtual memory.
 * If it's growing, initalize new page table entries as 
 * 'zero fill on demand' pages.  Region is one of: 0 (data segment),
 * 1 (stack segment), or 2 (shared memory segment)  - noemi
 */
expand(change, region)
	register int change;
	int region;
{
	register struct proc *p;
	register struct pte *base, *p0, *p2;
	struct pte proto;
	int p0lr, p2lr;
	struct pte *base0;
	size_t ods, oss;
	register int size;
	int ptsz;
	u_int v, virtaddr;

	p = u.u_procp;
	if (change == 0)
		return;
#if CLSIZE > 1
	if (change % CLSIZE)
		panic("expand");
#endif

#ifdef PGINPROF
	vmsizmon();
#endif

	/*
	 * Update the sizes to reflect the change.  Note that we may
	 * swap as a result of a ptexpand, but this will work, because
	 * the routines which swap out will get the current text and data
	 * sizes from the arguments they are passed, and when the process
	 * resumes the lengths in the proc structure are used to 
	 * build the new page tables.
	 */
	ods = u.u_dsize;
	oss = u.u_ssize;

	switch (region) {		/* changed for shared memory - nr */
		case 0:			/* data region */
			v = dptov(p, p->p_dsize);
			p->p_dsize += change;
			u.u_dsize += change;
			break;

		case 2:			/* shared memory */
			v = dptov(p, p->p_dsize);
			break;

		case 1:			/* stack */
		default:
			p->p_ssize += change;
			v = sptov(p, p->p_ssize-1);
			u.u_ssize += change;
			break;
	}

	/*
	 * If we are shrinking, give back the virtual memory.
	 */
	if (region != 2 && change < 0) {
		p0lr = u.u_pcb.pcb_p0lr;
		p2lr = u.u_pcb.pcb_p2lr;
		if (region == 0)
			base = u.u_pcb.pcb_p0br + p0lr + change;
		else
			base = u.u_pcb.pcb_p2br + p2lr;

		p->p_rssize -= vmemfree(base, -change);
	}

	if (region == 0 && p->p_smbeg)
		goto out;

	/*
	 * Compute the end of the text+data regions and the beginning
	 * of the stack region in the page tables,
	 * and expand the page tables if necessary.
	 */
	p0 = u.u_pcb.pcb_p0br + u.u_pcb.pcb_p0lr;
	p2 = u.u_pcb.pcb_p2br + u.u_pcb.pcb_p2lr;

	if (change > 0) {
		if (change > p2 - p0)
			ptexpand(clrnd(ctopt(change - (p2 - p0))),
					     ods, oss);
	 } else if ((ptsz = clrnd((-change + (p2 - p0)) / NPTEPG)) > 0)
		ptexpand(-ptsz, ods, oss);
out:
	p0lr = u.u_pcb.pcb_p0lr;
	p2lr = u.u_pcb.pcb_p2lr;

	/*
	 * Update the processor length registers and copies in the pcb.
	 */
	if (region == 0 || region == 2)	    /* changed for shared memory */
		setp0lr(p0lr + change);
	else
		setp2lr(p2lr - change);

	if (change > 0) {

		/*
		 * Compute the base of the allocated/freed region.
		 */
		if (region == 0 || region == 2)
			base = u.u_pcb.pcb_p0br + p0lr;
		else
			base = u.u_pcb.pcb_p2br + p2lr - change;

		/*
		 * Initialize zero fill on demand pte's.
		 */
		*(int *)&proto = PG_UW;
		proto.pg_fod = 1;
		((struct fpte *)&proto)->pg_fileno = PG_FZERO;
		cnt.v_nzfod += change;
		base0 = base;
		size = change;
		
		while (--change >= 0)
			*base++ = proto;
		/*
		** if data region locked, page in the expansion pages; this
		** will set the c_noswap bit in the corresponding cmap entries
		*/
		if ((p->p_lock & DATLOCK) && (region != 2)) {
			base = base0;
			change = size;
			if (region == 0)
				virtaddr = ctob(dptov(p, ptetodp(p, base)));
			else
				virtaddr = ctob(sptov(p, ptetosp(p, base)));
			while(--change >= 0) {
				while ( ! base->pg_v )
					pagein(virtaddr, 0); /* sleep */
				virtaddr += NBPG;
				base++;
			}
		}
	} else {
		size = -change;
		v -= size;
	}

	/*
	 * We changed mapping for the current process,
	 * so must update the hardware translation
	 */
	newptes(v, size);
}

/*
 * Create a duplicate copy of the current process
 * in process slot p, which has been partially initialized
 * by newproc().
 *
 * Could deadlock here if two large proc's get page tables
 * and then each gets part of his UPAGES if they then have
 * consumed all the available memory.  This can only happen when
 *	USRPTSIZE + UPAGES * NPROC > maxmem
 * which is impossible except on systems with tiny real memories,
 * when large procs stupidly fork() instead of vfork().
 */
procdup(p, isvfork)
	register struct proc *p;
{

	/*
	 * Allocate page tables for new process, waiting
	 * for memory to be free.
	 */
	while (vgetpt(p, vmemall) == 0) {
		kmapwnt++;
		slumber((caddr_t)kernelmap, 0, resboost);
	}
	/*
	 * Snapshot the current u. area pcb and get a u.
	 * for the new process, a copy of our u.
	 */
	resume(pcbb(u.u_procp));
	(void) vgetu(p, vmemall, Forkmap, &forkutl, &u);

	/*
	 * define new data cache key.
	 */
	ckey_cnt[forkutl.u_procp->p_ckey]++;
	forkutl.u_procp->p_dkey = getdatakey();

	/*
	 * Arrange for a non-local goto when the new process
	 * is started, to resume here, returning nonzero from setjmp.
	 */
	forkutl.u_pcb.pcb_sswap = (int *)&u.u_ssave;
	if (setjmp(&forkutl.u_ssave))
		/*
		 * Return 1 in child.
		 */
		return (1);

	/*
	 * If the new process is being created in vfork(), then
	 * exchange vm resources with it.  We want to end up with
	 * just a u. area and an empty p0 region, so initialize the
	 * prototypes in the other area before the exchange.
	 */
	if (isvfork) {
		forkutl.u_pcb.pcb_p0lr = u.u_pcb.pcb_p0lr;
		forkutl.u_pcb.pcb_p1lr = 0;
		forkutl.u_pcb.pcb_p2lr = P2PAGES - HIGHPAGES;
		vpassvm(u.u_procp, p, &u, &forkutl, Forkmap);
		/*
		 * Return 0 in parent.
		 */
		return (0);
	}
	/*
	 * A real fork; clear vm statistics of new process
	 * and link into the new text segment.
	 * Equivalent things happen during vfork() in vpassvm().
	 */
	big_bzero((caddr_t)&forkutl.u_ru, sizeof (struct rusage));
	big_bzero((caddr_t)&forkutl.u_cru, sizeof (struct rusage));
	forkutl.u_dmap = u.u_cdmap;
	forkutl.u_smap = u.u_csmap;
	forkutl.u_outime = 0;

	/*
	 * Attach to the text segment.
	 */
	if (p->p_textp) {
		p->p_textp->x_count++;
		xlink(p);
	}

	/*
	 * Duplicate data and stack space of current process
	 * in the new process.
	 */
	vmdup(p, dptopte(p, 0), dptov(p, 0), p->p_dsize, CDATA);
	vmdup(p, sptopte(p, p->p_ssize - 1), sptov(p, p->p_ssize - 1),
		p->p_ssize, CSTACK ) ;

	if (!isvfork)
		shmfork(p, u.u_procp);

	/*
	 * Return 0 in parent.
	 */
	return (0);
}

int dupflag = 0;

#define HAVE_FIRST_PTE  0x1
#define HAVE_SECOND_PTE 0x2
#define HAVE_THIRD_PTE  0x4
#define WAITING_FOR_PTE 0x80000000

vmdup(p, pte, v, count, type)
	register struct proc *p;
	register struct pte *pte;
	register unsigned v;
	register size_t count;
	int type;
{
	register struct proc *myp = u.u_procp;
	register struct pte *opte;
	register struct cmap *c;
	int using_flag;
	void (*copyproc)();
	extern void copyseg1();
	extern void copyseg2();
	extern void copyseg3();
#if CLSIZE > 1
	register int i;
#endif

	/*
	 * Get a pte to be used for mapping.
	 */
get_mapping_pte:
	if (! (dupflag & HAVE_FIRST_PTE) ) {
		using_flag = HAVE_FIRST_PTE;
		dupflag |= HAVE_FIRST_PTE;
		copyproc = copyseg1;
	} else if (! (dupflag & HAVE_SECOND_PTE) ) {
		using_flag = HAVE_SECOND_PTE;
		dupflag |= HAVE_SECOND_PTE;
		copyproc = copyseg2;
	} else if (! (dupflag & HAVE_THIRD_PTE) ) {
		using_flag = HAVE_THIRD_PTE;
		dupflag |= HAVE_THIRD_PTE;
		copyproc = copyseg3;
	} else {
		dupflag |= WAITING_FOR_PTE;
		slumber((caddr_t)&dupflag, 0, resboost);
		goto get_mapping_pte;
	}

	opte = vtopte(myp, v);
	while (count != 0) {
		count -= CLSIZE;
		if (opte->pg_fod) {
			v += CLSIZE;
#if CLSIZE > 1
			for (i = 0; i < CLSIZE; i++)
#endif
				*(int *)pte++ = *(int *)opte++;
			continue;
		}
		opte += CLSIZE;
		(void) vmemall(pte, CLSIZE, p, type);
		p->p_rssize += CLSIZE;
#if CLSIZE > 1
		for (i = 0; i < CLSIZE; i++) {
			(*copyproc)((caddr_t)ctob(v+i), (pte+i)->pg_pfnum);
			*(int *)(pte+i) |= (PG_V|PG_M) + PG_UW;
		}
#else
		(*copyproc)((caddr_t)ctob(v), (pte)->pg_pfnum);
		*(int *)(pte) |= (PG_V|PG_M) + PG_UW;
#endif
		v += CLSIZE;
		c = &cmap[pgtocm(pte->pg_pfnum)];
		MUNLOCK(c);
		pte += CLSIZE;
	}
	p->p_flag |= SPTECHG;
	dupflag &= ~using_flag;
	if (dupflag & WAITING_FOR_PTE) {
		dupflag &= ~WAITING_FOR_PTE;
		wakeup(&dupflag);
	}
}
