/*	vm_pt.c	6.1	83/07/29	*/

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/map.h"
#include "../sys/cmap.h"
#include "../sys/vm.h"
#include "../sys/buf.h"
#include "../sys/text.h"
#include "../sys/vfs.h"
#include "../sys/vnode.h"
#include "../sys/kernel.h"

#include "../ml/pte.h"

#include "../ml/mtpr.h"

#define unneeded
/*
 * Get page tables for process p.  Allocator
 * for memory is argument; process must be locked
 * from swapping if vmemall is used; if memall is
 * used, call will return w/o waiting for memory.
 * In any case an error return results if no user
 * page table space is available.
 */
vgetpt(p, pmemall)
	register struct proc *p;
	int (*pmemall)();
{
	register long a;
	register int i;

	if (p->p_szpt == 0)
		panic("vgetpt");
	/*
	 * Allocate space in the kernel map for this process.
	 * Then allocate page table pages, and initialize the
	 * process' p0br and addr pointer to be the kernel
	 * virtual addresses of the base of the page tables and
	 * the pte for the process pcb (at the base of the u.).
	 */
	a = rmalloc(kernelmap, (long)p->p_szpt);
	if (a == 0)
		return (0);
	if ((*pmemall)(&Usrptmap[a], p->p_szpt, p, CSYS) == 0) {
		rmfree(kernelmap, (long)p->p_szpt, a);
		return (0);
	}
	p->p_p0br = kmxtob(a);
	p->p_addr = uaddr(p);
	/*
	 * Now validate the system page table entries for the
	 * user page table pages, flushing old translations
	 * for these kernel virtual addresses.  Clear the new
	 * page table pages for clean post-mortems.
	 */
	vmaccess(&Usrptmap[a], (caddr_t)p->p_p0br, p->p_szpt);
	for (i = 0; i < p->p_szpt; i++)
	{
		clearseg(Usrptmap[a + i].pg_pfnum);
		mtpr (&((char *)usrpt)[(a+i) * NBPG], P1DC);
	}
	return (1);
}

/*
 * Initialize text portion of page table.
 */
vinitpt(p)
	register struct proc *p;
{
	register struct text *xp;
	register struct proc *q;
	register struct pte *pte;
	register int i;
	register int size;
	struct pte proto;
	int     a;

	xp = p->p_textp;
	if (xp == 0)
		return;
	size = xp->x_size;
	pte = tptopte(p, 0);
	/*
	 * If there is another instance of same text in core
	 * then just copy page tables from other process.
	 */
	if (q = xp->x_caddr) {
		bcopy((caddr_t)tptopte(q, 0), (caddr_t)pte,
		    (unsigned) (sizeof(struct pte) * size));
		goto done;
	}

	/*
	 * Initialize text page tables, zfod if we are loading
	 * the text now; unless the process is demand loaded,
	 * this will suffice as the text will henceforth either be
	 * read from a file or demand paged in.
	 */
	*(int *)&proto = PG_URKR;
	if (xp->x_flag & XLOAD) {
		proto.pg_fod = 1;
		((struct fpte *)&proto)->pg_fileno = PG_FZERO;
	}
	for (i = 0; i < size; i++)
		*pte++ = proto;
	if ((xp->x_flag & XPAGV) == 0)
		goto done;
	/*
	 * Text is demand loaded.  If process is not loaded (i.e. being
	 * swapped in) then retrieve page tables from swap area.  Otherwise
	 * this is the first time and we must initialize the page tables
	 * from the blocks in the file system.
	 */
	if (xp->x_flag & XLOAD)
		vinifod((struct fpte *)tptopte(p, 0), PG_FTEXT, xp->x_vptr,
		    (daddr_t)1, size);
	else
		swap(p, xp->x_ptdaddr, (caddr_t)tptopte(p, 0),
		    size * sizeof (struct pte), B_READ,
		    B_PAGET, swapdev_vp, 0);
done:
	/*
	 * In the case where we are overlaying ourself with new page
	 * table entries, old user-space translations should be flushed.
	 */

	if (p == u.u_procp) {
		unsigned int va = tptov(p, 0);
		newptes(va, size);
	} else
		p->p_flag |= SPTECHG;
}


/*
 * Update the page tables of all processes linked
 * to a particular text segment, by distributing
 * dpte to the the text page at virtual frame v.
 *
 * Note that invalidation in the translation buffer for
 * the current process is the responsibility of the caller.
 */
distpte(xp, tp, dpte)
	struct text *xp;
	register size_t tp;
	register struct pte *dpte;
{
	register struct proc *p;
	register struct pte *pte;
#if CLSIZE > 1
	register int i;
#endif

	for (p = xp->x_caddr; p; p = p->p_xlink) {
		pte = tptopte(p, tp);
		p->p_flag |= SPTECHG;
		if (pte != dpte)
#if CLSIZE > 1
			for (i = 0; i < CLSIZE; i++)
				pte[i] = dpte[i];
#else
			*pte = *dpte;
#endif
	}
}


/*
 * Initialize shared memory portion of page tables.
 *
 */

vinitshmpt(p, sptp, sp)
struct proc		*p;
register struct shm_pt	*sptp;
register struct shmid_ds *sp;
{

	register struct shm_pt		*qsptp;
	register struct pte		*pte;
	register 	size	=	btoc(sp->shm_segsz);
	register	i;

	struct proc	*q;
	struct pte	proto;
	int		a;

	if (sp == NULL)
		return;


/* If there is another instance of the same shared memory segment in core,
 * just copy its page tables.
 */

	if (q = sp->shm_proc) {
		qsptp = q->p_shmpt + sp->shm_pidx;
		bcopy((caddr_t)vtopte(q, qsptp->shm_sva),
		      (caddr_t)vtopte(p, sptp->shm_sva),
		      (unsigned)(sizeof(struct pte) * size));

		if (((*(int *)(vtopte(p, sptp->shm_sva))) & PG_PROT) !=
		    (sptp->shm_sflg & PG_PROT))
			for (pte = vtopte(p, sptp->shm_sva), i = 0;
			     i < size; i++, pte++)
			    *(int *)pte = ((*(int *)pte) & ~PG_PROT) |
				(sptp->shm_sflg & PG_PROT);
		for (pte = vtopte(p, sptp->shm_sva), i = 0; i < size; i++, pte++) {
		    shmanyum(sp, i, pte);
		    if (pte->pg_u || pte->pg_m)
			dishmpte(sp, i, pte);
		}
	}

/* If there is no other instance of the same shared memory segment in core,
 * but one has been swapped out, retrieve the page tables from the swap
 * space.
 */

	else if (sp->shm_perm.flag & SHM_SWAP) {

/*
 * Set ourselves non-swappable while we are reading from the swap device
 */
		p->p_flag |= SULOCK;
		swap(p, sp->shm_ptdaddr, (caddr_t)vtopte(p, sptp->shm_sva),
		     size * sizeof(struct pte), B_READ, B_PAGET, swapdev_vp, 0);
		p->p_flag &= ~SULOCK;

		if (((*(int *)(vtopte(p, sptp->shm_sva))) & PG_PROT) !=
		    (sptp->shm_sflg & PG_PROT))
			for (pte = vtopte(p, sptp->shm_sva), i = 0;
			     i < size; i++, pte++)
			    *(int *)pte = ((*(int *)pte) & ~PG_PROT) |
				(sptp->shm_sflg & PG_PROT);
	}

/* Otherwise, initialize the ptes to zero fill on demand.  Also, set the
 * non-cacheable bit and obtain the permissions from the shm_pt structure.
 */

	else {
		*(int *)&proto = sptp->shm_sflg & PG_PROT;
		proto.pg_fod = 1;
		((struct fpte *)&proto)->pg_fileno = PG_FZERO;

		for (pte = vtopte(p, sptp->shm_sva), i = 0; i < size; i++)
			*pte++ = proto;

		cnt.v_nzfod += size;
	}

/* In the case where we are overlaying ourself with new page table
 * entries, ole user-space translations should be flushed.
 */

	sp->shm_perm.flag &= ~SHM_SWAP;
	newptes(sptp->shm_sva, size);
}


/*
 * Update the page tables of all processes linked to a particular shared
 * memory segment.  Offset is the page offset from the first data page
 * for the shared memory segment.  Distribute dpte to the data page that
 * is offset pages from the start of the shared memory segment.
 */

dishmpte(sp, offset, dpte)
register struct shmid_ds	*sp;
register size_t			offset;
register struct pte		*dpte;
{

	register struct pte 	*pte;
	register struct proc 	*p;
	register struct shm_pt	*sptp;
#if CLSIZE > 1
	register		i;
#endif

	if (sp == NULL || (p = sp->shm_proc) == NULL)
		return;

	sptp = p->p_shmpt + sp->shm_pidx;

	do {
		pte = vtopte(p, sptp->shm_sva + offset);
		p->p_flag |= SPTECHG;

		if (pte != dpte)
#if CLSIZE > 1
			for (i = 0; i < CLSIZE; i++)
			    *(int *)(pte+i) = ((*(int *)(pte+i) & PG_PROT) |
				(*(int *)(dpte+i) & ~PG_PROT));
#else
			*(int *)(pte) = ((*(int *)(pte) & PG_PROT) |
			    (*(int *)(dpte) & ~PG_PROT));
#endif

	} while ((p = sptp->shm_link) && (sptp = p->p_shmpt + sptp->shm_idx));
}


/*
 * Release page tables of process p.
 */
vrelpt(p)
	register struct proc *p;
{
	register int a;

	if (p->p_szpt == 0)
		return;
	a = btokmx(p->p_p0br);
	(void) vmemfree(&Usrptmap[a], p->p_szpt);
	rmfree(kernelmap, (long)p->p_szpt, (long)a);
}

#define	Xu(a) { \
	register int t ; \
	t = up->u_pcb.a; up->u_pcb.a = uq ->u_pcb.a; uq->u_pcb.a = t; \
}
#define	Xup(a) { \
	register struct pte *tp; \
	tp = up->u_pcb.a; up->u_pcb.a = uq ->u_pcb.a; uq->u_pcb.a = tp; \
}
#define	Xp(a) { \
	register int t ; \
	t = p->a; p->a = q->a; q->a = t; \
}
#define	Xpp(a) { \
	register struct pte *tp; \
	tp = p->a; p->a = q->a; q->a = tp; \
}

/*
 * Pass the page tables of process p to process q.
 * Used during vfork().  P and q are not symmetric;
 * p is the giver and q the receiver; after calling vpasspt
 * p will be ``cleaned out''.  Thus before vfork() we call vpasspt
 * with the child as q and give it our resources; after vfork() we
 * call vpasspt with the child as p to steal our resources back.
 * We are cognizant of whether we are p or q because we have to
 * be careful to keep our u. area and restore the other u. area from
 * umap after we temporarily put our u. area in both p and q's page tables.
 */
vpasspt(p, q, up, uq, umap)
	register struct proc *p, *q;
	register struct user *up, *uq;
	register struct pte *umap;
{
	int s;
	register int i ;

	s = spl8();	/* conservative, and slightly paranoid */
	Xu(pcb_szpt); Xu(pcb_p0lr); Xu(pcb_p1lr); Xu(pcb_p2lr);
	Xup(pcb_p0br); Xup(pcb_p1br); Xup(pcb_p2br);

	/*
	 * The u. area is contained in the process' p2 region.
	 * Thus we map the current u. area into the process virtual space
	 * of both sets of page tables we will deal with so that it
	 * will stay with us as we rearrange memory management.
	 */
	if (up == &u) {
		for (i = 0; i < UPAGES; i++)
			q->p_addr[i] = p->p_addr[i];
	} else {
		for (i = 0; i < UPAGES; i++)
			p->p_addr[i] = q->p_addr[i];
	}
	mtpr(0, TBIA);
	/*
	 * Now have u. double mapped, and have flushed
	 * any stale translations to new u. area.
	 * Switch the page tables.
	 */
	Xpp(p_p0br); Xp(p_szpt); Xpp(p_addr);
	mtpr(u.u_pcb.pcb_p0br, P0BR);
	mtpr(u.u_pcb.pcb_p1br, P1BR);
	mtpr(u.u_pcb.pcb_p2br, P2BR);
	mtpr(u.u_pcb.pcb_p0lr, P0LR);
	mtpr(u.u_pcb.pcb_p1lr, P1LR);
	mtpr(u.u_pcb.pcb_p2lr, P2LR);
	/*
	 * Now running on the ``other'' set of page tables.
	 * Flush translation to insure that we get correct u.
	 * Resurrect the u. for the other process in the other
	 * (our old) set of page tables.  Thus the other u. has moved
	 * from its old (our current) set of page tables to our old
	 * (its current) set of page tables, while we have kept our
	 * u. by mapping it into the other page table and then keeping
	 * the other page table.
	 */
	/* mtpr(0, TBIA); Cliff can't see why this is needed since
	                  the code between here and the next TBIA
	                  doesn't reference the user's address space */
	if (up == &u) {
		for (i = 0; i < UPAGES; i++) {
			int pf = umap[i].pg_pfnum;
			register struct pte *pte = &q->p_addr[i];
			pte->pg_pfnum = pf;
		}
	} else {
		for (i = 0; i < UPAGES; i++) {
			int pf = umap[i].pg_pfnum;
			register struct pte *pte = &p->p_addr[i];
			pte->pg_pfnum = pf;
		}
	}
	mtpr(0, TBIA);
	splx(s);
}

/*
 * Compute number of pages to be allocated to the u. area
 * and data and stack area page tables, which are stored on the
 * disk immediately after the u. area.
 */
/*ARGSUSED*/
vusize(p, utl)
	register struct proc *p;
	struct user *utl;
{
	register int tsz = p->p_tsize / NPTEPG;

	/*
	 * We do not need page table space on the disk for page
	 * table pages wholly containing text.	This is well
	 * understood in the code in vmswap.c.
	 */
	return (clrnd(UPAGES +
	    clrnd(ctopt(p->p_tsize+p->p_dsize+p->p_ssize+UPAGES)) - tsz));
}

/*
 * Get u area for process p.  If a old u area is given,
 * then copy the new area from the old, else
 * swap in as specified in the proc structure.
 *
 * Since argument map/newu is potentially shared
 * when an old u. is provided we have to be careful not
 * to block after beginning to use them in this case.
 * (This is not true when called from swapin() with no old u.)
 */
vgetu(p, palloc, map, newu, oldu)
	register struct proc *p;
	int (*palloc)();
	register struct pte *map;
	register struct user *newu;
	register struct user *oldu;
{
	register int i;

	if ((*palloc)(p->p_addr, clrnd(UPAGES), p, CSYS) == 0)
		return (0);
	/*
	 * New u. pages are to be accessible in map/newu as well
	 * as in process p's virtual memory.
	 */
	for (i = 0; i < UPAGES; i++) {
		map[i] = p->p_addr[i];
		*(int *)(p->p_addr + i) |= PG_URKW | PG_V;
	}
	setredzone(p->p_addr, (caddr_t)0);
	vmaccess(map, (caddr_t)newu, UPAGES);
	/*
	 * New u.'s come from forking or inswap.
	 */
	if (oldu) {
		big_bcopy((caddr_t)oldu, (caddr_t)newu, UPAGES * NBPG);
		newu->u_procp = p;
		newu->u_madesyscalls = 0;
	} else {
		swap(p, p->p_swaddr, (caddr_t)&u, ctob(UPAGES),
		   B_READ, B_UAREA, swapdev_vp, 0);

		if (newu->u_tsize != p->p_tsize || newu->u_dsize != p->p_dsize ||
		    newu->u_ssize != p->p_ssize || newu->u_procp != p)
			panic("vgetu");
	}
	/*
	 * Initialize the pcb copies of the p0 and p2 region bases and
	 * software page table size from the information in the proc structure.
	 */
	newu->u_pcb.pcb_p0br = p->p_p0br;
	newu->u_pcb.pcb_p2br = initp2br(p->p_p0br + p->p_szpt * NPTEPG);
	newu->u_pcb.pcb_szpt = p->p_szpt;
	return (1);
}

/*
 * Release swap space for a u. area.
 */
vrelswu(p, utl)
	struct proc *p;
	struct user *utl;
{

	rmfree(swapmap, (long)ctod(vusize(p, utl)), p->p_swaddr);
	/* p->p_swaddr = 0; */	/* leave for post-mortems */
}

/*
 * Get swap space for a u. area.
 */
vgetswu(p, utl)
	struct proc *p;
	struct user *utl;
{

	p->p_swaddr = rmalloc(swapmap, (long)ctod(vusize(p, utl)));
	return (p->p_swaddr);
}

/*
 * Release u. area, swapping it out if desired.
 *
 * Note: we run on the old u. after it is released into swtch(),
 * and are safe because nothing can happen at interrupt time.
 */
vrelu(p, swapu)
	register struct proc *p;
{
	register int i;
	struct pte uu[UPAGES];

	if (swapu)
		swap(p, p->p_swaddr, (caddr_t)&u, ctob(UPAGES),
		    B_WRITE, B_UAREA, swapdev_vp, 0);
	for (i = 0; i < UPAGES; i++)
		uu[i] = p->p_addr[i];
	(void) vmemfree(uu, clrnd(UPAGES));
}

#ifdef unneeded
int	ptforceswap;
#endif
/*
 * Expand/shrink a page table, assigning new kernel virtual
 * space and copying the page table entries over both
 * in the system map and as necessary in the user page table space.
 */
ptexpand(change, ods, oss)
	register int change;
	size_t ods, oss;
{
	register struct proc *p = u.u_procp;
	register struct pte *p1, *p2;
	register int szpt, i;
	int knew, kold;
	int p2size, p0pages, p2pages, s;

	kold = btokmx(p->p_p0br);
	szpt = p->p_szpt;

	if (change == 0 || szpt+change <= 0 || change % CLSIZE)
		panic("ptexpand");
	/*
	 * Change is the number of new page table pages needed if we're
	 * growing; otherwise, it is the number of page table pages
	 * to free.  Kold is the old index in the kernelmap of the page tables.
	 * Allocate a new kernel map segment of size szpt+change for
	 * the page tables.
	 */
top:

#ifdef unneeded
	if (ptforceswap)
		goto bad;
#endif
	if ((knew=rmalloc(kernelmap, (long)(szpt+change))) == 0)
		goto bad;

	/*
	 * Check if shrinking or expanding page tables.
	 */
	if (change < 0) {
		/*
		 * Determine how many pages of page table entries of
		 * text/data/shm and stack that will still be valid.
		 */
		if (p->p_smbeg)
			p0pages = clrnd(ctopt(p->p_smend - p->p_p0br + 1));
		else
			p0pages = clrnd(ctopt(p->p_tsize + p->p_dsize));
		p2size  = p->p_ssize + UPAGES;
		p2pages = clrnd(ctopt(p2size));

		/*
		 * Check if p0 and p2 ptes have to be merged.
		 */
		if (p0pages + p2pages > szpt + change) {
			p2pages--;
			/*
			 * Copy the modulo of stack and u. pte's to 
			 * the top of the last text/data/shm page.
			 *
			 */
			p1 = u.u_pcb.pcb_p2br + u.u_pcb.pcb_p2lr + 
							(oss - p->p_ssize);
			p2 = p->p_p0br + p0pages * NPTEPG - p2size % NPTEPG;
			for (i = p2size % NPTEPG; i != 0; i--)
				*p2++ = *p1++;
		}

		/*
		 * Copy these pages to the newly allocated page table.
		 */
		if (p2pages)
			kmcopy(knew + p0pages, kold + szpt - p2pages, p2pages);
		kmcopy(knew, kold, p0pages);

		/*
		 * Free the page tables no longer in use.
		 */
		s = spl8();
		(void)vmemfree(&Usrptmap[kold+p0pages], -change);
		splx(s);
		goto skip;
	} else {
		p2pages = (P2PAGES - u.u_pcb.pcb_p2lr)/NPTEPG;
		p0pages = szpt - p2pages;
		p2size  = oss + UPAGES;
		if (memall(&Usrptmap[knew+p0pages], change, p, CSYS)
									== 0) {
			rmfree(kernelmap, (long)(szpt+change), (long)knew);
			goto bad;
		}

		/*
		 * Spages pages of u.+stack page tables go over unchanged.
		 * Tdpages of text+data page table may contain a few stack
		 * pages which need to go in one of the newly allocated pages;
		 * this is a rough cut.
		 */
		kmcopy(knew, kold, p0pages);
		kmcopy(knew+p0pages+change, kold+p0pages, p2pages);

		/*
		 * Validate and clear the newly allocated page table pages in
		 * the center of the new region of the kernelmap.
		 */
		i = knew + p0pages;
		p1 = &Usrptmap[i];
		p2 = p1 + change;
		while (p1 < p2) {
			/* tptov BELOW WORKS ONLY FOR VAX */
			mapin(p1, tptov(p, i), p1->pg_pfnum, 1,
			    (int)(PG_V|PG_KW));
			clearseg(p1->pg_pfnum);
			mtpr(&((char *)usrpt)[i*NBPG], P1DC);
			p1++;
			i++;
		}
		mtpr(0, TBIA);

		/*
		 * Move the stack and u. pte's which are before the newly
		 * allocated pages into the last of the newly allocated pages.
		 * They are taken from the end of the current p2 region,
		 * and moved to the end of the new p2 region.
		 */
		p1 = u.u_pcb.pcb_p2br + u.u_pcb.pcb_p2lr;
		p2 = initp2br(kmxtob(knew+szpt+change)) + u.u_pcb.pcb_p2lr;
		for (i = p2size % NPTEPG; i != 0; i--)
			*p2++ = *p1++;
	}

	/*
	 * Now switch to the new page tables.
	 */
skip:
#ifdef notdef
	mtpr(0, TBIA);	/* paranoid */	/* unneded - Hagai */
#endif
	p->p_p0br = kmxtob(knew);
	setp0br(p->p_p0br);
	p->p_p1br = kmxtob(knew);
	setp1br(p->p_p1br);
	u.u_pcb.pcb_p2br = initp2br(kmxtob(knew+szpt+change));
	setp2br(u.u_pcb.pcb_p2br);
	u.u_pcb.pcb_szpt += change;
	p->p_szpt += change;
	p->p_addr = uaddr(p);
	mtpr(0, TBIA);
	mtpr(0, PADC);

	/*
	 * Finally, free old kernelmap.
	 */
	if (szpt)
		rmfree(kernelmap, (long)szpt, (long)kold);
	return;

bad:
	/*
	** if proc has some region locked just blow it out of the water
	*/
	if (p->p_flag & SPLOCK) {
		psignal(p, SIGSEGV);
		return;
	}
	
	/*
	 * Swap out the process so that the unavailable 
	 * resource will be allocated upon swapin.
	 *
	 * When resume is executed for the process, 
	 * here is where it will resume.
	 */
	resume(pcbb(p));
	if (setjmp(&u.u_ssave))
		return;
	if (swapout(p, ods, oss) == 0) {
		/*
		 * No space to swap... it is inconvenient to try
		 * to exit, so just wait a bit and hope something
		 * turns up.  Could deadlock here.
		 *
		 * Set ourselves non-swappable so the scheduler won't
		 * try to swap this process here.
		 *
		 * SOMEDAY REFLECT ERROR BACK THROUGH expand TO CALLERS
		 * (grow, sbreak) SO CAN'T DEADLOCK HERE.
		 */
		p->p_flag |= SLOCK;
		slumber((caddr_t)&lbolt, 0, delayboost);
		p->p_flag &= ~SLOCK;
		goto top;
	}
	/*
	 * Set SSWAP bit, so that when process is swapped back in
	 * swapin will set u.u_pcb.pcb_sswap to u_sswap and force a
	 * return from the setjmp() above.
	 */
	p->p_flag |= SSWAP;
	spl8();
	swtch();
	/* no return */
}

kmcopy(to, from, count)
	register int to;
	int from;
	register int count;
{
	register struct pte *tp = &Usrptmap[to];
	register struct pte *fp = &Usrptmap[from];

	while (count != 0) {
		mapin(tp, tptov(u.u_procp, to), fp->pg_pfnum, 1,
		    (int)(*((int *)fp) & (PG_V|PG_PROT)));
		tp++;
		fp++;
		to++;
		count--;
	}
}
