/*	@(#)vm_sw.c 1.1 86/09/25 SMI; from UCB 4.18 83/05/18	*/

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/buf.h"
#include "../sys/conf.h"
#include "../sys/user.h"
#include "../sys/map.h"
#include "../sys/uio.h"
#include "../sys/vm.h"
#include "../sys/file.h"
#include "../sys/reboot.h"
#include "../sys/vnode.h"
#include "../sys/dsk.h"

struct	buf rswbuf;
/*
 * Indirect driver for multi-controller paging.
 */
swstrategy(bp)
	register struct buf *bp;
{
	int sz, off, seg;
	dev_t dev;
    	extern int boothowto ;

	/*
	 * A mini-root gets copied into the front of the swap
	 * and we run over top of the swap area just long
	 * enough for us to do a mkfs and restor of the real
	 * root (sure beats rewriting standalone restor).
	 */
#define	MINIROOTSIZE	4096
    	if (boothowto & RB_GENERIC)
		if (rootdev == dumpdev)
			bp->b_blkno += MINIROOTSIZE;
	sz = howmany(bp->b_bcount, DEV_BSIZE);
	if (bp->b_blkno+sz > nswap) {
		bp->b_flags |= B_ERROR;
		iodone(bp);
		return;
	}
	if (nswdev > 1) {
		off = bp->b_blkno % dmmax;
		if (off+sz > dmmax) {
			bp->b_flags |= B_ERROR;
			iodone(bp);
			return;
		}
		seg = bp->b_blkno / dmmax;
		dev = swdevt[seg % nswdev].sw_dev;
		seg /= nswdev;
		bp->b_blkno = seg*dmmax + off;
	} else
		dev = swdevt[0].sw_dev;
	bp->b_dev = dev;
	if (dev == 0)
		panic("swstrategy");
	(*bdevsw[major(dev)].d_strategy)(bp);
}

swread(dev, uio)
	dev_t dev;
	struct uio *uio;
{

	return (physio(swstrategy, &rswbuf, dev, B_READ, minphys, uio));
}

swwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{

	return (physio(swstrategy, &rswbuf, dev, B_WRITE, minphys, uio));
}

/*
 * System call swapon(name) enables swapping on device name,
 * which must be in the swdevsw.  Return EBUSY
 * if already swapping on this device.
 */
swapon()
{
	register struct a {
		char	*name;
	} *uap;
	struct vnode *vp;
	dev_t dev;
	register struct swdevt *sp;

	if (!suser())
		return;
	uap = (struct a *)u.u_ap;
	u.u_error =
	    lookupname(uap->name, UIO_USERSPACE, FOLLOW_LINK,
		(struct vnode **)0, &vp);
	if (u.u_error)
		return;
	if (vp->v_type != VBLK) {
		u.u_error = ENOTBLK;
		VN_RELE(vp);
		return;
	}
	dev = (dev_t)vp->v_rdev;
	VN_RELE(vp);
	if (major(dev) >= nblkdev) {
		u.u_error = ENXIO;
		return;
	}
	/*
	 * Search starting at second table entry,
	 * since first (primary swap area) is freed at boot.
	 */
	for (sp = &swdevt[1]; sp->sw_dev; sp++)
		if (sp->sw_dev == dev) {
			if (sp->sw_freed) {
				u.u_error = EBUSY;
				return;
			}
			swfree(sp - swdevt);
			swmaxpgio() ;
			return;
		}
	u.u_error = ENODEV;
}
/*
 * swmaxpgio() recomputes the maxpgio rate as a function of the current
 * number of 'freed' swap devices. This allows more paging operations per
 * second if there are multiple swap areas.
 */
swmaxpgio()
{
	register struct swdevt *swp;
	register int nswdev = 0 ;
	extern int maxpgio, maxpgiobase ;

	for (swp = swdevt; swp->sw_dev; swp++) {
		if (swp->sw_freed) nswdev ++ ;
	}

	if (nswdev > 1) {
		maxpgio = (maxpgiobase * (2 * nswdev - 1)) / 2;
	} else {
	        maxpgio = maxpgiobase ;
	}
}

/*
 * Swfree(index) frees the index'th portion of the swap map.
 * Each of the nswdev devices provides 1/nswdev'th of the swap
 * space, which is laid out with blocks of dmmax pages circularly
 * among the devices.
 */
swfree(index)
	int index;
{
	register swblk_t vsbase;
	register long blk;
	dev_t dev;
	register swblk_t dvbase;
	register int nblks;

	dev = swdevt[index].sw_dev;
	(*bdevsw[major(dev)].d_open)(dev, FREAD|FWRITE);
	swdevt[index].sw_freed = 1;
	nblks = swdevt[index].sw_nblks;

	for (dvbase = 0; dvbase < nblks; dvbase += dmmax) {
		blk = nblks - dvbase;
		if ((vsbase = index*dmmax + dvbase*nswdev) >= nswap)
			panic("swfree");
		if (blk > dmmax)
			blk = dmmax;
		if (vsbase == 0) {
			/*
			 * Can't free a block starting at 0 in the swapmap
			 * but need some space for argmap so use 1/2 this
			 * hunk which needs special treatment anyways.
			 */
			argdev = swdevt[0].sw_dev;
			VN_RELE(argdev_vp);
			argdev_vp = bdevvp(argdev);
			rminit(argmap, (long)(blk/2-ctod(CLSIZE)),
			    (long)ctod(CLSIZE), "argmap", ARGMAPSIZE);
			/*
			 * First of all chunks... initialize the swapmap
			 * the second half of the hunk.
			 */
			rminit(swapmap, (long)(blk/2), (long)(blk/2),
			    "swap", nswapmap);
		} else {
			if ( dvbase==0 )
				/*
				** Avoid wiping out geometry block,
				** disk label, or whatever else may be
				** in the first block of partition.
				*/
				rmfree(swapmap, blk-1, vsbase+1);
			else
				rmfree(swapmap, blk, vsbase);
		}
	}
}
