/*	vm_swap.c	6.1	83/07/29	*/

#include "../ml/pte.h"

#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/proc.h"
#include "../sys/text.h"
#include "../sys/map.h"
#include "../sys/buf.h"
#include "../sys/cmap.h"
#include "../sys/vm.h"
#include "../sys/lock.h"
#include "../sys/kernel.h"
#include "../ml/mtpr.h"

/*
 * Swap a process in.
 */
swapin(p)
	register struct proc *p;
{
	register struct text *xp;
	register int i, s;

	register struct shm_pt *sptp;

	if (xp = p->p_textp) 
		xlock(xp);

	if (p->p_smbeg)
		for(sptp = p->p_shmpt, i = 0; i < SHMSEG; i++, sptp++)
			if(sptp->shm_shmp)
				ipclock(&sptp->shm_shmp->shm_perm);

	if (p->p_smbeg)
		p->p_szpt = clrnd(ctopt((int)p->p_smend + p->p_ssize + UPAGES));
	else
		p->p_szpt  = clrnd(ctopt(p->p_tsize + p->p_dsize +
					 p->p_ssize + UPAGES));

	if (vgetpt(p, memall) == 0)
		goto nomem;
	if (vgetu(p, memall, Swapmap, &swaputl, (struct user *)0) == 0) {
		vrelpt(p);
		goto nomem;
	}

	for (i = 0; i < UPAGES; i++)
		mtpr ((caddr_t)&u+i*NBPG, P1DC);

	/*
	 * Update pte pointers to new p0br base
	 */
	if (p->p_smbeg)
		shmnewptes(p);

	swdspt(p, &swaputl, B_READ);

	/*
	 * Make sure swdspt didn't smash u. pte's
	 */
	for (i = 0; i < UPAGES; i++) {
		if (Swapmap[i].pg_pfnum != p->p_addr[i].pg_pfnum)
			panic("swapin");
	}
	if (p->p_smend && p->p_smend > p->p_addr)
		panic("swapin p_addr");

	vrelswu(p, &swaputl);
	if (xp) {
		xlink(p);
		xunlock(xp);
	}

	/* relink shared memory segments and do ipcunlock()s */
	if (p->p_smbeg)
		for (i = 0; i < SHMSEG; i++)
			if (p->p_shmpt[i].shm_shmp) {
			        s = spl8();
				shmlink(p, &p->p_shmpt[i],
						p->p_shmpt[i].shm_shmp);
				splx(s);
				ipcunlock(&(p->p_shmpt[i].shm_shmp->shm_perm));
			}
	p->p_rssize = 0;
	s = spl8();
	if (p->p_stat == SRUN)
		setrq(p);
	p->p_flag |= SLOAD;
	if (p->p_flag & SSWAP) {
		swaputl.u_pcb.pcb_sswap = (int *)&u.u_ssave;
		p->p_flag &= ~SSWAP;
	}
	splx(s);
	p->p_etime = time.tv_sec;
	multprog++;
	cnt.v_swpin++;
	return (1);

nomem:
	if (xp)
		xunlock(xp);

	if (p->p_smbeg)
		for(i=0; i<SHMSEG; i++)
			if(p->p_shmpt[i].shm_shmp)
				ipcunlock(&(p->p_shmpt[i].shm_shmp->shm_perm));

	return (0);
}

int	xswapwant, xswaplock;
/*
 * Swap out process p.
 * ds and ss are the old data size and the stack size
 * of the process, and are supplied during page table
 * expansion swaps.
 */
swapout(p, ds, ss)
	register struct proc *p;
	size_t ds, ss;
{
	register struct pte *map;
	register struct user *utl;
	register int a;
	int s;
	int rc = 1;

	if (p->p_lock & (PROCLOCK | TXTLOCK | DATLOCK))
		panic("swapout PROCLOCK");

	s = 1;
	map = Xswapmap;
	utl = &xswaputl;
	if (xswaplock & s)
		if ((xswaplock & 2) == 0) {
			s = 2;
			map = Xswap2map;
			utl = &xswap2utl;
		}
	p->p_flag |= SLOCK;
	a = spl8();
	while (xswaplock & s) {
		xswapwant |= s;
		slumber((caddr_t)map, 0, resboost);
	}
	xswaplock |= s;
	splx(a);
	uaccess(p, map, utl);
	utl->u_odsize = ds;
	utl->u_ossize = ss;
	if (vgetswu(p, utl) == 0) {
		swkill(p, "swapout: no swap space");
		rc = 0;
		p->p_flag &= ~SLOCK;
		goto out;
	}
	for (a = 0; a < UPAGES; a++)
		mtpr ((caddr_t)&u+a*NBPG, P1DC);
	utl->u_ru.ru_nswap++;
	if (p->p_textp) {
		if (p->p_textp->x_ccount == 1)
			p->p_textp->x_swrss = p->p_textp->x_rssize;
		xccdec(p->p_textp, p);
	}

	if (p->p_smbeg) {
		register struct shm_pt *sptp;

		for (sptp = p->p_shmpt; sptp < p->p_shmpt + SHMSEG; sptp++) {
			if (sptp->shm_shmp == NULL)
				continue;
			a = spl8();
			shmccdec(p, sptp, sptp->shm_shmp);
			splx(a);
		}
	}

	p->p_swrss = p->p_rssize;
	vsswap(p, dptopte(p, 0), CDATA, 0, ds, &utl->u_dmap);
	vsswap(p, sptopte(p, CLSIZE-1), CSTACK, 0, ss, &utl->u_smap);
	if (p->p_rssize != 0)
		panic("swapout rssize");
	swdspt(p, utl, B_WRITE);
	(void) spl8();		/* hack memory interlock XXX */
	vrelu(p, 1);

	if ((p->p_flag & SLOAD) && (p->p_stat != SRUN || p != u.u_procp))
		panic("swapout");

	p->p_flag &= ~SLOAD;
	vrelpt(p);
	p->p_flag &= ~SLOCK;
	p->p_etime = time.tv_sec;
	if (p->p_smbeg)
		p->p_smend = (struct pte *)(p->p_smend - p->p_p0br + 1);

	multprog--;
	cnt.v_swpout++;

	if (runout) {
		runout = 0;
		wakeup((caddr_t)&runout);
	}
out:
	xswaplock &= ~s;
	if (xswapwant & s) {
		xswapwant &= ~s;
		wakeup((caddr_t)map);
	}
	return (rc);
}

/*
 * Swap the data and stack page tables in or out.
 * Only hard thing is swapping out when new pt size is different than old.
 * If we are growing new pt pages, then we must spread pages with 2 swaps.
 * If we are shrinking pt pages, then we must merge stack pte's into last
 * data page so as not to lose them (and also do two swaps).
 */
swdspt(p, utl, rdwri)
	register struct proc *p;
	register struct user *utl;
{
	register int szpt, dszpt, tsz, ssz;
	int tdlast, slast, tdsz;
	int dsize, ssize, merge, shmfit;
	register struct pte *pte;
	register int i;

	/*
	 * Initially, set all values to what we want to become.
	 */
	dsize = p->p_dsize;
	ssize = p->p_ssize;
	tsz   = p->p_tsize / NPTEPG;
	ssz   = clrnd(ctopt(p->p_ssize + UPAGES));
	szpt  = clrnd(ctopt(p->p_tsize + p->p_dsize + p->p_ssize + UPAGES));
	dszpt = szpt - tsz;
	merge = 0;

	if (p->p_smend)
		szpt = clrnd(ctopt(p->p_smend - p->p_p0br + 1 + 
				   p->p_ssize + UPAGES));

	if (szpt == p->p_szpt) {
		/*
	 	 * Page table staying the same size...
		 * The pt stays the same size if:
		 * 1. No region has changed.
		 * 2. A region has changed but the change 
		 *    has not caused a expansion/shrinking of the pt. 
		 * 3. We are swapping in.
		 *
		 * We can get by with one swap if:
	 	 * 1. We do not have shm.
	 	 * 2. We do have shm but the pt size doesn't change due to it.
	 	 *    This means that shm resides in the "would be" gap between
	 	 *    data and stack ptes.
	 	 */

		swptstat.pteasy++;
		shmfit =
		    (szpt == clrnd(ctopt(p->p_tsize + dsize + ssize + UPAGES)));

		/*
		 * If were expanding, then set the values of dsize and 
		 * ssize to the old values.
		 */
		if (p->p_dsize > utl->u_odsize ||
		    p->p_ssize > utl->u_ossize) {
			dsize = utl->u_odsize;
			ssize = utl->u_ossize;
			ssz   = clrnd(ctopt(utl->u_ossize + UPAGES));
		}

		if (shmfit) {
			if (rdwri == B_READ)
				swpt(rdwri, p, 0, tsz, (p->p_szpt - tsz) * 
				     NBPG - UPAGES * sizeof (struct pte));
			else
				swpt(rdwri, p, 0, tsz, (p->p_szpt - tsz) * 
				     NBPG);
			goto check;
		} else {
			/*
			 * We were changing and we have shm...
			 *
			 * The rhs of the equation is the size of the
			 * page table in core if a merge was done.
			 */
			if (dszpt == clrnd(ctopt(p->p_tsize + dsize + ssize +
					         UPAGES)) - tsz)
				merge++;
		}
	} else if (szpt > p->p_szpt) {
		/*
		 * Page table expanding... 
	 	 * Reset to old values because in the
		 * expanding case the old values reflect the "relevant"
		 * in core sizes.
		 */
		swptstat.ptexpand++; 
		dsize = utl->u_odsize;
		ssize = utl->u_ossize;
		ssz   = clrnd(ctopt(utl->u_ossize + UPAGES));
		/*
		 * Merging possible if we have shm and expanding
		 * due to shm.
		 */
		if (p->p_smend && dszpt == clrnd(ctopt(p->p_tsize + 
						 dsize + ssize + UPAGES)) - tsz)
			merge++;
	} else {
		/*
	 	 * Page table shrinking... 
	 	 */
		swptstat.ptshrink++;
		merge++;
	}

	/*
	 * See if last data page  and
	 * last stack page must be merged... if so, copy
	 * stack pte's from last stack page to end of last
	 * data page, and decrease size of stack 
	 * pt to be swapped.
	 * Merging is possible in all cases except when data or
	 * stack is growing.
	 * In the merging process we might destroy some shm
	 * ptes if shm is present but we don't care.
	 */
	if (rdwri == B_WRITE && merge) {
		swptstat.ptpack++;
		tdlast = (p->p_tsize + dsize) % (NPTEPG * CLSIZE);
		slast  = (ssize + UPAGES) % (NPTEPG * CLSIZE);
		if (tdlast && slast && tdlast + slast <= (NPTEPG * CLSIZE))
			tdsz = clrnd(ctopt(p->p_tsize + dsize));
		else
			tdsz = clrnd(ctopt(p->p_tsize + dsize)) + 1;
		bcopy((caddr_t)sptopte(p, ssize - 1),
		    (caddr_t)&p->p_p0br[tdsz * NPTEPG - slast],
		    (unsigned)(ssize * sizeof (struct pte)));
		swpt(rdwri, p, 0, tsz, dszpt * NBPG);
		goto check;
	}

	if (ssz) {
		if (rdwri == B_READ)
			swpt(rdwri, p, dszpt - ssz, p->p_szpt - ssz,
			     ssz * NBPG - UPAGES * sizeof (struct pte));
		else
			swpt(rdwri, p, dszpt - ssz, p->p_szpt - ssz,
			     ssz * NBPG);
	}
	if (dsize) {
		/*
		 * If ssz is 0 then the last data page also
		 * contains the stack+UPAGES ptes.
		 */
		if (rdwri == B_READ && ssz == 0)
			swpt(rdwri, p, 0, tsz,
			     (clrnd(ctopt(p->p_tsize + dsize)) - tsz) *
			     NBPG - UPAGES * sizeof (struct pte));
		else
			swpt(rdwri, p, 0, tsz,
			     (clrnd(ctopt(p->p_tsize + dsize)) - tsz) *
			     NBPG);
	}
check:
	/*
	 * We must take care here since in the shrinking case
	 * the number of "relevant" * ptes changes in swapin and swapout.
	 */
	for (i = 0; i < dsize; i++) {
		pte = dptopte(p, i);
		uncache (pte);
		if (pte->pg_v || pte->pg_fod == 0 && (pte->pg_pfnum||pte->pg_m))
			panic("swdspt data");
	}
	for (i = 0; i < ssize; i++) {
		pte = sptopte(p, i);
		uncache (pte);
		if (pte->pg_v || pte->pg_fod == 0 && (pte->pg_pfnum||pte->pg_m))
			panic("swdspt stack");
	}
}

swpt(rdwri, p, doff, a, n)
	int rdwri;
	struct proc *p;
	int doff, a, n;
{

	if (n <= 0)
		return;
	swap(p, p->p_swaddr + ctod(UPAGES) + ctod(doff),
	    (caddr_t)&p->p_p0br[a * NPTEPG], n, rdwri, B_PAGET, swapdev_vp, 0);
}
