#ifndef lint
static  char sccsid[] = "@(#)spec_vnodeops.c 1.1 86/09/25 Copyr 1986 Sun Micro";
#endif

/*
 * Copyright (c) 1986 by Sun Microsystems, Inc.
 */


#include "../sys/param.h"
#include "../sys/systm.h"
#include "../sys/user.h"
#include "../sys/buf.h"
#include "../sys/kernel.h"
#include "../sys/vfs.h"
#include "../sys/vnode.h"
#include "../sys/uio.h"
#include "../sys/conf.h"
#include "../sys/file.h"
#include "../sys/cmap.h"
#include "../sys/tty.h"
#include "../specfs/snode.h"

#include "../krpc/lockmgr.h"

int spec_open();
int spec_close();
int spec_rdwr();
int spec_ioctl();
int spec_select();
int spec_getattr();
int spec_setattr();
int spec_access();
int spec_link();
int spec_inactive();
int spec_badop();
int spec_noop();
int spec_lockctl();
int spec_fsync();
int spec_fid();

struct vnodeops spec_vnodeops = {
	spec_open,
	spec_close,
	spec_rdwr,
	spec_ioctl,
	spec_select,
	spec_getattr,
	spec_setattr,
	spec_access,
	spec_noop,	/* lookup */
	spec_noop,	/* create */
	spec_noop,	/* remove */
	spec_link,
	spec_noop,	/* rename */
	spec_noop,	/* mkdir */
	spec_noop,	/* rmdir */
	spec_noop,	/* readdir */
	spec_noop,	/* symlink */
	spec_noop,	/* readlink */
	spec_fsync,
	spec_inactive,
	spec_badop,	/* bmap */
	spec_badop,	/* badop */
	spec_badop,	/* bread */
	spec_badop,	/* brelse */
	spec_lockctl,
	spec_fid,
};

/*
 * open a special file (device)
 * some weird stuff here having to do with clone and indirect devices:
 * When a file open operation happens (e.g. ufs_open) and the vnode has
 * type VDEV the open routine makes a spec vnode and calls us. When we
 * do the device open routine there are two possible strange results:
 * 1) an indirect device will return an error on open and return a new
 *    dev number. we have to make that into a spec vnode and call open
 *    on it again.
 * 2) a clone device will return a new dev number on open but no error.
 *    in this case we just make a new spec vnode out of the new dev number
 *    and return that.
 */
/*ARGSUSED*/
int
spec_open(vpp, flag, cred)
	struct vnode **vpp;
	int flag;
	struct ucred *cred;
{
	register struct snode *sp;
	dev_t dev;
	dev_t newdev;
	dev_t olddev;
	int error;
	label_t lqsave;

	/*
	 * Setjmp in case open is interrupted.
	 * If it is, close and return error.
	 */
	lqsave = u.u_qsave;
	if (setjmp(&u.u_qsave)) {
		u.u_qsave = lqsave;
		error = EINTR;
		(void) spec_close(*vpp, flag & FMASK, cred);
		return (error);
	}
	sp = VTOS(*vpp);

	/*
	 * Do open protocol for special type.
	 */
	olddev = dev = sp->s_dev;

	switch ((*vpp)->v_type) {

	case VCHR:

		newdev = dev;
		error = 0;
		do {
			dev = newdev;
			if ((u_int)major(dev) >= nchrdev)
				return (ENXIO);

			error = (*cdevsw[major(dev)].d_open)(dev,flag, &newdev);

			/*
			 * If this is an indirect device we need to do the
			 * open again.
			 */
		} while (newdev != dev && error == EAGAIN);

		if (olddev != newdev && error == 0) {
			register struct vnode *nvp;

			/*
			 * Allocate new snode with new minor device. Release
			 * old snode. Set vpp to point to new one.  This snode
			 * will go away when the last reference to it goes away.
			 * Warning: if you stat this, and try to match it with
			 * a name in the filesystem you will fail, unless you
			 * had previously put names in that match.
			 */
			nvp = specvp(*vpp, newdev);
			VN_RELE(*vpp);
			*vpp = nvp;
		}
		break;

	case VBLK:
		if ((u_int)major(dev) >= nblkdev)
			return (ENXIO);
		error = (*bdevsw[major(dev)].d_open)(dev, flag);
		break;

	case VFIFO:
		printf("spec_open: got a VFIFO???\n");

	case VSOCK:
		error = EOPNOTSUPP;
		break;
	
	default:
		error = 0;
		break;
	}
	u.u_qsave = lqsave;
	return (error);
}

/*ARGSUSED*/
int
spec_close(vp, flag, cred)
	struct vnode *vp;
	int flag;
	struct ucred *cred;
{
	register struct snode *sp;
	int (*cfunc)();
	dev_t dev;
	label_t lqsave;

	/*
	 * setjmp in case close is interrupted
	 */
	lqsave = u.u_qsave;
	if (setjmp(&u.u_qsave)) {
		u.u_qsave = lqsave;
		return (EINTR);
	}
	sp = VTOS(vp);
	dev = sp->s_dev;
	switch(vp->v_type) {

	case VCHR:
		cfunc = cdevsw[major(dev)].d_close;
		break;

	case VBLK:
		cfunc = bdevsw[major(dev)].d_close;
		/*
		 * On last close of a block device (that isn't mounted)
		 * we must invalidate any in core blocks, so that
		 * we can, for instance, change floppy disks.
		 */
		bflush(vp);
		binval(vp);
		break;

	case VFIFO:
		printf("spec_close: got a VFIFO???\n");

	default:
		u.u_qsave = lqsave;
		return (0);
	}
	/*
	 * Close the device.
	 */
	(*cfunc)(dev, flag);
	u.u_qsave = lqsave;
	return (0);
}

/*
 * read or write a vnode
 */
/*ARGSUSED*/
int
spec_rdwr(vp, uiop, rw, ioflag, cred)
	struct vnode *vp;
	register struct uio *uiop;
	enum uio_rw rw;
	int ioflag;
	struct ucred *cred;
{
	register struct snode *sp;
	extern int mem_no;
	int error = 0;

	if ( rw == UIO_READ ) {
		if ( uiop->uio_resid == 0)
			return (0);
	} else if ( rw != UIO_WRITE) {
		panic("spec_rdwr");
	}

	sp = VTOS(vp);
	if ((uiop->uio_offset < 0 || (uiop->uio_offset + uiop->uio_resid) < 0)
	    && !(vp->v_type == VCHR && mem_no == major((dev_t)sp->s_dev))) {
		return (EINVAL);
	}

	if (vp->v_type == VCHR) {
		register int dev;
		register struct tty *tp;
		register int mj;

		dev = (int)(dev_t)sp->s_dev;
		mj = major(dev);
		if (rw == UIO_READ) {
			smark(sp, SACC);
			if (tp = cdevsw[mj].d_ttys) {
				tp += minor(dev);
				return((*linesw[tp->t_line].l_read)(tp, uiop));
			}
			else 
				return((*cdevsw[mj].d_read)(dev, uiop));
		} else {
			smark(sp, SUPD|SCHG);
			if (tp = cdevsw[mj].d_ttys) {
				tp += minor(dev);
				return((*linesw[tp->t_line].l_write)(tp, uiop));
			}
			else
				return((*cdevsw[mj].d_write)(dev, uiop));
		}
	} else if (vp->v_type == VBLK) {
		struct vnode *blkvp;
		register struct buf *bp;
		daddr_t lbn, bn;
		register int n, on;
		int i;
		extern struct cmap *mfind();

		u.u_error = 0;
		blkvp = sp->s_bdevvp;
		if (rw == UIO_READ) {
			smark(sp, SACC);
			do {
				lbn = uiop->uio_offset / BLKDEV_IOSIZE;
				on = uiop->uio_offset % BLKDEV_IOSIZE;
				n = MIN((unsigned)(BLKDEV_IOSIZE - on),
					uiop->uio_resid);
				bn = lbn * (BLKDEV_IOSIZE/DEV_BSIZE);
				rablock = bn + (BLKDEV_IOSIZE/DEV_BSIZE);
				if ((long)bn<0) {
					bp = geteblk(BLKDEV_IOSIZE);
					clrbuf(bp);
				} else if (sp->s_lastr + 1 == lbn)
					bp = breada(blkvp, bn, BLKDEV_IOSIZE,
						rablock, BLKDEV_IOSIZE );
				else
					bp = bread(blkvp, bn, BLKDEV_IOSIZE);
				sp->s_lastr = lbn;
				n = MIN(n, bp->b_bcount - bp->b_resid);
				if (bp->b_flags & B_ERROR) {
					error = EIO;
					brelse(bp);
					goto bad;
				}
				u.u_error = uiomove(bp->b_un.b_addr+on, n, rw, uiop);
				if (n + on == BLKDEV_IOSIZE)
					bp->b_flags |= B_AGE;
				brelse(bp);
			} while (u.u_error == 0 && uiop->uio_resid > 0 && n != 0);
		} else { /* if ( rw = UIO_WRITE ) */
			if (uiop->uio_resid == 0)
				return (0);
			do {
				lbn = uiop->uio_offset / BLKDEV_IOSIZE;
				on = uiop->uio_offset % BLKDEV_IOSIZE;
				n = MIN((unsigned)(BLKDEV_IOSIZE - on), uiop->uio_resid);
				bn = lbn * (BLKDEV_IOSIZE/DEV_BSIZE);
				rablock = bn + (BLKDEV_IOSIZE/DEV_BSIZE);

				for (i = 0;
				     i < howmany(BLKDEV_IOSIZE, DEV_BSIZE);
				     i += CLBYTES/DEV_BSIZE) {
					if (mfind(blkvp, (daddr_t)(bn + i)))
						munhash(blkvp, (daddr_t)(bn + i));
				}
				if (n == BLKDEV_IOSIZE)
					bp = getblk(blkvp, bn, BLKDEV_IOSIZE);
				else
					bp = bread(blkvp, bn, BLKDEV_IOSIZE);
				n = MIN(n, bp->b_bcount - bp->b_resid);
				if (bp->b_flags & B_ERROR) {
					error = EIO;
					brelse(bp);
					goto bad;
				}
				u.u_error = uiomove(bp->b_un.b_addr+on, n, rw, uiop);
				if (ioflag & IO_SYNC)
					bwrite(bp);
				else if (n + on == BLKDEV_IOSIZE) {
					bp->b_flags |= B_AGE;
					bawrite(bp);
				} else
					bdwrite(bp);
				smark(sp, SUPD|SCHG);
			} while (u.u_error == 0 && uiop->uio_resid > 0 && n != 0);
		}
	} else {
		return (EOPNOTSUPP);
	}
	if (error == 0)				/* XXX */
		error = u.u_error;		/* XXX */
bad:
	return (error);
}

/*ARGSUSED*/
int
spec_ioctl(vp, com, data, flag, cred)
	struct vnode *vp;
	int com;
	caddr_t data;
	int flag;
	struct ucred *cred;
{
	register struct snode *sp;

	sp = VTOS(vp);
	if (vp->v_type != VCHR)
		panic("spec_ioctl");
	return ((*cdevsw[major(sp->s_dev)].d_ioctl)
			(sp->s_dev, com, data, flag));
}

/*ARGSUSED*/
int
spec_select(vp, which, cred)
	struct vnode *vp;
	int which;
	struct ucred *cred;
{
	register struct snode *sp;

	sp = VTOS(vp);
	if (vp->v_type != VCHR)
		panic("spec_select");
	return ((*cdevsw[major(sp->s_dev)].d_select)(sp->s_dev, which));
}

/*ARGSUSED*/
int
spec_inactive(vp, cred)
	struct vnode *vp;
	struct ucred *cred;
{

	(void) spec_fsync(vp, cred);
	/* XXX
	 * spec_fsync does a xxx_setattr which may set u.u_error. Blech.
	 */
	u.u_error = 0;
	sunsave(VTOS(vp));
	kmem_free((caddr_t)VTOS(vp), (u_int)sizeof (struct snode));
	return (0);
}

int
spec_getattr(vp, vap, cred)
	struct vnode *vp;
	register struct vattr *vap;
	struct ucred *cred;
{
	int error;
	register struct snode *sp;

	sp = VTOS(vp);
	error = VOP_GETATTR(sp->s_realvp, vap, cred);
	if (!error) {
		/* set current times from snode, even if older than vnode */
		vap->va_atime = sp->s_atime;
		vap->va_mtime = sp->s_mtime;
		vap->va_ctime = sp->s_ctime;

		/* set device-dependent blocksizes */
		switch (vap->va_type) {
		case VBLK:
			vap->va_blocksize = BLKDEV_IOSIZE;
			break;

		case VCHR:
			vap->va_blocksize = MAXBSIZE;
			break;
		}
	}
	return (error);
}

int
spec_setattr(vp, vap, cred)
	struct vnode *vp;
	register struct vattr *vap;
	struct ucred *cred;
{
	struct snode *sp;
	int error;
	register int chtime = 0;

	sp = VTOS(vp);
	error = VOP_SETATTR(sp->s_realvp, vap, cred);
	if (!error) {
		/* if times were changed, update snode */
		if (vap->va_atime.tv_sec != -1) {
			sp->s_atime = vap->va_atime;
			chtime++;
		}
		if (vap->va_mtime.tv_sec != -1) {
			sp->s_mtime = vap->va_mtime;
			chtime++;
		}
		if (chtime)
			sp->s_ctime = time;
	}
	return (error);
}

int
spec_access(vp, mode, cred)
	struct vnode *vp;
	int mode;
	struct ucred *cred;
{

	return (VOP_ACCESS(VTOS(vp)->s_realvp, mode, cred));
}

spec_link(vp, tdvp, tnm, cred)
	struct vnode *vp;
	struct vnode *tdvp;
	char *tnm;
	struct ucred *cred;
{

	return (VOP_LINK(VTOS(vp)->s_realvp, tdvp, tnm, cred));
}

/*
 * In order to sync out the snode times without multi-client problems,
 * make sure the times written out are never earlier than the times
 * already set in the vnode.
 */
int
spec_fsync(vp, cred)
	struct vnode *vp;
	struct ucred *cred;
{
	register int error;
	register struct snode *sp;
	struct vattr *vap;
	struct vattr *vatmp;

	sp = VTOS(vp);
	/* if times didn't change, don't flush anything */
	if ((sp->s_flag & (SACC|SUPD|SCHG)) == 0)
		return (0);

	vatmp = (struct vattr *)kmem_alloc((u_int)sizeof (*vatmp));
	error = VOP_GETATTR(sp->s_realvp, vatmp, cred);
	if (!error) {
		vap = (struct vattr *)kmem_alloc((u_int)sizeof (*vap));
		vattr_null(vap);
		vap->va_atime = timercmp(&vatmp->va_atime, &sp->s_atime, >) ?
		    vatmp->va_atime : sp->s_atime;
		vap->va_mtime = timercmp(&vatmp->va_mtime, &sp->s_mtime, >) ?
		    vatmp->va_mtime : sp->s_mtime;
		VOP_SETATTR(sp->s_realvp, vap, cred);
		kmem_free((caddr_t)vap, (u_int)sizeof (*vap));
	}
	kmem_free((caddr_t)vatmp, (u_int)sizeof (*vatmp));
	(void) VOP_FSYNC(sp->s_realvp, cred);
	return (0);
}

int
spec_badop()
{
	panic("spec_badop");
}

int
spec_noop()
{

	return (EINVAL);
}

/*
 * Record-locking requests are passed back to the real vnode handler.
 */
int
spec_lockctl(vp, ld, cmd, cred)
	struct vnode *vp;
	struct flock *ld;
	int cmd;
	struct ucred *cred;
{
	return (VOP_LOCKCTL(VTOS(vp)->s_realvp, ld, cmd, cred));
}

spec_fid(vp, fidpp)
	struct vnode *vp;
	struct fid **fidpp;
{
	return (VOP_FID(VTOS(vp)->s_realvp, fidpp));
}
